/*
 *  Hamlib Rotator backend - GS-232 interface protocol
 *  Copyright (c) 2001-2010 by Stephane Fillod
 *
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef _ROT_GS232A_H
#define _ROT_GS232A_H 1

#include "rotator.h"

extern const struct rot_caps gs232a_rot_caps;
extern const struct rot_caps gs232a_az_rot_caps;
extern const struct rot_caps gs232a_el_rot_caps;
extern const struct rot_caps gs232_generic_rot_caps;
extern const struct rot_caps gs232b_rot_caps;
extern const struct rot_caps gs232b_az_rot_caps;
extern const struct rot_caps gs232b_el_rot_caps;
extern const struct rot_caps f1tetracker_rot_caps;
extern const struct rot_caps gs23_rot_caps;
extern const struct rot_caps gs23_az_rot_caps;
extern const struct rot_caps gs232_rot_caps;
extern const struct rot_caps amsat_lvb_rot_caps;
extern const struct rot_caps st2_rot_caps;
extern const struct rot_caps gs232_af6sa_wrc_caps;

#endif /* _ROT_GS232A_H */
