/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (https://www.swig.org).
 * Version 4.4.1
 *
 * Do not make changes to this file unless you know what you are doing - modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


#define SWIG_VERSION 0x040401
#define SWIGPYTHON
#define SWIG_PYTHON_DIRECTOR_NO_VTABLE

#define SWIG_name    "_Hamlib"
/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

#if defined(__cplusplus) && __cplusplus >=201103L
# define SWIG_NOEXCEPT noexcept
#else
# define SWIG_NOEXCEPT throw()
#endif 

/* -----------------------------------------------------------------------------
 * swigcompat.swg
 *
 * Macros to provide support compatibility with older C and C++ standards.
 *
 * Note that SWIG expects __cplusplus to be defined to the appropriate C++ standard.
 * MSVC users are urged to check and examine the /Zc:__cplusplus compiler option.
 * See https://learn.microsoft.com/en-us/cpp/build/reference/zc-cplusplus.
 * ----------------------------------------------------------------------------- */

/* C99 and C++11 should provide snprintf, but define SWIG_NO_SNPRINTF
 * if you're missing it.
 */
#if ((defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L) || \
     (defined __cplusplus && __cplusplus >= 201103L) || \
     defined SWIG_HAVE_SNPRINTF) && \
    !defined SWIG_NO_SNPRINTF
# define SWIG_snprintf(O,S,F,A) snprintf(O,S,F,A)
# define SWIG_snprintf2(O,S,F,A,B) snprintf(O,S,F,A,B)
#else
/* Fallback versions ignore the buffer size, but most of our uses either have a
 * fixed maximum possible size or dynamically allocate a buffer that's large
 * enough.
 */
# define SWIG_snprintf(O,S,F,A) sprintf(O,F,A)
# define SWIG_snprintf2(O,S,F,A,B) sprintf(O,F,A,B)
#endif


#if defined(__GNUC__) && defined(_WIN32) && !defined(SWIG_PYTHON_NO_HYPOT_WORKAROUND)
/* Workaround for '::hypot' has not been declared', see https://bugs.python.org/issue11566 */
# include <math.h>
#endif

#if !defined(PY_SSIZE_T_CLEAN) && !defined(SWIG_NO_PY_SSIZE_T_CLEAN)
#define PY_SSIZE_T_CLEAN
#endif

#if __GNUC__ >= 7
#pragma GCC diagnostic push
#if defined(__cplusplus) && __cplusplus >=201703L
#pragma GCC diagnostic ignored "-Wregister" /* For python-2.7 headers that use register */
#endif
#endif

#if defined(_DEBUG) && defined(SWIG_PYTHON_INTERPRETER_NO_DEBUG)
/* Use debug wrappers with the Python release dll */

#if defined(_MSC_VER) && _MSC_VER >= 1929
/* Workaround compilation errors when redefining _DEBUG in MSVC 2019 version 16.10 and later
 * See https://github.com/swig/swig/issues/2090 */
# include <corecrt.h>
#endif

# undef _DEBUG
# include <Python.h>
# define _DEBUG 1
#else
# include <Python.h>
#endif


#if __GNUC__ >= 7
#pragma GCC diagnostic pop
#endif

#include <stdio.h>
#include <stdlib.h>

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "5"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4
#define SWIG_POINTER_CLEAR         0x8
#define SWIG_POINTER_RELEASE       (SWIG_POINTER_CLEAR | SWIG_POINTER_DISOWN)

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows returning the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
/* Runtime errors are < 0 */
#define SWIG_ERROR                 (-1)
/* Errors in range -1 to -99 are in swigerrors.swg (errors for all languages including those not using the runtime) */
/* Errors in range -100 to -199 are language specific errors defined in *errors.swg */
/* Errors < -200 are generic runtime specific errors */
#define SWIG_ERROR_RELEASE_NOT_OWNED (-200)

#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporary objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del object mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next array of casts | pointer to cast hashed by value */
  unsigned int            value;		/* index of the last valid element in the array | typename hash value */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
 * Hash function for type name strings, based on maRushPrime1Hash (http://amsoftware.narod.ru/algo2.html)
 */
SWIGRUNTIME unsigned int SWIG_Hash(const char *str, unsigned int len) {
  const unsigned char *data = (const unsigned char *)str;
  unsigned int hash = len, i = 0, k;
  int rem = (int)len;

  while (rem >= (int)sizeof(unsigned int)) {
    memcpy(&k, data, sizeof(unsigned int));
    k += i++;
    hash ^= k;
    hash *= 171717;
    data += sizeof(unsigned int);
    rem -= sizeof(unsigned int);
  }

  switch (rem) {
    case 3: k = (unsigned int)(data[2]) << 16;
            k |= (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 2: k = (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 1: k = (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
  }
  return hash;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  static const unsigned int scan_threshold = 4;
  if (ty) {
    swig_cast_info *head = ty->cast;
    unsigned int hash_value = 0;
    int hashed = 0;

    while (head) {

      if (strcmp(head->type->name, c) == 0) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;
        int search = 1;

        if (!hashed) {
          if (head->value < scan_threshold) {
            for (iter = first; iter <= last; iter++) {
              if (strcmp(iter->type->name, c) == 0) {
                return iter;
              }
            }
            search = 0;
          } else {
            hashed = 1;
            hash_value = SWIG_Hash(c, (unsigned int)strlen(c));
          }
        }

        if (search) {
          /* Binary search over sorted <'next'|'value'> pairs */
          do {
            iter = first + ((last - first) >> 1);
            if (iter->value < hash_value) {
              first = iter + 1;
            } else if (iter->value == hash_value) {

              if (strcmp(iter->next->type->name, c) == 0) {
                return iter->next;
              }

              /* Hash collision check */
              for (last = iter + 1; last->next && last->value == hash_value; last++) {
                if (strcmp(last->next->type->name, c) == 0) {
                  return last->next;
                }
              }
              for (first = iter - 1; first != head && first->value == hash_value; first--) {
                if (strcmp(first->next->type->name, c) == 0) {
                  return first->next;
                }
              }
              break;
            } else
              last = iter - 1;
          } while (first <= last);
        }
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Check the type by type address
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(const swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *head = ty->cast;
    while (head) {
      if (head->type == from) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;

        /* Binary search over sorted array of casts */
        do {
          iter = first + ((last - first) >> 1);
          if (iter->type < from) {
            first = iter + 1;
          } else if (iter->type == from) {
            return iter;
          } else
            last = iter - 1;
        } while (first <= last);
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  Choose the last
     name. It should be the most specific; a fully resolved name
     but not necessarily with default template parameters expanded. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *head = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (head) {
    swig_cast_info *cast;
    for (cast = head; (cast - head) <= head->value; cast++) {
      if (!cast->converter) {
        swig_type_info *tc = cast->type;
        if (!tc->clientdata) {
          SWIG_TypeClientData(tc, clientdata);
        }
      }
    }
    head = head->next;
  }
}

SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/* SWIG Errors applicable to all language modules, values are reserved from -1 to -99 */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13


#if PY_VERSION_HEX >= 0x03030000 && !defined(SWIG_NO_HEAPTYPES)
#if !defined(SWIG_HEAPTYPES)
#define SWIG_HEAPTYPES
#endif
#endif

/* Compatibility macros for Python 3 */
#if PY_VERSION_HEX >= 0x03000000

#define PyClass_Check(obj) PyObject_IsInstance(obj, (PyObject *)&PyType_Type)
#define PyInt_Check(x) PyLong_Check(x)
#define PyInt_AsLong(x) PyLong_AsLong(x)
#define PyInt_FromLong(x) PyLong_FromLong(x)
#define PyInt_FromSize_t(x) PyLong_FromSize_t(x)
#define PyString_Check(name) PyBytes_Check(name)
#define PyString_FromString(x) PyUnicode_FromString(x)
#define PyString_Format(fmt, args)  PyUnicode_Format(fmt, args)
#define PyString_AsString(str) PyBytes_AsString(str)
#define PyString_Size(str) PyBytes_Size(str)	
#define PyString_InternFromString(key) PyUnicode_InternFromString(key)
#define Py_TPFLAGS_HAVE_CLASS Py_TPFLAGS_BASETYPE
#define _PyLong_FromSsize_t(x) PyLong_FromSsize_t(x)

#endif

/* SWIG APIs for compatibility of both Python 2 & 3 */

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_FromFormat PyUnicode_FromFormat
#else
#  define SWIG_Python_str_FromFormat PyString_FromFormat
#endif

#if defined(SWIG_HEAPTYPES)
#if PY_VERSION_HEX < 0x030c0000
#include <structmember.h>
#define Py_READONLY READONLY
#define Py_T_PYSSIZET T_PYSSIZET
#endif
#endif

#include <stddef.h> /* For offsetof */


/* Wrapper around PyUnicode_AsUTF8AndSize - call Py_XDECREF on the returned pbytes when finished with the returned string */
SWIGINTERN const char *
SWIG_PyUnicode_AsUTF8AndSize(PyObject *str, Py_ssize_t *psize, PyObject **pbytes)
{
#if PY_VERSION_HEX >= 0x03030000
# if !defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030A0000
  *pbytes = NULL;
  return PyUnicode_AsUTF8AndSize(str, psize);
# else
  const char *chars;
  *pbytes = PyUnicode_AsUTF8String(str);
  chars = *pbytes ? PyBytes_AsString(*pbytes) : NULL;
  if (chars && psize)
    *psize = PyBytes_Size(*pbytes);
  return chars;
# endif
#else
  char *chars = NULL;
  *pbytes = NULL;
  PyString_AsStringAndSize(str, &chars, psize);
  return chars;
#endif
}

SWIGINTERN PyObject *
SWIG_Python_str_FromChar(const char *c)
{
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_FromString(c); 
#else
  return PyString_FromString(c);
#endif
}

#define SWIG_RUNTIME_MODULE "swig_runtime_data" SWIG_RUNTIME_VERSION

/* SWIGPY_USE_CAPSULE is no longer used within SWIG itself, but some user interface files check for it. */
# define SWIGPY_USE_CAPSULE
#ifdef SWIGPYTHON_BUILTIN
# define SWIGPY_CAPSULE_ATTR_NAME "type_pointer_capsule_builtin" SWIG_TYPE_TABLE_NAME
#else
# define SWIGPY_CAPSULE_ATTR_NAME "type_pointer_capsule" SWIG_TYPE_TABLE_NAME
#endif
#define SWIGPY_CAPSULE_NAME SWIG_RUNTIME_MODULE "." SWIGPY_CAPSULE_ATTR_NAME

#if PY_VERSION_HEX < 0x03020000
#define PyDescr_TYPE(x) (((PyDescrObject *)(x))->d_type)
#define PyDescr_NAME(x) (((PyDescrObject *)(x))->d_name)
#define Py_hash_t long
#endif

#if defined(Py_LIMITED_API)
# define PyTuple_GET_ITEM PyTuple_GetItem
/* Note that PyTuple_SetItem() has different semantics from PyTuple_SET_ITEM as it decref's the original tuple item, so in general they cannot be used
  interchangeably. However in SWIG-generated code PyTuple_SET_ITEM is only used with newly initialized tuples without any items and for them this does work. */
# define PyTuple_SET_ITEM PyTuple_SetItem
# define PyTuple_GET_SIZE PyTuple_Size
# define PyCFunction_GET_FLAGS PyCFunction_GetFlags
# define PyCFunction_GET_FUNCTION PyCFunction_GetFunction
# define PyCFunction_GET_SELF PyCFunction_GetSelf
# define PyList_GET_ITEM PyList_GetItem
# define PyList_SET_ITEM PyList_SetItem
# define PySliceObject PyObject
#endif

/* Increment and Decrement wrappers - for portability when using the stable abi and for performance otherwise */
#ifdef Py_LIMITED_API
# define SWIG_Py_INCREF Py_IncRef
# define SWIG_Py_XINCREF Py_IncRef
# define SWIG_Py_DECREF Py_DecRef
# define SWIG_Py_XDECREF Py_DecRef
#else
# define SWIG_Py_INCREF Py_INCREF
# define SWIG_Py_XINCREF Py_XINCREF
# define SWIG_Py_DECREF Py_DECREF
# define SWIG_Py_XDECREF Py_XDECREF
#endif

#if PY_VERSION_HEX >= 0x03000000
#if (PY_VERSION_HEX >= 0x030d00a6) && (!defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030d00a6)
# define SWIG_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
SWIGINTERN PyObject *
SWIG_PyType_GetFullyQualifiedName(PyTypeObject *type) {
  PyObject *result = NULL;
  PyObject *qualname = PyObject_GetAttrString((PyObject *)type, "__qualname__");
  if (qualname) {
    PyObject *mod = PyObject_GetAttrString((PyObject *)type, "__module__");
    if (mod) {
      if (PyUnicode_Check(mod) && PyUnicode_CompareWithASCIIString(mod, "builtins") && PyUnicode_CompareWithASCIIString(mod, "__main__")) {
        result = PyUnicode_FromFormat("%U%c%U", mod, '.', qualname);
        SWIG_Py_DECREF(qualname);
      } else {
        result = qualname;
      }
      SWIG_Py_DECREF(mod);
    } else {
      result = qualname;
    }
  }

  return result;
}
#endif
#endif

/* gh-114329 added PyList_GetItemRef() to Python 3.13.0a4 */
#if (PY_VERSION_HEX >= 0x030d00a4) && (!defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030d00a4)
# define SWIG_PyList_GetItemRef PyList_GetItemRef
#else
SWIGINTERN PyObject *
SWIG_PyList_GetItemRef(PyObject *op, Py_ssize_t index) {
  PyObject *item = PyList_GetItem(op, index);
  Py_XINCREF(item);
  return item;
}
#endif

/* gh-106004 added PyDict_GetItemRef() and PyDict_GetItemStringRef() to Python 3.13.0a1
   functions are renamed here for compatibility with abi3audit */
#if (PY_VERSION_HEX >= 0x030d00a1) && (!defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030d00a1)
# define SWIG_PyDict_GetItemRef PyDict_GetItemRef
# define SWIG_PyDict_GetItemStringRef PyDict_GetItemStringRef
#else
SWIGINTERN int
SWIG_PyDict_GetItemRef(PyObject *mp, PyObject *key, PyObject **result) {
#if PY_VERSION_HEX >= 0x03000000
  PyObject *item = PyDict_GetItemWithError(mp, key);
#else
  PyObject *item = _PyDict_GetItemWithError(mp, key);
#endif
  if (item != NULL) {
    *result = (PyObject *)(item);
    SWIG_Py_INCREF(*result);
    return 1;
  }
  if (!PyErr_Occurred()) {
    *result = NULL;
    return 0;
  }
  *result = NULL;
  return -1;
}

SWIGINTERN int
SWIG_PyDict_GetItemStringRef(PyObject *mp, const char *key, PyObject **result) {
  int res;
#if PY_VERSION_HEX >= 0x03000000
  PyObject *key_obj = PyUnicode_FromString(key);
#else
  PyObject *key_obj = PyString_FromString(key);
#endif
  if (key_obj == NULL) {
    *result = NULL;
    return -1;
  }
  res = SWIG_PyDict_GetItemRef(mp, key_obj, result);
  Py_DECREF(key_obj);
  return res;
}
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIME PyObject*
SWIG_Python_ErrorType(int code) {
  PyObject* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = PyExc_MemoryError;
    break;
  case SWIG_IOError:
    type = PyExc_IOError;
    break;
  case SWIG_RuntimeError:
    type = PyExc_RuntimeError;
    break;
  case SWIG_IndexError:
    type = PyExc_IndexError;
    break;
  case SWIG_TypeError:
    type = PyExc_TypeError;
    break;
  case SWIG_DivisionByZero:
    type = PyExc_ZeroDivisionError;
    break;
  case SWIG_OverflowError:
    type = PyExc_OverflowError;
    break;
  case SWIG_SyntaxError:
    type = PyExc_SyntaxError;
    break;
  case SWIG_ValueError:
    type = PyExc_ValueError;
    break;
  case SWIG_SystemError:
    type = PyExc_SystemError;
    break;
  case SWIG_AttributeError:
    type = PyExc_AttributeError;
    break;
  case SWIG_NullReferenceError:
    type = PyExc_TypeError;
    break;
  default:
    type = PyExc_RuntimeError;
  }
  return type;
}


SWIGRUNTIME void
SWIG_Python_AddErrorMsg(const char* mesg)
{
  PyObject *type = 0;
  PyObject *value = 0;
  PyObject *traceback = 0;

  if (PyErr_Occurred())
    PyErr_Fetch(&type, &value, &traceback);
  if (value) {
    PyObject *old_str = PyObject_Str(value);
    PyObject *bytes = NULL;
    const char *tmp = SWIG_PyUnicode_AsUTF8AndSize(old_str, NULL, &bytes);
    PyErr_Clear();
    SWIG_Py_XINCREF(type);
    if (tmp)
      PyErr_Format(type, "%s %s", tmp, mesg);
    else
      PyErr_Format(type, "%s", mesg);
    SWIG_Py_XDECREF(bytes);
    SWIG_Py_DECREF(old_str);
    SWIG_Py_DECREF(value);
  } else {
    PyErr_SetString(PyExc_RuntimeError, mesg);
  }
}

SWIGRUNTIME int
SWIG_Python_TypeErrorOccurred(PyObject *obj)
{
  PyObject *error;
  if (obj)
    return 0;
  error = PyErr_Occurred();
  return error && PyErr_GivenExceptionMatches(error, PyExc_TypeError);
}

SWIGRUNTIME void
SWIG_Python_RaiseOrModifyTypeError(const char *message)
{
  if (SWIG_Python_TypeErrorOccurred(NULL)) {
    /* Use existing TypeError to preserve stacktrace and enhance with given message */
    PyObject *newvalue;
    PyObject *type = NULL, *value = NULL, *traceback = NULL;
    PyErr_Fetch(&type, &value, &traceback);
#if PY_VERSION_HEX >= 0x03000000
    newvalue = PyUnicode_FromFormat("%S\nAdditional information:\n%s", value, message);
#else
    newvalue = PyString_FromFormat("%s\nAdditional information:\n%s", PyString_AsString(value), message);
#endif
    if (newvalue) {
      SWIG_Py_XDECREF(value);
      PyErr_Restore(type, newvalue, traceback);
    } else {
      PyErr_Restore(type, value, traceback);
    }
  } else {
    /* Raise TypeError using given message */
    PyErr_SetString(PyExc_TypeError, message);
  }
}

#if defined(SWIG_PYTHON_NO_THREADS)
#  if defined(SWIG_PYTHON_THREADS)
#    undef SWIG_PYTHON_THREADS
#  endif
#endif
#if defined(SWIG_PYTHON_THREADS) /* Threading support is enabled */
#  if !defined(SWIG_PYTHON_USE_GIL) && !defined(SWIG_PYTHON_NO_USE_GIL)
#    define SWIG_PYTHON_USE_GIL
#  endif
#  if defined(SWIG_PYTHON_USE_GIL) /* Use PyGILState threads calls */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      if PY_VERSION_HEX < 0x03070000
#        define SWIG_PYTHON_INITIALIZE_THREADS PyEval_InitThreads()
#      else
#        define SWIG_PYTHON_INITIALIZE_THREADS
#      endif
#    endif
#    ifdef __cplusplus /* C++ code */
       class SWIG_Python_Thread_Block {
         bool status;
         PyGILState_STATE state;
       public:
         void end() { if (status) { PyGILState_Release(state); status = false;} }
         SWIG_Python_Thread_Block() : status(true), state(PyGILState_Ensure()) {}
         ~SWIG_Python_Thread_Block() { end(); }
       };
       class SWIG_Python_Thread_Allow {
         bool status;
         PyThreadState *save;
       public:
         void end() { if (status) { status = false; PyEval_RestoreThread(save); }}
         SWIG_Python_Thread_Allow() : status(true), save(PyEval_SaveThread()) {}
         ~SWIG_Python_Thread_Allow() { end(); }
       };
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   SWIG_Python_Thread_Block _swig_thread_block
#      define SWIG_PYTHON_THREAD_END_BLOCK     _swig_thread_block.end()
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   SWIG_Python_Thread_Allow _swig_thread_allow
#      define SWIG_PYTHON_THREAD_END_ALLOW     _swig_thread_allow.end()
#    else /* C code */
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   PyGILState_STATE _swig_thread_block = PyGILState_Ensure()
#      define SWIG_PYTHON_THREAD_END_BLOCK     PyGILState_Release(_swig_thread_block)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   PyThreadState *_swig_thread_allow = PyEval_SaveThread()
#      define SWIG_PYTHON_THREAD_END_ALLOW     PyEval_RestoreThread(_swig_thread_allow)
#    endif
#  else /* Old thread way, not implemented, user must provide it */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      define SWIG_PYTHON_INITIALIZE_THREADS
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_BLOCK)
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_BLOCK)
#      define SWIG_PYTHON_THREAD_END_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_ALLOW)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_ALLOW)
#      define SWIG_PYTHON_THREAD_END_ALLOW
#    endif
#  endif
#else /* No thread support */
#  define SWIG_PYTHON_INITIALIZE_THREADS
#  define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#  define SWIG_PYTHON_THREAD_END_BLOCK
#  define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#  define SWIG_PYTHON_THREAD_END_ALLOW
#endif

/* -----------------------------------------------------------------------------
 * Python API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_PY_POINTER 4
#define SWIG_PY_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
  int type;
  const char *name;
  long lvalue;
  double dvalue;
  void   *pvalue;
  swig_type_info **ptype;
} swig_const_info;

#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * pyrun.swg
 *
 * This file contains the runtime support for Python modules
 * and includes code for managing global variables and pointer
 * type checking.
 *
 * ----------------------------------------------------------------------------- */

#if PY_VERSION_HEX < 0x02070000 /* 2.7.0 */
# error "This version of SWIG only supports Python >= 2.7"
#endif

#if PY_VERSION_HEX >= 0x03000000 && PY_VERSION_HEX < 0x03050000
# error "This version of SWIG only supports Python 3 >= 3.5"
#endif

/* Common SWIG API */

/* for raw pointers */
#define SWIG_Python_ConvertPtr(obj, pptr, type, flags)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Python_ConvertPtr(obj, pptr, type, flags)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, own)

#ifdef SWIGPYTHON_BUILTIN
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(self, ptr, type, flags)
#else
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(NULL, ptr, type, flags)
#endif

#define SWIG_InternalNewPointerObj(ptr, type, flags)	SWIG_Python_NewPointerObj(NULL, ptr, type, flags)

#define SWIG_CheckImplicit(ty)                          SWIG_Python_CheckImplicit(ty) 
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Python_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Python_ConvertFunctionPtr(obj, pptr, type)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Python_NewPointerObj(NULL, ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Python_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Python_SetModule(pointer)
#define SWIG_NewClientData(obj)                         SwigPyClientData_New(obj)

#define SWIG_SetErrorObj                                SWIG_Python_SetErrorObj                            
#define SWIG_SetErrorMsg                        	SWIG_Python_SetErrorMsg				   
#define SWIG_ErrorType(code)                    	SWIG_Python_ErrorType(code)                        
#define SWIG_Error(code, msg)            		SWIG_Python_SetErrorMsg(SWIG_ErrorType(code), msg) 
#define SWIG_fail                        		goto fail					   

/* Runtime API implementation */

/* Error manipulation */

SWIGINTERN void 
SWIG_Python_SetErrorObj(PyObject *errtype, PyObject *obj) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK; 
  PyErr_SetObject(errtype, obj);
  SWIG_Py_DECREF(obj);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

SWIGINTERN void 
SWIG_Python_SetErrorMsg(PyObject *errtype, const char *msg) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK;
  PyErr_SetString(errtype, msg);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

#define SWIG_Python_Raise(obj, type, desc)  SWIG_Python_SetErrorObj(SWIG_Python_ExceptionType(desc), obj)

/* Set a constant value */

#if defined(SWIGPYTHON_BUILTIN)

SWIGINTERN void
SwigPyBuiltin_AddPublicSymbol(PyObject *seq, const char *key) {
  PyObject *s = PyString_InternFromString(key);
  PyList_Append(seq, s);
  SWIG_Py_DECREF(s);
}

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, PyObject *public_interface, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  SWIG_Py_DECREF(obj);
  if (public_interface)
    SwigPyBuiltin_AddPublicSymbol(public_interface, name);
}

#else

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  SWIG_Py_DECREF(obj);
}

#endif

/* SWIG runtime data Python module */
static PyObject *Swig_runtime_data_module_global = NULL;

/* Create/obtain the single swig_runtime_data module which is used across different SWIG generated modules */
SWIGINTERN PyObject *
SWIG_runtime_data_module(void) {
  if (!Swig_runtime_data_module_global) {
#if (PY_VERSION_HEX >= 0x030d0000) && (!defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030d0000)
    /* free-threading note: the GIL is always enabled when this function is first called
       by SWIG_init, so there's no risk of race conditions */
    Swig_runtime_data_module_global = PyImport_AddModuleRef(SWIG_RUNTIME_MODULE);
#elif PY_VERSION_HEX >= 0x03000000
    Swig_runtime_data_module_global = PyImport_AddModule(SWIG_RUNTIME_MODULE);
    SWIG_Py_XINCREF(Swig_runtime_data_module_global);
#else
    static PyMethodDef swig_empty_runtime_method_table[] = { {NULL, NULL, 0, NULL} }; /* Sentinel */
    Swig_runtime_data_module_global = Py_InitModule(SWIG_RUNTIME_MODULE, swig_empty_runtime_method_table);
    SWIG_Py_XINCREF(Swig_runtime_data_module_global);
#endif
  }
  assert(Swig_runtime_data_module_global);
  return Swig_runtime_data_module_global;
}

/* Append a value to the result obj */
SWIGINTERN PyObject*
SWIG_Python_AppendOutput(PyObject* result, PyObject* obj, int is_void) {
  if (!result) {
    result = obj;
  } else if (result == Py_None && is_void) {
    SWIG_Py_DECREF(result);
    result = obj;
  } else {
    if (!PyList_Check(result)) {
      PyObject *o2 = result;
      result = PyList_New(1);
      if (result) {
        PyList_SET_ITEM(result, 0, o2);
      } else {
        SWIG_Py_DECREF(obj);
        return o2;
      }
    }
    PyList_Append(result,obj);
    SWIG_Py_DECREF(obj);
  }
  return result;
}

/* Unpack the argument tuple */

SWIGINTERN Py_ssize_t
SWIG_Python_UnpackTuple(PyObject *args, const char *name, Py_ssize_t min, Py_ssize_t max, PyObject **objs)
{
  if (!args) {
    if (!min && !max) {
      return 1;
    } else {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got none", 
		   name, (min == max ? "" : "at least "), (int)min);
      return 0;
    }
  }  
  if (!PyTuple_Check(args)) {
    if (min <= 1 && max >= 1) {
      Py_ssize_t i;
      objs[0] = args;
      for (i = 1; i < max; ++i) {
	objs[i] = 0;
      }
      return 2;
    }
    PyErr_SetString(PyExc_SystemError, "UnpackTuple() argument list is not a tuple");
    return 0;
  } else {
    Py_ssize_t l = PyTuple_GET_SIZE(args);
    if (l < min) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at least "), (int)min, (int)l);
      return 0;
    } else if (l > max) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at most "), (int)max, (int)l);
      return 0;
    } else {
      Py_ssize_t i;
      for (i = 0; i < l; ++i) {
	objs[i] = PyTuple_GET_ITEM(args, i);
      }
      for (; l < max; ++l) {
	objs[l] = 0;
      }
      return i + 1;
    }    
  }
}

SWIGINTERN int
SWIG_Python_CheckNoKeywords(PyObject *kwargs, const char *name) {
  int no_kwargs = 1;
  if (kwargs) {
    assert(PyDict_Check(kwargs));
    if (PyDict_Size(kwargs) > 0) {
      PyErr_Format(PyExc_TypeError, "%s() does not take keyword arguments", name);
      no_kwargs = 0;
    }
  }
  return no_kwargs;
}

/* A functor is a function object with one single object argument */
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunctionObjArgs(functor, obj, NULL);

/*
  Helper for static pointer initialization for both C and C++ code, for example
  static PyObject *SWIG_STATIC_POINTER(MyVar) = NewSomething(...);
*/
#ifdef __cplusplus
#define SWIG_STATIC_POINTER(var)  var
#else
#define SWIG_STATIC_POINTER(var)  var = 0; if (!var) var
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* Python-specific SWIG API */
#define SWIG_newvarlink()                             SWIG_Python_newvarlink()
#define SWIG_addvarlink(p, name, get_attr, set_attr)  SWIG_Python_addvarlink(p, name, get_attr, set_attr)
#define SWIG_InstallConstants(d, constants)           SWIG_Python_InstallConstants(d, constants)
 
/* -----------------------------------------------------------------------------
 * global variable support code.
 * ----------------------------------------------------------------------------- */
 
typedef struct swig_globalvar {   
  char       *name;                  /* Name of global variable */
  PyObject *(*get_attr)(void);       /* Return the current value */
  int       (*set_attr)(PyObject *); /* Set the value */
  struct swig_globalvar *next;
} swig_globalvar;

typedef struct swig_varlinkobject {
  PyObject_HEAD
  swig_globalvar *vars;
} swig_varlinkobject;

SWIGINTERN PyObject *
SwigVarLink_repr(PyObject *SWIGUNUSEDPARM(v)) {
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_InternFromString("<Swig global variables>");
#else
  return PyString_FromString("<Swig global variables>");
#endif
}

SWIGINTERN PyObject *
SwigVarLink_str(PyObject *o) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
#if PY_VERSION_HEX >= 0x03000000
  PyObject *str = PyUnicode_InternFromString("(");
  PyObject *tail;
  PyObject *joined;
  swig_globalvar *var;
  for (var = v->vars; var; var=var->next) {
    tail = PyUnicode_FromString(var->name);
    joined = PyUnicode_Concat(str, tail);
    SWIG_Py_DECREF(str);
    SWIG_Py_DECREF(tail);
    str = joined;
    if (var->next) {
        tail = PyUnicode_InternFromString(", ");
        joined = PyUnicode_Concat(str, tail);
        SWIG_Py_DECREF(str);
        SWIG_Py_DECREF(tail);
        str = joined;
    }
  }
  tail = PyUnicode_InternFromString(")");
  joined = PyUnicode_Concat(str, tail);
  SWIG_Py_DECREF(str);
  SWIG_Py_DECREF(tail);
  str = joined;
#else
  PyObject *str = PyString_FromString("(");
  swig_globalvar *var;
  for (var = v->vars; var; var=var->next) {
    PyString_ConcatAndDel(&str,PyString_FromString(var->name));
    if (var->next) PyString_ConcatAndDel(&str,PyString_FromString(", "));
  }
  PyString_ConcatAndDel(&str,PyString_FromString(")"));
#endif
  return str;
}

SWIGINTERN void
SwigVarLink_dealloc(PyObject *o) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  swig_globalvar *var = v->vars;
  while (var) {
    swig_globalvar *n = var->next;
    free(var->name);
    free(var);
    var = n;
  }
}

SWIGINTERN PyObject *
SwigVarLink_getattr(PyObject *o, char *n) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  PyObject *res = NULL;
  swig_globalvar *var = v->vars;
  while (var) {
    if (strcmp(var->name,n) == 0) {
      res = (*var->get_attr)();
      break;
    }
    var = var->next;
  }
  if (res == NULL && !PyErr_Occurred()) {
    PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
  }
  return res;
}

SWIGINTERN int
SwigVarLink_setattr(PyObject *o, char *n, PyObject *p) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  int res = 1;
  swig_globalvar *var = v->vars;
  while (var) {
    if (strcmp(var->name,n) == 0) {
      res = (*var->set_attr)(p);
      break;
    }
    var = var->next;
  }
  if (res == 1 && !PyErr_Occurred()) {
    PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
  }
  return res;
}

SWIGINTERN PyTypeObject*
SwigVarLink_TypeOnce(void) {
  static char SwigVarLink_doc[] = "Swig variable link object";
#ifndef SWIG_HEAPTYPES
  static PyTypeObject varlink_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                  /* ob_size */
#endif
      SWIG_RUNTIME_MODULE ".SwigVarLink", /* tp_name */
      sizeof(swig_varlinkobject),         /* tp_basicsize */
      0,                                  /* tp_itemsize */
      (destructor) SwigVarLink_dealloc,   /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                       /* tp_print */
#else
      (Py_ssize_t)0,                      /* tp_vectorcall_offset */
#endif
      (getattrfunc) SwigVarLink_getattr,  /* tp_getattr */
      (setattrfunc) SwigVarLink_setattr,  /* tp_setattr */
      0,                                  /* tp_compare */
      (reprfunc) SwigVarLink_repr,        /* tp_repr */
      0,                                  /* tp_as_number */
      0,                                  /* tp_as_sequence */
      0,                                  /* tp_as_mapping */
      0,                                  /* tp_hash */
      0,                                  /* tp_call */
      (reprfunc) SwigVarLink_str,         /* tp_str */
      0,                                  /* tp_getattro */
      0,                                  /* tp_setattro */
      0,                                  /* tp_as_buffer */
      0,                                  /* tp_flags */
      SwigVarLink_doc,                    /* tp_doc */
      0,                                  /* tp_traverse */
      0,                                  /* tp_clear */
      0,                                  /* tp_richcompare */
      0,                                  /* tp_weaklistoffset */
      0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* tp_iter -> tp_weaklist */
      0,                                  /* tp_del */
      0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                  /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                  /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                  /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030c0000
      0,                                  /* tp_watched */
#endif
#if PY_VERSION_HEX >= 0x030d00a4
      0,                                  /* tp_versions_used */
#endif
#ifdef COUNT_ALLOCS
      0,                                  /* tp_allocs */
      0,                                  /* tp_frees */
      0,                                  /* tp_maxalloc */
      0,                                  /* tp_prev */
      0                                   /* tp_next */
#endif
    };
    PyObject *runtime_data_module = SWIG_runtime_data_module();
    varlink_type = tmp;
    type_init = 1;
    if (PyType_Ready(&varlink_type) < 0)
      return NULL;
    if (PyModule_AddObject(runtime_data_module, "SwigVarLink", (PyObject *)&varlink_type) == 0)
      SWIG_Py_INCREF((PyObject *)&varlink_type);
  }
  return &varlink_type;
#else
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigVarLink_dealloc },
    { Py_tp_repr, (void *)SwigVarLink_repr },
    { Py_tp_getattr, (void *)SwigVarLink_getattr },
    { Py_tp_setattr, (void *)SwigVarLink_setattr },
    { Py_tp_str, (void *)SwigVarLink_str },
    { Py_tp_doc, (void *)SwigVarLink_doc },
    { 0, NULL }
  };
  PyType_Spec spec = {
    SWIG_RUNTIME_MODULE ".SwigVarLink",
    sizeof(swig_varlinkobject),
    0,
    Py_TPFLAGS_DEFAULT,
    slots
  };
  PyObject *pytype = PyType_FromSpec(&spec);
  PyObject *runtime_data_module = SWIG_runtime_data_module();
  if (pytype && PyModule_AddObject(runtime_data_module, "SwigVarLink", pytype) == 0)
    SWIG_Py_INCREF(pytype);
  return (PyTypeObject *)pytype;
#endif
}

SWIGRUNTIME PyTypeObject*
SwigVarLink_Type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigVarLink_TypeOnce();
  return type;
}

/* Create a variable linking object for use later */
SWIGINTERN PyObject *
SWIG_Python_newvarlink(void) {
  swig_varlinkobject *result = PyObject_New(swig_varlinkobject, SwigVarLink_Type());
  if (result) {
    result->vars = 0;
  }
  return ((PyObject*) result);
}

SWIGINTERN void 
SWIG_Python_addvarlink(PyObject *p, const char *name, PyObject *(*get_attr)(void), int (*set_attr)(PyObject *p)) {
  swig_varlinkobject *v = (swig_varlinkobject *) p;
  swig_globalvar *gv = (swig_globalvar *) malloc(sizeof(swig_globalvar));
  if (gv) {
    size_t size = strlen(name)+1;
    gv->name = (char *)malloc(size);
    if (gv->name) {
      memcpy(gv->name, name, size);
      gv->get_attr = get_attr;
      gv->set_attr = set_attr;
      gv->next = v->vars;
    }
  }
  v->vars = gv;
}


static PyObject *Swig_Globals_global = NULL;
  
SWIGINTERN PyObject *
SWIG_globals(void) {
  if (Swig_Globals_global == NULL) {
    Swig_Globals_global = SWIG_newvarlink();
  }
  return Swig_Globals_global;
}

#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------
 * Pointer declarations
 * ----------------------------------------------------------------------------- */

/* Flags for new pointer objects */
#define SWIG_POINTER_NOSHADOW       (SWIG_POINTER_OWN      << 1)
#define SWIG_POINTER_NEW            (SWIG_POINTER_NOSHADOW | SWIG_POINTER_OWN)

#define SWIG_POINTER_IMPLICIT_CONV  (SWIG_POINTER_DISOWN   << 1)

#define SWIG_BUILTIN_TP_INIT	    (SWIG_POINTER_OWN << 2)
#define SWIG_BUILTIN_INIT	    (SWIG_BUILTIN_TP_INIT | SWIG_POINTER_OWN)

#ifdef __cplusplus
extern "C" {
#endif

/* The python void return value */

SWIGRUNTIMEINLINE PyObject * 
SWIG_Py_Void(void)
{
  PyObject *none = Py_None;
  SWIG_Py_INCREF(none);
  return none;
}

/* SwigPyClientData */

typedef struct {
  PyObject *klass;
  PyObject *newraw;
  PyObject *newargs;
  PyObject *destroy;
  int delargs;
  int implicitconv;
  PyTypeObject *pytype;
} SwigPyClientData;

SWIGRUNTIMEINLINE int 
SWIG_Python_CheckImplicit(swig_type_info *ty)
{
  SwigPyClientData *data = (SwigPyClientData *)ty->clientdata;
  int fail = data ? data->implicitconv : 0;
  if (fail)
    PyErr_SetString(PyExc_TypeError, "Implicit conversion is prohibited for explicit constructors.");
  return fail;
}

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_ExceptionType(swig_type_info *desc) {
  SwigPyClientData *data = desc ? (SwigPyClientData *) desc->clientdata : 0;
  PyObject *klass = data ? data->klass : 0;
  return (klass ? klass : PyExc_RuntimeError);
}


SWIGRUNTIME SwigPyClientData * 
SwigPyClientData_New(PyObject* obj)
{
  if (!obj) {
    return 0;
  } else {
    SwigPyClientData *data = (SwigPyClientData *)malloc(sizeof(SwigPyClientData));
    /* the klass element */
    data->klass = obj;
    SWIG_Py_INCREF(data->klass);
    /* the newraw method and newargs arguments used to create a new raw instance */
    if (PyClass_Check(obj)) {
      data->newraw = 0;
      SWIG_Py_INCREF(obj);
      data->newargs = obj;
    } else {
      data->newraw = PyObject_GetAttrString(data->klass, "__new__");
      if (data->newraw) {
        data->newargs = PyTuple_New(1);
        if (data->newargs) {
          SWIG_Py_INCREF(obj);
          PyTuple_SET_ITEM(data->newargs, 0, obj);
        } else {
          SWIG_Py_DECREF(data->newraw);
          SWIG_Py_DECREF(data->klass);
          free(data);
          return 0;
        }
      } else {
        SWIG_Py_INCREF(obj);
        data->newargs = obj;
      }
    }
    /* the destroy method, aka as the C++ delete method */
    data->destroy = PyObject_GetAttrString(data->klass, "__swig_destroy__");
    if (PyErr_Occurred()) {
      PyErr_Clear();
      data->destroy = 0;
    }
    if (data->destroy) {
      data->delargs = !(PyCFunction_GET_FLAGS(data->destroy) & METH_O);
    } else {
      data->delargs = 0;
    }
    data->implicitconv = 0;
    data->pytype = 0;
    return data;
  }
}

SWIGRUNTIME void 
SwigPyClientData_Del(SwigPyClientData *data)
{
  SWIG_Py_XDECREF(data->klass);
  SWIG_Py_XDECREF(data->newraw);
  SWIG_Py_XDECREF(data->newargs);
  SWIG_Py_XDECREF(data->destroy);
  free(data);
}

/* =============== SwigPyObject =====================*/

typedef struct {
  PyObject_HEAD
  void *ptr;
  swig_type_info *ty;
  int own;
  PyObject *next;
  PyObject *swigdict;
  PyObject *weakreflist;
} SwigPyObject;


#ifdef SWIGPYTHON_BUILTIN

SWIGRUNTIME PyObject *
SwigPyObject_get___dict__(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;

  if (!sobj->swigdict)
    sobj->swigdict = PyDict_New();

  SWIG_Py_XINCREF(sobj->swigdict);
  return sobj->swigdict;
}

#endif

SWIGRUNTIME PyObject *
SwigPyObject_long(SwigPyObject *v)
{
  return PyLong_FromVoidPtr(v->ptr);
}

SWIGRUNTIME PyObject *
SwigPyObject_format(const char* fmt, SwigPyObject *v)
{
  PyObject *res = NULL;
  PyObject *args = PyTuple_New(1);
  if (args) {
    PyObject *val = SwigPyObject_long(v);
    if (val) {
      PyObject *ofmt;
      PyTuple_SET_ITEM(args, 0, val);
      ofmt = SWIG_Python_str_FromChar(fmt);
      if (ofmt) {
#if PY_VERSION_HEX >= 0x03000000
        res = PyUnicode_Format(ofmt,args);
#else
        res = PyString_Format(ofmt,args);
#endif
        SWIG_Py_DECREF(ofmt);
      }
    }
    SWIG_Py_DECREF(args);
  }
  return res;
}

SWIGRUNTIME PyObject *
SwigPyObject_oct(SwigPyObject *v)
{
  return SwigPyObject_format("%o",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_hex(SwigPyObject *v)
{
  return SwigPyObject_format("%x",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_repr(SwigPyObject *v)
{
  const char *name = SWIG_TypePrettyName(v->ty);
  PyObject *repr = SWIG_Python_str_FromFormat("<Swig Object of type '%s' at %p>", (name ? name : "unknown"), (void *)v);
  if (repr && v->next) {
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next);
    if (nrep) {
# if PY_VERSION_HEX >= 0x03000000
      PyObject *joined = PyUnicode_Concat(repr, nrep);
      SWIG_Py_DECREF(repr);
      SWIG_Py_DECREF(nrep);
      repr = joined;
# else
      PyString_ConcatAndDel(&repr,nrep);
# endif
    } else {
      SWIG_Py_DECREF(repr);
      repr = NULL;
    }
  }
  return repr;
}

/* We need a version taking two PyObject* parameters so it's a valid
 * PyCFunction to use in SwigPyObject_methods[]. */
SWIGRUNTIME PyObject *
SwigPyObject_repr2(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  return SwigPyObject_repr((SwigPyObject*)v);
}

SWIGRUNTIME int
SwigPyObject_compare(PyObject *v, PyObject *w)
{
  /* tp_compare is only called when both objects have the same type, so
   * the casts are guaranteed to be ok. */
  void *i = ((SwigPyObject *)v)->ptr;
  void *j = ((SwigPyObject *)w)->ptr;
  return (i < j) ? -1 : ((i > j) ? 1 : 0);
}

SWIGRUNTIMEINLINE int SwigPyObject_Check(PyObject *);

/* Added for Python 3.x, would it also be useful for Python 2.x? */
SWIGRUNTIME PyObject*
SwigPyObject_richcompare(PyObject *v, PyObject *w, int op)
{
  PyObject* res = NULL;
  if (!PyErr_Occurred()) {
    /* Per https://docs.python.org/3/c-api/typeobj.html#c.PyTypeObject.tp_richcompare
     * the first argument is guaranteed to be an instance of SwigPyObject, but the
     * second is not, so we typecheck that one. */
    if ((op != Py_EQ && op != Py_NE) || !SwigPyObject_Check(w)) {
      SWIG_Py_INCREF(Py_NotImplemented);
      return Py_NotImplemented;
    }
    res = PyBool_FromLong( (SwigPyObject_compare(v, w)==0) == (op == Py_EQ) ? 1 : 0);
  }
  return res;
}


SWIGRUNTIME PyTypeObject* SwigPyObject_TypeOnce(void);

#ifdef SWIGPYTHON_BUILTIN
static swig_type_info *SwigPyObject_stype = 0;
SWIGRUNTIME PyTypeObject*
SwigPyObject_Type(void) {
    SwigPyClientData *cd;
    assert(SwigPyObject_stype);
    cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
    assert(cd);
    assert(cd->pytype);
    return cd->pytype;
}
#else
SWIGRUNTIME PyTypeObject*
SwigPyObject_Type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyObject_TypeOnce();
  return type;
}
#endif

SWIGRUNTIMEINLINE int
SwigPyObject_Check(PyObject *op) {
  PyTypeObject *target_tp = SwigPyObject_Type();
  PyTypeObject *op_type = Py_TYPE(op);
#ifdef SWIGPYTHON_BUILTIN
  /* Only builtin types have SwigPyObject as a base type */
  return PyType_IsSubtype(op_type, target_tp);
#else
  /* Check for an exact match to SwigPyObject */
  if (op_type == target_tp) {
    return 1;
  } else {
    /* Fallback for multiple modules */
#if PY_VERSION_HEX >= 0x03000000
    int cmp;
    PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(op_type);
    if (!tpname)
      return 0;
    cmp = PyUnicode_CompareWithASCIIString(tpname, SWIG_RUNTIME_MODULE ".SwigPyObject");
    SWIG_Py_DECREF(tpname);
    return cmp == 0;
#else
    return strcmp(op_type->tp_name, SWIG_RUNTIME_MODULE ".SwigPyObject") == 0;
#endif
  }
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own);

static PyObject* Swig_Capsule_global = NULL;

SWIGRUNTIME void
SwigPyObject_dealloc(PyObject *v)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  PyObject *next = sobj->next;
  if (sobj->own == SWIG_POINTER_OWN) {
    swig_type_info *ty = sobj->ty;
    SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
    PyObject *destroy = data ? data->destroy : 0;
    if (destroy) {
      /* destroy is always a VARARGS method */
      PyObject *res;

      /* PyObject_CallFunction() has the potential to silently drop
         the active exception.  In cases of unnamed temporary
         variable or where we just finished iterating over a generator
         StopIteration will be active right now, and this needs to
         remain true upon return from SwigPyObject_dealloc.  So save
         and restore. */
      
      PyObject *type = NULL, *value = NULL, *traceback = NULL;
      PyErr_Fetch(&type, &value, &traceback);

      if (data->delargs) {
        /* we need to create a temporary object to carry the destroy operation */
        PyObject *tmp = SwigPyObject_New(sobj->ptr, ty, 0);
        if (tmp) {
          res = SWIG_Python_CallFunctor(destroy, tmp);
        } else {
          res = 0;
        }
        SWIG_Py_XDECREF(tmp);
      } else {
        PyCFunction meth = PyCFunction_GET_FUNCTION(destroy);
        PyObject *mself = PyCFunction_GET_SELF(destroy);
        res = ((*meth)(mself, v));
      }
      if (!res)
        PyErr_WriteUnraisable(destroy);

      PyErr_Restore(type, value, traceback);

      SWIG_Py_XDECREF(res);
    } 
#if !defined(SWIG_PYTHON_SILENT_MEMLEAK)
    else {
      const char *name = SWIG_TypePrettyName(ty);
      printf("swig/python detected a memory leak of type '%s', no destructor found.\n", (name ? name : "unknown"));
    }
#endif
    SWIG_Py_XDECREF(Swig_Capsule_global);
  }
  SWIG_Py_XDECREF(next);
  SWIG_Py_XDECREF(sobj->swigdict);
  PyObject_Free(v);
}

SWIGRUNTIME PyObject* 
SwigPyObject_append(PyObject* v, PyObject* next)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (!SwigPyObject_Check(next)) {
    PyErr_SetString(PyExc_TypeError, "Attempt to append a non SwigPyObject");
    return NULL;
  }
  ((SwigPyObject *)next)->next = sobj->next;
  sobj->next = next;
  SWIG_Py_INCREF(next);
  return SWIG_Py_Void();
}

SWIGRUNTIME PyObject* 
SwigPyObject_next(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (sobj->next) {    
    SWIG_Py_INCREF(sobj->next);
    return sobj->next;
  } else {
    return SWIG_Py_Void();
  }
}

SWIGINTERN PyObject*
SwigPyObject_disown(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = 0;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_acquire(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = SWIG_POINTER_OWN;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_own(PyObject *v, PyObject *args)
{
  PyObject *val = 0;
  if (!PyArg_UnpackTuple(args, "own", 0, 1, &val)) {
    return NULL;
  } else {
    SwigPyObject *sobj = (SwigPyObject *)v;
    PyObject *obj = PyBool_FromLong(sobj->own);
    if (val) {
      if (PyObject_IsTrue(val)) {
        SWIG_Py_DECREF(SwigPyObject_acquire(v,args));
      } else {
        SWIG_Py_DECREF(SwigPyObject_disown(v,args));
      }
    } 
    return obj;
  }
}

static PyMethodDef
SwigPyObject_methods[] = {
  {"disown",  SwigPyObject_disown,  METH_NOARGS,  "releases ownership of the pointer"},
  {"acquire", SwigPyObject_acquire, METH_NOARGS,  "acquires ownership of the pointer"},
  {"own",     SwigPyObject_own,     METH_VARARGS, "returns/sets ownership of the pointer"},
  {"append",  SwigPyObject_append,  METH_O,       "appends another 'this' object"},
  {"next",    SwigPyObject_next,    METH_NOARGS,  "returns the next 'this' object"},
  {"__repr__",SwigPyObject_repr2,   METH_NOARGS,  "returns object representation"},
  {0, 0, 0, 0}  
};

SWIGRUNTIME PyTypeObject*
SwigPyObject_TypeOnce(void) {
  static char SwigPyObject_doc[] = "Swig object holding a C/C++ pointer";
#ifndef SWIG_HEAPTYPES
  static PyNumberMethods SwigPyObject_as_number = {
    (binaryfunc)0, /*nb_add*/
    (binaryfunc)0, /*nb_subtract*/
    (binaryfunc)0, /*nb_multiply*/
    /* nb_divide removed in Python 3 */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc)0, /*nb_divide*/
#endif
    (binaryfunc)0, /*nb_remainder*/
    (binaryfunc)0, /*nb_divmod*/
    (ternaryfunc)0,/*nb_power*/
    (unaryfunc)0,  /*nb_negative*/
    (unaryfunc)0,  /*nb_positive*/
    (unaryfunc)0,  /*nb_absolute*/
    (inquiry)0,    /*nb_nonzero*/
    0,		   /*nb_invert*/
    0,		   /*nb_lshift*/
    0,		   /*nb_rshift*/
    0,		   /*nb_and*/
    0,		   /*nb_xor*/
    0,		   /*nb_or*/
#if PY_VERSION_HEX < 0x03000000
    0,   /*nb_coerce*/
#endif
    (unaryfunc)SwigPyObject_long, /*nb_int*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_long, /*nb_long*/
#else
    0, /*nb_reserved*/
#endif
    (unaryfunc)0,                 /*nb_float*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_oct,  /*nb_oct*/
    (unaryfunc)SwigPyObject_hex,  /*nb_hex*/
#endif
#if PY_VERSION_HEX >= 0x03050000 /* 3.5 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_matrix_multiply */
#elif PY_VERSION_HEX >= 0x03000000 /* 3.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index, nb_inplace_divide removed */
#else
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index */
#endif
  };

  static PyTypeObject swigpyobject_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      SWIG_RUNTIME_MODULE ".SwigPyObject",  /* tp_name */
      sizeof(SwigPyObject),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyObject_dealloc,     /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                         /* tp_print */
#else
      (Py_ssize_t)0,                        /* tp_vectorcall_offset */
#endif
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1, tp_compare in 3.0.0 but not used */
#else
      (cmpfunc)SwigPyObject_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyObject_repr,          /* tp_repr */
      &SwigPyObject_as_number,              /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      0,                                    /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      SwigPyObject_doc,                     /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      (richcmpfunc)SwigPyObject_richcompare,/* tp_richcompare */
      offsetof(SwigPyObject, weakreflist),  /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      SwigPyObject_methods,                 /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030c0000
      0,                                    /* tp_watched */
#endif
#if PY_VERSION_HEX >= 0x030d00a4
      0,                                    /* tp_versions_used */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    PyObject *runtime_data_module = SWIG_runtime_data_module();
    swigpyobject_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpyobject_type) != 0)
      return NULL;
    if (PyModule_AddObject(runtime_data_module, "SwigPyObject", (PyObject *)&swigpyobject_type) == 0)
      SWIG_Py_INCREF((PyObject *)&swigpyobject_type);
  }
  return &swigpyobject_type;
#else
  static PyMemberDef SwigPyObject_members[] = {
      { (char *)"__dictoffset__", Py_T_PYSSIZET, offsetof(SwigPyObject, swigdict), Py_READONLY, NULL },
      { (char *)"__weaklistoffset__", Py_T_PYSSIZET, offsetof(SwigPyObject, weakreflist), Py_READONLY, NULL },
      { NULL, 0, 0, 0, NULL }
  };
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigPyObject_dealloc },
    { Py_tp_repr, (void *)SwigPyObject_repr },
    { Py_tp_getattro, (void *)PyObject_GenericGetAttr },
    { Py_tp_doc, (void *)SwigPyObject_doc },
    { Py_tp_richcompare, (void *)SwigPyObject_richcompare },
    { Py_tp_methods, (void *)SwigPyObject_methods },
    { Py_nb_int, (void *)SwigPyObject_long },
    { Py_tp_members, (void *)SwigPyObject_members },
    { 0, NULL }
  };
  PyType_Spec spec = {
    SWIG_RUNTIME_MODULE ".SwigPyObject",
    sizeof(SwigPyObject),
    0,
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,
    slots
  };
  PyObject *pytype = PyType_FromSpec(&spec);
  PyObject *runtime_data_module = SWIG_runtime_data_module();
#if !defined(Py_LIMITED_API)
/* While this __dictoffset__ is only used with the builtin wrappers, SwigPyObject ought to be
   identical when created for use by proxy class wrappers in case it is shared across multiple modules. */
#if PY_VERSION_HEX < 0x03090000
    /* Workaround as __dictoffset__ and __weaklistoffset__ above are only supported from python-3.9 */
  if (pytype) {
    ((PyTypeObject *)pytype)->tp_dictoffset = offsetof(SwigPyObject, swigdict);
    ((PyTypeObject *)pytype)->tp_weaklistoffset = offsetof(SwigPyObject, weakreflist);
  }
#endif
#endif
  if (pytype && PyModule_AddObject(runtime_data_module, "SwigPyObject", pytype) == 0)
    SWIG_Py_INCREF(pytype);
  return (PyTypeObject *)pytype;
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own)
{
  SwigPyObject *sobj = PyObject_New(SwigPyObject, SwigPyObject_Type());
  if (sobj) {
    sobj->ptr  = ptr;
    sobj->ty   = ty;
    sobj->own  = own;
    sobj->next = 0;
    sobj->swigdict = 0;
    sobj->weakreflist = 0;
    if (own == SWIG_POINTER_OWN) {
      /* Obtain a reference to the Python capsule wrapping the module information, so that the
       * module information is correctly destroyed after all SWIG python objects have been freed
       * by the GC (and corresponding destructors invoked) */
      SWIG_Py_XINCREF(Swig_Capsule_global);
    }
  }
  return (PyObject *)sobj;
}

/* -----------------------------------------------------------------------------
 * Implements a simple Swig Packed type, and use it instead of string
 * ----------------------------------------------------------------------------- */

typedef struct {
  PyObject_HEAD
  void *pack;
  swig_type_info *ty;
  size_t size;
} SwigPyPacked;

SWIGRUNTIME PyObject *
SwigPyPacked_repr(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    return SWIG_Python_str_FromFormat("<Swig Packed at %s%s>", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromFormat("<Swig Packed %s>", v->ty->name);
  }  
}

SWIGRUNTIME PyObject *
SwigPyPacked_str(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))){
    return SWIG_Python_str_FromFormat("%s%s", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromChar(v->ty->name);
  }  
}

SWIGRUNTIME int
SwigPyPacked_compare(SwigPyPacked *v, SwigPyPacked *w)
{
  size_t i = v->size;
  size_t j = w->size;
  int s = (i < j) ? -1 : ((i > j) ? 1 : 0);
  return s ? s : strncmp((const char *)v->pack, (const char *)w->pack, 2*v->size);
}

SWIGRUNTIME PyTypeObject* SwigPyPacked_TypeOnce(void);

SWIGRUNTIME PyTypeObject*
SwigPyPacked_Type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyPacked_TypeOnce();
  return type;
}

SWIGRUNTIMEINLINE int
SwigPyPacked_Check(PyObject *op) {
  PyTypeObject *target_tp = SwigPyPacked_Type();
  PyTypeObject *op_type = Py_TYPE(op);
  /* Check for an exact match to SwigPyPacked */
  if (op_type == target_tp) {
    return 1;
  } else {
    /* Fallback for multiple modules */
#if PY_VERSION_HEX >= 0x03000000
    int cmp;
    PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(op_type);
    if (!tpname)
      return 0;
    cmp = PyUnicode_CompareWithASCIIString(tpname, SWIG_RUNTIME_MODULE ".SwigPyPacked");
    SWIG_Py_DECREF(tpname);
    return cmp == 0;
#else
    return strcmp(op_type->tp_name, SWIG_RUNTIME_MODULE ".SwigPyPacked") == 0;
#endif
  }
}

SWIGRUNTIME void
SwigPyPacked_dealloc(PyObject *v)
{
  if (SwigPyPacked_Check(v)) {
    SwigPyPacked *sobj = (SwigPyPacked *) v;
    free(sobj->pack);
  }
  PyObject_Free(v);
}

SWIGRUNTIME PyTypeObject*
SwigPyPacked_TypeOnce(void) {
  static char SwigPyPacked_doc[] = "Swig object holding a C/C++ function pointer";
#ifndef SWIG_HEAPTYPES
  static PyTypeObject swigpypacked_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      SWIG_RUNTIME_MODULE ".SwigPyPacked",  /* tp_name */
      sizeof(SwigPyPacked),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyPacked_dealloc,     /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                         /* tp_print */
#else
      (Py_ssize_t)0,                        /* tp_vectorcall_offset */
#endif
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1 */
#else
      (cmpfunc)SwigPyPacked_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyPacked_repr,          /* tp_repr */
      0,                                    /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      (reprfunc)SwigPyPacked_str,           /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      SwigPyPacked_doc,                     /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      0,                                    /* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      0,                                    /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030c0000
      0,                                    /* tp_watched */
#endif
#if PY_VERSION_HEX >= 0x030d00a4
      0,                                    /* tp_versions_used */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    PyObject *runtime_data_module = SWIG_runtime_data_module();
    swigpypacked_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpypacked_type) != 0)
      return NULL;
    if (PyModule_AddObject(runtime_data_module, "SwigPyPacked", (PyObject *)&swigpypacked_type) == 0)
      SWIG_Py_INCREF((PyObject *)&swigpypacked_type);
  }
  return &swigpypacked_type;
#else
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigPyPacked_dealloc },
    { Py_tp_repr, (void *)SwigPyPacked_repr },
    { Py_tp_str, (void *)SwigPyPacked_str },
    { Py_tp_getattro, (void *)PyObject_GenericGetAttr },
    { Py_tp_doc, (void *)SwigPyPacked_doc },
    { 0, NULL }
  };
  PyType_Spec spec = {
    SWIG_RUNTIME_MODULE ".SwigPyPacked",
    sizeof(SwigPyPacked),
    0,
    Py_TPFLAGS_DEFAULT,
    slots
  };
  PyObject *pytype = PyType_FromSpec(&spec);
  PyObject *runtime_data_module = SWIG_runtime_data_module();
  if (pytype && PyModule_AddObject(runtime_data_module, "SwigPyPacked", pytype) == 0)
    SWIG_Py_INCREF(pytype);
  return (PyTypeObject *)pytype;
#endif
}

SWIGRUNTIME PyObject *
SwigPyPacked_New(void *ptr, size_t size, swig_type_info *ty)
{
  SwigPyPacked *sobj = PyObject_New(SwigPyPacked, SwigPyPacked_Type());
  if (sobj) {
    void *pack = malloc(size);
    if (pack) {
      memcpy(pack, ptr, size);
      sobj->pack = pack;
      sobj->ty   = ty;
      sobj->size = size;
    } else {
      PyObject_Free((PyObject *)sobj);
      sobj = 0;
    }
  }
  return (PyObject *) sobj;
}

SWIGRUNTIME swig_type_info *
SwigPyPacked_UnpackData(PyObject *obj, void *ptr, size_t size)
{
  if (SwigPyPacked_Check(obj)) {
    SwigPyPacked *sobj = (SwigPyPacked *)obj;
    if (sobj->size != size) return 0;
    memcpy(ptr, sobj->pack, size);
    return sobj->ty;
  } else {
    return 0;
  }
}

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

static PyObject *Swig_This_global = NULL;

SWIGRUNTIME PyObject *
SWIG_This(void)
{
  if (Swig_This_global == NULL)
    Swig_This_global = SWIG_Python_str_FromChar("this");
  return Swig_This_global;
}

/* #define SWIG_PYTHON_SLOW_GETSET_THIS */

/* TODO: I don't know how to implement the fast getset in Python 3 right now */
#if PY_VERSION_HEX >= 0x03000000
#define SWIG_PYTHON_SLOW_GETSET_THIS 
#endif

/* Returns a borrowed reference to the 'this' object */
SWIGRUNTIME SwigPyObject *
SWIG_Python_GetSwigThis(PyObject *pyobj) 
{
  PyObject *obj;

  if (SwigPyObject_Check(pyobj))
    return (SwigPyObject *) pyobj;

#ifdef SWIGPYTHON_BUILTIN
  (void)obj;
  if (PyWeakref_CheckProxy(pyobj)) {
#if PY_VERSION_HEX >= 0x030d0000
    if (PyWeakref_GetRef(pyobj, &pyobj) > 0)
      Py_DECREF(pyobj);
    else
      pyobj = NULL;
#else
    pyobj = PyWeakref_GetObject(pyobj);
#endif
    if (pyobj && SwigPyObject_Check(pyobj))
      return (SwigPyObject*) pyobj;
  }
  return NULL;
#else

  obj = 0;

#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  if (PyInstance_Check(pyobj)) {
    obj = _PyInstance_Lookup(pyobj, SWIG_This());      
  } else {
    PyObject **dictptr = _PyObject_GetDictPtr(pyobj);
    if (dictptr != NULL) {
      PyObject *dict = *dictptr;
      obj = dict ? PyDict_GetItem(dict, SWIG_This()) : 0;
    } else {
      if (PyWeakref_CheckProxy(pyobj)) {
        PyObject *wobj = PyWeakref_GET_OBJECT(pyobj);
	return wobj ? SWIG_Python_GetSwigThis(wobj) : 0;
      }
      obj = PyObject_GetAttr(pyobj, SWIG_This());
      if (obj) {
	SWIG_Py_DECREF(obj);
      } else {
	if (PyErr_Occurred()) PyErr_Clear();
	return 0;
      }
    }
  }
#else
  obj = PyObject_GetAttr(pyobj, SWIG_This());
  if (obj) {
    SWIG_Py_DECREF(obj);
  } else {
    if (PyErr_Occurred()) PyErr_Clear();
    return 0;
  }
#endif
  if (obj && !SwigPyObject_Check(obj)) {
    /* a PyObject is called 'this', try to get the 'real this'
       SwigPyObject from it */ 
    return SWIG_Python_GetSwigThis(obj);
  }
  return (SwigPyObject *)obj;
#endif
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Python_AcquirePtr(PyObject *obj, int own) {
  if (own == SWIG_POINTER_OWN) {
    SwigPyObject *sobj = SWIG_Python_GetSwigThis(obj);
    if (sobj) {
      int oldown = sobj->own;
      sobj->own = own;
      return oldown;
    }
  }
  return 0;
}

/* Convert a pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPtrAndOwn(PyObject *obj, void **ptr, swig_type_info *ty, int flags, int *own) {
  int res;
  SwigPyObject *sobj;
  int implicit_conv = (flags & SWIG_POINTER_IMPLICIT_CONV) != 0;

  if (!obj)
    return SWIG_ERROR;
  if (obj == Py_None && !implicit_conv) {
    if (ptr)
      *ptr = 0;
    return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;
  }

  res = SWIG_ERROR;

  sobj = SWIG_Python_GetSwigThis(obj);
  if (own)
    *own = 0;
  while (sobj) {
    void *vptr = sobj->ptr;
    if (ty) {
      swig_type_info *to = sobj->ty;
      if (to == ty) {
        /* no type cast needed */
        if (ptr) *ptr = vptr;
        break;
      } else {
        swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
        if (!tc) {
          sobj = (SwigPyObject *)sobj->next;
        } else {
          if (ptr) {
            int newmemory = 0;
            *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
            if (newmemory == SWIG_CAST_NEW_MEMORY) {
              assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
              if (own)
                *own = *own | SWIG_CAST_NEW_MEMORY;
            }
          }
          break;
        }
      }
    } else {
      if (ptr) *ptr = vptr;
      break;
    }
  }
  if (sobj) {
    if (((flags & SWIG_POINTER_RELEASE) == SWIG_POINTER_RELEASE) && !sobj->own) {
      res = SWIG_ERROR_RELEASE_NOT_OWNED;
    } else {
      if (own)
        *own = *own | sobj->own;
      if (flags & SWIG_POINTER_DISOWN) {
        sobj->own = 0;
      }
      if (flags & SWIG_POINTER_CLEAR) {
        sobj->ptr = 0;
      }
      res = SWIG_OK;
    }
  } else {
    if (implicit_conv) {
      SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
      if (data && !data->implicitconv) {
        PyObject *klass = data->klass;
        if (klass) {
          PyObject *impconv;
          data->implicitconv = 1; /* avoid recursion and call 'explicit' constructors*/
          impconv = SWIG_Python_CallFunctor(klass, obj);
          data->implicitconv = 0;
          if (PyErr_Occurred()) {
            PyErr_Clear();
            impconv = 0;
          }
          if (impconv) {
            SwigPyObject *iobj = SWIG_Python_GetSwigThis(impconv);
            if (iobj) {
              void *vptr;
              res = SWIG_Python_ConvertPtrAndOwn((PyObject*)iobj, &vptr, ty, 0, 0);
              if (SWIG_IsOK(res)) {
                if (ptr) {
                  *ptr = vptr;
                  /* transfer the ownership to 'ptr' */
                  iobj->own = 0;
                  res = SWIG_AddCast(res);
                  res = SWIG_AddNewMask(res);
                } else {
                  res = SWIG_AddCast(res);		    
                }
              }
            }
            SWIG_Py_DECREF(impconv);
          }
        }
      }
      if (!SWIG_IsOK(res) && obj == Py_None) {
        if (ptr)
          *ptr = 0;
        if (PyErr_Occurred())
          PyErr_Clear();
        res = SWIG_OK;
      }
    }
  }
  return res;
}

/* Convert a function ptr value */

SWIGRUNTIME int
SWIG_Python_ConvertFunctionPtr(PyObject *obj, void **ptr, swig_type_info *ty) {
  if (!PyCFunction_Check(obj)) {
    return SWIG_ConvertPtr(obj, ptr, ty, 0);
  } else {
    void *vptr = 0;
    swig_cast_info *tc;

    /* here we get the method pointer for callbacks */
#ifndef Py_LIMITED_API
    const char *doc = (((PyCFunctionObject *)obj) -> m_ml -> ml_doc);
#else
    PyObject* pystr_doc = PyObject_GetAttrString(obj, "__doc__");
    PyObject *bytes = NULL;
    const char *doc = pystr_doc ? SWIG_PyUnicode_AsUTF8AndSize(pystr_doc, NULL, &bytes) : 0;
#endif
    const char *desc = doc ? strstr(doc, "swig_ptr: ") : 0;
    if (desc)
      desc = ty ? SWIG_UnpackVoidPtr(desc + 10, &vptr, ty->name) : 0;
#ifdef Py_LIMITED_API
    SWIG_Py_XDECREF(bytes);
    SWIG_Py_XDECREF(pystr_doc);
#endif
    if (!desc)
      return SWIG_ERROR;
    tc = SWIG_TypeCheck(desc,ty);
    if (tc) {
      int newmemory = 0;
      *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
      assert(!newmemory); /* newmemory handling not yet implemented */
    } else {
      return SWIG_ERROR;
    }
    return SWIG_OK;
  }
}

/* Convert a packed pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPacked(PyObject *obj, void *ptr, size_t sz, swig_type_info *ty) {
  swig_type_info *to = SwigPyPacked_UnpackData(obj, ptr, sz);
  if (!to) return SWIG_ERROR;
  if (ty) {
    if (to != ty) {
      /* check type cast? */
      swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
      if (!tc) return SWIG_ERROR;
    }
  }
  return SWIG_OK;
}  

/* -----------------------------------------------------------------------------
 * Create a new pointer object
 * ----------------------------------------------------------------------------- */

/*
  Create a new instance object, without calling __init__, and set the
  'this' attribute.
*/

SWIGRUNTIME PyObject* 
SWIG_Python_NewShadowInstance(SwigPyClientData *data, PyObject *swig_this)
{
  PyObject *inst = 0;
  PyObject *newraw = data->newraw;
  if (newraw) {
    inst = PyObject_Call(newraw, data->newargs, NULL);
    if (inst) {
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
      PyObject **dictptr = _PyObject_GetDictPtr(inst);
      if (dictptr != NULL) {
        PyObject *dict = *dictptr;
        if (dict == NULL) {
          dict = PyDict_New();
          *dictptr = dict;
        }
        if (dict) {
          PyDict_SetItem(dict, SWIG_This(), swig_this);
        } else{
          SWIG_Py_DECREF(inst);
          inst = 0;
        }
      }
#else
      if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
        SWIG_Py_DECREF(inst);
        inst = 0;
      }
#endif
    }
  } else {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *empty_args = PyTuple_New(0);
    if (empty_args) {
      PyObject *empty_kwargs = PyDict_New();
      if (empty_kwargs) {
#ifndef Py_LIMITED_API
        newfunc newfn = ((PyTypeObject *)data->newargs)->tp_new;
#else
        newfunc newfn = (newfunc)PyType_GetSlot((PyTypeObject *)data->newargs, Py_tp_new);
#endif
        inst = newfn((PyTypeObject *)data->newargs, empty_args, empty_kwargs);
        SWIG_Py_DECREF(empty_kwargs);
        if (inst) {
          if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
            SWIG_Py_DECREF(inst);
            inst = 0;
          } else {
            PyType_Modified(Py_TYPE(inst));
          }
        }
      }
      SWIG_Py_DECREF(empty_args);
    }
#else
    PyObject *dict = PyDict_New();
    if (dict) {
      PyDict_SetItem(dict, SWIG_This(), swig_this);
      inst = PyInstance_NewRaw(data->newargs, dict);
      SWIG_Py_DECREF(dict);
    }
#endif
  }
  return inst;
}

SWIGRUNTIME int
SWIG_Python_SetSwigThis(PyObject *inst, PyObject *swig_this)
{
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  PyObject **dictptr = _PyObject_GetDictPtr(inst);
  if (dictptr != NULL) {
    PyObject *dict = *dictptr;
    if (dict == NULL) {
      dict = PyDict_New();
      *dictptr = dict;
    }
    if (dict) {
      return PyDict_SetItem(dict, SWIG_This(), swig_this);
    } else{
      return -1;
    }
  }
#endif
  return PyObject_SetAttr(inst, SWIG_This(), swig_this);
} 


SWIGINTERN PyObject *
SWIG_Python_InitShadowInstance(PyObject *args) {
  PyObject *obj[2];
  if (!SWIG_Python_UnpackTuple(args, "swiginit", 2, 2, obj)) {
    return NULL;
  } else {
    SwigPyObject *sthis = SWIG_Python_GetSwigThis(obj[0]);
    if (sthis) {
      SWIG_Py_DECREF(SwigPyObject_append((PyObject*) sthis, obj[1]));
    } else {
      if (SWIG_Python_SetSwigThis(obj[0], obj[1]) != 0)
        return NULL;
    }
    return SWIG_Py_Void();
  }
}

/* Create a new pointer object */

SWIGRUNTIME PyObject *
SWIG_Python_NewPointerObj(PyObject *self, void *ptr, swig_type_info *type, int flags) {
  SwigPyClientData *clientdata;
  PyObject * robj;
  int own;

  if (!ptr)
    return SWIG_Py_Void();

  clientdata = type ? (SwigPyClientData *)(type->clientdata) : 0;
  own = (flags & SWIG_POINTER_OWN) ? SWIG_POINTER_OWN : 0;
  if (clientdata && clientdata->pytype) {
    SwigPyObject *newobj;
    if (flags & SWIG_BUILTIN_TP_INIT) {
      newobj = (SwigPyObject*) self;
      if (newobj->ptr) {
#ifndef Py_LIMITED_API
        allocfunc alloc = clientdata->pytype->tp_alloc;
#else
        allocfunc alloc = (allocfunc)PyType_GetSlot(clientdata->pytype, Py_tp_alloc);
#endif
        PyObject *next_self = alloc(clientdata->pytype, 0);
        while (newobj->next)
	  newobj = (SwigPyObject *) newobj->next;
        newobj->next = next_self;
        newobj = (SwigPyObject *)next_self;
        newobj->swigdict = 0;
        newobj->weakreflist = 0;
      }
    } else {
      newobj = PyObject_New(SwigPyObject, clientdata->pytype);
      if (newobj) {
        newobj->swigdict = 0;
        newobj->weakreflist = 0;
      }
    }
    if (newobj) {
      newobj->ptr = ptr;
      newobj->ty = type;
      newobj->own = own;
      newobj->next = 0;
      return (PyObject*) newobj;
    }
    return SWIG_Py_Void();
  }

  assert(!(flags & SWIG_BUILTIN_TP_INIT));

  robj = SwigPyObject_New(ptr, type, own);
  if (robj && clientdata && !(flags & SWIG_POINTER_NOSHADOW)) {
    PyObject *inst = SWIG_Python_NewShadowInstance(clientdata, robj);
    SWIG_Py_DECREF(robj);
    robj = inst;
  }
  return robj;
}

/* Create a new packed object */

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_NewPackedObj(void *ptr, size_t sz, swig_type_info *type) {
  return ptr ? SwigPyPacked_New((void *) ptr, sz, type) : SWIG_Py_Void();
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

#ifdef SWIG_LINK_RUNTIME
void *SWIG_ReturnGlobalTypeList(void *);
#endif

static PyObject *Swig_TypeCache_global = NULL;

/* The python cached type query */
SWIGRUNTIME PyObject *
SWIG_Python_TypeCache(void) {
  if (Swig_TypeCache_global == NULL) {
    Swig_TypeCache_global = PyDict_New();
  }
  return Swig_TypeCache_global;
}

SWIGRUNTIME swig_module_info *
SWIG_Python_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
#ifdef SWIG_LINK_RUNTIME
  static void *type_pointer = (void *)0;
  /* first check if module already created */
  if (!type_pointer) {
    type_pointer = SWIG_ReturnGlobalTypeList((void *)0);
  }
#else
  void *type_pointer = PyCapsule_Import(SWIGPY_CAPSULE_NAME, 0);
  if (PyErr_Occurred()) {
    PyErr_Clear();
    type_pointer = (void *)0;
  }
#endif
  return (swig_module_info *) type_pointer;
}


#if defined(SWIG_REFCNT_DEBUG)
#define SWIG_PYOBJ_REFCNT(OBJ) fprintf(stdout, "" #OBJ " count %ld\n", (OBJ ? Py_REFCNT(OBJ) : 0))
#else
#define SWIG_PYOBJ_REFCNT(OBJ)
#endif

static int interpreter_counter = 0; /* how many (sub-)interpreters are using swig_module's types */

SWIGRUNTIME void
SWIG_Python_DestroyModule(PyObject *obj)
{
  swig_module_info *swig_module = (swig_module_info *) PyCapsule_GetPointer(obj, SWIGPY_CAPSULE_NAME);
  swig_type_info **types = swig_module->types;
  size_t i;
  if (--interpreter_counter != 0) /* another sub-interpreter may still be using the swig_module's types */
    return;
  for (i = 0; i < swig_module->size; ++i) {
    swig_type_info *ty = types[i];
    if (ty->owndata) {
      SwigPyClientData *data = (SwigPyClientData *) ty->clientdata;
      ty->clientdata = 0;
      if (data) SwigPyClientData_Del(data);
    }
  }
  SWIG_PYOBJ_REFCNT(Swig_This_global);
  SWIG_Py_XDECREF(Swig_This_global);
  Swig_This_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_Globals_global);
  SWIG_Py_XDECREF(Swig_Globals_global);
  Swig_Globals_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_TypeCache_global);
  SWIG_Py_XDECREF(Swig_TypeCache_global);
  Swig_TypeCache_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_Capsule_global);
  Swig_Capsule_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_runtime_data_module_global);
  SWIG_Py_XDECREF(Swig_runtime_data_module_global);
  Swig_runtime_data_module_global = NULL;
}

SWIGRUNTIME void
SWIG_Python_SetModule(swig_module_info *swig_module) {
  PyObject *runtime_data_module = SWIG_runtime_data_module();
  PyObject *pointer = PyCapsule_New((void *) swig_module, SWIGPY_CAPSULE_NAME, SWIG_Python_DestroyModule);
  if (pointer && runtime_data_module) {
    if (PyModule_AddObject(runtime_data_module, SWIGPY_CAPSULE_ATTR_NAME, pointer) == 0) {
      ++interpreter_counter;
      Swig_Capsule_global = pointer;
    } else {
      SWIG_Py_DECREF(pointer);
    }
  } else {
    SWIG_Py_XDECREF(pointer);
  }
}

SWIGRUNTIME swig_type_info *
SWIG_Python_TypeQuery(const char *type)
{
  swig_type_info *descriptor;
  PyObject *cache = SWIG_Python_TypeCache();
  PyObject *obj;
  SWIG_PyDict_GetItemStringRef(cache, type, &obj);
  if (obj) {
    descriptor = (swig_type_info *) PyCapsule_GetPointer(obj, NULL);
  } else {
    swig_module_info *swig_module = SWIG_GetModule(0);
    descriptor = SWIG_TypeQueryModule(swig_module, swig_module, type);
    if (descriptor) {
      obj = PyCapsule_New((void*) descriptor, NULL, NULL);
      if (obj) PyDict_SetItemString(cache, type, obj);
    }
  }
  SWIG_Py_XDECREF(obj);
  return descriptor;
}

/* 
   For backward compatibility only
*/
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_arg_fail(arg)      SWIG_Python_ArgFail(arg)
#define SWIG_MustGetPtr(p, type, argnum, flags)  SWIG_Python_MustGetPtr(p, type, argnum, flags)

SWIGRUNTIME int
SWIG_Python_AddErrMesg(const char* mesg, int infront)
{  
  if (PyErr_Occurred()) {
    PyObject *type = 0;
    PyObject *value = 0;
    PyObject *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    if (value) {
      PyObject *old_str = PyObject_Str(value);
      PyObject *bytes = NULL;
      const char *tmp = SWIG_PyUnicode_AsUTF8AndSize(old_str, NULL, &bytes);
      const char *errmesg = tmp ? tmp : "Invalid error message";
      SWIG_Py_XINCREF(type);
      PyErr_Clear();
      if (infront) {
	PyErr_Format(type, "%s %s", mesg, errmesg);
      } else {
	PyErr_Format(type, "%s %s", errmesg, mesg);
      }
      SWIG_Py_XDECREF(bytes);
      SWIG_Py_DECREF(old_str);
    }
    return 1;
  } else {
    return 0;
  }
}
  
SWIGRUNTIME int
SWIG_Python_ArgFail(int argnum)
{
  if (PyErr_Occurred()) {
    /* add information about failing argument */
    char mesg[256];
    PyOS_snprintf(mesg, sizeof(mesg), "argument number %d:", argnum);
    return SWIG_Python_AddErrMesg(mesg, 1);
  } else {
    return 0;
  }
}

SWIGRUNTIMEINLINE const char *
SwigPyObject_GetDesc(PyObject *self)
{
  SwigPyObject *v = (SwigPyObject *)self;
  swig_type_info *ty = v ? v->ty : 0;
  return ty ? ty->str : "";
}

/* Convert a pointer value, signal an exception on a type mismatch */
SWIGRUNTIME void *
SWIG_Python_MustGetPtr(PyObject *obj, swig_type_info *ty, int SWIGUNUSEDPARM(argnum), int flags) {
  void *result;
  if (SWIG_Python_ConvertPtr(obj, &result, ty, flags) == -1) {
    PyErr_Clear();
  }
  return result;
}

#ifdef SWIGPYTHON_BUILTIN
SWIGRUNTIME int
SWIG_Python_NonDynamicSetAttr(PyObject *obj, PyObject *name, PyObject *value) {
  PyTypeObject *tp = Py_TYPE(obj);
  PyObject *descr;
  PyObject *encoded_name;
  descrsetfunc f;
  int res = -1;

# ifdef Py_USING_UNICODE
  if (PyString_Check(name)) {
    name = PyUnicode_Decode(PyString_AsString(name), PyString_Size(name), NULL, NULL);
    if (!name)
      return -1;
  } else if (!PyUnicode_Check(name))
# else
  if (!PyString_Check(name))
# endif
  {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(Py_TYPE(name));
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%S'", tpname);
    SWIG_Py_DECREF(tpname);
#else
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%s'", Py_TYPE(name)->tp_name);
#endif
    return -1;
  } else {
    SWIG_Py_INCREF(name);
  }

  if (!tp->tp_dict) {
    if (PyType_Ready(tp) != 0)
      goto done;
  }

  descr = _PyType_Lookup(tp, name);
  f = NULL;
  if (descr != NULL)
    f = Py_TYPE(descr)->tp_descr_set;
  if (!f) {
    if (PyString_Check(name)) {
      encoded_name = name;
      SWIG_Py_INCREF(name);
    } else {
      encoded_name = PyUnicode_AsUTF8String(name);
      if (!encoded_name)
        goto done;
    }
#if PY_VERSION_HEX >= 0x03000000
    {
      PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(tp);
      PyErr_Format(PyExc_AttributeError, "'%S' object has no attribute '%s'", tpname, PyString_AsString(encoded_name));
      SWIG_Py_DECREF(tpname);
    }
#else
    PyErr_Format(PyExc_AttributeError, "'%s' object has no attribute '%s'", tp->tp_name, PyString_AsString(encoded_name));
#endif
    SWIG_Py_DECREF(encoded_name);
  } else {
    res = f(descr, obj, value);
  }
  
  done:
  SWIG_Py_DECREF(name);
  return res;
}
#endif


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) do { if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } } while (0) 



  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg); SWIG_fail;; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_Amp swig_types[0]
#define SWIGTYPE_p_FIFO_RIG_s swig_types[1]
#define SWIGTYPE_p_FILE swig_types[2]
#define SWIGTYPE_p_Rig swig_types[3]
#define SWIGTYPE_p_Rot swig_types[4]
#define SWIGTYPE_p_agc_level_e swig_types[5]
#define SWIGTYPE_p_amp swig_types[6]
#define SWIGTYPE_p_amp_caps swig_types[7]
#define SWIGTYPE_p_amp_reset_e swig_types[8]
#define SWIGTYPE_p_amp_state swig_types[9]
#define SWIGTYPE_p_amp_type_e swig_types[10]
#define SWIGTYPE_p_ann_t swig_types[11]
#define SWIGTYPE_p_cal_table swig_types[12]
#define SWIGTYPE_p_cal_table_float swig_types[13]
#define SWIGTYPE_p_cal_table_float_table swig_types[14]
#define SWIGTYPE_p_cal_table_table swig_types[15]
#define SWIGTYPE_p_chan_list swig_types[16]
#define SWIGTYPE_p_chan_type_t swig_types[17]
#define SWIGTYPE_p_channel swig_types[18]
#define SWIGTYPE_p_channelArray swig_types[19]
#define SWIGTYPE_p_channel_cap swig_types[20]
#define SWIGTYPE_p_char swig_types[21]
#define SWIGTYPE_p_client_e swig_types[22]
#define SWIGTYPE_p_confparams swig_types[23]
#define SWIGTYPE_p_confparams_u swig_types[24]
#define SWIGTYPE_p_confparams_u_c swig_types[25]
#define SWIGTYPE_p_confparams_u_n swig_types[26]
#define SWIGTYPE_p_dcd_e swig_types[27]
#define SWIGTYPE_p_dcd_type_e swig_types[28]
#define SWIGTYPE_p_deferred_config_header swig_types[29]
#define SWIGTYPE_p_deferred_config_item swig_types[30]
#define SWIGTYPE_p_double swig_types[31]
#define SWIGTYPE_p_ext_list swig_types[32]
#define SWIGTYPE_p_f_p_struct_amp__int swig_types[33]
#define SWIGTYPE_p_f_p_struct_amp__p_char swig_types[34]
#define SWIGTYPE_p_f_p_struct_amp_double__int swig_types[35]
#define SWIGTYPE_p_f_p_struct_amp_enum_amp_reset_e__int swig_types[36]
#define SWIGTYPE_p_f_p_struct_amp_enum_powerstat_t__int swig_types[37]
#define SWIGTYPE_p_f_p_struct_amp_long_p_char__int swig_types[38]
#define SWIGTYPE_p_f_p_struct_amp_long_p_char_int__int swig_types[39]
#define SWIGTYPE_p_f_p_struct_amp_long_p_q_const__char__int swig_types[40]
#define SWIGTYPE_p_f_p_struct_amp_long_p_value_t__int swig_types[41]
#define SWIGTYPE_p_f_p_struct_amp_long_value_t__int swig_types[42]
#define SWIGTYPE_p_f_p_struct_amp_p_double__int swig_types[43]
#define SWIGTYPE_p_f_p_struct_amp_p_enum_powerstat_t__int swig_types[44]
#define SWIGTYPE_p_f_p_struct_amp_unsigned_long_long_p_value_t__int swig_types[45]
#define SWIGTYPE_p_f_p_struct_amp_unsigned_long_long_value_t__int swig_types[46]
#define SWIGTYPE_p_f_p_struct_s_rig__int swig_types[47]
#define SWIGTYPE_p_f_p_struct_s_rig__p_char swig_types[48]
#define SWIGTYPE_p_f_p_struct_s_rig_enum_powerstat_t__int swig_types[49]
#define SWIGTYPE_p_f_p_struct_s_rig_enum_reset_t__int swig_types[50]
#define SWIGTYPE_p_f_p_struct_s_rig_int__int swig_types[51]
#define SWIGTYPE_p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int swig_types[52]
#define SWIGTYPE_p_f_p_struct_s_rig_long_p_char__int swig_types[53]
#define SWIGTYPE_p_f_p_struct_s_rig_long_p_char_int__int swig_types[54]
#define SWIGTYPE_p_f_p_struct_s_rig_long_p_q_const__char__int swig_types[55]
#define SWIGTYPE_p_f_p_struct_s_rig_long_p_value_t__int swig_types[56]
#define SWIGTYPE_p_f_p_struct_s_rig_long_value_t__int swig_types[57]
#define SWIGTYPE_p_f_p_struct_s_rig_p_enum_powerstat_t__int swig_types[58]
#define SWIGTYPE_p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int swig_types[59]
#define SWIGTYPE_p_f_p_struct_s_rig_p_int__int swig_types[60]
#define SWIGTYPE_p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int swig_types[61]
#define SWIGTYPE_p_f_p_struct_s_rig_p_q_const__char__int swig_types[62]
#define SWIGTYPE_p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int swig_types[63]
#define SWIGTYPE_p_f_p_struct_s_rig_p_unsigned_int__int swig_types[64]
#define SWIGTYPE_p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int swig_types[65]
#define SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int swig_types[66]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int swig_types[67]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double__int swig_types[68]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double_p_void__int swig_types[69]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int swig_types[70]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int swig_types[71]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int swig_types[72]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int swig_types[73]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int swig_types[74]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int swig_types[75]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int swig_types[76]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int swig_types[77]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int swig_types[78]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int swig_types[79]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_int__int swig_types[80]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_p_int__int swig_types[81]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int swig_types[82]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_value_t__int swig_types[83]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int swig_types[84]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double__int swig_types[85]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int swig_types[86]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int swig_types[87]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int swig_types[88]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int swig_types[89]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int swig_types[90]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int swig_types[91]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int swig_types[92]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int swig_types[93]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int swig_types[94]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_int__int swig_types[95]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int swig_types[96]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int swig_types[97]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int swig_types[98]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int swig_types[99]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int swig_types[100]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int swig_types[101]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_void__int swig_types[102]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int swig_types[103]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int swig_types[104]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int swig_types[105]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int swig_types[106]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int swig_types[107]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int swig_types[108]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int swig_types[109]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int swig_types[110]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int swig_types[111]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int swig_types[112]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_long_long_value_t__int swig_types[113]
#define SWIGTYPE_p_f_p_struct_s_rot__int swig_types[114]
#define SWIGTYPE_p_f_p_struct_s_rot__p_char swig_types[115]
#define SWIGTYPE_p_f_p_struct_s_rot_float_float__int swig_types[116]
#define SWIGTYPE_p_f_p_struct_s_rot_int__int swig_types[117]
#define SWIGTYPE_p_f_p_struct_s_rot_int_int__int swig_types[118]
#define SWIGTYPE_p_f_p_struct_s_rot_long_int__int swig_types[119]
#define SWIGTYPE_p_f_p_struct_s_rot_long_p_char__int swig_types[120]
#define SWIGTYPE_p_f_p_struct_s_rot_long_p_char_int__int swig_types[121]
#define SWIGTYPE_p_f_p_struct_s_rot_long_p_int__int swig_types[122]
#define SWIGTYPE_p_f_p_struct_s_rot_long_p_q_const__char__int swig_types[123]
#define SWIGTYPE_p_f_p_struct_s_rot_long_p_value_t__int swig_types[124]
#define SWIGTYPE_p_f_p_struct_s_rot_long_value_t__int swig_types[125]
#define SWIGTYPE_p_f_p_struct_s_rot_p_enum_rot_status_t__int swig_types[126]
#define SWIGTYPE_p_f_p_struct_s_rot_p_float_p_float__int swig_types[127]
#define SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_int__int swig_types[128]
#define SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_int__int swig_types[129]
#define SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int swig_types[130]
#define SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_value_t__int swig_types[131]
#define SWIGTYPE_p_filter_list swig_types[132]
#define SWIGTYPE_p_float swig_types[133]
#define SWIGTYPE_p_freq_range_list swig_types[134]
#define SWIGTYPE_p_gran swig_types[135]
#define SWIGTYPE_p_hamlib_async_pipe swig_types[136]
#define SWIGTYPE_p_hamlib_band_t swig_types[137]
#define SWIGTYPE_p_hamlib_bandselect_t swig_types[138]
#define SWIGTYPE_p_hamlib_cache_t swig_types[139]
#define SWIGTYPE_p_hamlib_port swig_types[140]
#define SWIGTYPE_p_int swig_types[141]
#define SWIGTYPE_p_long swig_types[142]
#define SWIGTYPE_p_long_long swig_types[143]
#define SWIGTYPE_p_multicast_s swig_types[144]
#define SWIGTYPE_p_p_char swig_types[145]
#define SWIGTYPE_p_p_void swig_types[146]
#define SWIGTYPE_p_powerstat_t swig_types[147]
#define SWIGTYPE_p_pthread_mutex_t swig_types[148]
#define SWIGTYPE_p_pthread_t swig_types[149]
#define SWIGTYPE_p_ptt_t swig_types[150]
#define SWIGTYPE_p_ptt_type_e swig_types[151]
#define SWIGTYPE_p_reset_t swig_types[152]
#define SWIGTYPE_p_rig_callbacks swig_types[153]
#define SWIGTYPE_p_rig_caps swig_types[154]
#define SWIGTYPE_p_rig_debug_level_e swig_types[155]
#define SWIGTYPE_p_rig_port_e swig_types[156]
#define SWIGTYPE_p_rig_ptrx_t swig_types[157]
#define SWIGTYPE_p_rig_spectrum_avg_mode swig_types[158]
#define SWIGTYPE_p_rig_spectrum_line swig_types[159]
#define SWIGTYPE_p_rig_spectrum_mode_e swig_types[160]
#define SWIGTYPE_p_rig_spectrum_scope swig_types[161]
#define SWIGTYPE_p_rig_state swig_types[162]
#define SWIGTYPE_p_rig_type_t swig_types[163]
#define SWIGTYPE_p_rot_caps swig_types[164]
#define SWIGTYPE_p_rot_state swig_types[165]
#define SWIGTYPE_p_rot_status_t swig_types[166]
#define SWIGTYPE_p_rot_type_t swig_types[167]
#define SWIGTYPE_p_rptr_shift_t swig_types[168]
#define SWIGTYPE_p_s_rig swig_types[169]
#define SWIGTYPE_p_s_rot swig_types[170]
#define SWIGTYPE_p_scan_t swig_types[171]
#define SWIGTYPE_p_short swig_types[172]
#define SWIGTYPE_p_signed_char swig_types[173]
#define SWIGTYPE_p_sockaddr_in swig_types[174]
#define SWIGTYPE_p_split_t swig_types[175]
#define SWIGTYPE_p_time_t swig_types[176]
#define SWIGTYPE_p_timespec swig_types[177]
#define SWIGTYPE_p_toneArray swig_types[178]
#define SWIGTYPE_p_tuning_step_list swig_types[179]
#define SWIGTYPE_p_twiddle_state_t swig_types[180]
#define SWIGTYPE_p_unsigned_char swig_types[181]
#define SWIGTYPE_p_unsigned_int swig_types[182]
#define SWIGTYPE_p_unsigned_long swig_types[183]
#define SWIGTYPE_p_unsigned_long_long swig_types[184]
#define SWIGTYPE_p_unsigned_short swig_types[185]
#define SWIGTYPE_p_value_t swig_types[186]
#define SWIGTYPE_p_value_t_b swig_types[187]
#define SWIGTYPE_p_vfo_op_t swig_types[188]
#define SWIGTYPE_p_void swig_types[189]
static swig_type_info *swig_types[191];
static swig_module_info swig_module = {swig_types, 190, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#ifdef SWIG_TypeQuery
# undef SWIG_TypeQuery
#endif
#define SWIG_TypeQuery SWIG_Python_TypeQuery

/*-----------------------------------------------
              @(target):= _Hamlib.so
  ------------------------------------------------*/
#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_init    PyInit__Hamlib

#else
#  define SWIG_init    init_Hamlib

#endif

#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


/*
 *  Hamlib bindings - swig interface file
 *  Copyright (c) 2001-2008 by Stephane Fillod
 *
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <hamlib/config.h>

//mdblack98 -- Commented this out -- is anybody using the functions in misc.h?
//If so, then those functions should be moved to rig.h
//#include "misc.h"

#include <limits.h>



#include <stdint.h>		// Use the C99 official header


typedef const char * const_char_string;


#include <hamlib/rig.h>


SWIGINTERNINLINE PyObject*
  SWIG_From_int  (int value)
{
  return PyInt_FromLong((long) value);
}


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


SWIGINTERN int
SWIG_AsVal_double (PyObject *obj, double *val)
{
  int res = SWIG_TypeError;
  if (PyFloat_Check(obj)) {
    if (val) *val = PyFloat_AsDouble(obj);
    return SWIG_OK;
#if PY_VERSION_HEX < 0x03000000
  } else if (PyInt_Check(obj)) {
    if (val) *val = (double) PyInt_AsLong(obj);
    return SWIG_OK;
#endif
  } else if (PyLong_Check(obj)) {
    double v = PyLong_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    double d = PyFloat_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = d;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      long v = PyLong_AsLong(obj);
      if (!PyErr_Occurred()) {
	if (val) *val = v;
	return SWIG_AddCast(SWIG_AddCast(SWIG_OK));
      } else {
	PyErr_Clear();
      }
    }
  }
#endif
  return res;
}


#include <errno.h>


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx, cx, rd;
   errno = 0;
   fx = floor(x);
   cx = ceil(x);
   rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_long (PyObject *obj, long* val)
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else
#endif
  if (PyLong_Check(obj)) {
    long v = PyLong_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    long v = PyInt_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      // Largest double not larger than LONG_MAX (not portably calculated easily)
      // Note that double(LONG_MAX) is stored in a double rounded up by one (for 64-bit long)
      // 0x7ffffffffffffc00LL == (int64_t)std::nextafter(double(__uint128_t(LONG_MAX)+1), double(0))
      const double long_max = sizeof(long) == 8 ? 0x7ffffffffffffc00LL : LONG_MAX;
      // No equivalent needed for 64-bit double(LONG_MIN) is exactly LONG_MIN
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, long_max)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_int (PyObject * obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (int)(v);
    }
  }  
  return res;
}


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static swig_type_info* info = 0;
  if (!info) {
    info = SWIG_TypeQuery("_p_char");
  }
  return info;
}


SWIGINTERNINLINE PyObject *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > (size_t)PY_SSIZE_T_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_InternalNewPointerObj((char *)(carray), pchar_descriptor, 0) : SWIG_Py_Void();
    } else {
#if PY_VERSION_HEX >= 0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
      return PyBytes_FromStringAndSize(carray, (Py_ssize_t)(size));
#else
      return PyUnicode_DecodeUTF8(carray, (Py_ssize_t)(size), "surrogateescape");
#endif
#else
      return PyString_FromStringAndSize(carray, (Py_ssize_t)(size));
#endif
    }
  } else {
    return SWIG_Py_Void();
  }
}


SWIGINTERNINLINE PyObject * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}


SWIGINTERNINLINE PyObject*
  SWIG_From_unsigned_SS_int  (unsigned int value)
{
  return PyInt_FromSize_t((size_t) value);
}


  #define SWIG_From_long   PyInt_FromLong 

typedef union {
  struct {
    float min;                           
    float max;                           
    float step;                 
  } n;                                    
  struct {
    const char *combostr[16];                      
  } c;                                  
} confparams_u;

typedef struct {
  float min;                           
  float max;                           
  float step;                 
} confparams_u_n;

typedef struct {
  const char *combostr[16];                      
} confparams_u_c;


/* Getting isfinite working pre C99 across multiple platforms is non-trivial. Users can provide SWIG_isfinite on older platforms. */
#ifndef SWIG_isfinite
/* isfinite() is a macro for C99 */
# if defined(isfinite)
#  define SWIG_isfinite(X) (isfinite(X))
# elif defined(__cplusplus) && __cplusplus >= 201103L
/* Use a template so that this works whether isfinite() is std::isfinite() or
 * in the global namespace.  The reality seems to vary between compiler
 * versions.
 *
 * Make sure namespace std exists to avoid compiler warnings.
 *
 * extern "C++" is required as this fragment can end up inside an extern "C" { } block
 */
namespace std { }
extern "C++" template<typename T>
inline int SWIG_isfinite_func(T x) {
  using namespace std;
  return isfinite(x);
}
#  define SWIG_isfinite(X) (SWIG_isfinite_func(X))
# elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 2))
#  define SWIG_isfinite(X) (__builtin_isfinite(X))
# elif defined(_MSC_VER)
#  define SWIG_isfinite(X) (_finite(X))
# elif defined(__sun) && defined(__SVR4)
#  include <ieeefp.h>
#  define SWIG_isfinite(X) (finite(X))
# endif
#endif


/* Accept infinite as a valid float value unless we are unable to check if a value is finite */
#ifdef SWIG_isfinite
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX) && SWIG_isfinite(X))
#else
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX))
#endif


SWIGINTERN int
SWIG_AsVal_float (PyObject * obj, float *val)
{
  double v;
  int res = SWIG_AsVal_double (obj, &v);
  if (SWIG_IsOK(res)) {
    if (SWIG_Float_Overflow_Check(v)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (float)(v);
    }
  }  
  return res;
}


  #define SWIG_From_double   PyFloat_FromDouble 


SWIGINTERNINLINE PyObject *
SWIG_From_float  (float value)
{    
  return SWIG_From_double  (value);
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (PyObject *obj, unsigned long *val) 
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    long v = PyInt_AsLong(obj);
    if (v >= 0) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else
#endif
  if (PyLong_Check(obj)) {
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      // Largest double not larger than ULONG_MAX (not portably calculated easily)
      // Note that double(ULONG_MAX) is stored in a double rounded up by one (for 64-bit unsigned long)
      // 0xfffffffffffff800ULL == (uint64_t)std::nextafter(double(__uint128_t(ULONG_MAX)+1), double(0))
      const double ulong_max = sizeof(unsigned long) == 8 ? 0xfffffffffffff800ULL : ULONG_MAX;
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ulong_max)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_int (PyObject * obj, unsigned int *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UINT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (unsigned int)(v);
    }
  }  
  return res;
}


/* Return string from Python obj. NOTE: obj must remain in scope in order
   to use the returned cptr (but only when alloc is set to SWIG_OLDOBJ) */
SWIGINTERN int
SWIG_AsCharPtrAndSize(PyObject *obj, char **cptr, size_t *psize, int *alloc)
{
#if PY_VERSION_HEX >= 0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
  if (PyBytes_Check(obj))
#else
  if (PyUnicode_Check(obj))
#endif
#else  
  if (PyString_Check(obj))
#endif
  {
    char *cstr; Py_ssize_t len;
    PyObject *bytes = NULL;
    int ret = SWIG_OK;
    if (alloc)
      *alloc = SWIG_OLDOBJ;
#if PY_VERSION_HEX >= 0x03000000 && defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    if (PyBytes_AsStringAndSize(obj, &cstr, &len) == -1)
      return SWIG_TypeError;
#else
    cstr = (char *)SWIG_PyUnicode_AsUTF8AndSize(obj, &len, &bytes);
    if (!cstr)
      return SWIG_TypeError;
    /* The returned string is only duplicated if the char * returned is not owned and memory managed by obj */
    if (bytes && cptr) {
      if (alloc) {
        cstr = (char *)memcpy(malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
        *alloc = SWIG_NEWOBJ;
      } else {
        /* alloc must be set in order to clean up allocated memory */
        return SWIG_RuntimeError;
      }
    }
#endif
    if (cptr) *cptr = cstr;
    if (psize) *psize = len + 1;
    SWIG_Py_XDECREF(bytes);
    return ret;
  } else {
#if defined(SWIG_PYTHON_2_UNICODE)
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
#error "Cannot use both SWIG_PYTHON_2_UNICODE and SWIG_PYTHON_STRICT_BYTE_CHAR at once"
#endif
#if PY_VERSION_HEX < 0x03000000
    if (PyUnicode_Check(obj)) {
      char *cstr; Py_ssize_t len;
      if (!alloc && cptr) {
        return SWIG_RuntimeError;
      }
      obj = PyUnicode_AsUTF8String(obj);
      if (!obj)
        return SWIG_TypeError;
      if (PyString_AsStringAndSize(obj, &cstr, &len) != -1) {
        if (cptr) {
          if (alloc) *alloc = SWIG_NEWOBJ;
          *cptr = (char *)memcpy(malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
        }
        if (psize) *psize = len + 1;

        SWIG_Py_XDECREF(obj);
        return SWIG_OK;
      } else {
        SWIG_Py_XDECREF(obj);
      }
    }
#endif
#endif

    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *) vptr;
	if (psize) *psize = vptr ? (strlen((char *)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}




typedef struct {
  int l;                                
  unsigned char *d;                             
} value_t_b;


#if defined(LLONG_MAX) && !defined(SWIG_LONG_LONG_AVAILABLE)
#  define SWIG_LONG_LONG_AVAILABLE
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long_SS_long  (unsigned long long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLongLong(value) : PyInt_FromLong((long)(value));
}
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN int
SWIG_AsVal_unsigned_SS_long_SS_long (PyObject *obj, unsigned long long *val)
{
  int res = SWIG_TypeError;
  if (PyLong_Check(obj)) {
    unsigned long long v = PyLong_AsUnsignedLongLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      res = SWIG_OverflowError;
    }
  } else {
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long (obj,&v);
    if (SWIG_IsOK(res)) {
      if (val) *val = v;
      return res;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    const double mant_max = 1LL << DBL_MANT_DIG;
    double d;
    res = SWIG_AsVal_double (obj,&d);
    if (SWIG_IsOK(res) && !SWIG_CanCastAsInteger(&d, 0, mant_max))
      return SWIG_OverflowError;
    if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, mant_max)) {
      if (val) *val = (unsigned long long)(d);
      return SWIG_AddCast(res);
    }
    res = SWIG_TypeError;
  }
#endif
  return res;
}
#endif


SWIGINTERN int
SWIG_AsCharArray(PyObject * obj, char *val, size_t size)
{ 
  char* cptr = 0; size_t csize = 0; int alloc = SWIG_OLDOBJ;
  int res = SWIG_AsCharPtrAndSize(obj, &cptr, &csize, &alloc);
  if (SWIG_IsOK(res)) {
    /* special case of single char conversion when we don't need space for NUL */
    if (size == 1 && csize == 2 && cptr && !cptr[1]) --csize;
    if (csize <= size) {
      if (val) {
	if (csize) memcpy(val, cptr, csize*sizeof(char));
	if (csize < size) memset(val + csize, 0, (size - csize)*sizeof(char));
      }
      if (alloc == SWIG_NEWOBJ) {
	free((char*)cptr);
	res = SWIG_DelNewMask(res);
      }      
      return res;
    }
    if (alloc == SWIG_NEWOBJ) free((char*)cptr);
  }
  return SWIG_TypeError;
}


SWIGINTERN size_t
SWIG_strnlen(const char* s, size_t maxlen)
{
  const char *p;
  for (p = s; maxlen-- && *p; p++)
    ;
  return p - s;
}

SWIGINTERN struct channel *new_channel(vfo_t vfo,int ch){
		channel_t *chan;

		chan = (channel_t*)malloc(sizeof(channel_t));
		if (!chan)
			return NULL;
		memset(chan, 0, sizeof(channel_t));
		chan->vfo = vfo;
		chan->channel_num = ch;
		return chan;
	}
SWIGINTERN void delete_channel(struct channel *self){
		/* TODO: free ext_levels */
		free(self);
	}
typedef struct {
  int raw;                                                                           
  int val;                                                                         
} cal_table_table;

typedef struct {
  int raw;                                         
  float val;                                        
} cal_table_float_table;


SWIGINTERNINLINE int
SWIG_AsVal_size_t (PyObject * obj, size_t *val)
{
  int res = SWIG_TypeError;
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = (size_t)(v);
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else if (sizeof(size_t) <= sizeof(unsigned long long)) {
    unsigned long long v;
    res = SWIG_AsVal_unsigned_SS_long_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = (size_t)(v);
  }
#endif
  return res;
}


SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLong(value) : PyInt_FromLong((long)(value));
}


SWIGINTERNINLINE PyObject *
SWIG_From_size_t  (size_t value)
{    
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    return SWIG_From_unsigned_SS_long  ((unsigned long)(value));
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else {
    /* assume sizeof(size_t) <= sizeof(unsigned long long) */
    return SWIG_From_unsigned_SS_long_SS_long  ((unsigned long long)(value));
  }
#endif
}


SWIGINTERN int
SWIG_AsVal_short (PyObject * obj, short *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < SHRT_MIN || v > SHRT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (short)(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject *
SWIG_From_short  (short value)
{    
  return SWIG_From_long  (value);
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_char (PyObject * obj, unsigned char *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UCHAR_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (unsigned char)(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject *
SWIG_From_unsigned_SS_char  (unsigned char value)
{    
  return SWIG_From_unsigned_SS_long  (value);
}



#pragma SWIG nowarn=451

#ifdef SWIGPYTHON
	// This mirrors "struct rig_callbacks" from rig.h using Python types
	typedef struct {
		PyObject *freq_event;     /*!< Frequency change event */
		PyObject *freq_arg;       /*!< Frequency change argument */
		PyObject *mode_event;     /*!< Mode change event */
		PyObject *mode_arg;       /*!< Mode change argument */
		PyObject *vfo_event;      /*!< VFO change event */
		PyObject *vfo_arg;        /*!< VFO change argument */
		PyObject *ptt_event;      /*!< PTT change event */
		PyObject *ptt_arg;        /*!< PTT change argument */
		PyObject *dcd_event;      /*!< DCD change event */
		PyObject *dcd_arg;        /*!< DCD change argument */
		PyObject *pltune_event;   /*!< Pipeline tuning module freq/mode/width callback NOTE: the C struct doesn't have the _event suffix*/
		PyObject *pltune_arg;     /*!< Pipeline tuning argument */
		PyObject *spectrum_event; /*!< Spectrum line reception event */
		PyObject *spectrum_arg;   /*!< Spectrum line reception argument */
		/* etc.. */
	} python_callbacks_t;
#endif

typedef struct Rig {
	RIG *rig;
	struct rig_caps *caps;		/* shortcut to RIG->caps */
	struct rig_state *state;	/* shortcut to RIG->state */
	int error_status;
	int do_exception;
#ifdef SWIGPYTHON
	python_callbacks_t *python_callbacks;
#endif
} Rig;

typedef char * char_string;
typedef channel_t * channel_t_p;
typedef channel_t * const_channel_t_p;

#ifdef SWIGPYTHON
int rig_freq_cb_python(RIG *rig, vfo_t vfo, freq_t freq, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(3,
                                    PyLong_FromLong(vfo),
                                    PyFloat_FromDouble(freq),
                                    self->python_callbacks->freq_arg
                                    );

    PyObject_CallObject(self->python_callbacks->freq_event, python_arguments);
    Py_XDECREF(python_arguments);

    return RIG_OK;
}

int rig_mode_cb_python(RIG *rig, vfo_t vfo, rmode_t rmode, pbwidth_t pbwidth, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(4,
                                    PyLong_FromLong(vfo),
                                    PyLong_FromLong(rmode),
                                    PyLong_FromLong(pbwidth),
                                    self->python_callbacks->mode_arg
                                    );

    PyObject_CallObject(self->python_callbacks->mode_event, python_arguments);

    Py_XDECREF(python_arguments);

    return RIG_OK;
}

int rig_vfo_cb_python(RIG *rig, vfo_t vfo, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(2,
                                    PyLong_FromLong(vfo),
                                    self->python_callbacks->vfo_arg
                                    );

    PyObject_CallObject(self->python_callbacks->vfo_event, python_arguments);

    Py_XDECREF(python_arguments);

    return RIG_OK;
}

int rig_ptt_cb_python(RIG *rig, vfo_t vfo, ptt_t ptt, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(3,
                                    PyLong_FromLong(vfo),
                                    PyLong_FromLong(ptt),
                                    self->python_callbacks->ptt_arg
                                    );

    PyObject_CallObject(self->python_callbacks->ptt_event, python_arguments);

    Py_XDECREF(python_arguments);

    return RIG_OK;
}

int rig_dcd_cb_python(RIG *rig, vfo_t vfo, dcd_t dcd, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(3,
                                    PyLong_FromLong(vfo),
                                    PyLong_FromLong(dcd),
                                    self->python_callbacks->dcd_arg
                                    );

    PyObject_CallObject(self->python_callbacks->dcd_event, python_arguments);

    Py_XDECREF(python_arguments);

    return RIG_OK;
}

int rig_pltune_cb_python(RIG *rig, vfo_t vfo, freq_t *freq, rmode_t *rmode, pbwidth_t *pbwidth, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(1,
                                    self->python_callbacks->pltune_arg
                                    );

    PyObject_CallObject(self->python_callbacks->pltune_event, python_arguments);

    Py_XDECREF(python_arguments);

    return RIG_OK;
}

int rig_spectrum_cb_python(RIG *rig, struct rig_spectrum_line *rig_spectrum_line, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(1,
                                    self->python_callbacks->spectrum_arg
                                    );

    PyObject_CallObject(self->python_callbacks->spectrum_event, python_arguments);

    Py_XDECREF(python_arguments);

    return RIG_OK;
}
#endif

SWIGINTERN struct Rig *new_Rig(int rig_model){
		Rig *r;

		r = (Rig*)malloc(sizeof(Rig));
		if (!r)
			return NULL;
		r->rig = rig_init(rig_model);
		if (!r->rig) {
			free(r);
			return NULL;
		}

		r->python_callbacks = calloc(1, sizeof(python_callbacks_t));
		if (!r->python_callbacks) {
			free(r);
			return NULL;
		}

		/* install shortcuts */
		r->caps = r->rig->caps;
		r->state = rig_data_pointer(r->rig, RIG_PTRX_STATE);
		r->do_exception = 0;	/* default is disabled */
		r->error_status = RIG_OK;

		return r;
	}
SWIGINTERN void delete_Rig(struct Rig *self){
		rig_cleanup(self->rig);

		free(self->python_callbacks);

		free(self);
	}
SWIGINTERN void Rig_open(struct Rig *self){
		self->error_status = rig_open(self->rig);
	}
SWIGINTERN void Rig_close(struct Rig *self){
		self->error_status = rig_close(self->rig);
	}
SWIGINTERN void Rig_set_freq(struct Rig *self,vfo_t vfo,freq_t _freq_t){ self->error_status = rig_set_freq(self->rig  ,vfo, _freq_t); }
SWIGINTERN void Rig_set_mode(struct Rig *self,rmode_t _rmode_t_1,pbwidth_t _pbwidth_t_2,vfo_t vfo){ self->error_status = rig_set_mode(self->rig  ,vfo, _rmode_t_1, _pbwidth_t_2); }
SWIGINTERN void Rig_set_ptt(struct Rig *self,vfo_t vfo,ptt_t _ptt_t){ self->error_status = rig_set_ptt(self->rig  ,vfo, _ptt_t); }
SWIGINTERN void Rig_set_rptr_shift(struct Rig *self,vfo_t vfo,rptr_shift_t _rptr_shift_t){ self->error_status = rig_set_rptr_shift(self->rig  ,vfo, _rptr_shift_t); }
SWIGINTERN void Rig_set_rptr_offs(struct Rig *self,vfo_t vfo,shortfreq_t _shortfreq_t){ self->error_status = rig_set_rptr_offs(self->rig  ,vfo, _shortfreq_t); }
SWIGINTERN void Rig_set_ctcss_tone(struct Rig *self,vfo_t vfo,tone_t _tone_t){ self->error_status = rig_set_ctcss_tone(self->rig  ,vfo, _tone_t); }
SWIGINTERN void Rig_set_dcs_code(struct Rig *self,vfo_t vfo,tone_t _tone_t){ self->error_status = rig_set_dcs_code(self->rig  ,vfo, _tone_t); }
SWIGINTERN void Rig_set_ctcss_sql(struct Rig *self,vfo_t vfo,tone_t _tone_t){ self->error_status = rig_set_ctcss_sql(self->rig  ,vfo, _tone_t); }
SWIGINTERN void Rig_set_dcs_sql(struct Rig *self,vfo_t vfo,tone_t _tone_t){ self->error_status = rig_set_dcs_sql(self->rig  ,vfo, _tone_t); }
SWIGINTERN void Rig_set_split_freq(struct Rig *self,vfo_t vfo,freq_t _freq_t){ self->error_status = rig_set_split_freq(self->rig  ,vfo, _freq_t); }
SWIGINTERN void Rig_set_split_mode(struct Rig *self,rmode_t _rmode_t_1,pbwidth_t _pbwidth_t_2,vfo_t vfo){ self->error_status = rig_set_split_mode(self->rig  ,vfo, _rmode_t_1, _pbwidth_t_2); }
SWIGINTERN void Rig_set_split_freq_mode(struct Rig *self,freq_t _freq_t_1,rmode_t _rmode_t_2,pbwidth_t _pbwidth_t_3,vfo_t vfo){ self->error_status = rig_set_split_freq_mode(self->rig  ,vfo, _freq_t_1, _rmode_t_2, _pbwidth_t_3); }
SWIGINTERN void Rig_set_split_vfo(struct Rig *self,split_t _split_t_1,vfo_t _vfo_t_2,vfo_t vfo){ self->error_status = rig_set_split_vfo(self->rig  ,vfo, _split_t_1, _vfo_t_2); }
SWIGINTERN void Rig_set_rit(struct Rig *self,vfo_t vfo,shortfreq_t _shortfreq_t){ self->error_status = rig_set_rit(self->rig  ,vfo, _shortfreq_t); }
SWIGINTERN void Rig_set_xit(struct Rig *self,vfo_t vfo,shortfreq_t _shortfreq_t){ self->error_status = rig_set_xit(self->rig  ,vfo, _shortfreq_t); }
SWIGINTERN void Rig_set_ts(struct Rig *self,vfo_t vfo,shortfreq_t _shortfreq_t){ self->error_status = rig_set_ts(self->rig  ,vfo, _shortfreq_t); }
SWIGINTERN void Rig_set_ant(struct Rig *self,ant_t _ant_t_1,value_t _value_t_2,vfo_t vfo){ self->error_status = rig_set_ant(self->rig  ,vfo, _ant_t_1, _value_t_2); }
SWIGINTERN void Rig_set_func(struct Rig *self,setting_t _setting_t_1,int _int_2,vfo_t vfo){ self->error_status = rig_set_func(self->rig  ,vfo, _setting_t_1, _int_2); }
SWIGINTERN void Rig_set_ext_func(struct Rig *self,hamlib_token_t _hamlib_token_t_1,int _int_2,vfo_t vfo){ self->error_status = rig_set_ext_func(self->rig  ,vfo, _hamlib_token_t_1, _int_2); }
SWIGINTERN void Rig_set_bank(struct Rig *self,vfo_t vfo,int _int){ self->error_status = rig_set_bank(self->rig  ,vfo, _int); }
SWIGINTERN void Rig_set_mem(struct Rig *self,vfo_t vfo,int _int){ self->error_status = rig_set_mem(self->rig  ,vfo, _int); }
SWIGINTERN void Rig_send_dtmf(struct Rig *self,vfo_t vfo,const_char_string _const_char_string){ self->error_status = rig_send_dtmf(self->rig  ,vfo, _const_char_string); }
SWIGINTERN void Rig_send_morse(struct Rig *self,vfo_t vfo,const_char_string _const_char_string){ self->error_status = rig_send_morse(self->rig  ,vfo, _const_char_string); }
SWIGINTERN void Rig_vfo_op(struct Rig *self,vfo_t vfo,vfo_op_t _vfo_op_t){ self->error_status = rig_vfo_op(self->rig  ,vfo, _vfo_op_t); }
SWIGINTERN void Rig_scan(struct Rig *self,scan_t _scan_t_1,int _int_2,vfo_t vfo){ self->error_status = rig_scan(self->rig  ,vfo, _scan_t_1, _int_2); }
SWIGINTERN void Rig_set_level__SWIG_0(struct Rig *self,setting_t stg,int i,vfo_t vfo){ value_t val = {0}; 	if (((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.i = i;	 		self->error_status = rig_set_level(self->rig  ,vfo , stg, val);	 	}
SWIGINTERN void Rig_set_level__SWIG_1(struct Rig *self,setting_t stg,float f,vfo_t vfo){ value_t val = {0}; 	if (!((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.f = f;	 		self->error_status = rig_set_level(self->rig  ,vfo , stg, val);	 	}
SWIGINTERN void Rig_set_level__SWIG_2(struct Rig *self,char const *name,int i,vfo_t vfo){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_level(name);	 		if (!rig_has_set_level(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return;	 			}  			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				val.i = i;	 				break;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				val.i = (int)i;	 				break;	 			case RIG_CONF_STRING:		 				self->error_status = -RIG_EINVAL;		 				return;	 			default:	 				self->error_status = -RIG_ECONF;	 				return;	 			}	 			self->error_status = rig_set_ext_level(self->rig  ,vfo, cfp->token, val);	 			return;	 		}	 		if (((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.i = i;	 		self->error_status = rig_set_level(self->rig  ,vfo , stg, val);	 	}
SWIGINTERN void Rig_set_level__SWIG_3(struct Rig *self,char const *name,float f,vfo_t vfo){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_level(name);	 		if (!rig_has_set_level(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return;	 			}  			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				val.f = f;	 				break;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				val.i = (int)f;	 				break;	 			case RIG_CONF_STRING:		 				self->error_status = -RIG_EINVAL;		 				return;	 			default:	 				self->error_status = -RIG_ECONF;	 				return;	 			}	 			self->error_status = rig_set_ext_level(self->rig  ,vfo, cfp->token, val);	 			return;	 		}	 		if (!((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.f = f;	 		self->error_status = rig_set_level(self->rig  ,vfo , stg, val);	 	}
SWIGINTERN void Rig_set_level__SWIG_4(struct Rig *self,char const *name,char const *s,vfo_t vfo){ value_t val = {0};    		const struct confparams *cfp;  		cfp = rig_ext_lookup(self->rig, name);  		if (!cfp) {	 			self->error_status = -RIG_EINVAL;	  			return;  		}  		if (cfp->type != RIG_CONF_STRING) {  			self->error_status = -RIG_EINVAL;	  			return;  		}  		val.cs = s;  		self->error_status = rig_set_ext_level(self->rig ,vfo, cfp->token, val);  	}
SWIGINTERN void Rig_set_ext_level(struct Rig *self,hamlib_token_t _hamlib_token_t_1,value_t _value_t_2,vfo_t vfo){ self->error_status = rig_set_ext_level(self->rig  ,vfo, _hamlib_token_t_1, _value_t_2); }
SWIGINTERN int Rig_get_level_i__SWIG_0(struct Rig *self,setting_t stg,vfo_t vfo){ value_t val = {0}; 	if (((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return val.i;			 		}	 		self->error_status = rig_get_level(self->rig  ,vfo , stg, &val);	 		return val.i;	 	}
SWIGINTERN float Rig_get_level_f__SWIG_0(struct Rig *self,setting_t stg,vfo_t vfo){ value_t val = {0}; 	if (!((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return val.f;			 		}	 		self->error_status = rig_get_level(self->rig  ,vfo , stg, &val);	 		return val.f;	 	}
SWIGINTERN int Rig_get_level_i__SWIG_1(struct Rig *self,char const *name,vfo_t vfo){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_level(name);	 		if (!rig_has_get_level(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return val.i;		  			}  			if (cfp->type == RIG_CONF_STRING) {  				self->error_status = -RIG_EINVAL;	 				return val.i;	  			}	 			self->error_status = rig_get_ext_level(self->rig  ,vfo, cfp->token, &val);	 			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				return (int)val.f;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				return (int)val.i;	 			default:	 				self->error_status = -RIG_ECONF;	 				return val.i;	  			}	 		}	 		if (((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return val.i;	  		}	 		self->error_status = rig_get_level(self->rig  ,vfo , stg, &val);	 		return val.i;	 	}
SWIGINTERN float Rig_get_level_f__SWIG_1(struct Rig *self,char const *name,vfo_t vfo){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_level(name);	 		if (!rig_has_get_level(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return val.f;		  			}  			if (cfp->type == RIG_CONF_STRING) {  				self->error_status = -RIG_EINVAL;	 				return val.f;	  			}	 			self->error_status = rig_get_ext_level(self->rig  ,vfo, cfp->token, &val);	 			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				return (float)val.f;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				return (float)val.i;	 			default:	 				self->error_status = -RIG_ECONF;	 				return val.f;	  			}	 		}	 		if (!((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return val.f;	  		}	 		self->error_status = rig_get_level(self->rig  ,vfo , stg, &val);	 		return val.f;	 	}
SWIGINTERN void Rig_get_level__SWIG_0(struct Rig *self,char const *name,char *returnstr,vfo_t vfo){ value_t val = {0};   		const struct confparams *cfp;  		returnstr[0] = '\0';  		cfp = rig_ext_lookup(self->rig, name);  		if (!cfp || cfp->type != RIG_CONF_STRING) {	 			self->error_status = -RIG_EINVAL;	  			return;  		}  		val.s = returnstr;  		self->error_status = rig_get_ext_level(self->rig ,vfo, cfp->token, &val);  	}
SWIGINTERN void Rig_set_vfo(struct Rig *self,vfo_t _vfo_t){ self->error_status = rig_set_vfo(self->rig  , _vfo_t); }
SWIGINTERN void Rig_set_powerstat(struct Rig *self,powerstat_t _powerstat_t){ self->error_status = rig_set_powerstat(self->rig  , _powerstat_t); }
SWIGINTERN void Rig_has_get_level(struct Rig *self,setting_t _setting_t){ self->error_status = rig_has_get_level(self->rig  , _setting_t); }
SWIGINTERN void Rig_has_get_parm(struct Rig *self,setting_t _setting_t){ self->error_status = rig_has_get_parm(self->rig  , _setting_t); }
SWIGINTERN void Rig_has_set_parm(struct Rig *self,setting_t _setting_t){ self->error_status = rig_has_set_parm(self->rig  , _setting_t); }
SWIGINTERN void Rig_has_get_func(struct Rig *self,setting_t _setting_t){ self->error_status = rig_has_get_func(self->rig  , _setting_t); }
SWIGINTERN void Rig_has_set_func(struct Rig *self,setting_t _setting_t){ self->error_status = rig_has_set_func(self->rig  , _setting_t); }
SWIGINTERN void Rig_reset(struct Rig *self,reset_t _reset_t){ self->error_status = rig_reset(self->rig  , _reset_t); }
SWIGINTERN void Rig_has_scan(struct Rig *self,scan_t _scan_t){ self->error_status = rig_has_scan(self->rig  , _scan_t); }
SWIGINTERN void Rig_has_vfo_op(struct Rig *self,vfo_op_t _vfo_op_t){ self->error_status = rig_has_vfo_op(self->rig  , _vfo_op_t); }
SWIGINTERN void Rig_passband_normal(struct Rig *self,rmode_t _rmode_t){ self->error_status = rig_passband_normal(self->rig  , _rmode_t); }
SWIGINTERN void Rig_passband_narrow(struct Rig *self,rmode_t _rmode_t){ self->error_status = rig_passband_narrow(self->rig  , _rmode_t); }
SWIGINTERN void Rig_passband_wide(struct Rig *self,rmode_t _rmode_t){ self->error_status = rig_passband_wide(self->rig  , _rmode_t); }
SWIGINTERN void Rig_set_vfo_opt(struct Rig *self,int _int){ self->error_status = rig_set_vfo_opt(self->rig  , _int); }
SWIGINTERN void Rig_ext_token_lookup(struct Rig *self,const_char_string _const_char_string){ self->error_status = rig_ext_token_lookup(self->rig  , _const_char_string); }
SWIGINTERN void Rig_token_lookup(struct Rig *self,const_char_string _const_char_string){ self->error_status = rig_token_lookup(self->rig  , _const_char_string); }
SWIGINTERN void Rig_set_conf__SWIG_0(struct Rig *self,hamlib_token_t _hamlib_token_t_1,const_char_string _const_char_string_2){ self->error_status = rig_set_conf(self->rig  , _hamlib_token_t_1, _const_char_string_2); }
SWIGINTERN void Rig_set_ext_parm(struct Rig *self,hamlib_token_t _hamlib_token_t_1,value_t _value_t_2){ self->error_status = rig_set_ext_parm(self->rig  , _hamlib_token_t_1, _value_t_2); }
SWIGINTERN void Rig_set_parm__SWIG_0(struct Rig *self,setting_t stg,int i){ value_t val = {0}; 	if (((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.i = i;	 		self->error_status = rig_set_parm(self->rig   , stg, val);	 	}
SWIGINTERN void Rig_set_parm__SWIG_1(struct Rig *self,setting_t stg,float f){ value_t val = {0}; 	if (!((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.f = f;	 		self->error_status = rig_set_parm(self->rig   , stg, val);	 	}
SWIGINTERN void Rig_set_parm__SWIG_2(struct Rig *self,char const *name,int i){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_parm(name);	 		if (!rig_has_set_parm(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return;	 			}  			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				val.i = i;	 				break;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				val.i = (int)i;	 				break;	 			case RIG_CONF_STRING:		 				self->error_status = -RIG_EINVAL;		 				return;	 			default:	 				self->error_status = -RIG_ECONF;	 				return;	 			}	 			self->error_status = rig_set_ext_parm(self->rig  , cfp->token, val);	 			return;	 		}	 		if (((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.i = i;	 		self->error_status = rig_set_parm(self->rig   , stg, val);	 	}
SWIGINTERN void Rig_set_parm__SWIG_3(struct Rig *self,char const *name,float f){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_parm(name);	 		if (!rig_has_set_parm(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return;	 			}  			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				val.f = f;	 				break;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				val.i = (int)f;	 				break;	 			case RIG_CONF_STRING:		 				self->error_status = -RIG_EINVAL;		 				return;	 			default:	 				self->error_status = -RIG_ECONF;	 				return;	 			}	 			self->error_status = rig_set_ext_parm(self->rig  , cfp->token, val);	 			return;	 		}	 		if (!((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.f = f;	 		self->error_status = rig_set_parm(self->rig   , stg, val);	 	}
SWIGINTERN void Rig_set_parm__SWIG_4(struct Rig *self,char const *name,char const *s){ value_t val = {0};    		const struct confparams *cfp;  		cfp = rig_ext_lookup(self->rig, name);  		if (!cfp) {	 			self->error_status = -RIG_EINVAL;	  			return;  		}  		if (cfp->type != RIG_CONF_STRING) {  			self->error_status = -RIG_EINVAL;	  			return;  		}  		val.cs = s;  		self->error_status = rig_set_ext_parm(self->rig , cfp->token, val);  	}
SWIGINTERN int Rig_get_parm_i__SWIG_0(struct Rig *self,setting_t stg){ value_t val = {0}; 	if (((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return val.i;			 		}	 		self->error_status = rig_get_parm(self->rig   , stg, &val);	 		return val.i;	 	}
SWIGINTERN float Rig_get_parm_f__SWIG_0(struct Rig *self,setting_t stg){ value_t val = {0}; 	if (!((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return val.f;			 		}	 		self->error_status = rig_get_parm(self->rig   , stg, &val);	 		return val.f;	 	}
SWIGINTERN int Rig_get_parm_i__SWIG_1(struct Rig *self,char const *name){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_parm(name);	 		if (!rig_has_get_parm(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return val.i;		  			}  			if (cfp->type == RIG_CONF_STRING) {  				self->error_status = -RIG_EINVAL;	 				return val.i;	  			}	 			self->error_status = rig_get_ext_parm(self->rig  , cfp->token, &val);	 			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				return (int)val.f;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				return (int)val.i;	 			default:	 				self->error_status = -RIG_ECONF;	 				return val.i;	  			}	 		}	 		if (((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return val.i;	  		}	 		self->error_status = rig_get_parm(self->rig   , stg, &val);	 		return val.i;	 	}
SWIGINTERN float Rig_get_parm_f__SWIG_1(struct Rig *self,char const *name){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_parm(name);	 		if (!rig_has_get_parm(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return val.f;		  			}  			if (cfp->type == RIG_CONF_STRING) {  				self->error_status = -RIG_EINVAL;	 				return val.f;	  			}	 			self->error_status = rig_get_ext_parm(self->rig  , cfp->token, &val);	 			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				return (float)val.f;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				return (float)val.i;	 			default:	 				self->error_status = -RIG_ECONF;	 				return val.f;	  			}	 		}	 		if (!((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return val.f;	  		}	 		self->error_status = rig_get_parm(self->rig   , stg, &val);	 		return val.f;	 	}
SWIGINTERN void Rig_get_parm__SWIG_0(struct Rig *self,char const *name,char *returnstr){ value_t val = {0};   		const struct confparams *cfp;  		returnstr[0] = '\0';  		cfp = rig_ext_lookup(self->rig, name);  		if (!cfp || cfp->type != RIG_CONF_STRING) {	 			self->error_status = -RIG_EINVAL;	  			return;  		}  		val.s = returnstr;  		self->error_status = rig_get_ext_parm(self->rig , cfp->token, &val);  	}
SWIGINTERN void Rig_set_conf__SWIG_1(struct Rig *self,char const *name,char const *val){
		hamlib_token_t tok = rig_token_lookup(self->rig, name);
		if (tok == 0)
			self->error_status = -RIG_EINVAL;
		else
			self->error_status = rig_set_conf(self->rig, tok, val);
	}
SWIGINTERN freq_t Rig_get_freq(struct Rig *self,vfo_t vfo){ freq_t _freq_t; self->error_status = rig_get_freq(self->rig, vfo, &_freq_t); return _freq_t; }
SWIGINTERN vfo_t Rig_get_vfo(struct Rig *self){ vfo_t _vfo_t; self->error_status = rig_get_vfo(self->rig, &_vfo_t); return _vfo_t; }
SWIGINTERN ptt_t Rig_get_ptt(struct Rig *self,vfo_t vfo){ ptt_t _ptt_t; self->error_status = rig_get_ptt(self->rig, vfo, &_ptt_t); return _ptt_t; }
SWIGINTERN rptr_shift_t Rig_get_rptr_shift(struct Rig *self,vfo_t vfo){ rptr_shift_t _rptr_shift_t; self->error_status = rig_get_rptr_shift(self->rig, vfo, &_rptr_shift_t); return _rptr_shift_t; }
SWIGINTERN shortfreq_t Rig_get_rptr_offs(struct Rig *self,vfo_t vfo){ shortfreq_t _shortfreq_t; self->error_status = rig_get_rptr_offs(self->rig, vfo, &_shortfreq_t); return _shortfreq_t; }
SWIGINTERN tone_t Rig_get_ctcss_tone(struct Rig *self,vfo_t vfo){ tone_t _tone_t; self->error_status = rig_get_ctcss_tone(self->rig, vfo, &_tone_t); return _tone_t; }
SWIGINTERN tone_t Rig_get_dcs_code(struct Rig *self,vfo_t vfo){ tone_t _tone_t; self->error_status = rig_get_dcs_code(self->rig, vfo, &_tone_t); return _tone_t; }
SWIGINTERN tone_t Rig_get_ctcss_sql(struct Rig *self,vfo_t vfo){ tone_t _tone_t; self->error_status = rig_get_ctcss_sql(self->rig, vfo, &_tone_t); return _tone_t; }
SWIGINTERN tone_t Rig_get_dcs_sql(struct Rig *self,vfo_t vfo){ tone_t _tone_t; self->error_status = rig_get_dcs_sql(self->rig, vfo, &_tone_t); return _tone_t; }
SWIGINTERN freq_t Rig_get_split_freq(struct Rig *self,vfo_t vfo){ freq_t _freq_t; self->error_status = rig_get_split_freq(self->rig, vfo, &_freq_t); return _freq_t; }
SWIGINTERN void Rig_get_split_vfo(struct Rig *self,split_t *split,vfo_t *tx_vfo,vfo_t vfo){ self->error_status = rig_get_split_vfo(self->rig, vfo, split, tx_vfo); }
SWIGINTERN shortfreq_t Rig_get_rit(struct Rig *self,vfo_t vfo){ shortfreq_t _shortfreq_t; self->error_status = rig_get_rit(self->rig, vfo, &_shortfreq_t); return _shortfreq_t; }
SWIGINTERN shortfreq_t Rig_get_xit(struct Rig *self,vfo_t vfo){ shortfreq_t _shortfreq_t; self->error_status = rig_get_xit(self->rig, vfo, &_shortfreq_t); return _shortfreq_t; }
SWIGINTERN shortfreq_t Rig_get_ts(struct Rig *self,vfo_t vfo){ shortfreq_t _shortfreq_t; self->error_status = rig_get_ts(self->rig, vfo, &_shortfreq_t); return _shortfreq_t; }
SWIGINTERN void Rig_get_vfo_info(struct Rig *self,int *satmode,split_t *split,pbwidth_t *width,rmode_t *mode,freq_t *freq,vfo_t vfo){ self->error_status = rig_get_vfo_info(self->rig, vfo, freq, mode, width, split, satmode); }
SWIGINTERN int Rig_get_mem(struct Rig *self,vfo_t vfo){ int _int; self->error_status = rig_get_mem(self->rig, vfo, &_int); return _int; }
SWIGINTERN powerstat_t Rig_get_powerstat(struct Rig *self){ powerstat_t _powerstat_t; self->error_status = rig_get_powerstat(self->rig, &_powerstat_t); return _powerstat_t; }
SWIGINTERN dcd_t Rig_get_dcd(struct Rig *self,vfo_t vfo){ dcd_t _dcd_t; self->error_status = rig_get_dcd(self->rig, vfo, &_dcd_t); return _dcd_t; }
SWIGINTERN void Rig_set_freq_callback(struct Rig *self,PyObject *cb,PyObject *arg){
	void *callback;

	if (cb == Py_None) {
		callback = NULL;
	} else if (!PyCallable_Check(cb)) {
		SWIG_Python_RaiseOrModifyTypeError("The callback isn't a Python callable object");
		return;
	} else {
		callback = rig_freq_cb_python;
	};

	Py_XINCREF(cb);
	Py_XDECREF(self->python_callbacks->freq_event);

	Py_XINCREF(arg);
	Py_XDECREF(self->python_callbacks->freq_arg);

	self->python_callbacks->freq_event = cb;
	self->python_callbacks->freq_arg = arg;
	self->error_status = rig_set_freq_callback(self->rig, callback, self);

	return;
}
SWIGINTERN void Rig_set_mode_callback(struct Rig *self,PyObject *cb,PyObject *arg){
	void *callback;

	if (cb == Py_None) {
		callback = NULL;
	} else if (!PyCallable_Check(cb)) {
		SWIG_Python_RaiseOrModifyTypeError("The callback isn't a Python callable object");
		return;
	} else {
		callback = rig_mode_cb_python;
	};

	Py_XINCREF(cb);
	Py_XDECREF(self->python_callbacks->mode_event);

	Py_XINCREF(arg);
	Py_XDECREF(self->python_callbacks->mode_arg);

	self->python_callbacks->mode_event = cb;
	self->python_callbacks->mode_arg = arg;
	self->error_status = rig_set_mode_callback(self->rig, callback, self);

	return;
}
SWIGINTERN void Rig_set_vfo_callback(struct Rig *self,PyObject *cb,PyObject *arg){
	void *callback;

	if (cb == Py_None) {
		callback = NULL;
	} else if (!PyCallable_Check(cb)) {
		SWIG_Python_RaiseOrModifyTypeError("The callback isn't a Python callable object");
		return;
	} else {
		callback = rig_vfo_cb_python;
	};

	Py_XINCREF(cb);
	Py_XDECREF(self->python_callbacks->vfo_event);

	Py_XINCREF(arg);
	Py_XDECREF(self->python_callbacks->vfo_arg);

	self->python_callbacks->vfo_event = cb;
	self->python_callbacks->vfo_arg = arg;
	self->error_status = rig_set_vfo_callback(self->rig, callback, self);

	return;
}
SWIGINTERN void Rig_set_ptt_callback(struct Rig *self,PyObject *cb,PyObject *arg){
	void *callback;

	if (cb == Py_None) {
		callback = NULL;
	} else if (!PyCallable_Check(cb)) {
		SWIG_Python_RaiseOrModifyTypeError("The callback isn't a Python callable object");
		return;
	} else {
		callback = rig_ptt_cb_python;
	};

	Py_XINCREF(cb);
	Py_XDECREF(self->python_callbacks->ptt_event);

	Py_XINCREF(arg);
	Py_XDECREF(self->python_callbacks->ptt_arg);

	self->python_callbacks->ptt_event = cb;
	self->python_callbacks->ptt_arg = arg;
	self->error_status = rig_set_ptt_callback(self->rig, callback, self);

	return;
}
SWIGINTERN void Rig_set_dcd_callback(struct Rig *self,PyObject *cb,PyObject *arg){
	void *callback;

	if (cb == Py_None) {
		callback = NULL;
	} else if (!PyCallable_Check(cb)) {
		SWIG_Python_RaiseOrModifyTypeError("The callback isn't a Python callable object");
		return;
	} else {
		callback = rig_dcd_cb_python;
	};

	Py_XINCREF(cb);
	Py_XDECREF(self->python_callbacks->dcd_event);

	Py_XINCREF(arg);
	Py_XDECREF(self->python_callbacks->dcd_arg);

	self->python_callbacks->dcd_event = cb;
	self->python_callbacks->dcd_arg = arg;
	self->error_status = rig_set_dcd_callback(self->rig, callback, self);

	return;
}
SWIGINTERN void Rig_set_pltune_callback(struct Rig *self,PyObject *cb,PyObject *arg){
	void *callback;

	if (cb == Py_None) {
		callback = NULL;
	} else if (!PyCallable_Check(cb)) {
		SWIG_Python_RaiseOrModifyTypeError("The callback isn't a Python callable object");
		return;
	} else {
		callback = rig_pltune_cb_python;
	};

	Py_XINCREF(cb);
	Py_XDECREF(self->python_callbacks->pltune_event);

	Py_XINCREF(arg);
	Py_XDECREF(self->python_callbacks->pltune_arg);

	self->python_callbacks->pltune_event = cb;
	self->python_callbacks->pltune_arg = arg;
	self->error_status = rig_set_pltune_callback(self->rig, callback, self);

	return;
}
SWIGINTERN void Rig_set_spectrum_callback(struct Rig *self,PyObject *cb,PyObject *arg){
	void *callback;

	if (cb == Py_None) {
		callback = NULL;
	} else if (!PyCallable_Check(cb)) {
		SWIG_Python_RaiseOrModifyTypeError("The callback isn't a Python callable object");
		return;
	} else {
		callback = rig_spectrum_cb_python;
	};

	Py_XINCREF(cb);
	Py_XDECREF(self->python_callbacks->spectrum_event);

	Py_XINCREF(arg);
	Py_XDECREF(self->python_callbacks->spectrum_arg);

	self->python_callbacks->spectrum_event = cb;
	self->python_callbacks->spectrum_arg = arg;
	self->error_status = rig_set_spectrum_callback(self->rig, callback, self);

	return;
}
SWIGINTERN int Rig_mem_count(struct Rig *self){
		return rig_mem_count(self->rig);
	}
SWIGINTERN chan_t const *Rig_lookup_mem_caps(struct Rig *self,int channel_num){
		return rig_lookup_mem_caps(self->rig, channel_num);
	}
SWIGINTERN void Rig_set_channel(struct Rig *self,struct channel const *chan){
		self->error_status = rig_set_channel(self->rig, 0, chan);
	}
SWIGINTERN void Rig_get_channel__SWIG_0(struct Rig *self,struct channel *chan,int read_only){
		self->error_status = rig_get_channel(self->rig, 0, chan, read_only);
		/* TODO: handle ext_level's */
	}
SWIGINTERN void Rig_get_chan_all__SWIG_0(struct Rig *self,struct channel *chans){
		self->error_status = rig_get_chan_all(self->rig, 0, chans);
		/* TODO: handle ext_level's */
	}
SWIGINTERN struct channel *Rig_get_channel__SWIG_1(struct Rig *self,int read_only,int channel_num,vfo_t vfo){
		struct channel *chan;
		chan = new_channel(channel_num != INT_MAX ? vfo : (1u<<(29)), channel_num);
		if (!chan)
		{
			self->error_status = -RIG_ENOMEM;
			return NULL;
		}
		self->error_status = rig_get_channel(self->rig, 0, chan, read_only);
		/* TODO: copy ext_level's */
		return chan;
	}
SWIGINTERN void Rig_get_conf__SWIG_0(struct Rig *self,hamlib_token_t tok,char *returnstr){
		returnstr[0] = '\0';
		self->error_status = rig_get_conf2(self->rig, tok, returnstr, 256);
	}
SWIGINTERN void Rig_get_conf__SWIG_1(struct Rig *self,char const *name,char *returnstr){
		returnstr[0] = '\0';
		hamlib_token_t tok = rig_token_lookup(self->rig, name);
		if (tok == 0)
			self->error_status = -RIG_EINVAL;
		else
			self->error_status = rig_get_conf2(self->rig, tok, returnstr, 256);
	}
SWIGINTERN void Rig_recv_dtmf(struct Rig *self,char *returnstr,vfo_t vfo){
		int len = 256;
		self->error_status = rig_recv_dtmf(self->rig, vfo, returnstr, &len);
		returnstr[len] = '\0';
	}
SWIGINTERN char const *Rig_get_info(struct Rig *self){
		const char *s;
		s = rig_get_info(self->rig);
		self->error_status = s ? RIG_OK : -RIG_EINVAL;
		return s;
	}
SWIGINTERN int Rig_get_func(struct Rig *self,setting_t func,vfo_t vfo){
		int status;
		self->error_status = rig_get_func(self->rig, vfo, func, &status);
		return status;
	}
SWIGINTERN int Rig_get_ext_func(struct Rig *self,hamlib_token_t func,vfo_t vfo){
		int status;
		self->error_status = rig_get_ext_func(self->rig, vfo, func, &status);
		return status;
	}
SWIGINTERN PyObject *Rig_send_raw(struct Rig *self,PyObject *send_obj,PyObject *term_obj){
		char *send, *term;
		PyObject *bytes_obj;
		Py_ssize_t send_len;
		int reply_len = 256;
		char reply_buffer[256];
		int count;

		if (PyUnicode_Check(send_obj)) {
			bytes_obj = PyUnicode_AsUTF8String(send_obj);
			PyBytes_AsStringAndSize(bytes_obj, &send, &send_len);
			Py_XDECREF(bytes_obj);
		} else if (PyBytes_Check(send_obj)) {
			PyBytes_AsStringAndSize(send_obj, &send, &send_len);
		} else {
			SWIG_Python_RaiseOrModifyTypeError("Expected string or bytes for send argument");
			return NULL;
		}

		// Using NULL for length in PyUnicode_AsUTF8AndSize() and PyBytes_AsStringAndSize()
		// because we can't accept '\0' because there is no length for term in rig_send_raw()
		if (PyUnicode_Check(term_obj)) {
			bytes_obj = PyUnicode_AsUTF8String(term_obj);
			PyBytes_AsStringAndSize(bytes_obj, &term, NULL);
			Py_XDECREF(bytes_obj);
		} else if (PyBytes_Check(term_obj)) {
			PyBytes_AsStringAndSize(term_obj, &term, NULL);
		} else if (term_obj == Py_None) {
			term = NULL;
		} else {
			SWIG_Python_RaiseOrModifyTypeError("Expected string or bytes or NULL for term argument");
			return NULL;
		}

		count = rig_send_raw(self->rig, (unsigned char *)send, send_len, (unsigned char *)reply_buffer, reply_len, (unsigned char *)term);
		self->error_status = count < 0 ? count : RIG_OK;

		if (PyUnicode_Check(send_obj)) {
			return PyUnicode_FromStringAndSize(reply_buffer, count);
		} else {
			return PyBytes_FromStringAndSize(reply_buffer, count);
		}
	}
SWIGINTERN void Rig_get_level__SWIG_1(struct Rig *self,setting_t level,vfo_t vfo){ value_t val = {0}; self->error_status = rig_get_level(self->rig, vfo, level, &val);
			//if (RIG_LEVEL_IS_FLOAT(level))
			/* TODO: dynamic casting */
		}
SWIGINTERN void Rig_get_ext_level(struct Rig *self,hamlib_token_t token,vfo_t vfo){ value_t val = {0}; self->error_status = rig_get_level(self->rig, vfo, token, &val);
			//if (RIG_LEVEL_IS_FLOAT(level))
			/* TODO: dynamic casting */
		}
SWIGINTERN void Rig_get_parm__SWIG_1(struct Rig *self,setting_t parm){ value_t val = {0}; self->error_status = rig_get_parm(self->rig, parm, &val);
			//if (RIG_LEVEL_IS_FLOAT(parm))
			/* TODO: dynamic casting */
		}
SWIGINTERN void Rig_get_ext_parm(struct Rig *self,hamlib_token_t parm){ value_t val = {0}; self->error_status = rig_get_parm(self->rig, parm, &val);
			//if (RIG_LEVEL_IS_FLOAT(parm))
			/* TODO: dynamic casting */
		}

typedef struct channel channelArray;

SWIGINTERN channelArray *new_channelArray(size_t nelements){
    return (struct channel *)calloc(nelements, sizeof(struct channel));
  }
SWIGINTERN void delete_channelArray(channelArray *self){
    free((char*)self);
  }
SWIGINTERN struct channel *channelArray_cast(channelArray *self){
    return self;
  }
SWIGINTERN channelArray *channelArray_frompointer(struct channel *t){
    return (channelArray *)(t);
  }

typedef tone_t toneArray;

SWIGINTERN toneArray *new_toneArray(size_t nelements){
    return (tone_t *)calloc(nelements, sizeof(tone_t));
  }
SWIGINTERN void delete_toneArray(toneArray *self){
    free((char*)self);
  }
SWIGINTERN tone_t *toneArray_cast(toneArray *self){
    return self;
  }
SWIGINTERN toneArray *toneArray_frompointer(tone_t *t){
    return (toneArray *)(t);
  }


/*
 * these ones return 2 values, here is a perl example:
 * 	($mode, $width) = $rig->get_mode();
 */
void Rig_get_mode(Rig *self, rmode_t *mode, pbwidth_t *width, vfo_t vfo)
{
	self->error_status = rig_get_mode(self->rig, vfo, mode, width);
}

void Rig_get_split_mode(Rig *self, rmode_t *mode, pbwidth_t *width, vfo_t vfo)
{
	self->error_status = rig_get_split_mode(self->rig, vfo, mode, width);
}

/*
 * these ones return 3 values
 */
void Rig_get_split_freq_mode(Rig *self, vfo_t vfo, freq_t *tx_freq, rmode_t *tx_mode, pbwidth_t *tx_width)
{
	self->error_status = rig_get_split_freq_mode(self->rig, vfo, tx_freq, tx_mode, tx_width);
}


/*
 * these ones return 4 values
 */
void Rig_get_ant(Rig *self, ant_t *ant_rx, ant_t *ant_tx, ant_t *ant_curr, signed int *option, ant_t ant, vfo_t vfo)
{
	value_t value;
	self->error_status = rig_get_ant(self->rig, vfo, ant, &value, ant_curr, ant_tx, ant_rx);
	*option = value.i;
}

struct channel *Rig_get_chan_all(Rig *self)
{
	struct channel *chans;
	int nb_chans = rig_mem_count(self->rig);

	/* TODO: memleak: typemap to release that memory */
	chans = calloc(sizeof (struct channel), nb_chans);
	if (!chans)
	{
		self->error_status = -RIG_ENOMEM;
		return NULL;
	}
	self->error_status = rig_get_chan_all(self->rig, RIG_VFO_NONE, chans);
	/* TODO: copy ext_level's */
	return chans;
}



#include <hamlib/rotator.h>



typedef struct Rot {
	ROT *rot;
	struct rot_caps *caps;		/* shortcut to ROT->caps */
	struct rot_state *state;	/* shortcut to ROT->state */
	int error_status;
	int do_exception;
} Rot;


SWIGINTERN struct Rot *new_Rot(rot_model_t rot_model){
		Rot *r;

		r = (Rot*)malloc(sizeof(Rot));
		if (!r)
			return NULL;
		r->rot = rot_init(rot_model);
		if (!r->rot) {
			free(r);
			return NULL;
		}
		/* install shortcuts */
		r->caps = r->rot->caps;
		r->state = &r->rot->state; rot_data_pointer(r->rot, RIG_PTRX_ROTSTATE);
		r->do_exception = 0;    /* default is disabled */
		r->error_status = RIG_OK;
		return r;
	}
SWIGINTERN void delete_Rot(struct Rot *self){
		rot_cleanup(self->rot);
		free(self);
	}
SWIGINTERN void Rot_open(struct Rot *self){ self->error_status = rot_open(self->rot); }
SWIGINTERN void Rot_close(struct Rot *self){ self->error_status = rot_close(self->rot); }
SWIGINTERN void Rot_set_ext_func(struct Rot *self,hamlib_token_t _hamlib_token_t_1,int _int_2){ self->error_status = rot_set_ext_func(self->rot, _hamlib_token_t_1, _int_2); }
SWIGINTERN void Rot_set_ext_level(struct Rot *self,hamlib_token_t _hamlib_token_t_1,value_t _value_t_2){ self->error_status = rot_set_ext_level(self->rot, _hamlib_token_t_1, _value_t_2); }
SWIGINTERN void Rot_set_ext_parm(struct Rot *self,hamlib_token_t _hamlib_token_t_1,value_t _value_t_2){ self->error_status = rot_set_ext_parm(self->rot, _hamlib_token_t_1, _value_t_2); }
SWIGINTERN void Rot_set_func(struct Rot *self,setting_t _setting_t_1,int _int_2){ self->error_status = rot_set_func(self->rot, _setting_t_1, _int_2); }
SWIGINTERN void Rot_set_level(struct Rot *self,setting_t _setting_t_1,value_t _value_t_2){ self->error_status = rot_set_level(self->rot, _setting_t_1, _value_t_2); }
SWIGINTERN void Rot_set_parm(struct Rot *self,setting_t _setting_t_1,value_t _value_t_2){ self->error_status = rot_set_parm(self->rot, _setting_t_1, _value_t_2); }
SWIGINTERN void Rot_set_position(struct Rot *self,azimuth_t _azimuth_t_1,elevation_t _elevation_t_2){ self->error_status = rot_set_position(self->rot, _azimuth_t_1, _elevation_t_2); }
SWIGINTERN void Rot_stop(struct Rot *self){ self->error_status = rot_stop(self->rot); }
SWIGINTERN void Rot_park(struct Rot *self){ self->error_status = rot_park(self->rot); }
SWIGINTERN void Rot_reset(struct Rot *self,rot_reset_t _rot_reset_t){ self->error_status = rot_reset(self->rot, _rot_reset_t); }
SWIGINTERN void Rot_move(struct Rot *self,int _int_1,int _int_2){ self->error_status = rot_move(self->rot, _int_1, _int_2); }
SWIGINTERN void Rot_token_lookup(struct Rot *self,const_char_string _const_char_string){ self->error_status = rot_token_lookup(self->rot, _const_char_string); }
SWIGINTERN void Rot_set_conf__SWIG_0(struct Rot *self,char const *name,char const *val){
		hamlib_token_t tok = rot_token_lookup(self->rot, name);
		if (tok == 0)
			self->error_status = -RIG_EINVAL;
		else
			self->error_status = rot_set_conf(self->rot, tok, val);
	}
SWIGINTERN void Rot_set_conf__SWIG_1(struct Rot *self,hamlib_token_t _hamlib_token_t_1,const_char_string _const_char_string_2){ self->error_status = rot_set_conf(self->rot, _hamlib_token_t_1, _const_char_string_2); }
SWIGINTERN setting_t Rot_has_get_func(struct Rot *self,setting_t func){
		return rot_has_get_func(self->rot, func);
	}
SWIGINTERN setting_t Rot_has_set_func(struct Rot *self,setting_t func){
		return rot_has_set_func(self->rot, func);
	}
SWIGINTERN setting_t Rot_has_get_level(struct Rot *self,setting_t level){
		return rot_has_get_level(self->rot, level);
	}
SWIGINTERN setting_t Rot_has_set_level(struct Rot *self,setting_t level){
		return rot_has_set_level(self->rot, level);
	}
SWIGINTERN setting_t Rot_has_get_parm(struct Rot *self,setting_t parm){
		return rot_has_get_parm(self->rot, parm);
	}
SWIGINTERN setting_t Rot_has_set_parm(struct Rot *self,setting_t parm){
		return rot_has_set_parm(self->rot, parm);
	}
SWIGINTERN PyObject *Rot_get_func(struct Rot *self,hamlib_token_t token){
	int value;

	self->error_status =rot_get_func(self->rot, token, &value);
	if (self->error_status != RIG_OK)
		return Py_None;

	return PyLong_FromLong(value);
}
SWIGINTERN PyObject *Rot_get_ext_func(struct Rot *self,hamlib_token_t token){
	int value;

	self->error_status =rot_get_ext_func(self->rot, token, &value);
	if (self->error_status != RIG_OK)
		return Py_None;

	return PyLong_FromLong(value);
}
SWIGINTERN PyObject *Rot_get_level(struct Rot *self,hamlib_token_t token){
	value_t value;

	self->error_status =rot_get_level(self->rot, token, &value);
	if (self->error_status != RIG_OK)
		return Py_None;











	return PyLong_FromLong(value.i);
}
SWIGINTERN PyObject *Rot_get_ext_level(struct Rot *self,hamlib_token_t token){
	value_t value;

	self->error_status =rot_get_ext_level(self->rot, token, &value);
	if (self->error_status != RIG_OK)
		return Py_None;











	return PyLong_FromLong(value.i);
}
SWIGINTERN PyObject *Rot_get_parm(struct Rot *self,hamlib_token_t token){
	value_t value;

	self->error_status =rot_get_parm(self->rot, token, &value);
	if (self->error_status != RIG_OK)
		return Py_None;











	return PyLong_FromLong(value.i);
}
SWIGINTERN PyObject *Rot_get_ext_parm(struct Rot *self,hamlib_token_t token){
	value_t value;

	self->error_status =rot_get_ext_parm(self->rot, token, &value);
	if (self->error_status != RIG_OK)
		return Py_None;











	return PyLong_FromLong(value.i);
}
SWIGINTERN char const *Rot_get_conf__SWIG_0(struct Rot *self,hamlib_token_t tok){
                static char s[128] = "";
                self->error_status = rot_get_conf2(self->rot, tok, s, sizeof s);
                return s;
        }
SWIGINTERN char const *Rot_get_conf__SWIG_1(struct Rot *self,char const *name){
                hamlib_token_t tok = rot_token_lookup(self->rot, name);
                static char s[128] = "";
                if (tok == 0)
                        self->error_status = -RIG_EINVAL;
                else
                        self->error_status = rot_get_conf2(self->rot, tok, s, sizeof s);
                return s;
        }
SWIGINTERN char const *Rot_get_info(struct Rot *self){
                const char *s;
                s = rot_get_info(self->rot);
                self->error_status = s ? RIG_OK : -RIG_EINVAL;
                return s;
        }


/*
 * this one returns 2 values, here is a perl example:
 *      ($az, $elevation) = $rig->get_position();
 */
void Rot_get_position(Rot *self, azimuth_t *azimuth, elevation_t *elevation)
{
        self->error_status = rot_get_position(self->rot, azimuth, elevation);
}



#include <hamlib/amplifier.h>



typedef struct Amp {
	AMP *amp;
	struct amp_caps *caps;		/* shortcut to AMP->caps */
	struct amp_state *state;	/* shortcut to AMP->state */
	int error_status;
	int do_exception;
} Amp;


SWIGINTERN struct Amp *new_Amp(amp_model_t amp_model){
		Amp *r;

		r = (Amp*)malloc(sizeof(Amp));
		if (!r)
			return NULL;
		r->amp = amp_init(amp_model);
		if (!r->amp) {
			free(r);
			return NULL;
		}
		/* install shortcuts */
		r->caps = r->amp->caps;
		r->state = amp_data_pointer(r->amp, RIG_PTRX_AMPSTATE);
		r->do_exception = 0;    /* default is disabled */
		r->error_status = RIG_OK;
		return r;
	}
SWIGINTERN void delete_Amp(struct Amp *self){
		amp_cleanup(self->amp);
		free(self);
	}
SWIGINTERN void Amp_open(struct Amp *self){ self->error_status = amp_open(self->amp); }
SWIGINTERN void Amp_close(struct Amp *self){ self->error_status = amp_close(self->amp); }
SWIGINTERN void Amp_reset(struct Amp *self,amp_reset_t _amp_reset_t){ self->error_status = amp_reset(self->amp, _amp_reset_t); }
SWIGINTERN void Amp_token_lookup(struct Amp *self,const_char_string _const_char_string){ self->error_status = amp_token_lookup(self->amp, _const_char_string); }
SWIGINTERN setting_t Amp_has_get_level(struct Amp *self,setting_t level){
		return amp_has_get_level(self->amp, level);
	}
SWIGINTERN setting_t Amp_has_set_level(struct Amp *self,setting_t level){
		return amp_has_set_level(self->amp, level);
	}
SWIGINTERN PyObject *Amp_get_ext_level(struct Amp *self,hamlib_token_t token){
	value_t value;

	self->error_status =amp_get_ext_level(self->amp, token, &value);
	if (self->error_status != RIG_OK)
		return Py_None;











	return PyLong_FromLong(value.i);
}
SWIGINTERN void Amp_set_freq(struct Amp *self,freq_t _freq_t){ self->error_status = amp_set_freq(self->amp, _freq_t); }
SWIGINTERN void Amp_set_powerstat(struct Amp *self,powerstat_t _powerstat_t){ self->error_status = amp_set_powerstat(self->amp, _powerstat_t); }
SWIGINTERN void Amp_set_level(struct Amp *self,setting_t _setting_t_1,value_t _value_t_2){ self->error_status = amp_set_level(self->amp, _setting_t_1, _value_t_2); }
SWIGINTERN void Amp_set_ext_level(struct Amp *self,hamlib_token_t _hamlib_token_t_1,value_t _value_t_2){ self->error_status = amp_set_ext_level(self->amp, _hamlib_token_t_1, _value_t_2); }
SWIGINTERN void Amp_set_conf__SWIG_0(struct Amp *self,char const *name,char const *val){
		hamlib_token_t tok = amp_token_lookup(self->amp, name);
		if (tok == 0)
			self->error_status = -RIG_EINVAL;
		else
			self->error_status = amp_set_conf(self->amp, tok, val);
	}
SWIGINTERN void Amp_set_conf__SWIG_1(struct Amp *self,hamlib_token_t _hamlib_token_t_1,const_char_string _const_char_string_2){ self->error_status = amp_set_conf(self->amp, _hamlib_token_t_1, _const_char_string_2); }
SWIGINTERN char const *Amp_get_conf__SWIG_0(struct Amp *self,hamlib_token_t tok){
          static char s[128] = "";
          self->error_status = amp_get_conf2(self->amp, tok, s, sizeof s);
          return s;
  }
SWIGINTERN char const *Amp_get_conf__SWIG_1(struct Amp *self,char const *name){
          hamlib_token_t tok = amp_token_lookup(self->amp, name);
          static char s[128] = "";
          if (tok == 0)
                  self->error_status = -RIG_EINVAL;
          else
                  self->error_status = amp_get_conf2(self->amp, tok, s, sizeof s);
          return s;
  }
SWIGINTERN freq_t Amp_get_freq(struct Amp *self){ freq_t _freq_t; self->error_status = amp_get_freq(self->amp, &_freq_t); return _freq_t; }
SWIGINTERN char const *Amp_get_info(struct Amp *self){
          const char *s;
          s = amp_get_info(self->amp);
          self->error_status = s ? RIG_OK : -RIG_EINVAL;
          return s;
  }
SWIGINTERN powerstat_t Amp_get_powerstat(struct Amp *self){ powerstat_t _powerstat_t; self->error_status = amp_get_powerstat(self->amp, &_powerstat_t); return _powerstat_t; }
SWIGINTERN PyObject *Amp_get_level(struct Amp *self,setting_t level){
      value_t val;

      self->error_status = amp_get_level(self->amp, level, &val);
      if (self->error_status != RIG_OK)
          return Py_None;

      if (((level)&(AMP_LEVEL_SWR)))
          return PyFloat_FromDouble(val.f);
      else if (((level)&(AMP_LEVEL_FAULT)))
          return PyUnicode_FromString(val.s);

      return PyLong_FromLong(val.i);
  }
#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN int Swig_var_cookie_use_set(PyObject *_val) {
  {
    int val;
    int res = SWIG_AsVal_int(_val, &val);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in variable '""cookie_use""' of type '""int""'");
    }
    cookie_use = (int)(val);
  }
  return 0;
fail:
  return 1;
}


SWIGINTERN PyObject *Swig_var_cookie_use_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_From_int((int)(cookie_use));
  return pyobj;
}


SWIGINTERN int Swig_var_skip_init_set(PyObject *_val) {
  {
    int val;
    int res = SWIG_AsVal_int(_val, &val);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in variable '""skip_init""' of type '""int""'");
    }
    skip_init = (int)(val);
  }
  return 0;
fail:
  return 1;
}


SWIGINTERN PyObject *Swig_var_skip_init_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_From_int((int)(skip_init));
  return pyobj;
}


SWIGINTERN int Swig_var_hamlib_version_set(PyObject *_val SWIGUNUSED) {
  SWIG_Error(SWIG_AttributeError,"Variable hamlib_version is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_hamlib_version_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_FromCharPtr(hamlib_version);
  return pyobj;
}


SWIGINTERN int Swig_var_hamlib_copyright_set(PyObject *_val SWIGUNUSED) {
  SWIG_Error(SWIG_AttributeError,"Variable hamlib_copyright is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_hamlib_copyright_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_FromCharPtr(hamlib_copyright);
  return pyobj;
}


SWIGINTERN PyObject *_wrap_confparams_token_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct confparams *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "confparams_token_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_token_set" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "confparams_token_set" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  if (arg1) (arg1)->token = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_token_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  hamlib_token_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_token_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (hamlib_token_t) ((arg1)->token);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_name_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (char *) ((arg1)->name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_label_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_label_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (char *) ((arg1)->label);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_tooltip_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_tooltip_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (char *) ((arg1)->tooltip);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_dflt_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_dflt_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (char *) ((arg1)->dflt);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct confparams *arg1 = 0 ;
  enum rig_conf_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "confparams_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_type_set" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "confparams_type_set" "', argument " "2"" of type '" "enum rig_conf_e""'");
  } 
  arg2 = (enum rig_conf_e)(val2);
  if (arg1) (arg1)->type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum rig_conf_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_type_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (enum rig_conf_e) ((arg1)->type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_u_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  confparams_u *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (confparams_u *)& ((arg1)->u);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams_u, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_confparams(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct confparams *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_confparams", 0, 0, 0)) SWIG_fail;
  result = (struct confparams *)calloc(1, sizeof(struct confparams));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_confparams(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_confparams" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *confparams_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_confparams, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *confparams_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_confparams_u_n_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  confparams_u_n *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_get" "', argument " "1"" of type '" "confparams_u *""'"); 
  }
  arg1 = (confparams_u *)(argp1);
  result = (confparams_u_n *)& ((arg1)->n);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams_u_n, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_u_c_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  confparams_u_c *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_c_get" "', argument " "1"" of type '" "confparams_u *""'"); 
  }
  arg1 = (confparams_u *)(argp1);
  result = (confparams_u_c *)& ((arg1)->c);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams_u_c, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_confparams_u(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_confparams_u", 0, 0, 0)) SWIG_fail;
  result = (confparams_u *)calloc(1, sizeof(confparams_u));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams_u, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_confparams_u(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_confparams_u" "', argument " "1"" of type '" "confparams_u *""'"); 
  }
  arg1 = (confparams_u *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *confparams_u_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_confparams_u, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *confparams_u_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_confparams_u_c_combostr_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u_c *arg1 = 0 ;
  char **arg2 = (char **)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "confparams_u_c_combostr_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u_c, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_c_combostr_set" "', argument " "1"" of type '" "confparams_u_c *""'"); 
  }
  arg1 = (confparams_u_c *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "confparams_u_c_combostr_set" "', argument " "2"" of type '" "char const *[16]""'"); 
  } 
  arg2 = (char **)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)16; ++ii) *(char const * *)&arg1->combostr[ii] = *((char const * *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""combostr""' of type '""char const *[16]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_u_c_combostr_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u_c *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char **result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u_c, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_c_combostr_get" "', argument " "1"" of type '" "confparams_u_c *""'"); 
  }
  arg1 = (confparams_u_c *)(argp1);
  result = (char **)(char **) ((arg1)->combostr);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_confparams_u_c(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u_c *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_confparams_u_c", 0, 0, 0)) SWIG_fail;
  result = (confparams_u_c *)calloc(1, sizeof(confparams_u_c));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams_u_c, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_confparams_u_c(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u_c *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u_c, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_confparams_u_c" "', argument " "1"" of type '" "confparams_u_c *""'"); 
  }
  arg1 = (confparams_u_c *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *confparams_u_c_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_confparams_u_c, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *confparams_u_c_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_confparams_u_n_min_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u_n *arg1 = 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "confparams_u_n_min_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u_n, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_min_set" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "confparams_u_n_min_set" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = (float)(val2);
  if (arg1) (arg1)->min = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_u_n_min_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u_n *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u_n, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_min_get" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  result = (float) ((arg1)->min);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_u_n_max_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u_n *arg1 = 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "confparams_u_n_max_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u_n, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_max_set" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "confparams_u_n_max_set" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = (float)(val2);
  if (arg1) (arg1)->max = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_u_n_max_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u_n *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u_n, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_max_get" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  result = (float) ((arg1)->max);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_u_n_step_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u_n *arg1 = 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "confparams_u_n_step_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u_n, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_step_set" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "confparams_u_n_step_set" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = (float)(val2);
  if (arg1) (arg1)->step = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_confparams_u_n_step_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u_n *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u_n, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_step_get" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  result = (float) ((arg1)->step);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_confparams_u_n(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u_n *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_confparams_u_n", 0, 0, 0)) SWIG_fail;
  result = (confparams_u_n *)calloc(1, sizeof(confparams_u_n));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams_u_n, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_confparams_u_n(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  confparams_u_n *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_confparams_u_n, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_confparams_u_n" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *confparams_u_n_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_confparams_u_n, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *confparams_u_n_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_value_t_i_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "value_t_i_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_i_set" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "value_t_i_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->i = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_value_t_i_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_i_get" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  result = (int) ((arg1)->i);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_value_t_u_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "value_t_u_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_u_set" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "value_t_u_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->u = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_value_t_u_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_u_get" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  result = (unsigned int) ((arg1)->u);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_value_t_f_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t *arg1 = 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "value_t_f_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_f_set" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "value_t_f_set" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = (float)(val2);
  if (arg1) (arg1)->f = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_value_t_f_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_f_get" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  result = (float) ((arg1)->f);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_value_t_s_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "value_t_s_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_s_set" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "value_t_s_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->s);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->s = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->s = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_value_t_s_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_s_get" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  result = (char *) ((arg1)->s);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_value_t_cs_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_cs_get" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  result = (char *) ((arg1)->cs);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_value_t_b_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  value_t_b *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_b_get" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  result = (value_t_b *)& ((arg1)->b);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_value_t_b, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_value_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_value_t", 0, 0, 0)) SWIG_fail;
  result = (value_t *)calloc(1, sizeof(value_t));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_value_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_value_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_value_t" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *value_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_value_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *value_t_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_value_t_b_l_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t_b *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "value_t_b_l_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t_b, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_b_l_set" "', argument " "1"" of type '" "value_t_b *""'"); 
  }
  arg1 = (value_t_b *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "value_t_b_l_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->l = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_value_t_b_l_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t_b *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t_b, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_b_l_get" "', argument " "1"" of type '" "value_t_b *""'"); 
  }
  arg1 = (value_t_b *)(argp1);
  result = (int) ((arg1)->l);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_value_t_b_d_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t_b *arg1 = 0 ;
  unsigned char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "value_t_b_d_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t_b, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_b_d_set" "', argument " "1"" of type '" "value_t_b *""'"); 
  }
  arg1 = (value_t_b *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_unsigned_char, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "value_t_b_d_set" "', argument " "2"" of type '" "unsigned char *""'"); 
  }
  arg2 = (unsigned char *)(argp2);
  if (arg1) (arg1)->d = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_value_t_b_d_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t_b *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t_b, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_b_d_get" "', argument " "1"" of type '" "value_t_b *""'"); 
  }
  arg1 = (value_t_b *)(argp1);
  result = (unsigned char *) ((arg1)->d);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_value_t_b(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t_b *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_value_t_b", 0, 0, 0)) SWIG_fail;
  result = (value_t_b *)calloc(1, sizeof(value_t_b));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_value_t_b, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_value_t_b(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  value_t_b *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_value_t_b, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_value_t_b" "', argument " "1"" of type '" "value_t_b *""'"); 
  }
  arg1 = (value_t_b *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *value_t_b_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_value_t_b, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *value_t_b_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_freq_range_t_startf_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "freq_range_t_startf_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_startf_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_startf_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->startf = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_startf_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_startf_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (freq_t) ((arg1)->startf);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_endf_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "freq_range_t_endf_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_endf_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_endf_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->endf = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_endf_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_endf_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (freq_t) ((arg1)->endf);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_modes_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "freq_range_t_modes_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_modes_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_modes_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->modes = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_modes_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rmode_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_modes_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (rmode_t) ((arg1)->modes);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_low_power_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "freq_range_t_low_power_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_low_power_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_low_power_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->low_power = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_low_power_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_low_power_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (int) ((arg1)->low_power);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_high_power_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "freq_range_t_high_power_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_high_power_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_high_power_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->high_power = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_high_power_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_high_power_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (int) ((arg1)->high_power);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "freq_range_t_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_vfo_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_vfo_set" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  if (arg1) (arg1)->vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  vfo_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_vfo_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (vfo_t) ((arg1)->vfo);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_ant_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  ant_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "freq_range_t_ant_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_ant_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_ant_set" "', argument " "2"" of type '" "ant_t""'");
  } 
  arg2 = (ant_t)(val2);
  if (arg1) (arg1)->ant = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_ant_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  ant_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_ant_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (ant_t) ((arg1)->ant);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_label_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "freq_range_t_label_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_label_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "freq_range_t_label_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->label);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->label = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->label = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_freq_range_t_label_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_label_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (char *) ((arg1)->label);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_freq_range_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_freq_range_t", 0, 0, 0)) SWIG_fail;
  result = (struct freq_range_list *)calloc(1, sizeof(struct freq_range_list));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_freq_range_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_freq_range_list, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_freq_range_t" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *freq_range_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_freq_range_list, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *freq_range_t_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_tuning_step_list_modes_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct tuning_step_list *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "tuning_step_list_modes_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_tuning_step_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "tuning_step_list_modes_set" "', argument " "1"" of type '" "struct tuning_step_list *""'"); 
  }
  arg1 = (struct tuning_step_list *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "tuning_step_list_modes_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->modes = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_tuning_step_list_modes_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct tuning_step_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rmode_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_tuning_step_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "tuning_step_list_modes_get" "', argument " "1"" of type '" "struct tuning_step_list *""'"); 
  }
  arg1 = (struct tuning_step_list *)(argp1);
  result = (rmode_t) ((arg1)->modes);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_tuning_step_list_ts_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct tuning_step_list *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "tuning_step_list_ts_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_tuning_step_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "tuning_step_list_ts_set" "', argument " "1"" of type '" "struct tuning_step_list *""'"); 
  }
  arg1 = (struct tuning_step_list *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "tuning_step_list_ts_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->ts = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_tuning_step_list_ts_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct tuning_step_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  shortfreq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_tuning_step_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "tuning_step_list_ts_get" "', argument " "1"" of type '" "struct tuning_step_list *""'"); 
  }
  arg1 = (struct tuning_step_list *)(argp1);
  result = (shortfreq_t) ((arg1)->ts);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_tuning_step_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct tuning_step_list *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_tuning_step_list", 0, 0, 0)) SWIG_fail;
  result = (struct tuning_step_list *)calloc(1, sizeof(struct tuning_step_list));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_tuning_step_list, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_tuning_step_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct tuning_step_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_tuning_step_list, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_tuning_step_list" "', argument " "1"" of type '" "struct tuning_step_list *""'"); 
  }
  arg1 = (struct tuning_step_list *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *tuning_step_list_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_tuning_step_list, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *tuning_step_list_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_filter_list_modes_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct filter_list *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "filter_list_modes_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_filter_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "filter_list_modes_set" "', argument " "1"" of type '" "struct filter_list *""'"); 
  }
  arg1 = (struct filter_list *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "filter_list_modes_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->modes = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_filter_list_modes_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct filter_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rmode_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_filter_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "filter_list_modes_get" "', argument " "1"" of type '" "struct filter_list *""'"); 
  }
  arg1 = (struct filter_list *)(argp1);
  result = (rmode_t) ((arg1)->modes);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_filter_list_width_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct filter_list *arg1 = 0 ;
  pbwidth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "filter_list_width_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_filter_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "filter_list_width_set" "', argument " "1"" of type '" "struct filter_list *""'"); 
  }
  arg1 = (struct filter_list *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "filter_list_width_set" "', argument " "2"" of type '" "pbwidth_t""'");
  } 
  arg2 = (pbwidth_t)(val2);
  if (arg1) (arg1)->width = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_filter_list_width_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct filter_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  pbwidth_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_filter_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "filter_list_width_get" "', argument " "1"" of type '" "struct filter_list *""'"); 
  }
  arg1 = (struct filter_list *)(argp1);
  result = (pbwidth_t) ((arg1)->width);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_filter_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct filter_list *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_filter_list", 0, 0, 0)) SWIG_fail;
  result = (struct filter_list *)calloc(1, sizeof(struct filter_list));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_filter_list, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_filter_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct filter_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_filter_list, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_filter_list" "', argument " "1"" of type '" "struct filter_list *""'"); 
  }
  arg1 = (struct filter_list *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *filter_list_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_filter_list, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *filter_list_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_ext_list_token_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct ext_list *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "ext_list_token_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_ext_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ext_list_token_set" "', argument " "1"" of type '" "struct ext_list *""'"); 
  }
  arg1 = (struct ext_list *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ext_list_token_set" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  if (arg1) (arg1)->token = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ext_list_token_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct ext_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  hamlib_token_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_ext_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ext_list_token_get" "', argument " "1"" of type '" "struct ext_list *""'"); 
  }
  arg1 = (struct ext_list *)(argp1);
  result = (hamlib_token_t) ((arg1)->token);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ext_list_val_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct ext_list *arg1 = 0 ;
  value_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "ext_list_val_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_ext_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ext_list_val_set" "', argument " "1"" of type '" "struct ext_list *""'"); 
  }
  arg1 = (struct ext_list *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ext_list_val_set" "', argument " "2"" of type '" "value_t *""'"); 
  }
  arg2 = (value_t *)(argp2);
  if (arg1) (arg1)->val = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ext_list_val_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct ext_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  value_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_ext_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ext_list_val_get" "', argument " "1"" of type '" "struct ext_list *""'"); 
  }
  arg1 = (struct ext_list *)(argp1);
  result = (value_t *)& ((arg1)->val);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_value_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_ext_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct ext_list *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_ext_list", 0, 0, 0)) SWIG_fail;
  result = (struct ext_list *)calloc(1, sizeof(struct ext_list));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_ext_list, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_ext_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct ext_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_ext_list, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ext_list" "', argument " "1"" of type '" "struct ext_list *""'"); 
  }
  arg1 = (struct ext_list *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *ext_list_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_ext_list, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *ext_list_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_channel_channel_num_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_channel_num_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_channel_num_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_channel_num_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->channel_num = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_channel_num_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_channel_num_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (int) ((arg1)->channel_num);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_bank_num_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_bank_num_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_bank_num_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_bank_num_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->bank_num = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_bank_num_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_bank_num_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (int) ((arg1)->bank_num);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_vfo_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_vfo_set" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  if (arg1) (arg1)->vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  vfo_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_vfo_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (vfo_t) ((arg1)->vfo);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_ant_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  ant_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_ant_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ant_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_ant_set" "', argument " "2"" of type '" "ant_t""'");
  } 
  arg2 = (ant_t)(val2);
  if (arg1) (arg1)->ant = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_ant_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  ant_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ant_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (ant_t) ((arg1)->ant);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_freq_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_freq_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (freq_t) ((arg1)->freq);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_mode_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_mode_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rmode_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_mode_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (rmode_t) ((arg1)->mode);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_width_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  pbwidth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_width_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_width_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_width_set" "', argument " "2"" of type '" "pbwidth_t""'");
  } 
  arg2 = (pbwidth_t)(val2);
  if (arg1) (arg1)->width = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_width_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  pbwidth_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_width_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (pbwidth_t) ((arg1)->width);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_tx_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_tx_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_freq_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_tx_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->tx_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_tx_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_freq_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (freq_t) ((arg1)->tx_freq);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_tx_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_tx_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_mode_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_tx_mode_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->tx_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_tx_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rmode_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_mode_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (rmode_t) ((arg1)->tx_mode);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_tx_width_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  pbwidth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_tx_width_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_width_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_tx_width_set" "', argument " "2"" of type '" "pbwidth_t""'");
  } 
  arg2 = (pbwidth_t)(val2);
  if (arg1) (arg1)->tx_width = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_tx_width_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  pbwidth_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_width_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (pbwidth_t) ((arg1)->tx_width);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_split_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  split_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_split_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_split_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_split_set" "', argument " "2"" of type '" "split_t""'");
  } 
  arg2 = (split_t)(val2);
  if (arg1) (arg1)->split = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_split_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  split_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_split_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (split_t) ((arg1)->split);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_tx_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_tx_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_vfo_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_tx_vfo_set" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  if (arg1) (arg1)->tx_vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_tx_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  vfo_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_vfo_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (vfo_t) ((arg1)->tx_vfo);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_rptr_shift_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  rptr_shift_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_rptr_shift_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_rptr_shift_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_rptr_shift_set" "', argument " "2"" of type '" "rptr_shift_t""'");
  } 
  arg2 = (rptr_shift_t)(val2);
  if (arg1) (arg1)->rptr_shift = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_rptr_shift_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rptr_shift_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_rptr_shift_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (rptr_shift_t) ((arg1)->rptr_shift);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_rptr_offs_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_rptr_offs_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_rptr_offs_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_rptr_offs_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->rptr_offs = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_rptr_offs_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  shortfreq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_rptr_offs_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (shortfreq_t) ((arg1)->rptr_offs);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_tuning_step_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_tuning_step_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tuning_step_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_tuning_step_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->tuning_step = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_tuning_step_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  shortfreq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tuning_step_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (shortfreq_t) ((arg1)->tuning_step);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_rit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_rit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_rit_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_rit_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->rit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_rit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  shortfreq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_rit_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (shortfreq_t) ((arg1)->rit);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_xit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_xit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_xit_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_xit_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->xit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_xit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  shortfreq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_xit_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (shortfreq_t) ((arg1)->xit);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_funcs_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_funcs_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_funcs_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_funcs_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->funcs = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_funcs_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_funcs_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (setting_t) ((arg1)->funcs);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_levels_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  value_t *arg2 = (value_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_levels_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_levels_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "channel_levels_set" "', argument " "2"" of type '" "value_t [64]""'"); 
  } 
  arg2 = (value_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(value_t *)&arg1->levels[ii] = *((value_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""levels""' of type '""value_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_levels_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  value_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_levels_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (value_t *)(value_t *) ((arg1)->levels);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_value_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_ctcss_tone_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  tone_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_ctcss_tone_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ctcss_tone_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_ctcss_tone_set" "', argument " "2"" of type '" "tone_t""'");
  } 
  arg2 = (tone_t)(val2);
  if (arg1) (arg1)->ctcss_tone = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_ctcss_tone_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  tone_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ctcss_tone_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (tone_t) ((arg1)->ctcss_tone);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_ctcss_sql_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  tone_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_ctcss_sql_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ctcss_sql_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_ctcss_sql_set" "', argument " "2"" of type '" "tone_t""'");
  } 
  arg2 = (tone_t)(val2);
  if (arg1) (arg1)->ctcss_sql = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_ctcss_sql_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  tone_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ctcss_sql_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (tone_t) ((arg1)->ctcss_sql);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_dcs_code_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  tone_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_dcs_code_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_dcs_code_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_dcs_code_set" "', argument " "2"" of type '" "tone_t""'");
  } 
  arg2 = (tone_t)(val2);
  if (arg1) (arg1)->dcs_code = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_dcs_code_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  tone_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_dcs_code_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (tone_t) ((arg1)->dcs_code);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_dcs_sql_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  tone_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_dcs_sql_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_dcs_sql_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_dcs_sql_set" "', argument " "2"" of type '" "tone_t""'");
  } 
  arg2 = (tone_t)(val2);
  if (arg1) (arg1)->dcs_sql = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_dcs_sql_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  tone_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_dcs_sql_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (tone_t) ((arg1)->dcs_sql);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_scan_group_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_scan_group_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_scan_group_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_scan_group_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->scan_group = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_scan_group_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_scan_group_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (int) ((arg1)->scan_group);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_flags_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_flags_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_flags_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_flags_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->flags = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_flags_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_flags_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (unsigned int) ((arg1)->flags);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_channel_desc_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  char *arg2 = (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char temp2[30] ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_channel_desc_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_channel_desc_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  res2 = SWIG_AsCharArray(swig_obj[1], temp2, 30);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "channel_channel_desc_set" "', argument " "2"" of type '" "char [30]""'");
  }
  arg2 = (char *)(temp2);
  if (arg2) memcpy(arg1->channel_desc,arg2,30*sizeof(char));
  else memset(arg1->channel_desc,0,30*sizeof(char));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_channel_desc_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_channel_desc_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (char *)(char *) ((arg1)->channel_desc);
  {
    size_t size = SWIG_strnlen(result, 30);
    
    
    
    resultobj = SWIG_FromCharPtrAndSize(result, size);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_ext_levels_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  struct ext_list *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_ext_levels_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ext_levels_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_ext_list, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "channel_ext_levels_set" "', argument " "2"" of type '" "struct ext_list *""'"); 
  }
  arg2 = (struct ext_list *)(argp2);
  if (arg1) (arg1)->ext_levels = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_ext_levels_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct ext_list *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ext_levels_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (struct ext_list *) ((arg1)->ext_levels);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_ext_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_tag_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  char *arg2 = (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char temp2[32] ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_tag_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tag_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  res2 = SWIG_AsCharArray(swig_obj[1], temp2, 32);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "channel_tag_set" "', argument " "2"" of type '" "char [32]""'");
  }
  arg2 = (char *)(temp2);
  if (arg2) memcpy(arg1->tag,arg2,32*sizeof(char));
  else memset(arg1->tag,0,32*sizeof(char));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_tag_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tag_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (char *)(char *) ((arg1)->tag);
  {
    size_t size = SWIG_strnlen(result, 32);
    
    
    
    resultobj = SWIG_FromCharPtrAndSize(result, size);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_channel(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  vfo_t arg1 = (1u << (29)) ;
  int arg2 = 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  struct channel *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_channel", 0, 2, swig_obj)) SWIG_fail;
  if (swig_obj[0]) {
    ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_channel" "', argument " "1"" of type '" "vfo_t""'");
    } 
    arg1 = (vfo_t)(val1);
  }
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_channel" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = (struct channel *)new_channel(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_channel, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_channel(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_channel" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  delete_channel(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *channel_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_channel, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *channel_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_channel_cap_bank_num_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_bank_num_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_bank_num_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_bank_num_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->bank_num = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_bank_num_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_bank_num_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->bank_num);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_vfo_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_vfo_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_vfo_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->vfo);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_ant_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_ant_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ant_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_ant_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->ant = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_ant_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ant_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->ant);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_freq_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_freq_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_freq_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->freq);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_mode_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_mode_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_mode_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->mode);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_width_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_width_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_width_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_width_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->width = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_width_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_width_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->width);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_tx_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_tx_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_freq_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_tx_freq_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->tx_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_tx_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_freq_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->tx_freq);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_tx_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_tx_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_mode_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_tx_mode_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->tx_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_tx_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_mode_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->tx_mode);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_tx_width_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_tx_width_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_width_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_tx_width_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->tx_width = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_tx_width_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_width_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->tx_width);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_split_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_split_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_split_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_split_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->split = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_split_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_split_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->split);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_tx_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_tx_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_vfo_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_tx_vfo_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->tx_vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_tx_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_vfo_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->tx_vfo);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_rptr_shift_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_rptr_shift_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_rptr_shift_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_rptr_shift_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->rptr_shift = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_rptr_shift_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_rptr_shift_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->rptr_shift);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_rptr_offs_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_rptr_offs_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_rptr_offs_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_rptr_offs_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->rptr_offs = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_rptr_offs_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_rptr_offs_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->rptr_offs);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_tuning_step_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_tuning_step_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tuning_step_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_tuning_step_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->tuning_step = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_tuning_step_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tuning_step_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->tuning_step);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_rit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_rit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_rit_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_rit_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->rit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_rit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_rit_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->rit);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_xit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_xit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_xit_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_xit_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->xit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_xit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_xit_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->xit);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_funcs_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_funcs_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_funcs_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_funcs_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->funcs = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_funcs_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_funcs_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (setting_t) ((arg1)->funcs);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_levels_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_levels_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_levels_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_levels_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->levels = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_levels_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_levels_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (setting_t) ((arg1)->levels);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_ctcss_tone_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_ctcss_tone_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ctcss_tone_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_ctcss_tone_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->ctcss_tone = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_ctcss_tone_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ctcss_tone_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->ctcss_tone);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_ctcss_sql_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_ctcss_sql_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ctcss_sql_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_ctcss_sql_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->ctcss_sql = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_ctcss_sql_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ctcss_sql_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->ctcss_sql);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_dcs_code_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_dcs_code_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_dcs_code_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_dcs_code_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->dcs_code = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_dcs_code_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_dcs_code_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->dcs_code);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_dcs_sql_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_dcs_sql_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_dcs_sql_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_dcs_sql_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->dcs_sql = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_dcs_sql_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_dcs_sql_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->dcs_sql);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_scan_group_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_scan_group_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_scan_group_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_scan_group_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->scan_group = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_scan_group_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_scan_group_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->scan_group);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_flags_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_flags_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_flags_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_flags_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->flags = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_flags_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_flags_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->flags);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_channel_desc_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_channel_desc_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_channel_desc_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_channel_desc_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->channel_desc = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_channel_desc_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_channel_desc_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->channel_desc);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_ext_levels_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_ext_levels_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ext_levels_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_ext_levels_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->ext_levels = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_ext_levels_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ext_levels_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->ext_levels);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_tag_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "channel_cap_tag_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tag_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_tag_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->tag = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channel_cap_tag_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tag_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->tag);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_channel_cap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_channel_cap", 0, 0, 0)) SWIG_fail;
  result = (struct channel_cap *)calloc(1, sizeof(struct channel_cap));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_channel_cap, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_channel_cap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel_cap, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_channel_cap" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *channel_cap_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_channel_cap, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *channel_cap_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_chan_list_startc_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct chan_list *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "chan_list_startc_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_startc_set" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "chan_list_startc_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->startc = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_chan_list_startc_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct chan_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_startc_get" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  result = (int) ((arg1)->startc);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_chan_list_endc_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct chan_list *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "chan_list_endc_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_endc_set" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "chan_list_endc_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->endc = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_chan_list_endc_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct chan_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_endc_get" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  result = (int) ((arg1)->endc);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_chan_list_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct chan_list *arg1 = 0 ;
  chan_type_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "chan_list_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_type_set" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "chan_list_type_set" "', argument " "2"" of type '" "chan_type_t""'");
  } 
  arg2 = (chan_type_t)(val2);
  if (arg1) (arg1)->type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_chan_list_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct chan_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  chan_type_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_type_get" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  result = (chan_type_t) ((arg1)->type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_chan_list_mem_caps_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct chan_list *arg1 = 0 ;
  channel_cap_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "chan_list_mem_caps_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_mem_caps_set" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "chan_list_mem_caps_set" "', argument " "2"" of type '" "channel_cap_t *""'"); 
  }
  arg2 = (channel_cap_t *)(argp2);
  if (arg1) (arg1)->mem_caps = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_chan_list_mem_caps_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct chan_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  channel_cap_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_mem_caps_get" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  result = (channel_cap_t *)& ((arg1)->mem_caps);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_channel_cap, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_chan_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct chan_list *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_chan_list", 0, 0, 0)) SWIG_fail;
  result = (struct chan_list *)calloc(1, sizeof(struct chan_list));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_chan_list, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_chan_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct chan_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_chan_list, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_chan_list" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *chan_list_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_chan_list, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *chan_list_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_gran_min_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct gran *arg1 = 0 ;
  value_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "gran_min_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "gran_min_set" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "gran_min_set" "', argument " "2"" of type '" "value_t *""'"); 
  }
  arg2 = (value_t *)(argp2);
  if (arg1) (arg1)->min = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_gran_min_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct gran *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  value_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "gran_min_get" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  result = (value_t *)& ((arg1)->min);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_value_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_gran_max_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct gran *arg1 = 0 ;
  value_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "gran_max_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "gran_max_set" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "gran_max_set" "', argument " "2"" of type '" "value_t *""'"); 
  }
  arg2 = (value_t *)(argp2);
  if (arg1) (arg1)->max = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_gran_max_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct gran *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  value_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "gran_max_get" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  result = (value_t *)& ((arg1)->max);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_value_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_gran_step_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct gran *arg1 = 0 ;
  value_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "gran_step_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "gran_step_set" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "gran_step_set" "', argument " "2"" of type '" "value_t *""'"); 
  }
  arg2 = (value_t *)(argp2);
  if (arg1) (arg1)->step = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_gran_step_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct gran *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  value_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "gran_step_get" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  result = (value_t *)& ((arg1)->step);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_value_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_gran(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct gran *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_gran", 0, 0, 0)) SWIG_fail;
  result = (struct gran *)calloc(1, sizeof(struct gran));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_gran(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct gran *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_gran, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_gran" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *gran_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_gran, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *gran_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_cal_table_size_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct cal_table *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "cal_table_size_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_size_set" "', argument " "1"" of type '" "struct cal_table *""'"); 
  }
  arg1 = (struct cal_table *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "cal_table_size_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_cal_table_size_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct cal_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_size_get" "', argument " "1"" of type '" "struct cal_table *""'"); 
  }
  arg1 = (struct cal_table *)(argp1);
  result = (int) ((arg1)->size);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_cal_table_table_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct cal_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  cal_table_table *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_table_get" "', argument " "1"" of type '" "struct cal_table *""'"); 
  }
  arg1 = (struct cal_table *)(argp1);
  result = (cal_table_table *)(cal_table_table *) ((arg1)->table);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_table, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_cal_table(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct cal_table *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_cal_table", 0, 0, 0)) SWIG_fail;
  result = (struct cal_table *)calloc(1, sizeof(struct cal_table));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_cal_table(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct cal_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_cal_table" "', argument " "1"" of type '" "struct cal_table *""'"); 
  }
  arg1 = (struct cal_table *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *cal_table_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_cal_table, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *cal_table_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_cal_table_table_raw_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cal_table_table *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "cal_table_table_raw_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_table_raw_set" "', argument " "1"" of type '" "cal_table_table *""'"); 
  }
  arg1 = (cal_table_table *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "cal_table_table_raw_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->raw = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_cal_table_table_raw_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cal_table_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_table_raw_get" "', argument " "1"" of type '" "cal_table_table *""'"); 
  }
  arg1 = (cal_table_table *)(argp1);
  result = (int) ((arg1)->raw);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_cal_table_table_val_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cal_table_table *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "cal_table_table_val_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_table_val_set" "', argument " "1"" of type '" "cal_table_table *""'"); 
  }
  arg1 = (cal_table_table *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "cal_table_table_val_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->val = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_cal_table_table_val_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cal_table_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_table_val_get" "', argument " "1"" of type '" "cal_table_table *""'"); 
  }
  arg1 = (cal_table_table *)(argp1);
  result = (int) ((arg1)->val);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_cal_table_table(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cal_table_table *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_cal_table_table", 0, 0, 0)) SWIG_fail;
  result = (cal_table_table *)calloc(1, sizeof(cal_table_table));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_table, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_cal_table_table(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cal_table_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_table, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_cal_table_table" "', argument " "1"" of type '" "cal_table_table *""'"); 
  }
  arg1 = (cal_table_table *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *cal_table_table_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_cal_table_table, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *cal_table_table_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_cal_table_float_size_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct cal_table_float *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "cal_table_float_size_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_size_set" "', argument " "1"" of type '" "struct cal_table_float *""'"); 
  }
  arg1 = (struct cal_table_float *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "cal_table_float_size_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_cal_table_float_size_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct cal_table_float *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_size_get" "', argument " "1"" of type '" "struct cal_table_float *""'"); 
  }
  arg1 = (struct cal_table_float *)(argp1);
  result = (int) ((arg1)->size);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_cal_table_float_table_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct cal_table_float *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  cal_table_float_table *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_table_get" "', argument " "1"" of type '" "struct cal_table_float *""'"); 
  }
  arg1 = (struct cal_table_float *)(argp1);
  result = (cal_table_float_table *)(cal_table_float_table *) ((arg1)->table);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float_table, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_cal_table_float(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct cal_table_float *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_cal_table_float", 0, 0, 0)) SWIG_fail;
  result = (struct cal_table_float *)calloc(1, sizeof(struct cal_table_float));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_cal_table_float(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct cal_table_float *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_float, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_cal_table_float" "', argument " "1"" of type '" "struct cal_table_float *""'"); 
  }
  arg1 = (struct cal_table_float *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *cal_table_float_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_cal_table_float, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *cal_table_float_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_cal_table_float_table_raw_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cal_table_float_table *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "cal_table_float_table_raw_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_float_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_table_raw_set" "', argument " "1"" of type '" "cal_table_float_table *""'"); 
  }
  arg1 = (cal_table_float_table *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "cal_table_float_table_raw_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->raw = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_cal_table_float_table_raw_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cal_table_float_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_float_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_table_raw_get" "', argument " "1"" of type '" "cal_table_float_table *""'"); 
  }
  arg1 = (cal_table_float_table *)(argp1);
  result = (int) ((arg1)->raw);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_cal_table_float_table_val_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cal_table_float_table *arg1 = 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "cal_table_float_table_val_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_float_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_table_val_set" "', argument " "1"" of type '" "cal_table_float_table *""'"); 
  }
  arg1 = (cal_table_float_table *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "cal_table_float_table_val_set" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = (float)(val2);
  if (arg1) (arg1)->val = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_cal_table_float_table_val_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cal_table_float_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_float_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_table_val_get" "', argument " "1"" of type '" "cal_table_float_table *""'"); 
  }
  arg1 = (cal_table_float_table *)(argp1);
  result = (float) ((arg1)->val);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_cal_table_float_table(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cal_table_float_table *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_cal_table_float_table", 0, 0, 0)) SWIG_fail;
  result = (cal_table_float_table *)calloc(1, sizeof(cal_table_float_table));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float_table, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_cal_table_float_table(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cal_table_float_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_cal_table_float_table, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_cal_table_float_table" "', argument " "1"" of type '" "cal_table_float_table *""'"); 
  }
  arg1 = (cal_table_float_table *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *cal_table_float_table_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_cal_table_float_table, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *cal_table_float_table_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_rig_spectrum_scope_id_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_scope *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_scope_id_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_scope_id_set" "', argument " "1"" of type '" "struct rig_spectrum_scope *""'"); 
  }
  arg1 = (struct rig_spectrum_scope *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_scope_id_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->id = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_scope_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_scope *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_scope_id_get" "', argument " "1"" of type '" "struct rig_spectrum_scope *""'"); 
  }
  arg1 = (struct rig_spectrum_scope *)(argp1);
  result = (int) ((arg1)->id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_scope_name_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_scope *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_scope_name_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_scope_name_set" "', argument " "1"" of type '" "struct rig_spectrum_scope *""'"); 
  }
  arg1 = (struct rig_spectrum_scope *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_spectrum_scope_name_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->name);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->name = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->name = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_scope_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_scope *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_scope_name_get" "', argument " "1"" of type '" "struct rig_spectrum_scope *""'"); 
  }
  arg1 = (struct rig_spectrum_scope *)(argp1);
  result = (char *) ((arg1)->name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_rig_spectrum_scope(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_scope *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_rig_spectrum_scope", 0, 0, 0)) SWIG_fail;
  result = (struct rig_spectrum_scope *)calloc(1, sizeof(struct rig_spectrum_scope));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_scope, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_rig_spectrum_scope(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_scope *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_scope, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rig_spectrum_scope" "', argument " "1"" of type '" "struct rig_spectrum_scope *""'"); 
  }
  arg1 = (struct rig_spectrum_scope *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *rig_spectrum_scope_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_rig_spectrum_scope, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *rig_spectrum_scope_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_rig_spectrum_avg_mode_id_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_avg_mode *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_avg_mode_id_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_avg_mode_id_set" "', argument " "1"" of type '" "struct rig_spectrum_avg_mode *""'"); 
  }
  arg1 = (struct rig_spectrum_avg_mode *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_avg_mode_id_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->id = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_avg_mode_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_avg_mode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_avg_mode_id_get" "', argument " "1"" of type '" "struct rig_spectrum_avg_mode *""'"); 
  }
  arg1 = (struct rig_spectrum_avg_mode *)(argp1);
  result = (int) ((arg1)->id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_avg_mode_name_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_avg_mode *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_avg_mode_name_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_avg_mode_name_set" "', argument " "1"" of type '" "struct rig_spectrum_avg_mode *""'"); 
  }
  arg1 = (struct rig_spectrum_avg_mode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_spectrum_avg_mode_name_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->name);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->name = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->name = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_avg_mode_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_avg_mode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_avg_mode_name_get" "', argument " "1"" of type '" "struct rig_spectrum_avg_mode *""'"); 
  }
  arg1 = (struct rig_spectrum_avg_mode *)(argp1);
  result = (char *) ((arg1)->name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_rig_spectrum_avg_mode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_avg_mode *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_rig_spectrum_avg_mode", 0, 0, 0)) SWIG_fail;
  result = (struct rig_spectrum_avg_mode *)calloc(1, sizeof(struct rig_spectrum_avg_mode));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_avg_mode, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_rig_spectrum_avg_mode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_avg_mode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_avg_mode, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rig_spectrum_avg_mode" "', argument " "1"" of type '" "struct rig_spectrum_avg_mode *""'"); 
  }
  arg1 = (struct rig_spectrum_avg_mode *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *rig_spectrum_avg_mode_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_rig_spectrum_avg_mode, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *rig_spectrum_avg_mode_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_rig_spectrum_line_id_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_line_id_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_id_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_id_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->id = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_id_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (int) ((arg1)->id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_data_level_min_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_line_data_level_min_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_data_level_min_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_data_level_min_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->data_level_min = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_data_level_min_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_data_level_min_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (int) ((arg1)->data_level_min);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_data_level_max_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_line_data_level_max_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_data_level_max_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_data_level_max_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->data_level_max = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_data_level_max_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_data_level_max_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (int) ((arg1)->data_level_max);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_signal_strength_min_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_line_signal_strength_min_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_signal_strength_min_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_signal_strength_min_set" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  if (arg1) (arg1)->signal_strength_min = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_signal_strength_min_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_signal_strength_min_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (double) ((arg1)->signal_strength_min);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_signal_strength_max_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_line_signal_strength_max_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_signal_strength_max_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_signal_strength_max_set" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  if (arg1) (arg1)->signal_strength_max = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_signal_strength_max_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_signal_strength_max_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (double) ((arg1)->signal_strength_max);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_spectrum_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  enum rig_spectrum_mode_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_line_spectrum_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_spectrum_mode_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_spectrum_mode_set" "', argument " "2"" of type '" "enum rig_spectrum_mode_e""'");
  } 
  arg2 = (enum rig_spectrum_mode_e)(val2);
  if (arg1) (arg1)->spectrum_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_spectrum_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum rig_spectrum_mode_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_spectrum_mode_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (enum rig_spectrum_mode_e) ((arg1)->spectrum_mode);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_center_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_line_center_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_center_freq_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_center_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->center_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_center_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_center_freq_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (freq_t) ((arg1)->center_freq);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_span_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_line_span_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_span_freq_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_span_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->span_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_span_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_span_freq_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (freq_t) ((arg1)->span_freq);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_low_edge_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_line_low_edge_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_low_edge_freq_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_low_edge_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->low_edge_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_low_edge_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_low_edge_freq_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (freq_t) ((arg1)->low_edge_freq);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_high_edge_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_line_high_edge_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_high_edge_freq_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_high_edge_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->high_edge_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_high_edge_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_high_edge_freq_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (freq_t) ((arg1)->high_edge_freq);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_spectrum_data_length_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_line_spectrum_data_length_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_spectrum_data_length_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_spectrum_data_length_set" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  if (arg1) (arg1)->spectrum_data_length = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_spectrum_data_length_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  size_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_spectrum_data_length_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result =  ((arg1)->spectrum_data_length);
  resultobj = SWIG_From_size_t((size_t)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_spectrum_data_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  unsigned char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_spectrum_line_spectrum_data_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_spectrum_data_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_unsigned_char, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_spectrum_line_spectrum_data_set" "', argument " "2"" of type '" "unsigned char *""'"); 
  }
  arg2 = (unsigned char *)(argp2);
  if (arg1) (arg1)->spectrum_data = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_spectrum_line_spectrum_data_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_spectrum_data_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (unsigned char *) ((arg1)->spectrum_data);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_rig_spectrum_line(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_rig_spectrum_line", 0, 0, 0)) SWIG_fail;
  result = (struct rig_spectrum_line *)calloc(1, sizeof(struct rig_spectrum_line));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_line, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_rig_spectrum_line(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_spectrum_line, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rig_spectrum_line" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *rig_spectrum_line_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_rig_spectrum_line, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *rig_spectrum_line_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_deferred_config_item_nextt_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_item *arg1 = 0 ;
  struct deferred_config_item *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "deferred_config_item_nextt_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_deferred_config_item, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_item_nextt_set" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_deferred_config_item, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "deferred_config_item_nextt_set" "', argument " "2"" of type '" "struct deferred_config_item *""'"); 
  }
  arg2 = (struct deferred_config_item *)(argp2);
  if (arg1) (arg1)->nextt = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_deferred_config_item_nextt_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_item *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct deferred_config_item *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_deferred_config_item, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_item_nextt_get" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  result = (struct deferred_config_item *) ((arg1)->nextt);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_deferred_config_item, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_deferred_config_item_token_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_item *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "deferred_config_item_token_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_deferred_config_item, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_item_token_set" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "deferred_config_item_token_set" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  if (arg1) (arg1)->token = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_deferred_config_item_token_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_item *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  hamlib_token_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_deferred_config_item, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_item_token_get" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  result = (hamlib_token_t) ((arg1)->token);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_deferred_config_item_value_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_item *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "deferred_config_item_value_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_deferred_config_item, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_item_value_set" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "deferred_config_item_value_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->value);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->value = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->value = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_deferred_config_item_value_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_item *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_deferred_config_item, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_item_value_get" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  result = (char *) ((arg1)->value);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_deferred_config_item(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_item *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_deferred_config_item", 0, 0, 0)) SWIG_fail;
  result = (struct deferred_config_item *)calloc(1, sizeof(struct deferred_config_item));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_deferred_config_item, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_deferred_config_item(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_item *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_deferred_config_item, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_deferred_config_item" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *deferred_config_item_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_deferred_config_item, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *deferred_config_item_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_deferred_config_header_firstt_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_header *arg1 = 0 ;
  struct deferred_config_item *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "deferred_config_header_firstt_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_deferred_config_header, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_header_firstt_set" "', argument " "1"" of type '" "struct deferred_config_header *""'"); 
  }
  arg1 = (struct deferred_config_header *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_deferred_config_item, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "deferred_config_header_firstt_set" "', argument " "2"" of type '" "struct deferred_config_item *""'"); 
  }
  arg2 = (struct deferred_config_item *)(argp2);
  if (arg1) (arg1)->firstt = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_deferred_config_header_firstt_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_header *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct deferred_config_item *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_deferred_config_header, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_header_firstt_get" "', argument " "1"" of type '" "struct deferred_config_header *""'"); 
  }
  arg1 = (struct deferred_config_header *)(argp1);
  result = (struct deferred_config_item *) ((arg1)->firstt);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_deferred_config_item, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_deferred_config_header_lastt_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_header *arg1 = 0 ;
  struct deferred_config_item *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "deferred_config_header_lastt_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_deferred_config_header, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_header_lastt_set" "', argument " "1"" of type '" "struct deferred_config_header *""'"); 
  }
  arg1 = (struct deferred_config_header *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_deferred_config_item, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "deferred_config_header_lastt_set" "', argument " "2"" of type '" "struct deferred_config_item *""'"); 
  }
  arg2 = (struct deferred_config_item *)(argp2);
  if (arg1) (arg1)->lastt = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_deferred_config_header_lastt_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_header *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct deferred_config_item *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_deferred_config_header, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_header_lastt_get" "', argument " "1"" of type '" "struct deferred_config_header *""'"); 
  }
  arg1 = (struct deferred_config_header *)(argp1);
  result = (struct deferred_config_item *) ((arg1)->lastt);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_deferred_config_item, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_deferred_config_header(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_header *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_deferred_config_header", 0, 0, 0)) SWIG_fail;
  result = (struct deferred_config_header *)calloc(1, sizeof(struct deferred_config_header));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_deferred_config_header, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_deferred_config_header(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct deferred_config_header *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_deferred_config_header, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_deferred_config_header" "', argument " "1"" of type '" "struct deferred_config_header *""'"); 
  }
  arg1 = (struct deferred_config_header *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *deferred_config_header_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_deferred_config_header, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *deferred_config_header_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_rig_caps_rig_model_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  rig_model_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rig_model_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_model_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_rig_model_set" "', argument " "2"" of type '" "rig_model_t""'");
  } 
  arg2 = (rig_model_t)(val2);
  if (arg1) (arg1)->rig_model = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_model_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rig_model_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_model_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (rig_model_t) ((arg1)->rig_model);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_model_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_model_name_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->model_name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_mfg_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_mfg_name_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->mfg_name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_version_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_version_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->version);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_copyright_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_copyright_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->copyright);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_status_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  enum rig_status_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_status_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_status_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_status_set" "', argument " "2"" of type '" "enum rig_status_e""'");
  } 
  arg2 = (enum rig_status_e)(val2);
  if (arg1) (arg1)->status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_status_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum rig_status_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_status_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (enum rig_status_e) ((arg1)->status);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rig_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_type_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_rig_type_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->rig_type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_type_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->rig_type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_ptt_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  ptt_type_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_ptt_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_ptt_type_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_ptt_type_set" "', argument " "2"" of type '" "ptt_type_t""'");
  } 
  arg2 = (ptt_type_t)(val2);
  if (arg1) (arg1)->ptt_type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_ptt_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  ptt_type_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_ptt_type_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (ptt_type_t) ((arg1)->ptt_type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_dcd_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  dcd_type_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_dcd_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_dcd_type_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_dcd_type_set" "', argument " "2"" of type '" "dcd_type_t""'");
  } 
  arg2 = (dcd_type_t)(val2);
  if (arg1) (arg1)->dcd_type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_dcd_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  dcd_type_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_dcd_type_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (dcd_type_t) ((arg1)->dcd_type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_port_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  rig_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_port_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_port_type_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_port_type_set" "', argument " "2"" of type '" "rig_port_t""'");
  } 
  arg2 = (rig_port_t)(val2);
  if (arg1) (arg1)->port_type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_port_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rig_port_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_port_type_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (rig_port_t) ((arg1)->port_type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_serial_rate_min_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_serial_rate_min_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_rate_min_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_serial_rate_min_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_min = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_serial_rate_min_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_rate_min_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->serial_rate_min);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_serial_rate_max_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_serial_rate_max_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_rate_max_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_serial_rate_max_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_max = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_serial_rate_max_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_rate_max_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->serial_rate_max);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_serial_data_bits_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_serial_data_bits_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_data_bits_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_serial_data_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_data_bits = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_serial_data_bits_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_data_bits_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->serial_data_bits);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_serial_stop_bits_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_serial_stop_bits_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_stop_bits_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_serial_stop_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_stop_bits = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_serial_stop_bits_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_stop_bits_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->serial_stop_bits);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_serial_parity_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  enum serial_parity_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_serial_parity_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_parity_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_serial_parity_set" "', argument " "2"" of type '" "enum serial_parity_e""'");
  } 
  arg2 = (enum serial_parity_e)(val2);
  if (arg1) (arg1)->serial_parity = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_serial_parity_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum serial_parity_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_parity_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (enum serial_parity_e) ((arg1)->serial_parity);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_serial_handshake_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  enum serial_handshake_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_serial_handshake_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_handshake_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_serial_handshake_set" "', argument " "2"" of type '" "enum serial_handshake_e""'");
  } 
  arg2 = (enum serial_handshake_e)(val2);
  if (arg1) (arg1)->serial_handshake = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_serial_handshake_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum serial_handshake_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_handshake_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (enum serial_handshake_e) ((arg1)->serial_handshake);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_write_delay_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_write_delay_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_write_delay_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->write_delay = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_write_delay_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_write_delay_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->write_delay);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_post_write_delay_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_post_write_delay_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_post_write_delay_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_post_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->post_write_delay = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_post_write_delay_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_post_write_delay_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->post_write_delay);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_timeout_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_timeout_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_timeout_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_timeout_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->timeout = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_timeout_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_timeout_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->timeout);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_retry_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_retry_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_retry_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_retry_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->retry = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_retry_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_retry_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->retry);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_has_get_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_has_get_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_get_func_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_has_get_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_has_get_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_get_func_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_func);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_has_set_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_has_set_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_set_func_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_has_set_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_has_set_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_set_func_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_func);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_has_get_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_has_get_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_get_level_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_has_get_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_has_get_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_get_level_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_level);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_has_set_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_has_set_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_set_level_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_has_set_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_has_set_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_set_level_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_level);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_has_get_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_has_get_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_get_parm_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_has_get_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_has_get_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_get_parm_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_parm);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_has_set_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_has_set_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_set_parm_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_has_set_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_has_set_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_set_parm_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_parm);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_level_gran_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_level_gran_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_level_gran_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_level_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->level_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""level_gran""' of type '""gran_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_level_gran_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  gran_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_level_gran_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->level_gran);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_parm_gran_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_parm_gran_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_parm_gran_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_parm_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->parm_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""parm_gran""' of type '""gran_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_parm_gran_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  gran_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_parm_gran_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->parm_gran);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_extparms_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_extparms_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_extparms_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_extparms_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extparms = (struct confparams const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_extparms_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct confparams *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_extparms_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extparms);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_extlevels_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_extlevels_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_extlevels_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_extlevels_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extlevels = (struct confparams const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_extlevels_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct confparams *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_extlevels_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extlevels);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_extfuncs_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_extfuncs_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_extfuncs_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_extfuncs_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extfuncs = (struct confparams const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_extfuncs_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct confparams *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_extfuncs_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extfuncs);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_ext_tokens_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_ext_tokens_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_ext_tokens_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_ext_tokens_set" "', argument " "2"" of type '" "int *""'"); 
  }
  arg2 = (int *)(argp2);
  if (arg1) (arg1)->ext_tokens = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_ext_tokens_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_ext_tokens_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int *) ((arg1)->ext_tokens);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_ctcss_list_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  tone_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_ctcss_list_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_ctcss_list_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_unsigned_int, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_ctcss_list_set" "', argument " "2"" of type '" "tone_t *""'"); 
  }
  arg2 = (tone_t *)(argp2);
  if (arg1) (arg1)->ctcss_list = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_ctcss_list_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  tone_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_ctcss_list_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (tone_t *) ((arg1)->ctcss_list);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_dcs_list_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  tone_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_dcs_list_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_dcs_list_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_unsigned_int, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_dcs_list_set" "', argument " "2"" of type '" "tone_t *""'"); 
  }
  arg2 = (tone_t *)(argp2);
  if (arg1) (arg1)->dcs_list = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_dcs_list_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  tone_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_dcs_list_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (tone_t *) ((arg1)->dcs_list);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_preamp_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int *arg2 = (int *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_preamp_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_preamp_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_preamp_set" "', argument " "2"" of type '" "int [8]""'"); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(int *)&arg1->preamp[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""preamp""' of type '""int [8]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_preamp_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_preamp_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int *)(int *) ((arg1)->preamp);
  {
    int len,i;
    len = 8;
    resultobj = PyList_New(len);
    for (i = 0; i < len; i++) {
      PyList_SetItem(resultobj,i,PyInt_FromLong((long)result[i]));
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_attenuator_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int *arg2 = (int *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_attenuator_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_attenuator_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_attenuator_set" "', argument " "2"" of type '" "int [8]""'"); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(int *)&arg1->attenuator[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""attenuator""' of type '""int [8]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_attenuator_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_attenuator_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int *)(int *) ((arg1)->attenuator);
  {
    int len,i;
    len = 8;
    resultobj = PyList_New(len);
    for (i = 0; i < len; i++) {
      PyList_SetItem(resultobj,i,PyInt_FromLong((long)result[i]));
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_max_rit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_max_rit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_max_rit_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_max_rit_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->max_rit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_max_rit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  shortfreq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_max_rit_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (shortfreq_t) ((arg1)->max_rit);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_max_xit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_max_xit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_max_xit_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_max_xit_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->max_xit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_max_xit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  shortfreq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_max_xit_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (shortfreq_t) ((arg1)->max_xit);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_max_ifshift_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_max_ifshift_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_max_ifshift_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_max_ifshift_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->max_ifshift = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_max_ifshift_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  shortfreq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_max_ifshift_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (shortfreq_t) ((arg1)->max_ifshift);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_agc_level_count_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_agc_level_count_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_agc_level_count_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_agc_level_count_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->agc_level_count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_agc_level_count_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_agc_level_count_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->agc_level_count);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_agc_levels_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  enum agc_level_e *arg2 = (enum agc_level_e *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_agc_levels_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_agc_levels_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_agc_level_e, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_agc_levels_set" "', argument " "2"" of type '" "enum agc_level_e [8]""'"); 
  } 
  arg2 = (enum agc_level_e *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(enum agc_level_e *)&arg1->agc_levels[ii] = *((enum agc_level_e *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""agc_levels""' of type '""enum agc_level_e [8]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_agc_levels_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum agc_level_e *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_agc_levels_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (enum agc_level_e *)(enum agc_level_e *) ((arg1)->agc_levels);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_agc_level_e, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_announces_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  ann_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_announces_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_announces_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_announces_set" "', argument " "2"" of type '" "ann_t""'");
  } 
  arg2 = (ann_t)(val2);
  if (arg1) (arg1)->announces = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_announces_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  ann_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_announces_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (ann_t) ((arg1)->announces);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_vfo_ops_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  vfo_op_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_vfo_ops_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_vfo_ops_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_vfo_ops_set" "', argument " "2"" of type '" "vfo_op_t""'");
  } 
  arg2 = (vfo_op_t)(val2);
  if (arg1) (arg1)->vfo_ops = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_vfo_ops_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  vfo_op_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_vfo_ops_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (vfo_op_t) ((arg1)->vfo_ops);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_scan_ops_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  scan_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_scan_ops_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_scan_ops_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_scan_ops_set" "', argument " "2"" of type '" "scan_t""'");
  } 
  arg2 = (scan_t)(val2);
  if (arg1) (arg1)->scan_ops = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_scan_ops_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  scan_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_scan_ops_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (scan_t) ((arg1)->scan_ops);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_targetable_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_targetable_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_targetable_vfo_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_targetable_vfo_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->targetable_vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_targetable_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_targetable_vfo_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->targetable_vfo);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_transceive_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_transceive_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_transceive_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_transceive_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->transceive = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_transceive_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_transceive_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->transceive);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_bank_qty_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_bank_qty_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_bank_qty_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_bank_qty_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->bank_qty = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_bank_qty_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_bank_qty_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->bank_qty);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_chan_desc_sz_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_chan_desc_sz_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_chan_desc_sz_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_chan_desc_sz_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->chan_desc_sz = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_chan_desc_sz_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_chan_desc_sz_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->chan_desc_sz);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_chan_list_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  chan_t *arg2 = (chan_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_chan_list_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_chan_list_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_chan_list_set" "', argument " "2"" of type '" "chan_t [16]""'"); 
  } 
  arg2 = (chan_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)16; ++ii) *(chan_t *)&arg1->chan_list[ii] = *((chan_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""chan_list""' of type '""chan_t [16]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_chan_list_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  chan_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_chan_list_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (chan_t *)(chan_t *) ((arg1)->chan_list);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_chan_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rx_range_list1_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rx_range_list1_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list1_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_rx_range_list1_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list1[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list1""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rx_range_list1_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list1_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_tx_range_list1_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_tx_range_list1_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list1_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_tx_range_list1_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list1[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list1""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_tx_range_list1_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list1_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rx_range_list2_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rx_range_list2_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list2_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_rx_range_list2_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list2[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list2""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rx_range_list2_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list2_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_tx_range_list2_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_tx_range_list2_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list2_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_tx_range_list2_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list2[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list2""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_tx_range_list2_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list2_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rx_range_list3_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rx_range_list3_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list3_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_rx_range_list3_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list3[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list3""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rx_range_list3_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list3_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_tx_range_list3_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_tx_range_list3_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list3_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_tx_range_list3_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list3[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list3""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_tx_range_list3_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list3_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rx_range_list4_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rx_range_list4_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list4_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_rx_range_list4_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list4[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list4""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rx_range_list4_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list4_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_tx_range_list4_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_tx_range_list4_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list4_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_tx_range_list4_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list4[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list4""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_tx_range_list4_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list4_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rx_range_list5_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rx_range_list5_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list5_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_rx_range_list5_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list5[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list5""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rx_range_list5_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list5_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_tx_range_list5_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_tx_range_list5_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list5_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_tx_range_list5_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list5[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list5""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_tx_range_list5_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list5_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_tuning_steps_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  struct tuning_step_list *arg2 = (struct tuning_step_list *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_tuning_steps_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tuning_steps_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_tuning_step_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_tuning_steps_set" "', argument " "2"" of type '" "struct tuning_step_list [20]""'"); 
  } 
  arg2 = (struct tuning_step_list *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)20; ++ii) *(struct tuning_step_list *)&arg1->tuning_steps[ii] = *((struct tuning_step_list *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tuning_steps""' of type '""struct tuning_step_list [20]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_tuning_steps_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct tuning_step_list *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tuning_steps_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct tuning_step_list *)(struct tuning_step_list *) ((arg1)->tuning_steps);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_tuning_step_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_filters_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  struct filter_list *arg2 = (struct filter_list *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_filters_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_filters_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_filter_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_filters_set" "', argument " "2"" of type '" "struct filter_list [60]""'"); 
  } 
  arg2 = (struct filter_list *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)60; ++ii) *(struct filter_list *)&arg1->filters[ii] = *((struct filter_list *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""filters""' of type '""struct filter_list [60]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_filters_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct filter_list *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_filters_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct filter_list *)(struct filter_list *) ((arg1)->filters);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_filter_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_str_cal_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  cal_table_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_str_cal_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_str_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_cal_table, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_str_cal_set" "', argument " "2"" of type '" "cal_table_t *""'"); 
  }
  arg2 = (cal_table_t *)(argp2);
  if (arg1) (arg1)->str_cal = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_str_cal_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  cal_table_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_str_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_t *)& ((arg1)->str_cal);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_swr_cal_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  cal_table_float_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_swr_cal_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_swr_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_swr_cal_set" "', argument " "2"" of type '" "cal_table_float_t *""'"); 
  }
  arg2 = (cal_table_float_t *)(argp2);
  if (arg1) (arg1)->swr_cal = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_swr_cal_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  cal_table_float_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_swr_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_float_t *)& ((arg1)->swr_cal);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_alc_cal_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  cal_table_float_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_alc_cal_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_alc_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_alc_cal_set" "', argument " "2"" of type '" "cal_table_float_t *""'"); 
  }
  arg2 = (cal_table_float_t *)(argp2);
  if (arg1) (arg1)->alc_cal = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_alc_cal_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  cal_table_float_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_alc_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_float_t *)& ((arg1)->alc_cal);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rfpower_meter_cal_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  cal_table_float_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rfpower_meter_cal_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rfpower_meter_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_rfpower_meter_cal_set" "', argument " "2"" of type '" "cal_table_float_t *""'"); 
  }
  arg2 = (cal_table_float_t *)(argp2);
  if (arg1) (arg1)->rfpower_meter_cal = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rfpower_meter_cal_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  cal_table_float_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rfpower_meter_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_float_t *)& ((arg1)->rfpower_meter_cal);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_comp_meter_cal_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  cal_table_float_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_comp_meter_cal_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_comp_meter_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_comp_meter_cal_set" "', argument " "2"" of type '" "cal_table_float_t *""'"); 
  }
  arg2 = (cal_table_float_t *)(argp2);
  if (arg1) (arg1)->comp_meter_cal = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_comp_meter_cal_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  cal_table_float_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_comp_meter_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_float_t *)& ((arg1)->comp_meter_cal);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_vd_meter_cal_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  cal_table_float_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_vd_meter_cal_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_vd_meter_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_vd_meter_cal_set" "', argument " "2"" of type '" "cal_table_float_t *""'"); 
  }
  arg2 = (cal_table_float_t *)(argp2);
  if (arg1) (arg1)->vd_meter_cal = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_vd_meter_cal_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  cal_table_float_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_vd_meter_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_float_t *)& ((arg1)->vd_meter_cal);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_id_meter_cal_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  cal_table_float_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_id_meter_cal_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_id_meter_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_id_meter_cal_set" "', argument " "2"" of type '" "cal_table_float_t *""'"); 
  }
  arg2 = (cal_table_float_t *)(argp2);
  if (arg1) (arg1)->id_meter_cal = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_id_meter_cal_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  cal_table_float_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_id_meter_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_float_t *)& ((arg1)->id_meter_cal);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_spectrum_scopes_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  struct rig_spectrum_scope *arg2 = (struct rig_spectrum_scope *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_spectrum_scopes_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_scopes_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_spectrum_scopes_set" "', argument " "2"" of type '" "struct rig_spectrum_scope [4]""'"); 
  } 
  arg2 = (struct rig_spectrum_scope *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)4; ++ii) *(struct rig_spectrum_scope *)&arg1->spectrum_scopes[ii] = *((struct rig_spectrum_scope *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_scopes""' of type '""struct rig_spectrum_scope [4]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_spectrum_scopes_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct rig_spectrum_scope *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_scopes_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct rig_spectrum_scope *)(struct rig_spectrum_scope *) ((arg1)->spectrum_scopes);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_spectrum_modes_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  enum rig_spectrum_mode_e *arg2 = (enum rig_spectrum_mode_e *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_spectrum_modes_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_modes_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_rig_spectrum_mode_e, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_spectrum_modes_set" "', argument " "2"" of type '" "enum rig_spectrum_mode_e [5]""'"); 
  } 
  arg2 = (enum rig_spectrum_mode_e *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)5; ++ii) *(enum rig_spectrum_mode_e *)&arg1->spectrum_modes[ii] = *((enum rig_spectrum_mode_e *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_modes""' of type '""enum rig_spectrum_mode_e [5]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_spectrum_modes_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum rig_spectrum_mode_e *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_modes_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (enum rig_spectrum_mode_e *)(enum rig_spectrum_mode_e *) ((arg1)->spectrum_modes);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_mode_e, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_spectrum_spans_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  freq_t *arg2 = (freq_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_spectrum_spans_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_spans_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_spectrum_spans_set" "', argument " "2"" of type '" "freq_t [20]""'"); 
  } 
  arg2 = (freq_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)20; ++ii) *(freq_t *)&arg1->spectrum_spans[ii] = *((freq_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_spans""' of type '""freq_t [20]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_spectrum_spans_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_spans_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_t *)(freq_t *) ((arg1)->spectrum_spans);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_spectrum_avg_modes_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  struct rig_spectrum_avg_mode *arg2 = (struct rig_spectrum_avg_mode *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_spectrum_avg_modes_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_avg_modes_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_spectrum_avg_modes_set" "', argument " "2"" of type '" "struct rig_spectrum_avg_mode [12]""'"); 
  } 
  arg2 = (struct rig_spectrum_avg_mode *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)12; ++ii) *(struct rig_spectrum_avg_mode *)&arg1->spectrum_avg_modes[ii] = *((struct rig_spectrum_avg_mode *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_avg_modes""' of type '""struct rig_spectrum_avg_mode [12]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_spectrum_avg_modes_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct rig_spectrum_avg_mode *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_avg_modes_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct rig_spectrum_avg_mode *)(struct rig_spectrum_avg_mode *) ((arg1)->spectrum_avg_modes);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_spectrum_attenuator_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int *arg2 = (int *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_spectrum_attenuator_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_attenuator_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_spectrum_attenuator_set" "', argument " "2"" of type '" "int [8]""'"); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(int *)&arg1->spectrum_attenuator[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_attenuator""' of type '""int [8]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_spectrum_attenuator_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_attenuator_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int *)(int *) ((arg1)->spectrum_attenuator);
  {
    int len,i;
    len = 8;
    resultobj = PyList_New(len);
    for (i = 0; i < len; i++) {
      PyList_SetItem(resultobj,i,PyInt_FromLong((long)result[i]));
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_cfgparams_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_cfgparams_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_cfgparams_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_cfgparams_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->cfgparams = (struct confparams const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_cfgparams_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct confparams *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_cfgparams_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct confparams *) ((arg1)->cfgparams);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_priv_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_priv_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_priv_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_priv_set" "', argument " "2"" of type '" "void const *""'"); 
  }
  if (arg1) (arg1)->priv = (void const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_priv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_priv_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (void *) ((arg1)->priv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_init_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rig_init_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_init_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_rig_init_set" "', argument " "2"" of type '" "int (*)(RIG *)""'"); 
    }
  }
  if (arg1) (arg1)->rig_init = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_init_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_init_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *)) ((arg1)->rig_init);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_cleanup_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rig_cleanup_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_cleanup_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_rig_cleanup_set" "', argument " "2"" of type '" "int (*)(RIG *)""'"); 
    }
  }
  if (arg1) (arg1)->rig_cleanup = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_cleanup_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_cleanup_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *)) ((arg1)->rig_cleanup);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_open_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rig_open_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_open_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_rig_open_set" "', argument " "2"" of type '" "int (*)(RIG *)""'"); 
    }
  }
  if (arg1) (arg1)->rig_open = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_open_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_open_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *)) ((arg1)->rig_open);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_close_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rig_close_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_close_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_rig_close_set" "', argument " "2"" of type '" "int (*)(RIG *)""'"); 
    }
  }
  if (arg1) (arg1)->rig_close = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_close_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_close_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *)) ((arg1)->rig_close);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_freq_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_freq_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,freq_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_freq_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t)) ((arg1)->set_freq);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_freq_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_freq_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,freq_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_freq_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t *)) ((arg1)->get_freq);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,rmode_t,pbwidth_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,rmode_t,pbwidth_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,rmode_t,pbwidth_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,rmode_t,pbwidth_t)) ((arg1)->set_mode);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,rmode_t *,pbwidth_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,rmode_t *,pbwidth_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,rmode_t *,pbwidth_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,rmode_t *,pbwidth_t *)) ((arg1)->get_mode);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_vfo_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_vfo_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_vfo_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t)) ((arg1)->set_vfo);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_vfo_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_vfo_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_vfo_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t *)) ((arg1)->get_vfo);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ptt_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,ptt_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_ptt_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ptt_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ptt_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,ptt_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ptt = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ptt_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,ptt_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ptt_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,ptt_t)) ((arg1)->set_ptt);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ptt_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,ptt_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_ptt_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ptt_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ptt_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,ptt_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ptt = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ptt_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,ptt_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ptt_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,ptt_t *)) ((arg1)->get_ptt);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_dcd_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,dcd_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_dcd_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_dcd_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_dcd_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,dcd_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_dcd = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_dcd_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,dcd_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_dcd_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,dcd_t *)) ((arg1)->get_dcd);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_rptr_shift_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,rptr_shift_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_rptr_shift_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_rptr_shift_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_rptr_shift_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,rptr_shift_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_rptr_shift = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_rptr_shift_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,rptr_shift_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_rptr_shift_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,rptr_shift_t)) ((arg1)->set_rptr_shift);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_rptr_shift_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,rptr_shift_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_rptr_shift_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_rptr_shift_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_rptr_shift_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,rptr_shift_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_rptr_shift = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_rptr_shift_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,rptr_shift_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_rptr_shift_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,rptr_shift_t *)) ((arg1)->get_rptr_shift);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_rptr_offs_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_rptr_offs_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_rptr_offs_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_rptr_offs_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_rptr_offs = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_rptr_offs_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,shortfreq_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_rptr_offs_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t)) ((arg1)->set_rptr_offs);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_rptr_offs_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_rptr_offs_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_rptr_offs_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_rptr_offs_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_rptr_offs = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_rptr_offs_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_rptr_offs_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t *)) ((arg1)->get_rptr_offs);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_split_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_split_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_freq_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_split_freq_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_split_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_split_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,freq_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_freq_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t)) ((arg1)->set_split_freq);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_split_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_split_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_freq_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_split_freq_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_split_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_split_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,freq_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_freq_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t *)) ((arg1)->get_split_freq);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_split_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,rmode_t,pbwidth_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_split_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_split_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,rmode_t,pbwidth_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_split_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_split_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,rmode_t,pbwidth_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,rmode_t,pbwidth_t)) ((arg1)->set_split_mode);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_split_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,rmode_t *,pbwidth_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_split_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_split_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,rmode_t *,pbwidth_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_split_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_split_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,rmode_t *,pbwidth_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,rmode_t *,pbwidth_t *)) ((arg1)->get_split_mode);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_split_freq_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t,rmode_t,pbwidth_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_split_freq_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_freq_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_split_freq_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t,rmode_t,pbwidth_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_split_freq_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_split_freq_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,freq_t,rmode_t,pbwidth_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_freq_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t,rmode_t,pbwidth_t)) ((arg1)->set_split_freq_mode);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_split_freq_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_split_freq_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_freq_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_split_freq_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_split_freq_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_split_freq_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_freq_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *)) ((arg1)->get_split_freq_mode);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_split_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,split_t,vfo_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_split_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_vfo_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_split_vfo_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,split_t,vfo_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_split_vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_split_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,split_t,vfo_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_vfo_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,split_t,vfo_t)) ((arg1)->set_split_vfo);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_split_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,split_t *,vfo_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_split_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_vfo_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_split_vfo_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,split_t *,vfo_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_split_vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_split_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,split_t *,vfo_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_vfo_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,split_t *,vfo_t *)) ((arg1)->get_split_vfo);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_rit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_rit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_rit_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_rit_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_rit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_rit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,shortfreq_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_rit_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t)) ((arg1)->set_rit);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_rit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_rit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_rit_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_rit_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_rit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_rit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_rit_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t *)) ((arg1)->get_rit);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_xit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_xit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_xit_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_xit_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_xit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_xit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,shortfreq_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_xit_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t)) ((arg1)->set_xit);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_xit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_xit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_xit_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_xit_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_xit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_xit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_xit_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t *)) ((arg1)->get_xit);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ts_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_ts_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ts_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ts_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ts = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ts_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,shortfreq_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ts_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t)) ((arg1)->set_ts);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ts_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_ts_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ts_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ts_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ts = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ts_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ts_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t *)) ((arg1)->get_ts);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_dcs_code_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_dcs_code_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_dcs_code_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_dcs_code_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_dcs_code = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_dcs_code_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,tone_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_dcs_code_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t)) ((arg1)->set_dcs_code);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_dcs_code_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_dcs_code_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_dcs_code_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_dcs_code_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_dcs_code = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_dcs_code_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,tone_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_dcs_code_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t *)) ((arg1)->get_dcs_code);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_tone_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_tone_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_tone_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_tone_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_tone = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_tone_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,tone_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_tone_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t)) ((arg1)->set_tone);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_tone_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_tone_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_tone_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_tone_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_tone = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_tone_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,tone_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_tone_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t *)) ((arg1)->get_tone);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ctcss_tone_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_ctcss_tone_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ctcss_tone_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ctcss_tone_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ctcss_tone = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ctcss_tone_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,tone_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ctcss_tone_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t)) ((arg1)->set_ctcss_tone);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ctcss_tone_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_ctcss_tone_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ctcss_tone_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ctcss_tone_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ctcss_tone = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ctcss_tone_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,tone_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ctcss_tone_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t *)) ((arg1)->get_ctcss_tone);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_dcs_sql_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_dcs_sql_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_dcs_sql_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_dcs_sql_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_dcs_sql = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_dcs_sql_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,tone_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_dcs_sql_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t)) ((arg1)->set_dcs_sql);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_dcs_sql_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_dcs_sql_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_dcs_sql_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_dcs_sql_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_dcs_sql = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_dcs_sql_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,tone_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_dcs_sql_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t *)) ((arg1)->get_dcs_sql);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_tone_sql_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_tone_sql_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_tone_sql_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_tone_sql_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_tone_sql = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_tone_sql_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,tone_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_tone_sql_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t)) ((arg1)->set_tone_sql);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_tone_sql_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_tone_sql_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_tone_sql_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_tone_sql_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_tone_sql = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_tone_sql_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,tone_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_tone_sql_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t *)) ((arg1)->get_tone_sql);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ctcss_sql_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_ctcss_sql_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ctcss_sql_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ctcss_sql_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ctcss_sql = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ctcss_sql_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,tone_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ctcss_sql_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t)) ((arg1)->set_ctcss_sql);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ctcss_sql_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_ctcss_sql_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ctcss_sql_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ctcss_sql_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ctcss_sql = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ctcss_sql_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,tone_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ctcss_sql_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t *)) ((arg1)->get_ctcss_sql);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_power2mW_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,unsigned int *,float,freq_t,rmode_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_power2mW_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_power2mW_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_power2mW_set" "', argument " "2"" of type '" "int (*)(RIG *,unsigned int *,float,freq_t,rmode_t)""'"); 
    }
  }
  if (arg1) (arg1)->power2mW = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_power2mW_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,unsigned int *,float,freq_t,rmode_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_power2mW_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,unsigned int *,float,freq_t,rmode_t)) ((arg1)->power2mW);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_mW2power_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,float *,unsigned int,freq_t,rmode_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_mW2power_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_mW2power_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_mW2power_set" "', argument " "2"" of type '" "int (*)(RIG *,float *,unsigned int,freq_t,rmode_t)""'"); 
    }
  }
  if (arg1) (arg1)->mW2power = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_mW2power_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,float *,unsigned int,freq_t,rmode_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_mW2power_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,float *,unsigned int,freq_t,rmode_t)) ((arg1)->mW2power);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_powerstat_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,powerstat_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_powerstat_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_powerstat_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_enum_powerstat_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_powerstat_set" "', argument " "2"" of type '" "int (*)(RIG *,powerstat_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_powerstat = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_powerstat_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,powerstat_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_powerstat_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,powerstat_t)) ((arg1)->set_powerstat);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_enum_powerstat_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_powerstat_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,powerstat_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_powerstat_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_powerstat_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_enum_powerstat_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_powerstat_set" "', argument " "2"" of type '" "int (*)(RIG *,powerstat_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_powerstat = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_powerstat_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,powerstat_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_powerstat_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,powerstat_t *)) ((arg1)->get_powerstat);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_enum_powerstat_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_reset_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,reset_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_reset_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_reset_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_enum_reset_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_reset_set" "', argument " "2"" of type '" "int (*)(RIG *,reset_t)""'"); 
    }
  }
  if (arg1) (arg1)->reset = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_reset_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,reset_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_reset_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,reset_t)) ((arg1)->reset);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_enum_reset_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ant_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,ant_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_ant_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ant_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ant_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,ant_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ant = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ant_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,ant_t,value_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ant_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,ant_t,value_t)) ((arg1)->set_ant);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ant_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,ant_t,value_t *,ant_t *,ant_t *,ant_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_ant_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ant_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ant_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,ant_t,value_t *,ant_t *,ant_t *,ant_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ant = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ant_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,ant_t,value_t *,ant_t *,ant_t *,ant_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ant_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,ant_t,value_t *,ant_t *,ant_t *,ant_t *)) ((arg1)->get_ant);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,setting_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_level_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_level_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,setting_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,setting_t,value_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_level_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,setting_t,value_t)) ((arg1)->set_level);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,setting_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_level_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_level_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,setting_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,setting_t,value_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_level_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,setting_t,value_t *)) ((arg1)->get_level);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,setting_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_func_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_func_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,setting_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,setting_t,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_func_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,setting_t,int)) ((arg1)->set_func);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,setting_t,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_func_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_func_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,setting_t,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,setting_t,int *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_func_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,setting_t,int *)) ((arg1)->get_func);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,setting_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_parm_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_long_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_parm_set" "', argument " "2"" of type '" "int (*)(RIG *,setting_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,setting_t,value_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_parm_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,setting_t,value_t)) ((arg1)->set_parm);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_long_long_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,setting_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_parm_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_parm_set" "', argument " "2"" of type '" "int (*)(RIG *,setting_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,setting_t,value_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_parm_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,setting_t,value_t *)) ((arg1)->get_parm);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ext_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,hamlib_token_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_ext_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ext_level_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ext_level_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,hamlib_token_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ext_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,hamlib_token_t,value_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ext_level_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,hamlib_token_t,value_t)) ((arg1)->set_ext_level);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ext_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,hamlib_token_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_ext_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ext_level_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ext_level_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,hamlib_token_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ext_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,hamlib_token_t,value_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ext_level_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,hamlib_token_t,value_t *)) ((arg1)->get_ext_level);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ext_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,hamlib_token_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_ext_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ext_func_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ext_func_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,hamlib_token_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ext_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,hamlib_token_t,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ext_func_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,hamlib_token_t,int)) ((arg1)->set_ext_func);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ext_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,hamlib_token_t,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_ext_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ext_func_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ext_func_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,hamlib_token_t,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ext_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,hamlib_token_t,int *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ext_func_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,hamlib_token_t,int *)) ((arg1)->get_ext_func);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_p_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ext_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,hamlib_token_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_ext_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ext_parm_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ext_parm_set" "', argument " "2"" of type '" "int (*)(RIG *,hamlib_token_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_ext_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,hamlib_token_t,value_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ext_parm_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,hamlib_token_t,value_t)) ((arg1)->set_ext_parm);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_long_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ext_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,hamlib_token_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_ext_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ext_parm_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ext_parm_set" "', argument " "2"" of type '" "int (*)(RIG *,hamlib_token_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_ext_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,hamlib_token_t,value_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ext_parm_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,hamlib_token_t,value_t *)) ((arg1)->get_ext_parm);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_long_p_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_conf_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,hamlib_token_t,char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_conf_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_conf_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_long_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_conf_set" "', argument " "2"" of type '" "int (*)(RIG *,hamlib_token_t,char const *)""'"); 
    }
  }
  if (arg1) (arg1)->set_conf = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_conf_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,hamlib_token_t,char const *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_conf_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,hamlib_token_t,char const *)) ((arg1)->set_conf);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_long_p_q_const__char__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_conf_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,hamlib_token_t,char *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_conf_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_conf_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_long_p_char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_conf_set" "', argument " "2"" of type '" "int (*)(RIG *,hamlib_token_t,char *)""'"); 
    }
  }
  if (arg1) (arg1)->get_conf = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_conf_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,hamlib_token_t,char *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_conf_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,hamlib_token_t,char *)) ((arg1)->get_conf);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_long_p_char__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_send_dtmf_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_send_dtmf_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_send_dtmf_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_send_dtmf_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,char const *)""'"); 
    }
  }
  if (arg1) (arg1)->send_dtmf = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_send_dtmf_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,char const *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_send_dtmf_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,char const *)) ((arg1)->send_dtmf);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_recv_dtmf_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,char *,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_recv_dtmf_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_recv_dtmf_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_recv_dtmf_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,char *,int *)""'"); 
    }
  }
  if (arg1) (arg1)->recv_dtmf = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_recv_dtmf_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,char *,int *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_recv_dtmf_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,char *,int *)) ((arg1)->recv_dtmf);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_send_morse_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_send_morse_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_send_morse_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_send_morse_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,char const *)""'"); 
    }
  }
  if (arg1) (arg1)->send_morse = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_send_morse_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,char const *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_send_morse_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,char const *)) ((arg1)->send_morse);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_stop_morse_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_stop_morse_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_stop_morse_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_stop_morse_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t)""'"); 
    }
  }
  if (arg1) (arg1)->stop_morse = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_stop_morse_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_stop_morse_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t)) ((arg1)->stop_morse);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_wait_morse_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_wait_morse_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_wait_morse_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_wait_morse_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t)""'"); 
    }
  }
  if (arg1) (arg1)->wait_morse = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_wait_morse_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_wait_morse_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t)) ((arg1)->wait_morse);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_send_voice_mem_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_send_voice_mem_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_send_voice_mem_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_send_voice_mem_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->send_voice_mem = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_send_voice_mem_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_send_voice_mem_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,int)) ((arg1)->send_voice_mem);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_stop_voice_mem_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_stop_voice_mem_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_stop_voice_mem_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_stop_voice_mem_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t)""'"); 
    }
  }
  if (arg1) (arg1)->stop_voice_mem = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_stop_voice_mem_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_stop_voice_mem_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t)) ((arg1)->stop_voice_mem);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_bank_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_bank_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_bank_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_bank_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_bank = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_bank_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_bank_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,int)) ((arg1)->set_bank);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_mem_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_mem_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_mem_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_mem_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_mem = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_mem_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_mem_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,int)) ((arg1)->set_mem);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_mem_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_mem_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_mem_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_mem_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_mem = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_mem_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,int *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_mem_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,int *)) ((arg1)->get_mem);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_vfo_op_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,vfo_op_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_vfo_op_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_vfo_op_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_vfo_op_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,vfo_op_t)""'"); 
    }
  }
  if (arg1) (arg1)->vfo_op = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_vfo_op_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,vfo_op_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_vfo_op_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,vfo_op_t)) ((arg1)->vfo_op);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_scan_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,scan_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_scan_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_scan_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_scan_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,scan_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->scan = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_scan_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,scan_t,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_scan_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,scan_t,int)) ((arg1)->scan);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_trn_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_trn_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_trn_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_trn_set" "', argument " "2"" of type '" "int (*)(RIG *,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_trn = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_trn_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_trn_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int)) ((arg1)->set_trn);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_trn_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_trn_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_trn_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_trn_set" "', argument " "2"" of type '" "int (*)(RIG *,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_trn = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_trn_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,int *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_trn_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int *)) ((arg1)->get_trn);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_decode_event_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_decode_event_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_decode_event_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_decode_event_set" "', argument " "2"" of type '" "int (*)(RIG *)""'"); 
    }
  }
  if (arg1) (arg1)->decode_event = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_decode_event_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_decode_event_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *)) ((arg1)->decode_event);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_channel_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,channel_t const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_channel_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_channel_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_channel_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,channel_t const *)""'"); 
    }
  }
  if (arg1) (arg1)->set_channel = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_channel_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,channel_t const *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_channel_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,channel_t const *)) ((arg1)->set_channel);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_channel_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,channel_t *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_channel_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_channel_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_channel_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,channel_t *,int)""'"); 
    }
  }
  if (arg1) (arg1)->get_channel = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_channel_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,channel_t *,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_channel_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,channel_t *,int)) ((arg1)->get_channel);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_info_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  char *(*arg2)(RIG *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_info_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_info_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig__p_char);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_info_set" "', argument " "2"" of type '" "char const *(*)(RIG *)""'"); 
    }
  }
  if (arg1) (arg1)->get_info = (char const *(*)(RIG *))arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_info_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *(*result)(RIG *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_info_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *(*)(RIG *)) ((arg1)->get_info);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig__p_char);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_chan_all_cb_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,chan_cb_t,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_chan_all_cb_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_chan_all_cb_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_chan_all_cb_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,chan_cb_t,void *)""'"); 
    }
  }
  if (arg1) (arg1)->set_chan_all_cb = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_chan_all_cb_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,chan_cb_t,void *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_chan_all_cb_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,chan_cb_t,void *)) ((arg1)->set_chan_all_cb);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_chan_all_cb_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,chan_cb_t,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_chan_all_cb_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_chan_all_cb_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_chan_all_cb_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,chan_cb_t,void *)""'"); 
    }
  }
  if (arg1) (arg1)->get_chan_all_cb = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_chan_all_cb_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,chan_cb_t,void *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_chan_all_cb_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,chan_cb_t,void *)) ((arg1)->get_chan_all_cb);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_mem_all_cb_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_mem_all_cb_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_mem_all_cb_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_mem_all_cb_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *)""'"); 
    }
  }
  if (arg1) (arg1)->set_mem_all_cb = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_mem_all_cb_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_mem_all_cb_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *)) ((arg1)->set_mem_all_cb);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_mem_all_cb_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_mem_all_cb_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_mem_all_cb_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_mem_all_cb_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *)""'"); 
    }
  }
  if (arg1) (arg1)->get_mem_all_cb = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_mem_all_cb_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_mem_all_cb_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *)) ((arg1)->get_mem_all_cb);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_vfo_opt_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_vfo_opt_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_vfo_opt_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_vfo_opt_set" "', argument " "2"" of type '" "int (*)(RIG *,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_vfo_opt = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_vfo_opt_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_vfo_opt_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int)) ((arg1)->set_vfo_opt);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_get_vfo_info_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *,split_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_rig_get_vfo_info_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_get_vfo_info_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_rig_get_vfo_info_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *,split_t *)""'"); 
    }
  }
  if (arg1) (arg1)->rig_get_vfo_info = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_rig_get_vfo_info_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *,split_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_get_vfo_info_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *,split_t *)) ((arg1)->rig_get_vfo_info);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_clock_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int,int,int,int,int,int,double,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_clock_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_clock_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_clock_set" "', argument " "2"" of type '" "int (*)(RIG *,int,int,int,int,int,int,double,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_clock = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_clock_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,int,int,int,int,int,int,double,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_clock_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int,int,int,int,int,int,double,int)) ((arg1)->set_clock);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_clock_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int *,int *,int *,int *,int *,int *,double *,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_clock_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_clock_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_clock_set" "', argument " "2"" of type '" "int (*)(RIG *,int *,int *,int *,int *,int *,int *,double *,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_clock = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_clock_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,int *,int *,int *,int *,int *,int *,double *,int *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_clock_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int *,int *,int *,int *,int *,int *,double *,int *)) ((arg1)->get_clock);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_clone_combo_set_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_clone_combo_set_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->clone_combo_set);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_clone_combo_get_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_clone_combo_get_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->clone_combo_get);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_async_data_supported_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_async_data_supported_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_async_data_supported_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_async_data_supported_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->async_data_supported = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_async_data_supported_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_async_data_supported_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->async_data_supported);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_read_frame_direct_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,size_t,unsigned char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_read_frame_direct_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_read_frame_direct_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_read_frame_direct_set" "', argument " "2"" of type '" "int (*)(RIG *,size_t,unsigned char const *)""'"); 
    }
  }
  if (arg1) (arg1)->read_frame_direct = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_read_frame_direct_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,size_t,unsigned char const *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_read_frame_direct_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,size_t,unsigned char const *)) ((arg1)->read_frame_direct);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_is_async_frame_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,size_t,unsigned char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_is_async_frame_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_is_async_frame_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_is_async_frame_set" "', argument " "2"" of type '" "int (*)(RIG *,size_t,unsigned char const *)""'"); 
    }
  }
  if (arg1) (arg1)->is_async_frame = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_is_async_frame_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,size_t,unsigned char const *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_is_async_frame_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,size_t,unsigned char const *)) ((arg1)->is_async_frame);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_process_async_frame_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,size_t,unsigned char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_process_async_frame_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_process_async_frame_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_process_async_frame_set" "', argument " "2"" of type '" "int (*)(RIG *,size_t,unsigned char const *)""'"); 
    }
  }
  if (arg1) (arg1)->process_async_frame = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_process_async_frame_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,size_t,unsigned char const *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_process_async_frame_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,size_t,unsigned char const *)) ((arg1)->process_async_frame);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_hamlib_check_rig_caps_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_hamlib_check_rig_caps_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_hamlib_check_rig_caps_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_hamlib_check_rig_caps_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->hamlib_check_rig_caps);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->hamlib_check_rig_caps = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->hamlib_check_rig_caps = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_hamlib_check_rig_caps_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_hamlib_check_rig_caps_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->hamlib_check_rig_caps);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_conf2_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,hamlib_token_t,char *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_conf2_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_conf2_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_long_p_char_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_conf2_set" "', argument " "2"" of type '" "int (*)(RIG *,hamlib_token_t,char *,int)""'"); 
    }
  }
  if (arg1) (arg1)->get_conf2 = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_conf2_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,hamlib_token_t,char *,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_conf2_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,hamlib_token_t,char *,int)) ((arg1)->get_conf2);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_long_p_char_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_password_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_password_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_password_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_password_set" "', argument " "2"" of type '" "int (*)(RIG *,char const *)""'"); 
    }
  }
  if (arg1) (arg1)->password = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_password_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,char const *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_password_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,char const *)) ((arg1)->password);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_q_const__char__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_lock_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_set_lock_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_lock_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_lock_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_lock_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_set_lock_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_lock_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int)) ((arg1)->set_lock_mode);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_lock_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_get_lock_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_lock_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_lock_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_lock_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_get_lock_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(RIG *,int *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_lock_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int *)) ((arg1)->get_lock_mode);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_timeout_retry_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  short arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  short val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_timeout_retry_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_timeout_retry_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_timeout_retry_set" "', argument " "2"" of type '" "short""'");
  } 
  arg2 = (short)(val2);
  if (arg1) (arg1)->timeout_retry = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_timeout_retry_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  short result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_timeout_retry_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (short) ((arg1)->timeout_retry);
  resultobj = SWIG_From_short((short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_morse_qsize_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  short arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  short val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_caps_morse_qsize_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_morse_qsize_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_morse_qsize_set" "', argument " "2"" of type '" "short""'");
  } 
  arg2 = (short)(val2);
  if (arg1) (arg1)->morse_qsize = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_caps_morse_qsize_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  short result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_morse_qsize_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (short) ((arg1)->morse_qsize);
  resultobj = SWIG_From_short((short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_rig_caps(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_rig_caps", 0, 0, 0)) SWIG_fail;
  result = (struct rig_caps *)calloc(1, sizeof(struct rig_caps));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_caps, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_rig_caps(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_caps, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rig_caps" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *rig_caps_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_rig_caps, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *rig_caps_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_multicast_s_multicast_running_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "multicast_s_multicast_running_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_multicast_running_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "multicast_s_multicast_running_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->multicast_running = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_multicast_running_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_multicast_running_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result = (int) ((arg1)->multicast_running);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_sock_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "multicast_s_sock_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_sock_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "multicast_s_sock_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->sock = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_sock_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_sock_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result = (int) ((arg1)->sock);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_seqnumber_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "multicast_s_seqnumber_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_seqnumber_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "multicast_s_seqnumber_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->seqnumber = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_seqnumber_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_seqnumber_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result = (int) ((arg1)->seqnumber);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_runflag_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "multicast_s_runflag_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_runflag_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "multicast_s_runflag_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->runflag = (int volatile )arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_runflag_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_runflag_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result = (int)(int) ((arg1)->runflag);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_threadid_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  pthread_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "multicast_s_threadid_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_threadid_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_pthread_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "multicast_s_threadid_set" "', argument " "2"" of type '" "pthread_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "multicast_s_threadid_set" "', argument " "2"" of type '" "pthread_t""'");
    } else {
      arg2 = *((pthread_t *)(argp2));
    }
  }
  if (arg1) (arg1)->threadid = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_threadid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  pthread_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_threadid_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result =  ((arg1)->threadid);
  resultobj = SWIG_NewPointerObj((pthread_t *)memcpy((pthread_t *)calloc(1,sizeof(pthread_t)),&result,sizeof(pthread_t)), SWIGTYPE_p_pthread_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_mutex_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  pthread_mutex_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "multicast_s_mutex_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_mutex_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_pthread_mutex_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "multicast_s_mutex_set" "', argument " "2"" of type '" "pthread_mutex_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "multicast_s_mutex_set" "', argument " "2"" of type '" "pthread_mutex_t""'");
    } else {
      arg2 = *((pthread_mutex_t *)(argp2));
    }
  }
  if (arg1) (arg1)->mutex = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_mutex_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  pthread_mutex_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_mutex_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result =  ((arg1)->mutex);
  resultobj = SWIG_NewPointerObj((pthread_mutex_t *)memcpy((pthread_mutex_t *)calloc(1,sizeof(pthread_mutex_t)),&result,sizeof(pthread_mutex_t)), SWIGTYPE_p_pthread_mutex_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_mutex_initialized_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "multicast_s_mutex_initialized_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_mutex_initialized_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "multicast_s_mutex_initialized_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->mutex_initialized = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_mutex_initialized_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_mutex_initialized_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result = (int) ((arg1)->mutex_initialized);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_dest_addr_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  struct sockaddr_in arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "multicast_s_dest_addr_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_dest_addr_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_sockaddr_in,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "multicast_s_dest_addr_set" "', argument " "2"" of type '" "struct sockaddr_in""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "multicast_s_dest_addr_set" "', argument " "2"" of type '" "struct sockaddr_in""'");
    } else {
      arg2 = *((struct sockaddr_in *)(argp2));
    }
  }
  if (arg1) (arg1)->dest_addr = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_dest_addr_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct sockaddr_in result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_dest_addr_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result =  ((arg1)->dest_addr);
  resultobj = SWIG_NewPointerObj((struct sockaddr_in *)memcpy((struct sockaddr_in *)calloc(1,sizeof(struct sockaddr_in)),&result,sizeof(struct sockaddr_in)), SWIGTYPE_p_sockaddr_in, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_port_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "multicast_s_port_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_port_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "multicast_s_port_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->port = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_multicast_s_port_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_port_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result = (int) ((arg1)->port);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_multicast_s(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_multicast_s", 0, 0, 0)) SWIG_fail;
  result = (struct multicast_s *)calloc(1, sizeof(struct multicast_s));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_multicast_s, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_multicast_s(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_multicast_s, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_multicast_s" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *multicast_s_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_multicast_s, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *multicast_s_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_rig_callbacks_freq_event_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  freq_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_freq_event_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_freq_event_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_freq_event_set" "', argument " "2"" of type '" "freq_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->freq_event = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_freq_event_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_cb_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_freq_event_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (freq_cb_t) ((arg1)->freq_event);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double_p_void__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_freq_arg_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_freq_arg_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_freq_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_freq_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->freq_arg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_freq_arg_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_freq_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->freq_arg);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_mode_event_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  mode_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_mode_event_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_mode_event_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_mode_event_set" "', argument " "2"" of type '" "mode_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->mode_event = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_mode_event_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  mode_cb_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_mode_event_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (mode_cb_t) ((arg1)->mode_event);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_mode_arg_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_mode_arg_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_mode_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_mode_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->mode_arg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_mode_arg_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_mode_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->mode_arg);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_vfo_event_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  vfo_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_vfo_event_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_vfo_event_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_vfo_event_set" "', argument " "2"" of type '" "vfo_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->vfo_event = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_vfo_event_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  vfo_cb_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_vfo_event_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (vfo_cb_t) ((arg1)->vfo_event);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_void__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_vfo_arg_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_vfo_arg_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_vfo_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_vfo_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->vfo_arg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_vfo_arg_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_vfo_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->vfo_arg);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_ptt_event_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  ptt_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_ptt_event_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_ptt_event_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_ptt_event_set" "', argument " "2"" of type '" "ptt_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->ptt_event = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_ptt_event_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  ptt_cb_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_ptt_event_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (ptt_cb_t) ((arg1)->ptt_event);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_ptt_arg_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_ptt_arg_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_ptt_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_ptt_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->ptt_arg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_ptt_arg_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_ptt_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->ptt_arg);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_dcd_event_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  dcd_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_dcd_event_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_dcd_event_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_dcd_event_set" "', argument " "2"" of type '" "dcd_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->dcd_event = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_dcd_event_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  dcd_cb_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_dcd_event_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (dcd_cb_t) ((arg1)->dcd_event);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_dcd_arg_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_dcd_arg_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_dcd_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_dcd_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->dcd_arg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_dcd_arg_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_dcd_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->dcd_arg);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_pltune_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  pltune_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_pltune_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_pltune_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_pltune_set" "', argument " "2"" of type '" "pltune_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->pltune = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_pltune_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  pltune_cb_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_pltune_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (pltune_cb_t) ((arg1)->pltune);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_pltune_arg_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_pltune_arg_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_pltune_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_pltune_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->pltune_arg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_pltune_arg_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_pltune_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->pltune_arg);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_spectrum_event_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  spectrum_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_spectrum_event_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_spectrum_event_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_spectrum_event_set" "', argument " "2"" of type '" "spectrum_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->spectrum_event = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_spectrum_event_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  spectrum_cb_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_spectrum_event_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (spectrum_cb_t) ((arg1)->spectrum_event);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_spectrum_arg_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_callbacks_spectrum_arg_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_spectrum_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_spectrum_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->spectrum_arg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_callbacks_spectrum_arg_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_spectrum_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->spectrum_arg);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_rig_callbacks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_rig_callbacks", 0, 0, 0)) SWIG_fail;
  result = (struct rig_callbacks *)calloc(1, sizeof(struct rig_callbacks));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_callbacks, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_rig_callbacks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_callbacks, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rig_callbacks" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *rig_callbacks_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_rig_callbacks, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *rig_callbacks_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_netrigctl_get_vfo_mode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  RIG *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_s_rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "netrigctl_get_vfo_mode" "', argument " "1"" of type '" "RIG *""'"); 
  }
  arg1 = (RIG *)(argp1);
  result = (int)netrigctl_get_vfo_mode(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rigerror(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rigerror" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = (int)(val1);
  result = (char *)rigerror(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rigerror2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rigerror2" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = (int)(val1);
  result = (char *)rigerror2(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_set_debug(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  enum rig_debug_level_e arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_set_debug" "', argument " "1"" of type '" "enum rig_debug_level_e""'");
  } 
  arg1 = (enum rig_debug_level_e)(val1);
  rig_set_debug(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_get_debug(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  enum rig_debug_level_e *arg1 = 0 ;
  enum rig_debug_level_e temp1 ;
  int res1 = SWIG_TMPOBJ ;
  
  arg1 = &temp1;
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_get_debug", 0, 0, 0)) SWIG_fail;
  rig_get_debug(arg1);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_unsigned_SS_int((*arg1)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_rig_debug_level_e, new_flags), 1);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_set_debug_time_stamp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_set_debug_time_stamp" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = (int)(val1);
  rig_set_debug_time_stamp(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_add2debugmsgsave(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "add2debugmsgsave" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  add2debugmsgsave((char const *)arg1);
  resultobj = SWIG_Py_Void();
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strrmode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  rmode_t arg1 ;
  unsigned long long val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strrmode" "', argument " "1"" of type '" "rmode_t""'");
  } 
  arg1 = (rmode_t)(val1);
  result = (char *)rig_strrmode(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strrmodes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  rmode_t arg1 ;
  char *arg2 = 0 ;
  int arg3 ;
  unsigned long long val1 ;
  int ecode1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_strrmodes", 3, 3, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strrmodes" "', argument " "1"" of type '" "rmode_t""'");
  } 
  arg1 = (rmode_t)(val1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_strrmodes" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "rig_strrmodes" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  result = (int)rig_strrmodes(arg1,arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strvfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  vfo_t arg1 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strvfo" "', argument " "1"" of type '" "vfo_t""'");
  } 
  arg1 = (vfo_t)(val1);
  result = (char *)rig_strvfo(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strfunc(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  setting_t arg1 ;
  unsigned long long val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strfunc" "', argument " "1"" of type '" "setting_t""'");
  } 
  arg1 = (setting_t)(val1);
  result = (char *)rig_strfunc(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strlevel(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  setting_t arg1 ;
  unsigned long long val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strlevel" "', argument " "1"" of type '" "setting_t""'");
  } 
  arg1 = (setting_t)(val1);
  result = (char *)rig_strlevel(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strparm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  setting_t arg1 ;
  unsigned long long val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strparm" "', argument " "1"" of type '" "setting_t""'");
  } 
  arg1 = (setting_t)(val1);
  result = (char *)rig_strparm(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_stragclevel(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  enum agc_level_e arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_stragclevel" "', argument " "1"" of type '" "enum agc_level_e""'");
  } 
  arg1 = (enum agc_level_e)(val1);
  result = (char *)rig_stragclevel(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strptrshift(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  rptr_shift_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strptrshift" "', argument " "1"" of type '" "rptr_shift_t""'");
  } 
  arg1 = (rptr_shift_t)(val1);
  result = (char *)rig_strptrshift(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strvfop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  vfo_op_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strvfop" "', argument " "1"" of type '" "vfo_op_t""'");
  } 
  arg1 = (vfo_op_t)(val1);
  result = (char *)rig_strvfop(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strscan(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  scan_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strscan" "', argument " "1"" of type '" "scan_t""'");
  } 
  arg1 = (scan_t)(val1);
  result = (char *)rig_strscan(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strstatus(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  enum rig_status_e arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strstatus" "', argument " "1"" of type '" "enum rig_status_e""'");
  } 
  arg1 = (enum rig_status_e)(val1);
  result = (char *)rig_strstatus(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strmtype(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  chan_type_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strmtype" "', argument " "1"" of type '" "chan_type_t""'");
  } 
  arg1 = (chan_type_t)(val1);
  result = (char *)rig_strmtype(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strspectrummode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  enum rig_spectrum_mode_e arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strspectrummode" "', argument " "1"" of type '" "enum rig_spectrum_mode_e""'");
  } 
  arg1 = (enum rig_spectrum_mode_e)(val1);
  result = (char *)rig_strspectrummode(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_strcommstatus(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  rig_comm_status_t arg1 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strcommstatus" "', argument " "1"" of type '" "rig_comm_status_t""'");
  } 
  arg1 = (rig_comm_status_t)(val1);
  result = (char *)rig_strcommstatus(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_hl_usleep(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  rig_useconds_t arg1 ;
  unsigned long val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_unsigned_SS_long(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "hl_usleep" "', argument " "1"" of type '" "rig_useconds_t""'");
  } 
  arg1 = (rig_useconds_t)(val1);
  result = (int)hl_usleep(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_longlat2locator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  double arg1 ;
  double arg2 ;
  char *arg3 = 0 ;
  int arg4 ;
  double val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  char temp3[13+1] ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[3] ;
  int result;
  
  arg3 = (char *) temp3;
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "longlat2locator", 3, 3, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_double(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "longlat2locator" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = (double)(val1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "longlat2locator" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  ecode4 = SWIG_AsVal_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "longlat2locator" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = (int)(val4);
  result = (int)longlat2locator(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_int((int)(result));
  arg3[13] = 0;  
  resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_FromCharPtr(arg3), 0);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_locator2longlat(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  double *arg1 = 0 ;
  double *arg2 = 0 ;
  char *arg3 = 0 ;
  double temp1 ;
  int res1 = SWIG_TMPOBJ ;
  double temp2 ;
  int res2 = SWIG_TMPOBJ ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  arg1 = &temp1;
  arg2 = &temp2;
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res3 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "locator2longlat" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  result = (int)locator2longlat(arg1,arg2,(char const *)arg3);
  resultobj = SWIG_From_int((int)(result));
  if (SWIG_IsTmpObj(res1)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_double((*arg1)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_double, new_flags), 0);
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_double((*arg2)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_double, new_flags), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_vfo_comp_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_vfo_comp_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_comp_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_vfo_comp_set" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  if (arg1) (arg1)->vfo_comp = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_vfo_comp_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_comp_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (double) ((arg1)->vfo_comp);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_range_list_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_rx_range_list_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rx_range_list_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_range_list_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_range_list_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_tx_range_list_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tx_range_list_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_range_list_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tuning_steps_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  struct tuning_step_list *arg2 = (struct tuning_step_list *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_tuning_steps_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tuning_steps_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_tuning_step_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tuning_steps_set" "', argument " "2"" of type '" "struct tuning_step_list [20]""'"); 
  } 
  arg2 = (struct tuning_step_list *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)20; ++ii) *(struct tuning_step_list *)&arg1->tuning_steps[ii] = *((struct tuning_step_list *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tuning_steps""' of type '""struct tuning_step_list [20]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tuning_steps_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct tuning_step_list *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tuning_steps_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (struct tuning_step_list *)(struct tuning_step_list *) ((arg1)->tuning_steps);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_tuning_step_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_filters_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  struct filter_list *arg2 = (struct filter_list *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_filters_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_filters_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_filter_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_filters_set" "', argument " "2"" of type '" "struct filter_list [60]""'"); 
  } 
  arg2 = (struct filter_list *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)60; ++ii) *(struct filter_list *)&arg1->filters[ii] = *((struct filter_list *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""filters""' of type '""struct filter_list [60]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_filters_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct filter_list *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_filters_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (struct filter_list *)(struct filter_list *) ((arg1)->filters);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_filter_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_str_cal_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  cal_table_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_str_cal_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_str_cal_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_cal_table, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_str_cal_set" "', argument " "2"" of type '" "cal_table_t *""'"); 
  }
  arg2 = (cal_table_t *)(argp2);
  if (arg1) (arg1)->str_cal = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_str_cal_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  cal_table_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_str_cal_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (cal_table_t *)& ((arg1)->str_cal);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_cal_table, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_chan_list_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  chan_t *arg2 = (chan_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_chan_list_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_chan_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_chan_list_set" "', argument " "2"" of type '" "chan_t [16]""'"); 
  } 
  arg2 = (chan_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)16; ++ii) *(chan_t *)&arg1->chan_list[ii] = *((chan_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""chan_list""' of type '""chan_t [16]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_chan_list_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  chan_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_chan_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (chan_t *)(chan_t *) ((arg1)->chan_list);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_chan_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_max_rit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_max_rit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_max_rit_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_max_rit_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->max_rit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_max_rit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  shortfreq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_max_rit_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (shortfreq_t) ((arg1)->max_rit);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_max_xit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_max_xit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_max_xit_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_max_xit_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->max_xit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_max_xit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  shortfreq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_max_xit_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (shortfreq_t) ((arg1)->max_xit);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_max_ifshift_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_max_ifshift_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_max_ifshift_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_max_ifshift_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->max_ifshift = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_max_ifshift_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  shortfreq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_max_ifshift_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (shortfreq_t) ((arg1)->max_ifshift);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_announces_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  ann_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_announces_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_announces_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_announces_set" "', argument " "2"" of type '" "ann_t""'");
  } 
  arg2 = (ann_t)(val2);
  if (arg1) (arg1)->announces = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_announces_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  ann_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_announces_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (ann_t) ((arg1)->announces);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_preamp_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int *arg2 = (int *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_preamp_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_preamp_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_preamp_set" "', argument " "2"" of type '" "int [8]""'"); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(int *)&arg1->preamp[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""preamp""' of type '""int [8]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_preamp_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_preamp_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int *)(int *) ((arg1)->preamp);
  {
    int len,i;
    len = 8;
    resultobj = PyList_New(len);
    for (i = 0; i < len; i++) {
      PyList_SetItem(resultobj,i,PyInt_FromLong((long)result[i]));
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_attenuator_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int *arg2 = (int *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_attenuator_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_attenuator_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_attenuator_set" "', argument " "2"" of type '" "int [8]""'"); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(int *)&arg1->attenuator[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""attenuator""' of type '""int [8]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_attenuator_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_attenuator_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int *)(int *) ((arg1)->attenuator);
  {
    int len,i;
    len = 8;
    resultobj = PyList_New(len);
    for (i = 0; i < len; i++) {
      PyList_SetItem(resultobj,i,PyInt_FromLong((long)result[i]));
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_has_get_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_has_get_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_get_func_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_has_get_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_has_get_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_get_func_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_func);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_has_set_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_has_set_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_set_func_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_has_set_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_has_set_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_set_func_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_func);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_has_get_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_has_get_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_get_level_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_has_get_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_has_get_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_get_level_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_level);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_has_set_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_has_set_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_set_level_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_has_set_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_has_set_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_set_level_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_level);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_has_get_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_has_get_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_get_parm_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_has_get_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_has_get_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_get_parm_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_parm);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_has_set_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_has_set_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_set_parm_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_has_set_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_has_set_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_set_parm_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_parm);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_level_gran_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_level_gran_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_level_gran_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_level_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->level_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""level_gran""' of type '""gran_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_level_gran_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  gran_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_level_gran_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->level_gran);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_parm_gran_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_parm_gran_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_parm_gran_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_parm_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->parm_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""parm_gran""' of type '""gran_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_parm_gran_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  gran_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_parm_gran_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->parm_gran);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_transaction_active_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_transaction_active_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_transaction_active_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_transaction_active_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->transaction_active = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_transaction_active_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_transaction_active_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->transaction_active);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_current_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_current_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_vfo_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_current_vfo_set" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  if (arg1) (arg1)->current_vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_current_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  vfo_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_vfo_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (vfo_t) ((arg1)->current_vfo);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_vfo_list_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_vfo_list_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_vfo_list_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->vfo_list = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_vfo_list_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->vfo_list);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_comm_state_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_comm_state_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_comm_state_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_comm_state_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->comm_state = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_comm_state_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_comm_state_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->comm_state);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_priv_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_priv_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_priv_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_priv_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->priv = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_priv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_priv_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->priv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_obj_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_obj_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_obj_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_obj_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->obj = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_obj_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_obj_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->obj);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_async_data_enabled_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_async_data_enabled_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_enabled_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_async_data_enabled_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->async_data_enabled = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_async_data_enabled_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_enabled_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->async_data_enabled);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_poll_interval_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_poll_interval_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_poll_interval_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_poll_interval_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->poll_interval = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_poll_interval_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_poll_interval_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->poll_interval);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_current_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_current_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_freq_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_current_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->current_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_current_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_freq_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_t) ((arg1)->current_freq);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_current_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_current_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_mode_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_current_mode_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->current_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_current_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rmode_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_mode_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (rmode_t) ((arg1)->current_mode);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_current_width_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  pbwidth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_current_width_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_width_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_current_width_set" "', argument " "2"" of type '" "pbwidth_t""'");
  } 
  arg2 = (pbwidth_t)(val2);
  if (arg1) (arg1)->current_width = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_current_width_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  pbwidth_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_width_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (pbwidth_t) ((arg1)->current_width);
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_tx_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_vfo_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_tx_vfo_set" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  if (arg1) (arg1)->tx_vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  vfo_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_vfo_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (vfo_t) ((arg1)->tx_vfo);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_mode_list_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_mode_list_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_mode_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_mode_list_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->mode_list = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_mode_list_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rmode_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_mode_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (rmode_t) ((arg1)->mode_list);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_transmit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_transmit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_transmit_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_transmit_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->transmit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_transmit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_transmit_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->transmit);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_lo_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_lo_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_lo_freq_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_lo_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->lo_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_lo_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_lo_freq_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_t) ((arg1)->lo_freq);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_twiddle_time_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_twiddle_time_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_time_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_time_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_twiddle_time_set" "', argument " "2"" of type '" "time_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_twiddle_time_set" "', argument " "2"" of type '" "time_t""'");
    } else {
      arg2 = *((time_t *)(argp2));
    }
  }
  if (arg1) (arg1)->twiddle_time = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_twiddle_time_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  time_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_time_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->twiddle_time);
  resultobj = SWIG_NewPointerObj((time_t *)memcpy((time_t *)calloc(1,sizeof(time_t)),&result,sizeof(time_t)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_twiddle_timeout_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_twiddle_timeout_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_timeout_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_twiddle_timeout_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->twiddle_timeout = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_twiddle_timeout_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_timeout_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->twiddle_timeout);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_uplink_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_uplink_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_uplink_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_uplink_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->uplink = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_uplink_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_uplink_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->uplink);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_vfo_opt_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_vfo_opt_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_opt_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_vfo_opt_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->vfo_opt = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_vfo_opt_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_opt_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->vfo_opt);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_auto_power_on_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_auto_power_on_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_auto_power_on_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_auto_power_on_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->auto_power_on = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_auto_power_on_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_auto_power_on_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->auto_power_on);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_auto_power_off_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_auto_power_off_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_auto_power_off_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_auto_power_off_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->auto_power_off = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_auto_power_off_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_auto_power_off_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->auto_power_off);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_auto_disable_screensaver_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_auto_disable_screensaver_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_auto_disable_screensaver_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_auto_disable_screensaver_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->auto_disable_screensaver = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_auto_disable_screensaver_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_auto_disable_screensaver_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->auto_disable_screensaver);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_ptt_share_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_ptt_share_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_ptt_share_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_ptt_share_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->ptt_share = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_ptt_share_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_ptt_share_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->ptt_share);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_power_now_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_power_now_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_power_now_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_power_now_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->power_now = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_power_now_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_power_now_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->power_now);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_power_min_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_power_min_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_power_min_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_power_min_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->power_min = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_power_min_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_power_min_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->power_min);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_power_max_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_power_max_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_power_max_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_power_max_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->power_max = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_power_max_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_power_max_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->power_max);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_disable_yaesu_bandselect_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  unsigned char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_disable_yaesu_bandselect_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_disable_yaesu_bandselect_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_disable_yaesu_bandselect_set" "', argument " "2"" of type '" "unsigned char""'");
  } 
  arg2 = (unsigned char)(val2);
  if (arg1) (arg1)->disable_yaesu_bandselect = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_disable_yaesu_bandselect_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned char result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_disable_yaesu_bandselect_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (unsigned char) ((arg1)->disable_yaesu_bandselect);
  resultobj = SWIG_From_unsigned_SS_char((unsigned char)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_twiddle_rit_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_twiddle_rit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_rit_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_twiddle_rit_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->twiddle_rit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_twiddle_rit_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_rit_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->twiddle_rit);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_twiddle_state_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_twiddle_state_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_state_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_twiddle_state_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->twiddle_state = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_twiddle_state_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_state_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->twiddle_state);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_rx_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_vfo_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_rx_vfo_set" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  if (arg1) (arg1)->rx_vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  vfo_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_vfo_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (vfo_t) ((arg1)->rx_vfo);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_snapshot_packet_sequence_number_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_snapshot_packet_sequence_number_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_snapshot_packet_sequence_number_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_snapshot_packet_sequence_number_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->snapshot_packet_sequence_number = (unsigned int volatile )arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_snapshot_packet_sequence_number_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_snapshot_packet_sequence_number_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (unsigned int)(unsigned int) ((arg1)->snapshot_packet_sequence_number);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_publisher_run_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_multicast_publisher_run_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_publisher_run_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_multicast_publisher_run_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->multicast_publisher_run = (int volatile )arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_publisher_run_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_publisher_run_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int)(int) ((arg1)->multicast_publisher_run);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_publisher_priv_data_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_multicast_publisher_priv_data_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_publisher_priv_data_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_multicast_publisher_priv_data_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->multicast_publisher_priv_data = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_publisher_priv_data_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_publisher_priv_data_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->multicast_publisher_priv_data);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_async_data_handler_thread_run_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_async_data_handler_thread_run_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_handler_thread_run_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_async_data_handler_thread_run_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->async_data_handler_thread_run = (int volatile )arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_async_data_handler_thread_run_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_handler_thread_run_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int)(int) ((arg1)->async_data_handler_thread_run);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_async_data_handler_priv_data_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_async_data_handler_priv_data_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_handler_priv_data_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_async_data_handler_priv_data_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->async_data_handler_priv_data = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_async_data_handler_priv_data_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_handler_priv_data_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->async_data_handler_priv_data);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_poll_routine_thread_run_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_poll_routine_thread_run_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_poll_routine_thread_run_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_poll_routine_thread_run_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->poll_routine_thread_run = (int volatile )arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_poll_routine_thread_run_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_poll_routine_thread_run_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int)(int) ((arg1)->poll_routine_thread_run);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_poll_routine_priv_data_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_poll_routine_priv_data_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_poll_routine_priv_data_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_poll_routine_priv_data_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->poll_routine_priv_data = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_poll_routine_priv_data_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_poll_routine_priv_data_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->poll_routine_priv_data);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_mutex_set_transaction_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  pthread_mutex_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_mutex_set_transaction_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_mutex_set_transaction_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_pthread_mutex_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_mutex_set_transaction_set" "', argument " "2"" of type '" "pthread_mutex_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_mutex_set_transaction_set" "', argument " "2"" of type '" "pthread_mutex_t""'");
    } else {
      arg2 = *((pthread_mutex_t *)(argp2));
    }
  }
  if (arg1) (arg1)->mutex_set_transaction = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_mutex_set_transaction_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  pthread_mutex_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_mutex_set_transaction_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->mutex_set_transaction);
  resultobj = SWIG_NewPointerObj((pthread_mutex_t *)memcpy((pthread_mutex_t *)calloc(1,sizeof(pthread_mutex_t)),&result,sizeof(pthread_mutex_t)), SWIGTYPE_p_pthread_mutex_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rigport_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  hamlib_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_rigport_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rigport_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_hamlib_port,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rigport_set" "', argument " "2"" of type '" "hamlib_port_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_rigport_set" "', argument " "2"" of type '" "hamlib_port_t""'");
    } else {
      arg2 = *((hamlib_port_t *)(argp2));
    }
  }
  if (arg1) (arg1)->rigport = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rigport_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  hamlib_port_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rigport_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->rigport);
  resultobj = SWIG_NewPointerObj((hamlib_port_t *)memcpy((hamlib_port_t *)calloc(1,sizeof(hamlib_port_t)),&result,sizeof(hamlib_port_t)), SWIGTYPE_p_hamlib_port, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_pttport_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  hamlib_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_pttport_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_pttport_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_hamlib_port,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_pttport_set" "', argument " "2"" of type '" "hamlib_port_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_pttport_set" "', argument " "2"" of type '" "hamlib_port_t""'");
    } else {
      arg2 = *((hamlib_port_t *)(argp2));
    }
  }
  if (arg1) (arg1)->pttport = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_pttport_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  hamlib_port_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_pttport_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->pttport);
  resultobj = SWIG_NewPointerObj((hamlib_port_t *)memcpy((hamlib_port_t *)calloc(1,sizeof(hamlib_port_t)),&result,sizeof(hamlib_port_t)), SWIGTYPE_p_hamlib_port, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_dcdport_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  hamlib_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_dcdport_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dcdport_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_hamlib_port,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_dcdport_set" "', argument " "2"" of type '" "hamlib_port_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_dcdport_set" "', argument " "2"" of type '" "hamlib_port_t""'");
    } else {
      arg2 = *((hamlib_port_t *)(argp2));
    }
  }
  if (arg1) (arg1)->dcdport = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_dcdport_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  hamlib_port_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dcdport_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->dcdport);
  resultobj = SWIG_NewPointerObj((hamlib_port_t *)memcpy((hamlib_port_t *)calloc(1,sizeof(hamlib_port_t)),&result,sizeof(hamlib_port_t)), SWIGTYPE_p_hamlib_port, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_use_cached_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_use_cached_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_use_cached_freq_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_use_cached_freq_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->use_cached_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_use_cached_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_use_cached_freq_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->use_cached_freq);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_use_cached_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_use_cached_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_use_cached_mode_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_use_cached_mode_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->use_cached_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_use_cached_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_use_cached_mode_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->use_cached_mode);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_use_cached_ptt_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_use_cached_ptt_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_use_cached_ptt_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_use_cached_ptt_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->use_cached_ptt = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_use_cached_ptt_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_use_cached_ptt_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->use_cached_ptt);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_depth_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_depth_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_depth_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_depth_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->depth = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_depth_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_depth_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->depth);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_lock_mode_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_lock_mode_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_lock_mode_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_lock_mode_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->lock_mode = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_lock_mode_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_lock_mode_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->lock_mode);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_powerstat_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  powerstat_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_powerstat_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_powerstat_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_powerstat_set" "', argument " "2"" of type '" "powerstat_t""'");
  } 
  arg2 = (powerstat_t)(val2);
  if (arg1) (arg1)->powerstat = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_powerstat_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  powerstat_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_powerstat_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (powerstat_t) ((arg1)->powerstat);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tuner_control_pathname_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_tuner_control_pathname_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tuner_control_pathname_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tuner_control_pathname_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->tuner_control_pathname);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->tuner_control_pathname = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->tuner_control_pathname = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tuner_control_pathname_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tuner_control_pathname_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->tuner_control_pathname);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_client_version_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  char *arg2 = (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char temp2[32] ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_client_version_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_client_version_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharArray(swig_obj[1], temp2, 32);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_client_version_set" "', argument " "2"" of type '" "char [32]""'");
  }
  arg2 = (char *)(temp2);
  if (arg2) memcpy(arg1->client_version,arg2,32*sizeof(char));
  else memset(arg1->client_version,0,32*sizeof(char));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_client_version_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_client_version_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *)(char *) ((arg1)->client_version);
  {
    size_t size = SWIG_strnlen(result, 32);
    
    
    
    resultobj = SWIG_FromCharPtrAndSize(result, size);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_offset_vfoa_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_offset_vfoa_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_offset_vfoa_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_offset_vfoa_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->offset_vfoa = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_offset_vfoa_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_offset_vfoa_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_t) ((arg1)->offset_vfoa);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_offset_vfob_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_offset_vfob_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_offset_vfob_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_offset_vfob_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->offset_vfob = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_offset_vfob_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_offset_vfob_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_t) ((arg1)->offset_vfob);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  struct multicast_s *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_multicast_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_multicast_s, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_multicast_set" "', argument " "2"" of type '" "struct multicast_s *""'"); 
  }
  arg2 = (struct multicast_s *)(argp2);
  if (arg1) (arg1)->multicast = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct multicast_s *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (struct multicast_s *) ((arg1)->multicast);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_multicast_s, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rig_model_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  rig_model_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_rig_model_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rig_model_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_rig_model_set" "', argument " "2"" of type '" "rig_model_t""'");
  } 
  arg2 = (rig_model_t)(val2);
  if (arg1) (arg1)->rig_model = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rig_model_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rig_model_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rig_model_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (rig_model_t) ((arg1)->rig_model);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_model_name_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_model_name_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_model_name_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_model_name_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->model_name = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->model_name = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_model_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_model_name_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->model_name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_mfg_name_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_mfg_name_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_mfg_name_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_mfg_name_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->mfg_name = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->mfg_name = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_mfg_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_mfg_name_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->mfg_name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_version_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_version_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_version_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_version_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->version = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->version = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_version_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_version_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->version);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_copyright_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_copyright_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_copyright_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_copyright_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->copyright = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->copyright = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_copyright_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_copyright_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->copyright);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_status_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  enum rig_status_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_status_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_status_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_status_set" "', argument " "2"" of type '" "enum rig_status_e""'");
  } 
  arg2 = (enum rig_status_e)(val2);
  if (arg1) (arg1)->status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_status_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum rig_status_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_status_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (enum rig_status_e) ((arg1)->status);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rig_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_rig_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rig_type_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_rig_type_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->rig_type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rig_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rig_type_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->rig_type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_ptt_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  ptt_type_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_ptt_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_ptt_type_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_ptt_type_set" "', argument " "2"" of type '" "ptt_type_t""'");
  } 
  arg2 = (ptt_type_t)(val2);
  if (arg1) (arg1)->ptt_type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_ptt_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  ptt_type_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_ptt_type_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (ptt_type_t) ((arg1)->ptt_type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_dcd_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  dcd_type_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_dcd_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dcd_type_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_dcd_type_set" "', argument " "2"" of type '" "dcd_type_t""'");
  } 
  arg2 = (dcd_type_t)(val2);
  if (arg1) (arg1)->dcd_type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_dcd_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  dcd_type_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dcd_type_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (dcd_type_t) ((arg1)->dcd_type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_port_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  rig_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_port_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_port_type_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_port_type_set" "', argument " "2"" of type '" "rig_port_t""'");
  } 
  arg2 = (rig_port_t)(val2);
  if (arg1) (arg1)->port_type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_port_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rig_port_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_port_type_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (rig_port_t) ((arg1)->port_type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_serial_rate_min_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_serial_rate_min_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_rate_min_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_serial_rate_min_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_min = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_serial_rate_min_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_rate_min_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->serial_rate_min);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_serial_rate_max_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_serial_rate_max_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_rate_max_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_serial_rate_max_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_max = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_serial_rate_max_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_rate_max_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->serial_rate_max);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_serial_data_bits_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_serial_data_bits_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_data_bits_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_serial_data_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_data_bits = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_serial_data_bits_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_data_bits_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->serial_data_bits);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_serial_stop_bits_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_serial_stop_bits_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_stop_bits_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_serial_stop_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_stop_bits = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_serial_stop_bits_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_stop_bits_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->serial_stop_bits);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_serial_parity_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  enum serial_parity_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_serial_parity_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_parity_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_serial_parity_set" "', argument " "2"" of type '" "enum serial_parity_e""'");
  } 
  arg2 = (enum serial_parity_e)(val2);
  if (arg1) (arg1)->serial_parity = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_serial_parity_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum serial_parity_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_parity_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (enum serial_parity_e) ((arg1)->serial_parity);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_serial_handshake_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  enum serial_handshake_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_serial_handshake_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_handshake_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_serial_handshake_set" "', argument " "2"" of type '" "enum serial_handshake_e""'");
  } 
  arg2 = (enum serial_handshake_e)(val2);
  if (arg1) (arg1)->serial_handshake = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_serial_handshake_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum serial_handshake_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_handshake_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (enum serial_handshake_e) ((arg1)->serial_handshake);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_write_delay_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_write_delay_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_write_delay_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->write_delay = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_write_delay_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_write_delay_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->write_delay);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_post_write_delay_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_post_write_delay_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_post_write_delay_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_post_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->post_write_delay = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_post_write_delay_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_post_write_delay_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->post_write_delay);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_timeout_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_timeout_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_timeout_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_timeout_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->timeout = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_timeout_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_timeout_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->timeout);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_retry_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_retry_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_retry_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_retry_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->retry = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_retry_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_retry_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->retry);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_targetable_vfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_targetable_vfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_targetable_vfo_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_targetable_vfo_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->targetable_vfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_targetable_vfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_targetable_vfo_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->targetable_vfo);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_async_data_supported_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_async_data_supported_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_supported_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_async_data_supported_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->async_data_supported = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_async_data_supported_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_supported_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->async_data_supported);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_agc_level_count_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_agc_level_count_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_agc_level_count_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_agc_level_count_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->agc_level_count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_agc_level_count_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_agc_level_count_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->agc_level_count);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_agc_levels_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  enum agc_level_e *arg2 = (enum agc_level_e *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_agc_levels_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_agc_levels_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_agc_level_e, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_agc_levels_set" "', argument " "2"" of type '" "enum agc_level_e [8]""'"); 
  } 
  arg2 = (enum agc_level_e *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(enum agc_level_e *)&arg1->agc_levels[ii] = *((enum agc_level_e *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""agc_levels""' of type '""enum agc_level_e [8]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_agc_levels_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum agc_level_e *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_agc_levels_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (enum agc_level_e *)(enum agc_level_e *) ((arg1)->agc_levels);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_agc_level_e, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_ctcss_list_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  tone_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_ctcss_list_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_ctcss_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_unsigned_int, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_ctcss_list_set" "', argument " "2"" of type '" "tone_t *""'"); 
  }
  arg2 = (tone_t *)(argp2);
  if (arg1) (arg1)->ctcss_list = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_ctcss_list_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  tone_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_ctcss_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (tone_t *) ((arg1)->ctcss_list);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_dcs_list_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  tone_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_dcs_list_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dcs_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_unsigned_int, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_dcs_list_set" "', argument " "2"" of type '" "tone_t *""'"); 
  }
  arg2 = (tone_t *)(argp2);
  if (arg1) (arg1)->dcs_list = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_dcs_list_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  tone_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dcs_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (tone_t *) ((arg1)->dcs_list);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_vfo_ops_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  vfo_op_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_vfo_ops_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_ops_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_vfo_ops_set" "', argument " "2"" of type '" "vfo_op_t""'");
  } 
  arg2 = (vfo_op_t)(val2);
  if (arg1) (arg1)->vfo_ops = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_vfo_ops_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  vfo_op_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_ops_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (vfo_op_t) ((arg1)->vfo_ops);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_scan_ops_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  scan_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_scan_ops_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_scan_ops_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_scan_ops_set" "', argument " "2"" of type '" "scan_t""'");
  } 
  arg2 = (scan_t)(val2);
  if (arg1) (arg1)->scan_ops = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_scan_ops_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  scan_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_scan_ops_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (scan_t) ((arg1)->scan_ops);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_transceive_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_transceive_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_transceive_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_transceive_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->transceive = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_transceive_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_transceive_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->transceive);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_bank_qty_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_bank_qty_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_bank_qty_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_bank_qty_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->bank_qty = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_bank_qty_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_bank_qty_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->bank_qty);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_chan_desc_sz_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_chan_desc_sz_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_chan_desc_sz_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_chan_desc_sz_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->chan_desc_sz = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_chan_desc_sz_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_chan_desc_sz_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->chan_desc_sz);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_range_list1_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_rx_range_list1_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list1_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rx_range_list1_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list1[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list1""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_range_list1_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list1_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_range_list1_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_tx_range_list1_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list1_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tx_range_list1_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list1[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list1""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_range_list1_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list1_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_range_list2_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_rx_range_list2_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list2_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rx_range_list2_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list2[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list2""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_range_list2_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list2_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_range_list2_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_tx_range_list2_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list2_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tx_range_list2_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list2[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list2""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_range_list2_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list2_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_range_list3_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_rx_range_list3_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list3_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rx_range_list3_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list3[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list3""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_range_list3_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list3_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_range_list3_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_tx_range_list3_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list3_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tx_range_list3_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list3[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list3""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_range_list3_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list3_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_range_list4_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_rx_range_list4_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list4_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rx_range_list4_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list4[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list4""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_range_list4_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list4_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_range_list4_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_tx_range_list4_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list4_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tx_range_list4_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list4[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list4""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_range_list4_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list4_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_range_list5_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_rx_range_list5_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list5_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rx_range_list5_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list5[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list5""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_rx_range_list5_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list5_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_range_list5_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_tx_range_list5_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list5_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tx_range_list5_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list5[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list5""' of type '""freq_range_t [30]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_tx_range_list5_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_range_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list5_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_spectrum_scopes_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  struct rig_spectrum_scope *arg2 = (struct rig_spectrum_scope *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_spectrum_scopes_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_scopes_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_spectrum_scopes_set" "', argument " "2"" of type '" "struct rig_spectrum_scope [4]""'"); 
  } 
  arg2 = (struct rig_spectrum_scope *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)4; ++ii) *(struct rig_spectrum_scope *)&arg1->spectrum_scopes[ii] = *((struct rig_spectrum_scope *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_scopes""' of type '""struct rig_spectrum_scope [4]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_spectrum_scopes_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct rig_spectrum_scope *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_scopes_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (struct rig_spectrum_scope *)(struct rig_spectrum_scope *) ((arg1)->spectrum_scopes);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_spectrum_modes_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  enum rig_spectrum_mode_e *arg2 = (enum rig_spectrum_mode_e *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_spectrum_modes_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_modes_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_rig_spectrum_mode_e, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_spectrum_modes_set" "', argument " "2"" of type '" "enum rig_spectrum_mode_e [5]""'"); 
  } 
  arg2 = (enum rig_spectrum_mode_e *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)5; ++ii) *(enum rig_spectrum_mode_e *)&arg1->spectrum_modes[ii] = *((enum rig_spectrum_mode_e *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_modes""' of type '""enum rig_spectrum_mode_e [5]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_spectrum_modes_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum rig_spectrum_mode_e *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_modes_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (enum rig_spectrum_mode_e *)(enum rig_spectrum_mode_e *) ((arg1)->spectrum_modes);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_mode_e, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_spectrum_spans_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  freq_t *arg2 = (freq_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_spectrum_spans_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_spans_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_spectrum_spans_set" "', argument " "2"" of type '" "freq_t [20]""'"); 
  } 
  arg2 = (freq_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)20; ++ii) *(freq_t *)&arg1->spectrum_spans[ii] = *((freq_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_spans""' of type '""freq_t [20]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_spectrum_spans_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_spans_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_t *)(freq_t *) ((arg1)->spectrum_spans);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_spectrum_avg_modes_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  struct rig_spectrum_avg_mode *arg2 = (struct rig_spectrum_avg_mode *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_spectrum_avg_modes_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_avg_modes_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_spectrum_avg_modes_set" "', argument " "2"" of type '" "struct rig_spectrum_avg_mode [12]""'"); 
  } 
  arg2 = (struct rig_spectrum_avg_mode *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)12; ++ii) *(struct rig_spectrum_avg_mode *)&arg1->spectrum_avg_modes[ii] = *((struct rig_spectrum_avg_mode *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_avg_modes""' of type '""struct rig_spectrum_avg_mode [12]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_spectrum_avg_modes_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct rig_spectrum_avg_mode *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_avg_modes_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (struct rig_spectrum_avg_mode *)(struct rig_spectrum_avg_mode *) ((arg1)->spectrum_avg_modes);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_spectrum_attenuator_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int *arg2 = (int *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_spectrum_attenuator_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_attenuator_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_spectrum_attenuator_set" "', argument " "2"" of type '" "int [8]""'"); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(int *)&arg1->spectrum_attenuator[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_attenuator""' of type '""int [8]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_spectrum_attenuator_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_attenuator_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int *)(int *) ((arg1)->spectrum_attenuator);
  {
    int len,i;
    len = 8;
    resultobj = PyList_New(len);
    for (i = 0; i < len; i++) {
      PyList_SetItem(resultobj,i,PyInt_FromLong((long)result[i]));
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_morse_data_handler_thread_run_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_morse_data_handler_thread_run_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_morse_data_handler_thread_run_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_morse_data_handler_thread_run_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->morse_data_handler_thread_run = (int volatile )arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_morse_data_handler_thread_run_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_morse_data_handler_thread_run_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int)(int) ((arg1)->morse_data_handler_thread_run);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_morse_data_handler_priv_data_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_morse_data_handler_priv_data_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_morse_data_handler_priv_data_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_morse_data_handler_priv_data_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->morse_data_handler_priv_data = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_morse_data_handler_priv_data_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_morse_data_handler_priv_data_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->morse_data_handler_priv_data);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_fifo_morse_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  struct FIFO_RIG_s *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_fifo_morse_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_fifo_morse_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_FIFO_RIG_s, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_fifo_morse_set" "', argument " "2"" of type '" "struct FIFO_RIG_s *""'"); 
  }
  arg2 = (struct FIFO_RIG_s *)(argp2);
  if (arg1) (arg1)->fifo_morse = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_fifo_morse_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct FIFO_RIG_s *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_fifo_morse_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (struct FIFO_RIG_s *) ((arg1)->fifo_morse);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_FIFO_RIG_s, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_doppler_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_doppler_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_doppler_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_doppler_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->doppler = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_doppler_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_doppler_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->doppler);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_data_addr_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_multicast_data_addr_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_data_addr_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_multicast_data_addr_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->multicast_data_addr);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->multicast_data_addr = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->multicast_data_addr = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_data_addr_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_data_addr_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->multicast_data_addr);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_data_port_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_multicast_data_port_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_data_port_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_multicast_data_port_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->multicast_data_port = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_data_port_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_data_port_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->multicast_data_port);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_cmd_addr_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_multicast_cmd_addr_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_cmd_addr_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_multicast_cmd_addr_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->multicast_cmd_addr);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->multicast_cmd_addr = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->multicast_cmd_addr = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_cmd_addr_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_cmd_addr_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->multicast_cmd_addr);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_cmd_port_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_multicast_cmd_port_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_cmd_port_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_multicast_cmd_port_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->multicast_cmd_port = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_cmd_port_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_cmd_port_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->multicast_cmd_port);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_receiver_run_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_multicast_receiver_run_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_receiver_run_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_multicast_receiver_run_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->multicast_receiver_run = (int volatile )arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_receiver_run_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_receiver_run_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int)(int) ((arg1)->multicast_receiver_run);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_receiver_priv_data_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_multicast_receiver_priv_data_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_receiver_priv_data_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_multicast_receiver_priv_data_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->multicast_receiver_priv_data = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_multicast_receiver_priv_data_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_receiver_priv_data_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->multicast_receiver_priv_data);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_comm_status_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  rig_comm_status_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_comm_status_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_comm_status_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_comm_status_set" "', argument " "2"" of type '" "rig_comm_status_t""'");
  } 
  arg2 = (rig_comm_status_t)(val2);
  if (arg1) (arg1)->comm_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_comm_status_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rig_comm_status_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_comm_status_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (rig_comm_status_t) ((arg1)->comm_status);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_device_id_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  char *arg2 = (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char temp2[30] ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_device_id_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_device_id_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharArray(swig_obj[1], temp2, 30);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_device_id_set" "', argument " "2"" of type '" "char [30]""'");
  }
  arg2 = (char *)(temp2);
  if (arg2) memcpy(arg1->device_id,arg2,30*sizeof(char));
  else memset(arg1->device_id,0,30*sizeof(char));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_device_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_device_id_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *)(char *) ((arg1)->device_id);
  {
    size_t size = SWIG_strnlen(result, 30);
    
    
    
    resultobj = SWIG_FromCharPtrAndSize(result, size);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_dual_watch_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_dual_watch_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dual_watch_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_dual_watch_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->dual_watch = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_dual_watch_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dual_watch_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->dual_watch);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_post_ptt_delay_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_post_ptt_delay_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_post_ptt_delay_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_post_ptt_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->post_ptt_delay = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_post_ptt_delay_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_post_ptt_delay_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->post_ptt_delay);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_freq_event_elapsed_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  struct timespec arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_freq_event_elapsed_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_freq_event_elapsed_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_timespec,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_freq_event_elapsed_set" "', argument " "2"" of type '" "struct timespec""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_freq_event_elapsed_set" "', argument " "2"" of type '" "struct timespec""'");
    } else {
      arg2 = *((struct timespec *)(argp2));
    }
  }
  if (arg1) (arg1)->freq_event_elapsed = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_freq_event_elapsed_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct timespec result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_freq_event_elapsed_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->freq_event_elapsed);
  resultobj = SWIG_NewPointerObj((struct timespec *)memcpy((struct timespec *)calloc(1,sizeof(struct timespec)),&result,sizeof(struct timespec)), SWIGTYPE_p_timespec, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_freq_skip_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_freq_skip_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_freq_skip_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_freq_skip_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->freq_skip = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_freq_skip_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_freq_skip_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->freq_skip);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_client_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  client_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_client_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_client_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_client_set" "', argument " "2"" of type '" "client_t""'");
  } 
  arg2 = (client_t)(val2);
  if (arg1) (arg1)->client = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_client_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  client_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_client_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (client_t) ((arg1)->client);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_api_mutex_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  pthread_mutex_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rig_state_api_mutex_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_api_mutex_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_pthread_mutex_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_api_mutex_set" "', argument " "2"" of type '" "pthread_mutex_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_api_mutex_set" "', argument " "2"" of type '" "pthread_mutex_t""'");
    } else {
      arg2 = *((pthread_mutex_t *)(argp2));
    }
  }
  if (arg1) (arg1)->api_mutex = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rig_state_api_mutex_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  pthread_mutex_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_api_mutex_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->api_mutex);
  resultobj = SWIG_NewPointerObj((pthread_mutex_t *)memcpy((pthread_mutex_t *)calloc(1,sizeof(pthread_mutex_t)),&result,sizeof(pthread_mutex_t)), SWIGTYPE_p_pthread_mutex_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_rig_state(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_rig_state", 0, 0, 0)) SWIG_fail;
  result = (struct rig_state *)calloc(1, sizeof(struct rig_state));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_state, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_rig_state(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rig_state, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rig_state" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *rig_state_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_rig_state, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *rig_state_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Rig_rig_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  RIG *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_rig_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_rig_set" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_s_rig, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_rig_set" "', argument " "2"" of type '" "RIG *""'"); 
  }
  arg2 = (RIG *)(argp2);
  if (arg1) (arg1)->rig = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_rig_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  RIG *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_rig_get" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  result = (RIG *) ((arg1)->rig);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_s_rig, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_caps_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  struct rig_caps *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_caps_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_caps_set" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_rig_caps, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_caps_set" "', argument " "2"" of type '" "struct rig_caps *""'"); 
  }
  arg2 = (struct rig_caps *)(argp2);
  if (arg1) (arg1)->caps = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_caps_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct rig_caps *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_caps_get" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  result = (struct rig_caps *) ((arg1)->caps);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_caps, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_state_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  struct rig_state *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_state_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_state_set" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_rig_state, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_state_set" "', argument " "2"" of type '" "struct rig_state *""'"); 
  }
  arg2 = (struct rig_state *)(argp2);
  if (arg1) (arg1)->state = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_state_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct rig_state *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_state_get" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  result = (struct rig_state *) ((arg1)->state);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rig_state, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_error_status_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_error_status_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_error_status_set" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_error_status_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->error_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_error_status_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_error_status_get" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  result = (int) ((arg1)->error_status);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_do_exception_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_do_exception_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_do_exception_set" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_do_exception_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->do_exception = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_do_exception_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_do_exception_get" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  result = (int) ((arg1)->do_exception);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Rig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  struct Rig *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Rig" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = (int)(val1);
  result = (struct Rig *)new_Rig(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Rig, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Rig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Rig" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  delete_Rig(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_open(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_open" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rig_open(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_close(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_close" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rig_close(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_freq(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  freq_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_freq", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_freq" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_freq" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_freq" "', argument " "3"" of type '" "freq_t""'");
  } 
  arg3 = (freq_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_freq(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_mode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  rmode_t arg2 ;
  pbwidth_t arg3 = ((shortfreq_t) (0)) ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_mode", 2, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_mode" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_mode" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_mode" "', argument " "3"" of type '" "pbwidth_t""'");
    } 
    arg3 = (pbwidth_t)(val3);
  }
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_mode" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_mode(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_ptt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  ptt_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_ptt", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ptt" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ptt" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_ptt" "', argument " "3"" of type '" "ptt_t""'");
  } 
  arg3 = (ptt_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_ptt(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_rptr_shift(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  rptr_shift_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_rptr_shift", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_rptr_shift" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_rptr_shift" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_rptr_shift" "', argument " "3"" of type '" "rptr_shift_t""'");
  } 
  arg3 = (rptr_shift_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_rptr_shift(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_rptr_offs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  shortfreq_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_rptr_offs", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_rptr_offs" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_rptr_offs" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_rptr_offs" "', argument " "3"" of type '" "shortfreq_t""'");
  } 
  arg3 = (shortfreq_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_rptr_offs(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_ctcss_tone(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  tone_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_ctcss_tone", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ctcss_tone" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ctcss_tone" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_ctcss_tone" "', argument " "3"" of type '" "tone_t""'");
  } 
  arg3 = (tone_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_ctcss_tone(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_dcs_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  tone_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_dcs_code", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_dcs_code" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_dcs_code" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_dcs_code" "', argument " "3"" of type '" "tone_t""'");
  } 
  arg3 = (tone_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_dcs_code(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_ctcss_sql(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  tone_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_ctcss_sql", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ctcss_sql" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ctcss_sql" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_ctcss_sql" "', argument " "3"" of type '" "tone_t""'");
  } 
  arg3 = (tone_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_ctcss_sql(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_dcs_sql(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  tone_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_dcs_sql", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_dcs_sql" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_dcs_sql" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_dcs_sql" "', argument " "3"" of type '" "tone_t""'");
  } 
  arg3 = (tone_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_dcs_sql(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_split_freq(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  freq_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_split_freq", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_split_freq" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_split_freq" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_split_freq" "', argument " "3"" of type '" "freq_t""'");
  } 
  arg3 = (freq_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_split_freq(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_split_mode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  rmode_t arg2 ;
  pbwidth_t arg3 = ((shortfreq_t) (0)) ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_split_mode", 2, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_split_mode" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_split_mode" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_split_mode" "', argument " "3"" of type '" "pbwidth_t""'");
    } 
    arg3 = (pbwidth_t)(val3);
  }
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_split_mode" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_split_mode(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_split_freq_mode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  freq_t arg2 ;
  rmode_t arg3 ;
  pbwidth_t arg4 = ((shortfreq_t) (0)) ;
  vfo_t arg5 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  unsigned long long val3 ;
  int ecode3 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_split_freq_mode", 3, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_split_freq_mode" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_split_freq_mode" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_split_freq_mode" "', argument " "3"" of type '" "rmode_t""'");
  } 
  arg3 = (rmode_t)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_long(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_split_freq_mode" "', argument " "4"" of type '" "pbwidth_t""'");
    } 
    arg4 = (pbwidth_t)(val4);
  }
  if (swig_obj[4]) {
    ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[4], &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Rig_set_split_freq_mode" "', argument " "5"" of type '" "vfo_t""'");
    } 
    arg5 = (vfo_t)(val5);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_split_freq_mode(arg1,arg2,arg3,arg4,arg5);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_split_vfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  split_t arg2 ;
  vfo_t arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_split_vfo", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_split_vfo" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_split_vfo" "', argument " "2"" of type '" "split_t""'");
  } 
  arg2 = (split_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_split_vfo" "', argument " "3"" of type '" "vfo_t""'");
  } 
  arg3 = (vfo_t)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_split_vfo" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_split_vfo(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_rit(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  shortfreq_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_rit", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_rit" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_rit" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_rit" "', argument " "3"" of type '" "shortfreq_t""'");
  } 
  arg3 = (shortfreq_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_rit(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_xit(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  shortfreq_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_xit", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_xit" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_xit" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_xit" "', argument " "3"" of type '" "shortfreq_t""'");
  } 
  arg3 = (shortfreq_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_xit(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_ts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  shortfreq_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_ts", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ts" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ts" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_ts" "', argument " "3"" of type '" "shortfreq_t""'");
  } 
  arg3 = (shortfreq_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_ts(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_ant(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  ant_t arg2 ;
  value_t arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_ant", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ant" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ant" "', argument " "2"" of type '" "ant_t""'");
  } 
  arg2 = (ant_t)(val2);
  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_ant" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rig_set_ant" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_ant" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_ant(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  int arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_func", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_func" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_func" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_func" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_func(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_ext_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  int arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_ext_func", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ext_func" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ext_func" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_ext_func" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_ext_func" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_ext_func(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_bank(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_bank", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_bank" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_bank" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_bank" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_bank(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_mem(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_mem", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_mem" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_mem" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_mem" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_mem(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_send_dtmf(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  const_char_string arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_send_dtmf", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_send_dtmf" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_send_dtmf" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_send_dtmf" "', argument " "3"" of type '" "const_char_string""'");
  }
  arg3 = (const_char_string)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rig_send_dtmf(arg1,arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_send_morse(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  const_char_string arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_send_morse", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_send_morse" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_send_morse" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_send_morse" "', argument " "3"" of type '" "const_char_string""'");
  }
  arg3 = (const_char_string)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rig_send_morse(arg1,arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_vfo_op(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  vfo_op_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_vfo_op", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_vfo_op" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_vfo_op" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_vfo_op" "', argument " "3"" of type '" "vfo_op_t""'");
  } 
  arg3 = (vfo_op_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_vfo_op(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_scan(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  scan_t arg2 ;
  int arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_scan", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_scan" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_scan" "', argument " "2"" of type '" "scan_t""'");
  } 
  arg2 = (scan_t)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_scan" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_scan" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_scan(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_level__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  int arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_level" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_level__SWIG_0(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_level__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  float arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  ecode3 = SWIG_AsVal_float(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_level" "', argument " "3"" of type '" "float""'");
  } 
  arg3 = (float)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_level__SWIG_1(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_level__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_level" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_level" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_level__SWIG_2(arg1,(char const *)arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_level__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  float arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_level" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_float(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_level" "', argument " "3"" of type '" "float""'");
  } 
  arg3 = (float)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_level__SWIG_3(arg1,(char const *)arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_level__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_level" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_level" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_level__SWIG_4(arg1,(char const *)arg2,(char const *)arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_level(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rig_set_level", 0, 4, argv))) SWIG_fail;
  --argc;
  if ((argc >= 3) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          if (argc <= 3) {
            return _wrap_Rig_set_level__SWIG_0(self, argc, argv);
          }
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Rig_set_level__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  if ((argc >= 3) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_float(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          if (argc <= 3) {
            return _wrap_Rig_set_level__SWIG_1(self, argc, argv);
          }
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Rig_set_level__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if ((argc >= 3) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          if (argc <= 3) {
            return _wrap_Rig_set_level__SWIG_2(self, argc, argv);
          }
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Rig_set_level__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  if ((argc >= 3) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_float(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          if (argc <= 3) {
            return _wrap_Rig_set_level__SWIG_3(self, argc, argv);
          }
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Rig_set_level__SWIG_3(self, argc, argv);
          }
        }
      }
    }
  }
  if ((argc >= 3) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          if (argc <= 3) {
            return _wrap_Rig_set_level__SWIG_4(self, argc, argv);
          }
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Rig_set_level__SWIG_4(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rig_set_level'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::set_level(setting_t,int,vfo_t)\n"
    "    Rig::set_level(setting_t,float,vfo_t)\n"
    "    Rig::set_level(char const *,int,vfo_t)\n"
    "    Rig::set_level(char const *,float,vfo_t)\n"
    "    Rig::set_level(char const *,char const *,vfo_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rig_set_ext_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  value_t arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_ext_level", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ext_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ext_level" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_ext_level" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rig_set_ext_level" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_ext_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_ext_level(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_level_i__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_level_i" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_level_i" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_level_i" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_level_i__SWIG_0(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_level_f__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  float result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_level_f" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_level_f" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_level_f" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    result = (float)Rig_get_level_f__SWIG_0(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_level_i__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_level_i" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_level_i" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_level_i" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_level_i__SWIG_1(arg1,(char const *)arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_level_i(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rig_get_level_i", 0, 3, argv))) SWIG_fail;
  --argc;
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_level_i__SWIG_0(self, argc, argv);
        }
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_level_i__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_level_i__SWIG_1(self, argc, argv);
        }
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_level_i__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rig_get_level_i'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_level_i(setting_t,vfo_t)\n"
    "    Rig::get_level_i(char const *,vfo_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rig_get_level_f__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  float result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_level_f" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_level_f" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_level_f" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    result = (float)Rig_get_level_f__SWIG_1(arg1,(char const *)arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_float((float)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_level_f(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rig_get_level_f", 0, 3, argv))) SWIG_fail;
  --argc;
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_level_f__SWIG_0(self, argc, argv);
        }
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_level_f__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_level_f__SWIG_1(self, argc, argv);
        }
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_level_f__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rig_get_level_f'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_level_f(setting_t,vfo_t)\n"
    "    Rig::get_level_f(char const *,vfo_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rig_get_level__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char temp3[256+1] ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  arg3 = (char *) temp3;
  (void)self;
  if ((nobjs < 2) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_level" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_get_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_level__SWIG_0(arg1,(char const *)arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  arg3[256] = 0;  
  resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_FromCharPtr(arg3), 1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_vfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_vfo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_vfo" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_vfo" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_set_vfo(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_powerstat(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  powerstat_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_powerstat", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_powerstat" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_powerstat" "', argument " "2"" of type '" "powerstat_t""'");
  } 
  arg2 = (powerstat_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_set_powerstat(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_has_get_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_has_get_level", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_get_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_get_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_get_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_has_get_parm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_has_get_parm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_get_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_get_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_get_parm(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_has_set_parm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_has_set_parm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_set_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_set_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_set_parm(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_has_get_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_has_get_func", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_get_func" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_get_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_get_func(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_has_set_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_has_set_func", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_set_func" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_set_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_set_func(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  reset_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_reset", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_reset" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_reset" "', argument " "2"" of type '" "reset_t""'");
  } 
  arg2 = (reset_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_reset(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_has_scan(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  scan_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_has_scan", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_scan" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_scan" "', argument " "2"" of type '" "scan_t""'");
  } 
  arg2 = (scan_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_scan(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_has_vfo_op(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_op_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_has_vfo_op", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_vfo_op" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_vfo_op" "', argument " "2"" of type '" "vfo_op_t""'");
  } 
  arg2 = (vfo_op_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_vfo_op(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_passband_normal(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_passband_normal", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_passband_normal" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_passband_normal" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_passband_normal(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_passband_narrow(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_passband_narrow", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_passband_narrow" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_passband_narrow" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_passband_narrow(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_passband_wide(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_passband_wide", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_passband_wide" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_passband_wide" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_passband_wide(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_vfo_opt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_vfo_opt", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_vfo_opt" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_vfo_opt" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_set_vfo_opt(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_ext_token_lookup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  const_char_string arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_ext_token_lookup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_ext_token_lookup" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_ext_token_lookup" "', argument " "2"" of type '" "const_char_string""'");
  }
  arg2 = (const_char_string)(buf2);
  {
    arg1->error_status = RIG_OK;
    Rig_ext_token_lookup(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_token_lookup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  const_char_string arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_token_lookup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_token_lookup" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_token_lookup" "', argument " "2"" of type '" "const_char_string""'");
  }
  arg2 = (const_char_string)(buf2);
  {
    arg1->error_status = RIG_OK;
    Rig_token_lookup(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_conf__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  const_char_string arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_conf" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_conf" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_conf" "', argument " "3"" of type '" "const_char_string""'");
  }
  arg3 = (const_char_string)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_conf__SWIG_0(arg1,arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_ext_parm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_ext_parm", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ext_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ext_parm" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_ext_parm" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rig_set_ext_parm" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_ext_parm(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_parm__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_parm" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_parm__SWIG_0(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_parm__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  float arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  ecode3 = SWIG_AsVal_float(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_parm" "', argument " "3"" of type '" "float""'");
  } 
  arg3 = (float)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_parm__SWIG_1(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_parm__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_parm" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_parm" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_parm__SWIG_2(arg1,(char const *)arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_parm__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  float arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_parm" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_float(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_parm" "', argument " "3"" of type '" "float""'");
  } 
  arg3 = (float)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_parm__SWIG_3(arg1,(char const *)arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_parm__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_parm" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_parm" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_parm__SWIG_4(arg1,(char const *)arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_parm(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rig_set_parm", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_set_parm__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_float(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_set_parm__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_set_parm__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_float(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_set_parm__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Rig_set_parm__SWIG_4(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rig_set_parm'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::set_parm(setting_t,int)\n"
    "    Rig::set_parm(setting_t,float)\n"
    "    Rig::set_parm(char const *,int)\n"
    "    Rig::set_parm(char const *,float)\n"
    "    Rig::set_parm(char const *,char const *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rig_get_parm_i__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  int result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_parm_i" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_parm_i" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_parm_i__SWIG_0(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_parm_f__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  float result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_parm_f" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_parm_f" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (float)Rig_get_parm_f__SWIG_0(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_parm_i__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_parm_i" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_parm_i" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_parm_i__SWIG_1(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_parm_i(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rig_get_parm_i", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Rig_get_parm_i__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Rig_get_parm_i__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rig_get_parm_i'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_parm_i(setting_t)\n"
    "    Rig::get_parm_i(char const *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rig_get_parm_f__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  float result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_parm_f" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_parm_f" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    arg1->error_status = RIG_OK;
    result = (float)Rig_get_parm_f__SWIG_1(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_float((float)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_parm_f(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rig_get_parm_f", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Rig_get_parm_f__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Rig_get_parm_f__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rig_get_parm_f'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_parm_f(setting_t)\n"
    "    Rig::get_parm_f(char const *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rig_get_parm__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char temp3[256+1] ;
  
  arg3 = (char *) temp3;
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_parm" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    arg1->error_status = RIG_OK;
    Rig_get_parm__SWIG_0(arg1,(char const *)arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  arg3[256] = 0;  
  resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_FromCharPtr(arg3), 1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_conf__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_conf" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_conf" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_conf" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_conf__SWIG_1(arg1,(char const *)arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_conf(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rig_set_conf", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Rig_set_conf__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Rig_set_conf__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rig_set_conf'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::set_conf(hamlib_token_t,const_char_string)\n"
    "    Rig::set_conf(char const *,char const *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rig_get_freq(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  freq_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_freq", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_freq" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_freq" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (freq_t)Rig_get_freq(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_mode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  rmode_t *arg2 = 0 ;
  pbwidth_t *arg3 = 0 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rmode_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  pbwidth_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[2] ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_mode", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_mode" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_get_mode" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_mode(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_unsigned_SS_long_SS_long((*arg2)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_unsigned_long_long, new_flags), 1);
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg3)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags), 1);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_split_mode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  rmode_t *arg2 = 0 ;
  pbwidth_t *arg3 = 0 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rmode_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  pbwidth_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[2] ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_split_mode", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_split_mode" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_get_split_mode" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_split_mode(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_unsigned_SS_long_SS_long((*arg2)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_unsigned_long_long, new_flags), 1);
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg3)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags), 1);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_vfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  vfo_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_vfo" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (vfo_t)Rig_get_vfo(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_ptt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  ptt_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_ptt", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ptt" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ptt" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (ptt_t)Rig_get_ptt(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_rptr_shift(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  rptr_shift_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_rptr_shift", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_rptr_shift" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_rptr_shift" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (rptr_shift_t)Rig_get_rptr_shift(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_rptr_offs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  shortfreq_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_rptr_offs", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_rptr_offs" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_rptr_offs" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (shortfreq_t)Rig_get_rptr_offs(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_ctcss_tone(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  tone_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_ctcss_tone", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ctcss_tone" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ctcss_tone" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (tone_t)Rig_get_ctcss_tone(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_dcs_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  tone_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_dcs_code", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_dcs_code" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_dcs_code" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (tone_t)Rig_get_dcs_code(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_ctcss_sql(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  tone_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_ctcss_sql", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ctcss_sql" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ctcss_sql" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (tone_t)Rig_get_ctcss_sql(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_dcs_sql(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  tone_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_dcs_sql", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_dcs_sql" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_dcs_sql" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (tone_t)Rig_get_dcs_sql(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_split_freq(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  freq_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_split_freq", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_split_freq" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_split_freq" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (freq_t)Rig_get_split_freq(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_split_vfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  split_t *arg2 = 0 ;
  vfo_t *arg3 = 0 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  split_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  vfo_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[2] ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_split_vfo", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_split_vfo" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_get_split_vfo" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_split_vfo(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_split_t, new_flags), 1);
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_unsigned_SS_int((*arg3)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_unsigned_int, new_flags), 1);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_rit(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  shortfreq_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_rit", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_rit" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_rit" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (shortfreq_t)Rig_get_rit(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_xit(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  shortfreq_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_xit", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_xit" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_xit" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (shortfreq_t)Rig_get_xit(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_ts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  shortfreq_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_ts", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ts" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ts" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (shortfreq_t)Rig_get_ts(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_ant(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  ant_t *arg2 = 0 ;
  ant_t *arg3 = 0 ;
  ant_t *arg4 = 0 ;
  int *arg5 = 0 ;
  ant_t arg6 ;
  vfo_t arg7 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ant_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  ant_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  ant_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  int temp5 ;
  int res5 = SWIG_TMPOBJ ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[3] ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  arg5 = &temp5;
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_ant", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ant" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode6 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Rig_get_ant" "', argument " "6"" of type '" "ant_t""'");
  } 
  arg6 = (ant_t)(val6);
  if (swig_obj[2]) {
    ecode7 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Rig_get_ant" "', argument " "7"" of type '" "vfo_t""'");
    } 
    arg7 = (vfo_t)(val7);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_ant(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_unsigned_SS_int((*arg2)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_unsigned_int, new_flags), 1);
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_unsigned_SS_int((*arg3)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_unsigned_int, new_flags), 1);
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_unsigned_SS_int((*arg4)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_unsigned_int, new_flags), 1);
  }
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg5)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_int, new_flags), 1);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_vfo_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  int *arg2 = 0 ;
  split_t *arg3 = 0 ;
  pbwidth_t *arg4 = 0 ;
  rmode_t *arg5 = 0 ;
  freq_t *arg6 = 0 ;
  vfo_t arg7 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  split_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  pbwidth_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  rmode_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  freq_t temp6 ;
  int res6 = SWIG_TMPOBJ ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[2] ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  arg5 = &temp5;
  arg6 = &temp6;
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_vfo_info", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_vfo_info" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode7 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Rig_get_vfo_info" "', argument " "7"" of type '" "vfo_t""'");
    } 
    arg7 = (vfo_t)(val7);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_vfo_info(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags), 1);
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg3)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_split_t, new_flags), 1);
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg4)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_long, new_flags), 1);
  }
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_unsigned_SS_long_SS_long((*arg5)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_unsigned_long_long, new_flags), 1);
  }
  if (SWIG_IsTmpObj(res6)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_double((*arg6)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res6) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg6), SWIGTYPE_p_double, new_flags), 1);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_mem(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_mem", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_mem" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_mem" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_mem(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_powerstat(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  powerstat_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_powerstat" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (powerstat_t)Rig_get_powerstat(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_dcd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  dcd_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_dcd", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_dcd" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_dcd" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (dcd_t)Rig_get_dcd(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_freq_callback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  PyObject *arg2 = 0 ;
  PyObject *arg3 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_freq_callback", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_freq_callback" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  arg2 = swig_obj[1];
  if (swig_obj[2]) {
    arg3 = swig_obj[2];
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_freq_callback(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_mode_callback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  PyObject *arg2 = 0 ;
  PyObject *arg3 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_mode_callback", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_mode_callback" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  arg2 = swig_obj[1];
  if (swig_obj[2]) {
    arg3 = swig_obj[2];
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_mode_callback(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_vfo_callback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  PyObject *arg2 = 0 ;
  PyObject *arg3 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_vfo_callback", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_vfo_callback" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  arg2 = swig_obj[1];
  if (swig_obj[2]) {
    arg3 = swig_obj[2];
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_vfo_callback(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_ptt_callback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  PyObject *arg2 = 0 ;
  PyObject *arg3 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_ptt_callback", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ptt_callback" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  arg2 = swig_obj[1];
  if (swig_obj[2]) {
    arg3 = swig_obj[2];
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_ptt_callback(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_dcd_callback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  PyObject *arg2 = 0 ;
  PyObject *arg3 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_dcd_callback", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_dcd_callback" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  arg2 = swig_obj[1];
  if (swig_obj[2]) {
    arg3 = swig_obj[2];
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_dcd_callback(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_pltune_callback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  PyObject *arg2 = 0 ;
  PyObject *arg3 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_pltune_callback", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_pltune_callback" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  arg2 = swig_obj[1];
  if (swig_obj[2]) {
    arg3 = swig_obj[2];
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_pltune_callback(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_spectrum_callback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  PyObject *arg2 = 0 ;
  PyObject *arg3 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_spectrum_callback", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_spectrum_callback" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  arg2 = swig_obj[1];
  if (swig_obj[2]) {
    arg3 = swig_obj[2];
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_spectrum_callback(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_mem_count(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_mem_count" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_mem_count(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_lookup_mem_caps(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  int arg2 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  chan_t *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_lookup_mem_caps", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_lookup_mem_caps" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_lookup_mem_caps" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (chan_t *)Rig_lookup_mem_caps(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_chan_list, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_set_channel(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  struct channel *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_set_channel", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_channel" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_channel" "', argument " "2"" of type '" "struct channel const *""'"); 
  }
  arg2 = (struct channel *)(argp2);
  {
    arg1->error_status = RIG_OK;
    Rig_set_channel(arg1,(struct channel const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_channel__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  struct channel *arg2 = 0 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_channel" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_channel" "', argument " "2"" of type '" "struct channel *""'"); 
  }
  arg2 = (struct channel *)(argp2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_channel" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_get_channel__SWIG_0(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_chan_all__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  struct channel *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_chan_all" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_chan_all" "', argument " "2"" of type '" "struct channel *""'"); 
  }
  arg2 = (struct channel *)(argp2);
  {
    arg1->error_status = RIG_OK;
    Rig_get_chan_all__SWIG_0(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_chan_all__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct channel *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_chan_all" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (struct channel *)Rig_get_chan_all(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_channel, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_chan_all(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rig_get_chan_all", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Rig_get_chan_all__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_channel, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Rig_get_chan_all__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rig_get_chan_all'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_chan_all(struct channel *)\n"
    "    Rig::get_chan_all()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rig_get_channel__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  int arg2 ;
  int arg3 = INT_MAX ;
  vfo_t arg4 = (1u << (28)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  struct channel *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_channel" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_channel" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_channel" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_get_channel" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    result = (struct channel *)Rig_get_channel__SWIG_1(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_channel, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_channel(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rig_get_channel", 0, 4, argv))) SWIG_fail;
  --argc;
  if ((argc >= 2) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_channel__SWIG_1(self, argc, argv);
        }
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          if (argc <= 3) {
            return _wrap_Rig_get_channel__SWIG_1(self, argc, argv);
          }
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Rig_get_channel__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_channel, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_channel__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rig_get_channel'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_channel(struct channel *,int)\n"
    "    Rig::get_channel(int,int,vfo_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rig_get_conf__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  char temp3[256+1] ;
  
  arg3 = (char *) temp3;
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_conf" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_conf" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_get_conf__SWIG_0(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  arg3[256] = 0;  
  resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_FromCharPtr(arg3), 1);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_conf__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char temp3[256+1] ;
  
  arg3 = (char *) temp3;
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_conf" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_conf" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    arg1->error_status = RIG_OK;
    Rig_get_conf__SWIG_1(arg1,(char const *)arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  arg3[256] = 0;  
  resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_FromCharPtr(arg3), 1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_conf(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rig_get_conf", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Rig_get_conf__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Rig_get_conf__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rig_get_conf'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_conf(hamlib_token_t,char *)\n"
    "    Rig::get_conf(char const *,char *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rig_recv_dtmf(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char temp2[256+1] ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[2] ;
  
  arg2 = (char *) temp2;
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_recv_dtmf", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_recv_dtmf" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (swig_obj[1]) {
    ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_recv_dtmf" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_recv_dtmf(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  arg2[256] = 0;  
  resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_FromCharPtr(arg2), 1);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_info" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Rig_get_info(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_func", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_func" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_func" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_func(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_ext_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_ext_func", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ext_func" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ext_func" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_ext_func" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_ext_func(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_send_raw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  PyObject *arg2 = 0 ;
  PyObject *arg3 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_send_raw", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_send_raw" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  arg2 = swig_obj[1];
  if (swig_obj[2]) {
    arg3 = swig_obj[2];
  }
  {
    arg1->error_status = RIG_OK;
    result = (PyObject *)Rig_send_raw(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = result;
  if (!resultobj) SWIG_fail;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_level__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_level" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_level__SWIG_1(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_level(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rig_get_level", 0, 3, argv))) SWIG_fail;
  --argc;
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_level__SWIG_1(self, argc, argv);
        }
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_level__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_level__SWIG_0(self, argc, argv);
        }
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_level__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rig_get_level'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_level(char const *,char *,vfo_t)\n"
    "    Rig::get_level(setting_t,vfo_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rig_get_ext_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_ext_level", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ext_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ext_level" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_ext_level" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_ext_level(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_parm__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_get_parm__SWIG_1(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rig_get_parm(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rig_get_parm", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Rig_get_parm__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Rig_get_parm__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rig_get_parm'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_parm(char const *,char *)\n"
    "    Rig::get_parm(setting_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rig_get_ext_parm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rig_get_ext_parm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ext_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ext_parm" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_get_ext_parm(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Rig_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Rig, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Rig_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_channelArray(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  channelArray *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_channelArray" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (channelArray *)new_channelArray(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_channelArray, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_channelArray(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  channelArray *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channelArray, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_channelArray" "', argument " "1"" of type '" "channelArray *""'"); 
  }
  arg1 = (channelArray *)(argp1);
  delete_channelArray(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channelArray_cast(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  channelArray *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct channel *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channelArray, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channelArray_cast" "', argument " "1"" of type '" "channelArray *""'"); 
  }
  arg1 = (channelArray *)(argp1);
  result = (struct channel *)channelArray_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_channel, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_channelArray_frompointer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  channelArray *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channelArray_frompointer" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (channelArray *)channelArray_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_channelArray, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *channelArray_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_channelArray, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *channelArray_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_toneArray(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  toneArray *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_toneArray" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (toneArray *)new_toneArray(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_toneArray, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_toneArray(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  toneArray *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_toneArray, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_toneArray" "', argument " "1"" of type '" "toneArray *""'"); 
  }
  arg1 = (toneArray *)(argp1);
  delete_toneArray(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_toneArray_cast(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  toneArray *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  tone_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_toneArray, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toneArray_cast" "', argument " "1"" of type '" "toneArray *""'"); 
  }
  arg1 = (toneArray *)(argp1);
  result = (tone_t *)toneArray_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_toneArray_frompointer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  tone_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  toneArray *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toneArray_frompointer" "', argument " "1"" of type '" "tone_t *""'"); 
  }
  arg1 = (tone_t *)(argp1);
  result = (toneArray *)toneArray_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_toneArray, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *toneArray_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_toneArray, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *toneArray_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_rot_caps_rot_model_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  rot_model_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_rot_model_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_model_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_rot_model_set" "', argument " "2"" of type '" "rot_model_t""'");
  } 
  arg2 = (rot_model_t)(val2);
  if (arg1) (arg1)->rot_model = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_rot_model_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rot_model_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_model_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (rot_model_t) ((arg1)->rot_model);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_model_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_model_name_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (char *) ((arg1)->model_name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_mfg_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_mfg_name_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (char *) ((arg1)->mfg_name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_version_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_version_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (char *) ((arg1)->version);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_copyright_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_copyright_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (char *) ((arg1)->copyright);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_status_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  enum rig_status_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_status_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_status_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_status_set" "', argument " "2"" of type '" "enum rig_status_e""'");
  } 
  arg2 = (enum rig_status_e)(val2);
  if (arg1) (arg1)->status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_status_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum rig_status_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_status_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (enum rig_status_e) ((arg1)->status);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_rot_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_rot_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_type_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_rot_type_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->rot_type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_rot_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_type_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->rot_type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_port_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  enum rig_port_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_port_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_port_type_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_port_type_set" "', argument " "2"" of type '" "enum rig_port_e""'");
  } 
  arg2 = (enum rig_port_e)(val2);
  if (arg1) (arg1)->port_type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_port_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum rig_port_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_port_type_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (enum rig_port_e) ((arg1)->port_type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_serial_rate_min_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_serial_rate_min_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_rate_min_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_serial_rate_min_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_min = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_serial_rate_min_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_rate_min_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->serial_rate_min);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_serial_rate_max_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_serial_rate_max_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_rate_max_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_serial_rate_max_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_max = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_serial_rate_max_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_rate_max_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->serial_rate_max);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_serial_data_bits_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_serial_data_bits_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_data_bits_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_serial_data_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_data_bits = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_serial_data_bits_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_data_bits_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->serial_data_bits);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_serial_stop_bits_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_serial_stop_bits_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_stop_bits_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_serial_stop_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_stop_bits = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_serial_stop_bits_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_stop_bits_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->serial_stop_bits);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_serial_parity_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  enum serial_parity_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_serial_parity_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_parity_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_serial_parity_set" "', argument " "2"" of type '" "enum serial_parity_e""'");
  } 
  arg2 = (enum serial_parity_e)(val2);
  if (arg1) (arg1)->serial_parity = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_serial_parity_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum serial_parity_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_parity_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (enum serial_parity_e) ((arg1)->serial_parity);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_serial_handshake_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  enum serial_handshake_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_serial_handshake_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_handshake_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_serial_handshake_set" "', argument " "2"" of type '" "enum serial_handshake_e""'");
  } 
  arg2 = (enum serial_handshake_e)(val2);
  if (arg1) (arg1)->serial_handshake = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_serial_handshake_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum serial_handshake_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_handshake_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (enum serial_handshake_e) ((arg1)->serial_handshake);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_write_delay_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_write_delay_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_write_delay_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->write_delay = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_write_delay_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_write_delay_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->write_delay);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_post_write_delay_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_post_write_delay_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_post_write_delay_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_post_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->post_write_delay = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_post_write_delay_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_post_write_delay_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->post_write_delay);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_timeout_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_timeout_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_timeout_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_timeout_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->timeout = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_timeout_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_timeout_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->timeout);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_retry_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_retry_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_retry_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_retry_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->retry = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_retry_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_retry_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->retry);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_get_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_has_get_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_get_func_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_get_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_get_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_get_func_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_func);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_set_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_has_set_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_set_func_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_set_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_set_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_set_func_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_func);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_get_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_has_get_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_get_level_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_get_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_get_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_get_level_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_level);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_set_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_has_set_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_set_level_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_set_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_set_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_set_level_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_level);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_get_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_has_get_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_get_parm_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_get_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_get_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_get_parm_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_parm);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_set_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_has_set_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_set_parm_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_set_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_set_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_set_parm_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_parm);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_status_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  rot_status_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_has_status_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_status_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_status_set" "', argument " "2"" of type '" "rot_status_t""'");
  } 
  arg2 = (rot_status_t)(val2);
  if (arg1) (arg1)->has_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_has_status_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rot_status_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_status_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (rot_status_t) ((arg1)->has_status);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_level_gran_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_level_gran_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_level_gran_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_level_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->level_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""level_gran""' of type '""gran_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_level_gran_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  gran_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_level_gran_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->level_gran);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_parm_gran_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_parm_gran_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_parm_gran_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_parm_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->parm_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""parm_gran""' of type '""gran_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_parm_gran_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  gran_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_parm_gran_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->parm_gran);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_extparms_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_extparms_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_extparms_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_extparms_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extparms = (struct confparams const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_extparms_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct confparams *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_extparms_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extparms);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_extlevels_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_extlevels_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_extlevels_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_extlevels_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extlevels = (struct confparams const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_extlevels_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct confparams *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_extlevels_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extlevels);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_extfuncs_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_extfuncs_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_extfuncs_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_extfuncs_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extfuncs = (struct confparams const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_extfuncs_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct confparams *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_extfuncs_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extfuncs);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_ext_tokens_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  PyObject *swig_obj[1] ;
  
  arg2 = &temp2;
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_ext_tokens_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  if (arg1) (arg1)->ext_tokens = arg2;
  resultobj = SWIG_Py_Void();
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags), 1);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_ext_tokens_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_ext_tokens_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int *) ((arg1)->ext_tokens);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_min_az_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  azimuth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_min_az_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_min_az_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_min_az_set" "', argument " "2"" of type '" "azimuth_t""'");
  } 
  arg2 = (azimuth_t)(val2);
  if (arg1) (arg1)->min_az = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_min_az_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  azimuth_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_min_az_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (azimuth_t) ((arg1)->min_az);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_max_az_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  azimuth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_max_az_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_max_az_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_max_az_set" "', argument " "2"" of type '" "azimuth_t""'");
  } 
  arg2 = (azimuth_t)(val2);
  if (arg1) (arg1)->max_az = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_max_az_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  azimuth_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_max_az_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (azimuth_t) ((arg1)->max_az);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_min_el_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  elevation_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_min_el_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_min_el_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_min_el_set" "', argument " "2"" of type '" "elevation_t""'");
  } 
  arg2 = (elevation_t)(val2);
  if (arg1) (arg1)->min_el = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_min_el_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  elevation_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_min_el_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (elevation_t) ((arg1)->min_el);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_max_el_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  elevation_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_max_el_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_max_el_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_max_el_set" "', argument " "2"" of type '" "elevation_t""'");
  } 
  arg2 = (elevation_t)(val2);
  if (arg1) (arg1)->max_el = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_max_el_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  elevation_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_max_el_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (elevation_t) ((arg1)->max_el);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_cfgparams_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_cfgparams_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_cfgparams_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_cfgparams_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->cfgparams = (struct confparams const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_cfgparams_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct confparams *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_cfgparams_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (struct confparams *) ((arg1)->cfgparams);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_priv_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_priv_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_priv_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_priv_set" "', argument " "2"" of type '" "void const *""'"); 
  }
  if (arg1) (arg1)->priv = (void const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_priv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_priv_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (void *) ((arg1)->priv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_rot_init_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_rot_init_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_init_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_rot_init_set" "', argument " "2"" of type '" "int (*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->rot_init = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_rot_init_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_init_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *)) ((arg1)->rot_init);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_rot_cleanup_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_rot_cleanup_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_cleanup_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_rot_cleanup_set" "', argument " "2"" of type '" "int (*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->rot_cleanup = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_rot_cleanup_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_cleanup_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *)) ((arg1)->rot_cleanup);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_rot_open_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_rot_open_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_open_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_rot_open_set" "', argument " "2"" of type '" "int (*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->rot_open = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_rot_open_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_open_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *)) ((arg1)->rot_open);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_rot_close_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_rot_close_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_close_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_rot_close_set" "', argument " "2"" of type '" "int (*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->rot_close = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_rot_close_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_close_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *)) ((arg1)->rot_close);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_conf_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_set_conf_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_conf_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_conf_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,char const *)""'"); 
    }
  }
  if (arg1) (arg1)->set_conf = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_conf_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,hamlib_token_t,char const *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_conf_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,char const *)) ((arg1)->set_conf);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_p_q_const__char__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_conf_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,char *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_get_conf_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_conf_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_p_char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_conf_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,char *)""'"); 
    }
  }
  if (arg1) (arg1)->get_conf = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_conf_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,hamlib_token_t,char *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_conf_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,char *)) ((arg1)->get_conf);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_p_char__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_position_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,azimuth_t,elevation_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_set_position_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_position_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_float_float__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_position_set" "', argument " "2"" of type '" "int (*)(ROT *,azimuth_t,elevation_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_position = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_position_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,azimuth_t,elevation_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_position_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,azimuth_t,elevation_t)) ((arg1)->set_position);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_float_float__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_position_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,azimuth_t *,elevation_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_get_position_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_position_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_p_float_p_float__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_position_set" "', argument " "2"" of type '" "int (*)(ROT *,azimuth_t *,elevation_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_position = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_position_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,azimuth_t *,elevation_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_position_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,azimuth_t *,elevation_t *)) ((arg1)->get_position);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_p_float_p_float__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_stop_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_stop_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_stop_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_stop_set" "', argument " "2"" of type '" "int (*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->stop = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_stop_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_stop_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *)) ((arg1)->stop);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_park_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_park_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_park_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_park_set" "', argument " "2"" of type '" "int (*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->park = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_park_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_park_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *)) ((arg1)->park);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_reset_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,rot_reset_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_reset_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_reset_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_reset_set" "', argument " "2"" of type '" "int (*)(ROT *,rot_reset_t)""'"); 
    }
  }
  if (arg1) (arg1)->reset = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_reset_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,rot_reset_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_reset_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,rot_reset_t)) ((arg1)->reset);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_move_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,int,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_move_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_move_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_int_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_move_set" "', argument " "2"" of type '" "int (*)(ROT *,int,int)""'"); 
    }
  }
  if (arg1) (arg1)->move = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_move_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,int,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_move_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,int,int)) ((arg1)->move);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_int_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_info_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  char *(*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_get_info_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_info_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__p_char);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_info_set" "', argument " "2"" of type '" "char const *(*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->get_info = (char const *(*)(ROT *))arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_info_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *(*result)(ROT *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_info_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (char *(*)(ROT *)) ((arg1)->get_info);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__p_char);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,setting_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_set_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_level_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_level_set" "', argument " "2"" of type '" "int (*)(ROT *,setting_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,setting_t,value_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_level_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,setting_t,value_t)) ((arg1)->set_level);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,setting_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_get_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_level_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_level_set" "', argument " "2"" of type '" "int (*)(ROT *,setting_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,setting_t,value_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_level_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,setting_t,value_t *)) ((arg1)->get_level);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,setting_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_set_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_func_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_func_set" "', argument " "2"" of type '" "int (*)(ROT *,setting_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,setting_t,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_func_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,setting_t,int)) ((arg1)->set_func);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,setting_t,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_get_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_func_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_func_set" "', argument " "2"" of type '" "int (*)(ROT *,setting_t,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,setting_t,int *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_func_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,setting_t,int *)) ((arg1)->get_func);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,setting_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_set_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_parm_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_parm_set" "', argument " "2"" of type '" "int (*)(ROT *,setting_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,setting_t,value_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_parm_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,setting_t,value_t)) ((arg1)->set_parm);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,setting_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_get_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_parm_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_parm_set" "', argument " "2"" of type '" "int (*)(ROT *,setting_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,setting_t,value_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_parm_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,setting_t,value_t *)) ((arg1)->get_parm);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_ext_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_set_ext_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_ext_level_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_ext_level_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_ext_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,hamlib_token_t,value_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_ext_level_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,value_t)) ((arg1)->set_ext_level);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_ext_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_get_ext_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_ext_level_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_ext_level_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_ext_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,hamlib_token_t,value_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_ext_level_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,value_t *)) ((arg1)->get_ext_level);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_p_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_ext_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_set_ext_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_ext_func_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_ext_func_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_ext_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,hamlib_token_t,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_ext_func_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,int)) ((arg1)->set_ext_func);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_ext_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_get_ext_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_ext_func_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_ext_func_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_ext_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,hamlib_token_t,int *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_ext_func_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,int *)) ((arg1)->get_ext_func);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_p_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_ext_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_set_ext_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_ext_parm_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_ext_parm_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_set_ext_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,hamlib_token_t,value_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_ext_parm_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,value_t)) ((arg1)->set_ext_parm);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_ext_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_get_ext_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_ext_parm_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_ext_parm_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_ext_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,hamlib_token_t,value_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_ext_parm_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,value_t *)) ((arg1)->get_ext_parm);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_p_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_status_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,rot_status_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_get_status_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_status_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_p_enum_rot_status_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_status_set" "', argument " "2"" of type '" "int (*)(ROT *,rot_status_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_status_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,rot_status_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_status_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,rot_status_t *)) ((arg1)->get_status);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_p_enum_rot_status_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_conf2_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,char *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_caps_get_conf2_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_conf2_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_p_char_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_conf2_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,char *,int)""'"); 
    }
  }
  if (arg1) (arg1)->get_conf2 = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_caps_get_conf2_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(ROT *,hamlib_token_t,char *,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_conf2_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,char *,int)) ((arg1)->get_conf2);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_p_char_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_rot_caps(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_rot_caps", 0, 0, 0)) SWIG_fail;
  result = (struct rot_caps *)calloc(1, sizeof(struct rot_caps));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rot_caps, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_rot_caps(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_caps, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rot_caps" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *rot_caps_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_rot_caps, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *rot_caps_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_qrb(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  double arg1 ;
  double arg2 ;
  double arg3 ;
  double arg4 ;
  double *arg5 = 0 ;
  double *arg6 = 0 ;
  double val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  double temp5 ;
  int res5 = SWIG_TMPOBJ ;
  double temp6 ;
  int res6 = SWIG_TMPOBJ ;
  PyObject *swig_obj[4] ;
  int result;
  
  arg5 = &temp5;
  arg6 = &temp6;
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "qrb", 4, 4, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_double(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "qrb" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = (double)(val1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "qrb" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "qrb" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = (double)(val3);
  ecode4 = SWIG_AsVal_double(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "qrb" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = (double)(val4);
  result = (int)qrb(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_int((int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_double((*arg5)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_double, new_flags), 0);
  }
  if (SWIG_IsTmpObj(res6)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_double((*arg6)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res6) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg6), SWIGTYPE_p_double, new_flags), 0);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_distance_long_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  double arg1 ;
  double val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_double(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "distance_long_path" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = (double)(val1);
  result = (double)distance_long_path(arg1);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_azimuth_long_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  double arg1 ;
  double val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_double(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "azimuth_long_path" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = (double)(val1);
  result = (double)azimuth_long_path(arg1);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_dms2dec(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int arg1 ;
  int arg2 ;
  double arg3 ;
  int arg4 ;
  int val1 ;
  int ecode1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  double result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "dms2dec", 4, 4, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "dms2dec" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = (int)(val1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "dms2dec" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "dms2dec" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = (double)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "dms2dec" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = (int)(val4);
  result = (double)dms2dec(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_dec2dms(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  double arg1 ;
  int *arg2 = 0 ;
  int *arg3 = 0 ;
  double *arg4 = 0 ;
  int *arg5 = 0 ;
  double val1 ;
  int ecode1 = 0 ;
  int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  int temp3 ;
  int res3 = SWIG_TMPOBJ ;
  double temp4 ;
  int res4 = SWIG_TMPOBJ ;
  int temp5 ;
  int res5 = SWIG_TMPOBJ ;
  PyObject *swig_obj[1] ;
  int result;
  
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  arg5 = &temp5;
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_double(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "dec2dms" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = (double)(val1);
  result = (int)dec2dms(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_int((int)(result));
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags), 0);
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg3)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags), 0);
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg4)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_double, new_flags), 0);
  }
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg5)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_int, new_flags), 0);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_dec2dmmm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  double arg1 ;
  int *arg2 = 0 ;
  double *arg3 = 0 ;
  int *arg4 = 0 ;
  double val1 ;
  int ecode1 = 0 ;
  int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  double temp3 ;
  int res3 = SWIG_TMPOBJ ;
  int temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject *swig_obj[1] ;
  int result;
  
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_double(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "dec2dmmm" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = (double)(val1);
  result = (int)dec2dmmm(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_int((int)(result));
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags), 0);
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg3)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_double, new_flags), 0);
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg4)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags), 0);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_dmmm2dec(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int arg1 ;
  double arg2 ;
  int arg3 ;
  int val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  double result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "dmmm2dec", 3, 3, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "dmmm2dec" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = (int)(val1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "dmmm2dec" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "dmmm2dec" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  result = (double)dmmm2dec(arg1,arg2,arg3);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_min_az_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  azimuth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_min_az_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_min_az_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_min_az_set" "', argument " "2"" of type '" "azimuth_t""'");
  } 
  arg2 = (azimuth_t)(val2);
  if (arg1) (arg1)->min_az = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_min_az_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  azimuth_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_min_az_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (azimuth_t) ((arg1)->min_az);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_max_az_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  azimuth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_max_az_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_max_az_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_max_az_set" "', argument " "2"" of type '" "azimuth_t""'");
  } 
  arg2 = (azimuth_t)(val2);
  if (arg1) (arg1)->max_az = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_max_az_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  azimuth_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_max_az_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (azimuth_t) ((arg1)->max_az);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_min_el_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  elevation_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_min_el_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_min_el_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_min_el_set" "', argument " "2"" of type '" "elevation_t""'");
  } 
  arg2 = (elevation_t)(val2);
  if (arg1) (arg1)->min_el = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_min_el_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  elevation_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_min_el_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (elevation_t) ((arg1)->min_el);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_max_el_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  elevation_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_max_el_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_max_el_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_max_el_set" "', argument " "2"" of type '" "elevation_t""'");
  } 
  arg2 = (elevation_t)(val2);
  if (arg1) (arg1)->max_el = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_max_el_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  elevation_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_max_el_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (elevation_t) ((arg1)->max_el);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_south_zero_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_south_zero_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_south_zero_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_south_zero_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->south_zero = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_south_zero_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_south_zero_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (int) ((arg1)->south_zero);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_az_offset_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  azimuth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_az_offset_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_az_offset_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_az_offset_set" "', argument " "2"" of type '" "azimuth_t""'");
  } 
  arg2 = (azimuth_t)(val2);
  if (arg1) (arg1)->az_offset = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_az_offset_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  azimuth_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_az_offset_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (azimuth_t) ((arg1)->az_offset);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_el_offset_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  elevation_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_el_offset_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_el_offset_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_el_offset_set" "', argument " "2"" of type '" "elevation_t""'");
  } 
  arg2 = (elevation_t)(val2);
  if (arg1) (arg1)->el_offset = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_el_offset_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  elevation_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_el_offset_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (elevation_t) ((arg1)->el_offset);
  resultobj = SWIG_From_float((float)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_get_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_has_get_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_get_func_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_get_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_get_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_get_func_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_func);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_set_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_has_set_func_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_set_func_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_set_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_set_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_set_func_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_func);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_get_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_has_get_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_get_level_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_get_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_get_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_get_level_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_level);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_set_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_has_set_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_set_level_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_set_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_set_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_set_level_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_level);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_get_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_has_get_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_get_parm_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_get_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_get_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_get_parm_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_parm);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_set_parm_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_has_set_parm_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_set_parm_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_set_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_parm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_set_parm_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_set_parm_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_parm);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_status_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  rot_status_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_has_status_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_status_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_status_set" "', argument " "2"" of type '" "rot_status_t""'");
  } 
  arg2 = (rot_status_t)(val2);
  if (arg1) (arg1)->has_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_has_status_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  rot_status_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_status_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (rot_status_t) ((arg1)->has_status);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_level_gran_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_level_gran_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_level_gran_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_level_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->level_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""level_gran""' of type '""gran_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_level_gran_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  gran_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_level_gran_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->level_gran);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_parm_gran_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_parm_gran_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_parm_gran_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_parm_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->parm_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""parm_gran""' of type '""gran_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_parm_gran_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  gran_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_parm_gran_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->parm_gran);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_comm_state_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_comm_state_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_comm_state_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_comm_state_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->comm_state = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_comm_state_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_comm_state_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (int) ((arg1)->comm_state);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_priv_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_priv_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_priv_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_priv_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->priv = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_priv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_priv_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (void *) ((arg1)->priv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_obj_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_obj_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_obj_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_obj_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->obj = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_obj_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_obj_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (void *) ((arg1)->obj);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_current_speed_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_current_speed_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_current_speed_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_current_speed_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->current_speed = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_current_speed_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_current_speed_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (int) ((arg1)->current_speed);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_rotport_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  hamlib_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_rotport_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_rotport_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_hamlib_port,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_rotport_set" "', argument " "2"" of type '" "hamlib_port_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rot_state_rotport_set" "', argument " "2"" of type '" "hamlib_port_t""'");
    } else {
      arg2 = *((hamlib_port_t *)(argp2));
    }
  }
  if (arg1) (arg1)->rotport = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_rotport_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  hamlib_port_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_rotport_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result =  ((arg1)->rotport);
  resultobj = SWIG_NewPointerObj((hamlib_port_t *)memcpy((hamlib_port_t *)calloc(1,sizeof(hamlib_port_t)),&result,sizeof(hamlib_port_t)), SWIGTYPE_p_hamlib_port, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_rotport2_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  hamlib_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_rotport2_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_rotport2_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_hamlib_port,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_rotport2_set" "', argument " "2"" of type '" "hamlib_port_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rot_state_rotport2_set" "', argument " "2"" of type '" "hamlib_port_t""'");
    } else {
      arg2 = *((hamlib_port_t *)(argp2));
    }
  }
  if (arg1) (arg1)->rotport2 = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_rotport2_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  hamlib_port_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_rotport2_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result =  ((arg1)->rotport2);
  resultobj = SWIG_NewPointerObj((hamlib_port_t *)memcpy((hamlib_port_t *)calloc(1,sizeof(hamlib_port_t)),&result,sizeof(hamlib_port_t)), SWIGTYPE_p_hamlib_port, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_pstrotator_handler_priv_data_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void **arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_pstrotator_handler_priv_data_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_pstrotator_handler_priv_data_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_p_void, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_pstrotator_handler_priv_data_set" "', argument " "2"" of type '" "void **""'"); 
  }
  arg2 = (void **)(argp2);
  if (arg1) (arg1)->pstrotator_handler_priv_data = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_pstrotator_handler_priv_data_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void **result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_pstrotator_handler_priv_data_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (void **) ((arg1)->pstrotator_handler_priv_data);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_config_queue_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  deferred_config_header_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rot_state_config_queue_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_config_queue_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_deferred_config_header, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_config_queue_set" "', argument " "2"" of type '" "deferred_config_header_t *""'"); 
  }
  arg2 = (deferred_config_header_t *)(argp2);
  if (arg1) (arg1)->config_queue = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_rot_state_config_queue_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  deferred_config_header_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_config_queue_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (deferred_config_header_t *)& ((arg1)->config_queue);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_deferred_config_header, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_rot_state(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_rot_state", 0, 0, 0)) SWIG_fail;
  result = (struct rot_state *)calloc(1, sizeof(struct rot_state));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rot_state, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_rot_state(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_rot_state, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rot_state" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *rot_state_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_rot_state, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *rot_state_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Rot_rot_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  ROT *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_rot_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_rot_set" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_s_rot, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rot_rot_set" "', argument " "2"" of type '" "ROT *""'"); 
  }
  arg2 = (ROT *)(argp2);
  if (arg1) (arg1)->rot = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_rot_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  ROT *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_rot_get" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  result = (ROT *) ((arg1)->rot);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_s_rot, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_caps_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  struct rot_caps *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_caps_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_caps_set" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_rot_caps, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rot_caps_set" "', argument " "2"" of type '" "struct rot_caps *""'"); 
  }
  arg2 = (struct rot_caps *)(argp2);
  if (arg1) (arg1)->caps = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_caps_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct rot_caps *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_caps_get" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  result = (struct rot_caps *) ((arg1)->caps);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rot_caps, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_state_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  struct rot_state *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_state_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_state_set" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_rot_state, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rot_state_set" "', argument " "2"" of type '" "struct rot_state *""'"); 
  }
  arg2 = (struct rot_state *)(argp2);
  if (arg1) (arg1)->state = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_state_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct rot_state *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_state_get" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  result = (struct rot_state *) ((arg1)->state);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_rot_state, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_error_status_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_error_status_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_error_status_set" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_error_status_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->error_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_error_status_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_error_status_get" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  result = (int) ((arg1)->error_status);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_do_exception_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_do_exception_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_do_exception_set" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_do_exception_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->do_exception = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_do_exception_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_do_exception_get" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  result = (int) ((arg1)->do_exception);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Rot(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  rot_model_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  struct Rot *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Rot" "', argument " "1"" of type '" "rot_model_t""'");
  } 
  arg1 = (rot_model_t)(val1);
  result = (struct Rot *)new_Rot(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Rot, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Rot(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Rot" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  delete_Rot(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_open(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_open" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rot_open(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_close(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_close" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rot_close(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_set_ext_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_set_ext_func", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_ext_func" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_ext_func" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rot_set_ext_func" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rot_set_ext_func(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_set_ext_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_set_ext_level", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_ext_level" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_ext_level" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rot_set_ext_level" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rot_set_ext_level" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Rot_set_ext_level(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_set_ext_parm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_set_ext_parm", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_ext_parm" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_ext_parm" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rot_set_ext_parm" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rot_set_ext_parm" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Rot_set_ext_parm(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_set_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_set_func", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_func" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rot_set_func" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rot_set_func(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_set_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_set_level", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_level" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rot_set_level" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rot_set_level" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Rot_set_level(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_set_parm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_set_parm", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_parm" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rot_set_parm" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rot_set_parm" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Rot_set_parm(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_set_position(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  azimuth_t arg2 ;
  elevation_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_set_position", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_position" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_position" "', argument " "2"" of type '" "azimuth_t""'");
  } 
  arg2 = (azimuth_t)(val2);
  ecode3 = SWIG_AsVal_float(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rot_set_position" "', argument " "3"" of type '" "elevation_t""'");
  } 
  arg3 = (elevation_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rot_set_position(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_get_position(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  azimuth_t *arg2 = 0 ;
  elevation_t *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  azimuth_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  elevation_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject *swig_obj[1] ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_position" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rot_get_position(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_float((*arg2)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_float, new_flags), 1);
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_float((*arg3)), 1);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_float, new_flags), 1);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_stop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_stop" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rot_stop(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_park(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_park" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rot_park(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  rot_reset_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_reset", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_reset" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_reset" "', argument " "2"" of type '" "rot_reset_t""'");
  } 
  arg2 = (rot_reset_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rot_reset(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_move(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  int arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_move", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_move" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_move" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rot_move" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rot_move(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_token_lookup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  const_char_string arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_token_lookup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_token_lookup" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rot_token_lookup" "', argument " "2"" of type '" "const_char_string""'");
  }
  arg2 = (const_char_string)(buf2);
  {
    arg1->error_status = RIG_OK;
    Rot_token_lookup(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_set_conf__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_conf" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rot_set_conf" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rot_set_conf" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rot_set_conf__SWIG_0(arg1,(char const *)arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_set_conf__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  const_char_string arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_conf" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_conf" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rot_set_conf" "', argument " "3"" of type '" "const_char_string""'");
  }
  arg3 = (const_char_string)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rot_set_conf__SWIG_1(arg1,arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_set_conf(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rot_set_conf", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rot, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Rot_set_conf__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rot, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Rot_set_conf__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rot_set_conf'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rot::set_conf(char const *,char const *)\n"
    "    Rot::set_conf(hamlib_token_t,const_char_string)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rot_has_get_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  setting_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_has_get_func", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_has_get_func" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_has_get_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Rot_has_get_func(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_has_set_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  setting_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_has_set_func", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_has_set_func" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_has_set_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Rot_has_set_func(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_has_get_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  setting_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_has_get_level", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_has_get_level" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_has_get_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Rot_has_get_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_has_set_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  setting_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_has_set_level", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_has_set_level" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_has_set_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Rot_has_set_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_has_get_parm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  setting_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_has_get_parm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_has_get_parm" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_has_get_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Rot_has_get_parm(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_has_set_parm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  setting_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_has_set_parm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_has_set_parm" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_has_set_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Rot_has_set_parm(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_get_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_get_func", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_func" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_get_func" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (PyObject *)Rot_get_func(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = result;
  if (!resultobj) SWIG_fail;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_get_ext_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_get_ext_func", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_ext_func" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_get_ext_func" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (PyObject *)Rot_get_ext_func(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = result;
  if (!resultobj) SWIG_fail;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_get_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_get_level", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_level" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_get_level" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (PyObject *)Rot_get_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = result;
  if (!resultobj) SWIG_fail;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_get_ext_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_get_ext_level", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_ext_level" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_get_ext_level" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (PyObject *)Rot_get_ext_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = result;
  if (!resultobj) SWIG_fail;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_get_parm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_get_parm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_parm" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_get_parm" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (PyObject *)Rot_get_parm(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = result;
  if (!resultobj) SWIG_fail;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_get_ext_parm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Rot_get_ext_parm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_ext_parm" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_get_ext_parm" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (PyObject *)Rot_get_ext_parm(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = result;
  if (!resultobj) SWIG_fail;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_get_conf__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  char *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_conf" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_get_conf" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Rot_get_conf__SWIG_0(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_get_conf__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_conf" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rot_get_conf" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Rot_get_conf__SWIG_1(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rot_get_conf(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Rot_get_conf", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rot, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Rot_get_conf__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rot, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Rot_get_conf__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Rot_get_conf'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rot::get_conf(hamlib_token_t)\n"
    "    Rot::get_conf(char const *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Rot_get_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_info" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Rot_get_info(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Rot_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Rot, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Rot_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_amp_caps_amp_model_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  amp_model_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_amp_model_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_model_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_amp_model_set" "', argument " "2"" of type '" "amp_model_t""'");
  } 
  arg2 = (amp_model_t)(val2);
  if (arg1) (arg1)->amp_model = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_model_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  amp_model_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_model_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (amp_model_t) ((arg1)->amp_model);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_model_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_model_name_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (char *) ((arg1)->model_name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_mfg_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_mfg_name_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (char *) ((arg1)->mfg_name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_version_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_version_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (char *) ((arg1)->version);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_copyright_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_copyright_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (char *) ((arg1)->copyright);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_status_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  enum rig_status_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_status_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_status_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_status_set" "', argument " "2"" of type '" "enum rig_status_e""'");
  } 
  arg2 = (enum rig_status_e)(val2);
  if (arg1) (arg1)->status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_status_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum rig_status_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_status_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (enum rig_status_e) ((arg1)->status);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_amp_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_type_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_amp_type_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->amp_type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_type_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->amp_type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_port_type_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  enum rig_port_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_port_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_port_type_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_port_type_set" "', argument " "2"" of type '" "enum rig_port_e""'");
  } 
  arg2 = (enum rig_port_e)(val2);
  if (arg1) (arg1)->port_type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_port_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum rig_port_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_port_type_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (enum rig_port_e) ((arg1)->port_type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_serial_rate_min_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_serial_rate_min_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_rate_min_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_serial_rate_min_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_min = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_serial_rate_min_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_rate_min_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->serial_rate_min);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_serial_rate_max_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_serial_rate_max_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_rate_max_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_serial_rate_max_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_max = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_serial_rate_max_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_rate_max_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->serial_rate_max);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_serial_data_bits_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_serial_data_bits_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_data_bits_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_serial_data_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_data_bits = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_serial_data_bits_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_data_bits_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->serial_data_bits);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_serial_stop_bits_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_serial_stop_bits_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_stop_bits_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_serial_stop_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_stop_bits = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_serial_stop_bits_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_stop_bits_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->serial_stop_bits);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_serial_parity_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  enum serial_parity_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_serial_parity_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_parity_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_serial_parity_set" "', argument " "2"" of type '" "enum serial_parity_e""'");
  } 
  arg2 = (enum serial_parity_e)(val2);
  if (arg1) (arg1)->serial_parity = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_serial_parity_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum serial_parity_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_parity_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (enum serial_parity_e) ((arg1)->serial_parity);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_serial_handshake_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  enum serial_handshake_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_serial_handshake_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_handshake_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_serial_handshake_set" "', argument " "2"" of type '" "enum serial_handshake_e""'");
  } 
  arg2 = (enum serial_handshake_e)(val2);
  if (arg1) (arg1)->serial_handshake = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_serial_handshake_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum serial_handshake_e result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_handshake_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (enum serial_handshake_e) ((arg1)->serial_handshake);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_write_delay_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_write_delay_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_write_delay_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->write_delay = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_write_delay_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_write_delay_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->write_delay);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_post_write_delay_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_post_write_delay_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_post_write_delay_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_post_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->post_write_delay = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_post_write_delay_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_post_write_delay_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->post_write_delay);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_timeout_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_timeout_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_timeout_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_timeout_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->timeout = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_timeout_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_timeout_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->timeout);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_retry_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_retry_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_retry_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_retry_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->retry = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_retry_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_retry_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->retry);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_cfgparams_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_cfgparams_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_cfgparams_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_cfgparams_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->cfgparams = (struct confparams const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_cfgparams_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct confparams *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_cfgparams_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (struct confparams *) ((arg1)->cfgparams);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_priv_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_priv_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_priv_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_priv_set" "', argument " "2"" of type '" "void const *""'"); 
  }
  if (arg1) (arg1)->priv = (void const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_priv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_priv_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (void *) ((arg1)->priv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_model_macro_name_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_amp_model_macro_name_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_model_macro_name_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_amp_model_macro_name_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->amp_model_macro_name = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->amp_model_macro_name = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_model_macro_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_model_macro_name_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (char *) ((arg1)->amp_model_macro_name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_has_get_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_has_get_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_has_get_level_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_has_get_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_has_get_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_has_get_level_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_level);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_has_set_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_has_set_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_has_set_level_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_has_set_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_has_set_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_has_set_level_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_level);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_level_gran_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_level_gran_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_level_gran_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_level_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->level_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""level_gran""' of type '""gran_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_level_gran_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  gran_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_level_gran_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->level_gran);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_parm_gran_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_parm_gran_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_parm_gran_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_parm_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->parm_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""parm_gran""' of type '""gran_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_parm_gran_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  gran_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_parm_gran_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->parm_gran);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_init_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_amp_init_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_init_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_amp_init_set" "', argument " "2"" of type '" "int (*)(AMP *)""'"); 
    }
  }
  if (arg1) (arg1)->amp_init = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_init_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_init_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *)) ((arg1)->amp_init);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_cleanup_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_amp_cleanup_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_cleanup_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_amp_cleanup_set" "', argument " "2"" of type '" "int (*)(AMP *)""'"); 
    }
  }
  if (arg1) (arg1)->amp_cleanup = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_cleanup_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_cleanup_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *)) ((arg1)->amp_cleanup);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_open_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_amp_open_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_open_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_amp_open_set" "', argument " "2"" of type '" "int (*)(AMP *)""'"); 
    }
  }
  if (arg1) (arg1)->amp_open = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_open_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_open_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *)) ((arg1)->amp_open);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_close_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_amp_close_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_close_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_amp_close_set" "', argument " "2"" of type '" "int (*)(AMP *)""'"); 
    }
  }
  if (arg1) (arg1)->amp_close = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_amp_close_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_close_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *)) ((arg1)->amp_close);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_set_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,freq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_set_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_freq_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_double__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_set_freq_set" "', argument " "2"" of type '" "int (*)(AMP *,freq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_set_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *,freq_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_freq_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,freq_t)) ((arg1)->set_freq);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_double__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_freq_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,freq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_get_freq_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_freq_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_p_double__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_freq_set" "', argument " "2"" of type '" "int (*)(AMP *,freq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_freq = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_freq_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *,freq_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_freq_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,freq_t *)) ((arg1)->get_freq);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_p_double__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_set_conf_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,hamlib_token_t,char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_set_conf_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_conf_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_long_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_set_conf_set" "', argument " "2"" of type '" "int (*)(AMP *,hamlib_token_t,char const *)""'"); 
    }
  }
  if (arg1) (arg1)->set_conf = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_set_conf_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *,hamlib_token_t,char const *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_conf_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,hamlib_token_t,char const *)) ((arg1)->set_conf);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_long_p_q_const__char__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_conf2_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,hamlib_token_t,char *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_get_conf2_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_conf2_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_long_p_char_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_conf2_set" "', argument " "2"" of type '" "int (*)(AMP *,hamlib_token_t,char *,int)""'"); 
    }
  }
  if (arg1) (arg1)->get_conf2 = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_conf2_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *,hamlib_token_t,char *,int) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_conf2_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,hamlib_token_t,char *,int)) ((arg1)->get_conf2);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_long_p_char_int__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_conf_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,hamlib_token_t,char *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_get_conf_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_conf_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_long_p_char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_conf_set" "', argument " "2"" of type '" "int (*)(AMP *,hamlib_token_t,char *)""'"); 
    }
  }
  if (arg1) (arg1)->get_conf = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_conf_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *,hamlib_token_t,char *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_conf_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,hamlib_token_t,char *)) ((arg1)->get_conf);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_long_p_char__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_reset_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,amp_reset_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_reset_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_reset_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_enum_amp_reset_e__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_reset_set" "', argument " "2"" of type '" "int (*)(AMP *,amp_reset_t)""'"); 
    }
  }
  if (arg1) (arg1)->reset = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_reset_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *,amp_reset_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_reset_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,amp_reset_t)) ((arg1)->reset);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_enum_amp_reset_e__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,setting_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_get_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_level_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_unsigned_long_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_level_set" "', argument " "2"" of type '" "int (*)(AMP *,setting_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *,setting_t,value_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_level_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,setting_t,value_t *)) ((arg1)->get_level);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_unsigned_long_long_p_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_set_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,setting_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_set_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_level_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_unsigned_long_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_set_level_set" "', argument " "2"" of type '" "int (*)(AMP *,setting_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_set_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *,setting_t,value_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_level_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,setting_t,value_t)) ((arg1)->set_level);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_unsigned_long_long_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_ext_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,hamlib_token_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_get_ext_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_ext_level_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_ext_level_set" "', argument " "2"" of type '" "int (*)(AMP *,hamlib_token_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_ext_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *,hamlib_token_t,value_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_ext_level_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,hamlib_token_t,value_t *)) ((arg1)->get_ext_level);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_long_p_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_set_ext_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,hamlib_token_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_set_ext_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_ext_level_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_set_ext_level_set" "', argument " "2"" of type '" "int (*)(AMP *,hamlib_token_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_set_ext_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *,hamlib_token_t,value_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_ext_level_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,hamlib_token_t,value_t)) ((arg1)->set_ext_level);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_long_value_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_set_powerstat_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,powerstat_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_set_powerstat_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_powerstat_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_enum_powerstat_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_set_powerstat_set" "', argument " "2"" of type '" "int (*)(AMP *,powerstat_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_powerstat = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_set_powerstat_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *,powerstat_t) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_powerstat_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,powerstat_t)) ((arg1)->set_powerstat);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_enum_powerstat_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_powerstat_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,powerstat_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_get_powerstat_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_powerstat_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_p_enum_powerstat_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_powerstat_set" "', argument " "2"" of type '" "int (*)(AMP *,powerstat_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_powerstat = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_powerstat_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int (*result)(AMP *,powerstat_t *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_powerstat_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,powerstat_t *)) ((arg1)->get_powerstat);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_p_enum_powerstat_t__int);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_info_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  char *(*arg2)(AMP *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_get_info_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_info_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp__p_char);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_info_set" "', argument " "2"" of type '" "char const *(*)(AMP *)""'"); 
    }
  }
  if (arg1) (arg1)->get_info = (char const *(*)(AMP *))arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_get_info_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *(*result)(AMP *) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_info_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (char *(*)(AMP *)) ((arg1)->get_info);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp__p_char);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_levels_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_levels_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_levels_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_levels_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->levels = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_levels_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_levels_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (setting_t) ((arg1)->levels);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_ext_levels_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_ext_levels_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_ext_levels_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_ext_levels_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->ext_levels = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_ext_levels_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_ext_levels_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (unsigned int) ((arg1)->ext_levels);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_extlevels_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_extlevels_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_extlevels_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_extlevels_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extlevels = (struct confparams const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_extlevels_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct confparams *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_extlevels_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extlevels);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_extparms_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_caps_extparms_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_extparms_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_extparms_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extparms = (struct confparams const *)arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_caps_extparms_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct confparams *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_extparms_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extparms);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_confparams, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_amp_caps(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_amp_caps", 0, 0, 0)) SWIG_fail;
  result = (struct amp_caps *)calloc(1, sizeof(struct amp_caps));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_amp_caps, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_amp_caps(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_caps, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_amp_caps" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *amp_caps_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_amp_caps, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *amp_caps_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_amp_state_comm_state_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_state_comm_state_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_comm_state_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_state_comm_state_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->comm_state = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_comm_state_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_comm_state_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (int) ((arg1)->comm_state);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_priv_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_state_priv_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_priv_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_state_priv_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->priv = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_priv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_priv_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (void *) ((arg1)->priv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_obj_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_state_obj_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_obj_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_state_obj_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->obj = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_obj_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_obj_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (void *) ((arg1)->obj);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_has_get_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_state_has_get_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_has_get_level_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_state_has_get_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_has_get_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_has_get_level_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_level);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_has_set_level_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_state_has_set_level_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_has_set_level_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_state_has_set_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_level = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_has_set_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  setting_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_has_set_level_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_level);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_level_gran_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_state_level_gran_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_level_gran_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_state_level_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->level_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""level_gran""' of type '""gran_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_level_gran_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  gran_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_level_gran_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->level_gran);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_parm_gran_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_state_parm_gran_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_parm_gran_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_state_parm_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->parm_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""parm_gran""' of type '""gran_t [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_parm_gran_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  gran_t *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_parm_gran_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->parm_gran);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_gran, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_ampport_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  hamlib_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "amp_state_ampport_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_ampport_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_hamlib_port,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_state_ampport_set" "', argument " "2"" of type '" "hamlib_port_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "amp_state_ampport_set" "', argument " "2"" of type '" "hamlib_port_t""'");
    } else {
      arg2 = *((hamlib_port_t *)(argp2));
    }
  }
  if (arg1) (arg1)->ampport = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_amp_state_ampport_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  hamlib_port_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_ampport_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result =  ((arg1)->ampport);
  resultobj = SWIG_NewPointerObj((hamlib_port_t *)memcpy((hamlib_port_t *)calloc(1,sizeof(hamlib_port_t)),&result,sizeof(hamlib_port_t)), SWIGTYPE_p_hamlib_port, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_amp_state(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_amp_state", 0, 0, 0)) SWIG_fail;
  result = (struct amp_state *)calloc(1, sizeof(struct amp_state));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_amp_state, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_amp_state(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_amp_state, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_amp_state" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *amp_state_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_amp_state, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *amp_state_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Amp_caps_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  struct amp_caps *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_caps_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_caps_set" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_amp_caps, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Amp_caps_set" "', argument " "2"" of type '" "struct amp_caps *""'"); 
  }
  arg2 = (struct amp_caps *)(argp2);
  if (arg1) (arg1)->caps = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_caps_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct amp_caps *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_caps_get" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  result = (struct amp_caps *) ((arg1)->caps);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_amp_caps, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_state_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  struct amp_state *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_state_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_state_set" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_amp_state, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Amp_state_set" "', argument " "2"" of type '" "struct amp_state *""'"); 
  }
  arg2 = (struct amp_state *)(argp2);
  if (arg1) (arg1)->state = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_state_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  struct amp_state *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_state_get" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  result = (struct amp_state *) ((arg1)->state);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_amp_state, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_error_status_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_error_status_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_error_status_set" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_error_status_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->error_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_error_status_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_error_status_get" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  result = (int) ((arg1)->error_status);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_do_exception_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_do_exception_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_do_exception_set" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_do_exception_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->do_exception = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_do_exception_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_do_exception_get" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  result = (int) ((arg1)->do_exception);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Amp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  amp_model_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  struct Amp *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Amp" "', argument " "1"" of type '" "amp_model_t""'");
  } 
  arg1 = (amp_model_t)(val1);
  result = (struct Amp *)new_Amp(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Amp, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Amp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Amp" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  delete_Amp(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_open(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_open" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Amp_open(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_close(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_close" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Amp_close(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  amp_reset_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_reset", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_reset" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_reset" "', argument " "2"" of type '" "amp_reset_t""'");
  } 
  arg2 = (amp_reset_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Amp_reset(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_token_lookup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  const_char_string arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_token_lookup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_token_lookup" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Amp_token_lookup" "', argument " "2"" of type '" "const_char_string""'");
  }
  arg2 = (const_char_string)(buf2);
  {
    arg1->error_status = RIG_OK;
    Amp_token_lookup(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_has_get_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  setting_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_has_get_level", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_has_get_level" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_has_get_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Amp_has_get_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_has_set_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  setting_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_has_set_level", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_has_set_level" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_has_set_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Amp_has_set_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_get_ext_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_get_ext_level", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_get_ext_level" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_get_ext_level" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (PyObject *)Amp_get_ext_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = result;
  if (!resultobj) SWIG_fail;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_set_freq(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_set_freq", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_set_freq" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_set_freq" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Amp_set_freq(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_set_powerstat(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  powerstat_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_set_powerstat", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_set_powerstat" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_set_powerstat" "', argument " "2"" of type '" "powerstat_t""'");
  } 
  arg2 = (powerstat_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Amp_set_powerstat(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_set_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  setting_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_set_level", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_set_level" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_set_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Amp_set_level" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Amp_set_level" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Amp_set_level(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_set_ext_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  hamlib_token_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_set_ext_level", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_set_ext_level" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_set_ext_level" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Amp_set_ext_level" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Amp_set_ext_level" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Amp_set_ext_level(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_set_conf__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_set_conf" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Amp_set_conf" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Amp_set_conf" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  {
    arg1->error_status = RIG_OK;
    Amp_set_conf__SWIG_0(arg1,(char const *)arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_set_conf__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  hamlib_token_t arg2 ;
  const_char_string arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_set_conf" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_set_conf" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Amp_set_conf" "', argument " "3"" of type '" "const_char_string""'");
  }
  arg3 = (const_char_string)(buf3);
  {
    arg1->error_status = RIG_OK;
    Amp_set_conf__SWIG_1(arg1,arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_set_conf(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Amp_set_conf", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Amp, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Amp_set_conf__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Amp, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Amp_set_conf__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Amp_set_conf'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Amp::set_conf(char const *,char const *)\n"
    "    Amp::set_conf(hamlib_token_t,const_char_string)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Amp_get_conf__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  char *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_get_conf" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_get_conf" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Amp_get_conf__SWIG_0(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_get_conf__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_get_conf" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Amp_get_conf" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Amp_get_conf__SWIG_1(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_get_conf(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Amp_get_conf", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Amp, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Amp_get_conf__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Amp, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Amp_get_conf__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Amp_get_conf'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Amp::get_conf(hamlib_token_t)\n"
    "    Amp::get_conf(char const *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Amp_get_freq(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  freq_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_get_freq" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (freq_t)Amp_get_freq(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_get_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_get_info" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Amp_get_info(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_get_powerstat(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  powerstat_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_get_powerstat" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (powerstat_t)Amp_get_powerstat(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Amp_get_level(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct Amp *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Amp_get_level", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_get_level" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_get_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (PyObject *)Amp_get_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  resultobj = result;
  if (!resultobj) SWIG_fail;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Amp_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Amp, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Amp_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

static PyMethodDef SwigMethods[] = {
	 { "confparams_token_set", _wrap_confparams_token_set, METH_VARARGS, NULL},
	 { "confparams_token_get", _wrap_confparams_token_get, METH_O, NULL},
	 { "confparams_name_get", _wrap_confparams_name_get, METH_O, NULL},
	 { "confparams_label_get", _wrap_confparams_label_get, METH_O, NULL},
	 { "confparams_tooltip_get", _wrap_confparams_tooltip_get, METH_O, NULL},
	 { "confparams_dflt_get", _wrap_confparams_dflt_get, METH_O, NULL},
	 { "confparams_type_set", _wrap_confparams_type_set, METH_VARARGS, NULL},
	 { "confparams_type_get", _wrap_confparams_type_get, METH_O, NULL},
	 { "confparams_u_get", _wrap_confparams_u_get, METH_O, NULL},
	 { "new_confparams", _wrap_new_confparams, METH_NOARGS, NULL},
	 { "delete_confparams", _wrap_delete_confparams, METH_O, NULL},
	 { "confparams_swigregister", confparams_swigregister, METH_O, NULL},
	 { "confparams_swiginit", confparams_swiginit, METH_VARARGS, NULL},
	 { "confparams_u_n_get", _wrap_confparams_u_n_get, METH_O, NULL},
	 { "confparams_u_c_get", _wrap_confparams_u_c_get, METH_O, NULL},
	 { "new_confparams_u", _wrap_new_confparams_u, METH_NOARGS, NULL},
	 { "delete_confparams_u", _wrap_delete_confparams_u, METH_O, NULL},
	 { "confparams_u_swigregister", confparams_u_swigregister, METH_O, NULL},
	 { "confparams_u_swiginit", confparams_u_swiginit, METH_VARARGS, NULL},
	 { "confparams_u_c_combostr_set", _wrap_confparams_u_c_combostr_set, METH_VARARGS, NULL},
	 { "confparams_u_c_combostr_get", _wrap_confparams_u_c_combostr_get, METH_O, NULL},
	 { "new_confparams_u_c", _wrap_new_confparams_u_c, METH_NOARGS, NULL},
	 { "delete_confparams_u_c", _wrap_delete_confparams_u_c, METH_O, NULL},
	 { "confparams_u_c_swigregister", confparams_u_c_swigregister, METH_O, NULL},
	 { "confparams_u_c_swiginit", confparams_u_c_swiginit, METH_VARARGS, NULL},
	 { "confparams_u_n_min_set", _wrap_confparams_u_n_min_set, METH_VARARGS, NULL},
	 { "confparams_u_n_min_get", _wrap_confparams_u_n_min_get, METH_O, NULL},
	 { "confparams_u_n_max_set", _wrap_confparams_u_n_max_set, METH_VARARGS, NULL},
	 { "confparams_u_n_max_get", _wrap_confparams_u_n_max_get, METH_O, NULL},
	 { "confparams_u_n_step_set", _wrap_confparams_u_n_step_set, METH_VARARGS, NULL},
	 { "confparams_u_n_step_get", _wrap_confparams_u_n_step_get, METH_O, NULL},
	 { "new_confparams_u_n", _wrap_new_confparams_u_n, METH_NOARGS, NULL},
	 { "delete_confparams_u_n", _wrap_delete_confparams_u_n, METH_O, NULL},
	 { "confparams_u_n_swigregister", confparams_u_n_swigregister, METH_O, NULL},
	 { "confparams_u_n_swiginit", confparams_u_n_swiginit, METH_VARARGS, NULL},
	 { "value_t_i_set", _wrap_value_t_i_set, METH_VARARGS, NULL},
	 { "value_t_i_get", _wrap_value_t_i_get, METH_O, NULL},
	 { "value_t_u_set", _wrap_value_t_u_set, METH_VARARGS, NULL},
	 { "value_t_u_get", _wrap_value_t_u_get, METH_O, NULL},
	 { "value_t_f_set", _wrap_value_t_f_set, METH_VARARGS, NULL},
	 { "value_t_f_get", _wrap_value_t_f_get, METH_O, NULL},
	 { "value_t_s_set", _wrap_value_t_s_set, METH_VARARGS, NULL},
	 { "value_t_s_get", _wrap_value_t_s_get, METH_O, NULL},
	 { "value_t_cs_get", _wrap_value_t_cs_get, METH_O, NULL},
	 { "value_t_b_get", _wrap_value_t_b_get, METH_O, NULL},
	 { "new_value_t", _wrap_new_value_t, METH_NOARGS, NULL},
	 { "delete_value_t", _wrap_delete_value_t, METH_O, NULL},
	 { "value_t_swigregister", value_t_swigregister, METH_O, NULL},
	 { "value_t_swiginit", value_t_swiginit, METH_VARARGS, NULL},
	 { "value_t_b_l_set", _wrap_value_t_b_l_set, METH_VARARGS, NULL},
	 { "value_t_b_l_get", _wrap_value_t_b_l_get, METH_O, NULL},
	 { "value_t_b_d_set", _wrap_value_t_b_d_set, METH_VARARGS, NULL},
	 { "value_t_b_d_get", _wrap_value_t_b_d_get, METH_O, NULL},
	 { "new_value_t_b", _wrap_new_value_t_b, METH_NOARGS, NULL},
	 { "delete_value_t_b", _wrap_delete_value_t_b, METH_O, NULL},
	 { "value_t_b_swigregister", value_t_b_swigregister, METH_O, NULL},
	 { "value_t_b_swiginit", value_t_b_swiginit, METH_VARARGS, NULL},
	 { "freq_range_t_startf_set", _wrap_freq_range_t_startf_set, METH_VARARGS, NULL},
	 { "freq_range_t_startf_get", _wrap_freq_range_t_startf_get, METH_O, NULL},
	 { "freq_range_t_endf_set", _wrap_freq_range_t_endf_set, METH_VARARGS, NULL},
	 { "freq_range_t_endf_get", _wrap_freq_range_t_endf_get, METH_O, NULL},
	 { "freq_range_t_modes_set", _wrap_freq_range_t_modes_set, METH_VARARGS, NULL},
	 { "freq_range_t_modes_get", _wrap_freq_range_t_modes_get, METH_O, NULL},
	 { "freq_range_t_low_power_set", _wrap_freq_range_t_low_power_set, METH_VARARGS, NULL},
	 { "freq_range_t_low_power_get", _wrap_freq_range_t_low_power_get, METH_O, NULL},
	 { "freq_range_t_high_power_set", _wrap_freq_range_t_high_power_set, METH_VARARGS, NULL},
	 { "freq_range_t_high_power_get", _wrap_freq_range_t_high_power_get, METH_O, NULL},
	 { "freq_range_t_vfo_set", _wrap_freq_range_t_vfo_set, METH_VARARGS, NULL},
	 { "freq_range_t_vfo_get", _wrap_freq_range_t_vfo_get, METH_O, NULL},
	 { "freq_range_t_ant_set", _wrap_freq_range_t_ant_set, METH_VARARGS, NULL},
	 { "freq_range_t_ant_get", _wrap_freq_range_t_ant_get, METH_O, NULL},
	 { "freq_range_t_label_set", _wrap_freq_range_t_label_set, METH_VARARGS, NULL},
	 { "freq_range_t_label_get", _wrap_freq_range_t_label_get, METH_O, NULL},
	 { "new_freq_range_t", _wrap_new_freq_range_t, METH_NOARGS, NULL},
	 { "delete_freq_range_t", _wrap_delete_freq_range_t, METH_O, NULL},
	 { "freq_range_t_swigregister", freq_range_t_swigregister, METH_O, NULL},
	 { "freq_range_t_swiginit", freq_range_t_swiginit, METH_VARARGS, NULL},
	 { "tuning_step_list_modes_set", _wrap_tuning_step_list_modes_set, METH_VARARGS, NULL},
	 { "tuning_step_list_modes_get", _wrap_tuning_step_list_modes_get, METH_O, NULL},
	 { "tuning_step_list_ts_set", _wrap_tuning_step_list_ts_set, METH_VARARGS, NULL},
	 { "tuning_step_list_ts_get", _wrap_tuning_step_list_ts_get, METH_O, NULL},
	 { "new_tuning_step_list", _wrap_new_tuning_step_list, METH_NOARGS, NULL},
	 { "delete_tuning_step_list", _wrap_delete_tuning_step_list, METH_O, NULL},
	 { "tuning_step_list_swigregister", tuning_step_list_swigregister, METH_O, NULL},
	 { "tuning_step_list_swiginit", tuning_step_list_swiginit, METH_VARARGS, NULL},
	 { "filter_list_modes_set", _wrap_filter_list_modes_set, METH_VARARGS, NULL},
	 { "filter_list_modes_get", _wrap_filter_list_modes_get, METH_O, NULL},
	 { "filter_list_width_set", _wrap_filter_list_width_set, METH_VARARGS, NULL},
	 { "filter_list_width_get", _wrap_filter_list_width_get, METH_O, NULL},
	 { "new_filter_list", _wrap_new_filter_list, METH_NOARGS, NULL},
	 { "delete_filter_list", _wrap_delete_filter_list, METH_O, NULL},
	 { "filter_list_swigregister", filter_list_swigregister, METH_O, NULL},
	 { "filter_list_swiginit", filter_list_swiginit, METH_VARARGS, NULL},
	 { "ext_list_token_set", _wrap_ext_list_token_set, METH_VARARGS, NULL},
	 { "ext_list_token_get", _wrap_ext_list_token_get, METH_O, NULL},
	 { "ext_list_val_set", _wrap_ext_list_val_set, METH_VARARGS, NULL},
	 { "ext_list_val_get", _wrap_ext_list_val_get, METH_O, NULL},
	 { "new_ext_list", _wrap_new_ext_list, METH_NOARGS, NULL},
	 { "delete_ext_list", _wrap_delete_ext_list, METH_O, NULL},
	 { "ext_list_swigregister", ext_list_swigregister, METH_O, NULL},
	 { "ext_list_swiginit", ext_list_swiginit, METH_VARARGS, NULL},
	 { "channel_channel_num_set", _wrap_channel_channel_num_set, METH_VARARGS, NULL},
	 { "channel_channel_num_get", _wrap_channel_channel_num_get, METH_O, NULL},
	 { "channel_bank_num_set", _wrap_channel_bank_num_set, METH_VARARGS, NULL},
	 { "channel_bank_num_get", _wrap_channel_bank_num_get, METH_O, NULL},
	 { "channel_vfo_set", _wrap_channel_vfo_set, METH_VARARGS, NULL},
	 { "channel_vfo_get", _wrap_channel_vfo_get, METH_O, NULL},
	 { "channel_ant_set", _wrap_channel_ant_set, METH_VARARGS, NULL},
	 { "channel_ant_get", _wrap_channel_ant_get, METH_O, NULL},
	 { "channel_freq_set", _wrap_channel_freq_set, METH_VARARGS, NULL},
	 { "channel_freq_get", _wrap_channel_freq_get, METH_O, NULL},
	 { "channel_mode_set", _wrap_channel_mode_set, METH_VARARGS, NULL},
	 { "channel_mode_get", _wrap_channel_mode_get, METH_O, NULL},
	 { "channel_width_set", _wrap_channel_width_set, METH_VARARGS, NULL},
	 { "channel_width_get", _wrap_channel_width_get, METH_O, NULL},
	 { "channel_tx_freq_set", _wrap_channel_tx_freq_set, METH_VARARGS, NULL},
	 { "channel_tx_freq_get", _wrap_channel_tx_freq_get, METH_O, NULL},
	 { "channel_tx_mode_set", _wrap_channel_tx_mode_set, METH_VARARGS, NULL},
	 { "channel_tx_mode_get", _wrap_channel_tx_mode_get, METH_O, NULL},
	 { "channel_tx_width_set", _wrap_channel_tx_width_set, METH_VARARGS, NULL},
	 { "channel_tx_width_get", _wrap_channel_tx_width_get, METH_O, NULL},
	 { "channel_split_set", _wrap_channel_split_set, METH_VARARGS, NULL},
	 { "channel_split_get", _wrap_channel_split_get, METH_O, NULL},
	 { "channel_tx_vfo_set", _wrap_channel_tx_vfo_set, METH_VARARGS, NULL},
	 { "channel_tx_vfo_get", _wrap_channel_tx_vfo_get, METH_O, NULL},
	 { "channel_rptr_shift_set", _wrap_channel_rptr_shift_set, METH_VARARGS, NULL},
	 { "channel_rptr_shift_get", _wrap_channel_rptr_shift_get, METH_O, NULL},
	 { "channel_rptr_offs_set", _wrap_channel_rptr_offs_set, METH_VARARGS, NULL},
	 { "channel_rptr_offs_get", _wrap_channel_rptr_offs_get, METH_O, NULL},
	 { "channel_tuning_step_set", _wrap_channel_tuning_step_set, METH_VARARGS, NULL},
	 { "channel_tuning_step_get", _wrap_channel_tuning_step_get, METH_O, NULL},
	 { "channel_rit_set", _wrap_channel_rit_set, METH_VARARGS, NULL},
	 { "channel_rit_get", _wrap_channel_rit_get, METH_O, NULL},
	 { "channel_xit_set", _wrap_channel_xit_set, METH_VARARGS, NULL},
	 { "channel_xit_get", _wrap_channel_xit_get, METH_O, NULL},
	 { "channel_funcs_set", _wrap_channel_funcs_set, METH_VARARGS, NULL},
	 { "channel_funcs_get", _wrap_channel_funcs_get, METH_O, NULL},
	 { "channel_levels_set", _wrap_channel_levels_set, METH_VARARGS, NULL},
	 { "channel_levels_get", _wrap_channel_levels_get, METH_O, NULL},
	 { "channel_ctcss_tone_set", _wrap_channel_ctcss_tone_set, METH_VARARGS, NULL},
	 { "channel_ctcss_tone_get", _wrap_channel_ctcss_tone_get, METH_O, NULL},
	 { "channel_ctcss_sql_set", _wrap_channel_ctcss_sql_set, METH_VARARGS, NULL},
	 { "channel_ctcss_sql_get", _wrap_channel_ctcss_sql_get, METH_O, NULL},
	 { "channel_dcs_code_set", _wrap_channel_dcs_code_set, METH_VARARGS, NULL},
	 { "channel_dcs_code_get", _wrap_channel_dcs_code_get, METH_O, NULL},
	 { "channel_dcs_sql_set", _wrap_channel_dcs_sql_set, METH_VARARGS, NULL},
	 { "channel_dcs_sql_get", _wrap_channel_dcs_sql_get, METH_O, NULL},
	 { "channel_scan_group_set", _wrap_channel_scan_group_set, METH_VARARGS, NULL},
	 { "channel_scan_group_get", _wrap_channel_scan_group_get, METH_O, NULL},
	 { "channel_flags_set", _wrap_channel_flags_set, METH_VARARGS, NULL},
	 { "channel_flags_get", _wrap_channel_flags_get, METH_O, NULL},
	 { "channel_channel_desc_set", _wrap_channel_channel_desc_set, METH_VARARGS, NULL},
	 { "channel_channel_desc_get", _wrap_channel_channel_desc_get, METH_O, NULL},
	 { "channel_ext_levels_set", _wrap_channel_ext_levels_set, METH_VARARGS, NULL},
	 { "channel_ext_levels_get", _wrap_channel_ext_levels_get, METH_O, NULL},
	 { "channel_tag_set", _wrap_channel_tag_set, METH_VARARGS, NULL},
	 { "channel_tag_get", _wrap_channel_tag_get, METH_O, NULL},
	 { "new_channel", _wrap_new_channel, METH_VARARGS, NULL},
	 { "delete_channel", _wrap_delete_channel, METH_O, NULL},
	 { "channel_swigregister", channel_swigregister, METH_O, NULL},
	 { "channel_swiginit", channel_swiginit, METH_VARARGS, NULL},
	 { "channel_cap_bank_num_set", _wrap_channel_cap_bank_num_set, METH_VARARGS, NULL},
	 { "channel_cap_bank_num_get", _wrap_channel_cap_bank_num_get, METH_O, NULL},
	 { "channel_cap_vfo_set", _wrap_channel_cap_vfo_set, METH_VARARGS, NULL},
	 { "channel_cap_vfo_get", _wrap_channel_cap_vfo_get, METH_O, NULL},
	 { "channel_cap_ant_set", _wrap_channel_cap_ant_set, METH_VARARGS, NULL},
	 { "channel_cap_ant_get", _wrap_channel_cap_ant_get, METH_O, NULL},
	 { "channel_cap_freq_set", _wrap_channel_cap_freq_set, METH_VARARGS, NULL},
	 { "channel_cap_freq_get", _wrap_channel_cap_freq_get, METH_O, NULL},
	 { "channel_cap_mode_set", _wrap_channel_cap_mode_set, METH_VARARGS, NULL},
	 { "channel_cap_mode_get", _wrap_channel_cap_mode_get, METH_O, NULL},
	 { "channel_cap_width_set", _wrap_channel_cap_width_set, METH_VARARGS, NULL},
	 { "channel_cap_width_get", _wrap_channel_cap_width_get, METH_O, NULL},
	 { "channel_cap_tx_freq_set", _wrap_channel_cap_tx_freq_set, METH_VARARGS, NULL},
	 { "channel_cap_tx_freq_get", _wrap_channel_cap_tx_freq_get, METH_O, NULL},
	 { "channel_cap_tx_mode_set", _wrap_channel_cap_tx_mode_set, METH_VARARGS, NULL},
	 { "channel_cap_tx_mode_get", _wrap_channel_cap_tx_mode_get, METH_O, NULL},
	 { "channel_cap_tx_width_set", _wrap_channel_cap_tx_width_set, METH_VARARGS, NULL},
	 { "channel_cap_tx_width_get", _wrap_channel_cap_tx_width_get, METH_O, NULL},
	 { "channel_cap_split_set", _wrap_channel_cap_split_set, METH_VARARGS, NULL},
	 { "channel_cap_split_get", _wrap_channel_cap_split_get, METH_O, NULL},
	 { "channel_cap_tx_vfo_set", _wrap_channel_cap_tx_vfo_set, METH_VARARGS, NULL},
	 { "channel_cap_tx_vfo_get", _wrap_channel_cap_tx_vfo_get, METH_O, NULL},
	 { "channel_cap_rptr_shift_set", _wrap_channel_cap_rptr_shift_set, METH_VARARGS, NULL},
	 { "channel_cap_rptr_shift_get", _wrap_channel_cap_rptr_shift_get, METH_O, NULL},
	 { "channel_cap_rptr_offs_set", _wrap_channel_cap_rptr_offs_set, METH_VARARGS, NULL},
	 { "channel_cap_rptr_offs_get", _wrap_channel_cap_rptr_offs_get, METH_O, NULL},
	 { "channel_cap_tuning_step_set", _wrap_channel_cap_tuning_step_set, METH_VARARGS, NULL},
	 { "channel_cap_tuning_step_get", _wrap_channel_cap_tuning_step_get, METH_O, NULL},
	 { "channel_cap_rit_set", _wrap_channel_cap_rit_set, METH_VARARGS, NULL},
	 { "channel_cap_rit_get", _wrap_channel_cap_rit_get, METH_O, NULL},
	 { "channel_cap_xit_set", _wrap_channel_cap_xit_set, METH_VARARGS, NULL},
	 { "channel_cap_xit_get", _wrap_channel_cap_xit_get, METH_O, NULL},
	 { "channel_cap_funcs_set", _wrap_channel_cap_funcs_set, METH_VARARGS, NULL},
	 { "channel_cap_funcs_get", _wrap_channel_cap_funcs_get, METH_O, NULL},
	 { "channel_cap_levels_set", _wrap_channel_cap_levels_set, METH_VARARGS, NULL},
	 { "channel_cap_levels_get", _wrap_channel_cap_levels_get, METH_O, NULL},
	 { "channel_cap_ctcss_tone_set", _wrap_channel_cap_ctcss_tone_set, METH_VARARGS, NULL},
	 { "channel_cap_ctcss_tone_get", _wrap_channel_cap_ctcss_tone_get, METH_O, NULL},
	 { "channel_cap_ctcss_sql_set", _wrap_channel_cap_ctcss_sql_set, METH_VARARGS, NULL},
	 { "channel_cap_ctcss_sql_get", _wrap_channel_cap_ctcss_sql_get, METH_O, NULL},
	 { "channel_cap_dcs_code_set", _wrap_channel_cap_dcs_code_set, METH_VARARGS, NULL},
	 { "channel_cap_dcs_code_get", _wrap_channel_cap_dcs_code_get, METH_O, NULL},
	 { "channel_cap_dcs_sql_set", _wrap_channel_cap_dcs_sql_set, METH_VARARGS, NULL},
	 { "channel_cap_dcs_sql_get", _wrap_channel_cap_dcs_sql_get, METH_O, NULL},
	 { "channel_cap_scan_group_set", _wrap_channel_cap_scan_group_set, METH_VARARGS, NULL},
	 { "channel_cap_scan_group_get", _wrap_channel_cap_scan_group_get, METH_O, NULL},
	 { "channel_cap_flags_set", _wrap_channel_cap_flags_set, METH_VARARGS, NULL},
	 { "channel_cap_flags_get", _wrap_channel_cap_flags_get, METH_O, NULL},
	 { "channel_cap_channel_desc_set", _wrap_channel_cap_channel_desc_set, METH_VARARGS, NULL},
	 { "channel_cap_channel_desc_get", _wrap_channel_cap_channel_desc_get, METH_O, NULL},
	 { "channel_cap_ext_levels_set", _wrap_channel_cap_ext_levels_set, METH_VARARGS, NULL},
	 { "channel_cap_ext_levels_get", _wrap_channel_cap_ext_levels_get, METH_O, NULL},
	 { "channel_cap_tag_set", _wrap_channel_cap_tag_set, METH_VARARGS, NULL},
	 { "channel_cap_tag_get", _wrap_channel_cap_tag_get, METH_O, NULL},
	 { "new_channel_cap", _wrap_new_channel_cap, METH_NOARGS, NULL},
	 { "delete_channel_cap", _wrap_delete_channel_cap, METH_O, NULL},
	 { "channel_cap_swigregister", channel_cap_swigregister, METH_O, NULL},
	 { "channel_cap_swiginit", channel_cap_swiginit, METH_VARARGS, NULL},
	 { "chan_list_startc_set", _wrap_chan_list_startc_set, METH_VARARGS, NULL},
	 { "chan_list_startc_get", _wrap_chan_list_startc_get, METH_O, NULL},
	 { "chan_list_endc_set", _wrap_chan_list_endc_set, METH_VARARGS, NULL},
	 { "chan_list_endc_get", _wrap_chan_list_endc_get, METH_O, NULL},
	 { "chan_list_type_set", _wrap_chan_list_type_set, METH_VARARGS, NULL},
	 { "chan_list_type_get", _wrap_chan_list_type_get, METH_O, NULL},
	 { "chan_list_mem_caps_set", _wrap_chan_list_mem_caps_set, METH_VARARGS, NULL},
	 { "chan_list_mem_caps_get", _wrap_chan_list_mem_caps_get, METH_O, NULL},
	 { "new_chan_list", _wrap_new_chan_list, METH_NOARGS, NULL},
	 { "delete_chan_list", _wrap_delete_chan_list, METH_O, NULL},
	 { "chan_list_swigregister", chan_list_swigregister, METH_O, NULL},
	 { "chan_list_swiginit", chan_list_swiginit, METH_VARARGS, NULL},
	 { "gran_min_set", _wrap_gran_min_set, METH_VARARGS, NULL},
	 { "gran_min_get", _wrap_gran_min_get, METH_O, NULL},
	 { "gran_max_set", _wrap_gran_max_set, METH_VARARGS, NULL},
	 { "gran_max_get", _wrap_gran_max_get, METH_O, NULL},
	 { "gran_step_set", _wrap_gran_step_set, METH_VARARGS, NULL},
	 { "gran_step_get", _wrap_gran_step_get, METH_O, NULL},
	 { "new_gran", _wrap_new_gran, METH_NOARGS, NULL},
	 { "delete_gran", _wrap_delete_gran, METH_O, NULL},
	 { "gran_swigregister", gran_swigregister, METH_O, NULL},
	 { "gran_swiginit", gran_swiginit, METH_VARARGS, NULL},
	 { "cal_table_size_set", _wrap_cal_table_size_set, METH_VARARGS, NULL},
	 { "cal_table_size_get", _wrap_cal_table_size_get, METH_O, NULL},
	 { "cal_table_table_get", _wrap_cal_table_table_get, METH_O, NULL},
	 { "new_cal_table", _wrap_new_cal_table, METH_NOARGS, NULL},
	 { "delete_cal_table", _wrap_delete_cal_table, METH_O, NULL},
	 { "cal_table_swigregister", cal_table_swigregister, METH_O, NULL},
	 { "cal_table_swiginit", cal_table_swiginit, METH_VARARGS, NULL},
	 { "cal_table_table_raw_set", _wrap_cal_table_table_raw_set, METH_VARARGS, NULL},
	 { "cal_table_table_raw_get", _wrap_cal_table_table_raw_get, METH_O, NULL},
	 { "cal_table_table_val_set", _wrap_cal_table_table_val_set, METH_VARARGS, NULL},
	 { "cal_table_table_val_get", _wrap_cal_table_table_val_get, METH_O, NULL},
	 { "new_cal_table_table", _wrap_new_cal_table_table, METH_NOARGS, NULL},
	 { "delete_cal_table_table", _wrap_delete_cal_table_table, METH_O, NULL},
	 { "cal_table_table_swigregister", cal_table_table_swigregister, METH_O, NULL},
	 { "cal_table_table_swiginit", cal_table_table_swiginit, METH_VARARGS, NULL},
	 { "cal_table_float_size_set", _wrap_cal_table_float_size_set, METH_VARARGS, NULL},
	 { "cal_table_float_size_get", _wrap_cal_table_float_size_get, METH_O, NULL},
	 { "cal_table_float_table_get", _wrap_cal_table_float_table_get, METH_O, NULL},
	 { "new_cal_table_float", _wrap_new_cal_table_float, METH_NOARGS, NULL},
	 { "delete_cal_table_float", _wrap_delete_cal_table_float, METH_O, NULL},
	 { "cal_table_float_swigregister", cal_table_float_swigregister, METH_O, NULL},
	 { "cal_table_float_swiginit", cal_table_float_swiginit, METH_VARARGS, NULL},
	 { "cal_table_float_table_raw_set", _wrap_cal_table_float_table_raw_set, METH_VARARGS, NULL},
	 { "cal_table_float_table_raw_get", _wrap_cal_table_float_table_raw_get, METH_O, NULL},
	 { "cal_table_float_table_val_set", _wrap_cal_table_float_table_val_set, METH_VARARGS, NULL},
	 { "cal_table_float_table_val_get", _wrap_cal_table_float_table_val_get, METH_O, NULL},
	 { "new_cal_table_float_table", _wrap_new_cal_table_float_table, METH_NOARGS, NULL},
	 { "delete_cal_table_float_table", _wrap_delete_cal_table_float_table, METH_O, NULL},
	 { "cal_table_float_table_swigregister", cal_table_float_table_swigregister, METH_O, NULL},
	 { "cal_table_float_table_swiginit", cal_table_float_table_swiginit, METH_VARARGS, NULL},
	 { "rig_spectrum_scope_id_set", _wrap_rig_spectrum_scope_id_set, METH_VARARGS, NULL},
	 { "rig_spectrum_scope_id_get", _wrap_rig_spectrum_scope_id_get, METH_O, NULL},
	 { "rig_spectrum_scope_name_set", _wrap_rig_spectrum_scope_name_set, METH_VARARGS, NULL},
	 { "rig_spectrum_scope_name_get", _wrap_rig_spectrum_scope_name_get, METH_O, NULL},
	 { "new_rig_spectrum_scope", _wrap_new_rig_spectrum_scope, METH_NOARGS, NULL},
	 { "delete_rig_spectrum_scope", _wrap_delete_rig_spectrum_scope, METH_O, NULL},
	 { "rig_spectrum_scope_swigregister", rig_spectrum_scope_swigregister, METH_O, NULL},
	 { "rig_spectrum_scope_swiginit", rig_spectrum_scope_swiginit, METH_VARARGS, NULL},
	 { "rig_spectrum_avg_mode_id_set", _wrap_rig_spectrum_avg_mode_id_set, METH_VARARGS, NULL},
	 { "rig_spectrum_avg_mode_id_get", _wrap_rig_spectrum_avg_mode_id_get, METH_O, NULL},
	 { "rig_spectrum_avg_mode_name_set", _wrap_rig_spectrum_avg_mode_name_set, METH_VARARGS, NULL},
	 { "rig_spectrum_avg_mode_name_get", _wrap_rig_spectrum_avg_mode_name_get, METH_O, NULL},
	 { "new_rig_spectrum_avg_mode", _wrap_new_rig_spectrum_avg_mode, METH_NOARGS, NULL},
	 { "delete_rig_spectrum_avg_mode", _wrap_delete_rig_spectrum_avg_mode, METH_O, NULL},
	 { "rig_spectrum_avg_mode_swigregister", rig_spectrum_avg_mode_swigregister, METH_O, NULL},
	 { "rig_spectrum_avg_mode_swiginit", rig_spectrum_avg_mode_swiginit, METH_VARARGS, NULL},
	 { "rig_spectrum_line_id_set", _wrap_rig_spectrum_line_id_set, METH_VARARGS, NULL},
	 { "rig_spectrum_line_id_get", _wrap_rig_spectrum_line_id_get, METH_O, NULL},
	 { "rig_spectrum_line_data_level_min_set", _wrap_rig_spectrum_line_data_level_min_set, METH_VARARGS, NULL},
	 { "rig_spectrum_line_data_level_min_get", _wrap_rig_spectrum_line_data_level_min_get, METH_O, NULL},
	 { "rig_spectrum_line_data_level_max_set", _wrap_rig_spectrum_line_data_level_max_set, METH_VARARGS, NULL},
	 { "rig_spectrum_line_data_level_max_get", _wrap_rig_spectrum_line_data_level_max_get, METH_O, NULL},
	 { "rig_spectrum_line_signal_strength_min_set", _wrap_rig_spectrum_line_signal_strength_min_set, METH_VARARGS, NULL},
	 { "rig_spectrum_line_signal_strength_min_get", _wrap_rig_spectrum_line_signal_strength_min_get, METH_O, NULL},
	 { "rig_spectrum_line_signal_strength_max_set", _wrap_rig_spectrum_line_signal_strength_max_set, METH_VARARGS, NULL},
	 { "rig_spectrum_line_signal_strength_max_get", _wrap_rig_spectrum_line_signal_strength_max_get, METH_O, NULL},
	 { "rig_spectrum_line_spectrum_mode_set", _wrap_rig_spectrum_line_spectrum_mode_set, METH_VARARGS, NULL},
	 { "rig_spectrum_line_spectrum_mode_get", _wrap_rig_spectrum_line_spectrum_mode_get, METH_O, NULL},
	 { "rig_spectrum_line_center_freq_set", _wrap_rig_spectrum_line_center_freq_set, METH_VARARGS, NULL},
	 { "rig_spectrum_line_center_freq_get", _wrap_rig_spectrum_line_center_freq_get, METH_O, NULL},
	 { "rig_spectrum_line_span_freq_set", _wrap_rig_spectrum_line_span_freq_set, METH_VARARGS, NULL},
	 { "rig_spectrum_line_span_freq_get", _wrap_rig_spectrum_line_span_freq_get, METH_O, NULL},
	 { "rig_spectrum_line_low_edge_freq_set", _wrap_rig_spectrum_line_low_edge_freq_set, METH_VARARGS, NULL},
	 { "rig_spectrum_line_low_edge_freq_get", _wrap_rig_spectrum_line_low_edge_freq_get, METH_O, NULL},
	 { "rig_spectrum_line_high_edge_freq_set", _wrap_rig_spectrum_line_high_edge_freq_set, METH_VARARGS, NULL},
	 { "rig_spectrum_line_high_edge_freq_get", _wrap_rig_spectrum_line_high_edge_freq_get, METH_O, NULL},
	 { "rig_spectrum_line_spectrum_data_length_set", _wrap_rig_spectrum_line_spectrum_data_length_set, METH_VARARGS, NULL},
	 { "rig_spectrum_line_spectrum_data_length_get", _wrap_rig_spectrum_line_spectrum_data_length_get, METH_O, NULL},
	 { "rig_spectrum_line_spectrum_data_set", _wrap_rig_spectrum_line_spectrum_data_set, METH_VARARGS, NULL},
	 { "rig_spectrum_line_spectrum_data_get", _wrap_rig_spectrum_line_spectrum_data_get, METH_O, NULL},
	 { "new_rig_spectrum_line", _wrap_new_rig_spectrum_line, METH_NOARGS, NULL},
	 { "delete_rig_spectrum_line", _wrap_delete_rig_spectrum_line, METH_O, NULL},
	 { "rig_spectrum_line_swigregister", rig_spectrum_line_swigregister, METH_O, NULL},
	 { "rig_spectrum_line_swiginit", rig_spectrum_line_swiginit, METH_VARARGS, NULL},
	 { "deferred_config_item_nextt_set", _wrap_deferred_config_item_nextt_set, METH_VARARGS, NULL},
	 { "deferred_config_item_nextt_get", _wrap_deferred_config_item_nextt_get, METH_O, NULL},
	 { "deferred_config_item_token_set", _wrap_deferred_config_item_token_set, METH_VARARGS, NULL},
	 { "deferred_config_item_token_get", _wrap_deferred_config_item_token_get, METH_O, NULL},
	 { "deferred_config_item_value_set", _wrap_deferred_config_item_value_set, METH_VARARGS, NULL},
	 { "deferred_config_item_value_get", _wrap_deferred_config_item_value_get, METH_O, NULL},
	 { "new_deferred_config_item", _wrap_new_deferred_config_item, METH_NOARGS, NULL},
	 { "delete_deferred_config_item", _wrap_delete_deferred_config_item, METH_O, NULL},
	 { "deferred_config_item_swigregister", deferred_config_item_swigregister, METH_O, NULL},
	 { "deferred_config_item_swiginit", deferred_config_item_swiginit, METH_VARARGS, NULL},
	 { "deferred_config_header_firstt_set", _wrap_deferred_config_header_firstt_set, METH_VARARGS, NULL},
	 { "deferred_config_header_firstt_get", _wrap_deferred_config_header_firstt_get, METH_O, NULL},
	 { "deferred_config_header_lastt_set", _wrap_deferred_config_header_lastt_set, METH_VARARGS, NULL},
	 { "deferred_config_header_lastt_get", _wrap_deferred_config_header_lastt_get, METH_O, NULL},
	 { "new_deferred_config_header", _wrap_new_deferred_config_header, METH_NOARGS, NULL},
	 { "delete_deferred_config_header", _wrap_delete_deferred_config_header, METH_O, NULL},
	 { "deferred_config_header_swigregister", deferred_config_header_swigregister, METH_O, NULL},
	 { "deferred_config_header_swiginit", deferred_config_header_swiginit, METH_VARARGS, NULL},
	 { "rig_caps_rig_model_set", _wrap_rig_caps_rig_model_set, METH_VARARGS, NULL},
	 { "rig_caps_rig_model_get", _wrap_rig_caps_rig_model_get, METH_O, NULL},
	 { "rig_caps_model_name_get", _wrap_rig_caps_model_name_get, METH_O, NULL},
	 { "rig_caps_mfg_name_get", _wrap_rig_caps_mfg_name_get, METH_O, NULL},
	 { "rig_caps_version_get", _wrap_rig_caps_version_get, METH_O, NULL},
	 { "rig_caps_copyright_get", _wrap_rig_caps_copyright_get, METH_O, NULL},
	 { "rig_caps_status_set", _wrap_rig_caps_status_set, METH_VARARGS, NULL},
	 { "rig_caps_status_get", _wrap_rig_caps_status_get, METH_O, NULL},
	 { "rig_caps_rig_type_set", _wrap_rig_caps_rig_type_set, METH_VARARGS, NULL},
	 { "rig_caps_rig_type_get", _wrap_rig_caps_rig_type_get, METH_O, NULL},
	 { "rig_caps_ptt_type_set", _wrap_rig_caps_ptt_type_set, METH_VARARGS, NULL},
	 { "rig_caps_ptt_type_get", _wrap_rig_caps_ptt_type_get, METH_O, NULL},
	 { "rig_caps_dcd_type_set", _wrap_rig_caps_dcd_type_set, METH_VARARGS, NULL},
	 { "rig_caps_dcd_type_get", _wrap_rig_caps_dcd_type_get, METH_O, NULL},
	 { "rig_caps_port_type_set", _wrap_rig_caps_port_type_set, METH_VARARGS, NULL},
	 { "rig_caps_port_type_get", _wrap_rig_caps_port_type_get, METH_O, NULL},
	 { "rig_caps_serial_rate_min_set", _wrap_rig_caps_serial_rate_min_set, METH_VARARGS, NULL},
	 { "rig_caps_serial_rate_min_get", _wrap_rig_caps_serial_rate_min_get, METH_O, NULL},
	 { "rig_caps_serial_rate_max_set", _wrap_rig_caps_serial_rate_max_set, METH_VARARGS, NULL},
	 { "rig_caps_serial_rate_max_get", _wrap_rig_caps_serial_rate_max_get, METH_O, NULL},
	 { "rig_caps_serial_data_bits_set", _wrap_rig_caps_serial_data_bits_set, METH_VARARGS, NULL},
	 { "rig_caps_serial_data_bits_get", _wrap_rig_caps_serial_data_bits_get, METH_O, NULL},
	 { "rig_caps_serial_stop_bits_set", _wrap_rig_caps_serial_stop_bits_set, METH_VARARGS, NULL},
	 { "rig_caps_serial_stop_bits_get", _wrap_rig_caps_serial_stop_bits_get, METH_O, NULL},
	 { "rig_caps_serial_parity_set", _wrap_rig_caps_serial_parity_set, METH_VARARGS, NULL},
	 { "rig_caps_serial_parity_get", _wrap_rig_caps_serial_parity_get, METH_O, NULL},
	 { "rig_caps_serial_handshake_set", _wrap_rig_caps_serial_handshake_set, METH_VARARGS, NULL},
	 { "rig_caps_serial_handshake_get", _wrap_rig_caps_serial_handshake_get, METH_O, NULL},
	 { "rig_caps_write_delay_set", _wrap_rig_caps_write_delay_set, METH_VARARGS, NULL},
	 { "rig_caps_write_delay_get", _wrap_rig_caps_write_delay_get, METH_O, NULL},
	 { "rig_caps_post_write_delay_set", _wrap_rig_caps_post_write_delay_set, METH_VARARGS, NULL},
	 { "rig_caps_post_write_delay_get", _wrap_rig_caps_post_write_delay_get, METH_O, NULL},
	 { "rig_caps_timeout_set", _wrap_rig_caps_timeout_set, METH_VARARGS, NULL},
	 { "rig_caps_timeout_get", _wrap_rig_caps_timeout_get, METH_O, NULL},
	 { "rig_caps_retry_set", _wrap_rig_caps_retry_set, METH_VARARGS, NULL},
	 { "rig_caps_retry_get", _wrap_rig_caps_retry_get, METH_O, NULL},
	 { "rig_caps_has_get_func_set", _wrap_rig_caps_has_get_func_set, METH_VARARGS, NULL},
	 { "rig_caps_has_get_func_get", _wrap_rig_caps_has_get_func_get, METH_O, NULL},
	 { "rig_caps_has_set_func_set", _wrap_rig_caps_has_set_func_set, METH_VARARGS, NULL},
	 { "rig_caps_has_set_func_get", _wrap_rig_caps_has_set_func_get, METH_O, NULL},
	 { "rig_caps_has_get_level_set", _wrap_rig_caps_has_get_level_set, METH_VARARGS, NULL},
	 { "rig_caps_has_get_level_get", _wrap_rig_caps_has_get_level_get, METH_O, NULL},
	 { "rig_caps_has_set_level_set", _wrap_rig_caps_has_set_level_set, METH_VARARGS, NULL},
	 { "rig_caps_has_set_level_get", _wrap_rig_caps_has_set_level_get, METH_O, NULL},
	 { "rig_caps_has_get_parm_set", _wrap_rig_caps_has_get_parm_set, METH_VARARGS, NULL},
	 { "rig_caps_has_get_parm_get", _wrap_rig_caps_has_get_parm_get, METH_O, NULL},
	 { "rig_caps_has_set_parm_set", _wrap_rig_caps_has_set_parm_set, METH_VARARGS, NULL},
	 { "rig_caps_has_set_parm_get", _wrap_rig_caps_has_set_parm_get, METH_O, NULL},
	 { "rig_caps_level_gran_set", _wrap_rig_caps_level_gran_set, METH_VARARGS, NULL},
	 { "rig_caps_level_gran_get", _wrap_rig_caps_level_gran_get, METH_O, NULL},
	 { "rig_caps_parm_gran_set", _wrap_rig_caps_parm_gran_set, METH_VARARGS, NULL},
	 { "rig_caps_parm_gran_get", _wrap_rig_caps_parm_gran_get, METH_O, NULL},
	 { "rig_caps_extparms_set", _wrap_rig_caps_extparms_set, METH_VARARGS, NULL},
	 { "rig_caps_extparms_get", _wrap_rig_caps_extparms_get, METH_O, NULL},
	 { "rig_caps_extlevels_set", _wrap_rig_caps_extlevels_set, METH_VARARGS, NULL},
	 { "rig_caps_extlevels_get", _wrap_rig_caps_extlevels_get, METH_O, NULL},
	 { "rig_caps_extfuncs_set", _wrap_rig_caps_extfuncs_set, METH_VARARGS, NULL},
	 { "rig_caps_extfuncs_get", _wrap_rig_caps_extfuncs_get, METH_O, NULL},
	 { "rig_caps_ext_tokens_set", _wrap_rig_caps_ext_tokens_set, METH_VARARGS, NULL},
	 { "rig_caps_ext_tokens_get", _wrap_rig_caps_ext_tokens_get, METH_O, NULL},
	 { "rig_caps_ctcss_list_set", _wrap_rig_caps_ctcss_list_set, METH_VARARGS, NULL},
	 { "rig_caps_ctcss_list_get", _wrap_rig_caps_ctcss_list_get, METH_O, NULL},
	 { "rig_caps_dcs_list_set", _wrap_rig_caps_dcs_list_set, METH_VARARGS, NULL},
	 { "rig_caps_dcs_list_get", _wrap_rig_caps_dcs_list_get, METH_O, NULL},
	 { "rig_caps_preamp_set", _wrap_rig_caps_preamp_set, METH_VARARGS, NULL},
	 { "rig_caps_preamp_get", _wrap_rig_caps_preamp_get, METH_O, NULL},
	 { "rig_caps_attenuator_set", _wrap_rig_caps_attenuator_set, METH_VARARGS, NULL},
	 { "rig_caps_attenuator_get", _wrap_rig_caps_attenuator_get, METH_O, NULL},
	 { "rig_caps_max_rit_set", _wrap_rig_caps_max_rit_set, METH_VARARGS, NULL},
	 { "rig_caps_max_rit_get", _wrap_rig_caps_max_rit_get, METH_O, NULL},
	 { "rig_caps_max_xit_set", _wrap_rig_caps_max_xit_set, METH_VARARGS, NULL},
	 { "rig_caps_max_xit_get", _wrap_rig_caps_max_xit_get, METH_O, NULL},
	 { "rig_caps_max_ifshift_set", _wrap_rig_caps_max_ifshift_set, METH_VARARGS, NULL},
	 { "rig_caps_max_ifshift_get", _wrap_rig_caps_max_ifshift_get, METH_O, NULL},
	 { "rig_caps_agc_level_count_set", _wrap_rig_caps_agc_level_count_set, METH_VARARGS, NULL},
	 { "rig_caps_agc_level_count_get", _wrap_rig_caps_agc_level_count_get, METH_O, NULL},
	 { "rig_caps_agc_levels_set", _wrap_rig_caps_agc_levels_set, METH_VARARGS, NULL},
	 { "rig_caps_agc_levels_get", _wrap_rig_caps_agc_levels_get, METH_O, NULL},
	 { "rig_caps_announces_set", _wrap_rig_caps_announces_set, METH_VARARGS, NULL},
	 { "rig_caps_announces_get", _wrap_rig_caps_announces_get, METH_O, NULL},
	 { "rig_caps_vfo_ops_set", _wrap_rig_caps_vfo_ops_set, METH_VARARGS, NULL},
	 { "rig_caps_vfo_ops_get", _wrap_rig_caps_vfo_ops_get, METH_O, NULL},
	 { "rig_caps_scan_ops_set", _wrap_rig_caps_scan_ops_set, METH_VARARGS, NULL},
	 { "rig_caps_scan_ops_get", _wrap_rig_caps_scan_ops_get, METH_O, NULL},
	 { "rig_caps_targetable_vfo_set", _wrap_rig_caps_targetable_vfo_set, METH_VARARGS, NULL},
	 { "rig_caps_targetable_vfo_get", _wrap_rig_caps_targetable_vfo_get, METH_O, NULL},
	 { "rig_caps_transceive_set", _wrap_rig_caps_transceive_set, METH_VARARGS, NULL},
	 { "rig_caps_transceive_get", _wrap_rig_caps_transceive_get, METH_O, NULL},
	 { "rig_caps_bank_qty_set", _wrap_rig_caps_bank_qty_set, METH_VARARGS, NULL},
	 { "rig_caps_bank_qty_get", _wrap_rig_caps_bank_qty_get, METH_O, NULL},
	 { "rig_caps_chan_desc_sz_set", _wrap_rig_caps_chan_desc_sz_set, METH_VARARGS, NULL},
	 { "rig_caps_chan_desc_sz_get", _wrap_rig_caps_chan_desc_sz_get, METH_O, NULL},
	 { "rig_caps_chan_list_set", _wrap_rig_caps_chan_list_set, METH_VARARGS, NULL},
	 { "rig_caps_chan_list_get", _wrap_rig_caps_chan_list_get, METH_O, NULL},
	 { "rig_caps_rx_range_list1_set", _wrap_rig_caps_rx_range_list1_set, METH_VARARGS, NULL},
	 { "rig_caps_rx_range_list1_get", _wrap_rig_caps_rx_range_list1_get, METH_O, NULL},
	 { "rig_caps_tx_range_list1_set", _wrap_rig_caps_tx_range_list1_set, METH_VARARGS, NULL},
	 { "rig_caps_tx_range_list1_get", _wrap_rig_caps_tx_range_list1_get, METH_O, NULL},
	 { "rig_caps_rx_range_list2_set", _wrap_rig_caps_rx_range_list2_set, METH_VARARGS, NULL},
	 { "rig_caps_rx_range_list2_get", _wrap_rig_caps_rx_range_list2_get, METH_O, NULL},
	 { "rig_caps_tx_range_list2_set", _wrap_rig_caps_tx_range_list2_set, METH_VARARGS, NULL},
	 { "rig_caps_tx_range_list2_get", _wrap_rig_caps_tx_range_list2_get, METH_O, NULL},
	 { "rig_caps_rx_range_list3_set", _wrap_rig_caps_rx_range_list3_set, METH_VARARGS, NULL},
	 { "rig_caps_rx_range_list3_get", _wrap_rig_caps_rx_range_list3_get, METH_O, NULL},
	 { "rig_caps_tx_range_list3_set", _wrap_rig_caps_tx_range_list3_set, METH_VARARGS, NULL},
	 { "rig_caps_tx_range_list3_get", _wrap_rig_caps_tx_range_list3_get, METH_O, NULL},
	 { "rig_caps_rx_range_list4_set", _wrap_rig_caps_rx_range_list4_set, METH_VARARGS, NULL},
	 { "rig_caps_rx_range_list4_get", _wrap_rig_caps_rx_range_list4_get, METH_O, NULL},
	 { "rig_caps_tx_range_list4_set", _wrap_rig_caps_tx_range_list4_set, METH_VARARGS, NULL},
	 { "rig_caps_tx_range_list4_get", _wrap_rig_caps_tx_range_list4_get, METH_O, NULL},
	 { "rig_caps_rx_range_list5_set", _wrap_rig_caps_rx_range_list5_set, METH_VARARGS, NULL},
	 { "rig_caps_rx_range_list5_get", _wrap_rig_caps_rx_range_list5_get, METH_O, NULL},
	 { "rig_caps_tx_range_list5_set", _wrap_rig_caps_tx_range_list5_set, METH_VARARGS, NULL},
	 { "rig_caps_tx_range_list5_get", _wrap_rig_caps_tx_range_list5_get, METH_O, NULL},
	 { "rig_caps_tuning_steps_set", _wrap_rig_caps_tuning_steps_set, METH_VARARGS, NULL},
	 { "rig_caps_tuning_steps_get", _wrap_rig_caps_tuning_steps_get, METH_O, NULL},
	 { "rig_caps_filters_set", _wrap_rig_caps_filters_set, METH_VARARGS, NULL},
	 { "rig_caps_filters_get", _wrap_rig_caps_filters_get, METH_O, NULL},
	 { "rig_caps_str_cal_set", _wrap_rig_caps_str_cal_set, METH_VARARGS, NULL},
	 { "rig_caps_str_cal_get", _wrap_rig_caps_str_cal_get, METH_O, NULL},
	 { "rig_caps_swr_cal_set", _wrap_rig_caps_swr_cal_set, METH_VARARGS, NULL},
	 { "rig_caps_swr_cal_get", _wrap_rig_caps_swr_cal_get, METH_O, NULL},
	 { "rig_caps_alc_cal_set", _wrap_rig_caps_alc_cal_set, METH_VARARGS, NULL},
	 { "rig_caps_alc_cal_get", _wrap_rig_caps_alc_cal_get, METH_O, NULL},
	 { "rig_caps_rfpower_meter_cal_set", _wrap_rig_caps_rfpower_meter_cal_set, METH_VARARGS, NULL},
	 { "rig_caps_rfpower_meter_cal_get", _wrap_rig_caps_rfpower_meter_cal_get, METH_O, NULL},
	 { "rig_caps_comp_meter_cal_set", _wrap_rig_caps_comp_meter_cal_set, METH_VARARGS, NULL},
	 { "rig_caps_comp_meter_cal_get", _wrap_rig_caps_comp_meter_cal_get, METH_O, NULL},
	 { "rig_caps_vd_meter_cal_set", _wrap_rig_caps_vd_meter_cal_set, METH_VARARGS, NULL},
	 { "rig_caps_vd_meter_cal_get", _wrap_rig_caps_vd_meter_cal_get, METH_O, NULL},
	 { "rig_caps_id_meter_cal_set", _wrap_rig_caps_id_meter_cal_set, METH_VARARGS, NULL},
	 { "rig_caps_id_meter_cal_get", _wrap_rig_caps_id_meter_cal_get, METH_O, NULL},
	 { "rig_caps_spectrum_scopes_set", _wrap_rig_caps_spectrum_scopes_set, METH_VARARGS, NULL},
	 { "rig_caps_spectrum_scopes_get", _wrap_rig_caps_spectrum_scopes_get, METH_O, NULL},
	 { "rig_caps_spectrum_modes_set", _wrap_rig_caps_spectrum_modes_set, METH_VARARGS, NULL},
	 { "rig_caps_spectrum_modes_get", _wrap_rig_caps_spectrum_modes_get, METH_O, NULL},
	 { "rig_caps_spectrum_spans_set", _wrap_rig_caps_spectrum_spans_set, METH_VARARGS, NULL},
	 { "rig_caps_spectrum_spans_get", _wrap_rig_caps_spectrum_spans_get, METH_O, NULL},
	 { "rig_caps_spectrum_avg_modes_set", _wrap_rig_caps_spectrum_avg_modes_set, METH_VARARGS, NULL},
	 { "rig_caps_spectrum_avg_modes_get", _wrap_rig_caps_spectrum_avg_modes_get, METH_O, NULL},
	 { "rig_caps_spectrum_attenuator_set", _wrap_rig_caps_spectrum_attenuator_set, METH_VARARGS, NULL},
	 { "rig_caps_spectrum_attenuator_get", _wrap_rig_caps_spectrum_attenuator_get, METH_O, NULL},
	 { "rig_caps_cfgparams_set", _wrap_rig_caps_cfgparams_set, METH_VARARGS, NULL},
	 { "rig_caps_cfgparams_get", _wrap_rig_caps_cfgparams_get, METH_O, NULL},
	 { "rig_caps_priv_set", _wrap_rig_caps_priv_set, METH_VARARGS, NULL},
	 { "rig_caps_priv_get", _wrap_rig_caps_priv_get, METH_O, NULL},
	 { "rig_caps_rig_init_set", _wrap_rig_caps_rig_init_set, METH_VARARGS, NULL},
	 { "rig_caps_rig_init_get", _wrap_rig_caps_rig_init_get, METH_O, NULL},
	 { "rig_caps_rig_cleanup_set", _wrap_rig_caps_rig_cleanup_set, METH_VARARGS, NULL},
	 { "rig_caps_rig_cleanup_get", _wrap_rig_caps_rig_cleanup_get, METH_O, NULL},
	 { "rig_caps_rig_open_set", _wrap_rig_caps_rig_open_set, METH_VARARGS, NULL},
	 { "rig_caps_rig_open_get", _wrap_rig_caps_rig_open_get, METH_O, NULL},
	 { "rig_caps_rig_close_set", _wrap_rig_caps_rig_close_set, METH_VARARGS, NULL},
	 { "rig_caps_rig_close_get", _wrap_rig_caps_rig_close_get, METH_O, NULL},
	 { "rig_caps_set_freq_set", _wrap_rig_caps_set_freq_set, METH_VARARGS, NULL},
	 { "rig_caps_set_freq_get", _wrap_rig_caps_set_freq_get, METH_O, NULL},
	 { "rig_caps_get_freq_set", _wrap_rig_caps_get_freq_set, METH_VARARGS, NULL},
	 { "rig_caps_get_freq_get", _wrap_rig_caps_get_freq_get, METH_O, NULL},
	 { "rig_caps_set_mode_set", _wrap_rig_caps_set_mode_set, METH_VARARGS, NULL},
	 { "rig_caps_set_mode_get", _wrap_rig_caps_set_mode_get, METH_O, NULL},
	 { "rig_caps_get_mode_set", _wrap_rig_caps_get_mode_set, METH_VARARGS, NULL},
	 { "rig_caps_get_mode_get", _wrap_rig_caps_get_mode_get, METH_O, NULL},
	 { "rig_caps_set_vfo_set", _wrap_rig_caps_set_vfo_set, METH_VARARGS, NULL},
	 { "rig_caps_set_vfo_get", _wrap_rig_caps_set_vfo_get, METH_O, NULL},
	 { "rig_caps_get_vfo_set", _wrap_rig_caps_get_vfo_set, METH_VARARGS, NULL},
	 { "rig_caps_get_vfo_get", _wrap_rig_caps_get_vfo_get, METH_O, NULL},
	 { "rig_caps_set_ptt_set", _wrap_rig_caps_set_ptt_set, METH_VARARGS, NULL},
	 { "rig_caps_set_ptt_get", _wrap_rig_caps_set_ptt_get, METH_O, NULL},
	 { "rig_caps_get_ptt_set", _wrap_rig_caps_get_ptt_set, METH_VARARGS, NULL},
	 { "rig_caps_get_ptt_get", _wrap_rig_caps_get_ptt_get, METH_O, NULL},
	 { "rig_caps_get_dcd_set", _wrap_rig_caps_get_dcd_set, METH_VARARGS, NULL},
	 { "rig_caps_get_dcd_get", _wrap_rig_caps_get_dcd_get, METH_O, NULL},
	 { "rig_caps_set_rptr_shift_set", _wrap_rig_caps_set_rptr_shift_set, METH_VARARGS, NULL},
	 { "rig_caps_set_rptr_shift_get", _wrap_rig_caps_set_rptr_shift_get, METH_O, NULL},
	 { "rig_caps_get_rptr_shift_set", _wrap_rig_caps_get_rptr_shift_set, METH_VARARGS, NULL},
	 { "rig_caps_get_rptr_shift_get", _wrap_rig_caps_get_rptr_shift_get, METH_O, NULL},
	 { "rig_caps_set_rptr_offs_set", _wrap_rig_caps_set_rptr_offs_set, METH_VARARGS, NULL},
	 { "rig_caps_set_rptr_offs_get", _wrap_rig_caps_set_rptr_offs_get, METH_O, NULL},
	 { "rig_caps_get_rptr_offs_set", _wrap_rig_caps_get_rptr_offs_set, METH_VARARGS, NULL},
	 { "rig_caps_get_rptr_offs_get", _wrap_rig_caps_get_rptr_offs_get, METH_O, NULL},
	 { "rig_caps_set_split_freq_set", _wrap_rig_caps_set_split_freq_set, METH_VARARGS, NULL},
	 { "rig_caps_set_split_freq_get", _wrap_rig_caps_set_split_freq_get, METH_O, NULL},
	 { "rig_caps_get_split_freq_set", _wrap_rig_caps_get_split_freq_set, METH_VARARGS, NULL},
	 { "rig_caps_get_split_freq_get", _wrap_rig_caps_get_split_freq_get, METH_O, NULL},
	 { "rig_caps_set_split_mode_set", _wrap_rig_caps_set_split_mode_set, METH_VARARGS, NULL},
	 { "rig_caps_set_split_mode_get", _wrap_rig_caps_set_split_mode_get, METH_O, NULL},
	 { "rig_caps_get_split_mode_set", _wrap_rig_caps_get_split_mode_set, METH_VARARGS, NULL},
	 { "rig_caps_get_split_mode_get", _wrap_rig_caps_get_split_mode_get, METH_O, NULL},
	 { "rig_caps_set_split_freq_mode_set", _wrap_rig_caps_set_split_freq_mode_set, METH_VARARGS, NULL},
	 { "rig_caps_set_split_freq_mode_get", _wrap_rig_caps_set_split_freq_mode_get, METH_O, NULL},
	 { "rig_caps_get_split_freq_mode_set", _wrap_rig_caps_get_split_freq_mode_set, METH_VARARGS, NULL},
	 { "rig_caps_get_split_freq_mode_get", _wrap_rig_caps_get_split_freq_mode_get, METH_O, NULL},
	 { "rig_caps_set_split_vfo_set", _wrap_rig_caps_set_split_vfo_set, METH_VARARGS, NULL},
	 { "rig_caps_set_split_vfo_get", _wrap_rig_caps_set_split_vfo_get, METH_O, NULL},
	 { "rig_caps_get_split_vfo_set", _wrap_rig_caps_get_split_vfo_set, METH_VARARGS, NULL},
	 { "rig_caps_get_split_vfo_get", _wrap_rig_caps_get_split_vfo_get, METH_O, NULL},
	 { "rig_caps_set_rit_set", _wrap_rig_caps_set_rit_set, METH_VARARGS, NULL},
	 { "rig_caps_set_rit_get", _wrap_rig_caps_set_rit_get, METH_O, NULL},
	 { "rig_caps_get_rit_set", _wrap_rig_caps_get_rit_set, METH_VARARGS, NULL},
	 { "rig_caps_get_rit_get", _wrap_rig_caps_get_rit_get, METH_O, NULL},
	 { "rig_caps_set_xit_set", _wrap_rig_caps_set_xit_set, METH_VARARGS, NULL},
	 { "rig_caps_set_xit_get", _wrap_rig_caps_set_xit_get, METH_O, NULL},
	 { "rig_caps_get_xit_set", _wrap_rig_caps_get_xit_set, METH_VARARGS, NULL},
	 { "rig_caps_get_xit_get", _wrap_rig_caps_get_xit_get, METH_O, NULL},
	 { "rig_caps_set_ts_set", _wrap_rig_caps_set_ts_set, METH_VARARGS, NULL},
	 { "rig_caps_set_ts_get", _wrap_rig_caps_set_ts_get, METH_O, NULL},
	 { "rig_caps_get_ts_set", _wrap_rig_caps_get_ts_set, METH_VARARGS, NULL},
	 { "rig_caps_get_ts_get", _wrap_rig_caps_get_ts_get, METH_O, NULL},
	 { "rig_caps_set_dcs_code_set", _wrap_rig_caps_set_dcs_code_set, METH_VARARGS, NULL},
	 { "rig_caps_set_dcs_code_get", _wrap_rig_caps_set_dcs_code_get, METH_O, NULL},
	 { "rig_caps_get_dcs_code_set", _wrap_rig_caps_get_dcs_code_set, METH_VARARGS, NULL},
	 { "rig_caps_get_dcs_code_get", _wrap_rig_caps_get_dcs_code_get, METH_O, NULL},
	 { "rig_caps_set_tone_set", _wrap_rig_caps_set_tone_set, METH_VARARGS, NULL},
	 { "rig_caps_set_tone_get", _wrap_rig_caps_set_tone_get, METH_O, NULL},
	 { "rig_caps_get_tone_set", _wrap_rig_caps_get_tone_set, METH_VARARGS, NULL},
	 { "rig_caps_get_tone_get", _wrap_rig_caps_get_tone_get, METH_O, NULL},
	 { "rig_caps_set_ctcss_tone_set", _wrap_rig_caps_set_ctcss_tone_set, METH_VARARGS, NULL},
	 { "rig_caps_set_ctcss_tone_get", _wrap_rig_caps_set_ctcss_tone_get, METH_O, NULL},
	 { "rig_caps_get_ctcss_tone_set", _wrap_rig_caps_get_ctcss_tone_set, METH_VARARGS, NULL},
	 { "rig_caps_get_ctcss_tone_get", _wrap_rig_caps_get_ctcss_tone_get, METH_O, NULL},
	 { "rig_caps_set_dcs_sql_set", _wrap_rig_caps_set_dcs_sql_set, METH_VARARGS, NULL},
	 { "rig_caps_set_dcs_sql_get", _wrap_rig_caps_set_dcs_sql_get, METH_O, NULL},
	 { "rig_caps_get_dcs_sql_set", _wrap_rig_caps_get_dcs_sql_set, METH_VARARGS, NULL},
	 { "rig_caps_get_dcs_sql_get", _wrap_rig_caps_get_dcs_sql_get, METH_O, NULL},
	 { "rig_caps_set_tone_sql_set", _wrap_rig_caps_set_tone_sql_set, METH_VARARGS, NULL},
	 { "rig_caps_set_tone_sql_get", _wrap_rig_caps_set_tone_sql_get, METH_O, NULL},
	 { "rig_caps_get_tone_sql_set", _wrap_rig_caps_get_tone_sql_set, METH_VARARGS, NULL},
	 { "rig_caps_get_tone_sql_get", _wrap_rig_caps_get_tone_sql_get, METH_O, NULL},
	 { "rig_caps_set_ctcss_sql_set", _wrap_rig_caps_set_ctcss_sql_set, METH_VARARGS, NULL},
	 { "rig_caps_set_ctcss_sql_get", _wrap_rig_caps_set_ctcss_sql_get, METH_O, NULL},
	 { "rig_caps_get_ctcss_sql_set", _wrap_rig_caps_get_ctcss_sql_set, METH_VARARGS, NULL},
	 { "rig_caps_get_ctcss_sql_get", _wrap_rig_caps_get_ctcss_sql_get, METH_O, NULL},
	 { "rig_caps_power2mW_set", _wrap_rig_caps_power2mW_set, METH_VARARGS, NULL},
	 { "rig_caps_power2mW_get", _wrap_rig_caps_power2mW_get, METH_O, NULL},
	 { "rig_caps_mW2power_set", _wrap_rig_caps_mW2power_set, METH_VARARGS, NULL},
	 { "rig_caps_mW2power_get", _wrap_rig_caps_mW2power_get, METH_O, NULL},
	 { "rig_caps_set_powerstat_set", _wrap_rig_caps_set_powerstat_set, METH_VARARGS, NULL},
	 { "rig_caps_set_powerstat_get", _wrap_rig_caps_set_powerstat_get, METH_O, NULL},
	 { "rig_caps_get_powerstat_set", _wrap_rig_caps_get_powerstat_set, METH_VARARGS, NULL},
	 { "rig_caps_get_powerstat_get", _wrap_rig_caps_get_powerstat_get, METH_O, NULL},
	 { "rig_caps_reset_set", _wrap_rig_caps_reset_set, METH_VARARGS, NULL},
	 { "rig_caps_reset_get", _wrap_rig_caps_reset_get, METH_O, NULL},
	 { "rig_caps_set_ant_set", _wrap_rig_caps_set_ant_set, METH_VARARGS, NULL},
	 { "rig_caps_set_ant_get", _wrap_rig_caps_set_ant_get, METH_O, NULL},
	 { "rig_caps_get_ant_set", _wrap_rig_caps_get_ant_set, METH_VARARGS, NULL},
	 { "rig_caps_get_ant_get", _wrap_rig_caps_get_ant_get, METH_O, NULL},
	 { "rig_caps_set_level_set", _wrap_rig_caps_set_level_set, METH_VARARGS, NULL},
	 { "rig_caps_set_level_get", _wrap_rig_caps_set_level_get, METH_O, NULL},
	 { "rig_caps_get_level_set", _wrap_rig_caps_get_level_set, METH_VARARGS, NULL},
	 { "rig_caps_get_level_get", _wrap_rig_caps_get_level_get, METH_O, NULL},
	 { "rig_caps_set_func_set", _wrap_rig_caps_set_func_set, METH_VARARGS, NULL},
	 { "rig_caps_set_func_get", _wrap_rig_caps_set_func_get, METH_O, NULL},
	 { "rig_caps_get_func_set", _wrap_rig_caps_get_func_set, METH_VARARGS, NULL},
	 { "rig_caps_get_func_get", _wrap_rig_caps_get_func_get, METH_O, NULL},
	 { "rig_caps_set_parm_set", _wrap_rig_caps_set_parm_set, METH_VARARGS, NULL},
	 { "rig_caps_set_parm_get", _wrap_rig_caps_set_parm_get, METH_O, NULL},
	 { "rig_caps_get_parm_set", _wrap_rig_caps_get_parm_set, METH_VARARGS, NULL},
	 { "rig_caps_get_parm_get", _wrap_rig_caps_get_parm_get, METH_O, NULL},
	 { "rig_caps_set_ext_level_set", _wrap_rig_caps_set_ext_level_set, METH_VARARGS, NULL},
	 { "rig_caps_set_ext_level_get", _wrap_rig_caps_set_ext_level_get, METH_O, NULL},
	 { "rig_caps_get_ext_level_set", _wrap_rig_caps_get_ext_level_set, METH_VARARGS, NULL},
	 { "rig_caps_get_ext_level_get", _wrap_rig_caps_get_ext_level_get, METH_O, NULL},
	 { "rig_caps_set_ext_func_set", _wrap_rig_caps_set_ext_func_set, METH_VARARGS, NULL},
	 { "rig_caps_set_ext_func_get", _wrap_rig_caps_set_ext_func_get, METH_O, NULL},
	 { "rig_caps_get_ext_func_set", _wrap_rig_caps_get_ext_func_set, METH_VARARGS, NULL},
	 { "rig_caps_get_ext_func_get", _wrap_rig_caps_get_ext_func_get, METH_O, NULL},
	 { "rig_caps_set_ext_parm_set", _wrap_rig_caps_set_ext_parm_set, METH_VARARGS, NULL},
	 { "rig_caps_set_ext_parm_get", _wrap_rig_caps_set_ext_parm_get, METH_O, NULL},
	 { "rig_caps_get_ext_parm_set", _wrap_rig_caps_get_ext_parm_set, METH_VARARGS, NULL},
	 { "rig_caps_get_ext_parm_get", _wrap_rig_caps_get_ext_parm_get, METH_O, NULL},
	 { "rig_caps_set_conf_set", _wrap_rig_caps_set_conf_set, METH_VARARGS, NULL},
	 { "rig_caps_set_conf_get", _wrap_rig_caps_set_conf_get, METH_O, NULL},
	 { "rig_caps_get_conf_set", _wrap_rig_caps_get_conf_set, METH_VARARGS, NULL},
	 { "rig_caps_get_conf_get", _wrap_rig_caps_get_conf_get, METH_O, NULL},
	 { "rig_caps_send_dtmf_set", _wrap_rig_caps_send_dtmf_set, METH_VARARGS, NULL},
	 { "rig_caps_send_dtmf_get", _wrap_rig_caps_send_dtmf_get, METH_O, NULL},
	 { "rig_caps_recv_dtmf_set", _wrap_rig_caps_recv_dtmf_set, METH_VARARGS, NULL},
	 { "rig_caps_recv_dtmf_get", _wrap_rig_caps_recv_dtmf_get, METH_O, NULL},
	 { "rig_caps_send_morse_set", _wrap_rig_caps_send_morse_set, METH_VARARGS, NULL},
	 { "rig_caps_send_morse_get", _wrap_rig_caps_send_morse_get, METH_O, NULL},
	 { "rig_caps_stop_morse_set", _wrap_rig_caps_stop_morse_set, METH_VARARGS, NULL},
	 { "rig_caps_stop_morse_get", _wrap_rig_caps_stop_morse_get, METH_O, NULL},
	 { "rig_caps_wait_morse_set", _wrap_rig_caps_wait_morse_set, METH_VARARGS, NULL},
	 { "rig_caps_wait_morse_get", _wrap_rig_caps_wait_morse_get, METH_O, NULL},
	 { "rig_caps_send_voice_mem_set", _wrap_rig_caps_send_voice_mem_set, METH_VARARGS, NULL},
	 { "rig_caps_send_voice_mem_get", _wrap_rig_caps_send_voice_mem_get, METH_O, NULL},
	 { "rig_caps_stop_voice_mem_set", _wrap_rig_caps_stop_voice_mem_set, METH_VARARGS, NULL},
	 { "rig_caps_stop_voice_mem_get", _wrap_rig_caps_stop_voice_mem_get, METH_O, NULL},
	 { "rig_caps_set_bank_set", _wrap_rig_caps_set_bank_set, METH_VARARGS, NULL},
	 { "rig_caps_set_bank_get", _wrap_rig_caps_set_bank_get, METH_O, NULL},
	 { "rig_caps_set_mem_set", _wrap_rig_caps_set_mem_set, METH_VARARGS, NULL},
	 { "rig_caps_set_mem_get", _wrap_rig_caps_set_mem_get, METH_O, NULL},
	 { "rig_caps_get_mem_set", _wrap_rig_caps_get_mem_set, METH_VARARGS, NULL},
	 { "rig_caps_get_mem_get", _wrap_rig_caps_get_mem_get, METH_O, NULL},
	 { "rig_caps_vfo_op_set", _wrap_rig_caps_vfo_op_set, METH_VARARGS, NULL},
	 { "rig_caps_vfo_op_get", _wrap_rig_caps_vfo_op_get, METH_O, NULL},
	 { "rig_caps_scan_set", _wrap_rig_caps_scan_set, METH_VARARGS, NULL},
	 { "rig_caps_scan_get", _wrap_rig_caps_scan_get, METH_O, NULL},
	 { "rig_caps_set_trn_set", _wrap_rig_caps_set_trn_set, METH_VARARGS, NULL},
	 { "rig_caps_set_trn_get", _wrap_rig_caps_set_trn_get, METH_O, NULL},
	 { "rig_caps_get_trn_set", _wrap_rig_caps_get_trn_set, METH_VARARGS, NULL},
	 { "rig_caps_get_trn_get", _wrap_rig_caps_get_trn_get, METH_O, NULL},
	 { "rig_caps_decode_event_set", _wrap_rig_caps_decode_event_set, METH_VARARGS, NULL},
	 { "rig_caps_decode_event_get", _wrap_rig_caps_decode_event_get, METH_O, NULL},
	 { "rig_caps_set_channel_set", _wrap_rig_caps_set_channel_set, METH_VARARGS, NULL},
	 { "rig_caps_set_channel_get", _wrap_rig_caps_set_channel_get, METH_O, NULL},
	 { "rig_caps_get_channel_set", _wrap_rig_caps_get_channel_set, METH_VARARGS, NULL},
	 { "rig_caps_get_channel_get", _wrap_rig_caps_get_channel_get, METH_O, NULL},
	 { "rig_caps_get_info_set", _wrap_rig_caps_get_info_set, METH_VARARGS, NULL},
	 { "rig_caps_get_info_get", _wrap_rig_caps_get_info_get, METH_O, NULL},
	 { "rig_caps_set_chan_all_cb_set", _wrap_rig_caps_set_chan_all_cb_set, METH_VARARGS, NULL},
	 { "rig_caps_set_chan_all_cb_get", _wrap_rig_caps_set_chan_all_cb_get, METH_O, NULL},
	 { "rig_caps_get_chan_all_cb_set", _wrap_rig_caps_get_chan_all_cb_set, METH_VARARGS, NULL},
	 { "rig_caps_get_chan_all_cb_get", _wrap_rig_caps_get_chan_all_cb_get, METH_O, NULL},
	 { "rig_caps_set_mem_all_cb_set", _wrap_rig_caps_set_mem_all_cb_set, METH_VARARGS, NULL},
	 { "rig_caps_set_mem_all_cb_get", _wrap_rig_caps_set_mem_all_cb_get, METH_O, NULL},
	 { "rig_caps_get_mem_all_cb_set", _wrap_rig_caps_get_mem_all_cb_set, METH_VARARGS, NULL},
	 { "rig_caps_get_mem_all_cb_get", _wrap_rig_caps_get_mem_all_cb_get, METH_O, NULL},
	 { "rig_caps_set_vfo_opt_set", _wrap_rig_caps_set_vfo_opt_set, METH_VARARGS, NULL},
	 { "rig_caps_set_vfo_opt_get", _wrap_rig_caps_set_vfo_opt_get, METH_O, NULL},
	 { "rig_caps_rig_get_vfo_info_set", _wrap_rig_caps_rig_get_vfo_info_set, METH_VARARGS, NULL},
	 { "rig_caps_rig_get_vfo_info_get", _wrap_rig_caps_rig_get_vfo_info_get, METH_O, NULL},
	 { "rig_caps_set_clock_set", _wrap_rig_caps_set_clock_set, METH_VARARGS, NULL},
	 { "rig_caps_set_clock_get", _wrap_rig_caps_set_clock_get, METH_O, NULL},
	 { "rig_caps_get_clock_set", _wrap_rig_caps_get_clock_set, METH_VARARGS, NULL},
	 { "rig_caps_get_clock_get", _wrap_rig_caps_get_clock_get, METH_O, NULL},
	 { "rig_caps_clone_combo_set_get", _wrap_rig_caps_clone_combo_set_get, METH_O, NULL},
	 { "rig_caps_clone_combo_get_get", _wrap_rig_caps_clone_combo_get_get, METH_O, NULL},
	 { "rig_caps_async_data_supported_set", _wrap_rig_caps_async_data_supported_set, METH_VARARGS, NULL},
	 { "rig_caps_async_data_supported_get", _wrap_rig_caps_async_data_supported_get, METH_O, NULL},
	 { "rig_caps_read_frame_direct_set", _wrap_rig_caps_read_frame_direct_set, METH_VARARGS, NULL},
	 { "rig_caps_read_frame_direct_get", _wrap_rig_caps_read_frame_direct_get, METH_O, NULL},
	 { "rig_caps_is_async_frame_set", _wrap_rig_caps_is_async_frame_set, METH_VARARGS, NULL},
	 { "rig_caps_is_async_frame_get", _wrap_rig_caps_is_async_frame_get, METH_O, NULL},
	 { "rig_caps_process_async_frame_set", _wrap_rig_caps_process_async_frame_set, METH_VARARGS, NULL},
	 { "rig_caps_process_async_frame_get", _wrap_rig_caps_process_async_frame_get, METH_O, NULL},
	 { "rig_caps_hamlib_check_rig_caps_set", _wrap_rig_caps_hamlib_check_rig_caps_set, METH_VARARGS, NULL},
	 { "rig_caps_hamlib_check_rig_caps_get", _wrap_rig_caps_hamlib_check_rig_caps_get, METH_O, NULL},
	 { "rig_caps_get_conf2_set", _wrap_rig_caps_get_conf2_set, METH_VARARGS, NULL},
	 { "rig_caps_get_conf2_get", _wrap_rig_caps_get_conf2_get, METH_O, NULL},
	 { "rig_caps_password_set", _wrap_rig_caps_password_set, METH_VARARGS, NULL},
	 { "rig_caps_password_get", _wrap_rig_caps_password_get, METH_O, NULL},
	 { "rig_caps_set_lock_mode_set", _wrap_rig_caps_set_lock_mode_set, METH_VARARGS, NULL},
	 { "rig_caps_set_lock_mode_get", _wrap_rig_caps_set_lock_mode_get, METH_O, NULL},
	 { "rig_caps_get_lock_mode_set", _wrap_rig_caps_get_lock_mode_set, METH_VARARGS, NULL},
	 { "rig_caps_get_lock_mode_get", _wrap_rig_caps_get_lock_mode_get, METH_O, NULL},
	 { "rig_caps_timeout_retry_set", _wrap_rig_caps_timeout_retry_set, METH_VARARGS, NULL},
	 { "rig_caps_timeout_retry_get", _wrap_rig_caps_timeout_retry_get, METH_O, NULL},
	 { "rig_caps_morse_qsize_set", _wrap_rig_caps_morse_qsize_set, METH_VARARGS, NULL},
	 { "rig_caps_morse_qsize_get", _wrap_rig_caps_morse_qsize_get, METH_O, NULL},
	 { "new_rig_caps", _wrap_new_rig_caps, METH_NOARGS, NULL},
	 { "delete_rig_caps", _wrap_delete_rig_caps, METH_O, NULL},
	 { "rig_caps_swigregister", rig_caps_swigregister, METH_O, NULL},
	 { "rig_caps_swiginit", rig_caps_swiginit, METH_VARARGS, NULL},
	 { "multicast_s_multicast_running_set", _wrap_multicast_s_multicast_running_set, METH_VARARGS, NULL},
	 { "multicast_s_multicast_running_get", _wrap_multicast_s_multicast_running_get, METH_O, NULL},
	 { "multicast_s_sock_set", _wrap_multicast_s_sock_set, METH_VARARGS, NULL},
	 { "multicast_s_sock_get", _wrap_multicast_s_sock_get, METH_O, NULL},
	 { "multicast_s_seqnumber_set", _wrap_multicast_s_seqnumber_set, METH_VARARGS, NULL},
	 { "multicast_s_seqnumber_get", _wrap_multicast_s_seqnumber_get, METH_O, NULL},
	 { "multicast_s_runflag_set", _wrap_multicast_s_runflag_set, METH_VARARGS, NULL},
	 { "multicast_s_runflag_get", _wrap_multicast_s_runflag_get, METH_O, NULL},
	 { "multicast_s_threadid_set", _wrap_multicast_s_threadid_set, METH_VARARGS, NULL},
	 { "multicast_s_threadid_get", _wrap_multicast_s_threadid_get, METH_O, NULL},
	 { "multicast_s_mutex_set", _wrap_multicast_s_mutex_set, METH_VARARGS, NULL},
	 { "multicast_s_mutex_get", _wrap_multicast_s_mutex_get, METH_O, NULL},
	 { "multicast_s_mutex_initialized_set", _wrap_multicast_s_mutex_initialized_set, METH_VARARGS, NULL},
	 { "multicast_s_mutex_initialized_get", _wrap_multicast_s_mutex_initialized_get, METH_O, NULL},
	 { "multicast_s_dest_addr_set", _wrap_multicast_s_dest_addr_set, METH_VARARGS, NULL},
	 { "multicast_s_dest_addr_get", _wrap_multicast_s_dest_addr_get, METH_O, NULL},
	 { "multicast_s_port_set", _wrap_multicast_s_port_set, METH_VARARGS, NULL},
	 { "multicast_s_port_get", _wrap_multicast_s_port_get, METH_O, NULL},
	 { "new_multicast_s", _wrap_new_multicast_s, METH_NOARGS, NULL},
	 { "delete_multicast_s", _wrap_delete_multicast_s, METH_O, NULL},
	 { "multicast_s_swigregister", multicast_s_swigregister, METH_O, NULL},
	 { "multicast_s_swiginit", multicast_s_swiginit, METH_VARARGS, NULL},
	 { "rig_callbacks_freq_event_set", _wrap_rig_callbacks_freq_event_set, METH_VARARGS, NULL},
	 { "rig_callbacks_freq_event_get", _wrap_rig_callbacks_freq_event_get, METH_O, NULL},
	 { "rig_callbacks_freq_arg_set", _wrap_rig_callbacks_freq_arg_set, METH_VARARGS, NULL},
	 { "rig_callbacks_freq_arg_get", _wrap_rig_callbacks_freq_arg_get, METH_O, NULL},
	 { "rig_callbacks_mode_event_set", _wrap_rig_callbacks_mode_event_set, METH_VARARGS, NULL},
	 { "rig_callbacks_mode_event_get", _wrap_rig_callbacks_mode_event_get, METH_O, NULL},
	 { "rig_callbacks_mode_arg_set", _wrap_rig_callbacks_mode_arg_set, METH_VARARGS, NULL},
	 { "rig_callbacks_mode_arg_get", _wrap_rig_callbacks_mode_arg_get, METH_O, NULL},
	 { "rig_callbacks_vfo_event_set", _wrap_rig_callbacks_vfo_event_set, METH_VARARGS, NULL},
	 { "rig_callbacks_vfo_event_get", _wrap_rig_callbacks_vfo_event_get, METH_O, NULL},
	 { "rig_callbacks_vfo_arg_set", _wrap_rig_callbacks_vfo_arg_set, METH_VARARGS, NULL},
	 { "rig_callbacks_vfo_arg_get", _wrap_rig_callbacks_vfo_arg_get, METH_O, NULL},
	 { "rig_callbacks_ptt_event_set", _wrap_rig_callbacks_ptt_event_set, METH_VARARGS, NULL},
	 { "rig_callbacks_ptt_event_get", _wrap_rig_callbacks_ptt_event_get, METH_O, NULL},
	 { "rig_callbacks_ptt_arg_set", _wrap_rig_callbacks_ptt_arg_set, METH_VARARGS, NULL},
	 { "rig_callbacks_ptt_arg_get", _wrap_rig_callbacks_ptt_arg_get, METH_O, NULL},
	 { "rig_callbacks_dcd_event_set", _wrap_rig_callbacks_dcd_event_set, METH_VARARGS, NULL},
	 { "rig_callbacks_dcd_event_get", _wrap_rig_callbacks_dcd_event_get, METH_O, NULL},
	 { "rig_callbacks_dcd_arg_set", _wrap_rig_callbacks_dcd_arg_set, METH_VARARGS, NULL},
	 { "rig_callbacks_dcd_arg_get", _wrap_rig_callbacks_dcd_arg_get, METH_O, NULL},
	 { "rig_callbacks_pltune_set", _wrap_rig_callbacks_pltune_set, METH_VARARGS, NULL},
	 { "rig_callbacks_pltune_get", _wrap_rig_callbacks_pltune_get, METH_O, NULL},
	 { "rig_callbacks_pltune_arg_set", _wrap_rig_callbacks_pltune_arg_set, METH_VARARGS, NULL},
	 { "rig_callbacks_pltune_arg_get", _wrap_rig_callbacks_pltune_arg_get, METH_O, NULL},
	 { "rig_callbacks_spectrum_event_set", _wrap_rig_callbacks_spectrum_event_set, METH_VARARGS, NULL},
	 { "rig_callbacks_spectrum_event_get", _wrap_rig_callbacks_spectrum_event_get, METH_O, NULL},
	 { "rig_callbacks_spectrum_arg_set", _wrap_rig_callbacks_spectrum_arg_set, METH_VARARGS, NULL},
	 { "rig_callbacks_spectrum_arg_get", _wrap_rig_callbacks_spectrum_arg_get, METH_O, NULL},
	 { "new_rig_callbacks", _wrap_new_rig_callbacks, METH_NOARGS, NULL},
	 { "delete_rig_callbacks", _wrap_delete_rig_callbacks, METH_O, NULL},
	 { "rig_callbacks_swigregister", rig_callbacks_swigregister, METH_O, NULL},
	 { "rig_callbacks_swiginit", rig_callbacks_swiginit, METH_VARARGS, NULL},
	 { "netrigctl_get_vfo_mode", _wrap_netrigctl_get_vfo_mode, METH_O, NULL},
	 { "rigerror", _wrap_rigerror, METH_O, NULL},
	 { "rigerror2", _wrap_rigerror2, METH_O, NULL},
	 { "rig_set_debug", _wrap_rig_set_debug, METH_O, NULL},
	 { "rig_get_debug", _wrap_rig_get_debug, METH_NOARGS, NULL},
	 { "rig_set_debug_time_stamp", _wrap_rig_set_debug_time_stamp, METH_O, NULL},
	 { "add2debugmsgsave", _wrap_add2debugmsgsave, METH_O, NULL},
	 { "rig_strrmode", _wrap_rig_strrmode, METH_O, NULL},
	 { "rig_strrmodes", _wrap_rig_strrmodes, METH_VARARGS, NULL},
	 { "rig_strvfo", _wrap_rig_strvfo, METH_O, NULL},
	 { "rig_strfunc", _wrap_rig_strfunc, METH_O, NULL},
	 { "rig_strlevel", _wrap_rig_strlevel, METH_O, NULL},
	 { "rig_strparm", _wrap_rig_strparm, METH_O, NULL},
	 { "rig_stragclevel", _wrap_rig_stragclevel, METH_O, NULL},
	 { "rig_strptrshift", _wrap_rig_strptrshift, METH_O, NULL},
	 { "rig_strvfop", _wrap_rig_strvfop, METH_O, NULL},
	 { "rig_strscan", _wrap_rig_strscan, METH_O, NULL},
	 { "rig_strstatus", _wrap_rig_strstatus, METH_O, NULL},
	 { "rig_strmtype", _wrap_rig_strmtype, METH_O, NULL},
	 { "rig_strspectrummode", _wrap_rig_strspectrummode, METH_O, NULL},
	 { "rig_strcommstatus", _wrap_rig_strcommstatus, METH_O, NULL},
	 { "hl_usleep", _wrap_hl_usleep, METH_O, NULL},
	 { "longlat2locator", _wrap_longlat2locator, METH_VARARGS, NULL},
	 { "locator2longlat", _wrap_locator2longlat, METH_O, NULL},
	 { "rig_state_vfo_comp_set", _wrap_rig_state_vfo_comp_set, METH_VARARGS, NULL},
	 { "rig_state_vfo_comp_get", _wrap_rig_state_vfo_comp_get, METH_O, NULL},
	 { "rig_state_rx_range_list_set", _wrap_rig_state_rx_range_list_set, METH_VARARGS, NULL},
	 { "rig_state_rx_range_list_get", _wrap_rig_state_rx_range_list_get, METH_O, NULL},
	 { "rig_state_tx_range_list_set", _wrap_rig_state_tx_range_list_set, METH_VARARGS, NULL},
	 { "rig_state_tx_range_list_get", _wrap_rig_state_tx_range_list_get, METH_O, NULL},
	 { "rig_state_tuning_steps_set", _wrap_rig_state_tuning_steps_set, METH_VARARGS, NULL},
	 { "rig_state_tuning_steps_get", _wrap_rig_state_tuning_steps_get, METH_O, NULL},
	 { "rig_state_filters_set", _wrap_rig_state_filters_set, METH_VARARGS, NULL},
	 { "rig_state_filters_get", _wrap_rig_state_filters_get, METH_O, NULL},
	 { "rig_state_str_cal_set", _wrap_rig_state_str_cal_set, METH_VARARGS, NULL},
	 { "rig_state_str_cal_get", _wrap_rig_state_str_cal_get, METH_O, NULL},
	 { "rig_state_chan_list_set", _wrap_rig_state_chan_list_set, METH_VARARGS, NULL},
	 { "rig_state_chan_list_get", _wrap_rig_state_chan_list_get, METH_O, NULL},
	 { "rig_state_max_rit_set", _wrap_rig_state_max_rit_set, METH_VARARGS, NULL},
	 { "rig_state_max_rit_get", _wrap_rig_state_max_rit_get, METH_O, NULL},
	 { "rig_state_max_xit_set", _wrap_rig_state_max_xit_set, METH_VARARGS, NULL},
	 { "rig_state_max_xit_get", _wrap_rig_state_max_xit_get, METH_O, NULL},
	 { "rig_state_max_ifshift_set", _wrap_rig_state_max_ifshift_set, METH_VARARGS, NULL},
	 { "rig_state_max_ifshift_get", _wrap_rig_state_max_ifshift_get, METH_O, NULL},
	 { "rig_state_announces_set", _wrap_rig_state_announces_set, METH_VARARGS, NULL},
	 { "rig_state_announces_get", _wrap_rig_state_announces_get, METH_O, NULL},
	 { "rig_state_preamp_set", _wrap_rig_state_preamp_set, METH_VARARGS, NULL},
	 { "rig_state_preamp_get", _wrap_rig_state_preamp_get, METH_O, NULL},
	 { "rig_state_attenuator_set", _wrap_rig_state_attenuator_set, METH_VARARGS, NULL},
	 { "rig_state_attenuator_get", _wrap_rig_state_attenuator_get, METH_O, NULL},
	 { "rig_state_has_get_func_set", _wrap_rig_state_has_get_func_set, METH_VARARGS, NULL},
	 { "rig_state_has_get_func_get", _wrap_rig_state_has_get_func_get, METH_O, NULL},
	 { "rig_state_has_set_func_set", _wrap_rig_state_has_set_func_set, METH_VARARGS, NULL},
	 { "rig_state_has_set_func_get", _wrap_rig_state_has_set_func_get, METH_O, NULL},
	 { "rig_state_has_get_level_set", _wrap_rig_state_has_get_level_set, METH_VARARGS, NULL},
	 { "rig_state_has_get_level_get", _wrap_rig_state_has_get_level_get, METH_O, NULL},
	 { "rig_state_has_set_level_set", _wrap_rig_state_has_set_level_set, METH_VARARGS, NULL},
	 { "rig_state_has_set_level_get", _wrap_rig_state_has_set_level_get, METH_O, NULL},
	 { "rig_state_has_get_parm_set", _wrap_rig_state_has_get_parm_set, METH_VARARGS, NULL},
	 { "rig_state_has_get_parm_get", _wrap_rig_state_has_get_parm_get, METH_O, NULL},
	 { "rig_state_has_set_parm_set", _wrap_rig_state_has_set_parm_set, METH_VARARGS, NULL},
	 { "rig_state_has_set_parm_get", _wrap_rig_state_has_set_parm_get, METH_O, NULL},
	 { "rig_state_level_gran_set", _wrap_rig_state_level_gran_set, METH_VARARGS, NULL},
	 { "rig_state_level_gran_get", _wrap_rig_state_level_gran_get, METH_O, NULL},
	 { "rig_state_parm_gran_set", _wrap_rig_state_parm_gran_set, METH_VARARGS, NULL},
	 { "rig_state_parm_gran_get", _wrap_rig_state_parm_gran_get, METH_O, NULL},
	 { "rig_state_transaction_active_set", _wrap_rig_state_transaction_active_set, METH_VARARGS, NULL},
	 { "rig_state_transaction_active_get", _wrap_rig_state_transaction_active_get, METH_O, NULL},
	 { "rig_state_current_vfo_set", _wrap_rig_state_current_vfo_set, METH_VARARGS, NULL},
	 { "rig_state_current_vfo_get", _wrap_rig_state_current_vfo_get, METH_O, NULL},
	 { "rig_state_vfo_list_set", _wrap_rig_state_vfo_list_set, METH_VARARGS, NULL},
	 { "rig_state_vfo_list_get", _wrap_rig_state_vfo_list_get, METH_O, NULL},
	 { "rig_state_comm_state_set", _wrap_rig_state_comm_state_set, METH_VARARGS, NULL},
	 { "rig_state_comm_state_get", _wrap_rig_state_comm_state_get, METH_O, NULL},
	 { "rig_state_priv_set", _wrap_rig_state_priv_set, METH_VARARGS, NULL},
	 { "rig_state_priv_get", _wrap_rig_state_priv_get, METH_O, NULL},
	 { "rig_state_obj_set", _wrap_rig_state_obj_set, METH_VARARGS, NULL},
	 { "rig_state_obj_get", _wrap_rig_state_obj_get, METH_O, NULL},
	 { "rig_state_async_data_enabled_set", _wrap_rig_state_async_data_enabled_set, METH_VARARGS, NULL},
	 { "rig_state_async_data_enabled_get", _wrap_rig_state_async_data_enabled_get, METH_O, NULL},
	 { "rig_state_poll_interval_set", _wrap_rig_state_poll_interval_set, METH_VARARGS, NULL},
	 { "rig_state_poll_interval_get", _wrap_rig_state_poll_interval_get, METH_O, NULL},
	 { "rig_state_current_freq_set", _wrap_rig_state_current_freq_set, METH_VARARGS, NULL},
	 { "rig_state_current_freq_get", _wrap_rig_state_current_freq_get, METH_O, NULL},
	 { "rig_state_current_mode_set", _wrap_rig_state_current_mode_set, METH_VARARGS, NULL},
	 { "rig_state_current_mode_get", _wrap_rig_state_current_mode_get, METH_O, NULL},
	 { "rig_state_current_width_set", _wrap_rig_state_current_width_set, METH_VARARGS, NULL},
	 { "rig_state_current_width_get", _wrap_rig_state_current_width_get, METH_O, NULL},
	 { "rig_state_tx_vfo_set", _wrap_rig_state_tx_vfo_set, METH_VARARGS, NULL},
	 { "rig_state_tx_vfo_get", _wrap_rig_state_tx_vfo_get, METH_O, NULL},
	 { "rig_state_mode_list_set", _wrap_rig_state_mode_list_set, METH_VARARGS, NULL},
	 { "rig_state_mode_list_get", _wrap_rig_state_mode_list_get, METH_O, NULL},
	 { "rig_state_transmit_set", _wrap_rig_state_transmit_set, METH_VARARGS, NULL},
	 { "rig_state_transmit_get", _wrap_rig_state_transmit_get, METH_O, NULL},
	 { "rig_state_lo_freq_set", _wrap_rig_state_lo_freq_set, METH_VARARGS, NULL},
	 { "rig_state_lo_freq_get", _wrap_rig_state_lo_freq_get, METH_O, NULL},
	 { "rig_state_twiddle_time_set", _wrap_rig_state_twiddle_time_set, METH_VARARGS, NULL},
	 { "rig_state_twiddle_time_get", _wrap_rig_state_twiddle_time_get, METH_O, NULL},
	 { "rig_state_twiddle_timeout_set", _wrap_rig_state_twiddle_timeout_set, METH_VARARGS, NULL},
	 { "rig_state_twiddle_timeout_get", _wrap_rig_state_twiddle_timeout_get, METH_O, NULL},
	 { "rig_state_uplink_set", _wrap_rig_state_uplink_set, METH_VARARGS, NULL},
	 { "rig_state_uplink_get", _wrap_rig_state_uplink_get, METH_O, NULL},
	 { "rig_state_vfo_opt_set", _wrap_rig_state_vfo_opt_set, METH_VARARGS, NULL},
	 { "rig_state_vfo_opt_get", _wrap_rig_state_vfo_opt_get, METH_O, NULL},
	 { "rig_state_auto_power_on_set", _wrap_rig_state_auto_power_on_set, METH_VARARGS, NULL},
	 { "rig_state_auto_power_on_get", _wrap_rig_state_auto_power_on_get, METH_O, NULL},
	 { "rig_state_auto_power_off_set", _wrap_rig_state_auto_power_off_set, METH_VARARGS, NULL},
	 { "rig_state_auto_power_off_get", _wrap_rig_state_auto_power_off_get, METH_O, NULL},
	 { "rig_state_auto_disable_screensaver_set", _wrap_rig_state_auto_disable_screensaver_set, METH_VARARGS, NULL},
	 { "rig_state_auto_disable_screensaver_get", _wrap_rig_state_auto_disable_screensaver_get, METH_O, NULL},
	 { "rig_state_ptt_share_set", _wrap_rig_state_ptt_share_set, METH_VARARGS, NULL},
	 { "rig_state_ptt_share_get", _wrap_rig_state_ptt_share_get, METH_O, NULL},
	 { "rig_state_power_now_set", _wrap_rig_state_power_now_set, METH_VARARGS, NULL},
	 { "rig_state_power_now_get", _wrap_rig_state_power_now_get, METH_O, NULL},
	 { "rig_state_power_min_set", _wrap_rig_state_power_min_set, METH_VARARGS, NULL},
	 { "rig_state_power_min_get", _wrap_rig_state_power_min_get, METH_O, NULL},
	 { "rig_state_power_max_set", _wrap_rig_state_power_max_set, METH_VARARGS, NULL},
	 { "rig_state_power_max_get", _wrap_rig_state_power_max_get, METH_O, NULL},
	 { "rig_state_disable_yaesu_bandselect_set", _wrap_rig_state_disable_yaesu_bandselect_set, METH_VARARGS, NULL},
	 { "rig_state_disable_yaesu_bandselect_get", _wrap_rig_state_disable_yaesu_bandselect_get, METH_O, NULL},
	 { "rig_state_twiddle_rit_set", _wrap_rig_state_twiddle_rit_set, METH_VARARGS, NULL},
	 { "rig_state_twiddle_rit_get", _wrap_rig_state_twiddle_rit_get, METH_O, NULL},
	 { "rig_state_twiddle_state_set", _wrap_rig_state_twiddle_state_set, METH_VARARGS, NULL},
	 { "rig_state_twiddle_state_get", _wrap_rig_state_twiddle_state_get, METH_O, NULL},
	 { "rig_state_rx_vfo_set", _wrap_rig_state_rx_vfo_set, METH_VARARGS, NULL},
	 { "rig_state_rx_vfo_get", _wrap_rig_state_rx_vfo_get, METH_O, NULL},
	 { "rig_state_snapshot_packet_sequence_number_set", _wrap_rig_state_snapshot_packet_sequence_number_set, METH_VARARGS, NULL},
	 { "rig_state_snapshot_packet_sequence_number_get", _wrap_rig_state_snapshot_packet_sequence_number_get, METH_O, NULL},
	 { "rig_state_multicast_publisher_run_set", _wrap_rig_state_multicast_publisher_run_set, METH_VARARGS, NULL},
	 { "rig_state_multicast_publisher_run_get", _wrap_rig_state_multicast_publisher_run_get, METH_O, NULL},
	 { "rig_state_multicast_publisher_priv_data_set", _wrap_rig_state_multicast_publisher_priv_data_set, METH_VARARGS, NULL},
	 { "rig_state_multicast_publisher_priv_data_get", _wrap_rig_state_multicast_publisher_priv_data_get, METH_O, NULL},
	 { "rig_state_async_data_handler_thread_run_set", _wrap_rig_state_async_data_handler_thread_run_set, METH_VARARGS, NULL},
	 { "rig_state_async_data_handler_thread_run_get", _wrap_rig_state_async_data_handler_thread_run_get, METH_O, NULL},
	 { "rig_state_async_data_handler_priv_data_set", _wrap_rig_state_async_data_handler_priv_data_set, METH_VARARGS, NULL},
	 { "rig_state_async_data_handler_priv_data_get", _wrap_rig_state_async_data_handler_priv_data_get, METH_O, NULL},
	 { "rig_state_poll_routine_thread_run_set", _wrap_rig_state_poll_routine_thread_run_set, METH_VARARGS, NULL},
	 { "rig_state_poll_routine_thread_run_get", _wrap_rig_state_poll_routine_thread_run_get, METH_O, NULL},
	 { "rig_state_poll_routine_priv_data_set", _wrap_rig_state_poll_routine_priv_data_set, METH_VARARGS, NULL},
	 { "rig_state_poll_routine_priv_data_get", _wrap_rig_state_poll_routine_priv_data_get, METH_O, NULL},
	 { "rig_state_mutex_set_transaction_set", _wrap_rig_state_mutex_set_transaction_set, METH_VARARGS, NULL},
	 { "rig_state_mutex_set_transaction_get", _wrap_rig_state_mutex_set_transaction_get, METH_O, NULL},
	 { "rig_state_rigport_set", _wrap_rig_state_rigport_set, METH_VARARGS, NULL},
	 { "rig_state_rigport_get", _wrap_rig_state_rigport_get, METH_O, NULL},
	 { "rig_state_pttport_set", _wrap_rig_state_pttport_set, METH_VARARGS, NULL},
	 { "rig_state_pttport_get", _wrap_rig_state_pttport_get, METH_O, NULL},
	 { "rig_state_dcdport_set", _wrap_rig_state_dcdport_set, METH_VARARGS, NULL},
	 { "rig_state_dcdport_get", _wrap_rig_state_dcdport_get, METH_O, NULL},
	 { "rig_state_use_cached_freq_set", _wrap_rig_state_use_cached_freq_set, METH_VARARGS, NULL},
	 { "rig_state_use_cached_freq_get", _wrap_rig_state_use_cached_freq_get, METH_O, NULL},
	 { "rig_state_use_cached_mode_set", _wrap_rig_state_use_cached_mode_set, METH_VARARGS, NULL},
	 { "rig_state_use_cached_mode_get", _wrap_rig_state_use_cached_mode_get, METH_O, NULL},
	 { "rig_state_use_cached_ptt_set", _wrap_rig_state_use_cached_ptt_set, METH_VARARGS, NULL},
	 { "rig_state_use_cached_ptt_get", _wrap_rig_state_use_cached_ptt_get, METH_O, NULL},
	 { "rig_state_depth_set", _wrap_rig_state_depth_set, METH_VARARGS, NULL},
	 { "rig_state_depth_get", _wrap_rig_state_depth_get, METH_O, NULL},
	 { "rig_state_lock_mode_set", _wrap_rig_state_lock_mode_set, METH_VARARGS, NULL},
	 { "rig_state_lock_mode_get", _wrap_rig_state_lock_mode_get, METH_O, NULL},
	 { "rig_state_powerstat_set", _wrap_rig_state_powerstat_set, METH_VARARGS, NULL},
	 { "rig_state_powerstat_get", _wrap_rig_state_powerstat_get, METH_O, NULL},
	 { "rig_state_tuner_control_pathname_set", _wrap_rig_state_tuner_control_pathname_set, METH_VARARGS, NULL},
	 { "rig_state_tuner_control_pathname_get", _wrap_rig_state_tuner_control_pathname_get, METH_O, NULL},
	 { "rig_state_client_version_set", _wrap_rig_state_client_version_set, METH_VARARGS, NULL},
	 { "rig_state_client_version_get", _wrap_rig_state_client_version_get, METH_O, NULL},
	 { "rig_state_offset_vfoa_set", _wrap_rig_state_offset_vfoa_set, METH_VARARGS, NULL},
	 { "rig_state_offset_vfoa_get", _wrap_rig_state_offset_vfoa_get, METH_O, NULL},
	 { "rig_state_offset_vfob_set", _wrap_rig_state_offset_vfob_set, METH_VARARGS, NULL},
	 { "rig_state_offset_vfob_get", _wrap_rig_state_offset_vfob_get, METH_O, NULL},
	 { "rig_state_multicast_set", _wrap_rig_state_multicast_set, METH_VARARGS, NULL},
	 { "rig_state_multicast_get", _wrap_rig_state_multicast_get, METH_O, NULL},
	 { "rig_state_rig_model_set", _wrap_rig_state_rig_model_set, METH_VARARGS, NULL},
	 { "rig_state_rig_model_get", _wrap_rig_state_rig_model_get, METH_O, NULL},
	 { "rig_state_model_name_set", _wrap_rig_state_model_name_set, METH_VARARGS, NULL},
	 { "rig_state_model_name_get", _wrap_rig_state_model_name_get, METH_O, NULL},
	 { "rig_state_mfg_name_set", _wrap_rig_state_mfg_name_set, METH_VARARGS, NULL},
	 { "rig_state_mfg_name_get", _wrap_rig_state_mfg_name_get, METH_O, NULL},
	 { "rig_state_version_set", _wrap_rig_state_version_set, METH_VARARGS, NULL},
	 { "rig_state_version_get", _wrap_rig_state_version_get, METH_O, NULL},
	 { "rig_state_copyright_set", _wrap_rig_state_copyright_set, METH_VARARGS, NULL},
	 { "rig_state_copyright_get", _wrap_rig_state_copyright_get, METH_O, NULL},
	 { "rig_state_status_set", _wrap_rig_state_status_set, METH_VARARGS, NULL},
	 { "rig_state_status_get", _wrap_rig_state_status_get, METH_O, NULL},
	 { "rig_state_rig_type_set", _wrap_rig_state_rig_type_set, METH_VARARGS, NULL},
	 { "rig_state_rig_type_get", _wrap_rig_state_rig_type_get, METH_O, NULL},
	 { "rig_state_ptt_type_set", _wrap_rig_state_ptt_type_set, METH_VARARGS, NULL},
	 { "rig_state_ptt_type_get", _wrap_rig_state_ptt_type_get, METH_O, NULL},
	 { "rig_state_dcd_type_set", _wrap_rig_state_dcd_type_set, METH_VARARGS, NULL},
	 { "rig_state_dcd_type_get", _wrap_rig_state_dcd_type_get, METH_O, NULL},
	 { "rig_state_port_type_set", _wrap_rig_state_port_type_set, METH_VARARGS, NULL},
	 { "rig_state_port_type_get", _wrap_rig_state_port_type_get, METH_O, NULL},
	 { "rig_state_serial_rate_min_set", _wrap_rig_state_serial_rate_min_set, METH_VARARGS, NULL},
	 { "rig_state_serial_rate_min_get", _wrap_rig_state_serial_rate_min_get, METH_O, NULL},
	 { "rig_state_serial_rate_max_set", _wrap_rig_state_serial_rate_max_set, METH_VARARGS, NULL},
	 { "rig_state_serial_rate_max_get", _wrap_rig_state_serial_rate_max_get, METH_O, NULL},
	 { "rig_state_serial_data_bits_set", _wrap_rig_state_serial_data_bits_set, METH_VARARGS, NULL},
	 { "rig_state_serial_data_bits_get", _wrap_rig_state_serial_data_bits_get, METH_O, NULL},
	 { "rig_state_serial_stop_bits_set", _wrap_rig_state_serial_stop_bits_set, METH_VARARGS, NULL},
	 { "rig_state_serial_stop_bits_get", _wrap_rig_state_serial_stop_bits_get, METH_O, NULL},
	 { "rig_state_serial_parity_set", _wrap_rig_state_serial_parity_set, METH_VARARGS, NULL},
	 { "rig_state_serial_parity_get", _wrap_rig_state_serial_parity_get, METH_O, NULL},
	 { "rig_state_serial_handshake_set", _wrap_rig_state_serial_handshake_set, METH_VARARGS, NULL},
	 { "rig_state_serial_handshake_get", _wrap_rig_state_serial_handshake_get, METH_O, NULL},
	 { "rig_state_write_delay_set", _wrap_rig_state_write_delay_set, METH_VARARGS, NULL},
	 { "rig_state_write_delay_get", _wrap_rig_state_write_delay_get, METH_O, NULL},
	 { "rig_state_post_write_delay_set", _wrap_rig_state_post_write_delay_set, METH_VARARGS, NULL},
	 { "rig_state_post_write_delay_get", _wrap_rig_state_post_write_delay_get, METH_O, NULL},
	 { "rig_state_timeout_set", _wrap_rig_state_timeout_set, METH_VARARGS, NULL},
	 { "rig_state_timeout_get", _wrap_rig_state_timeout_get, METH_O, NULL},
	 { "rig_state_retry_set", _wrap_rig_state_retry_set, METH_VARARGS, NULL},
	 { "rig_state_retry_get", _wrap_rig_state_retry_get, METH_O, NULL},
	 { "rig_state_targetable_vfo_set", _wrap_rig_state_targetable_vfo_set, METH_VARARGS, NULL},
	 { "rig_state_targetable_vfo_get", _wrap_rig_state_targetable_vfo_get, METH_O, NULL},
	 { "rig_state_async_data_supported_set", _wrap_rig_state_async_data_supported_set, METH_VARARGS, NULL},
	 { "rig_state_async_data_supported_get", _wrap_rig_state_async_data_supported_get, METH_O, NULL},
	 { "rig_state_agc_level_count_set", _wrap_rig_state_agc_level_count_set, METH_VARARGS, NULL},
	 { "rig_state_agc_level_count_get", _wrap_rig_state_agc_level_count_get, METH_O, NULL},
	 { "rig_state_agc_levels_set", _wrap_rig_state_agc_levels_set, METH_VARARGS, NULL},
	 { "rig_state_agc_levels_get", _wrap_rig_state_agc_levels_get, METH_O, NULL},
	 { "rig_state_ctcss_list_set", _wrap_rig_state_ctcss_list_set, METH_VARARGS, NULL},
	 { "rig_state_ctcss_list_get", _wrap_rig_state_ctcss_list_get, METH_O, NULL},
	 { "rig_state_dcs_list_set", _wrap_rig_state_dcs_list_set, METH_VARARGS, NULL},
	 { "rig_state_dcs_list_get", _wrap_rig_state_dcs_list_get, METH_O, NULL},
	 { "rig_state_vfo_ops_set", _wrap_rig_state_vfo_ops_set, METH_VARARGS, NULL},
	 { "rig_state_vfo_ops_get", _wrap_rig_state_vfo_ops_get, METH_O, NULL},
	 { "rig_state_scan_ops_set", _wrap_rig_state_scan_ops_set, METH_VARARGS, NULL},
	 { "rig_state_scan_ops_get", _wrap_rig_state_scan_ops_get, METH_O, NULL},
	 { "rig_state_transceive_set", _wrap_rig_state_transceive_set, METH_VARARGS, NULL},
	 { "rig_state_transceive_get", _wrap_rig_state_transceive_get, METH_O, NULL},
	 { "rig_state_bank_qty_set", _wrap_rig_state_bank_qty_set, METH_VARARGS, NULL},
	 { "rig_state_bank_qty_get", _wrap_rig_state_bank_qty_get, METH_O, NULL},
	 { "rig_state_chan_desc_sz_set", _wrap_rig_state_chan_desc_sz_set, METH_VARARGS, NULL},
	 { "rig_state_chan_desc_sz_get", _wrap_rig_state_chan_desc_sz_get, METH_O, NULL},
	 { "rig_state_rx_range_list1_set", _wrap_rig_state_rx_range_list1_set, METH_VARARGS, NULL},
	 { "rig_state_rx_range_list1_get", _wrap_rig_state_rx_range_list1_get, METH_O, NULL},
	 { "rig_state_tx_range_list1_set", _wrap_rig_state_tx_range_list1_set, METH_VARARGS, NULL},
	 { "rig_state_tx_range_list1_get", _wrap_rig_state_tx_range_list1_get, METH_O, NULL},
	 { "rig_state_rx_range_list2_set", _wrap_rig_state_rx_range_list2_set, METH_VARARGS, NULL},
	 { "rig_state_rx_range_list2_get", _wrap_rig_state_rx_range_list2_get, METH_O, NULL},
	 { "rig_state_tx_range_list2_set", _wrap_rig_state_tx_range_list2_set, METH_VARARGS, NULL},
	 { "rig_state_tx_range_list2_get", _wrap_rig_state_tx_range_list2_get, METH_O, NULL},
	 { "rig_state_rx_range_list3_set", _wrap_rig_state_rx_range_list3_set, METH_VARARGS, NULL},
	 { "rig_state_rx_range_list3_get", _wrap_rig_state_rx_range_list3_get, METH_O, NULL},
	 { "rig_state_tx_range_list3_set", _wrap_rig_state_tx_range_list3_set, METH_VARARGS, NULL},
	 { "rig_state_tx_range_list3_get", _wrap_rig_state_tx_range_list3_get, METH_O, NULL},
	 { "rig_state_rx_range_list4_set", _wrap_rig_state_rx_range_list4_set, METH_VARARGS, NULL},
	 { "rig_state_rx_range_list4_get", _wrap_rig_state_rx_range_list4_get, METH_O, NULL},
	 { "rig_state_tx_range_list4_set", _wrap_rig_state_tx_range_list4_set, METH_VARARGS, NULL},
	 { "rig_state_tx_range_list4_get", _wrap_rig_state_tx_range_list4_get, METH_O, NULL},
	 { "rig_state_rx_range_list5_set", _wrap_rig_state_rx_range_list5_set, METH_VARARGS, NULL},
	 { "rig_state_rx_range_list5_get", _wrap_rig_state_rx_range_list5_get, METH_O, NULL},
	 { "rig_state_tx_range_list5_set", _wrap_rig_state_tx_range_list5_set, METH_VARARGS, NULL},
	 { "rig_state_tx_range_list5_get", _wrap_rig_state_tx_range_list5_get, METH_O, NULL},
	 { "rig_state_spectrum_scopes_set", _wrap_rig_state_spectrum_scopes_set, METH_VARARGS, NULL},
	 { "rig_state_spectrum_scopes_get", _wrap_rig_state_spectrum_scopes_get, METH_O, NULL},
	 { "rig_state_spectrum_modes_set", _wrap_rig_state_spectrum_modes_set, METH_VARARGS, NULL},
	 { "rig_state_spectrum_modes_get", _wrap_rig_state_spectrum_modes_get, METH_O, NULL},
	 { "rig_state_spectrum_spans_set", _wrap_rig_state_spectrum_spans_set, METH_VARARGS, NULL},
	 { "rig_state_spectrum_spans_get", _wrap_rig_state_spectrum_spans_get, METH_O, NULL},
	 { "rig_state_spectrum_avg_modes_set", _wrap_rig_state_spectrum_avg_modes_set, METH_VARARGS, NULL},
	 { "rig_state_spectrum_avg_modes_get", _wrap_rig_state_spectrum_avg_modes_get, METH_O, NULL},
	 { "rig_state_spectrum_attenuator_set", _wrap_rig_state_spectrum_attenuator_set, METH_VARARGS, NULL},
	 { "rig_state_spectrum_attenuator_get", _wrap_rig_state_spectrum_attenuator_get, METH_O, NULL},
	 { "rig_state_morse_data_handler_thread_run_set", _wrap_rig_state_morse_data_handler_thread_run_set, METH_VARARGS, NULL},
	 { "rig_state_morse_data_handler_thread_run_get", _wrap_rig_state_morse_data_handler_thread_run_get, METH_O, NULL},
	 { "rig_state_morse_data_handler_priv_data_set", _wrap_rig_state_morse_data_handler_priv_data_set, METH_VARARGS, NULL},
	 { "rig_state_morse_data_handler_priv_data_get", _wrap_rig_state_morse_data_handler_priv_data_get, METH_O, NULL},
	 { "rig_state_fifo_morse_set", _wrap_rig_state_fifo_morse_set, METH_VARARGS, NULL},
	 { "rig_state_fifo_morse_get", _wrap_rig_state_fifo_morse_get, METH_O, NULL},
	 { "rig_state_doppler_set", _wrap_rig_state_doppler_set, METH_VARARGS, NULL},
	 { "rig_state_doppler_get", _wrap_rig_state_doppler_get, METH_O, NULL},
	 { "rig_state_multicast_data_addr_set", _wrap_rig_state_multicast_data_addr_set, METH_VARARGS, NULL},
	 { "rig_state_multicast_data_addr_get", _wrap_rig_state_multicast_data_addr_get, METH_O, NULL},
	 { "rig_state_multicast_data_port_set", _wrap_rig_state_multicast_data_port_set, METH_VARARGS, NULL},
	 { "rig_state_multicast_data_port_get", _wrap_rig_state_multicast_data_port_get, METH_O, NULL},
	 { "rig_state_multicast_cmd_addr_set", _wrap_rig_state_multicast_cmd_addr_set, METH_VARARGS, NULL},
	 { "rig_state_multicast_cmd_addr_get", _wrap_rig_state_multicast_cmd_addr_get, METH_O, NULL},
	 { "rig_state_multicast_cmd_port_set", _wrap_rig_state_multicast_cmd_port_set, METH_VARARGS, NULL},
	 { "rig_state_multicast_cmd_port_get", _wrap_rig_state_multicast_cmd_port_get, METH_O, NULL},
	 { "rig_state_multicast_receiver_run_set", _wrap_rig_state_multicast_receiver_run_set, METH_VARARGS, NULL},
	 { "rig_state_multicast_receiver_run_get", _wrap_rig_state_multicast_receiver_run_get, METH_O, NULL},
	 { "rig_state_multicast_receiver_priv_data_set", _wrap_rig_state_multicast_receiver_priv_data_set, METH_VARARGS, NULL},
	 { "rig_state_multicast_receiver_priv_data_get", _wrap_rig_state_multicast_receiver_priv_data_get, METH_O, NULL},
	 { "rig_state_comm_status_set", _wrap_rig_state_comm_status_set, METH_VARARGS, NULL},
	 { "rig_state_comm_status_get", _wrap_rig_state_comm_status_get, METH_O, NULL},
	 { "rig_state_device_id_set", _wrap_rig_state_device_id_set, METH_VARARGS, NULL},
	 { "rig_state_device_id_get", _wrap_rig_state_device_id_get, METH_O, NULL},
	 { "rig_state_dual_watch_set", _wrap_rig_state_dual_watch_set, METH_VARARGS, NULL},
	 { "rig_state_dual_watch_get", _wrap_rig_state_dual_watch_get, METH_O, NULL},
	 { "rig_state_post_ptt_delay_set", _wrap_rig_state_post_ptt_delay_set, METH_VARARGS, NULL},
	 { "rig_state_post_ptt_delay_get", _wrap_rig_state_post_ptt_delay_get, METH_O, NULL},
	 { "rig_state_freq_event_elapsed_set", _wrap_rig_state_freq_event_elapsed_set, METH_VARARGS, NULL},
	 { "rig_state_freq_event_elapsed_get", _wrap_rig_state_freq_event_elapsed_get, METH_O, NULL},
	 { "rig_state_freq_skip_set", _wrap_rig_state_freq_skip_set, METH_VARARGS, NULL},
	 { "rig_state_freq_skip_get", _wrap_rig_state_freq_skip_get, METH_O, NULL},
	 { "rig_state_client_set", _wrap_rig_state_client_set, METH_VARARGS, NULL},
	 { "rig_state_client_get", _wrap_rig_state_client_get, METH_O, NULL},
	 { "rig_state_api_mutex_set", _wrap_rig_state_api_mutex_set, METH_VARARGS, NULL},
	 { "rig_state_api_mutex_get", _wrap_rig_state_api_mutex_get, METH_O, NULL},
	 { "new_rig_state", _wrap_new_rig_state, METH_NOARGS, NULL},
	 { "delete_rig_state", _wrap_delete_rig_state, METH_O, NULL},
	 { "rig_state_swigregister", rig_state_swigregister, METH_O, NULL},
	 { "rig_state_swiginit", rig_state_swiginit, METH_VARARGS, NULL},
	 { "Rig_rig_set", _wrap_Rig_rig_set, METH_VARARGS, NULL},
	 { "Rig_rig_get", _wrap_Rig_rig_get, METH_O, NULL},
	 { "Rig_caps_set", _wrap_Rig_caps_set, METH_VARARGS, NULL},
	 { "Rig_caps_get", _wrap_Rig_caps_get, METH_O, NULL},
	 { "Rig_state_set", _wrap_Rig_state_set, METH_VARARGS, NULL},
	 { "Rig_state_get", _wrap_Rig_state_get, METH_O, NULL},
	 { "Rig_error_status_set", _wrap_Rig_error_status_set, METH_VARARGS, NULL},
	 { "Rig_error_status_get", _wrap_Rig_error_status_get, METH_O, NULL},
	 { "Rig_do_exception_set", _wrap_Rig_do_exception_set, METH_VARARGS, NULL},
	 { "Rig_do_exception_get", _wrap_Rig_do_exception_get, METH_O, NULL},
	 { "new_Rig", _wrap_new_Rig, METH_O, NULL},
	 { "delete_Rig", _wrap_delete_Rig, METH_O, NULL},
	 { "Rig_open", _wrap_Rig_open, METH_O, NULL},
	 { "Rig_close", _wrap_Rig_close, METH_O, NULL},
	 { "Rig_set_freq", _wrap_Rig_set_freq, METH_VARARGS, NULL},
	 { "Rig_set_mode", _wrap_Rig_set_mode, METH_VARARGS, NULL},
	 { "Rig_set_ptt", _wrap_Rig_set_ptt, METH_VARARGS, NULL},
	 { "Rig_set_rptr_shift", _wrap_Rig_set_rptr_shift, METH_VARARGS, NULL},
	 { "Rig_set_rptr_offs", _wrap_Rig_set_rptr_offs, METH_VARARGS, NULL},
	 { "Rig_set_ctcss_tone", _wrap_Rig_set_ctcss_tone, METH_VARARGS, NULL},
	 { "Rig_set_dcs_code", _wrap_Rig_set_dcs_code, METH_VARARGS, NULL},
	 { "Rig_set_ctcss_sql", _wrap_Rig_set_ctcss_sql, METH_VARARGS, NULL},
	 { "Rig_set_dcs_sql", _wrap_Rig_set_dcs_sql, METH_VARARGS, NULL},
	 { "Rig_set_split_freq", _wrap_Rig_set_split_freq, METH_VARARGS, NULL},
	 { "Rig_set_split_mode", _wrap_Rig_set_split_mode, METH_VARARGS, NULL},
	 { "Rig_set_split_freq_mode", _wrap_Rig_set_split_freq_mode, METH_VARARGS, NULL},
	 { "Rig_set_split_vfo", _wrap_Rig_set_split_vfo, METH_VARARGS, NULL},
	 { "Rig_set_rit", _wrap_Rig_set_rit, METH_VARARGS, NULL},
	 { "Rig_set_xit", _wrap_Rig_set_xit, METH_VARARGS, NULL},
	 { "Rig_set_ts", _wrap_Rig_set_ts, METH_VARARGS, NULL},
	 { "Rig_set_ant", _wrap_Rig_set_ant, METH_VARARGS, NULL},
	 { "Rig_set_func", _wrap_Rig_set_func, METH_VARARGS, NULL},
	 { "Rig_set_ext_func", _wrap_Rig_set_ext_func, METH_VARARGS, NULL},
	 { "Rig_set_bank", _wrap_Rig_set_bank, METH_VARARGS, NULL},
	 { "Rig_set_mem", _wrap_Rig_set_mem, METH_VARARGS, NULL},
	 { "Rig_send_dtmf", _wrap_Rig_send_dtmf, METH_VARARGS, NULL},
	 { "Rig_send_morse", _wrap_Rig_send_morse, METH_VARARGS, NULL},
	 { "Rig_vfo_op", _wrap_Rig_vfo_op, METH_VARARGS, NULL},
	 { "Rig_scan", _wrap_Rig_scan, METH_VARARGS, NULL},
	 { "Rig_set_level", _wrap_Rig_set_level, METH_VARARGS, NULL},
	 { "Rig_set_ext_level", _wrap_Rig_set_ext_level, METH_VARARGS, NULL},
	 { "Rig_get_level_i", _wrap_Rig_get_level_i, METH_VARARGS, NULL},
	 { "Rig_get_level_f", _wrap_Rig_get_level_f, METH_VARARGS, NULL},
	 { "Rig_set_vfo", _wrap_Rig_set_vfo, METH_VARARGS, NULL},
	 { "Rig_set_powerstat", _wrap_Rig_set_powerstat, METH_VARARGS, NULL},
	 { "Rig_has_get_level", _wrap_Rig_has_get_level, METH_VARARGS, NULL},
	 { "Rig_has_get_parm", _wrap_Rig_has_get_parm, METH_VARARGS, NULL},
	 { "Rig_has_set_parm", _wrap_Rig_has_set_parm, METH_VARARGS, NULL},
	 { "Rig_has_get_func", _wrap_Rig_has_get_func, METH_VARARGS, NULL},
	 { "Rig_has_set_func", _wrap_Rig_has_set_func, METH_VARARGS, NULL},
	 { "Rig_reset", _wrap_Rig_reset, METH_VARARGS, NULL},
	 { "Rig_has_scan", _wrap_Rig_has_scan, METH_VARARGS, NULL},
	 { "Rig_has_vfo_op", _wrap_Rig_has_vfo_op, METH_VARARGS, NULL},
	 { "Rig_passband_normal", _wrap_Rig_passband_normal, METH_VARARGS, NULL},
	 { "Rig_passband_narrow", _wrap_Rig_passband_narrow, METH_VARARGS, NULL},
	 { "Rig_passband_wide", _wrap_Rig_passband_wide, METH_VARARGS, NULL},
	 { "Rig_set_vfo_opt", _wrap_Rig_set_vfo_opt, METH_VARARGS, NULL},
	 { "Rig_ext_token_lookup", _wrap_Rig_ext_token_lookup, METH_VARARGS, NULL},
	 { "Rig_token_lookup", _wrap_Rig_token_lookup, METH_VARARGS, NULL},
	 { "Rig_set_ext_parm", _wrap_Rig_set_ext_parm, METH_VARARGS, NULL},
	 { "Rig_set_parm", _wrap_Rig_set_parm, METH_VARARGS, NULL},
	 { "Rig_get_parm_i", _wrap_Rig_get_parm_i, METH_VARARGS, NULL},
	 { "Rig_get_parm_f", _wrap_Rig_get_parm_f, METH_VARARGS, NULL},
	 { "Rig_set_conf", _wrap_Rig_set_conf, METH_VARARGS, NULL},
	 { "Rig_get_freq", _wrap_Rig_get_freq, METH_VARARGS, NULL},
	 { "Rig_get_mode", _wrap_Rig_get_mode, METH_VARARGS, NULL},
	 { "Rig_get_split_mode", _wrap_Rig_get_split_mode, METH_VARARGS, NULL},
	 { "Rig_get_vfo", _wrap_Rig_get_vfo, METH_O, NULL},
	 { "Rig_get_ptt", _wrap_Rig_get_ptt, METH_VARARGS, NULL},
	 { "Rig_get_rptr_shift", _wrap_Rig_get_rptr_shift, METH_VARARGS, NULL},
	 { "Rig_get_rptr_offs", _wrap_Rig_get_rptr_offs, METH_VARARGS, NULL},
	 { "Rig_get_ctcss_tone", _wrap_Rig_get_ctcss_tone, METH_VARARGS, NULL},
	 { "Rig_get_dcs_code", _wrap_Rig_get_dcs_code, METH_VARARGS, NULL},
	 { "Rig_get_ctcss_sql", _wrap_Rig_get_ctcss_sql, METH_VARARGS, NULL},
	 { "Rig_get_dcs_sql", _wrap_Rig_get_dcs_sql, METH_VARARGS, NULL},
	 { "Rig_get_split_freq", _wrap_Rig_get_split_freq, METH_VARARGS, NULL},
	 { "Rig_get_split_vfo", _wrap_Rig_get_split_vfo, METH_VARARGS, NULL},
	 { "Rig_get_rit", _wrap_Rig_get_rit, METH_VARARGS, NULL},
	 { "Rig_get_xit", _wrap_Rig_get_xit, METH_VARARGS, NULL},
	 { "Rig_get_ts", _wrap_Rig_get_ts, METH_VARARGS, NULL},
	 { "Rig_get_ant", _wrap_Rig_get_ant, METH_VARARGS, NULL},
	 { "Rig_get_vfo_info", _wrap_Rig_get_vfo_info, METH_VARARGS, NULL},
	 { "Rig_get_mem", _wrap_Rig_get_mem, METH_VARARGS, NULL},
	 { "Rig_get_powerstat", _wrap_Rig_get_powerstat, METH_O, NULL},
	 { "Rig_get_dcd", _wrap_Rig_get_dcd, METH_VARARGS, NULL},
	 { "Rig_set_freq_callback", _wrap_Rig_set_freq_callback, METH_VARARGS, NULL},
	 { "Rig_set_mode_callback", _wrap_Rig_set_mode_callback, METH_VARARGS, NULL},
	 { "Rig_set_vfo_callback", _wrap_Rig_set_vfo_callback, METH_VARARGS, NULL},
	 { "Rig_set_ptt_callback", _wrap_Rig_set_ptt_callback, METH_VARARGS, NULL},
	 { "Rig_set_dcd_callback", _wrap_Rig_set_dcd_callback, METH_VARARGS, NULL},
	 { "Rig_set_pltune_callback", _wrap_Rig_set_pltune_callback, METH_VARARGS, NULL},
	 { "Rig_set_spectrum_callback", _wrap_Rig_set_spectrum_callback, METH_VARARGS, NULL},
	 { "Rig_mem_count", _wrap_Rig_mem_count, METH_O, NULL},
	 { "Rig_lookup_mem_caps", _wrap_Rig_lookup_mem_caps, METH_VARARGS, NULL},
	 { "Rig_set_channel", _wrap_Rig_set_channel, METH_VARARGS, NULL},
	 { "Rig_get_chan_all", _wrap_Rig_get_chan_all, METH_VARARGS, NULL},
	 { "Rig_get_channel", _wrap_Rig_get_channel, METH_VARARGS, NULL},
	 { "Rig_get_conf", _wrap_Rig_get_conf, METH_VARARGS, NULL},
	 { "Rig_recv_dtmf", _wrap_Rig_recv_dtmf, METH_VARARGS, NULL},
	 { "Rig_get_info", _wrap_Rig_get_info, METH_O, NULL},
	 { "Rig_get_func", _wrap_Rig_get_func, METH_VARARGS, NULL},
	 { "Rig_get_ext_func", _wrap_Rig_get_ext_func, METH_VARARGS, NULL},
	 { "Rig_send_raw", _wrap_Rig_send_raw, METH_VARARGS, NULL},
	 { "Rig_get_level", _wrap_Rig_get_level, METH_VARARGS, NULL},
	 { "Rig_get_ext_level", _wrap_Rig_get_ext_level, METH_VARARGS, NULL},
	 { "Rig_get_parm", _wrap_Rig_get_parm, METH_VARARGS, NULL},
	 { "Rig_get_ext_parm", _wrap_Rig_get_ext_parm, METH_VARARGS, NULL},
	 { "Rig_swigregister", Rig_swigregister, METH_O, NULL},
	 { "Rig_swiginit", Rig_swiginit, METH_VARARGS, NULL},
	 { "new_channelArray", _wrap_new_channelArray, METH_O, NULL},
	 { "delete_channelArray", _wrap_delete_channelArray, METH_O, NULL},
	 { "channelArray_cast", _wrap_channelArray_cast, METH_O, NULL},
	 { "channelArray_frompointer", _wrap_channelArray_frompointer, METH_O, NULL},
	 { "channelArray_swigregister", channelArray_swigregister, METH_O, NULL},
	 { "channelArray_swiginit", channelArray_swiginit, METH_VARARGS, NULL},
	 { "new_toneArray", _wrap_new_toneArray, METH_O, NULL},
	 { "delete_toneArray", _wrap_delete_toneArray, METH_O, NULL},
	 { "toneArray_cast", _wrap_toneArray_cast, METH_O, NULL},
	 { "toneArray_frompointer", _wrap_toneArray_frompointer, METH_O, NULL},
	 { "toneArray_swigregister", toneArray_swigregister, METH_O, NULL},
	 { "toneArray_swiginit", toneArray_swiginit, METH_VARARGS, NULL},
	 { "rot_caps_rot_model_set", _wrap_rot_caps_rot_model_set, METH_VARARGS, NULL},
	 { "rot_caps_rot_model_get", _wrap_rot_caps_rot_model_get, METH_O, NULL},
	 { "rot_caps_model_name_get", _wrap_rot_caps_model_name_get, METH_O, NULL},
	 { "rot_caps_mfg_name_get", _wrap_rot_caps_mfg_name_get, METH_O, NULL},
	 { "rot_caps_version_get", _wrap_rot_caps_version_get, METH_O, NULL},
	 { "rot_caps_copyright_get", _wrap_rot_caps_copyright_get, METH_O, NULL},
	 { "rot_caps_status_set", _wrap_rot_caps_status_set, METH_VARARGS, NULL},
	 { "rot_caps_status_get", _wrap_rot_caps_status_get, METH_O, NULL},
	 { "rot_caps_rot_type_set", _wrap_rot_caps_rot_type_set, METH_VARARGS, NULL},
	 { "rot_caps_rot_type_get", _wrap_rot_caps_rot_type_get, METH_O, NULL},
	 { "rot_caps_port_type_set", _wrap_rot_caps_port_type_set, METH_VARARGS, NULL},
	 { "rot_caps_port_type_get", _wrap_rot_caps_port_type_get, METH_O, NULL},
	 { "rot_caps_serial_rate_min_set", _wrap_rot_caps_serial_rate_min_set, METH_VARARGS, NULL},
	 { "rot_caps_serial_rate_min_get", _wrap_rot_caps_serial_rate_min_get, METH_O, NULL},
	 { "rot_caps_serial_rate_max_set", _wrap_rot_caps_serial_rate_max_set, METH_VARARGS, NULL},
	 { "rot_caps_serial_rate_max_get", _wrap_rot_caps_serial_rate_max_get, METH_O, NULL},
	 { "rot_caps_serial_data_bits_set", _wrap_rot_caps_serial_data_bits_set, METH_VARARGS, NULL},
	 { "rot_caps_serial_data_bits_get", _wrap_rot_caps_serial_data_bits_get, METH_O, NULL},
	 { "rot_caps_serial_stop_bits_set", _wrap_rot_caps_serial_stop_bits_set, METH_VARARGS, NULL},
	 { "rot_caps_serial_stop_bits_get", _wrap_rot_caps_serial_stop_bits_get, METH_O, NULL},
	 { "rot_caps_serial_parity_set", _wrap_rot_caps_serial_parity_set, METH_VARARGS, NULL},
	 { "rot_caps_serial_parity_get", _wrap_rot_caps_serial_parity_get, METH_O, NULL},
	 { "rot_caps_serial_handshake_set", _wrap_rot_caps_serial_handshake_set, METH_VARARGS, NULL},
	 { "rot_caps_serial_handshake_get", _wrap_rot_caps_serial_handshake_get, METH_O, NULL},
	 { "rot_caps_write_delay_set", _wrap_rot_caps_write_delay_set, METH_VARARGS, NULL},
	 { "rot_caps_write_delay_get", _wrap_rot_caps_write_delay_get, METH_O, NULL},
	 { "rot_caps_post_write_delay_set", _wrap_rot_caps_post_write_delay_set, METH_VARARGS, NULL},
	 { "rot_caps_post_write_delay_get", _wrap_rot_caps_post_write_delay_get, METH_O, NULL},
	 { "rot_caps_timeout_set", _wrap_rot_caps_timeout_set, METH_VARARGS, NULL},
	 { "rot_caps_timeout_get", _wrap_rot_caps_timeout_get, METH_O, NULL},
	 { "rot_caps_retry_set", _wrap_rot_caps_retry_set, METH_VARARGS, NULL},
	 { "rot_caps_retry_get", _wrap_rot_caps_retry_get, METH_O, NULL},
	 { "rot_caps_has_get_func_set", _wrap_rot_caps_has_get_func_set, METH_VARARGS, NULL},
	 { "rot_caps_has_get_func_get", _wrap_rot_caps_has_get_func_get, METH_O, NULL},
	 { "rot_caps_has_set_func_set", _wrap_rot_caps_has_set_func_set, METH_VARARGS, NULL},
	 { "rot_caps_has_set_func_get", _wrap_rot_caps_has_set_func_get, METH_O, NULL},
	 { "rot_caps_has_get_level_set", _wrap_rot_caps_has_get_level_set, METH_VARARGS, NULL},
	 { "rot_caps_has_get_level_get", _wrap_rot_caps_has_get_level_get, METH_O, NULL},
	 { "rot_caps_has_set_level_set", _wrap_rot_caps_has_set_level_set, METH_VARARGS, NULL},
	 { "rot_caps_has_set_level_get", _wrap_rot_caps_has_set_level_get, METH_O, NULL},
	 { "rot_caps_has_get_parm_set", _wrap_rot_caps_has_get_parm_set, METH_VARARGS, NULL},
	 { "rot_caps_has_get_parm_get", _wrap_rot_caps_has_get_parm_get, METH_O, NULL},
	 { "rot_caps_has_set_parm_set", _wrap_rot_caps_has_set_parm_set, METH_VARARGS, NULL},
	 { "rot_caps_has_set_parm_get", _wrap_rot_caps_has_set_parm_get, METH_O, NULL},
	 { "rot_caps_has_status_set", _wrap_rot_caps_has_status_set, METH_VARARGS, NULL},
	 { "rot_caps_has_status_get", _wrap_rot_caps_has_status_get, METH_O, NULL},
	 { "rot_caps_level_gran_set", _wrap_rot_caps_level_gran_set, METH_VARARGS, NULL},
	 { "rot_caps_level_gran_get", _wrap_rot_caps_level_gran_get, METH_O, NULL},
	 { "rot_caps_parm_gran_set", _wrap_rot_caps_parm_gran_set, METH_VARARGS, NULL},
	 { "rot_caps_parm_gran_get", _wrap_rot_caps_parm_gran_get, METH_O, NULL},
	 { "rot_caps_extparms_set", _wrap_rot_caps_extparms_set, METH_VARARGS, NULL},
	 { "rot_caps_extparms_get", _wrap_rot_caps_extparms_get, METH_O, NULL},
	 { "rot_caps_extlevels_set", _wrap_rot_caps_extlevels_set, METH_VARARGS, NULL},
	 { "rot_caps_extlevels_get", _wrap_rot_caps_extlevels_get, METH_O, NULL},
	 { "rot_caps_extfuncs_set", _wrap_rot_caps_extfuncs_set, METH_VARARGS, NULL},
	 { "rot_caps_extfuncs_get", _wrap_rot_caps_extfuncs_get, METH_O, NULL},
	 { "rot_caps_ext_tokens_set", _wrap_rot_caps_ext_tokens_set, METH_O, NULL},
	 { "rot_caps_ext_tokens_get", _wrap_rot_caps_ext_tokens_get, METH_O, NULL},
	 { "rot_caps_min_az_set", _wrap_rot_caps_min_az_set, METH_VARARGS, NULL},
	 { "rot_caps_min_az_get", _wrap_rot_caps_min_az_get, METH_O, NULL},
	 { "rot_caps_max_az_set", _wrap_rot_caps_max_az_set, METH_VARARGS, NULL},
	 { "rot_caps_max_az_get", _wrap_rot_caps_max_az_get, METH_O, NULL},
	 { "rot_caps_min_el_set", _wrap_rot_caps_min_el_set, METH_VARARGS, NULL},
	 { "rot_caps_min_el_get", _wrap_rot_caps_min_el_get, METH_O, NULL},
	 { "rot_caps_max_el_set", _wrap_rot_caps_max_el_set, METH_VARARGS, NULL},
	 { "rot_caps_max_el_get", _wrap_rot_caps_max_el_get, METH_O, NULL},
	 { "rot_caps_cfgparams_set", _wrap_rot_caps_cfgparams_set, METH_VARARGS, NULL},
	 { "rot_caps_cfgparams_get", _wrap_rot_caps_cfgparams_get, METH_O, NULL},
	 { "rot_caps_priv_set", _wrap_rot_caps_priv_set, METH_VARARGS, NULL},
	 { "rot_caps_priv_get", _wrap_rot_caps_priv_get, METH_O, NULL},
	 { "rot_caps_rot_init_set", _wrap_rot_caps_rot_init_set, METH_VARARGS, NULL},
	 { "rot_caps_rot_init_get", _wrap_rot_caps_rot_init_get, METH_O, NULL},
	 { "rot_caps_rot_cleanup_set", _wrap_rot_caps_rot_cleanup_set, METH_VARARGS, NULL},
	 { "rot_caps_rot_cleanup_get", _wrap_rot_caps_rot_cleanup_get, METH_O, NULL},
	 { "rot_caps_rot_open_set", _wrap_rot_caps_rot_open_set, METH_VARARGS, NULL},
	 { "rot_caps_rot_open_get", _wrap_rot_caps_rot_open_get, METH_O, NULL},
	 { "rot_caps_rot_close_set", _wrap_rot_caps_rot_close_set, METH_VARARGS, NULL},
	 { "rot_caps_rot_close_get", _wrap_rot_caps_rot_close_get, METH_O, NULL},
	 { "rot_caps_set_conf_set", _wrap_rot_caps_set_conf_set, METH_VARARGS, NULL},
	 { "rot_caps_set_conf_get", _wrap_rot_caps_set_conf_get, METH_O, NULL},
	 { "rot_caps_get_conf_set", _wrap_rot_caps_get_conf_set, METH_VARARGS, NULL},
	 { "rot_caps_get_conf_get", _wrap_rot_caps_get_conf_get, METH_O, NULL},
	 { "rot_caps_set_position_set", _wrap_rot_caps_set_position_set, METH_VARARGS, NULL},
	 { "rot_caps_set_position_get", _wrap_rot_caps_set_position_get, METH_O, NULL},
	 { "rot_caps_get_position_set", _wrap_rot_caps_get_position_set, METH_VARARGS, NULL},
	 { "rot_caps_get_position_get", _wrap_rot_caps_get_position_get, METH_O, NULL},
	 { "rot_caps_stop_set", _wrap_rot_caps_stop_set, METH_VARARGS, NULL},
	 { "rot_caps_stop_get", _wrap_rot_caps_stop_get, METH_O, NULL},
	 { "rot_caps_park_set", _wrap_rot_caps_park_set, METH_VARARGS, NULL},
	 { "rot_caps_park_get", _wrap_rot_caps_park_get, METH_O, NULL},
	 { "rot_caps_reset_set", _wrap_rot_caps_reset_set, METH_VARARGS, NULL},
	 { "rot_caps_reset_get", _wrap_rot_caps_reset_get, METH_O, NULL},
	 { "rot_caps_move_set", _wrap_rot_caps_move_set, METH_VARARGS, NULL},
	 { "rot_caps_move_get", _wrap_rot_caps_move_get, METH_O, NULL},
	 { "rot_caps_get_info_set", _wrap_rot_caps_get_info_set, METH_VARARGS, NULL},
	 { "rot_caps_get_info_get", _wrap_rot_caps_get_info_get, METH_O, NULL},
	 { "rot_caps_set_level_set", _wrap_rot_caps_set_level_set, METH_VARARGS, NULL},
	 { "rot_caps_set_level_get", _wrap_rot_caps_set_level_get, METH_O, NULL},
	 { "rot_caps_get_level_set", _wrap_rot_caps_get_level_set, METH_VARARGS, NULL},
	 { "rot_caps_get_level_get", _wrap_rot_caps_get_level_get, METH_O, NULL},
	 { "rot_caps_set_func_set", _wrap_rot_caps_set_func_set, METH_VARARGS, NULL},
	 { "rot_caps_set_func_get", _wrap_rot_caps_set_func_get, METH_O, NULL},
	 { "rot_caps_get_func_set", _wrap_rot_caps_get_func_set, METH_VARARGS, NULL},
	 { "rot_caps_get_func_get", _wrap_rot_caps_get_func_get, METH_O, NULL},
	 { "rot_caps_set_parm_set", _wrap_rot_caps_set_parm_set, METH_VARARGS, NULL},
	 { "rot_caps_set_parm_get", _wrap_rot_caps_set_parm_get, METH_O, NULL},
	 { "rot_caps_get_parm_set", _wrap_rot_caps_get_parm_set, METH_VARARGS, NULL},
	 { "rot_caps_get_parm_get", _wrap_rot_caps_get_parm_get, METH_O, NULL},
	 { "rot_caps_set_ext_level_set", _wrap_rot_caps_set_ext_level_set, METH_VARARGS, NULL},
	 { "rot_caps_set_ext_level_get", _wrap_rot_caps_set_ext_level_get, METH_O, NULL},
	 { "rot_caps_get_ext_level_set", _wrap_rot_caps_get_ext_level_set, METH_VARARGS, NULL},
	 { "rot_caps_get_ext_level_get", _wrap_rot_caps_get_ext_level_get, METH_O, NULL},
	 { "rot_caps_set_ext_func_set", _wrap_rot_caps_set_ext_func_set, METH_VARARGS, NULL},
	 { "rot_caps_set_ext_func_get", _wrap_rot_caps_set_ext_func_get, METH_O, NULL},
	 { "rot_caps_get_ext_func_set", _wrap_rot_caps_get_ext_func_set, METH_VARARGS, NULL},
	 { "rot_caps_get_ext_func_get", _wrap_rot_caps_get_ext_func_get, METH_O, NULL},
	 { "rot_caps_set_ext_parm_set", _wrap_rot_caps_set_ext_parm_set, METH_VARARGS, NULL},
	 { "rot_caps_set_ext_parm_get", _wrap_rot_caps_set_ext_parm_get, METH_O, NULL},
	 { "rot_caps_get_ext_parm_set", _wrap_rot_caps_get_ext_parm_set, METH_VARARGS, NULL},
	 { "rot_caps_get_ext_parm_get", _wrap_rot_caps_get_ext_parm_get, METH_O, NULL},
	 { "rot_caps_get_status_set", _wrap_rot_caps_get_status_set, METH_VARARGS, NULL},
	 { "rot_caps_get_status_get", _wrap_rot_caps_get_status_get, METH_O, NULL},
	 { "rot_caps_get_conf2_set", _wrap_rot_caps_get_conf2_set, METH_VARARGS, NULL},
	 { "rot_caps_get_conf2_get", _wrap_rot_caps_get_conf2_get, METH_O, NULL},
	 { "new_rot_caps", _wrap_new_rot_caps, METH_NOARGS, NULL},
	 { "delete_rot_caps", _wrap_delete_rot_caps, METH_O, NULL},
	 { "rot_caps_swigregister", rot_caps_swigregister, METH_O, NULL},
	 { "rot_caps_swiginit", rot_caps_swiginit, METH_VARARGS, NULL},
	 { "qrb", _wrap_qrb, METH_VARARGS, NULL},
	 { "distance_long_path", _wrap_distance_long_path, METH_O, NULL},
	 { "azimuth_long_path", _wrap_azimuth_long_path, METH_O, NULL},
	 { "dms2dec", _wrap_dms2dec, METH_VARARGS, NULL},
	 { "dec2dms", _wrap_dec2dms, METH_O, NULL},
	 { "dec2dmmm", _wrap_dec2dmmm, METH_O, NULL},
	 { "dmmm2dec", _wrap_dmmm2dec, METH_VARARGS, NULL},
	 { "rot_state_min_az_set", _wrap_rot_state_min_az_set, METH_VARARGS, NULL},
	 { "rot_state_min_az_get", _wrap_rot_state_min_az_get, METH_O, NULL},
	 { "rot_state_max_az_set", _wrap_rot_state_max_az_set, METH_VARARGS, NULL},
	 { "rot_state_max_az_get", _wrap_rot_state_max_az_get, METH_O, NULL},
	 { "rot_state_min_el_set", _wrap_rot_state_min_el_set, METH_VARARGS, NULL},
	 { "rot_state_min_el_get", _wrap_rot_state_min_el_get, METH_O, NULL},
	 { "rot_state_max_el_set", _wrap_rot_state_max_el_set, METH_VARARGS, NULL},
	 { "rot_state_max_el_get", _wrap_rot_state_max_el_get, METH_O, NULL},
	 { "rot_state_south_zero_set", _wrap_rot_state_south_zero_set, METH_VARARGS, NULL},
	 { "rot_state_south_zero_get", _wrap_rot_state_south_zero_get, METH_O, NULL},
	 { "rot_state_az_offset_set", _wrap_rot_state_az_offset_set, METH_VARARGS, NULL},
	 { "rot_state_az_offset_get", _wrap_rot_state_az_offset_get, METH_O, NULL},
	 { "rot_state_el_offset_set", _wrap_rot_state_el_offset_set, METH_VARARGS, NULL},
	 { "rot_state_el_offset_get", _wrap_rot_state_el_offset_get, METH_O, NULL},
	 { "rot_state_has_get_func_set", _wrap_rot_state_has_get_func_set, METH_VARARGS, NULL},
	 { "rot_state_has_get_func_get", _wrap_rot_state_has_get_func_get, METH_O, NULL},
	 { "rot_state_has_set_func_set", _wrap_rot_state_has_set_func_set, METH_VARARGS, NULL},
	 { "rot_state_has_set_func_get", _wrap_rot_state_has_set_func_get, METH_O, NULL},
	 { "rot_state_has_get_level_set", _wrap_rot_state_has_get_level_set, METH_VARARGS, NULL},
	 { "rot_state_has_get_level_get", _wrap_rot_state_has_get_level_get, METH_O, NULL},
	 { "rot_state_has_set_level_set", _wrap_rot_state_has_set_level_set, METH_VARARGS, NULL},
	 { "rot_state_has_set_level_get", _wrap_rot_state_has_set_level_get, METH_O, NULL},
	 { "rot_state_has_get_parm_set", _wrap_rot_state_has_get_parm_set, METH_VARARGS, NULL},
	 { "rot_state_has_get_parm_get", _wrap_rot_state_has_get_parm_get, METH_O, NULL},
	 { "rot_state_has_set_parm_set", _wrap_rot_state_has_set_parm_set, METH_VARARGS, NULL},
	 { "rot_state_has_set_parm_get", _wrap_rot_state_has_set_parm_get, METH_O, NULL},
	 { "rot_state_has_status_set", _wrap_rot_state_has_status_set, METH_VARARGS, NULL},
	 { "rot_state_has_status_get", _wrap_rot_state_has_status_get, METH_O, NULL},
	 { "rot_state_level_gran_set", _wrap_rot_state_level_gran_set, METH_VARARGS, NULL},
	 { "rot_state_level_gran_get", _wrap_rot_state_level_gran_get, METH_O, NULL},
	 { "rot_state_parm_gran_set", _wrap_rot_state_parm_gran_set, METH_VARARGS, NULL},
	 { "rot_state_parm_gran_get", _wrap_rot_state_parm_gran_get, METH_O, NULL},
	 { "rot_state_comm_state_set", _wrap_rot_state_comm_state_set, METH_VARARGS, NULL},
	 { "rot_state_comm_state_get", _wrap_rot_state_comm_state_get, METH_O, NULL},
	 { "rot_state_priv_set", _wrap_rot_state_priv_set, METH_VARARGS, NULL},
	 { "rot_state_priv_get", _wrap_rot_state_priv_get, METH_O, NULL},
	 { "rot_state_obj_set", _wrap_rot_state_obj_set, METH_VARARGS, NULL},
	 { "rot_state_obj_get", _wrap_rot_state_obj_get, METH_O, NULL},
	 { "rot_state_current_speed_set", _wrap_rot_state_current_speed_set, METH_VARARGS, NULL},
	 { "rot_state_current_speed_get", _wrap_rot_state_current_speed_get, METH_O, NULL},
	 { "rot_state_rotport_set", _wrap_rot_state_rotport_set, METH_VARARGS, NULL},
	 { "rot_state_rotport_get", _wrap_rot_state_rotport_get, METH_O, NULL},
	 { "rot_state_rotport2_set", _wrap_rot_state_rotport2_set, METH_VARARGS, NULL},
	 { "rot_state_rotport2_get", _wrap_rot_state_rotport2_get, METH_O, NULL},
	 { "rot_state_pstrotator_handler_priv_data_set", _wrap_rot_state_pstrotator_handler_priv_data_set, METH_VARARGS, NULL},
	 { "rot_state_pstrotator_handler_priv_data_get", _wrap_rot_state_pstrotator_handler_priv_data_get, METH_O, NULL},
	 { "rot_state_config_queue_set", _wrap_rot_state_config_queue_set, METH_VARARGS, NULL},
	 { "rot_state_config_queue_get", _wrap_rot_state_config_queue_get, METH_O, NULL},
	 { "new_rot_state", _wrap_new_rot_state, METH_NOARGS, NULL},
	 { "delete_rot_state", _wrap_delete_rot_state, METH_O, NULL},
	 { "rot_state_swigregister", rot_state_swigregister, METH_O, NULL},
	 { "rot_state_swiginit", rot_state_swiginit, METH_VARARGS, NULL},
	 { "Rot_rot_set", _wrap_Rot_rot_set, METH_VARARGS, NULL},
	 { "Rot_rot_get", _wrap_Rot_rot_get, METH_O, NULL},
	 { "Rot_caps_set", _wrap_Rot_caps_set, METH_VARARGS, NULL},
	 { "Rot_caps_get", _wrap_Rot_caps_get, METH_O, NULL},
	 { "Rot_state_set", _wrap_Rot_state_set, METH_VARARGS, NULL},
	 { "Rot_state_get", _wrap_Rot_state_get, METH_O, NULL},
	 { "Rot_error_status_set", _wrap_Rot_error_status_set, METH_VARARGS, NULL},
	 { "Rot_error_status_get", _wrap_Rot_error_status_get, METH_O, NULL},
	 { "Rot_do_exception_set", _wrap_Rot_do_exception_set, METH_VARARGS, NULL},
	 { "Rot_do_exception_get", _wrap_Rot_do_exception_get, METH_O, NULL},
	 { "new_Rot", _wrap_new_Rot, METH_O, NULL},
	 { "delete_Rot", _wrap_delete_Rot, METH_O, NULL},
	 { "Rot_open", _wrap_Rot_open, METH_O, NULL},
	 { "Rot_close", _wrap_Rot_close, METH_O, NULL},
	 { "Rot_set_ext_func", _wrap_Rot_set_ext_func, METH_VARARGS, NULL},
	 { "Rot_set_ext_level", _wrap_Rot_set_ext_level, METH_VARARGS, NULL},
	 { "Rot_set_ext_parm", _wrap_Rot_set_ext_parm, METH_VARARGS, NULL},
	 { "Rot_set_func", _wrap_Rot_set_func, METH_VARARGS, NULL},
	 { "Rot_set_level", _wrap_Rot_set_level, METH_VARARGS, NULL},
	 { "Rot_set_parm", _wrap_Rot_set_parm, METH_VARARGS, NULL},
	 { "Rot_set_position", _wrap_Rot_set_position, METH_VARARGS, NULL},
	 { "Rot_get_position", _wrap_Rot_get_position, METH_O, NULL},
	 { "Rot_stop", _wrap_Rot_stop, METH_O, NULL},
	 { "Rot_park", _wrap_Rot_park, METH_O, NULL},
	 { "Rot_reset", _wrap_Rot_reset, METH_VARARGS, NULL},
	 { "Rot_move", _wrap_Rot_move, METH_VARARGS, NULL},
	 { "Rot_token_lookup", _wrap_Rot_token_lookup, METH_VARARGS, NULL},
	 { "Rot_set_conf", _wrap_Rot_set_conf, METH_VARARGS, NULL},
	 { "Rot_has_get_func", _wrap_Rot_has_get_func, METH_VARARGS, NULL},
	 { "Rot_has_set_func", _wrap_Rot_has_set_func, METH_VARARGS, NULL},
	 { "Rot_has_get_level", _wrap_Rot_has_get_level, METH_VARARGS, NULL},
	 { "Rot_has_set_level", _wrap_Rot_has_set_level, METH_VARARGS, NULL},
	 { "Rot_has_get_parm", _wrap_Rot_has_get_parm, METH_VARARGS, NULL},
	 { "Rot_has_set_parm", _wrap_Rot_has_set_parm, METH_VARARGS, NULL},
	 { "Rot_get_func", _wrap_Rot_get_func, METH_VARARGS, NULL},
	 { "Rot_get_ext_func", _wrap_Rot_get_ext_func, METH_VARARGS, NULL},
	 { "Rot_get_level", _wrap_Rot_get_level, METH_VARARGS, NULL},
	 { "Rot_get_ext_level", _wrap_Rot_get_ext_level, METH_VARARGS, NULL},
	 { "Rot_get_parm", _wrap_Rot_get_parm, METH_VARARGS, NULL},
	 { "Rot_get_ext_parm", _wrap_Rot_get_ext_parm, METH_VARARGS, NULL},
	 { "Rot_get_conf", _wrap_Rot_get_conf, METH_VARARGS, NULL},
	 { "Rot_get_info", _wrap_Rot_get_info, METH_O, NULL},
	 { "Rot_swigregister", Rot_swigregister, METH_O, NULL},
	 { "Rot_swiginit", Rot_swiginit, METH_VARARGS, NULL},
	 { "amp_caps_amp_model_set", _wrap_amp_caps_amp_model_set, METH_VARARGS, NULL},
	 { "amp_caps_amp_model_get", _wrap_amp_caps_amp_model_get, METH_O, NULL},
	 { "amp_caps_model_name_get", _wrap_amp_caps_model_name_get, METH_O, NULL},
	 { "amp_caps_mfg_name_get", _wrap_amp_caps_mfg_name_get, METH_O, NULL},
	 { "amp_caps_version_get", _wrap_amp_caps_version_get, METH_O, NULL},
	 { "amp_caps_copyright_get", _wrap_amp_caps_copyright_get, METH_O, NULL},
	 { "amp_caps_status_set", _wrap_amp_caps_status_set, METH_VARARGS, NULL},
	 { "amp_caps_status_get", _wrap_amp_caps_status_get, METH_O, NULL},
	 { "amp_caps_amp_type_set", _wrap_amp_caps_amp_type_set, METH_VARARGS, NULL},
	 { "amp_caps_amp_type_get", _wrap_amp_caps_amp_type_get, METH_O, NULL},
	 { "amp_caps_port_type_set", _wrap_amp_caps_port_type_set, METH_VARARGS, NULL},
	 { "amp_caps_port_type_get", _wrap_amp_caps_port_type_get, METH_O, NULL},
	 { "amp_caps_serial_rate_min_set", _wrap_amp_caps_serial_rate_min_set, METH_VARARGS, NULL},
	 { "amp_caps_serial_rate_min_get", _wrap_amp_caps_serial_rate_min_get, METH_O, NULL},
	 { "amp_caps_serial_rate_max_set", _wrap_amp_caps_serial_rate_max_set, METH_VARARGS, NULL},
	 { "amp_caps_serial_rate_max_get", _wrap_amp_caps_serial_rate_max_get, METH_O, NULL},
	 { "amp_caps_serial_data_bits_set", _wrap_amp_caps_serial_data_bits_set, METH_VARARGS, NULL},
	 { "amp_caps_serial_data_bits_get", _wrap_amp_caps_serial_data_bits_get, METH_O, NULL},
	 { "amp_caps_serial_stop_bits_set", _wrap_amp_caps_serial_stop_bits_set, METH_VARARGS, NULL},
	 { "amp_caps_serial_stop_bits_get", _wrap_amp_caps_serial_stop_bits_get, METH_O, NULL},
	 { "amp_caps_serial_parity_set", _wrap_amp_caps_serial_parity_set, METH_VARARGS, NULL},
	 { "amp_caps_serial_parity_get", _wrap_amp_caps_serial_parity_get, METH_O, NULL},
	 { "amp_caps_serial_handshake_set", _wrap_amp_caps_serial_handshake_set, METH_VARARGS, NULL},
	 { "amp_caps_serial_handshake_get", _wrap_amp_caps_serial_handshake_get, METH_O, NULL},
	 { "amp_caps_write_delay_set", _wrap_amp_caps_write_delay_set, METH_VARARGS, NULL},
	 { "amp_caps_write_delay_get", _wrap_amp_caps_write_delay_get, METH_O, NULL},
	 { "amp_caps_post_write_delay_set", _wrap_amp_caps_post_write_delay_set, METH_VARARGS, NULL},
	 { "amp_caps_post_write_delay_get", _wrap_amp_caps_post_write_delay_get, METH_O, NULL},
	 { "amp_caps_timeout_set", _wrap_amp_caps_timeout_set, METH_VARARGS, NULL},
	 { "amp_caps_timeout_get", _wrap_amp_caps_timeout_get, METH_O, NULL},
	 { "amp_caps_retry_set", _wrap_amp_caps_retry_set, METH_VARARGS, NULL},
	 { "amp_caps_retry_get", _wrap_amp_caps_retry_get, METH_O, NULL},
	 { "amp_caps_cfgparams_set", _wrap_amp_caps_cfgparams_set, METH_VARARGS, NULL},
	 { "amp_caps_cfgparams_get", _wrap_amp_caps_cfgparams_get, METH_O, NULL},
	 { "amp_caps_priv_set", _wrap_amp_caps_priv_set, METH_VARARGS, NULL},
	 { "amp_caps_priv_get", _wrap_amp_caps_priv_get, METH_O, NULL},
	 { "amp_caps_amp_model_macro_name_set", _wrap_amp_caps_amp_model_macro_name_set, METH_VARARGS, NULL},
	 { "amp_caps_amp_model_macro_name_get", _wrap_amp_caps_amp_model_macro_name_get, METH_O, NULL},
	 { "amp_caps_has_get_level_set", _wrap_amp_caps_has_get_level_set, METH_VARARGS, NULL},
	 { "amp_caps_has_get_level_get", _wrap_amp_caps_has_get_level_get, METH_O, NULL},
	 { "amp_caps_has_set_level_set", _wrap_amp_caps_has_set_level_set, METH_VARARGS, NULL},
	 { "amp_caps_has_set_level_get", _wrap_amp_caps_has_set_level_get, METH_O, NULL},
	 { "amp_caps_level_gran_set", _wrap_amp_caps_level_gran_set, METH_VARARGS, NULL},
	 { "amp_caps_level_gran_get", _wrap_amp_caps_level_gran_get, METH_O, NULL},
	 { "amp_caps_parm_gran_set", _wrap_amp_caps_parm_gran_set, METH_VARARGS, NULL},
	 { "amp_caps_parm_gran_get", _wrap_amp_caps_parm_gran_get, METH_O, NULL},
	 { "amp_caps_amp_init_set", _wrap_amp_caps_amp_init_set, METH_VARARGS, NULL},
	 { "amp_caps_amp_init_get", _wrap_amp_caps_amp_init_get, METH_O, NULL},
	 { "amp_caps_amp_cleanup_set", _wrap_amp_caps_amp_cleanup_set, METH_VARARGS, NULL},
	 { "amp_caps_amp_cleanup_get", _wrap_amp_caps_amp_cleanup_get, METH_O, NULL},
	 { "amp_caps_amp_open_set", _wrap_amp_caps_amp_open_set, METH_VARARGS, NULL},
	 { "amp_caps_amp_open_get", _wrap_amp_caps_amp_open_get, METH_O, NULL},
	 { "amp_caps_amp_close_set", _wrap_amp_caps_amp_close_set, METH_VARARGS, NULL},
	 { "amp_caps_amp_close_get", _wrap_amp_caps_amp_close_get, METH_O, NULL},
	 { "amp_caps_set_freq_set", _wrap_amp_caps_set_freq_set, METH_VARARGS, NULL},
	 { "amp_caps_set_freq_get", _wrap_amp_caps_set_freq_get, METH_O, NULL},
	 { "amp_caps_get_freq_set", _wrap_amp_caps_get_freq_set, METH_VARARGS, NULL},
	 { "amp_caps_get_freq_get", _wrap_amp_caps_get_freq_get, METH_O, NULL},
	 { "amp_caps_set_conf_set", _wrap_amp_caps_set_conf_set, METH_VARARGS, NULL},
	 { "amp_caps_set_conf_get", _wrap_amp_caps_set_conf_get, METH_O, NULL},
	 { "amp_caps_get_conf2_set", _wrap_amp_caps_get_conf2_set, METH_VARARGS, NULL},
	 { "amp_caps_get_conf2_get", _wrap_amp_caps_get_conf2_get, METH_O, NULL},
	 { "amp_caps_get_conf_set", _wrap_amp_caps_get_conf_set, METH_VARARGS, NULL},
	 { "amp_caps_get_conf_get", _wrap_amp_caps_get_conf_get, METH_O, NULL},
	 { "amp_caps_reset_set", _wrap_amp_caps_reset_set, METH_VARARGS, NULL},
	 { "amp_caps_reset_get", _wrap_amp_caps_reset_get, METH_O, NULL},
	 { "amp_caps_get_level_set", _wrap_amp_caps_get_level_set, METH_VARARGS, NULL},
	 { "amp_caps_get_level_get", _wrap_amp_caps_get_level_get, METH_O, NULL},
	 { "amp_caps_set_level_set", _wrap_amp_caps_set_level_set, METH_VARARGS, NULL},
	 { "amp_caps_set_level_get", _wrap_amp_caps_set_level_get, METH_O, NULL},
	 { "amp_caps_get_ext_level_set", _wrap_amp_caps_get_ext_level_set, METH_VARARGS, NULL},
	 { "amp_caps_get_ext_level_get", _wrap_amp_caps_get_ext_level_get, METH_O, NULL},
	 { "amp_caps_set_ext_level_set", _wrap_amp_caps_set_ext_level_set, METH_VARARGS, NULL},
	 { "amp_caps_set_ext_level_get", _wrap_amp_caps_set_ext_level_get, METH_O, NULL},
	 { "amp_caps_set_powerstat_set", _wrap_amp_caps_set_powerstat_set, METH_VARARGS, NULL},
	 { "amp_caps_set_powerstat_get", _wrap_amp_caps_set_powerstat_get, METH_O, NULL},
	 { "amp_caps_get_powerstat_set", _wrap_amp_caps_get_powerstat_set, METH_VARARGS, NULL},
	 { "amp_caps_get_powerstat_get", _wrap_amp_caps_get_powerstat_get, METH_O, NULL},
	 { "amp_caps_get_info_set", _wrap_amp_caps_get_info_set, METH_VARARGS, NULL},
	 { "amp_caps_get_info_get", _wrap_amp_caps_get_info_get, METH_O, NULL},
	 { "amp_caps_levels_set", _wrap_amp_caps_levels_set, METH_VARARGS, NULL},
	 { "amp_caps_levels_get", _wrap_amp_caps_levels_get, METH_O, NULL},
	 { "amp_caps_ext_levels_set", _wrap_amp_caps_ext_levels_set, METH_VARARGS, NULL},
	 { "amp_caps_ext_levels_get", _wrap_amp_caps_ext_levels_get, METH_O, NULL},
	 { "amp_caps_extlevels_set", _wrap_amp_caps_extlevels_set, METH_VARARGS, NULL},
	 { "amp_caps_extlevels_get", _wrap_amp_caps_extlevels_get, METH_O, NULL},
	 { "amp_caps_extparms_set", _wrap_amp_caps_extparms_set, METH_VARARGS, NULL},
	 { "amp_caps_extparms_get", _wrap_amp_caps_extparms_get, METH_O, NULL},
	 { "new_amp_caps", _wrap_new_amp_caps, METH_NOARGS, NULL},
	 { "delete_amp_caps", _wrap_delete_amp_caps, METH_O, NULL},
	 { "amp_caps_swigregister", amp_caps_swigregister, METH_O, NULL},
	 { "amp_caps_swiginit", amp_caps_swiginit, METH_VARARGS, NULL},
	 { "amp_state_comm_state_set", _wrap_amp_state_comm_state_set, METH_VARARGS, NULL},
	 { "amp_state_comm_state_get", _wrap_amp_state_comm_state_get, METH_O, NULL},
	 { "amp_state_priv_set", _wrap_amp_state_priv_set, METH_VARARGS, NULL},
	 { "amp_state_priv_get", _wrap_amp_state_priv_get, METH_O, NULL},
	 { "amp_state_obj_set", _wrap_amp_state_obj_set, METH_VARARGS, NULL},
	 { "amp_state_obj_get", _wrap_amp_state_obj_get, METH_O, NULL},
	 { "amp_state_has_get_level_set", _wrap_amp_state_has_get_level_set, METH_VARARGS, NULL},
	 { "amp_state_has_get_level_get", _wrap_amp_state_has_get_level_get, METH_O, NULL},
	 { "amp_state_has_set_level_set", _wrap_amp_state_has_set_level_set, METH_VARARGS, NULL},
	 { "amp_state_has_set_level_get", _wrap_amp_state_has_set_level_get, METH_O, NULL},
	 { "amp_state_level_gran_set", _wrap_amp_state_level_gran_set, METH_VARARGS, NULL},
	 { "amp_state_level_gran_get", _wrap_amp_state_level_gran_get, METH_O, NULL},
	 { "amp_state_parm_gran_set", _wrap_amp_state_parm_gran_set, METH_VARARGS, NULL},
	 { "amp_state_parm_gran_get", _wrap_amp_state_parm_gran_get, METH_O, NULL},
	 { "amp_state_ampport_set", _wrap_amp_state_ampport_set, METH_VARARGS, NULL},
	 { "amp_state_ampport_get", _wrap_amp_state_ampport_get, METH_O, NULL},
	 { "new_amp_state", _wrap_new_amp_state, METH_NOARGS, NULL},
	 { "delete_amp_state", _wrap_delete_amp_state, METH_O, NULL},
	 { "amp_state_swigregister", amp_state_swigregister, METH_O, NULL},
	 { "amp_state_swiginit", amp_state_swiginit, METH_VARARGS, NULL},
	 { "Amp_caps_set", _wrap_Amp_caps_set, METH_VARARGS, NULL},
	 { "Amp_caps_get", _wrap_Amp_caps_get, METH_O, NULL},
	 { "Amp_state_set", _wrap_Amp_state_set, METH_VARARGS, NULL},
	 { "Amp_state_get", _wrap_Amp_state_get, METH_O, NULL},
	 { "Amp_error_status_set", _wrap_Amp_error_status_set, METH_VARARGS, NULL},
	 { "Amp_error_status_get", _wrap_Amp_error_status_get, METH_O, NULL},
	 { "Amp_do_exception_set", _wrap_Amp_do_exception_set, METH_VARARGS, NULL},
	 { "Amp_do_exception_get", _wrap_Amp_do_exception_get, METH_O, NULL},
	 { "new_Amp", _wrap_new_Amp, METH_O, NULL},
	 { "delete_Amp", _wrap_delete_Amp, METH_O, NULL},
	 { "Amp_open", _wrap_Amp_open, METH_O, NULL},
	 { "Amp_close", _wrap_Amp_close, METH_O, NULL},
	 { "Amp_reset", _wrap_Amp_reset, METH_VARARGS, NULL},
	 { "Amp_token_lookup", _wrap_Amp_token_lookup, METH_VARARGS, NULL},
	 { "Amp_has_get_level", _wrap_Amp_has_get_level, METH_VARARGS, NULL},
	 { "Amp_has_set_level", _wrap_Amp_has_set_level, METH_VARARGS, NULL},
	 { "Amp_get_ext_level", _wrap_Amp_get_ext_level, METH_VARARGS, NULL},
	 { "Amp_set_freq", _wrap_Amp_set_freq, METH_VARARGS, NULL},
	 { "Amp_set_powerstat", _wrap_Amp_set_powerstat, METH_VARARGS, NULL},
	 { "Amp_set_level", _wrap_Amp_set_level, METH_VARARGS, NULL},
	 { "Amp_set_ext_level", _wrap_Amp_set_ext_level, METH_VARARGS, NULL},
	 { "Amp_set_conf", _wrap_Amp_set_conf, METH_VARARGS, NULL},
	 { "Amp_get_conf", _wrap_Amp_get_conf, METH_VARARGS, NULL},
	 { "Amp_get_freq", _wrap_Amp_get_freq, METH_O, NULL},
	 { "Amp_get_info", _wrap_Amp_get_info, METH_O, NULL},
	 { "Amp_get_powerstat", _wrap_Amp_get_powerstat, METH_O, NULL},
	 { "Amp_get_level", _wrap_Amp_get_level, METH_VARARGS, NULL},
	 { "Amp_swigregister", Amp_swigregister, METH_O, NULL},
	 { "Amp_swiginit", Amp_swiginit, METH_VARARGS, NULL},
	 { NULL, NULL, 0, NULL }
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static void *_p_channelArrayTo_p_channel(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((struct channel *)  ((channelArray *) x));
}
static swig_type_info _swigt__p_Amp = {"_p_Amp", "Amp *|struct Amp *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_FIFO_RIG_s = {"_p_FIFO_RIG_s", "struct FIFO_RIG_s *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_FILE = {"_p_FILE", "FILE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Rig = {"_p_Rig", "Rig *|struct Rig *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Rot = {"_p_Rot", "Rot *|struct Rot *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_agc_level_e = {"_p_agc_level_e", "enum agc_level_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_amp = {"_p_amp", "AMP *|struct amp *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_amp_caps = {"_p_amp_caps", "struct amp_caps *|amp_caps *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_amp_reset_e = {"_p_amp_reset_e", "amp_reset_t *|enum amp_reset_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_amp_state = {"_p_amp_state", "struct amp_state *|amp_state *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_amp_type_e = {"_p_amp_type_e", "amp_type_t *|enum amp_type_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ann_t = {"_p_ann_t", "ann_t *|enum ann_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_cal_table = {"_p_cal_table", "cal_table_t *|struct cal_table *|cal_table *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_cal_table_float = {"_p_cal_table_float", "cal_table_float_t *|struct cal_table_float *|cal_table_float *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_cal_table_float_table = {"_p_cal_table_float_table", "cal_table_float_table *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_cal_table_table = {"_p_cal_table_table", "cal_table_table *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_chan_list = {"_p_chan_list", "chan_t *|struct chan_list *|chan_list *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_chan_type_t = {"_p_chan_type_t", "chan_type_t *|enum chan_type_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_channel = {"_p_channel", "channel_t *|struct channel *|channel *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_channelArray = {"_p_channelArray", "channelArray *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_channel_cap = {"_p_channel_cap", "channel_cap_t *|struct channel_cap *|channel_cap *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_client_e = {"_p_client_e", "client_t *|enum client_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_confparams = {"_p_confparams", "struct confparams *|confparams *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_confparams_u = {"_p_confparams_u", "confparams_u *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_confparams_u_c = {"_p_confparams_u_c", "confparams_u_c *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_confparams_u_n = {"_p_confparams_u_n", "confparams_u_n *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_dcd_e = {"_p_dcd_e", "dcd_t *|enum dcd_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_dcd_type_e = {"_p_dcd_type_e", "dcd_type_t *|enum dcd_type_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_deferred_config_header = {"_p_deferred_config_header", "deferred_config_header_t *|struct deferred_config_header *|deferred_config_header *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_deferred_config_item = {"_p_deferred_config_item", "deferred_config_item_t *|struct deferred_config_item *|deferred_config_item *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_double = {"_p_double", "freq_t *|double *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ext_list = {"_p_ext_list", "struct ext_list *|ext_list *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp__int = {"_p_f_p_struct_amp__int", "int (*)(AMP *)|int (*)(struct amp *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp__p_char = {"_p_f_p_struct_amp__p_char", "char *(*)(AMP *)|char *(*)(struct amp *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_double__int = {"_p_f_p_struct_amp_double__int", "int (*)(AMP *,freq_t)|int (*)(struct amp *,double)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_enum_amp_reset_e__int = {"_p_f_p_struct_amp_enum_amp_reset_e__int", "int (*)(AMP *,amp_reset_t)|int (*)(struct amp *,enum amp_reset_e)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_enum_powerstat_t__int = {"_p_f_p_struct_amp_enum_powerstat_t__int", "int (*)(AMP *,powerstat_t)|int (*)(struct amp *,enum powerstat_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_long_p_char__int = {"_p_f_p_struct_amp_long_p_char__int", "int (*)(AMP *,hamlib_token_t,char *)|int (*)(struct amp *,long,char *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_long_p_char_int__int = {"_p_f_p_struct_amp_long_p_char_int__int", "int (*)(AMP *,hamlib_token_t,char *,int)|int (*)(struct amp *,long,char *,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_long_p_q_const__char__int = {"_p_f_p_struct_amp_long_p_q_const__char__int", "int (*)(AMP *,hamlib_token_t,char const *)|int (*)(struct amp *,long,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_long_p_value_t__int = {"_p_f_p_struct_amp_long_p_value_t__int", "int (*)(AMP *,hamlib_token_t,value_t *)|int (*)(struct amp *,long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_long_value_t__int = {"_p_f_p_struct_amp_long_value_t__int", "int (*)(AMP *,hamlib_token_t,value_t)|int (*)(struct amp *,long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_p_double__int = {"_p_f_p_struct_amp_p_double__int", "int (*)(AMP *,freq_t *)|int (*)(struct amp *,double *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_p_enum_powerstat_t__int = {"_p_f_p_struct_amp_p_enum_powerstat_t__int", "int (*)(AMP *,powerstat_t *)|int (*)(struct amp *,enum powerstat_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_unsigned_long_long_p_value_t__int = {"_p_f_p_struct_amp_unsigned_long_long_p_value_t__int", "int (*)(AMP *,setting_t,value_t *)|int (*)(struct amp *,unsigned long long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_unsigned_long_long_value_t__int = {"_p_f_p_struct_amp_unsigned_long_long_value_t__int", "int (*)(AMP *,setting_t,value_t)|int (*)(struct amp *,unsigned long long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig__int = {"_p_f_p_struct_s_rig__int", "int (*)(RIG *)|int (*)(struct s_rig *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig__p_char = {"_p_f_p_struct_s_rig__p_char", "char *(*)(RIG *)|char *(*)(struct s_rig *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_enum_powerstat_t__int = {"_p_f_p_struct_s_rig_enum_powerstat_t__int", "int (*)(RIG *,powerstat_t)|int (*)(struct s_rig *,enum powerstat_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_enum_reset_t__int = {"_p_f_p_struct_s_rig_enum_reset_t__int", "int (*)(RIG *,reset_t)|int (*)(struct s_rig *,enum reset_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_int__int = {"_p_f_p_struct_s_rig_int__int", "int (*)(RIG *,int)|int (*)(struct s_rig *,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int = {"_p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int", "int (*)(RIG *,int,int,int,int,int,int,double,int)|int (*)(struct s_rig *,int,int,int,int,int,int,double,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_long_p_char__int = {"_p_f_p_struct_s_rig_long_p_char__int", "int (*)(RIG *,hamlib_token_t,char *)|int (*)(struct s_rig *,long,char *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_long_p_char_int__int = {"_p_f_p_struct_s_rig_long_p_char_int__int", "int (*)(RIG *,hamlib_token_t,char *,int)|int (*)(struct s_rig *,long,char *,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_long_p_q_const__char__int = {"_p_f_p_struct_s_rig_long_p_q_const__char__int", "int (*)(RIG *,hamlib_token_t,char const *)|int (*)(struct s_rig *,long,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_long_p_value_t__int = {"_p_f_p_struct_s_rig_long_p_value_t__int", "int (*)(RIG *,hamlib_token_t,value_t *)|int (*)(struct s_rig *,long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_long_value_t__int = {"_p_f_p_struct_s_rig_long_value_t__int", "int (*)(RIG *,hamlib_token_t,value_t)|int (*)(struct s_rig *,long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_enum_powerstat_t__int = {"_p_f_p_struct_s_rig_p_enum_powerstat_t__int", "int (*)(RIG *,powerstat_t *)|int (*)(struct s_rig *,enum powerstat_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int = {"_p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int", "int (*)(RIG *,float *,unsigned int,freq_t,rmode_t)|int (*)(struct s_rig *,float *,unsigned int,double,unsigned long long)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_int__int = {"_p_f_p_struct_s_rig_p_int__int", "int (*)(RIG *,int *)|int (*)(struct s_rig *,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int = {"_p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int", "int (*)(RIG *,int *,int *,int *,int *,int *,int *,double *,int *)|int (*)(struct s_rig *,int *,int *,int *,int *,int *,int *,double *,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_q_const__char__int = {"_p_f_p_struct_s_rig_p_q_const__char__int", "int (*)(RIG *,char const *)|int (*)(struct s_rig *,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int = {"_p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int", "spectrum_cb_t|int (*)(struct s_rig *,struct rig_spectrum_line *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_unsigned_int__int = {"_p_f_p_struct_s_rig_p_unsigned_int__int", "int (*)(RIG *,vfo_t *)|int (*)(struct s_rig *,unsigned int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int = {"_p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int", "int (*)(RIG *,unsigned int *,float,freq_t,rmode_t)|int (*)(struct s_rig *,unsigned int *,float,double,unsigned long long)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int = {"_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int", "int (*)(RIG *,size_t,unsigned char const *)|int (*)(struct s_rig *,size_t,unsigned char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int__int = {"_p_f_p_struct_s_rig_unsigned_int__int", "int (*)(RIG *,vfo_t)|int (*)(struct s_rig *,unsigned int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_double__int = {"_p_f_p_struct_s_rig_unsigned_int_double__int", "int (*)(RIG *,vfo_t,freq_t)|int (*)(struct s_rig *,unsigned int,double)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_double_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_double_p_void__int", "freq_cb_t|int (*)(struct s_rig *,unsigned int,double,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int = {"_p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int", "int (*)(RIG *,vfo_t,freq_t,rmode_t,pbwidth_t)|int (*)(struct s_rig *,unsigned int,double,unsigned long long,long)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int", "dcd_cb_t|int (*)(struct s_rig *,unsigned int,enum dcd_e,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int", "int (*)(RIG *,vfo_t,ptt_t)|int (*)(struct s_rig *,unsigned int,enum ptt_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int", "ptt_cb_t|int (*)(struct s_rig *,unsigned int,enum ptt_t,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int", "int (*)(RIG *,vfo_t,rptr_shift_t)|int (*)(struct s_rig *,unsigned int,enum rptr_shift_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int", "int (*)(RIG *,vfo_t,scan_t,int)|int (*)(struct s_rig *,unsigned int,enum scan_t,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int", "int (*)(RIG *,vfo_t,split_t,vfo_t)|int (*)(struct s_rig *,unsigned int,enum split_t,unsigned int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int", "int (*)(RIG *,vfo_t,vfo_op_t)|int (*)(struct s_rig *,unsigned int,enum vfo_op_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_int__int = {"_p_f_p_struct_s_rig_unsigned_int_int__int", "int (*)(RIG *,vfo_t,int)|int (*)(struct s_rig *,unsigned int,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_long__int = {"_p_f_p_struct_s_rig_unsigned_int_long__int", "int (*)(RIG *,vfo_t,shortfreq_t)|int (*)(struct s_rig *,unsigned int,long)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_long_int__int = {"_p_f_p_struct_s_rig_unsigned_int_long_int__int", "int (*)(RIG *,vfo_t,hamlib_token_t,int)|int (*)(struct s_rig *,unsigned int,long,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_long_p_int__int = {"_p_f_p_struct_s_rig_unsigned_int_long_p_int__int", "int (*)(RIG *,vfo_t,hamlib_token_t,int *)|int (*)(struct s_rig *,unsigned int,long,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int = {"_p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int", "int (*)(RIG *,vfo_t,hamlib_token_t,value_t *)|int (*)(struct s_rig *,unsigned int,long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_long_value_t__int = {"_p_f_p_struct_s_rig_unsigned_int_long_value_t__int", "int (*)(RIG *,vfo_t,hamlib_token_t,value_t)|int (*)(struct s_rig *,unsigned int,long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int = {"_p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int", "int (*)(RIG *,vfo_t,char *,int *)|int (*)(struct s_rig *,unsigned int,char *,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_double__int = {"_p_f_p_struct_s_rig_unsigned_int_p_double__int", "int (*)(RIG *,vfo_t,freq_t *)|int (*)(struct s_rig *,unsigned int,double *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int = {"_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int", "int (*)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *)|int (*)(struct s_rig *,unsigned int,double *,unsigned long long *,long *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int = {"_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int", "int (*)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *,split_t *)|int (*)(struct s_rig *,unsigned int,double *,unsigned long long *,long *,enum split_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int", "pltune_cb_t|int (*)(struct s_rig *,unsigned int,double *,unsigned long long *,long *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int = {"_p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int", "int (*)(RIG *,vfo_t,dcd_t *)|int (*)(struct s_rig *,unsigned int,enum dcd_e *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int = {"_p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int", "int (*)(RIG *,vfo_t,ptt_t *)|int (*)(struct s_rig *,unsigned int,enum ptt_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int = {"_p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int", "int (*)(RIG *,vfo_t,rptr_shift_t *)|int (*)(struct s_rig *,unsigned int,enum rptr_shift_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int = {"_p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int", "int (*)(RIG *,vfo_t,split_t *,vfo_t *)|int (*)(struct s_rig *,unsigned int,enum split_t *,unsigned int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int", "int (*)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *)|int (*)(struct s_rig *,unsigned int,int (*)(struct s_rig *,unsigned int,struct channel **,int,struct chan_list const *,void *),int (*)(struct s_rig *,struct confparams const *,value_t *,void *),void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int", "int (*)(RIG *,vfo_t,chan_cb_t,void *)|int (*)(struct s_rig *,unsigned int,int (*)(struct s_rig *,unsigned int,struct channel **,int,struct chan_list const *,void *),void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_int__int = {"_p_f_p_struct_s_rig_unsigned_int_p_int__int", "int (*)(RIG *,vfo_t,int *)|int (*)(struct s_rig *,unsigned int,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_long__int = {"_p_f_p_struct_s_rig_unsigned_int_p_long__int", "int (*)(RIG *,vfo_t,shortfreq_t *)|int (*)(struct s_rig *,unsigned int,long *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int = {"_p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int", "int (*)(RIG *,vfo_t,char const *)|int (*)(struct s_rig *,unsigned int,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int = {"_p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int", "int (*)(RIG *,vfo_t,channel_t const *)|int (*)(struct s_rig *,unsigned int,struct channel const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int = {"_p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int", "int (*)(RIG *,vfo_t,channel_t *,int)|int (*)(struct s_rig *,unsigned int,struct channel *,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int = {"_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int", "int (*)(RIG *,vfo_t,tone_t *)|int (*)(struct s_rig *,unsigned int,unsigned int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int = {"_p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int", "int (*)(RIG *,vfo_t,rmode_t *,pbwidth_t *)|int (*)(struct s_rig *,unsigned int,unsigned long long *,long *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_p_void__int", "vfo_cb_t|int (*)(struct s_rig *,unsigned int,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int", "int (*)(RIG *,vfo_t,tone_t)|int (*)(struct s_rig *,unsigned int,unsigned int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int", "int (*)(RIG *,vfo_t,ant_t,value_t *,ant_t *,ant_t *,ant_t *)|int (*)(struct s_rig *,unsigned int,unsigned int,value_t *,unsigned int *,unsigned int *,unsigned int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int", "int (*)(RIG *,vfo_t,ant_t,value_t)|int (*)(struct s_rig *,unsigned int,unsigned int,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int", "int (*)(RIG *,vfo_t,setting_t,int)|int (*)(struct s_rig *,unsigned int,unsigned long long,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int", "int (*)(RIG *,vfo_t,rmode_t,pbwidth_t)|int (*)(struct s_rig *,unsigned int,unsigned long long,long)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int", "mode_cb_t|int (*)(struct s_rig *,unsigned int,unsigned long long,long,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int", "int (*)(RIG *,vfo_t,setting_t,int *)|int (*)(struct s_rig *,unsigned int,unsigned long long,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int", "int (*)(RIG *,vfo_t,setting_t,value_t *)|int (*)(struct s_rig *,unsigned int,unsigned long long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int", "int (*)(RIG *,vfo_t,setting_t,value_t)|int (*)(struct s_rig *,unsigned int,unsigned long long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int = {"_p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int", "int (*)(RIG *,setting_t,value_t *)|int (*)(struct s_rig *,unsigned long long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_long_long_value_t__int = {"_p_f_p_struct_s_rig_unsigned_long_long_value_t__int", "int (*)(RIG *,setting_t,value_t)|int (*)(struct s_rig *,unsigned long long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot__int = {"_p_f_p_struct_s_rot__int", "int (*)(ROT *)|int (*)(struct s_rot *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot__p_char = {"_p_f_p_struct_s_rot__p_char", "char *(*)(ROT *)|char *(*)(struct s_rot *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_float_float__int = {"_p_f_p_struct_s_rot_float_float__int", "int (*)(ROT *,azimuth_t,elevation_t)|int (*)(struct s_rot *,float,float)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_int__int = {"_p_f_p_struct_s_rot_int__int", "int (*)(ROT *,rot_reset_t)|int (*)(struct s_rot *,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_int_int__int = {"_p_f_p_struct_s_rot_int_int__int", "int (*)(ROT *,int,int)|int (*)(struct s_rot *,int,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_int__int = {"_p_f_p_struct_s_rot_long_int__int", "int (*)(ROT *,hamlib_token_t,int)|int (*)(struct s_rot *,long,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_p_char__int = {"_p_f_p_struct_s_rot_long_p_char__int", "int (*)(ROT *,hamlib_token_t,char *)|int (*)(struct s_rot *,long,char *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_p_char_int__int = {"_p_f_p_struct_s_rot_long_p_char_int__int", "int (*)(ROT *,hamlib_token_t,char *,int)|int (*)(struct s_rot *,long,char *,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_p_int__int = {"_p_f_p_struct_s_rot_long_p_int__int", "int (*)(ROT *,hamlib_token_t,int *)|int (*)(struct s_rot *,long,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_p_q_const__char__int = {"_p_f_p_struct_s_rot_long_p_q_const__char__int", "int (*)(ROT *,hamlib_token_t,char const *)|int (*)(struct s_rot *,long,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_p_value_t__int = {"_p_f_p_struct_s_rot_long_p_value_t__int", "int (*)(ROT *,hamlib_token_t,value_t *)|int (*)(struct s_rot *,long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_value_t__int = {"_p_f_p_struct_s_rot_long_value_t__int", "int (*)(ROT *,hamlib_token_t,value_t)|int (*)(struct s_rot *,long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_p_enum_rot_status_t__int = {"_p_f_p_struct_s_rot_p_enum_rot_status_t__int", "int (*)(ROT *,rot_status_t *)|int (*)(struct s_rot *,enum rot_status_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_p_float_p_float__int = {"_p_f_p_struct_s_rot_p_float_p_float__int", "int (*)(ROT *,azimuth_t *,elevation_t *)|int (*)(struct s_rot *,float *,float *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_unsigned_long_long_int__int = {"_p_f_p_struct_s_rot_unsigned_long_long_int__int", "int (*)(ROT *,setting_t,int)|int (*)(struct s_rot *,unsigned long long,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_unsigned_long_long_p_int__int = {"_p_f_p_struct_s_rot_unsigned_long_long_p_int__int", "int (*)(ROT *,setting_t,int *)|int (*)(struct s_rot *,unsigned long long,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int = {"_p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int", "int (*)(ROT *,setting_t,value_t *)|int (*)(struct s_rot *,unsigned long long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_unsigned_long_long_value_t__int = {"_p_f_p_struct_s_rot_unsigned_long_long_value_t__int", "int (*)(ROT *,setting_t,value_t)|int (*)(struct s_rot *,unsigned long long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_filter_list = {"_p_filter_list", "struct filter_list *|filter_list *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_float = {"_p_float", "agc_time_t *|azimuth_t *|elevation_t *|swr_t *|float *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_freq_range_list = {"_p_freq_range_list", "freq_range_t *|struct freq_range_list *|freq_range_list *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_gran = {"_p_gran", "gran_t *|struct gran *|gran *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_hamlib_async_pipe = {"_p_hamlib_async_pipe", "hamlib_async_pipe_t *|struct hamlib_async_pipe *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_hamlib_band_t = {"_p_hamlib_band_t", "hamlib_band_t *|enum hamlib_band_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_hamlib_bandselect_t = {"_p_hamlib_bandselect_t", "hamlib_bandselect_t *|enum hamlib_bandselect_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_hamlib_cache_t = {"_p_hamlib_cache_t", "hamlib_cache_t *|enum hamlib_cache_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_hamlib_port = {"_p_hamlib_port", "hamlib_port_t *|struct hamlib_port *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "amp_model_t *|int32_t *|int_fast16_t *|int_fast32_t *|int_least32_t *|intptr_t *|rot_model_t *|rot_reset_t *|tune_value_t *|int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_long = {"_p_long", "hamlib_token_t *|pbwidth_t *|shortfreq_t *|long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_long_long = {"_p_long_long", "int64_t *|int_fast64_t *|int_least64_t *|intmax_t *|long long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_multicast_s = {"_p_multicast_s", "struct multicast_s *|multicast_s *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_char = {"_p_p_char", "char **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_void = {"_p_p_void", "void **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_powerstat_t = {"_p_powerstat_t", "powerstat_t *|enum powerstat_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_pthread_mutex_t = {"_p_pthread_mutex_t", "pthread_mutex_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_pthread_t = {"_p_pthread_t", "pthread_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ptt_t = {"_p_ptt_t", "ptt_t *|enum ptt_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ptt_type_e = {"_p_ptt_type_e", "ptt_type_t *|enum ptt_type_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_reset_t = {"_p_reset_t", "reset_t *|enum reset_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_callbacks = {"_p_rig_callbacks", "struct rig_callbacks *|rig_callbacks *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_caps = {"_p_rig_caps", "struct rig_caps *|rig_caps *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_debug_level_e = {"_p_rig_debug_level_e", "enum rig_debug_level_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_port_e = {"_p_rig_port_e", "rig_port_t *|enum rig_port_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_ptrx_t = {"_p_rig_ptrx_t", "rig_ptrx_t *|enum rig_ptrx_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_spectrum_avg_mode = {"_p_rig_spectrum_avg_mode", "struct rig_spectrum_avg_mode *|rig_spectrum_avg_mode *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_spectrum_line = {"_p_rig_spectrum_line", "struct rig_spectrum_line *|rig_spectrum_line *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_spectrum_mode_e = {"_p_rig_spectrum_mode_e", "enum rig_spectrum_mode_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_spectrum_scope = {"_p_rig_spectrum_scope", "struct rig_spectrum_scope *|rig_spectrum_scope *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_state = {"_p_rig_state", "struct rig_state *|rig_state *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_type_t = {"_p_rig_type_t", "rig_type_t *|enum rig_type_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rot_caps = {"_p_rot_caps", "struct rot_caps *|rot_caps *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rot_state = {"_p_rot_state", "struct rot_state *|rot_state *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rot_status_t = {"_p_rot_status_t", "rot_status_t *|enum rot_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rot_type_t = {"_p_rot_type_t", "rot_type_t *|enum rot_type_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rptr_shift_t = {"_p_rptr_shift_t", "rptr_shift_t *|enum rptr_shift_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_s_rig = {"_p_s_rig", "RIG *|struct s_rig *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_s_rot = {"_p_s_rot", "ROT *|struct s_rot *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_scan_t = {"_p_scan_t", "scan_t *|enum scan_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_short = {"_p_short", "int16_t *|int_least16_t *|short *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_signed_char = {"_p_signed_char", "int8_t *|int_fast8_t *|int_least8_t *|signed char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_sockaddr_in = {"_p_sockaddr_in", "struct sockaddr_in *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_split_t = {"_p_split_t", "split_t *|enum split_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_time_t = {"_p_time_t", "time_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_timespec = {"_p_timespec", "struct timespec *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_toneArray = {"_p_toneArray", "toneArray *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_tuning_step_list = {"_p_tuning_step_list", "struct tuning_step_list *|tuning_step_list *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_twiddle_state_t = {"_p_twiddle_state_t", "twiddle_state_t *|enum twiddle_state_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_char = {"_p_unsigned_char", "uint8_t *|uint_fast8_t *|uint_least8_t *|unsigned char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_int = {"_p_unsigned_int", "ant_t *|rig_comm_status_t *|rig_model_t *|tone_t *|uint32_t *|uint_fast16_t *|uint_fast32_t *|uint_least32_t *|uintptr_t *|vfo_t *|unsigned int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_long = {"_p_unsigned_long", "rig_useconds_t *|unsigned long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_long_long = {"_p_unsigned_long_long", "rig_level_e *|rmode_t *|setting_t *|uint64_t *|uint_fast64_t *|uint_least64_t *|uintmax_t *|unsigned long long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_short = {"_p_unsigned_short", "uint16_t *|uint_least16_t *|unsigned short *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_value_t = {"_p_value_t", "value_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_value_t_b = {"_p_value_t_b", "value_t_b *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_vfo_op_t = {"_p_vfo_op_t", "vfo_op_t *|enum vfo_op_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_void = {"_p_void", "void *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_Amp,
  &_swigt__p_FIFO_RIG_s,
  &_swigt__p_FILE,
  &_swigt__p_Rig,
  &_swigt__p_Rot,
  &_swigt__p_agc_level_e,
  &_swigt__p_amp,
  &_swigt__p_amp_caps,
  &_swigt__p_amp_reset_e,
  &_swigt__p_amp_state,
  &_swigt__p_amp_type_e,
  &_swigt__p_ann_t,
  &_swigt__p_cal_table,
  &_swigt__p_cal_table_float,
  &_swigt__p_cal_table_float_table,
  &_swigt__p_cal_table_table,
  &_swigt__p_chan_list,
  &_swigt__p_chan_type_t,
  &_swigt__p_channel,
  &_swigt__p_channelArray,
  &_swigt__p_channel_cap,
  &_swigt__p_char,
  &_swigt__p_client_e,
  &_swigt__p_confparams,
  &_swigt__p_confparams_u,
  &_swigt__p_confparams_u_c,
  &_swigt__p_confparams_u_n,
  &_swigt__p_dcd_e,
  &_swigt__p_dcd_type_e,
  &_swigt__p_deferred_config_header,
  &_swigt__p_deferred_config_item,
  &_swigt__p_double,
  &_swigt__p_ext_list,
  &_swigt__p_f_p_struct_amp__int,
  &_swigt__p_f_p_struct_amp__p_char,
  &_swigt__p_f_p_struct_amp_double__int,
  &_swigt__p_f_p_struct_amp_enum_amp_reset_e__int,
  &_swigt__p_f_p_struct_amp_enum_powerstat_t__int,
  &_swigt__p_f_p_struct_amp_long_p_char__int,
  &_swigt__p_f_p_struct_amp_long_p_char_int__int,
  &_swigt__p_f_p_struct_amp_long_p_q_const__char__int,
  &_swigt__p_f_p_struct_amp_long_p_value_t__int,
  &_swigt__p_f_p_struct_amp_long_value_t__int,
  &_swigt__p_f_p_struct_amp_p_double__int,
  &_swigt__p_f_p_struct_amp_p_enum_powerstat_t__int,
  &_swigt__p_f_p_struct_amp_unsigned_long_long_p_value_t__int,
  &_swigt__p_f_p_struct_amp_unsigned_long_long_value_t__int,
  &_swigt__p_f_p_struct_s_rig__int,
  &_swigt__p_f_p_struct_s_rig__p_char,
  &_swigt__p_f_p_struct_s_rig_enum_powerstat_t__int,
  &_swigt__p_f_p_struct_s_rig_enum_reset_t__int,
  &_swigt__p_f_p_struct_s_rig_int__int,
  &_swigt__p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int,
  &_swigt__p_f_p_struct_s_rig_long_p_char__int,
  &_swigt__p_f_p_struct_s_rig_long_p_char_int__int,
  &_swigt__p_f_p_struct_s_rig_long_p_q_const__char__int,
  &_swigt__p_f_p_struct_s_rig_long_p_value_t__int,
  &_swigt__p_f_p_struct_s_rig_long_value_t__int,
  &_swigt__p_f_p_struct_s_rig_p_enum_powerstat_t__int,
  &_swigt__p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int,
  &_swigt__p_f_p_struct_s_rig_p_int__int,
  &_swigt__p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int,
  &_swigt__p_f_p_struct_s_rig_p_q_const__char__int,
  &_swigt__p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int,
  &_swigt__p_f_p_struct_s_rig_p_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int,
  &_swigt__p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_double__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_double_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_long__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_long_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_long_p_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_long_value_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_double__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_long__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_long_long_value_t__int,
  &_swigt__p_f_p_struct_s_rot__int,
  &_swigt__p_f_p_struct_s_rot__p_char,
  &_swigt__p_f_p_struct_s_rot_float_float__int,
  &_swigt__p_f_p_struct_s_rot_int__int,
  &_swigt__p_f_p_struct_s_rot_int_int__int,
  &_swigt__p_f_p_struct_s_rot_long_int__int,
  &_swigt__p_f_p_struct_s_rot_long_p_char__int,
  &_swigt__p_f_p_struct_s_rot_long_p_char_int__int,
  &_swigt__p_f_p_struct_s_rot_long_p_int__int,
  &_swigt__p_f_p_struct_s_rot_long_p_q_const__char__int,
  &_swigt__p_f_p_struct_s_rot_long_p_value_t__int,
  &_swigt__p_f_p_struct_s_rot_long_value_t__int,
  &_swigt__p_f_p_struct_s_rot_p_enum_rot_status_t__int,
  &_swigt__p_f_p_struct_s_rot_p_float_p_float__int,
  &_swigt__p_f_p_struct_s_rot_unsigned_long_long_int__int,
  &_swigt__p_f_p_struct_s_rot_unsigned_long_long_p_int__int,
  &_swigt__p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int,
  &_swigt__p_f_p_struct_s_rot_unsigned_long_long_value_t__int,
  &_swigt__p_filter_list,
  &_swigt__p_float,
  &_swigt__p_freq_range_list,
  &_swigt__p_gran,
  &_swigt__p_hamlib_async_pipe,
  &_swigt__p_hamlib_band_t,
  &_swigt__p_hamlib_bandselect_t,
  &_swigt__p_hamlib_cache_t,
  &_swigt__p_hamlib_port,
  &_swigt__p_int,
  &_swigt__p_long,
  &_swigt__p_long_long,
  &_swigt__p_multicast_s,
  &_swigt__p_p_char,
  &_swigt__p_p_void,
  &_swigt__p_powerstat_t,
  &_swigt__p_pthread_mutex_t,
  &_swigt__p_pthread_t,
  &_swigt__p_ptt_t,
  &_swigt__p_ptt_type_e,
  &_swigt__p_reset_t,
  &_swigt__p_rig_callbacks,
  &_swigt__p_rig_caps,
  &_swigt__p_rig_debug_level_e,
  &_swigt__p_rig_port_e,
  &_swigt__p_rig_ptrx_t,
  &_swigt__p_rig_spectrum_avg_mode,
  &_swigt__p_rig_spectrum_line,
  &_swigt__p_rig_spectrum_mode_e,
  &_swigt__p_rig_spectrum_scope,
  &_swigt__p_rig_state,
  &_swigt__p_rig_type_t,
  &_swigt__p_rot_caps,
  &_swigt__p_rot_state,
  &_swigt__p_rot_status_t,
  &_swigt__p_rot_type_t,
  &_swigt__p_rptr_shift_t,
  &_swigt__p_s_rig,
  &_swigt__p_s_rot,
  &_swigt__p_scan_t,
  &_swigt__p_short,
  &_swigt__p_signed_char,
  &_swigt__p_sockaddr_in,
  &_swigt__p_split_t,
  &_swigt__p_time_t,
  &_swigt__p_timespec,
  &_swigt__p_toneArray,
  &_swigt__p_tuning_step_list,
  &_swigt__p_twiddle_state_t,
  &_swigt__p_unsigned_char,
  &_swigt__p_unsigned_int,
  &_swigt__p_unsigned_long,
  &_swigt__p_unsigned_long_long,
  &_swigt__p_unsigned_short,
  &_swigt__p_value_t,
  &_swigt__p_value_t_b,
  &_swigt__p_vfo_op_t,
  &_swigt__p_void,
};

static swig_cast_info _swigc__p_Amp[] = {  {&_swigt__p_Amp, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_FIFO_RIG_s[] = {  {&_swigt__p_FIFO_RIG_s, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_FILE[] = {  {&_swigt__p_FILE, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Rig[] = {  {&_swigt__p_Rig, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Rot[] = {  {&_swigt__p_Rot, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_agc_level_e[] = {  {&_swigt__p_agc_level_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_amp[] = {  {&_swigt__p_amp, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_amp_caps[] = {  {&_swigt__p_amp_caps, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_amp_reset_e[] = {  {&_swigt__p_amp_reset_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_amp_state[] = {  {&_swigt__p_amp_state, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_amp_type_e[] = {  {&_swigt__p_amp_type_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ann_t[] = {  {&_swigt__p_ann_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_cal_table[] = {  {&_swigt__p_cal_table, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_cal_table_float[] = {  {&_swigt__p_cal_table_float, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_cal_table_float_table[] = {  {&_swigt__p_cal_table_float_table, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_cal_table_table[] = {  {&_swigt__p_cal_table_table, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_chan_list[] = {  {&_swigt__p_chan_list, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_chan_type_t[] = {  {&_swigt__p_chan_type_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_channel[] = {  {&_swigt__p_channel, 0, 0, 0},  {&_swigt__p_channelArray, _p_channelArrayTo_p_channel, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_channelArray[] = {  {&_swigt__p_channelArray, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_channel_cap[] = {  {&_swigt__p_channel_cap, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_client_e[] = {  {&_swigt__p_client_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_confparams[] = {  {&_swigt__p_confparams, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_confparams_u[] = {  {&_swigt__p_confparams_u, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_confparams_u_c[] = {  {&_swigt__p_confparams_u_c, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_confparams_u_n[] = {  {&_swigt__p_confparams_u_n, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_dcd_e[] = {  {&_swigt__p_dcd_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_dcd_type_e[] = {  {&_swigt__p_dcd_type_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_deferred_config_header[] = {  {&_swigt__p_deferred_config_header, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_deferred_config_item[] = {  {&_swigt__p_deferred_config_item, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_double[] = {  {&_swigt__p_double, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ext_list[] = {  {&_swigt__p_ext_list, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp__int[] = {  {&_swigt__p_f_p_struct_amp__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp__p_char[] = {  {&_swigt__p_f_p_struct_amp__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_double__int[] = {  {&_swigt__p_f_p_struct_amp_double__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_enum_amp_reset_e__int[] = {  {&_swigt__p_f_p_struct_amp_enum_amp_reset_e__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_enum_powerstat_t__int[] = {  {&_swigt__p_f_p_struct_amp_enum_powerstat_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_long_p_char__int[] = {  {&_swigt__p_f_p_struct_amp_long_p_char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_long_p_char_int__int[] = {  {&_swigt__p_f_p_struct_amp_long_p_char_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_long_p_q_const__char__int[] = {  {&_swigt__p_f_p_struct_amp_long_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_amp_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_long_value_t__int[] = {  {&_swigt__p_f_p_struct_amp_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_p_double__int[] = {  {&_swigt__p_f_p_struct_amp_p_double__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_p_enum_powerstat_t__int[] = {  {&_swigt__p_f_p_struct_amp_p_enum_powerstat_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_unsigned_long_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_amp_unsigned_long_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_unsigned_long_long_value_t__int[] = {  {&_swigt__p_f_p_struct_amp_unsigned_long_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig__int[] = {  {&_swigt__p_f_p_struct_s_rig__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig__p_char[] = {  {&_swigt__p_f_p_struct_s_rig__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_enum_powerstat_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_enum_powerstat_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_enum_reset_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_enum_reset_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_long_p_char__int[] = {  {&_swigt__p_f_p_struct_s_rig_long_p_char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_long_p_char_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_long_p_char_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_long_p_q_const__char__int[] = {  {&_swigt__p_f_p_struct_s_rig_long_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_long_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_enum_powerstat_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_enum_powerstat_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_q_const__char__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int[] = {  {&_swigt__p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_double__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_double__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_double_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_double_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_long_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_long_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_long_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_long_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_long_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_double__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_double__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_long_long_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_long_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot__int[] = {  {&_swigt__p_f_p_struct_s_rot__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot__p_char[] = {  {&_swigt__p_f_p_struct_s_rot__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_float_float__int[] = {  {&_swigt__p_f_p_struct_s_rot_float_float__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_int_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_int_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_p_char__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_p_char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_p_char_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_p_char_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_p_q_const__char__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_p_enum_rot_status_t__int[] = {  {&_swigt__p_f_p_struct_s_rot_p_enum_rot_status_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_p_float_p_float__int[] = {  {&_swigt__p_f_p_struct_s_rot_p_float_p_float__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_unsigned_long_long_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_unsigned_long_long_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_unsigned_long_long_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_unsigned_long_long_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_unsigned_long_long_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rot_unsigned_long_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_filter_list[] = {  {&_swigt__p_filter_list, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_float[] = {  {&_swigt__p_float, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_freq_range_list[] = {  {&_swigt__p_freq_range_list, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_gran[] = {  {&_swigt__p_gran, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_hamlib_async_pipe[] = {  {&_swigt__p_hamlib_async_pipe, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_hamlib_band_t[] = {  {&_swigt__p_hamlib_band_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_hamlib_bandselect_t[] = {  {&_swigt__p_hamlib_bandselect_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_hamlib_cache_t[] = {  {&_swigt__p_hamlib_cache_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_hamlib_port[] = {  {&_swigt__p_hamlib_port, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_long[] = {  {&_swigt__p_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_long_long[] = {  {&_swigt__p_long_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_multicast_s[] = {  {&_swigt__p_multicast_s, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_char[] = {  {&_swigt__p_p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_void[] = {  {&_swigt__p_p_void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_powerstat_t[] = {  {&_swigt__p_powerstat_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_pthread_mutex_t[] = {  {&_swigt__p_pthread_mutex_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_pthread_t[] = {  {&_swigt__p_pthread_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ptt_t[] = {  {&_swigt__p_ptt_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ptt_type_e[] = {  {&_swigt__p_ptt_type_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_reset_t[] = {  {&_swigt__p_reset_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_callbacks[] = {  {&_swigt__p_rig_callbacks, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_caps[] = {  {&_swigt__p_rig_caps, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_debug_level_e[] = {  {&_swigt__p_rig_debug_level_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_port_e[] = {  {&_swigt__p_rig_port_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_ptrx_t[] = {  {&_swigt__p_rig_ptrx_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_spectrum_avg_mode[] = {  {&_swigt__p_rig_spectrum_avg_mode, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_spectrum_line[] = {  {&_swigt__p_rig_spectrum_line, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_spectrum_mode_e[] = {  {&_swigt__p_rig_spectrum_mode_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_spectrum_scope[] = {  {&_swigt__p_rig_spectrum_scope, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_state[] = {  {&_swigt__p_rig_state, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_type_t[] = {  {&_swigt__p_rig_type_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rot_caps[] = {  {&_swigt__p_rot_caps, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rot_state[] = {  {&_swigt__p_rot_state, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rot_status_t[] = {  {&_swigt__p_rot_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rot_type_t[] = {  {&_swigt__p_rot_type_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rptr_shift_t[] = {  {&_swigt__p_rptr_shift_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_s_rig[] = {  {&_swigt__p_s_rig, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_s_rot[] = {  {&_swigt__p_s_rot, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_scan_t[] = {  {&_swigt__p_scan_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_short[] = {  {&_swigt__p_short, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_signed_char[] = {  {&_swigt__p_signed_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_sockaddr_in[] = {  {&_swigt__p_sockaddr_in, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_split_t[] = {  {&_swigt__p_split_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_time_t[] = {  {&_swigt__p_time_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_timespec[] = {  {&_swigt__p_timespec, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_toneArray[] = {  {&_swigt__p_toneArray, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_tuning_step_list[] = {  {&_swigt__p_tuning_step_list, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_twiddle_state_t[] = {  {&_swigt__p_twiddle_state_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_char[] = {  {&_swigt__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_int[] = {  {&_swigt__p_unsigned_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_long[] = {  {&_swigt__p_unsigned_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_long_long[] = {  {&_swigt__p_unsigned_long_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_short[] = {  {&_swigt__p_unsigned_short, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_value_t[] = {  {&_swigt__p_value_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_value_t_b[] = {  {&_swigt__p_value_t_b, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_vfo_op_t[] = {  {&_swigt__p_vfo_op_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_void[] = {  {&_swigt__p_void, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_Amp,
  _swigc__p_FIFO_RIG_s,
  _swigc__p_FILE,
  _swigc__p_Rig,
  _swigc__p_Rot,
  _swigc__p_agc_level_e,
  _swigc__p_amp,
  _swigc__p_amp_caps,
  _swigc__p_amp_reset_e,
  _swigc__p_amp_state,
  _swigc__p_amp_type_e,
  _swigc__p_ann_t,
  _swigc__p_cal_table,
  _swigc__p_cal_table_float,
  _swigc__p_cal_table_float_table,
  _swigc__p_cal_table_table,
  _swigc__p_chan_list,
  _swigc__p_chan_type_t,
  _swigc__p_channel,
  _swigc__p_channelArray,
  _swigc__p_channel_cap,
  _swigc__p_char,
  _swigc__p_client_e,
  _swigc__p_confparams,
  _swigc__p_confparams_u,
  _swigc__p_confparams_u_c,
  _swigc__p_confparams_u_n,
  _swigc__p_dcd_e,
  _swigc__p_dcd_type_e,
  _swigc__p_deferred_config_header,
  _swigc__p_deferred_config_item,
  _swigc__p_double,
  _swigc__p_ext_list,
  _swigc__p_f_p_struct_amp__int,
  _swigc__p_f_p_struct_amp__p_char,
  _swigc__p_f_p_struct_amp_double__int,
  _swigc__p_f_p_struct_amp_enum_amp_reset_e__int,
  _swigc__p_f_p_struct_amp_enum_powerstat_t__int,
  _swigc__p_f_p_struct_amp_long_p_char__int,
  _swigc__p_f_p_struct_amp_long_p_char_int__int,
  _swigc__p_f_p_struct_amp_long_p_q_const__char__int,
  _swigc__p_f_p_struct_amp_long_p_value_t__int,
  _swigc__p_f_p_struct_amp_long_value_t__int,
  _swigc__p_f_p_struct_amp_p_double__int,
  _swigc__p_f_p_struct_amp_p_enum_powerstat_t__int,
  _swigc__p_f_p_struct_amp_unsigned_long_long_p_value_t__int,
  _swigc__p_f_p_struct_amp_unsigned_long_long_value_t__int,
  _swigc__p_f_p_struct_s_rig__int,
  _swigc__p_f_p_struct_s_rig__p_char,
  _swigc__p_f_p_struct_s_rig_enum_powerstat_t__int,
  _swigc__p_f_p_struct_s_rig_enum_reset_t__int,
  _swigc__p_f_p_struct_s_rig_int__int,
  _swigc__p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int,
  _swigc__p_f_p_struct_s_rig_long_p_char__int,
  _swigc__p_f_p_struct_s_rig_long_p_char_int__int,
  _swigc__p_f_p_struct_s_rig_long_p_q_const__char__int,
  _swigc__p_f_p_struct_s_rig_long_p_value_t__int,
  _swigc__p_f_p_struct_s_rig_long_value_t__int,
  _swigc__p_f_p_struct_s_rig_p_enum_powerstat_t__int,
  _swigc__p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int,
  _swigc__p_f_p_struct_s_rig_p_int__int,
  _swigc__p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int,
  _swigc__p_f_p_struct_s_rig_p_q_const__char__int,
  _swigc__p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int,
  _swigc__p_f_p_struct_s_rig_p_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int,
  _swigc__p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_double__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_double_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_long__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_long_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_long_p_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_long_value_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_double__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_long__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_long_long_value_t__int,
  _swigc__p_f_p_struct_s_rot__int,
  _swigc__p_f_p_struct_s_rot__p_char,
  _swigc__p_f_p_struct_s_rot_float_float__int,
  _swigc__p_f_p_struct_s_rot_int__int,
  _swigc__p_f_p_struct_s_rot_int_int__int,
  _swigc__p_f_p_struct_s_rot_long_int__int,
  _swigc__p_f_p_struct_s_rot_long_p_char__int,
  _swigc__p_f_p_struct_s_rot_long_p_char_int__int,
  _swigc__p_f_p_struct_s_rot_long_p_int__int,
  _swigc__p_f_p_struct_s_rot_long_p_q_const__char__int,
  _swigc__p_f_p_struct_s_rot_long_p_value_t__int,
  _swigc__p_f_p_struct_s_rot_long_value_t__int,
  _swigc__p_f_p_struct_s_rot_p_enum_rot_status_t__int,
  _swigc__p_f_p_struct_s_rot_p_float_p_float__int,
  _swigc__p_f_p_struct_s_rot_unsigned_long_long_int__int,
  _swigc__p_f_p_struct_s_rot_unsigned_long_long_p_int__int,
  _swigc__p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int,
  _swigc__p_f_p_struct_s_rot_unsigned_long_long_value_t__int,
  _swigc__p_filter_list,
  _swigc__p_float,
  _swigc__p_freq_range_list,
  _swigc__p_gran,
  _swigc__p_hamlib_async_pipe,
  _swigc__p_hamlib_band_t,
  _swigc__p_hamlib_bandselect_t,
  _swigc__p_hamlib_cache_t,
  _swigc__p_hamlib_port,
  _swigc__p_int,
  _swigc__p_long,
  _swigc__p_long_long,
  _swigc__p_multicast_s,
  _swigc__p_p_char,
  _swigc__p_p_void,
  _swigc__p_powerstat_t,
  _swigc__p_pthread_mutex_t,
  _swigc__p_pthread_t,
  _swigc__p_ptt_t,
  _swigc__p_ptt_type_e,
  _swigc__p_reset_t,
  _swigc__p_rig_callbacks,
  _swigc__p_rig_caps,
  _swigc__p_rig_debug_level_e,
  _swigc__p_rig_port_e,
  _swigc__p_rig_ptrx_t,
  _swigc__p_rig_spectrum_avg_mode,
  _swigc__p_rig_spectrum_line,
  _swigc__p_rig_spectrum_mode_e,
  _swigc__p_rig_spectrum_scope,
  _swigc__p_rig_state,
  _swigc__p_rig_type_t,
  _swigc__p_rot_caps,
  _swigc__p_rot_state,
  _swigc__p_rot_status_t,
  _swigc__p_rot_type_t,
  _swigc__p_rptr_shift_t,
  _swigc__p_s_rig,
  _swigc__p_s_rot,
  _swigc__p_scan_t,
  _swigc__p_short,
  _swigc__p_signed_char,
  _swigc__p_sockaddr_in,
  _swigc__p_split_t,
  _swigc__p_time_t,
  _swigc__p_timespec,
  _swigc__p_toneArray,
  _swigc__p_tuning_step_list,
  _swigc__p_twiddle_state_t,
  _swigc__p_unsigned_char,
  _swigc__p_unsigned_int,
  _swigc__p_unsigned_long,
  _swigc__p_unsigned_long_long,
  _swigc__p_unsigned_short,
  _swigc__p_value_t,
  _swigc__p_value_t_b,
  _swigc__p_vfo_op_t,
  _swigc__p_void,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_const_info swig_const_table[] = {
{0, 0, 0, 0.0, 0, 0}};

#ifdef __cplusplus
}
#endif
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is made tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast dependencies. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns.
 *
 * We loop through the cast[] array associated with the type and mark casts
 * which have not been defined in previously loaded modules by assigning
 * cast pointer value to cast->next. We also hash cast->type->name string
 * and store the value in the cast->value field. If we encounter swig_cast_info
 * structure that represents a cast to self we move it to the beginning
 * of the cast array. One trick we need to do is making sure the type pointer
 * in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the list (because the cast->type) pointer will
 *     be correct.
 *
 * Once the cast array has been set up AND it does have new casts that need
 * to be added we sort non-self cast entries to move filtered out entries
 * to the end of the array and to arrange the rest in the increasing order
 * of their type pointer values. We store the index of the last added entry
 * in the cast->value field of the entry[0] (overwriting the name hash). Then
 * we sort <next|value> fields of the remaining entries to arrange hash values
 * in the increasing order. This way cast->next->type->name field matches
 * the cast->value hash.
 *
 * Example:
 * Array of casts for type stored at 0x5000, cast to type stored at 0x3000
 * has already been loaded
 *
 *      After sweep-and-hash:   After sort-by-type:    After sort-by-hash:
 *        ________________       ________________       ________________
 *       |                |     |                |     |                |
 * Entry |  type = 0x5000 |     |  type = 0x5000 |     |  type = 0x5000 |
 *   0   |                |     |                |     |                |
 *       |  next = Entry0 |     |  next = Entry0 |     |  next = Entry0 |
 *       | value = 1212   |     | value = 3      |     | value = 3      |
 *       |                |     |                |     |                |
 *       |================|     |================|     |================|
 *       |                |     |                |     |                |
 * Entry |  type = 0x2000 |     |  type = 0x1000 |     |  type = 0x1000 |
 *   1   |                |     |                |     |                |
 *       |  next = Entry1 |     |  next = Entry1 |     |  next = Entry3 |
 *       | value = 3434   |     | value = 4545   |     | value = 2323   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x3000 |     |  type = 0x2000 |     |  type = 0x2000 |
 *   2   |                |     |                |     |                |
 *       |  next = 0      |     |  next = Entry2 |     |  next = Entry2 |
 *       | value = 0      |     | value = 3434   |     | value = 3434   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x1000 |     |  type = 0x4000 |     |  type = 0x4000 |
 *   3   |                |     |                |     |                |
 *       |  next = Entry3 |     |  next = Entry3 |     |  next = Entry1 |
 *       | value = 4545   |     | value = 2323   |     | value = 4545   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x4000 |     |  type = 0x3000 |     |  type = 0x3000 |
 *   4   |                |     |                |     |                |
 *       |  next = Entry4 |     |  next = 0      |     |  next = 0      |
 *       | value = 2323   |     | value = 0      |     | value = 0      |
 *       |________________|     |________________|     |________________|
 *
 * Once the cast array has been initialized, we use cast[0]->next field to link
 * it into the list of cast arrays for the type.
 *              ____    ____    ____
 * type->cast->|next|->|next|->|next|->0
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *
 * Subsequent cast resolution works as follows:
 *
 * 1. Check whether the type matches the first entry in the current cast array.
 * 2. If not, then do a binary search over the (0:cast->value] interval using
 *    either type address or the hash value of the type name.
 * 3. If not found, then move over to the next cast array (cast[0]->next).
 *
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif

#ifndef SWIG_INIT_CLIENT_DATA_TYPE
#define SWIG_INIT_CLIENT_DATA_TYPE void *
#endif

/*
 * Sort function that puts cast entries with nonzero 'next' at the front
 * of the array while ordering them by addresses of their 'type' structs.
 */
SWIGINTERN int SWIG_CastCmpStruct(const void *pa, const void *pb) {
  swig_cast_info *pca = (swig_cast_info *)pa;
  swig_cast_info *pcb = (swig_cast_info *)pb;
  if (pca->type < pcb->type)
  return (pca->next || pcb->next == 0) ? -1 : 1;
  if (pca->type > pcb->type)
  return (pcb->next || pca->next == 0) ? 1 : -1;
  return 0;
}

/*
 * Shell-sort 'next' and 'value' field pairs to order them by 'value'.
 */
SWIGINTERN void SWIG_CastHashSort(swig_cast_info *cast, int size) {
  const int hmax = size/9;
  int h, i;
  for(h = 1; h <= hmax; h = 3*h+1);
  for(; h > 0; h /= 3)
  {
    for(i = h; i < size; ++i)
    {
      swig_cast_info *p = cast[i].next;
      unsigned int hash = cast[i].value;
      int j = i;
      while(j >= h && hash < cast[j-h].value)
      {
        cast[j].next = cast[j-h].next;
        cast[j].value = cast[j-h].value;
        j -= h;
      }
      cast[j].next = p;
      cast[j].value = hash;
    }
  }
}

SWIGRUNTIME void
SWIG_InitializeModule(SWIG_INIT_CLIENT_DATA_TYPE clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *target_type;
    swig_cast_info *cast, *first;
    int num_mapped = 0;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = first = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      target_type = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        target_type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
        if (target_type) {
          /* Target type already defined in another module */
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: found cast %s\n", target_type->name);
#endif
          if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
            printf("SWIG_InitializeModule: skip old type %s\n", target_type->name);
#endif
            cast->type = target_type;
            target_type = 0;
          } else {
            /* Check if this cast is already in the list */
            swig_cast_info *ocast = SWIG_TypeCheck(target_type->name, type);
#ifdef SWIGRUNTIME_DEBUG
            if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", target_type->name);
#endif
            if (!ocast) target_type = 0;
          }
        }
      }
      
      if (!target_type) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        /* Set inclusion mark for sorting */
        cast->next = cast;
        num_mapped++;
        
        if (type == cast->type) {
#ifdef SWIGRUNTIME_DEBUG
          printf("%s : self cast at pos [%li]\n", type->name, cast - first);
#endif
          if (cast - first) {
            /* Move cast to itself to the first entry in the array */
            
            swig_cast_info tmp = *cast;
            *cast = *first;
            *first = tmp;
          }
          first++;
          
        } else {
          cast->value = SWIG_Hash(cast->type->name, (unsigned int)strlen(cast->type->name));
        }
      }
      cast++;
    }
    
    if (num_mapped) {
      if (cast - first) {
        swig_cast_info *tmp;
        
        /* Sort casts by type address for binary search in SWIG_TypeCheckStruct */
        qsort(first, cast - first, sizeof(swig_cast_info), SWIG_CastCmpStruct);
        
        /* Remap back links for added entries */
        cast = swig_module.cast_initial[i] + num_mapped;
        for (tmp = first; tmp < cast; tmp++) {
          tmp->next = tmp;
        }
      }
      
      /* Set the value field of the first entry to the index of the last added entry */
      cast = swig_module.cast_initial[i];
      cast->value = num_mapped - 1;
      
      num_mapped -= (int)(first - cast);
      if (num_mapped > 1) {
        /* Sort <'next','value'> pairs by 'value' for binary search in SWIG_TypeCheck */
        
        SWIG_CastHashSort(first, num_mapped);
      }
      
      first = type->cast;
      if (first) {
        /* Link the current set into the list of cast arrays */
        cast->next = first->next;
        first->next = cast;
      } else {
        cast->next = 0;
        type->cast = cast;
      }
    }
    
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      swig_cast_info *head, *cast;
      head = swig_module.types[i]->cast;
      while (head) {
        for (cast = head; (cast - head) <= head->value; cast++) {
          if (!cast->converter) {
            if (cast->type && !cast->type->clientdata)
            SWIG_TypeClientData(cast->type, swig_module.types[i]->clientdata);
          }
        }
        head = head->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#ifdef __cplusplus
extern "C" {
#endif
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  SWIGINTERN void
  SWIG_Python_InstallConstants(PyObject *d, swig_const_info constants[]) {
    PyObject *obj = 0;
    size_t i;
    for (i = 0; constants[i].type; ++i) {
      switch(constants[i].type) {
      case SWIG_PY_POINTER:
        obj = SWIG_InternalNewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_PY_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        PyDict_SetItemString(d, constants[i].name, obj);
        SWIG_Py_DECREF(obj);
      }
    }
  }
  
  /* -----------------------------------------------------------------------------
   * Patch %callback methods' docstrings to hold the callback ptrs
   * -----------------------------------------------------------------------------*/
  
  SWIGINTERN void
  SWIG_Python_FixMethods(PyMethodDef *methods, const swig_const_info *const_table, swig_type_info **types, swig_type_info **types_initial) {
    size_t i;
    for (i = 0; methods[i].ml_name; ++i) {
      const char *c = methods[i].ml_doc;
      if (!c) continue;
      c = strstr(c, "swig_ptr: ");
      if (c) {
        int j;
        const swig_const_info *ci = 0;
        const char *name = c + 10;
        for (j = 0; const_table[j].type; ++j) {
          if (strncmp(const_table[j].name, name, 
              strlen(const_table[j].name)) == 0) {
            ci = &(const_table[j]);
            break;
          }
        }
        if (ci) {
          void *ptr = (ci->type == SWIG_PY_POINTER) ? ci->pvalue : 0;
          if (ptr) {
            size_t shift = (ci->ptype) - types;
            swig_type_info *ty = types_initial[shift];
            size_t ldoc = (c - methods[i].ml_doc);
            size_t lptr = strlen(ty->name)+2*sizeof(void*)+2;
            char *ndoc = (char*)malloc(ldoc + lptr + 10);
            if (ndoc) {
              char *buff = ndoc;
              memcpy(buff, methods[i].ml_doc, ldoc);
              buff += ldoc;
              memcpy(buff, "swig_ptr: ", 10);
              buff += 10;
              SWIG_PackVoidPtr(buff, ptr, ty->name, lptr);
              methods[i].ml_doc = ndoc;
            }
          }
        }
      }
    }
  } 
  
#ifdef __cplusplus
}
#endif




/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

SWIGINTERN int SWIG_mod_exec(PyObject *module);

#ifdef __cplusplus
extern "C"
#endif

SWIGEXPORT 
#if PY_VERSION_HEX >= 0x03000000
PyObject*
#else
void
#endif
SWIG_init(void) {
#if PY_VERSION_HEX >= 0x03000000
  static PyModuleDef_Slot SwigSlots[] = {
    {
      Py_mod_exec, (void *)SWIG_mod_exec 
    },
#ifdef SWIGPYTHON_NOGIL
#ifdef Py_GIL_DISABLED
    {
      Py_mod_gil, Py_MOD_GIL_NOT_USED 
    },
#endif
#endif
    {
      0, NULL 
    }
  };
  
  static struct PyModuleDef SWIG_module = {
    PyModuleDef_HEAD_INIT,
    SWIG_name,
    NULL,
    0,
    SwigMethods,
    SwigSlots,
    NULL,
    NULL,
    NULL
  };
  
  return PyModuleDef_Init(&SWIG_module);
#else
  PyObject *m = Py_InitModule(SWIG_name, SwigMethods);
  if (m && SWIG_mod_exec(m) != 0) {
    Py_DECREF(m);
  }
#endif
}

SWIGINTERN int SWIG_mod_exec(PyObject *m) {
  PyObject *d, *md, *globals;
  
#if defined(SWIGPYTHON_BUILTIN)
  static SwigPyClientData SwigPyObject_clientdata = {
    0, 0, 0, 0, 0, 0, 0
  };
  static PyGetSetDef this_getset_def = {
    (char *)"this", &SwigPyBuiltin_ThisClosure, NULL, NULL, NULL
  };
  static SwigPyGetSet thisown_getset_closure = {
    SwigPyObject_own,
    SwigPyObject_own
  };
  static PyGetSetDef thisown_getset_def = {
    (char *)"thisown", SwigPyBuiltin_GetterClosure, SwigPyBuiltin_SetterClosure, NULL, &thisown_getset_closure
  };
  PyTypeObject *builtin_pytype;
  int builtin_base_count;
  swig_type_info *builtin_basetype;
  PyObject *tuple;
  PyGetSetDescrObject *static_getset;
  PyTypeObject *metatype;
  PyTypeObject *swigpyobject;
  SwigPyClientData *cd;
  PyObject *public_interface, *public_symbol;
  PyObject *this_descr;
  PyObject *thisown_descr;
  PyObject *self = 0;
  int i;
  
  (void)builtin_pytype;
  (void)builtin_base_count;
  (void)builtin_basetype;
  (void)tuple;
  (void)static_getset;
  (void)self;
  
  /* Metaclass is used to implement static member variables */
  metatype = SwigPyObjectType_Type();
  assert(metatype);
  
  SwigPyStaticVar_Type();
#endif
  
  (void)globals;
  
  /* Create singletons now to avoid potential deadlocks with multi-threaded usage after module initialization */
  SWIG_runtime_data_module();
  SWIG_This();
  SWIG_Python_TypeCache();
  SwigPyPacked_Type();
  SwigVarLink_Type();
#ifndef SWIGPYTHON_BUILTIN
  SwigPyObject_Type();
#endif
  
  /* Fix SwigMethods to carry the callback ptrs when needed */
  SWIG_Python_FixMethods(SwigMethods, swig_const_table, swig_types, swig_type_initial);
  
#ifdef SWIGPYTHON_NOGIL
#ifdef Py_GIL_DISABLED
  PyUnstable_Module_SetGIL(m, Py_MOD_GIL_NOT_USED);
#endif
#endif
  
  md = d = PyModule_GetDict(m);
  (void)md;
  
  SWIG_InitializeModule(0);
  
#ifdef SWIGPYTHON_BUILTIN
  swigpyobject = SwigPyObject_TypeOnce();
  
  SwigPyObject_stype = SWIG_MangledTypeQuery("_p_SwigPyObject");
  assert(SwigPyObject_stype);
  cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
  if (!cd) {
    SwigPyObject_stype->clientdata = &SwigPyObject_clientdata;
    SwigPyObject_clientdata.pytype = swigpyobject;
  } else if (swigpyobject->tp_basicsize != cd->pytype->tp_basicsize) {
    PyErr_SetString(PyExc_RuntimeError, "Import error: attempted to load two incompatible swig-generated modules.");
    return -1;
  }
  
  /* All objects have a 'this' attribute */
  this_descr = PyDescr_NewGetSet(SwigPyObject_Type(), &this_getset_def);
  (void)this_descr;
  
  /* All objects have a 'thisown' attribute */
  thisown_descr = PyDescr_NewGetSet(SwigPyObject_Type(), &thisown_getset_def);
  (void)thisown_descr;
  
  public_interface = PyList_New(0);
  public_symbol = 0;
  (void)public_symbol;
  
  PyDict_SetItemString(md, "__all__", public_interface);
  SWIG_Py_DECREF(public_interface);
  for (i = 0; SwigMethods[i].ml_name != NULL; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, SwigMethods[i].ml_name);
  for (i = 0; swig_const_table[i].name != 0; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, swig_const_table[i].name);
#endif
  
  SWIG_InstallConstants(d,swig_const_table);
  
  SWIG_Python_SetConstant(d, "MAX_MODELS_PER_BACKEND",SWIG_From_int((int)(1000)));
  SWIG_Python_SetConstant(d, "RIG_MODEL_NONE",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DUMMY",SWIG_From_int((int)((1000*(0)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_NETRIGCTL",SWIG_From_int((int)((1000*(0)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ARMSTRONG",SWIG_From_int((int)((1000*(0)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FLRIG",SWIG_From_int((int)((1000*(0)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TRXMANAGER_RIG",SWIG_From_int((int)((1000*(0)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DUMMY_NOVFO",SWIG_From_int((int)((1000*(0)+(6)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TCI1X",SWIG_From_int((int)((1000*(0)+(7)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ACLOG",SWIG_From_int((int)((1000*(0)+(8)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SDRSHARP",SWIG_From_int((int)((1000*(0)+(9)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_QUISK",SWIG_From_int((int)((1000*(0)+(10)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_GQRX",SWIG_From_int((int)((1000*(0)+(11)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT847",SWIG_From_int((int)((1000*(1)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT1000",SWIG_From_int((int)((1000*(1)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT1000D",SWIG_From_int((int)((1000*(1)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT1000MPMKV",SWIG_From_int((int)((1000*(1)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT747",SWIG_From_int((int)((1000*(1)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT757",SWIG_From_int((int)((1000*(1)+(6)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT757GXII",SWIG_From_int((int)((1000*(1)+(7)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT575",SWIG_From_int((int)((1000*(1)+(8)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT767",SWIG_From_int((int)((1000*(1)+(9)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT736R",SWIG_From_int((int)((1000*(1)+(10)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT840",SWIG_From_int((int)((1000*(1)+(11)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT820",SWIG_From_int((int)((1000*(1)+(12)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT900",SWIG_From_int((int)((1000*(1)+(13)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT920",SWIG_From_int((int)((1000*(1)+(14)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT890",SWIG_From_int((int)((1000*(1)+(15)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT990",SWIG_From_int((int)((1000*(1)+(16)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FRG100",SWIG_From_int((int)((1000*(1)+(17)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FRG9600",SWIG_From_int((int)((1000*(1)+(18)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FRG8800",SWIG_From_int((int)((1000*(1)+(19)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT817",SWIG_From_int((int)((1000*(1)+(20)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT100",SWIG_From_int((int)((1000*(1)+(21)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT857",SWIG_From_int((int)((1000*(1)+(22)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT897",SWIG_From_int((int)((1000*(1)+(23)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT1000MP",SWIG_From_int((int)((1000*(1)+(24)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT1000MPMKVFLD",SWIG_From_int((int)((1000*(1)+(25)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_VR5000",SWIG_From_int((int)((1000*(1)+(26)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT450",SWIG_From_int((int)((1000*(1)+(27)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT950",SWIG_From_int((int)((1000*(1)+(28)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT2000",SWIG_From_int((int)((1000*(1)+(29)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT9000",SWIG_From_int((int)((1000*(1)+(30)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT980",SWIG_From_int((int)((1000*(1)+(31)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FTDX5000",SWIG_From_int((int)((1000*(1)+(32)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_VX1700",SWIG_From_int((int)((1000*(1)+(33)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FTDX1200",SWIG_From_int((int)((1000*(1)+(34)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT991",SWIG_From_int((int)((1000*(1)+(35)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT891",SWIG_From_int((int)((1000*(1)+(36)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FTDX3000",SWIG_From_int((int)((1000*(1)+(37)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT847UNI",SWIG_From_int((int)((1000*(1)+(38)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT600",SWIG_From_int((int)((1000*(1)+(39)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FTDX101D",SWIG_From_int((int)((1000*(1)+(40)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT818",SWIG_From_int((int)((1000*(1)+(41)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FTDX10",SWIG_From_int((int)((1000*(1)+(42)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT897D",SWIG_From_int((int)((1000*(1)+(43)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FTDX101MP",SWIG_From_int((int)((1000*(1)+(44)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_MCHFQRP",SWIG_From_int((int)((1000*(1)+(45)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT450D",SWIG_From_int((int)((1000*(1)+(46)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT650",SWIG_From_int((int)((1000*(1)+(47)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT990UNI",SWIG_From_int((int)((1000*(1)+(48)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT710",SWIG_From_int((int)((1000*(1)+(49)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FT9000OLD",SWIG_From_int((int)((1000*(1)+(50)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FTX1",SWIG_From_int((int)((1000*(1)+(51)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS50",SWIG_From_int((int)((1000*(2)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS440",SWIG_From_int((int)((1000*(2)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS450S",SWIG_From_int((int)((1000*(2)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS570D",SWIG_From_int((int)((1000*(2)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS690S",SWIG_From_int((int)((1000*(2)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS711",SWIG_From_int((int)((1000*(2)+(6)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS790",SWIG_From_int((int)((1000*(2)+(7)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS811",SWIG_From_int((int)((1000*(2)+(8)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS850",SWIG_From_int((int)((1000*(2)+(9)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS870S",SWIG_From_int((int)((1000*(2)+(10)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS940",SWIG_From_int((int)((1000*(2)+(11)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS950S",SWIG_From_int((int)((1000*(2)+(12)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS950SDX",SWIG_From_int((int)((1000*(2)+(13)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS2000",SWIG_From_int((int)((1000*(2)+(14)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_R5000",SWIG_From_int((int)((1000*(2)+(15)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS570S",SWIG_From_int((int)((1000*(2)+(16)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_THD7A",SWIG_From_int((int)((1000*(2)+(17)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_THD7AG",SWIG_From_int((int)((1000*(2)+(18)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_THF6A",SWIG_From_int((int)((1000*(2)+(19)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_THF7E",SWIG_From_int((int)((1000*(2)+(20)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_K2",SWIG_From_int((int)((1000*(2)+(21)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS930",SWIG_From_int((int)((1000*(2)+(22)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_THG71",SWIG_From_int((int)((1000*(2)+(23)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS680S",SWIG_From_int((int)((1000*(2)+(24)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS140S",SWIG_From_int((int)((1000*(2)+(25)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TMD700",SWIG_From_int((int)((1000*(2)+(26)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TMV7",SWIG_From_int((int)((1000*(2)+(27)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS480",SWIG_From_int((int)((1000*(2)+(28)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_K3",SWIG_From_int((int)((1000*(2)+(29)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TRC80",SWIG_From_int((int)((1000*(2)+(30)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS590S",SWIG_From_int((int)((1000*(2)+(31)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TRANSFOX",SWIG_From_int((int)((1000*(2)+(32)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_THD72A",SWIG_From_int((int)((1000*(2)+(33)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TMD710",SWIG_From_int((int)((1000*(2)+(34)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TMV71",SWIG_From_int((int)((1000*(2)+(35)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_F6K",SWIG_From_int((int)((1000*(2)+(36)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS590SG",SWIG_From_int((int)((1000*(2)+(37)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_XG3",SWIG_From_int((int)((1000*(2)+(38)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS990S",SWIG_From_int((int)((1000*(2)+(39)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_HPSDR",SWIG_From_int((int)((1000*(2)+(40)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TS890S",SWIG_From_int((int)((1000*(2)+(41)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_THD74",SWIG_From_int((int)((1000*(2)+(42)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_K3S",SWIG_From_int((int)((1000*(2)+(43)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_KX2",SWIG_From_int((int)((1000*(2)+(44)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_KX3",SWIG_From_int((int)((1000*(2)+(45)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PT8000A",SWIG_From_int((int)((1000*(2)+(46)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_K4",SWIG_From_int((int)((1000*(2)+(47)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_POWERSDR",SWIG_From_int((int)((1000*(2)+(48)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_MALACHITE",SWIG_From_int((int)((1000*(2)+(49)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_LAB599_TX500",SWIG_From_int((int)((1000*(2)+(50)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SDRUNO",SWIG_From_int((int)((1000*(2)+(51)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_QRPLABS",SWIG_From_int((int)((1000*(2)+(52)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FX4",SWIG_From_int((int)((1000*(2)+(53)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_THETIS",SWIG_From_int((int)((1000*(2)+(54)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TRUSDX",SWIG_From_int((int)((1000*(2)+(55)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SDRCONSOLE",SWIG_From_int((int)((1000*(2)+(56)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_QRPLABS_QMX",SWIG_From_int((int)((1000*(2)+(57)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC1271",SWIG_From_int((int)((1000*(3)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC1275",SWIG_From_int((int)((1000*(3)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC271",SWIG_From_int((int)((1000*(3)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC275",SWIG_From_int((int)((1000*(3)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC375",SWIG_From_int((int)((1000*(3)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC471",SWIG_From_int((int)((1000*(3)+(6)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC475",SWIG_From_int((int)((1000*(3)+(7)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC575",SWIG_From_int((int)((1000*(3)+(8)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC706",SWIG_From_int((int)((1000*(3)+(9)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC706MKII",SWIG_From_int((int)((1000*(3)+(10)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC706MKIIG",SWIG_From_int((int)((1000*(3)+(11)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC707",SWIG_From_int((int)((1000*(3)+(12)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC718",SWIG_From_int((int)((1000*(3)+(13)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC725",SWIG_From_int((int)((1000*(3)+(14)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC726",SWIG_From_int((int)((1000*(3)+(15)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC728",SWIG_From_int((int)((1000*(3)+(16)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC729",SWIG_From_int((int)((1000*(3)+(17)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC731",SWIG_From_int((int)((1000*(3)+(18)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC735",SWIG_From_int((int)((1000*(3)+(19)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC736",SWIG_From_int((int)((1000*(3)+(20)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC737",SWIG_From_int((int)((1000*(3)+(21)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC738",SWIG_From_int((int)((1000*(3)+(22)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC746",SWIG_From_int((int)((1000*(3)+(23)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC751",SWIG_From_int((int)((1000*(3)+(24)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC751A",SWIG_From_int((int)((1000*(3)+(25)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC756",SWIG_From_int((int)((1000*(3)+(26)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC756PRO",SWIG_From_int((int)((1000*(3)+(27)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC761",SWIG_From_int((int)((1000*(3)+(28)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC765",SWIG_From_int((int)((1000*(3)+(29)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC775",SWIG_From_int((int)((1000*(3)+(30)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC781",SWIG_From_int((int)((1000*(3)+(31)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC820",SWIG_From_int((int)((1000*(3)+(32)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC821H",SWIG_From_int((int)((1000*(3)+(34)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC970",SWIG_From_int((int)((1000*(3)+(35)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR10",SWIG_From_int((int)((1000*(3)+(36)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR71",SWIG_From_int((int)((1000*(3)+(37)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR72",SWIG_From_int((int)((1000*(3)+(38)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR75",SWIG_From_int((int)((1000*(3)+(39)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR7000",SWIG_From_int((int)((1000*(3)+(40)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR7100",SWIG_From_int((int)((1000*(3)+(41)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR8500",SWIG_From_int((int)((1000*(3)+(42)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR9000",SWIG_From_int((int)((1000*(3)+(43)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC910",SWIG_From_int((int)((1000*(3)+(44)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC78",SWIG_From_int((int)((1000*(3)+(45)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC746PRO",SWIG_From_int((int)((1000*(3)+(46)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC756PROII",SWIG_From_int((int)((1000*(3)+(47)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICID1",SWIG_From_int((int)((1000*(3)+(54)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC703",SWIG_From_int((int)((1000*(3)+(55)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC7800",SWIG_From_int((int)((1000*(3)+(56)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC756PROIII",SWIG_From_int((int)((1000*(3)+(57)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR20",SWIG_From_int((int)((1000*(3)+(58)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC7000",SWIG_From_int((int)((1000*(3)+(60)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC7200",SWIG_From_int((int)((1000*(3)+(61)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC7700",SWIG_From_int((int)((1000*(3)+(62)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC7600",SWIG_From_int((int)((1000*(3)+(63)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC92D",SWIG_From_int((int)((1000*(3)+(65)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR9500",SWIG_From_int((int)((1000*(3)+(66)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC7410",SWIG_From_int((int)((1000*(3)+(67)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC9100",SWIG_From_int((int)((1000*(3)+(68)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICRX7",SWIG_From_int((int)((1000*(3)+(69)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC7100",SWIG_From_int((int)((1000*(3)+(70)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ID5100",SWIG_From_int((int)((1000*(3)+(71)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC2730",SWIG_From_int((int)((1000*(3)+(72)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC7300",SWIG_From_int((int)((1000*(3)+(73)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PERSEUS",SWIG_From_int((int)((1000*(3)+(74)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC785x",SWIG_From_int((int)((1000*(3)+(75)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_X108G",SWIG_From_int((int)((1000*(3)+(76)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR6",SWIG_From_int((int)((1000*(3)+(77)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC7610",SWIG_From_int((int)((1000*(3)+(78)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR8600",SWIG_From_int((int)((1000*(3)+(79)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICR30",SWIG_From_int((int)((1000*(3)+(80)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC9700",SWIG_From_int((int)((1000*(3)+(81)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ID4100",SWIG_From_int((int)((1000*(3)+(82)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ID31",SWIG_From_int((int)((1000*(3)+(83)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ID51",SWIG_From_int((int)((1000*(3)+(84)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC705",SWIG_From_int((int)((1000*(3)+(85)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ICF8101",SWIG_From_int((int)((1000*(3)+(86)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_X6100",SWIG_From_int((int)((1000*(3)+(87)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_G90",SWIG_From_int((int)((1000*(3)+(88)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_X5105",SWIG_From_int((int)((1000*(3)+(89)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC905",SWIG_From_int((int)((1000*(3)+(90)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_X6200",SWIG_From_int((int)((1000*(3)+(91)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC7760",SWIG_From_int((int)((1000*(3)+(92)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ID52PLUS",SWIG_From_int((int)((1000*(3)+(93)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC7300MK2",SWIG_From_int((int)((1000*(3)+(94)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_MINISCOUT",SWIG_From_int((int)((1000*(3)+(48)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_XPLORER",SWIG_From_int((int)((1000*(3)+(49)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_OS535",SWIG_From_int((int)((1000*(3)+(52)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_OS456",SWIG_From_int((int)((1000*(3)+(53)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_OMNIVI",SWIG_From_int((int)((1000*(3)+(50)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_OMNIVIP",SWIG_From_int((int)((1000*(3)+(51)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PARAGON2",SWIG_From_int((int)((1000*(3)+(59)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DELTAII",SWIG_From_int((int)((1000*(3)+(64)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PCR1000",SWIG_From_int((int)((1000*(4)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PCR100",SWIG_From_int((int)((1000*(4)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PCR1500",SWIG_From_int((int)((1000*(4)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PCR2500",SWIG_From_int((int)((1000*(4)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR8200",SWIG_From_int((int)((1000*(5)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR8000",SWIG_From_int((int)((1000*(5)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR7030",SWIG_From_int((int)((1000*(5)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR5000",SWIG_From_int((int)((1000*(5)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR3030",SWIG_From_int((int)((1000*(5)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR3000A",SWIG_From_int((int)((1000*(5)+(6)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR3000",SWIG_From_int((int)((1000*(5)+(7)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR2700",SWIG_From_int((int)((1000*(5)+(8)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR2500",SWIG_From_int((int)((1000*(5)+(9)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR16",SWIG_From_int((int)((1000*(5)+(10)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SDU5500",SWIG_From_int((int)((1000*(5)+(11)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SDU5000",SWIG_From_int((int)((1000*(5)+(12)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR8600",SWIG_From_int((int)((1000*(5)+(13)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR5000A",SWIG_From_int((int)((1000*(5)+(14)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_AR7030P",SWIG_From_int((int)((1000*(5)+(15)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SR2200",SWIG_From_int((int)((1000*(5)+(16)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_JST145",SWIG_From_int((int)((1000*(6)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_JST245",SWIG_From_int((int)((1000*(6)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_CMH530",SWIG_From_int((int)((1000*(6)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_NRD345",SWIG_From_int((int)((1000*(6)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_NRD525",SWIG_From_int((int)((1000*(6)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_NRD535",SWIG_From_int((int)((1000*(6)+(6)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_NRD545",SWIG_From_int((int)((1000*(6)+(7)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RS64",SWIG_From_int((int)((1000*(7)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RS2005",SWIG_From_int((int)((1000*(7)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RS2006",SWIG_From_int((int)((1000*(7)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RS2035",SWIG_From_int((int)((1000*(7)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RS2042",SWIG_From_int((int)((1000*(7)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RS2041",SWIG_From_int((int)((1000*(7)+(6)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BC780",SWIG_From_int((int)((1000*(8)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BC245",SWIG_From_int((int)((1000*(8)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BC895",SWIG_From_int((int)((1000*(8)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PRO2052",SWIG_From_int((int)((1000*(8)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BC235",SWIG_From_int((int)((1000*(8)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BC250",SWIG_From_int((int)((1000*(8)+(6)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BC785",SWIG_From_int((int)((1000*(8)+(7)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BC786",SWIG_From_int((int)((1000*(8)+(8)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BCT8",SWIG_From_int((int)((1000*(8)+(9)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BCD396T",SWIG_From_int((int)((1000*(8)+(10)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BCD996T",SWIG_From_int((int)((1000*(8)+(11)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BC898",SWIG_From_int((int)((1000*(8)+(12)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DKR8",SWIG_From_int((int)((1000*(9)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DKR8A",SWIG_From_int((int)((1000*(9)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DKR8B",SWIG_From_int((int)((1000*(9)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_HF150",SWIG_From_int((int)((1000*(10)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_HF225",SWIG_From_int((int)((1000*(10)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_HF250",SWIG_From_int((int)((1000*(10)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_HF235",SWIG_From_int((int)((1000*(10)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RA3790",SWIG_From_int((int)((1000*(11)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RA3720",SWIG_From_int((int)((1000*(11)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RA6790",SWIG_From_int((int)((1000*(11)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RA3710",SWIG_From_int((int)((1000*(11)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RA3702",SWIG_From_int((int)((1000*(11)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_HF1000",SWIG_From_int((int)((1000*(12)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_HF1000A",SWIG_From_int((int)((1000*(12)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_WJ8711",SWIG_From_int((int)((1000*(12)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_WJ8888",SWIG_From_int((int)((1000*(12)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ESM500",SWIG_From_int((int)((1000*(13)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_EK890",SWIG_From_int((int)((1000*(13)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_EK891",SWIG_From_int((int)((1000*(13)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_EK895",SWIG_From_int((int)((1000*(13)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_EK070",SWIG_From_int((int)((1000*(13)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TRP7000",SWIG_From_int((int)((1000*(14)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TRP8000",SWIG_From_int((int)((1000*(14)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TRP9000",SWIG_From_int((int)((1000*(14)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TRP8255",SWIG_From_int((int)((1000*(14)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_WR1000",SWIG_From_int((int)((1000*(15)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_WR1500",SWIG_From_int((int)((1000*(15)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_WR1550",SWIG_From_int((int)((1000*(15)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_WR3100",SWIG_From_int((int)((1000*(15)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_WR3150",SWIG_From_int((int)((1000*(15)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_WR3500",SWIG_From_int((int)((1000*(15)+(6)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_WR3700",SWIG_From_int((int)((1000*(15)+(7)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_G303",SWIG_From_int((int)((1000*(15)+(8)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_G313",SWIG_From_int((int)((1000*(15)+(9)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_G305",SWIG_From_int((int)((1000*(15)+(10)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_G315",SWIG_From_int((int)((1000*(15)+(11)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TT550",SWIG_From_int((int)((1000*(16)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TT538",SWIG_From_int((int)((1000*(16)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RX320",SWIG_From_int((int)((1000*(16)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RX340",SWIG_From_int((int)((1000*(16)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RX350",SWIG_From_int((int)((1000*(16)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TT526",SWIG_From_int((int)((1000*(16)+(6)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TT516",SWIG_From_int((int)((1000*(16)+(7)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TT565",SWIG_From_int((int)((1000*(16)+(8)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TT585",SWIG_From_int((int)((1000*(16)+(9)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TT588",SWIG_From_int((int)((1000*(16)+(11)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RX331",SWIG_From_int((int)((1000*(16)+(12)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_TT599",SWIG_From_int((int)((1000*(16)+(13)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DX77",SWIG_From_int((int)((1000*(17)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DXSR8",SWIG_From_int((int)((1000*(17)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_505DSP",SWIG_From_int((int)((1000*(18)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_GNURADIO",SWIG_From_int((int)((1000*(20)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_MC4020",SWIG_From_int((int)((1000*(20)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_GRAUDIO",SWIG_From_int((int)((1000*(20)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_GRAUDIOIQ",SWIG_From_int((int)((1000*(20)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_USRP_G",SWIG_From_int((int)((1000*(20)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_MICROTUNE_4937",SWIG_From_int((int)((1000*(21)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_MICROTUNE_4702",SWIG_From_int((int)((1000*(21)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_MICROTUNE_4707",SWIG_From_int((int)((1000*(21)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DSP10",SWIG_From_int((int)((1000*(22)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SDR1000",SWIG_From_int((int)((1000*(23)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SDR1000RFE",SWIG_From_int((int)((1000*(23)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DTTSP",SWIG_From_int((int)((1000*(23)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DTTSP_UDP",SWIG_From_int((int)((1000*(23)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SMARTSDR_A",SWIG_From_int((int)((1000*(23)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SMARTSDR_B",SWIG_From_int((int)((1000*(23)+(6)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SMARTSDR_C",SWIG_From_int((int)((1000*(23)+(7)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SMARTSDR_D",SWIG_From_int((int)((1000*(23)+(8)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SMARTSDR_E",SWIG_From_int((int)((1000*(23)+(9)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SMARTSDR_F",SWIG_From_int((int)((1000*(23)+(10)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SMARTSDR_G",SWIG_From_int((int)((1000*(23)+(11)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SMARTSDR_H",SWIG_From_int((int)((1000*(23)+(12)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_EKD500",SWIG_From_int((int)((1000*(24)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ELEKTOR304",SWIG_From_int((int)((1000*(25)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DRT1",SWIG_From_int((int)((1000*(25)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DWT",SWIG_From_int((int)((1000*(25)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_USRP0",SWIG_From_int((int)((1000*(25)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_USRP",SWIG_From_int((int)((1000*(25)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DDS60",SWIG_From_int((int)((1000*(25)+(6)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ELEKTOR507",SWIG_From_int((int)((1000*(25)+(7)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_MINIVNA",SWIG_From_int((int)((1000*(25)+(8)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SI570AVRUSB",SWIG_From_int((int)((1000*(25)+(9)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PMSDR",SWIG_From_int((int)((1000*(25)+(10)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SI570PICUSB",SWIG_From_int((int)((1000*(25)+(11)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FIFISDR",SWIG_From_int((int)((1000*(25)+(12)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FUNCUBEDONGLE",SWIG_From_int((int)((1000*(25)+(13)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_HIQSDR",SWIG_From_int((int)((1000*(25)+(14)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FASDR",SWIG_From_int((int)((1000*(25)+(15)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SI570PEABERRY1",SWIG_From_int((int)((1000*(25)+(16)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SI570PEABERRY2",SWIG_From_int((int)((1000*(25)+(17)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_FUNCUBEDONGLEPLUS",SWIG_From_int((int)((1000*(25)+(18)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_RSHFIQ",SWIG_From_int((int)((1000*(25)+(19)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_V4L",SWIG_From_int((int)((1000*(26)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_V4L2",SWIG_From_int((int)((1000*(26)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ESMC",SWIG_From_int((int)((1000*(27)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_EB200",SWIG_From_int((int)((1000*(27)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_XK2100",SWIG_From_int((int)((1000*(27)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_EK89X",SWIG_From_int((int)((1000*(27)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_XK852",SWIG_From_int((int)((1000*(27)+(5)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PRM8060",SWIG_From_int((int)((1000*(28)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PRM8070",SWIG_From_int((int)((1000*(28)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ADT_200A",SWIG_From_int((int)((1000*(29)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC_M700PRO",SWIG_From_int((int)((1000*(30)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC_M802",SWIG_From_int((int)((1000*(30)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC_M710",SWIG_From_int((int)((1000*(30)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_IC_M803",SWIG_From_int((int)((1000*(30)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DORJI_DRA818V",SWIG_From_int((int)((1000*(31)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_DORJI_DRA818U",SWIG_From_int((int)((1000*(31)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BARRETT_2050",SWIG_From_int((int)((1000*(32)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BARRETT_950",SWIG_From_int((int)((1000*(32)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BARRETT_4050",SWIG_From_int((int)((1000*(32)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_BARRETT_4100",SWIG_From_int((int)((1000*(32)+(4)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ELAD_FDM_DUO",SWIG_From_int((int)((1000*(33)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_CODAN_ENVOY",SWIG_From_int((int)((1000*(34)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_CODAN_NGT",SWIG_From_int((int)((1000*(34)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_CODAN_2110",SWIG_From_int((int)((1000*(34)+(3)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_GS100",SWIG_From_int((int)((1000*(35)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_MDS4710",SWIG_From_int((int)((1000*(36)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_MDS9710",SWIG_From_int((int)((1000*(36)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_ATD578UVIII",SWIG_From_int((int)((1000*(37)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_MICOM2",SWIG_From_int((int)((1000*(38)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_CTX10",SWIG_From_int((int)((1000*(39)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PMR171",SWIG_From_int((int)((1000*(40)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_Q900",SWIG_From_int((int)((1000*(40)+(2)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_SIMPLECAT",SWIG_From_int((int)((1000*(41)+(1)))));
  SWIG_Python_SetConstant(d, "RIG_MODEL_PRC138",SWIG_From_int((int)((1000*(42)+(1)))));
  SWIG_Python_SetConstant(d, "BUILTINFUNC",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "HAMLIB_SECRET_LENGTH",SWIG_From_int((int)(32)));
  SWIG_Python_SetConstant(d, "HAMLIB_COOKIE_SIZE",SWIG_From_int((int)(37)));
  globals = SWIG_globals();
  if (!globals) {
    PyErr_SetString(PyExc_TypeError, "Failure to create SWIG globals.");
    return -1;
  }
  PyDict_SetItemString(md, "cvar", globals);
  SWIG_Python_SetConstant(d, "RIG_OK",SWIG_From_int((int)(RIG_OK)));
  SWIG_Python_SetConstant(d, "RIG_EINVAL",SWIG_From_int((int)(RIG_EINVAL)));
  SWIG_Python_SetConstant(d, "RIG_ECONF",SWIG_From_int((int)(RIG_ECONF)));
  SWIG_Python_SetConstant(d, "RIG_ENOMEM",SWIG_From_int((int)(RIG_ENOMEM)));
  SWIG_Python_SetConstant(d, "RIG_ENIMPL",SWIG_From_int((int)(RIG_ENIMPL)));
  SWIG_Python_SetConstant(d, "RIG_ETIMEOUT",SWIG_From_int((int)(RIG_ETIMEOUT)));
  SWIG_Python_SetConstant(d, "RIG_EIO",SWIG_From_int((int)(RIG_EIO)));
  SWIG_Python_SetConstant(d, "RIG_EINTERNAL",SWIG_From_int((int)(RIG_EINTERNAL)));
  SWIG_Python_SetConstant(d, "RIG_EPROTO",SWIG_From_int((int)(RIG_EPROTO)));
  SWIG_Python_SetConstant(d, "RIG_ERJCTED",SWIG_From_int((int)(RIG_ERJCTED)));
  SWIG_Python_SetConstant(d, "RIG_ETRUNC",SWIG_From_int((int)(RIG_ETRUNC)));
  SWIG_Python_SetConstant(d, "RIG_ENAVAIL",SWIG_From_int((int)(RIG_ENAVAIL)));
  SWIG_Python_SetConstant(d, "RIG_ENTARGET",SWIG_From_int((int)(RIG_ENTARGET)));
  SWIG_Python_SetConstant(d, "RIG_BUSERROR",SWIG_From_int((int)(RIG_BUSERROR)));
  SWIG_Python_SetConstant(d, "RIG_BUSBUSY",SWIG_From_int((int)(RIG_BUSBUSY)));
  SWIG_Python_SetConstant(d, "RIG_EARG",SWIG_From_int((int)(RIG_EARG)));
  SWIG_Python_SetConstant(d, "RIG_EVFO",SWIG_From_int((int)(RIG_EVFO)));
  SWIG_Python_SetConstant(d, "RIG_EDOM",SWIG_From_int((int)(RIG_EDOM)));
  SWIG_Python_SetConstant(d, "RIG_EDEPRECATED",SWIG_From_int((int)(RIG_EDEPRECATED)));
  SWIG_Python_SetConstant(d, "RIG_ESECURITY",SWIG_From_int((int)(RIG_ESECURITY)));
  SWIG_Python_SetConstant(d, "RIG_EPOWER",SWIG_From_int((int)(RIG_EPOWER)));
  SWIG_Python_SetConstant(d, "RIG_ELIMIT",SWIG_From_int((int)(RIG_ELIMIT)));
  SWIG_Python_SetConstant(d, "RIG_EACCESS",SWIG_From_int((int)(RIG_EACCESS)));
  SWIG_Python_SetConstant(d, "RIG_EEND",SWIG_From_int((int)(RIG_EEND)));
  SWIG_Python_SetConstant(d, "NETRIGCTL_RET",SWIG_FromCharPtr("RPRT "));
  SWIG_Python_SetConstant(d, "RIG_DEBUG_NONE",SWIG_From_int((int)(RIG_DEBUG_NONE)));
  SWIG_Python_SetConstant(d, "RIG_DEBUG_BUG",SWIG_From_int((int)(RIG_DEBUG_BUG)));
  SWIG_Python_SetConstant(d, "RIG_DEBUG_ERR",SWIG_From_int((int)(RIG_DEBUG_ERR)));
  SWIG_Python_SetConstant(d, "RIG_DEBUG_WARN",SWIG_From_int((int)(RIG_DEBUG_WARN)));
  SWIG_Python_SetConstant(d, "RIG_DEBUG_VERBOSE",SWIG_From_int((int)(RIG_DEBUG_VERBOSE)));
  SWIG_Python_SetConstant(d, "RIG_DEBUG_TRACE",SWIG_From_int((int)(RIG_DEBUG_TRACE)));
  SWIG_Python_SetConstant(d, "RIG_DEBUG_CACHE",SWIG_From_int((int)(RIG_DEBUG_CACHE)));
  SWIG_Python_SetConstant(d, "HAMLIB_RIGNAMSIZ",SWIG_From_int((int)(30)));
  SWIG_Python_SetConstant(d, "HAMLIB_RIGVERSIZ",SWIG_From_int((int)(8)));
  SWIG_Python_SetConstant(d, "HAMLIB_FILPATHLEN",SWIG_From_int((int)(512)));
  SWIG_Python_SetConstant(d, "HAMLIB_FRQRANGESIZ",SWIG_From_int((int)(30)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAXCHANDESC",SWIG_From_int((int)(30)));
  SWIG_Python_SetConstant(d, "HAMLIB_TSLSTSIZ",SWIG_From_int((int)(20)));
  SWIG_Python_SetConstant(d, "HAMLIB_FLTLSTSIZ",SWIG_From_int((int)(60)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAXDBLSTSIZ",SWIG_From_int((int)(8)));
  SWIG_Python_SetConstant(d, "HAMLIB_CHANLSTSIZ",SWIG_From_int((int)(16)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_AGC_LEVELS",SWIG_From_int((int)(8)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_SPECTRUM_SCOPES",SWIG_From_int((int)(4)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_SPECTRUM_MODES",SWIG_From_int((int)(5)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_SPECTRUM_AVG_MODES",SWIG_From_int((int)(12)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_SPECTRUM_SPANS",SWIG_From_int((int)(20)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_SPECTRUM_DATA",SWIG_From_int((int)(2048)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_CAL_LENGTH",SWIG_From_int((int)(32)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_MODES",SWIG_From_int((int)(63)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_VFOS",SWIG_From_int((int)(31)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_ROTORS",SWIG_From_int((int)(31)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_VFO_OPS",SWIG_From_int((int)(31)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_RSCANS",SWIG_From_int((int)(31)));
  SWIG_Python_SetConstant(d, "HAMLIB_MAX_SNAPSHOT_PACKET_SIZE",SWIG_From_int((int)(16384)));
  SWIG_Python_SetConstant(d, "CTCSS_LIST_SIZE",SWIG_From_int((int)(60)));
  SWIG_Python_SetConstant(d, "DCS_LIST_SIZE",SWIG_From_int((int)(128)));
  SWIG_Python_SetConstant(d, "RIG_PORT_NONE",SWIG_From_int((int)(RIG_PORT_NONE)));
  SWIG_Python_SetConstant(d, "RIG_PORT_SERIAL",SWIG_From_int((int)(RIG_PORT_SERIAL)));
  SWIG_Python_SetConstant(d, "RIG_PORT_NETWORK",SWIG_From_int((int)(RIG_PORT_NETWORK)));
  SWIG_Python_SetConstant(d, "RIG_PORT_DEVICE",SWIG_From_int((int)(RIG_PORT_DEVICE)));
  SWIG_Python_SetConstant(d, "RIG_PORT_PACKET",SWIG_From_int((int)(RIG_PORT_PACKET)));
  SWIG_Python_SetConstant(d, "RIG_PORT_DTMF",SWIG_From_int((int)(RIG_PORT_DTMF)));
  SWIG_Python_SetConstant(d, "RIG_PORT_ULTRA",SWIG_From_int((int)(RIG_PORT_ULTRA)));
  SWIG_Python_SetConstant(d, "RIG_PORT_RPC",SWIG_From_int((int)(RIG_PORT_RPC)));
  SWIG_Python_SetConstant(d, "RIG_PORT_PARALLEL",SWIG_From_int((int)(RIG_PORT_PARALLEL)));
  SWIG_Python_SetConstant(d, "RIG_PORT_USB",SWIG_From_int((int)(RIG_PORT_USB)));
  SWIG_Python_SetConstant(d, "RIG_PORT_UDP_NETWORK",SWIG_From_int((int)(RIG_PORT_UDP_NETWORK)));
  SWIG_Python_SetConstant(d, "RIG_PORT_CM108",SWIG_From_int((int)(RIG_PORT_CM108)));
  SWIG_Python_SetConstant(d, "RIG_PORT_GPIO",SWIG_From_int((int)(RIG_PORT_GPIO)));
  SWIG_Python_SetConstant(d, "RIG_PORT_GPION",SWIG_From_int((int)(RIG_PORT_GPION)));
  SWIG_Python_SetConstant(d, "RIG_PARITY_NONE",SWIG_From_int((int)(RIG_PARITY_NONE)));
  SWIG_Python_SetConstant(d, "RIG_PARITY_ODD",SWIG_From_int((int)(RIG_PARITY_ODD)));
  SWIG_Python_SetConstant(d, "RIG_PARITY_EVEN",SWIG_From_int((int)(RIG_PARITY_EVEN)));
  SWIG_Python_SetConstant(d, "RIG_PARITY_MARK",SWIG_From_int((int)(RIG_PARITY_MARK)));
  SWIG_Python_SetConstant(d, "RIG_PARITY_SPACE",SWIG_From_int((int)(RIG_PARITY_SPACE)));
  SWIG_Python_SetConstant(d, "RIG_HANDSHAKE_NONE",SWIG_From_int((int)(RIG_HANDSHAKE_NONE)));
  SWIG_Python_SetConstant(d, "RIG_HANDSHAKE_XONXOFF",SWIG_From_int((int)(RIG_HANDSHAKE_XONXOFF)));
  SWIG_Python_SetConstant(d, "RIG_HANDSHAKE_HARDWARE",SWIG_From_int((int)(RIG_HANDSHAKE_HARDWARE)));
  SWIG_Python_SetConstant(d, "RIG_SIGNAL_UNSET",SWIG_From_int((int)(RIG_SIGNAL_UNSET)));
  SWIG_Python_SetConstant(d, "RIG_SIGNAL_ON",SWIG_From_int((int)(RIG_SIGNAL_ON)));
  SWIG_Python_SetConstant(d, "RIG_SIGNAL_OFF",SWIG_From_int((int)(RIG_SIGNAL_OFF)));
  SWIG_Python_SetConstant(d, "RIG_FLAG_RECEIVER",SWIG_From_int((int)(RIG_FLAG_RECEIVER)));
  SWIG_Python_SetConstant(d, "RIG_FLAG_TRANSMITTER",SWIG_From_int((int)(RIG_FLAG_TRANSMITTER)));
  SWIG_Python_SetConstant(d, "RIG_FLAG_SCANNER",SWIG_From_int((int)(RIG_FLAG_SCANNER)));
  SWIG_Python_SetConstant(d, "RIG_FLAG_MOBILE",SWIG_From_int((int)(RIG_FLAG_MOBILE)));
  SWIG_Python_SetConstant(d, "RIG_FLAG_HANDHELD",SWIG_From_int((int)(RIG_FLAG_HANDHELD)));
  SWIG_Python_SetConstant(d, "RIG_FLAG_COMPUTER",SWIG_From_int((int)(RIG_FLAG_COMPUTER)));
  SWIG_Python_SetConstant(d, "RIG_FLAG_TRUNKING",SWIG_From_int((int)(RIG_FLAG_TRUNKING)));
  SWIG_Python_SetConstant(d, "RIG_FLAG_APRS",SWIG_From_int((int)(RIG_FLAG_APRS)));
  SWIG_Python_SetConstant(d, "RIG_FLAG_TNC",SWIG_From_int((int)(RIG_FLAG_TNC)));
  SWIG_Python_SetConstant(d, "RIG_FLAG_DXCLUSTER",SWIG_From_int((int)(RIG_FLAG_DXCLUSTER)));
  SWIG_Python_SetConstant(d, "RIG_FLAG_TUNER",SWIG_From_int((int)(RIG_FLAG_TUNER)));
  SWIG_Python_SetConstant(d, "RIG_AGC_OFF",SWIG_From_int((int)(RIG_AGC_OFF)));
  SWIG_Python_SetConstant(d, "RIG_AGC_SUPERFAST",SWIG_From_int((int)(RIG_AGC_SUPERFAST)));
  SWIG_Python_SetConstant(d, "RIG_AGC_FAST",SWIG_From_int((int)(RIG_AGC_FAST)));
  SWIG_Python_SetConstant(d, "RIG_AGC_SLOW",SWIG_From_int((int)(RIG_AGC_SLOW)));
  SWIG_Python_SetConstant(d, "RIG_AGC_USER",SWIG_From_int((int)(RIG_AGC_USER)));
  SWIG_Python_SetConstant(d, "RIG_AGC_MEDIUM",SWIG_From_int((int)(RIG_AGC_MEDIUM)));
  SWIG_Python_SetConstant(d, "RIG_AGC_AUTO",SWIG_From_int((int)(RIG_AGC_AUTO)));
  SWIG_Python_SetConstant(d, "RIG_AGC_LONG",SWIG_From_int((int)(RIG_AGC_LONG)));
  SWIG_Python_SetConstant(d, "RIG_AGC_ON",SWIG_From_int((int)(RIG_AGC_ON)));
  SWIG_Python_SetConstant(d, "RIG_AGC_NONE",SWIG_From_int((int)(RIG_AGC_NONE)));
  SWIG_Python_SetConstant(d, "RIG_TYPE_OTHER",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "RIG_STATUS_ALPHA",SWIG_From_int((int)(RIG_STATUS_ALPHA)));
  SWIG_Python_SetConstant(d, "RIG_STATUS_UNTESTED",SWIG_From_int((int)(RIG_STATUS_UNTESTED)));
  SWIG_Python_SetConstant(d, "RIG_STATUS_BETA",SWIG_From_int((int)(RIG_STATUS_BETA)));
  SWIG_Python_SetConstant(d, "RIG_STATUS_STABLE",SWIG_From_int((int)(RIG_STATUS_STABLE)));
  SWIG_Python_SetConstant(d, "RIG_STATUS_BUGGY",SWIG_From_int((int)(RIG_STATUS_BUGGY)));
  SWIG_Python_SetConstant(d, "RIG_RPT_SHIFT_NONE",SWIG_From_int((int)(RIG_RPT_SHIFT_NONE)));
  SWIG_Python_SetConstant(d, "RIG_RPT_SHIFT_MINUS",SWIG_From_int((int)(RIG_RPT_SHIFT_MINUS)));
  SWIG_Python_SetConstant(d, "RIG_RPT_SHIFT_PLUS",SWIG_From_int((int)(RIG_RPT_SHIFT_PLUS)));
  SWIG_Python_SetConstant(d, "RIG_SPLIT_OFF",SWIG_From_int((int)(RIG_SPLIT_OFF)));
  SWIG_Python_SetConstant(d, "RIG_SPLIT_ON",SWIG_From_int((int)(RIG_SPLIT_ON)));
  SWIG_Python_SetConstant(d, "RIG_VFO_NONE",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "RIG_VFO_A",SWIG_From_unsigned_SS_int((unsigned int)((1u << (0)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_B",SWIG_From_unsigned_SS_int((unsigned int)((1u << (1)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_C",SWIG_From_unsigned_SS_int((unsigned int)((1u << (2)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_SUB_A",SWIG_From_unsigned_SS_int((unsigned int)((1u << (21)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_SUB_B",SWIG_From_unsigned_SS_int((unsigned int)((1u << (22)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_SUB_C",SWIG_From_unsigned_SS_int((unsigned int)((1u << (3)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_MAIN_A",SWIG_From_unsigned_SS_int((unsigned int)((1u << (23)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_MAIN_B",SWIG_From_unsigned_SS_int((unsigned int)((1u << (24)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_MAIN_C",SWIG_From_unsigned_SS_int((unsigned int)((1u << (4)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_OTHER",SWIG_From_unsigned_SS_int((unsigned int)((1u << (5)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_SUB",SWIG_From_unsigned_SS_int((unsigned int)((1u << (25)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_MAIN",SWIG_From_unsigned_SS_int((unsigned int)((1u << (26)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_VFO",SWIG_From_unsigned_SS_int((unsigned int)((1u << (27)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_MEM",SWIG_From_unsigned_SS_int((unsigned int)((1u << (28)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_CURR",SWIG_From_unsigned_SS_int((unsigned int)((1u << (29)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_TX_FLAG",SWIG_From_unsigned_SS_int((unsigned int)((1u << (30)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_ALL",SWIG_From_unsigned_SS_int((unsigned int)((1u << (31)))));
  SWIG_Python_SetConstant(d, "RIG_VFO_TX",SWIG_From_unsigned_SS_int((unsigned int)((((1u << (29)))|(1u << (30))))));
  SWIG_Python_SetConstant(d, "RIG_VFO_RX",SWIG_From_unsigned_SS_int((unsigned int)((1u << (29)))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_NONE",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_FREQ",SWIG_From_int((int)((1 << 0))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_MODE",SWIG_From_int((int)((1 << 1))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_PURE",SWIG_From_int((int)((1 << 2))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_TONE",SWIG_From_int((int)((1 << 3))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_FUNC",SWIG_From_int((int)((1 << 4))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_LEVEL",SWIG_From_int((int)((1 << 5))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_RITXIT",SWIG_From_int((int)((1 << 6))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_PTT",SWIG_From_int((int)((1 << 7))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_MEM",SWIG_From_int((int)((1 << 8))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_BANK",SWIG_From_int((int)((1 << 9))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_ANT",SWIG_From_int((int)((1 << 10))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_ROOFING",SWIG_From_int((int)((1 << 11))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_SPECTRUM",SWIG_From_int((int)((1 << 12))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_BAND",SWIG_From_int((int)((1 << 13))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_COMMON",SWIG_From_int((int)(((1 << 6)|(1 << 7)|(1 << 8)|(1 << 9)))));
  SWIG_Python_SetConstant(d, "RIG_TARGETABLE_ALL",SWIG_From_int((int)(0x7fffffff)));
  SWIG_Python_SetConstant(d, "VFO_DUAL",SWIG_From_unsigned_SS_int((unsigned int)(((1u << (23))|(1u << (24))|(1u << (21))|(1u << (22))))));
  SWIG_Python_SetConstant(d, "RIG_DCD_OFF",SWIG_From_int((int)(RIG_DCD_OFF)));
  SWIG_Python_SetConstant(d, "RIG_DCD_ON",SWIG_From_int((int)(RIG_DCD_ON)));
  SWIG_Python_SetConstant(d, "RIG_DCD_NONE",SWIG_From_int((int)(RIG_DCD_NONE)));
  SWIG_Python_SetConstant(d, "RIG_DCD_RIG",SWIG_From_int((int)(RIG_DCD_RIG)));
  SWIG_Python_SetConstant(d, "RIG_DCD_SERIAL_DSR",SWIG_From_int((int)(RIG_DCD_SERIAL_DSR)));
  SWIG_Python_SetConstant(d, "RIG_DCD_SERIAL_CTS",SWIG_From_int((int)(RIG_DCD_SERIAL_CTS)));
  SWIG_Python_SetConstant(d, "RIG_DCD_SERIAL_CAR",SWIG_From_int((int)(RIG_DCD_SERIAL_CAR)));
  SWIG_Python_SetConstant(d, "RIG_DCD_PARALLEL",SWIG_From_int((int)(RIG_DCD_PARALLEL)));
  SWIG_Python_SetConstant(d, "RIG_DCD_CM108",SWIG_From_int((int)(RIG_DCD_CM108)));
  SWIG_Python_SetConstant(d, "RIG_DCD_GPIO",SWIG_From_int((int)(RIG_DCD_GPIO)));
  SWIG_Python_SetConstant(d, "RIG_DCD_GPION",SWIG_From_int((int)(RIG_DCD_GPION)));
  SWIG_Python_SetConstant(d, "RIG_PTT_OFF",SWIG_From_int((int)(RIG_PTT_OFF)));
  SWIG_Python_SetConstant(d, "RIG_PTT_ON",SWIG_From_int((int)(RIG_PTT_ON)));
  SWIG_Python_SetConstant(d, "RIG_PTT_ON_MIC",SWIG_From_int((int)(RIG_PTT_ON_MIC)));
  SWIG_Python_SetConstant(d, "RIG_PTT_ON_DATA",SWIG_From_int((int)(RIG_PTT_ON_DATA)));
  SWIG_Python_SetConstant(d, "RIG_PTT_NONE",SWIG_From_int((int)(RIG_PTT_NONE)));
  SWIG_Python_SetConstant(d, "RIG_PTT_RIG",SWIG_From_int((int)(RIG_PTT_RIG)));
  SWIG_Python_SetConstant(d, "RIG_PTT_SERIAL_DTR",SWIG_From_int((int)(RIG_PTT_SERIAL_DTR)));
  SWIG_Python_SetConstant(d, "RIG_PTT_SERIAL_RTS",SWIG_From_int((int)(RIG_PTT_SERIAL_RTS)));
  SWIG_Python_SetConstant(d, "RIG_PTT_PARALLEL",SWIG_From_int((int)(RIG_PTT_PARALLEL)));
  SWIG_Python_SetConstant(d, "RIG_PTT_RIG_MICDATA",SWIG_From_int((int)(RIG_PTT_RIG_MICDATA)));
  SWIG_Python_SetConstant(d, "RIG_PTT_CM108",SWIG_From_int((int)(RIG_PTT_CM108)));
  SWIG_Python_SetConstant(d, "RIG_PTT_GPIO",SWIG_From_int((int)(RIG_PTT_GPIO)));
  SWIG_Python_SetConstant(d, "RIG_PTT_GPION",SWIG_From_int((int)(RIG_PTT_GPION)));
  SWIG_Python_SetConstant(d, "RIG_POWER_OFF",SWIG_From_int((int)(RIG_POWER_OFF)));
  SWIG_Python_SetConstant(d, "RIG_POWER_ON",SWIG_From_int((int)(RIG_POWER_ON)));
  SWIG_Python_SetConstant(d, "RIG_POWER_STANDBY",SWIG_From_int((int)(RIG_POWER_STANDBY)));
  SWIG_Python_SetConstant(d, "RIG_POWER_OPERATE",SWIG_From_int((int)(RIG_POWER_OPERATE)));
  SWIG_Python_SetConstant(d, "RIG_POWER_UNKNOWN",SWIG_From_int((int)(RIG_POWER_UNKNOWN)));
  SWIG_Python_SetConstant(d, "RIG_RESET_NONE",SWIG_From_int((int)(RIG_RESET_NONE)));
  SWIG_Python_SetConstant(d, "RIG_RESET_SOFT",SWIG_From_int((int)(RIG_RESET_SOFT)));
  SWIG_Python_SetConstant(d, "RIG_RESET_VFO",SWIG_From_int((int)(RIG_RESET_VFO)));
  SWIG_Python_SetConstant(d, "RIG_RESET_MCALL",SWIG_From_int((int)(RIG_RESET_MCALL)));
  SWIG_Python_SetConstant(d, "RIG_RESET_MASTER",SWIG_From_int((int)(RIG_RESET_MASTER)));
  SWIG_Python_SetConstant(d, "RIG_CLIENT_UNKNOWN",SWIG_From_int((int)(RIG_CLIENT_UNKNOWN)));
  SWIG_Python_SetConstant(d, "RIG_CLIENT_WSJTX",SWIG_From_int((int)(RIG_CLIENT_WSJTX)));
  SWIG_Python_SetConstant(d, "RIG_CLIENT_GPREDICT",SWIG_From_int((int)(RIG_CLIENT_GPREDICT)));
  SWIG_Python_SetConstant(d, "RIG_OP_NONE",SWIG_From_int((int)(RIG_OP_NONE)));
  SWIG_Python_SetConstant(d, "RIG_OP_CPY",SWIG_From_int((int)(RIG_OP_CPY)));
  SWIG_Python_SetConstant(d, "RIG_OP_XCHG",SWIG_From_int((int)(RIG_OP_XCHG)));
  SWIG_Python_SetConstant(d, "RIG_OP_FROM_VFO",SWIG_From_int((int)(RIG_OP_FROM_VFO)));
  SWIG_Python_SetConstant(d, "RIG_OP_TO_VFO",SWIG_From_int((int)(RIG_OP_TO_VFO)));
  SWIG_Python_SetConstant(d, "RIG_OP_MCL",SWIG_From_int((int)(RIG_OP_MCL)));
  SWIG_Python_SetConstant(d, "RIG_OP_UP",SWIG_From_int((int)(RIG_OP_UP)));
  SWIG_Python_SetConstant(d, "RIG_OP_DOWN",SWIG_From_int((int)(RIG_OP_DOWN)));
  SWIG_Python_SetConstant(d, "RIG_OP_BAND_UP",SWIG_From_int((int)(RIG_OP_BAND_UP)));
  SWIG_Python_SetConstant(d, "RIG_OP_BAND_DOWN",SWIG_From_int((int)(RIG_OP_BAND_DOWN)));
  SWIG_Python_SetConstant(d, "RIG_OP_LEFT",SWIG_From_int((int)(RIG_OP_LEFT)));
  SWIG_Python_SetConstant(d, "RIG_OP_RIGHT",SWIG_From_int((int)(RIG_OP_RIGHT)));
  SWIG_Python_SetConstant(d, "RIG_OP_TUNE",SWIG_From_int((int)(RIG_OP_TUNE)));
  SWIG_Python_SetConstant(d, "RIG_OP_TOGGLE",SWIG_From_int((int)(RIG_OP_TOGGLE)));
  SWIG_Python_SetConstant(d, "RIG_BAND_160M",SWIG_From_int((int)(RIG_BAND_160M)));
  SWIG_Python_SetConstant(d, "RIG_BAND_80M",SWIG_From_int((int)(RIG_BAND_80M)));
  SWIG_Python_SetConstant(d, "RIG_BAND_60M",SWIG_From_int((int)(RIG_BAND_60M)));
  SWIG_Python_SetConstant(d, "RIG_BAND_40M",SWIG_From_int((int)(RIG_BAND_40M)));
  SWIG_Python_SetConstant(d, "RIG_BAND_30M",SWIG_From_int((int)(RIG_BAND_30M)));
  SWIG_Python_SetConstant(d, "RIG_BAND_20M",SWIG_From_int((int)(RIG_BAND_20M)));
  SWIG_Python_SetConstant(d, "RIG_BAND_17M",SWIG_From_int((int)(RIG_BAND_17M)));
  SWIG_Python_SetConstant(d, "RIG_BAND_15M",SWIG_From_int((int)(RIG_BAND_15M)));
  SWIG_Python_SetConstant(d, "RIG_BAND_12M",SWIG_From_int((int)(RIG_BAND_12M)));
  SWIG_Python_SetConstant(d, "RIG_BAND_10M",SWIG_From_int((int)(RIG_BAND_10M)));
  SWIG_Python_SetConstant(d, "RIG_BAND_6M",SWIG_From_int((int)(RIG_BAND_6M)));
  SWIG_Python_SetConstant(d, "RIG_BAND_GEN",SWIG_From_int((int)(RIG_BAND_GEN)));
  SWIG_Python_SetConstant(d, "RIG_BAND_MW",SWIG_From_int((int)(RIG_BAND_MW)));
  SWIG_Python_SetConstant(d, "RIG_BAND_UNUSED",SWIG_From_int((int)(RIG_BAND_UNUSED)));
  SWIG_Python_SetConstant(d, "RIG_BAND_AIR",SWIG_From_int((int)(RIG_BAND_AIR)));
  SWIG_Python_SetConstant(d, "RIG_BAND_144MHZ",SWIG_From_int((int)(RIG_BAND_144MHZ)));
  SWIG_Python_SetConstant(d, "RIG_BAND_430MHZ",SWIG_From_int((int)(RIG_BAND_430MHZ)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_UNUSED",SWIG_From_int((int)(RIG_BANDSELECT_UNUSED)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_2200M",SWIG_From_int((int)(RIG_BANDSELECT_2200M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_600M",SWIG_From_int((int)(RIG_BANDSELECT_600M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_160M",SWIG_From_int((int)(RIG_BANDSELECT_160M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_80M",SWIG_From_int((int)(RIG_BANDSELECT_80M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_60M",SWIG_From_int((int)(RIG_BANDSELECT_60M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_40M",SWIG_From_int((int)(RIG_BANDSELECT_40M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_30M",SWIG_From_int((int)(RIG_BANDSELECT_30M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_20M",SWIG_From_int((int)(RIG_BANDSELECT_20M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_17M",SWIG_From_int((int)(RIG_BANDSELECT_17M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_15M",SWIG_From_int((int)(RIG_BANDSELECT_15M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_12M",SWIG_From_int((int)(RIG_BANDSELECT_12M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_10M",SWIG_From_int((int)(RIG_BANDSELECT_10M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_6M",SWIG_From_int((int)(RIG_BANDSELECT_6M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_WFM",SWIG_From_int((int)(RIG_BANDSELECT_WFM)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_GEN",SWIG_From_int((int)(RIG_BANDSELECT_GEN)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_MW",SWIG_From_int((int)(RIG_BANDSELECT_MW)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_AIR",SWIG_From_int((int)(RIG_BANDSELECT_AIR)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_4M",SWIG_From_int((int)(RIG_BANDSELECT_4M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_2M",SWIG_From_int((int)(RIG_BANDSELECT_2M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_1_25M",SWIG_From_int((int)(RIG_BANDSELECT_1_25M)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_70CM",SWIG_From_int((int)(RIG_BANDSELECT_70CM)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_33CM",SWIG_From_int((int)(RIG_BANDSELECT_33CM)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_23CM",SWIG_From_int((int)(RIG_BANDSELECT_23CM)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_13CM",SWIG_From_int((int)(RIG_BANDSELECT_13CM)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_9CM",SWIG_From_int((int)(RIG_BANDSELECT_9CM)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_5CM",SWIG_From_int((int)(RIG_BANDSELECT_5CM)));
  SWIG_Python_SetConstant(d, "RIG_BANDSELECT_3CM",SWIG_From_int((int)(RIG_BANDSELECT_3CM)));
  SWIG_Python_SetConstant(d, "RIG_SCAN_NONE",SWIG_From_int((int)(RIG_SCAN_NONE)));
  SWIG_Python_SetConstant(d, "RIG_SCAN_MEM",SWIG_From_int((int)(RIG_SCAN_MEM)));
  SWIG_Python_SetConstant(d, "RIG_SCAN_SLCT",SWIG_From_int((int)(RIG_SCAN_SLCT)));
  SWIG_Python_SetConstant(d, "RIG_SCAN_PRIO",SWIG_From_int((int)(RIG_SCAN_PRIO)));
  SWIG_Python_SetConstant(d, "RIG_SCAN_PROG",SWIG_From_int((int)(RIG_SCAN_PROG)));
  SWIG_Python_SetConstant(d, "RIG_SCAN_DELTA",SWIG_From_int((int)(RIG_SCAN_DELTA)));
  SWIG_Python_SetConstant(d, "RIG_SCAN_VFO",SWIG_From_int((int)(RIG_SCAN_VFO)));
  SWIG_Python_SetConstant(d, "RIG_SCAN_PLT",SWIG_From_int((int)(RIG_SCAN_PLT)));
  SWIG_Python_SetConstant(d, "RIG_SCAN_STOP",SWIG_From_int((int)(RIG_SCAN_STOP)));
  SWIG_Python_SetConstant(d, "RIG_CONF_END",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "RIG_CONF_STRING",SWIG_From_int((int)(RIG_CONF_STRING)));
  SWIG_Python_SetConstant(d, "RIG_CONF_COMBO",SWIG_From_int((int)(RIG_CONF_COMBO)));
  SWIG_Python_SetConstant(d, "RIG_CONF_NUMERIC",SWIG_From_int((int)(RIG_CONF_NUMERIC)));
  SWIG_Python_SetConstant(d, "RIG_CONF_CHECKBUTTON",SWIG_From_int((int)(RIG_CONF_CHECKBUTTON)));
  SWIG_Python_SetConstant(d, "RIG_CONF_BUTTON",SWIG_From_int((int)(RIG_CONF_BUTTON)));
  SWIG_Python_SetConstant(d, "RIG_CONF_BINARY",SWIG_From_int((int)(RIG_CONF_BINARY)));
  SWIG_Python_SetConstant(d, "RIG_CONF_INT",SWIG_From_int((int)(RIG_CONF_INT)));
  SWIG_Python_SetConstant(d, "RIG_ANN_NONE",SWIG_From_int((int)(RIG_ANN_NONE)));
  SWIG_Python_SetConstant(d, "RIG_ANN_OFF",SWIG_From_int((int)(RIG_ANN_OFF)));
  SWIG_Python_SetConstant(d, "RIG_ANN_FREQ",SWIG_From_int((int)(RIG_ANN_FREQ)));
  SWIG_Python_SetConstant(d, "RIG_ANN_RXMODE",SWIG_From_int((int)(RIG_ANN_RXMODE)));
  SWIG_Python_SetConstant(d, "RIG_ANN_CW",SWIG_From_int((int)(RIG_ANN_CW)));
  SWIG_Python_SetConstant(d, "RIG_ANN_ENG",SWIG_From_int((int)(RIG_ANN_ENG)));
  SWIG_Python_SetConstant(d, "RIG_ANN_JAP",SWIG_From_int((int)(RIG_ANN_JAP)));
  SWIG_Python_SetConstant(d, "RIG_ANT_NONE",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "RIG_ANT_MAX",SWIG_From_int((int)(32)));
  SWIG_Python_SetConstant(d, "RIG_AGC_LAST",SWIG_From_int((int)(99999)));
  SWIG_Python_SetConstant(d, "RIG_METER_NONE",SWIG_From_int((int)(RIG_METER_NONE)));
  SWIG_Python_SetConstant(d, "RIG_METER_SWR",SWIG_From_int((int)(RIG_METER_SWR)));
  SWIG_Python_SetConstant(d, "RIG_METER_COMP",SWIG_From_int((int)(RIG_METER_COMP)));
  SWIG_Python_SetConstant(d, "RIG_METER_ALC",SWIG_From_int((int)(RIG_METER_ALC)));
  SWIG_Python_SetConstant(d, "RIG_METER_IC",SWIG_From_int((int)(RIG_METER_IC)));
  SWIG_Python_SetConstant(d, "RIG_METER_DB",SWIG_From_int((int)(RIG_METER_DB)));
  SWIG_Python_SetConstant(d, "RIG_METER_PO",SWIG_From_int((int)(RIG_METER_PO)));
  SWIG_Python_SetConstant(d, "RIG_METER_VDD",SWIG_From_int((int)(RIG_METER_VDD)));
  SWIG_Python_SetConstant(d, "RIG_METER_TEMP",SWIG_From_int((int)(RIG_METER_TEMP)));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_NONE",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_PREAMP",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (0)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_ATT",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (1)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_VOXDELAY",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (2)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_AF",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (3)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_RF",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (4)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_SQL",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (5)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_IF",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (6)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_APF",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (7)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_NR",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (8)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_PBT_IN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (9)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_PBT_OUT",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (10)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_CWPITCH",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (11)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_RFPOWER",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (12)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_MICGAIN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (13)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_KEYSPD",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (14)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_NOTCHF",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (15)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_COMP",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (16)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_AGC",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (17)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_BKINDL",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (18)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_BALANCE",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (19)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_METER",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (20)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_VOXGAIN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (21)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_ANTIVOX",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (22)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_SLOPE_LOW",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (23)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_SLOPE_HIGH",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (24)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_BKIN_DLYMS",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (25)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_RAWSTR",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (26)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_SWR",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (28)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_ALC",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (29)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_STRENGTH",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (30)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_RFPOWER_METER",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (32)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_COMP_METER",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (33)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_VD_METER",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (34)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_ID_METER",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (35)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_NOTCHF_RAW",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (36)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_MONITOR_GAIN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (37)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_NB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (38)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_RFPOWER_METER_WATTS",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (39)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_SPECTRUM_MODE",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (40)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_SPECTRUM_SPAN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (41)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_SPECTRUM_EDGE_LOW",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (42)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_SPECTRUM_EDGE_HIGH",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (43)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_SPECTRUM_SPEED",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (44)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_SPECTRUM_REF",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (45)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_SPECTRUM_AVG",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (46)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_SPECTRUM_ATT",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (47)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_TEMP_METER",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (48)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_BAND_SELECT",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (49)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_USB_AF",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (50)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_USB_AF_INPUT",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (51)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_AGC_TIME",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (52)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_53",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (53)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_54",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (54)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_55",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (55)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_56",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (56)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_57",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (57)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_58",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (58)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_59",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (59)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_60",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (60)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_61",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (61)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_62",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (62)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_63",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (63)))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_FLOAT_LIST",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))));
  SWIG_Python_SetConstant(d, "RIG_LEVEL_READONLY_LIST",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(((1ull << (28))|(1ull << (29))|(1ull << (30))|(1ull << (26))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (48))|(1ull << (32))|(1ull << (39))))));
  SWIG_Python_SetConstant(d, "RIG_PARM_NONE",SWIG_From_int((int)(RIG_PARM_NONE)));
  SWIG_Python_SetConstant(d, "RIG_PARM_ANN",SWIG_From_int((int)(RIG_PARM_ANN)));
  SWIG_Python_SetConstant(d, "RIG_PARM_APO",SWIG_From_int((int)(RIG_PARM_APO)));
  SWIG_Python_SetConstant(d, "RIG_PARM_BACKLIGHT",SWIG_From_int((int)(RIG_PARM_BACKLIGHT)));
  SWIG_Python_SetConstant(d, "RIG_PARM_BEEP",SWIG_From_int((int)(RIG_PARM_BEEP)));
  SWIG_Python_SetConstant(d, "RIG_PARM_TIME",SWIG_From_int((int)(RIG_PARM_TIME)));
  SWIG_Python_SetConstant(d, "RIG_PARM_BAT",SWIG_From_int((int)(RIG_PARM_BAT)));
  SWIG_Python_SetConstant(d, "RIG_PARM_KEYLIGHT",SWIG_From_int((int)(RIG_PARM_KEYLIGHT)));
  SWIG_Python_SetConstant(d, "RIG_PARM_SCREENSAVER",SWIG_From_int((int)(RIG_PARM_SCREENSAVER)));
  SWIG_Python_SetConstant(d, "RIG_PARM_AFIF",SWIG_From_int((int)(RIG_PARM_AFIF)));
  SWIG_Python_SetConstant(d, "RIG_PARM_BANDSELECT",SWIG_From_int((int)(RIG_PARM_BANDSELECT)));
  SWIG_Python_SetConstant(d, "RIG_PARM_KEYERTYPE",SWIG_From_int((int)(RIG_PARM_KEYERTYPE)));
  SWIG_Python_SetConstant(d, "RIG_PARM_AFIF_LAN",SWIG_From_int((int)(RIG_PARM_AFIF_LAN)));
  SWIG_Python_SetConstant(d, "RIG_PARM_AFIF_WLAN",SWIG_From_int((int)(RIG_PARM_AFIF_WLAN)));
  SWIG_Python_SetConstant(d, "RIG_PARM_AFIF_ACC",SWIG_From_int((int)(RIG_PARM_AFIF_ACC)));
  SWIG_Python_SetConstant(d, "RIG_KEYERTYPE_STRAIGHT",SWIG_From_int((int)(RIG_KEYERTYPE_STRAIGHT)));
  SWIG_Python_SetConstant(d, "RIG_KEYERTYPE_BUG",SWIG_From_int((int)(RIG_KEYERTYPE_BUG)));
  SWIG_Python_SetConstant(d, "RIG_KEYERTYPE_PADDLE",SWIG_From_int((int)(RIG_KEYERTYPE_PADDLE)));
  SWIG_Python_SetConstant(d, "RIG_KEYERTYPE_UNKNOWN",SWIG_From_int((int)(RIG_KEYERTYPE_UNKNOWN)));
  SWIG_Python_SetConstant(d, "RIG_COOKIE_GET",SWIG_From_int((int)(RIG_COOKIE_GET)));
  SWIG_Python_SetConstant(d, "RIG_COOKIE_RELEASE",SWIG_From_int((int)(RIG_COOKIE_RELEASE)));
  SWIG_Python_SetConstant(d, "RIG_COOKIE_RENEW",SWIG_From_int((int)(RIG_COOKIE_RENEW)));
  SWIG_Python_SetConstant(d, "RIG_MULTICAST_POLL",SWIG_From_int((int)(RIG_MULTICAST_POLL)));
  SWIG_Python_SetConstant(d, "RIG_MULTICAST_TRANSCEIVE",SWIG_From_int((int)(RIG_MULTICAST_TRANSCEIVE)));
  SWIG_Python_SetConstant(d, "RIG_MULTICAST_SPECTRUM",SWIG_From_int((int)(RIG_MULTICAST_SPECTRUM)));
  SWIG_Python_SetConstant(d, "RIG_SETTING_MAX",SWIG_From_int((int)(64)));
  SWIG_Python_SetConstant(d, "RIG_TRN_OFF",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "RIG_TRN_RIG",SWIG_From_int((int)(1)));
  SWIG_Python_SetConstant(d, "RIG_TRN_POLL",SWIG_From_int((int)(2)));
  SWIG_Python_SetConstant(d, "RIG_FUNC_NONE",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "RIG_FUNC_FAGC",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (0)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_NB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (1)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_COMP",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (2)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_VOX",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (3)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_TONE",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (4)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_TSQL",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (5)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_SBKIN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (6)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_FBKIN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (7)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_ANF",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (8)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_NR",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (9)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_AIP",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (10)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_APF",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (11)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_MON",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (12)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_MN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (13)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_RF",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (14)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_ARO",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (15)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_LOCK",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (16)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_MUTE",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (17)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_VSC",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (18)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_REV",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (19)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_SQL",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (20)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_ABM",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (21)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BC",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (22)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_MBC",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (23)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_RIT",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (24)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_AFC",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (25)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_SATMODE",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (26)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_SCOPE",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (27)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_RESUME",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (28)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_TBURST",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (29)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_TUNER",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (30)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_XIT",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (31)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_NB2",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (32)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_CSQL",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (33)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_AFLT",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (34)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_ANL",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (35)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BC2",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (36)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_DUAL_WATCH",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (37)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_DIVERSITY",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (38)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_DSQL",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (39)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_SCEN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (40)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_SLICE",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (41)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_TRANSCEIVE",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (42)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_SPECTRUM",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (43)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_SPECTRUM_HOLD",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (44)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_SEND_MORSE",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (45)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_SEND_VOICE_MEM",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (46)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_OVF_STATUS",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (47)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_SYNC",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (48)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT49",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (49)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT50",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (50)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT51",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (51)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT52",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (52)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT53",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (53)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT54",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (54)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT55",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (55)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT56",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (56)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT57",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (57)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT58",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (58)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT59",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (59)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT60",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (60)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT61",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (61)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT62",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (62)))));
  SWIG_Python_SetConstant(d, "RIG_FUNC_BIT63",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (63)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_NONE",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "RIG_MODE_AM",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (0)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_CW",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (1)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_USB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (2)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_LSB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (3)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_RTTY",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (4)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_FM",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (5)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_WFM",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (6)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_CWR",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (7)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_RTTYR",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (8)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_AMS",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (9)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_PKTLSB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (10)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_PKTUSB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (11)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_PKTFM",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (12)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_ECSSUSB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (13)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_ECSSLSB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (14)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_FAX",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (15)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_SAM",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (16)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_SAL",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (17)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_SAH",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (18)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_DSB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (19)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_FMN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (21)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_PKTAM",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (22)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_P25",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (23)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_DSTAR",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (24)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_DPMR",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (25)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_NXDNVN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (26)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_NXDN_N",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (27)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_DCR",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (28)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_AMN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (29)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_PSK",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (30)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_PSKR",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (31)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_DD",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (32)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_C4FM",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (33)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_PKTFMN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (34)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_SPEC",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (35)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_CWN",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (36)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_IQ",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (37)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_ISBUSB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (38)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_ISBLSB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (39)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_USBD1",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (40)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_USBD2",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (41)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_USBD3",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (42)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_LSBD1",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (43)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_LSBD2",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (44)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_LSBD3",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (45)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_WFMS",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (46)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT47",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (47)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT48",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (48)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT49",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (49)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT50",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (50)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT51",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (51)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT52",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (52)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT53",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (53)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT54",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (54)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT55",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (55)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT56",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (56)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT57",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (57)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT58",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (58)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT59",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (59)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT60",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (60)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT61",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (61)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_BIT62",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (62)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_TESTS_MAX",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (63)))));
  SWIG_Python_SetConstant(d, "RIG_MODE_ALL",SWIG_From_int((int)((0xffffffff))));
  SWIG_Python_SetConstant(d, "RIG_MODE_SSB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(((1ull << (2))|(1ull << (3))))));
  SWIG_Python_SetConstant(d, "RIG_MODE_PKTSSB",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(((1ull << (11))|(1ull << (10))))));
  SWIG_Python_SetConstant(d, "RIG_MODE_ECSS",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(((1ull << (13))|(1ull << (14))))));
  SWIG_Python_SetConstant(d, "RIG_TS_ANY",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "RIG_FLT_ANY",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "DEBUGMSGSAVE_SIZE",SWIG_From_int((int)(24000)));
  SWIG_Python_SetConstant(d, "RIG_CHFLAG_NONE",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "RIG_CHFLAG_SKIP",SWIG_From_int((int)((1 << 0))));
  SWIG_Python_SetConstant(d, "RIG_CHFLAG_DATA",SWIG_From_int((int)((1 << 1))));
  SWIG_Python_SetConstant(d, "RIG_CHFLAG_PSKIP",SWIG_From_int((int)((1 << 2))));
  SWIG_Python_SetConstant(d, "RIG_MTYPE_NONE",SWIG_From_int((int)(RIG_MTYPE_NONE)));
  SWIG_Python_SetConstant(d, "RIG_MTYPE_MEM",SWIG_From_int((int)(RIG_MTYPE_MEM)));
  SWIG_Python_SetConstant(d, "RIG_MTYPE_EDGE",SWIG_From_int((int)(RIG_MTYPE_EDGE)));
  SWIG_Python_SetConstant(d, "RIG_MTYPE_CALL",SWIG_From_int((int)(RIG_MTYPE_CALL)));
  SWIG_Python_SetConstant(d, "RIG_MTYPE_MEMOPAD",SWIG_From_int((int)(RIG_MTYPE_MEMOPAD)));
  SWIG_Python_SetConstant(d, "RIG_MTYPE_SAT",SWIG_From_int((int)(RIG_MTYPE_SAT)));
  SWIG_Python_SetConstant(d, "RIG_MTYPE_BAND",SWIG_From_int((int)(RIG_MTYPE_BAND)));
  SWIG_Python_SetConstant(d, "RIG_MTYPE_PRIO",SWIG_From_int((int)(RIG_MTYPE_PRIO)));
  SWIG_Python_SetConstant(d, "RIG_MTYPE_VOICE",SWIG_From_int((int)(RIG_MTYPE_VOICE)));
  SWIG_Python_SetConstant(d, "RIG_MTYPE_MORSE",SWIG_From_int((int)(RIG_MTYPE_MORSE)));
  SWIG_Python_SetConstant(d, "RIG_MTYPE_SPLIT",SWIG_From_int((int)(RIG_MTYPE_SPLIT)));
  SWIG_Python_SetConstant(d, "RIG_MEM_CAPS_ALL",SWIG_From_int((int)(-1)));
  SWIG_Python_SetConstant(d, "RIG_SPECTRUM_MODE_NONE",SWIG_From_int((int)(RIG_SPECTRUM_MODE_NONE)));
  SWIG_Python_SetConstant(d, "RIG_SPECTRUM_MODE_CENTER",SWIG_From_int((int)(RIG_SPECTRUM_MODE_CENTER)));
  SWIG_Python_SetConstant(d, "RIG_SPECTRUM_MODE_FIXED",SWIG_From_int((int)(RIG_SPECTRUM_MODE_FIXED)));
  SWIG_Python_SetConstant(d, "RIG_SPECTRUM_MODE_CENTER_SCROLL",SWIG_From_int((int)(RIG_SPECTRUM_MODE_CENTER_SCROLL)));
  SWIG_Python_SetConstant(d, "RIG_SPECTRUM_MODE_FIXED_SCROLL",SWIG_From_int((int)(RIG_SPECTRUM_MODE_FIXED_SCROLL)));
  SWIG_Python_SetConstant(d, "HAMLIB_CHECK_RIG_CAPS",SWIG_FromCharPtr("HAMLIB_CHECK_RIG_CAPS"));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_INIT",SWIG_From_int((int)(RIG_FUNCTION_INIT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_CLEANUP",SWIG_From_int((int)(RIG_FUNCTION_CLEANUP)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_OPEN",SWIG_From_int((int)(RIG_FUNCTION_OPEN)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_CLOSE",SWIG_From_int((int)(RIG_FUNCTION_CLOSE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_FREQ",SWIG_From_int((int)(RIG_FUNCTION_SET_FREQ)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_FREQ",SWIG_From_int((int)(RIG_FUNCTION_GET_FREQ)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_MODE",SWIG_From_int((int)(RIG_FUNCTION_SET_MODE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_MODE",SWIG_From_int((int)(RIG_FUNCTION_GET_MODE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_VFO",SWIG_From_int((int)(RIG_FUNCTION_SET_VFO)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_VFO",SWIG_From_int((int)(RIG_FUNCTION_GET_VFO)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_PTT",SWIG_From_int((int)(RIG_FUNCTION_SET_PTT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_PTT",SWIG_From_int((int)(RIG_FUNCTION_GET_PTT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_DCD",SWIG_From_int((int)(RIG_FUNCTION_GET_DCD)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_RPTR_SHIFT",SWIG_From_int((int)(RIG_FUNCTION_SET_RPTR_SHIFT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_RPTR_SHIFT",SWIG_From_int((int)(RIG_FUNCTION_GET_RPTR_SHIFT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_RPTR_OFFS",SWIG_From_int((int)(RIG_FUNCTION_SET_RPTR_OFFS)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_RPTR_OFFS",SWIG_From_int((int)(RIG_FUNCTION_GET_RPTR_OFFS)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_SPLIT_FREQ",SWIG_From_int((int)(RIG_FUNCTION_SET_SPLIT_FREQ)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_SPLIT_FREQ",SWIG_From_int((int)(RIG_FUNCTION_GET_SPLIT_FREQ)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_SPLIT_MODE",SWIG_From_int((int)(RIG_FUNCTION_SET_SPLIT_MODE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_SPLIT_FREQ_MODE",SWIG_From_int((int)(RIG_FUNCTION_SET_SPLIT_FREQ_MODE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_SPLIT_FREQ_MODE",SWIG_From_int((int)(RIG_FUNCTION_GET_SPLIT_FREQ_MODE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_SPLIT_VFO",SWIG_From_int((int)(RIG_FUNCTION_SET_SPLIT_VFO)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_SPLIT_VFO",SWIG_From_int((int)(RIG_FUNCTION_GET_SPLIT_VFO)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_RIT",SWIG_From_int((int)(RIG_FUNCTION_SET_RIT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_RIT",SWIG_From_int((int)(RIG_FUNCTION_GET_RIT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_XIT",SWIG_From_int((int)(RIG_FUNCTION_SET_XIT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_XIT",SWIG_From_int((int)(RIG_FUNCTION_GET_XIT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_TS",SWIG_From_int((int)(RIG_FUNCTION_SET_TS)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_TS",SWIG_From_int((int)(RIG_FUNCTION_GET_TS)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_DCS_CODE",SWIG_From_int((int)(RIG_FUNCTION_SET_DCS_CODE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_DCS_CODE",SWIG_From_int((int)(RIG_FUNCTION_GET_DCS_CODE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_TONE",SWIG_From_int((int)(RIG_FUNCTION_SET_TONE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_TONE",SWIG_From_int((int)(RIG_FUNCTION_GET_TONE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_CTCSS_TONE",SWIG_From_int((int)(RIG_FUNCTION_SET_CTCSS_TONE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_CTCSS_TONE",SWIG_From_int((int)(RIG_FUNCTION_GET_CTCSS_TONE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_DCS_SQL",SWIG_From_int((int)(RIG_FUNCTION_SET_DCS_SQL)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_DCS_SQL",SWIG_From_int((int)(RIG_FUNCTION_GET_DCS_SQL)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_TONE_SQL",SWIG_From_int((int)(RIG_FUNCTION_SET_TONE_SQL)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_TONE_SQL",SWIG_From_int((int)(RIG_FUNCTION_GET_TONE_SQL)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_CTCSS_SQL",SWIG_From_int((int)(RIG_FUNCTION_SET_CTCSS_SQL)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_CTCSS_SQL",SWIG_From_int((int)(RIG_FUNCTION_GET_CTCSS_SQL)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_POWER2MW",SWIG_From_int((int)(RIG_FUNCTION_POWER2MW)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_MW2POWER",SWIG_From_int((int)(RIG_FUNCTION_MW2POWER)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_POWERSTAT",SWIG_From_int((int)(RIG_FUNCTION_SET_POWERSTAT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_POWERSTAT",SWIG_From_int((int)(RIG_FUNCTION_GET_POWERSTAT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_RESET",SWIG_From_int((int)(RIG_FUNCTION_RESET)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_ANT",SWIG_From_int((int)(RIG_FUNCTION_SET_ANT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_ANT",SWIG_From_int((int)(RIG_FUNCTION_GET_ANT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_LEVEL",SWIG_From_int((int)(RIG_FUNCTION_SET_LEVEL)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_LEVEL",SWIG_From_int((int)(RIG_FUNCTION_GET_LEVEL)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_FUNC",SWIG_From_int((int)(RIG_FUNCTION_SET_FUNC)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_FUNC",SWIG_From_int((int)(RIG_FUNCTION_GET_FUNC)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_PARM",SWIG_From_int((int)(RIG_FUNCTION_SET_PARM)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_PARM",SWIG_From_int((int)(RIG_FUNCTION_GET_PARM)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_EXT_LEVEL",SWIG_From_int((int)(RIG_FUNCTION_SET_EXT_LEVEL)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_EXT_LEVEL",SWIG_From_int((int)(RIG_FUNCTION_GET_EXT_LEVEL)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_EXT_FUNC",SWIG_From_int((int)(RIG_FUNCTION_SET_EXT_FUNC)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_EXT_FUNC",SWIG_From_int((int)(RIG_FUNCTION_GET_EXT_FUNC)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_EXT_PARM",SWIG_From_int((int)(RIG_FUNCTION_SET_EXT_PARM)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_EXT_PARM",SWIG_From_int((int)(RIG_FUNCTION_GET_EXT_PARM)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_CONF",SWIG_From_int((int)(RIG_FUNCTION_SET_CONF)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_CONF",SWIG_From_int((int)(RIG_FUNCTION_GET_CONF)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SEND_DTMF",SWIG_From_int((int)(RIG_FUNCTION_SEND_DTMF)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SEND_MORSE",SWIG_From_int((int)(RIG_FUNCTION_SEND_MORSE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_STOP_MORSE",SWIG_From_int((int)(RIG_FUNCTION_STOP_MORSE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_WAIT_MORSE",SWIG_From_int((int)(RIG_FUNCTION_WAIT_MORSE)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SEND_VOICE_MEM",SWIG_From_int((int)(RIG_FUNCTION_SEND_VOICE_MEM)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_BANK",SWIG_From_int((int)(RIG_FUNCTION_SET_BANK)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_MEM",SWIG_From_int((int)(RIG_FUNCTION_SET_MEM)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_MEM",SWIG_From_int((int)(RIG_FUNCTION_GET_MEM)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_VFO_OP",SWIG_From_int((int)(RIG_FUNCTION_VFO_OP)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SCAN",SWIG_From_int((int)(RIG_FUNCTION_SCAN)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_TRN",SWIG_From_int((int)(RIG_FUNCTION_SET_TRN)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_TRN",SWIG_From_int((int)(RIG_FUNCTION_GET_TRN)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_DECODE_EVENT",SWIG_From_int((int)(RIG_FUNCTION_DECODE_EVENT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_CHANNEL",SWIG_From_int((int)(RIG_FUNCTION_SET_CHANNEL)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_CHANNEL",SWIG_From_int((int)(RIG_FUNCTION_GET_CHANNEL)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_INFO",SWIG_From_int((int)(RIG_FUNCTION_GET_INFO)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_CHAN_ALL_CB",SWIG_From_int((int)(RIG_FUNCTION_SET_CHAN_ALL_CB)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_CHAN_ALL_CB",SWIG_From_int((int)(RIG_FUNCTION_GET_CHAN_ALL_CB)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_MEM_ALL_CB",SWIG_From_int((int)(RIG_FUNCTION_SET_MEM_ALL_CB)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_MEM_ALL_CB",SWIG_From_int((int)(RIG_FUNCTION_GET_MEM_ALL_CB)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_SET_VFO_OPT",SWIG_From_int((int)(RIG_FUNCTION_SET_VFO_OPT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_READ_FRAME_DIRECT",SWIG_From_int((int)(RIG_FUNCTION_READ_FRAME_DIRECT)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_IS_ASYNC_FRAME",SWIG_From_int((int)(RIG_FUNCTION_IS_ASYNC_FRAME)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_PROCESS_ASYNC_FRAME",SWIG_From_int((int)(RIG_FUNCTION_PROCESS_ASYNC_FRAME)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_GET_CONF2",SWIG_From_int((int)(RIG_FUNCTION_GET_CONF2)));
  SWIG_Python_SetConstant(d, "RIG_FUNCTION_STOP_VOICE_MEM",SWIG_From_int((int)(RIG_FUNCTION_STOP_VOICE_MEM)));
  SWIG_Python_SetConstant(d, "RIG_CAPS_TARGETABLE_VFO",SWIG_From_int((int)(RIG_CAPS_TARGETABLE_VFO)));
  SWIG_Python_SetConstant(d, "RIG_CAPS_RIG_MODEL",SWIG_From_int((int)(RIG_CAPS_RIG_MODEL)));
  SWIG_Python_SetConstant(d, "RIG_CAPS_PORT_TYPE",SWIG_From_int((int)(RIG_CAPS_PORT_TYPE)));
  SWIG_Python_SetConstant(d, "RIG_CAPS_PTT_TYPE",SWIG_From_int((int)(RIG_CAPS_PTT_TYPE)));
  SWIG_Python_SetConstant(d, "RIG_CAPS_HAS_GET_LEVEL",SWIG_From_int((int)(RIG_CAPS_HAS_GET_LEVEL)));
  SWIG_Python_SetConstant(d, "RIG_CAPS_HAS_SET_LEVEL",SWIG_From_int((int)(RIG_CAPS_HAS_SET_LEVEL)));
  SWIG_Python_SetConstant(d, "RIG_CAPS_VERSION_CPTR",SWIG_From_int((int)(RIG_CAPS_VERSION_CPTR)));
  SWIG_Python_SetConstant(d, "RIG_CAPS_MFG_NAME_CPTR",SWIG_From_int((int)(RIG_CAPS_MFG_NAME_CPTR)));
  SWIG_Python_SetConstant(d, "RIG_CAPS_MODEL_NAME_CPTR",SWIG_From_int((int)(RIG_CAPS_MODEL_NAME_CPTR)));
  SWIG_Python_SetConstant(d, "RIG_CAPS_STATUS_CPTR",SWIG_From_int((int)(RIG_CAPS_STATUS_CPTR)));
  SWIG_Python_SetConstant(d, "RIG_PTRX_NONE",SWIG_From_int((int)(RIG_PTRX_NONE)));
  SWIG_Python_SetConstant(d, "RIG_PTRX_RIGPORT",SWIG_From_int((int)(RIG_PTRX_RIGPORT)));
  SWIG_Python_SetConstant(d, "RIG_PTRX_PTTPORT",SWIG_From_int((int)(RIG_PTRX_PTTPORT)));
  SWIG_Python_SetConstant(d, "RIG_PTRX_DCDPORT",SWIG_From_int((int)(RIG_PTRX_DCDPORT)));
  SWIG_Python_SetConstant(d, "RIG_PTRX_CACHE",SWIG_From_int((int)(RIG_PTRX_CACHE)));
  SWIG_Python_SetConstant(d, "RIG_PTRX_AMPPORT",SWIG_From_int((int)(RIG_PTRX_AMPPORT)));
  SWIG_Python_SetConstant(d, "RIG_PTRX_ROTPORT",SWIG_From_int((int)(RIG_PTRX_ROTPORT)));
  SWIG_Python_SetConstant(d, "RIG_PTRX_ROTPORT2",SWIG_From_int((int)(RIG_PTRX_ROTPORT2)));
  SWIG_Python_SetConstant(d, "RIG_PTRX_STATE",SWIG_From_int((int)(RIG_PTRX_STATE)));
  SWIG_Python_SetConstant(d, "RIG_PTRX_AMPSTATE",SWIG_From_int((int)(RIG_PTRX_AMPSTATE)));
  SWIG_Python_SetConstant(d, "RIG_PTRX_ROTSTATE",SWIG_From_int((int)(RIG_PTRX_ROTSTATE)));
  SWIG_Python_SetConstant(d, "RIG_PTRX_MAXIMUM",SWIG_From_int((int)(RIG_PTRX_MAXIMUM)));
  SWIG_Python_SetConstant(d, "HAMLIB_ELAPSED_GET",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "HAMLIB_ELAPSED_SET",SWIG_From_int((int)(1)));
  SWIG_Python_SetConstant(d, "HAMLIB_ELAPSED_INVALIDATE",SWIG_From_int((int)(2)));
  SWIG_Python_SetConstant(d, "HAMLIB_CACHE_ALWAYS",SWIG_From_int((int)((-1))));
  SWIG_Python_SetConstant(d, "HAMLIB_CACHE_ALL",SWIG_From_int((int)(HAMLIB_CACHE_ALL)));
  SWIG_Python_SetConstant(d, "HAMLIB_CACHE_VFO",SWIG_From_int((int)(HAMLIB_CACHE_VFO)));
  SWIG_Python_SetConstant(d, "HAMLIB_CACHE_FREQ",SWIG_From_int((int)(HAMLIB_CACHE_FREQ)));
  SWIG_Python_SetConstant(d, "HAMLIB_CACHE_MODE",SWIG_From_int((int)(HAMLIB_CACHE_MODE)));
  SWIG_Python_SetConstant(d, "HAMLIB_CACHE_PTT",SWIG_From_int((int)(HAMLIB_CACHE_PTT)));
  SWIG_Python_SetConstant(d, "HAMLIB_CACHE_SPLIT",SWIG_From_int((int)(HAMLIB_CACHE_SPLIT)));
  SWIG_Python_SetConstant(d, "HAMLIB_CACHE_WIDTH",SWIG_From_int((int)(HAMLIB_CACHE_WIDTH)));
  SWIG_Python_SetConstant(d, "TWIDDLE_OFF",SWIG_From_int((int)(TWIDDLE_OFF)));
  SWIG_Python_SetConstant(d, "TWIDDLE_ON",SWIG_From_int((int)(TWIDDLE_ON)));
  SWIG_Python_SetConstant(d, "RIG_COMM_STATUS_OK",SWIG_From_int((int)(0x00)));
  SWIG_Python_SetConstant(d, "RIG_COMM_STATUS_CONNECTING",SWIG_From_int((int)(0x01)));
  SWIG_Python_SetConstant(d, "RIG_COMM_STATUS_DISCONNECTED",SWIG_From_int((int)(0x02)));
  SWIG_Python_SetConstant(d, "RIG_COMM_STATUS_TERMINATED",SWIG_From_int((int)(0x03)));
  SWIG_Python_SetConstant(d, "RIG_COMM_STATUS_WARNING",SWIG_From_int((int)(0x04)));
  SWIG_Python_SetConstant(d, "RIG_COMM_STATUS_ERROR",SWIG_From_int((int)(0x05)));
  SWIG_Python_SetConstant(d, "HAMLIB_SETTINGS_FILE",SWIG_FromCharPtr("hamlib_settings"));
  SWIG_Python_SetConstant(d, "GPIO1",SWIG_From_int((int)(GPIO1)));
  SWIG_Python_SetConstant(d, "GPIO2",SWIG_From_int((int)(GPIO2)));
  SWIG_Python_SetConstant(d, "GPIO3",SWIG_From_int((int)(GPIO3)));
  SWIG_Python_SetConstant(d, "GPIO4",SWIG_From_int((int)(GPIO4)));
  SWIG_Python_SetConstant(d, "NETROTCTL_RET",SWIG_FromCharPtr("RPRT "));
  SWIG_Python_SetConstant(d, "ROT_RESET_ALL",SWIG_From_int((int)(1)));
  SWIG_Python_SetConstant(d, "ROT_FLAG_AZIMUTH",SWIG_From_int((int)(ROT_FLAG_AZIMUTH)));
  SWIG_Python_SetConstant(d, "ROT_FLAG_ELEVATION",SWIG_From_int((int)(ROT_FLAG_ELEVATION)));
  SWIG_Python_SetConstant(d, "ROT_TYPE_OTHER",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "ROT_MOVE_UP",SWIG_From_int((int)((1 << 1))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_DOWN",SWIG_From_int((int)((1 << 2))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_LEFT",SWIG_From_int((int)((1 << 3))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_CCW",SWIG_From_int((int)((1 << 3))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_RIGHT",SWIG_From_int((int)((1 << 4))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_CW",SWIG_From_int((int)((1 << 4))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_UP_LEFT",SWIG_From_int((int)((1 << 5))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_UP_CCW",SWIG_From_int((int)((1 << 5))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_UP_RIGHT",SWIG_From_int((int)((1 << 6))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_UP_CW",SWIG_From_int((int)((1 << 6))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_DOWN_LEFT",SWIG_From_int((int)((1 << 7))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_DOWN_CCW",SWIG_From_int((int)((1 << 7))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_DOWN_RIGHT",SWIG_From_int((int)((1 << 8))));
  SWIG_Python_SetConstant(d, "ROT_MOVE_DOWN_CW",SWIG_From_int((int)((1 << 8))));
  SWIG_Python_SetConstant(d, "ROT_STATUS_NONE",SWIG_From_int((int)(ROT_STATUS_NONE)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_BUSY",SWIG_From_int((int)(ROT_STATUS_BUSY)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_MOVING",SWIG_From_int((int)(ROT_STATUS_MOVING)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_MOVING_AZ",SWIG_From_int((int)(ROT_STATUS_MOVING_AZ)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_MOVING_LEFT",SWIG_From_int((int)(ROT_STATUS_MOVING_LEFT)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_MOVING_RIGHT",SWIG_From_int((int)(ROT_STATUS_MOVING_RIGHT)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_MOVING_EL",SWIG_From_int((int)(ROT_STATUS_MOVING_EL)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_MOVING_UP",SWIG_From_int((int)(ROT_STATUS_MOVING_UP)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_MOVING_DOWN",SWIG_From_int((int)(ROT_STATUS_MOVING_DOWN)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_LIMIT_UP",SWIG_From_int((int)(ROT_STATUS_LIMIT_UP)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_LIMIT_DOWN",SWIG_From_int((int)(ROT_STATUS_LIMIT_DOWN)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_LIMIT_LEFT",SWIG_From_int((int)(ROT_STATUS_LIMIT_LEFT)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_LIMIT_RIGHT",SWIG_From_int((int)(ROT_STATUS_LIMIT_RIGHT)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_OVERLAP_UP",SWIG_From_int((int)(ROT_STATUS_OVERLAP_UP)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_OVERLAP_DOWN",SWIG_From_int((int)(ROT_STATUS_OVERLAP_DOWN)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_OVERLAP_LEFT",SWIG_From_int((int)(ROT_STATUS_OVERLAP_LEFT)));
  SWIG_Python_SetConstant(d, "ROT_STATUS_OVERLAP_RIGHT",SWIG_From_int((int)(ROT_STATUS_OVERLAP_RIGHT)));
  SWIG_Python_SetConstant(d, "ROT_SPEED_NOCHANGE",SWIG_From_int((int)((-1))));
  SWIG_Python_SetConstant(d, "ROT_LEVEL_NONE",SWIG_From_int((int)(ROT_LEVEL_NONE)));
  SWIG_Python_SetConstant(d, "ROT_LEVEL_SPEED",SWIG_From_int((int)(ROT_LEVEL_SPEED)));
  SWIG_Python_SetConstant(d, "ROT_LEVEL_63",SWIG_From_int((int)(ROT_LEVEL_63)));
  SWIG_Python_SetConstant(d, "ROT_LEVEL_FLOAT_LIST",SWIG_From_int((int)((0))));
  SWIG_Python_SetConstant(d, "ROT_LEVEL_READONLY_LIST",SWIG_From_int((int)((0))));
  SWIG_Python_SetConstant(d, "ROT_PARM_NONE",SWIG_From_int((int)(ROT_PARM_NONE)));
  SWIG_Python_SetConstant(d, "ROT_PARM_FLOAT_LIST",SWIG_From_int((int)((0))));
  SWIG_Python_SetConstant(d, "ROT_PARM_READONLY_LIST",SWIG_From_int((int)((0))));
  SWIG_Python_SetConstant(d, "ROT_FUNC_NONE",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "ROT_FUNC_BIT63",SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (63)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_NONE",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "ROT_MODEL_DUMMY",SWIG_From_int((int)((100*(0)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_NETROTCTL",SWIG_From_int((int)((100*(0)+(2)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_PSTROTATOR",SWIG_From_int((int)((100*(0)+(3)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_SATROTCTL",SWIG_From_int((int)((100*(0)+(4)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_EASYCOMM1",SWIG_From_int((int)((100*(2)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_EASYCOMM2",SWIG_From_int((int)((100*(2)+(2)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_EASYCOMM3",SWIG_From_int((int)((100*(2)+(4)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_FODTRACK",SWIG_From_int((int)((100*(3)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_ROTOREZ",SWIG_From_int((int)((100*(4)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_ROTORCARD",SWIG_From_int((int)((100*(4)+(2)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_DCU",SWIG_From_int((int)((100*(4)+(3)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_ERC",SWIG_From_int((int)((100*(4)+(4)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_RT21",SWIG_From_int((int)((100*(4)+(5)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_YRC1",SWIG_From_int((int)((100*(4)+(6)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_SARTEK1",SWIG_From_int((int)((100*(5)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_GS232A",SWIG_From_int((int)((100*(6)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_GS232_GENERIC",SWIG_From_int((int)((100*(6)+(2)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_GS232B",SWIG_From_int((int)((100*(6)+(3)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_F1TETRACKER",SWIG_From_int((int)((100*(6)+(4)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_GS23",SWIG_From_int((int)((100*(6)+(5)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_GS232",SWIG_From_int((int)((100*(6)+(6)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_LVB",SWIG_From_int((int)((100*(6)+(7)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_ST2",SWIG_From_int((int)((100*(6)+(8)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_GS232A_AZ",SWIG_From_int((int)((100*(6)+(9)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_GS232A_EL",SWIG_From_int((int)((100*(6)+(10)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_GS232B_AZ",SWIG_From_int((int)((100*(6)+(11)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_GS232B_EL",SWIG_From_int((int)((100*(6)+(12)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_GS23_AZ",SWIG_From_int((int)((100*(6)+(13)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_AF6SA_WRC",SWIG_From_int((int)((100*(6)+(14)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_ARRAYSOLUTIONS_SAL_12_20_30",SWIG_From_int((int)((100*(327)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_PCROTOR",SWIG_From_int((int)((100*(7)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_HD1780",SWIG_From_int((int)((100*(8)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_SPID_ROT2PROG",SWIG_From_int((int)((100*(9)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_SPID_ROT1PROG",SWIG_From_int((int)((100*(9)+(2)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_SPID_MD01_ROT2PROG",SWIG_From_int((int)((100*(9)+(3)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_RC2800",SWIG_From_int((int)((100*(10)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_RC2800_EARLY_AZ",SWIG_From_int((int)((100*(10)+(2)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_RC2800_EARLY_AZEL",SWIG_From_int((int)((100*(10)+(3)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_RCI_AZEL",SWIG_From_int((int)((100*(11)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_RCI_AZ",SWIG_From_int((int)((100*(11)+(2)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_IF100",SWIG_From_int((int)((100*(12)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_TS7400",SWIG_From_int((int)((100*(13)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_NEXSTAR",SWIG_From_int((int)((100*(14)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_ETHER6",SWIG_From_int((int)((100*(15)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_CNCTRK",SWIG_From_int((int)((100*(16)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_PROSISTEL_D_AZ",SWIG_From_int((int)((100*(17)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_PROSISTEL_D_EL",SWIG_From_int((int)((100*(17)+(2)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_PROSISTEL_COMBI_TRACK_AZEL",SWIG_From_int((int)((100*(17)+(3)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_PROSISTEL_D_EL_CBOXAZ",SWIG_From_int((int)((100*(17)+(4)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_MEADE",SWIG_From_int((int)((100*(18)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_IOPTRON",SWIG_From_int((int)((100*(19)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_INDI",SWIG_From_int((int)((100*(20)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_SATEL",SWIG_From_int((int)((100*(21)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_RADANT",SWIG_From_int((int)((100*(22)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_ANDROIDSENSOR",SWIG_From_int((int)((100*(23)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_GRBLTRK_SER",SWIG_From_int((int)((100*(24)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_GRBLTRK_NET",SWIG_From_int((int)((100*(24)+(2)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_FLIR",SWIG_From_int((int)((100*(25)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_APEX_SHARED_LOOP",SWIG_From_int((int)((100*(26)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_SAEBRTRACK",SWIG_From_int((int)((100*(27)+(1)))));
  SWIG_Python_SetConstant(d, "ROT_MODEL_SKYWATCHER",SWIG_From_int((int)((100*(28)+(1)))));
  SWIG_Python_SetConstant(d, "AMP_MODEL_NONE",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "AMP_MODEL_DUMMY",SWIG_From_int((int)(((0)*100+(1)))));
  SWIG_Python_SetConstant(d, "AMP_MODEL_NETAMPCTL",SWIG_From_int((int)(((0)*100+(2)))));
  SWIG_Python_SetConstant(d, "AMP_MODEL_ELECRAFT_KPA1500",SWIG_From_int((int)(((2)*100+(1)))));
  SWIG_Python_SetConstant(d, "AMP_MODEL_GEMINI_DX1200",SWIG_From_int((int)(((3)*100+(1)))));
  SWIG_Python_SetConstant(d, "AMP_MODEL_EXPERT_FA",SWIG_From_int((int)(((4)*100+(1)))));
  SWIG_Python_SetConstant(d, "NETAMPCTL_RET",SWIG_FromCharPtr("RPRT "));
  SWIG_Python_SetConstant(d, "AMP_RESET_MEM",SWIG_From_int((int)(AMP_RESET_MEM)));
  SWIG_Python_SetConstant(d, "AMP_RESET_FAULT",SWIG_From_int((int)(AMP_RESET_FAULT)));
  SWIG_Python_SetConstant(d, "AMP_RESET_AMP",SWIG_From_int((int)(AMP_RESET_AMP)));
  SWIG_Python_SetConstant(d, "AMP_FLAG_1",SWIG_From_int((int)(AMP_FLAG_1)));
  SWIG_Python_SetConstant(d, "AMP_FLAG_2",SWIG_From_int((int)(AMP_FLAG_2)));
  SWIG_Python_SetConstant(d, "AMP_TYPE_OTHER",SWIG_From_int((int)(0)));
  SWIG_Python_SetConstant(d, "AMP_LEVEL_NONE",SWIG_From_int((int)(AMP_LEVEL_NONE)));
  SWIG_Python_SetConstant(d, "AMP_LEVEL_SWR",SWIG_From_int((int)(AMP_LEVEL_SWR)));
  SWIG_Python_SetConstant(d, "AMP_LEVEL_NH",SWIG_From_int((int)(AMP_LEVEL_NH)));
  SWIG_Python_SetConstant(d, "AMP_LEVEL_PF",SWIG_From_int((int)(AMP_LEVEL_PF)));
  SWIG_Python_SetConstant(d, "AMP_LEVEL_PWR_INPUT",SWIG_From_int((int)(AMP_LEVEL_PWR_INPUT)));
  SWIG_Python_SetConstant(d, "AMP_LEVEL_PWR_FWD",SWIG_From_int((int)(AMP_LEVEL_PWR_FWD)));
  SWIG_Python_SetConstant(d, "AMP_LEVEL_PWR_REFLECTED",SWIG_From_int((int)(AMP_LEVEL_PWR_REFLECTED)));
  SWIG_Python_SetConstant(d, "AMP_LEVEL_PWR_PEAK",SWIG_From_int((int)(AMP_LEVEL_PWR_PEAK)));
  SWIG_Python_SetConstant(d, "AMP_LEVEL_FAULT",SWIG_From_int((int)(AMP_LEVEL_FAULT)));
  SWIG_Python_SetConstant(d, "AMP_LEVEL_PWR",SWIG_From_int((int)(AMP_LEVEL_PWR)));
  SWIG_addvarlink(globals, "cookie_use", Swig_var_cookie_use_get, Swig_var_cookie_use_set);
  SWIG_addvarlink(globals, "skip_init", Swig_var_skip_init_get, Swig_var_skip_init_set);
  SWIG_addvarlink(globals, "hamlib_version", Swig_var_hamlib_version_get, Swig_var_hamlib_version_set);
  SWIG_addvarlink(globals, "hamlib_copyright", Swig_var_hamlib_copyright_get, Swig_var_hamlib_copyright_set);
  return 0;
}

