/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (https://www.swig.org).
 * Version 4.4.1
 *
 * Do not make changes to this file unless you know what you are doing - modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


#define SWIG_VERSION 0x040401
#define SWIGTCL
/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

#if defined(__cplusplus) && __cplusplus >=201103L
# define SWIG_NOEXCEPT noexcept
#else
# define SWIG_NOEXCEPT throw()
#endif 

/* -----------------------------------------------------------------------------
 * swigcompat.swg
 *
 * Macros to provide support compatibility with older C and C++ standards.
 *
 * Note that SWIG expects __cplusplus to be defined to the appropriate C++ standard.
 * MSVC users are urged to check and examine the /Zc:__cplusplus compiler option.
 * See https://learn.microsoft.com/en-us/cpp/build/reference/zc-cplusplus.
 * ----------------------------------------------------------------------------- */

/* C99 and C++11 should provide snprintf, but define SWIG_NO_SNPRINTF
 * if you're missing it.
 */
#if ((defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L) || \
     (defined __cplusplus && __cplusplus >= 201103L) || \
     defined SWIG_HAVE_SNPRINTF) && \
    !defined SWIG_NO_SNPRINTF
# define SWIG_snprintf(O,S,F,A) snprintf(O,S,F,A)
# define SWIG_snprintf2(O,S,F,A,B) snprintf(O,S,F,A,B)
#else
/* Fallback versions ignore the buffer size, but most of our uses either have a
 * fixed maximum possible size or dynamically allocate a buffer that's large
 * enough.
 */
# define SWIG_snprintf(O,S,F,A) sprintf(O,F,A)
# define SWIG_snprintf2(O,S,F,A,B) sprintf(O,F,A,B)
#endif

/* tcl.h has to appear first */
#include <stdio.h>
#include <tcl.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

/* Check, if Tcl version supports Tcl_Size,
   which was introduced in Tcl 8.7 and 9.
*/
#ifndef TCL_SIZE_MAX
    #include <limits.h>
    #define TCL_SIZE_MAX INT_MAX

    #ifndef Tcl_Size
        typedef int Tcl_Size;
    #endif

    #define TCL_SIZE_MODIFIER ""
    #define Tcl_GetSizeIntFromObj Tcl_GetIntFromObj
    #define Tcl_NewSizeIntObj     Tcl_NewIntObj
#endif

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "5"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4
#define SWIG_POINTER_CLEAR         0x8
#define SWIG_POINTER_RELEASE       (SWIG_POINTER_CLEAR | SWIG_POINTER_DISOWN)

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows returning the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
/* Runtime errors are < 0 */
#define SWIG_ERROR                 (-1)
/* Errors in range -1 to -99 are in swigerrors.swg (errors for all languages including those not using the runtime) */
/* Errors in range -100 to -199 are language specific errors defined in *errors.swg */
/* Errors < -200 are generic runtime specific errors */
#define SWIG_ERROR_RELEASE_NOT_OWNED (-200)

#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporary objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del object mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next array of casts | pointer to cast hashed by value */
  unsigned int            value;		/* index of the last valid element in the array | typename hash value */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
 * Hash function for type name strings, based on maRushPrime1Hash (http://amsoftware.narod.ru/algo2.html)
 */
SWIGRUNTIME unsigned int SWIG_Hash(const char *str, unsigned int len) {
  const unsigned char *data = (const unsigned char *)str;
  unsigned int hash = len, i = 0, k;
  int rem = (int)len;

  while (rem >= (int)sizeof(unsigned int)) {
    memcpy(&k, data, sizeof(unsigned int));
    k += i++;
    hash ^= k;
    hash *= 171717;
    data += sizeof(unsigned int);
    rem -= sizeof(unsigned int);
  }

  switch (rem) {
    case 3: k = (unsigned int)(data[2]) << 16;
            k |= (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 2: k = (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 1: k = (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
  }
  return hash;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  static const unsigned int scan_threshold = 4;
  if (ty) {
    swig_cast_info *head = ty->cast;
    unsigned int hash_value = 0;
    int hashed = 0;

    while (head) {

      if (strcmp(head->type->name, c) == 0) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;
        int search = 1;

        if (!hashed) {
          if (head->value < scan_threshold) {
            for (iter = first; iter <= last; iter++) {
              if (strcmp(iter->type->name, c) == 0) {
                return iter;
              }
            }
            search = 0;
          } else {
            hashed = 1;
            hash_value = SWIG_Hash(c, (unsigned int)strlen(c));
          }
        }

        if (search) {
          /* Binary search over sorted <'next'|'value'> pairs */
          do {
            iter = first + ((last - first) >> 1);
            if (iter->value < hash_value) {
              first = iter + 1;
            } else if (iter->value == hash_value) {

              if (strcmp(iter->next->type->name, c) == 0) {
                return iter->next;
              }

              /* Hash collision check */
              for (last = iter + 1; last->next && last->value == hash_value; last++) {
                if (strcmp(last->next->type->name, c) == 0) {
                  return last->next;
                }
              }
              for (first = iter - 1; first != head && first->value == hash_value; first--) {
                if (strcmp(first->next->type->name, c) == 0) {
                  return first->next;
                }
              }
              break;
            } else
              last = iter - 1;
          } while (first <= last);
        }
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Check the type by type address
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(const swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *head = ty->cast;
    while (head) {
      if (head->type == from) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;

        /* Binary search over sorted array of casts */
        do {
          iter = first + ((last - first) >> 1);
          if (iter->type < from) {
            first = iter + 1;
          } else if (iter->type == from) {
            return iter;
          } else
            last = iter - 1;
        } while (first <= last);
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  Choose the last
     name. It should be the most specific; a fully resolved name
     but not necessarily with default template parameters expanded. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *head = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (head) {
    swig_cast_info *cast;
    for (cast = head; (cast - head) <= head->value; cast++) {
      if (!cast->converter) {
        swig_type_info *tc = cast->type;
        if (!tc->clientdata) {
          SWIG_TypeClientData(tc, clientdata);
        }
      }
    }
    head = head->next;
  }
}

SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/* SWIG Errors applicable to all language modules, values are reserved from -1 to -99 */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13


/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGINTERN const char*
SWIG_Tcl_ErrorType(int code) {
  const char* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = "MemoryError";
    break;
  case SWIG_IOError:
    type = "IOError";
    break;
  case SWIG_RuntimeError:
    type = "RuntimeError";
    break;
  case SWIG_IndexError:
    type = "IndexError";
    break;
  case SWIG_TypeError:
    type = "TypeError";
    break;
  case SWIG_DivisionByZero:
    type = "ZeroDivisionError";
    break;
  case SWIG_OverflowError:
    type = "OverflowError";
    break;
  case SWIG_SyntaxError:
    type = "SyntaxError";
    break;
  case SWIG_ValueError:
    type = "ValueError";
    break;
  case SWIG_SystemError:
    type = "SystemError";
    break;
  case SWIG_AttributeError:
    type = "AttributeError";
    break;
  case SWIG_NullReferenceError:
    type = "NullReferenceError";
    break;
  default:
    type = "RuntimeError";
  }
  return type;
}


SWIGINTERN void
SWIG_Tcl_SetErrorObj(Tcl_Interp *interp, const char *ctype, Tcl_Obj *obj)
{
  Tcl_ResetResult(interp);
  Tcl_SetObjResult(interp, obj);
  Tcl_SetErrorCode(interp, "SWIG", ctype, (char *)NULL);
}

SWIGINTERN void
SWIG_Tcl_SetErrorMsg(Tcl_Interp *interp, const char *ctype, const char *mesg)
{
  Tcl_ResetResult(interp);
  Tcl_SetErrorCode(interp, "SWIG", ctype, (char *)NULL);
  Tcl_AppendResult(interp, ctype, " ", mesg, (char *)NULL);
  /*
  Tcl_AddErrorInfo(interp, ctype);
  Tcl_AddErrorInfo(interp, " ");
  Tcl_AddErrorInfo(interp, mesg);
  */
}

SWIGINTERNINLINE void
SWIG_Tcl_AddErrorMsg(Tcl_Interp *interp, const char* mesg)
{
  Tcl_AddErrorInfo(interp, mesg);
}



/* -----------------------------------------------------------------------------
 * SWIG API. Portion that goes into the runtime
 * ----------------------------------------------------------------------------- */
#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_TCL_POINTER 4
#define SWIG_TCL_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
    int type;
    const char *name;
    long lvalue;
    double dvalue;
    void   *pvalue;
    swig_type_info **ptype;
} swig_const_info;

typedef int   (*swig_wrapper)(ClientData, Tcl_Interp *, int, Tcl_Obj *const []);
typedef int   (*swig_wrapper_func)(ClientData, Tcl_Interp *, int, Tcl_Obj *const []);
typedef char *(*swig_variable_func)(ClientData, Tcl_Interp *, char *, char *, int);
typedef void  (*swig_delete_func)(ClientData);

typedef struct swig_method {
  const char     *name;
  swig_wrapper   method;
} swig_method;

typedef struct swig_attribute {
  const char     *name;
  swig_wrapper   getmethod;
  swig_wrapper   setmethod;
} swig_attribute;

typedef struct swig_class {
  const char         *name;
  swig_type_info   **type;
  swig_wrapper       constructor;
  void              (*destructor)(void *);
  swig_method        *methods;
  swig_attribute     *attributes;
  struct swig_class **bases;
  const char              **base_names;
  swig_module_info   *module;
  Tcl_HashTable       hashtable;
} swig_class;

typedef struct swig_instance {
  Tcl_Obj       *thisptr;
  void          *thisvalue;
  swig_class   *classptr;
  int            destroy;
  Tcl_Command    cmdtok;
} swig_instance;

/* Structure for command table */
typedef struct {
  const char *name;
  int       (*wrapper)(ClientData, Tcl_Interp *, int, Tcl_Obj *const []);
  ClientData  clientdata;
} swig_command_info;

/* Structure for variable linking table */
typedef struct {
  const char *name;
  void *addr;
  char * (*get)(ClientData, Tcl_Interp *, char *, char *, int);
  char * (*set)(ClientData, Tcl_Interp *, char *, char *, int);
} swig_var_info;


/* -----------------------------------------------------------------------------*
 *  Install a constant object 
 * -----------------------------------------------------------------------------*/

static Tcl_HashTable   swigconstTable;
static int             swigconstTableinit = 0;

SWIGINTERN void
SWIG_Tcl_SetConstantObj(Tcl_Interp *interp, const char* name, Tcl_Obj *obj) {
  int newobj;
  Tcl_Obj *varNamePtr = Tcl_NewStringObj(name, -1);
  Tcl_IncrRefCount(varNamePtr);
  Tcl_ObjSetVar2(interp,varNamePtr, NULL, obj, TCL_GLOBAL_ONLY);
  Tcl_DecrRefCount(varNamePtr);
  Tcl_SetHashValue(Tcl_CreateHashEntry(&swigconstTable, name, &newobj), (ClientData) obj);
}

SWIGINTERN Tcl_Obj *
SWIG_Tcl_GetConstantObj(const char *key) {
  Tcl_HashEntry *entryPtr;
  if (!swigconstTableinit) return 0;
  entryPtr = Tcl_FindHashEntry(&swigconstTable, key);
  if (entryPtr) {
    return (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
  }
  return 0;
}

#ifdef __cplusplus
}
#endif



/* -----------------------------------------------------------------------------
 * tclrun.swg
 *
 * This file contains the runtime support for Tcl modules and includes
 * code for managing global variables and pointer type checking.
 * ----------------------------------------------------------------------------- */

/* Common SWIG API */

/* for raw pointers */
#define SWIG_ConvertPtr(oc, ptr, ty, flags)             SWIG_Tcl_ConvertPtr(interp, oc, ptr, ty, flags)
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Tcl_NewPointerObj(ptr, type, flags)

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Tcl_ConvertPacked(interp, obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Tcl_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_Tcl_ConvertPtr(interp, obj, pptr, type, flags)
#define SWIG_NewInstanceObj(thisvalue, type, flags)     SWIG_Tcl_NewInstanceObj(interp, thisvalue, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Tcl_ConvertPtr(interp, obj, pptr, type, 0)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Tcl_NewPointerObj(ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Tcl_ConvertPacked(interp,obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Tcl_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Tcl_GetModule((Tcl_Interp *) (clientdata))	     
#define SWIG_SetModule(clientdata, pointer)          	SWIG_Tcl_SetModule((Tcl_Interp *) (clientdata), pointer)


/* Error manipulation */

#define SWIG_ErrorType(code)                            SWIG_Tcl_ErrorType(code)                                      
#define SWIG_Error(code, msg)            		SWIG_Tcl_SetErrorMsg(interp, SWIG_Tcl_ErrorType(code), msg)
#define SWIG_fail                        		goto fail						    


/* Tcl-specific SWIG API */

#define SWIG_Acquire(ptr)                               SWIG_Tcl_Acquire(ptr)                                     
#define SWIG_MethodCommand                           	SWIG_Tcl_MethodCommand				       
#define SWIG_Disown(ptr)                             	SWIG_Tcl_Disown(ptr)				       
#define SWIG_ConvertPtrFromString(c, ptr, ty, flags) 	SWIG_Tcl_ConvertPtrFromString(interp, c, ptr, ty, flags)  
#define SWIG_MakePtr(c, ptr, ty, flags)              	SWIG_Tcl_MakePtr(c, ptr, ty, flags)		       
#define SWIG_PointerTypeFromString(c)                	SWIG_Tcl_PointerTypeFromString(c)			       
#define SWIG_GetArgs                                 	SWIG_Tcl_GetArgs					       
#define SWIG_GetConstantObj(key)                     	SWIG_Tcl_GetConstantObj(key)			       
#define SWIG_ObjectConstructor                       	SWIG_Tcl_ObjectConstructor				       
#define SWIG_Thisown(ptr)                            	SWIG_Tcl_Thisown(ptr)				       
#define SWIG_ObjectDelete                            	SWIG_Tcl_ObjectDelete				       


#define SWIG_TCL_DECL_ARGS_2(arg1, arg2)                (Tcl_Interp *interp SWIGUNUSED, arg1, arg2)
#define SWIG_TCL_CALL_ARGS_2(arg1, arg2)                (interp, arg1, arg2)
/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

/* For backward compatibility only */
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_GetConstant        SWIG_GetConstantObj
#define SWIG_Tcl_GetConstant    SWIG_Tcl_GetConstantObj

#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >= 5)
#define SWIG_TCL_HASHTABLE_INIT {0, {0, 0, 0, 0}, 0, 0, 0, 0, 0, 0, 0, 0, 0}
#else
#define SWIG_TCL_HASHTABLE_INIT {0}
#endif

#include "assert.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Object support */

SWIGRUNTIME Tcl_HashTable*
SWIG_Tcl_ObjectTable(void) {
  static Tcl_HashTable  swigobjectTable;
  static int            swigobjectTableinit = 0;
  if (!swigobjectTableinit) {
    Tcl_InitHashTable(&swigobjectTable, TCL_ONE_WORD_KEYS);
    swigobjectTableinit = 1;
  }
  return &swigobjectTable;
}

/* Acquire ownership of a pointer */
SWIGRUNTIME void
SWIG_Tcl_Acquire(void *ptr) {
  int newobj;
  Tcl_CreateHashEntry(SWIG_Tcl_ObjectTable(), (char *) ptr, &newobj);
}

SWIGRUNTIME int
SWIG_Tcl_Thisown(void *ptr) {
  if (Tcl_FindHashEntry(SWIG_Tcl_ObjectTable(), (char *) ptr)) {
    return 1;
  }
  return 0;
}

/* Disown a pointer.  Returns 1 if we owned it to begin with */
SWIGRUNTIME int
SWIG_Tcl_Disown(void *ptr) {
  Tcl_HashEntry *entryPtr = Tcl_FindHashEntry(SWIG_Tcl_ObjectTable(), (char *) ptr);
  if (entryPtr) {
    Tcl_DeleteHashEntry(entryPtr);
    return 1;
  }
  return 0;
}

SWIGRUNTIME void SWIG_Tcl_ObjectDelete(ClientData clientData);

/* Convert a pointer value */
SWIGRUNTIME int
SWIG_Tcl_ConvertPtrFromString(Tcl_Interp *interp, const char *c, void **ptr, swig_type_info *ty, int flags) {
  swig_cast_info *tc;
  const char *cmd_name;
  /* Pointer values must start with leading underscore */
  while (*c != '_') {
    *ptr = (void *) 0;
    if (strcmp(c,"NULL") == 0)
      return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;

    /* Empty string: not a pointer */
    if (*c == 0) return SWIG_ERROR; 

    /* Hmmm. It could be an object name. */

    /* Check if this is a command at all. Prevents <c> cget -this         */
    /* from being called when c is not a command, firing the unknown proc */
    if (Tcl_VarEval(interp,"info commands ", c, (char *) NULL) == TCL_OK) {
      Tcl_Obj *result = Tcl_GetObjResult(interp);
      if (*(Tcl_GetString(result)) == 0) {
        /* It's not a command, so it can't be a pointer */
        Tcl_ResetResult(interp);
        return SWIG_ERROR;
      }
    } else {
      /* This will only fail if the argument is multiple words. */
      /* Multiple words are also not commands.                  */
      Tcl_ResetResult(interp);
      return SWIG_ERROR;
    }

    /* Check if this is really a SWIG pointer */
    if (Tcl_VarEval(interp,c," cget -this", (char *) NULL) != TCL_OK) {
      Tcl_ResetResult(interp);
      return SWIG_ERROR;
    }

    c = Tcl_GetString(Tcl_GetObjResult(interp));
  }
  cmd_name = c;

  c++;
  c = SWIG_UnpackData(c,ptr,sizeof(void *));

  if (ty) {
    Tcl_CmdInfo info;
    tc = c ? SWIG_TypeCheck(c,ty) : 0;
    if (!tc) {
      return SWIG_ERROR;
    }
    if (Tcl_GetCommandInfo(interp, cmd_name, &info)) {
      /* When creating a pointer string, SWIG_Tcl_NewInstanceObj calls Tcl_CreateObjCommand and sets
       * info.objClientData to an instance of swig_instance. Detecting when we can cast any info.objClientData
       * to swig_instance is not simple as it may be an unrelated command; we use deleteProc to determine this. */
      if (info.deleteProc == SWIG_Tcl_ObjectDelete) {
        swig_instance *inst = (swig_instance *)info.objClientData;
        if (!inst->thisvalue) {
          *ptr = 0;
        }
        assert(inst->thisvalue == *ptr);
        if (((flags & SWIG_POINTER_RELEASE) == SWIG_POINTER_RELEASE) && !SWIG_Thisown(inst->thisvalue)) {
          return SWIG_ERROR_RELEASE_NOT_OWNED;
        } else {
          if (flags & SWIG_POINTER_DISOWN) {
            SWIG_Disown((void *) *ptr);
          }
          if (flags & SWIG_POINTER_CLEAR) {
            inst->thisvalue = 0;
          }
          {
            int newmemory = 0;
            *ptr = SWIG_TypeCast(tc,(void *) *ptr, &newmemory);
            assert(!newmemory); /* newmemory handling not yet implemented */
          }
        }
      }
    }
  }

  return SWIG_OK;
}

/* Convert a pointer value */
SWIGRUNTIMEINLINE int
SWIG_Tcl_ConvertPtr(Tcl_Interp *interp, Tcl_Obj *oc, void **ptr, swig_type_info *ty, int flags) {
  return SWIG_Tcl_ConvertPtrFromString(interp, Tcl_GetString(oc), ptr, ty, flags);
}

/* Convert a pointer value */
SWIGRUNTIME char *
SWIG_Tcl_PointerTypeFromString(char *c) {
  char d;
  /* Pointer values must start with leading underscore. NULL has no type */
  if (*c != '_') {
    return 0;
  }
  c++;
  /* Extract hex value from pointer */
  while ((d = *c)) {
    if (!(((d >= '0') && (d <= '9')) || ((d >= 'a') && (d <= 'f')))) break;
    c++;
  }
  return c;
}

/* Convert a packed pointer value */
SWIGRUNTIME int
SWIG_Tcl_ConvertPacked(Tcl_Interp *SWIGUNUSEDPARM(interp) , Tcl_Obj *obj, void *ptr, int sz, swig_type_info *ty) {
  swig_cast_info *tc;
  const char  *c;

  if (!obj) goto type_error;
  c = Tcl_GetString(obj);
  /* Pointer values must start with leading underscore */
  if (*c != '_') goto type_error;
  c++;
  c = SWIG_UnpackData(c,ptr,sz);
  if (ty) {
    tc = SWIG_TypeCheck(c,ty);
    if (!tc) goto type_error;
  }
  return SWIG_OK;

 type_error:

  return SWIG_ERROR;
}


/* Take a pointer and convert it to a string */
SWIGRUNTIME void
SWIG_Tcl_MakePtr(char *c, void *ptr, swig_type_info *ty, int SWIGUNUSEDPARM(flags)) {
  if (ptr) {
    *(c++) = '_';
    c = SWIG_PackData(c,&ptr,sizeof(void *));
    strcpy(c,ty->name);
  } else {
    strcpy(c,"NULL");
  }
}

/* Create a new pointer object */
SWIGRUNTIMEINLINE Tcl_Obj *
SWIG_Tcl_NewPointerObj(void *ptr, swig_type_info *type, int flags) {
  Tcl_Obj *robj;
  char result[SWIG_BUFFER_SIZE];
  SWIG_MakePtr(result,ptr,type,flags);
  robj = Tcl_NewStringObj(result,-1);
  return robj;
}

SWIGRUNTIME Tcl_Obj *
SWIG_Tcl_NewPackedObj(void *ptr, int sz, swig_type_info *type) {
  char result[1024];
  char *r = result;
  if ((2*sz + 1 + strlen(type->name)) > 1000) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  strcpy(r,type->name);
  return Tcl_NewStringObj(result,-1);
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

SWIGRUNTIME swig_module_info * 
SWIG_Tcl_GetModule(Tcl_Interp *interp) {
  const char *data;
  swig_module_info *ret = 0;
  
  /* first check if pointer already created */
  data = Tcl_GetVar(interp, (char *)"swig_runtime_data_type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME, TCL_GLOBAL_ONLY);
  if (data) {
    SWIG_UnpackData(data, &ret, sizeof(swig_type_info **));
  }

  return ret;
}

SWIGRUNTIME void
SWIG_Tcl_SetModule(Tcl_Interp *interp, swig_module_info *module) {
  char buf[SWIG_BUFFER_SIZE];
  char *data;

  /* create a new pointer */
  data = SWIG_PackData(buf, &module, sizeof(swig_type_info **));
  *data = 0;
  Tcl_SetVar(interp, (char *)"swig_runtime_data_type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME, buf, TCL_GLOBAL_ONLY);
}

/* -----------------------------------------------------------------------------*
 *  Object auxiliaries
 * -----------------------------------------------------------------------------*/


SWIGRUNTIME void
SWIG_Tcl_ObjectDelete(ClientData clientData) {
  swig_instance *si = (swig_instance *) clientData;
  if (!si) return;
  if (si->destroy && SWIG_Disown(si->thisvalue)) {
    if (si->classptr->destructor) {
      (si->classptr->destructor)(si->thisvalue);
    }
  }
  Tcl_DecrRefCount(si->thisptr);
  free(si);
}

/* Function to invoke object methods given an instance */
SWIGRUNTIME int
SWIG_Tcl_MethodCommand(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *const _objv[]) {
  char *method,   *attrname;
  swig_instance   *inst = (swig_instance *) clientData;
  swig_method     *meth;
  swig_attribute  *attr;
  Tcl_Obj         *oldarg;
  Tcl_Obj         **objv;
  int              rcode;
  swig_class      *cls;
  swig_class      *cls_stack[64];
  int              cls_stack_bi[64];
  int              cls_stack_top = 0;
  int              numconf = 2;
  int              bi;

  objv = (Tcl_Obj **) _objv;
  if (objc < 2) {
    Tcl_SetResult(interp, (char *) "wrong # args.", TCL_STATIC);
    return TCL_ERROR;
  }
  method = Tcl_GetString(objv[1]);
  if (strcmp(method,"-acquire") == 0) {
    inst->destroy = 1;
    SWIG_Acquire(inst->thisvalue);
    return TCL_OK;
  }
  if (strcmp(method,"-disown") == 0) {
    if (inst->destroy) {
      SWIG_Disown(inst->thisvalue);
    }
    inst->destroy = 0;
    return TCL_OK;
  }
  if (strcmp(method,"-delete") == 0) {
    Tcl_DeleteCommandFromToken(interp,inst->cmdtok);
    return TCL_OK;
  }
  cls_stack[cls_stack_top] = inst->classptr;
  cls_stack_bi[cls_stack_top] = -1;
  while (1) {
    Tcl_HashEntry* hashentry;
    bi = cls_stack_bi[cls_stack_top];
    cls = cls_stack[cls_stack_top];
    if (bi != -1) {
      if (!cls->bases[bi] && cls->base_names[bi]) {
        /* lookup and cache the base class */
	swig_type_info *info = SWIG_TypeQueryModule(cls->module, cls->module, cls->base_names[bi]);
	if (info) cls->bases[bi] = (swig_class *) info->clientdata;
      }
      cls = cls->bases[bi];
      if (cls) {
        cls_stack_bi[cls_stack_top]++;
        cls_stack_top++;
        cls_stack[cls_stack_top] = cls;
        cls_stack_bi[cls_stack_top] = -1;
        continue;
      }
    }
    if (!cls) {
      cls_stack_top--;
      if (cls_stack_top < 0) break;
      else continue;
    }
    cls_stack_bi[cls_stack_top]++;

    hashentry = Tcl_FindHashEntry(&(cls->hashtable), method);
    if (hashentry) {
        ClientData cd = Tcl_GetHashValue(hashentry);
        swig_wrapper method_wrapper = (swig_wrapper)cd;
        oldarg = objv[1];
        objv[1] = inst->thisptr;
        Tcl_IncrRefCount(inst->thisptr);
        rcode = (method_wrapper)(clientData,interp,objc,objv);
        objv[1] = oldarg;
        Tcl_DecrRefCount(inst->thisptr);
        return rcode;
    }
    /* Check class methods for a match */
    if (strcmp(method,"cget") == 0) {
      if (objc < 3) {
        Tcl_SetResult(interp, (char *) "wrong # args.", TCL_STATIC);
        return TCL_ERROR;
      }
      attrname = Tcl_GetString(objv[2]);
      attr = cls->attributes;
      while (attr && attr->name) {
        if ((strcmp(attr->name, attrname) == 0) && (attr->getmethod)) {
          oldarg = objv[1];
          objv[1] = inst->thisptr;
          Tcl_IncrRefCount(inst->thisptr);
          rcode = (*attr->getmethod)(clientData,interp,2, objv);
          objv[1] = oldarg;
          Tcl_DecrRefCount(inst->thisptr);
          return rcode;
        }
        attr++;
      }
      if (strcmp(attrname, "-this") == 0) {
        Tcl_SetObjResult(interp, Tcl_DuplicateObj(inst->thisptr));
        return TCL_OK;
      }
      if (strcmp(attrname, "-thisown") == 0) {
        if (SWIG_Thisown(inst->thisvalue)) {
          Tcl_SetResult(interp,(char*)"1",TCL_STATIC);
        } else {
          Tcl_SetResult(interp,(char*)"0",TCL_STATIC);
        }
        return TCL_OK;
      }
    } else if (strcmp(method, "configure") == 0) {
      int i;
      if (objc < 4) {
        Tcl_SetResult(interp, (char *) "wrong # args.", TCL_STATIC);
        return TCL_ERROR;
      }
      i = 2;
      while (i < objc) {
        attrname = Tcl_GetString(objv[i]);
        attr = cls->attributes;
        while (attr && attr->name) {
          if ((strcmp(attr->name, attrname) == 0) && (attr->setmethod)) {
            oldarg = objv[i];
            objv[i] = inst->thisptr;
            Tcl_IncrRefCount(inst->thisptr);
            rcode = (*attr->setmethod)(clientData,interp,3, &objv[i-1]);
            objv[i] = oldarg;
            Tcl_DecrRefCount(inst->thisptr);
            if (rcode != TCL_OK) return rcode;
            numconf += 2;
          }
          attr++;
        }
        i+=2;
      }
    }
  }
  if (strcmp(method,"configure") == 0) {
    if (numconf >= objc) {
      return TCL_OK;
    } else {
      Tcl_SetResult(interp,(char *) "Invalid attribute name.", TCL_STATIC);
      return TCL_ERROR;
    }
  }
  if (strcmp(method,"cget") == 0) {
    Tcl_SetResult(interp,(char *) "Invalid attribute name.", TCL_STATIC);
    return TCL_ERROR;
  }
  Tcl_SetResult(interp, (char *) "Invalid method. Must be one of: configure cget -acquire -disown -delete", TCL_STATIC);
  cls = inst->classptr;
  bi = 0;
  while (cls) {
    meth = cls->methods;
    while (meth && meth->name) {
      char *cr = (char *) Tcl_GetStringResult(interp);
      size_t meth_len = strlen(meth->name);
      char* where = strchr(cr,':');
      while(where) {
        where = strstr(where, meth->name);
        if(where) {
          if(where[-1] == ' ' && (where[meth_len] == ' ' || where[meth_len]==0)) {
            break;
          } else {
            where++;
          }
        }
      }

      if (!where)
        Tcl_AppendElement(interp, (char *) meth->name);
      meth++;
    }
    cls = inst->classptr->bases[bi++];
  }
  return TCL_ERROR;
}

/* This function takes the current result and turns it into an object command */
SWIGRUNTIME Tcl_Obj *
SWIG_Tcl_NewInstanceObj(Tcl_Interp *interp, void *thisvalue, swig_type_info *type, int flags) {
  Tcl_Obj *robj = SWIG_NewPointerObj(thisvalue, type,0);
  /* Check to see if this pointer belongs to a class or not */
  if (thisvalue && (type->clientdata) && (interp)) {
    Tcl_CmdInfo    ci;
    int has_command;
    char          *name;
    name = Tcl_GetString(robj);
    has_command = Tcl_GetCommandInfo(interp, name, &ci);
    if (!has_command || flags) {
      swig_instance *newinst = (swig_instance *) malloc(sizeof(swig_instance));
      newinst->thisptr = Tcl_DuplicateObj(robj);
      Tcl_IncrRefCount(newinst->thisptr);
      newinst->thisvalue = thisvalue;
      newinst->classptr = (swig_class *) type->clientdata;
      newinst->destroy = flags;
      newinst->cmdtok = Tcl_CreateObjCommand(interp, Tcl_GetString(robj), (swig_wrapper_func) SWIG_Tcl_MethodCommand, (ClientData) newinst, (swig_delete_func) SWIG_Tcl_ObjectDelete);
      if (flags) {
        SWIG_Acquire(thisvalue);
      }
    } else {
      swig_instance *inst = (swig_instance *)ci.objClientData;
      /* Restore thisvalue as SWIG_POINTER_CLEAR may have been used to set it to zero.
         Occurs when the C pointer is re-used by the memory allocator and the command has
         been created and not destroyed - bug?? - see cpp11_std_unique_ptr_runme.tcl test. */
      if (inst->thisvalue != thisvalue) {
        assert(inst->thisvalue == 0);
        inst->thisvalue = thisvalue;
      }
    }
  }
  return robj;
}

/* Function to create objects */
SWIGRUNTIME int
SWIG_Tcl_ObjectConstructor(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj          *newObj = 0;
  void             *thisvalue = 0;
  swig_instance   *newinst = 0;
  swig_class      *classptr = (swig_class *) clientData;
  swig_wrapper     cons = 0;
  char             *name = 0;
  int               firstarg = 0;
  int               thisarg = 0;
  int               destroy = 1;

  if (!classptr) {
    Tcl_SetResult(interp, (char *) "swig: internal runtime error. No class object defined.", TCL_STATIC);
    return TCL_ERROR;
  }
  cons = classptr->constructor;
  if (objc > 1) {
    char *s = Tcl_GetString(objv[1]);
    if (strcmp(s,"-this") == 0) {
      thisarg = 2;
      cons = 0;
    } else if (strcmp(s,"-args") == 0) {
      firstarg = 1;
    } else if (objc == 2) {
      firstarg = 1;
      name = s;
    } else if (objc >= 3) {
      char *s1;
      name = s;
      s1 = Tcl_GetString(objv[2]);
      if (strcmp(s1,"-this") == 0) {
	thisarg = 3;
	cons = 0;
      } else {
	firstarg = 1;
      }
    }
  }
  if (cons) {
    int result;
    result = (*cons)(0, interp, objc-firstarg, &objv[firstarg]);
    if (result != TCL_OK) {
      return result;
    }
    newObj = Tcl_DuplicateObj(Tcl_GetObjResult(interp));
    if (!name) name = Tcl_GetString(newObj);
  } else if (thisarg > 0) {
    if (thisarg < objc) {
      destroy = 0;
      newObj = Tcl_DuplicateObj(objv[thisarg]);
      if (!name) name = Tcl_GetString(newObj);
    } else {
      Tcl_SetResult(interp, (char *) "wrong # args.", TCL_STATIC);
      return TCL_ERROR;
    }
  } else {
    Tcl_SetResult(interp, (char *) "No constructor available.", TCL_STATIC);
    return TCL_ERROR;
  }
  if (SWIG_Tcl_ConvertPtr(interp,newObj, (void **) &thisvalue, *(classptr->type), 0) != SWIG_OK) {
    Tcl_DecrRefCount(newObj);
    return TCL_ERROR;
  }
  newinst = (swig_instance *) malloc(sizeof(swig_instance));
  newinst->thisptr = newObj;
  Tcl_IncrRefCount(newObj);
  newinst->thisvalue = thisvalue;
  newinst->classptr = classptr;
  newinst->destroy = destroy;
  if (destroy) {
    SWIG_Acquire(thisvalue);
  }
  newinst->cmdtok = Tcl_CreateObjCommand(interp,name, (swig_wrapper) SWIG_Tcl_MethodCommand, (ClientData) newinst, (swig_delete_func) SWIG_Tcl_ObjectDelete);
  return TCL_OK;
}

/* -----------------------------------------------------------------------------*
 *   Get arguments 
 * -----------------------------------------------------------------------------*/
SWIGRUNTIME int
SWIG_Tcl_GetArgs(Tcl_Interp *interp, int objc, Tcl_Obj *const objv[], const char *fmt, ...) {
  int        argno = 0, opt = 0;
  long       tempi;
  double     tempd;
  const char *c;
  va_list    ap;
  void      *vptr;
  Tcl_Obj   *obj = 0;
  swig_type_info *ty;

  va_start(ap,fmt);
  for (c = fmt; (*c && (*c != ':') && (*c != ';')); c++,argno++) {
    if (*c == '|') {
      opt = 1;
      c++;
    }
    if (argno >= (objc-1)) {
      if (!opt) {
        Tcl_SetResult(interp, (char *) "Wrong number of arguments ", TCL_STATIC);
        goto argerror;
      } else {
        va_end(ap);
        return TCL_OK;
      }
    }

    vptr = va_arg(ap,void *);
    if (vptr) {
      if (isupper(*c)) {
        obj = SWIG_Tcl_GetConstantObj(Tcl_GetString(objv[argno+1]));
        if (!obj) obj = objv[argno+1];
      } else {
        obj = objv[argno+1];
      }
      switch(*c) {
      case 'i': case 'I':
      case 'l': case 'L':
      case 'h': case 'H':
      case 'b': case 'B':
        if (Tcl_GetLongFromObj(interp,obj,&tempi) != TCL_OK) goto argerror;
        if ((*c == 'i') || (*c == 'I')) *((int *)vptr) = (int)tempi;
        else if ((*c == 'l') || (*c == 'L')) *((long *)vptr) = (long)tempi;
        else if ((*c == 'h') || (*c == 'H')) *((short*)vptr) = (short)tempi;
        else if ((*c == 'b') || (*c == 'B')) *((unsigned char *)vptr) = (unsigned char)tempi;
        break;
      case 'f': case 'F':
      case 'd': case 'D':
        if (Tcl_GetDoubleFromObj(interp,obj,&tempd) != TCL_OK) goto argerror;
        if ((*c == 'f') || (*c == 'F')) *((float *) vptr) = (float)tempd;
        else if ((*c == 'd') || (*c == 'D')) *((double*) vptr) = tempd;
        break;
      case 's': case 'S':
        if (*(c+1) == '#') {
          Tcl_Size *vlptr = (Tcl_Size *) va_arg(ap, void *);
          *((char **) vptr) = Tcl_GetStringFromObj(obj, vlptr);
          c++;
        } else {
          *((char **)vptr) = Tcl_GetString(obj);
        }
        break;
      case 'c': case 'C':
        *((char *)vptr) = *(Tcl_GetString(obj));
        break;
      case 'p': case 'P':
        ty = (swig_type_info *) va_arg(ap, void *);
        if (SWIG_Tcl_ConvertPtr(interp, obj, (void **) vptr, ty, 0) != SWIG_OK) goto argerror;
        break;
      case 'o': case 'O':
        *((Tcl_Obj **)vptr) = objv[argno+1];
        break;
      default:
        break;
      }
    }
  }

  if ((*c != ';') && ((objc-1) > argno)) {
    Tcl_SetResult(interp, (char *) "Wrong # args.", TCL_STATIC);
    goto argerror;
  }
  va_end(ap);
  return TCL_OK;

 argerror:
  {
    char temp[32];
    SWIG_snprintf(temp, sizeof(temp), "%d", argno+1);
    c = strchr(fmt,':');
    if (!c) c = strchr(fmt,';');
    if (!c) c = (char *)"";
    Tcl_AppendResult(interp,c," argument ", temp, (char *)NULL);
    va_end(ap);
    return TCL_ERROR;
  }
}

#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) do { if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } } while (0) 



  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg); return TCL_ERROR;; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_Amp swig_types[0]
#define SWIGTYPE_p_FIFO_RIG_s swig_types[1]
#define SWIGTYPE_p_Rig swig_types[2]
#define SWIGTYPE_p_Rot swig_types[3]
#define SWIGTYPE_p_agc_level_e swig_types[4]
#define SWIGTYPE_p_amp swig_types[5]
#define SWIGTYPE_p_amp_caps swig_types[6]
#define SWIGTYPE_p_amp_reset_e swig_types[7]
#define SWIGTYPE_p_amp_state swig_types[8]
#define SWIGTYPE_p_amp_type_e swig_types[9]
#define SWIGTYPE_p_ann_t swig_types[10]
#define SWIGTYPE_p_cal_table swig_types[11]
#define SWIGTYPE_p_cal_table_float swig_types[12]
#define SWIGTYPE_p_cal_table_float_table swig_types[13]
#define SWIGTYPE_p_cal_table_table swig_types[14]
#define SWIGTYPE_p_chan_list swig_types[15]
#define SWIGTYPE_p_chan_type_t swig_types[16]
#define SWIGTYPE_p_channel swig_types[17]
#define SWIGTYPE_p_channelArray swig_types[18]
#define SWIGTYPE_p_channel_cap swig_types[19]
#define SWIGTYPE_p_char swig_types[20]
#define SWIGTYPE_p_client_e swig_types[21]
#define SWIGTYPE_p_confparams swig_types[22]
#define SWIGTYPE_p_confparams_u swig_types[23]
#define SWIGTYPE_p_confparams_u_c swig_types[24]
#define SWIGTYPE_p_confparams_u_n swig_types[25]
#define SWIGTYPE_p_dcd_e swig_types[26]
#define SWIGTYPE_p_dcd_type_e swig_types[27]
#define SWIGTYPE_p_deferred_config_header swig_types[28]
#define SWIGTYPE_p_deferred_config_item swig_types[29]
#define SWIGTYPE_p_double swig_types[30]
#define SWIGTYPE_p_ext_list swig_types[31]
#define SWIGTYPE_p_f_p_struct_amp__int swig_types[32]
#define SWIGTYPE_p_f_p_struct_amp__p_char swig_types[33]
#define SWIGTYPE_p_f_p_struct_amp_double__int swig_types[34]
#define SWIGTYPE_p_f_p_struct_amp_enum_amp_reset_e__int swig_types[35]
#define SWIGTYPE_p_f_p_struct_amp_enum_powerstat_t__int swig_types[36]
#define SWIGTYPE_p_f_p_struct_amp_long_p_char__int swig_types[37]
#define SWIGTYPE_p_f_p_struct_amp_long_p_char_int__int swig_types[38]
#define SWIGTYPE_p_f_p_struct_amp_long_p_q_const__char__int swig_types[39]
#define SWIGTYPE_p_f_p_struct_amp_long_p_value_t__int swig_types[40]
#define SWIGTYPE_p_f_p_struct_amp_long_value_t__int swig_types[41]
#define SWIGTYPE_p_f_p_struct_amp_p_double__int swig_types[42]
#define SWIGTYPE_p_f_p_struct_amp_p_enum_powerstat_t__int swig_types[43]
#define SWIGTYPE_p_f_p_struct_amp_unsigned_long_long_p_value_t__int swig_types[44]
#define SWIGTYPE_p_f_p_struct_amp_unsigned_long_long_value_t__int swig_types[45]
#define SWIGTYPE_p_f_p_struct_s_rig__int swig_types[46]
#define SWIGTYPE_p_f_p_struct_s_rig__p_char swig_types[47]
#define SWIGTYPE_p_f_p_struct_s_rig_enum_powerstat_t__int swig_types[48]
#define SWIGTYPE_p_f_p_struct_s_rig_enum_reset_t__int swig_types[49]
#define SWIGTYPE_p_f_p_struct_s_rig_int__int swig_types[50]
#define SWIGTYPE_p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int swig_types[51]
#define SWIGTYPE_p_f_p_struct_s_rig_long_p_char__int swig_types[52]
#define SWIGTYPE_p_f_p_struct_s_rig_long_p_char_int__int swig_types[53]
#define SWIGTYPE_p_f_p_struct_s_rig_long_p_q_const__char__int swig_types[54]
#define SWIGTYPE_p_f_p_struct_s_rig_long_p_value_t__int swig_types[55]
#define SWIGTYPE_p_f_p_struct_s_rig_long_value_t__int swig_types[56]
#define SWIGTYPE_p_f_p_struct_s_rig_p_enum_powerstat_t__int swig_types[57]
#define SWIGTYPE_p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int swig_types[58]
#define SWIGTYPE_p_f_p_struct_s_rig_p_int__int swig_types[59]
#define SWIGTYPE_p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int swig_types[60]
#define SWIGTYPE_p_f_p_struct_s_rig_p_q_const__char__int swig_types[61]
#define SWIGTYPE_p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int swig_types[62]
#define SWIGTYPE_p_f_p_struct_s_rig_p_unsigned_int__int swig_types[63]
#define SWIGTYPE_p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int swig_types[64]
#define SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int swig_types[65]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int swig_types[66]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double__int swig_types[67]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double_p_void__int swig_types[68]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int swig_types[69]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int swig_types[70]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int swig_types[71]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int swig_types[72]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int swig_types[73]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int swig_types[74]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int swig_types[75]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int swig_types[76]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int swig_types[77]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int swig_types[78]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_int__int swig_types[79]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_p_int__int swig_types[80]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int swig_types[81]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_value_t__int swig_types[82]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int swig_types[83]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double__int swig_types[84]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int swig_types[85]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int swig_types[86]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int swig_types[87]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int swig_types[88]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int swig_types[89]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int swig_types[90]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int swig_types[91]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int swig_types[92]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int swig_types[93]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_int__int swig_types[94]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int swig_types[95]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int swig_types[96]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int swig_types[97]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int swig_types[98]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int swig_types[99]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int swig_types[100]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_void__int swig_types[101]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int swig_types[102]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int swig_types[103]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int swig_types[104]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int swig_types[105]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int swig_types[106]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int swig_types[107]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int swig_types[108]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int swig_types[109]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int swig_types[110]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int swig_types[111]
#define SWIGTYPE_p_f_p_struct_s_rig_unsigned_long_long_value_t__int swig_types[112]
#define SWIGTYPE_p_f_p_struct_s_rot__int swig_types[113]
#define SWIGTYPE_p_f_p_struct_s_rot__p_char swig_types[114]
#define SWIGTYPE_p_f_p_struct_s_rot_float_float__int swig_types[115]
#define SWIGTYPE_p_f_p_struct_s_rot_int__int swig_types[116]
#define SWIGTYPE_p_f_p_struct_s_rot_int_int__int swig_types[117]
#define SWIGTYPE_p_f_p_struct_s_rot_long_int__int swig_types[118]
#define SWIGTYPE_p_f_p_struct_s_rot_long_p_char__int swig_types[119]
#define SWIGTYPE_p_f_p_struct_s_rot_long_p_char_int__int swig_types[120]
#define SWIGTYPE_p_f_p_struct_s_rot_long_p_int__int swig_types[121]
#define SWIGTYPE_p_f_p_struct_s_rot_long_p_q_const__char__int swig_types[122]
#define SWIGTYPE_p_f_p_struct_s_rot_long_p_value_t__int swig_types[123]
#define SWIGTYPE_p_f_p_struct_s_rot_long_value_t__int swig_types[124]
#define SWIGTYPE_p_f_p_struct_s_rot_p_enum_rot_status_t__int swig_types[125]
#define SWIGTYPE_p_f_p_struct_s_rot_p_float_p_float__int swig_types[126]
#define SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_int__int swig_types[127]
#define SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_int__int swig_types[128]
#define SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int swig_types[129]
#define SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_value_t__int swig_types[130]
#define SWIGTYPE_p_filter_list swig_types[131]
#define SWIGTYPE_p_float swig_types[132]
#define SWIGTYPE_p_freq_range_list swig_types[133]
#define SWIGTYPE_p_gran swig_types[134]
#define SWIGTYPE_p_hamlib_async_pipe swig_types[135]
#define SWIGTYPE_p_hamlib_band_t swig_types[136]
#define SWIGTYPE_p_hamlib_bandselect_t swig_types[137]
#define SWIGTYPE_p_hamlib_cache_t swig_types[138]
#define SWIGTYPE_p_hamlib_port swig_types[139]
#define SWIGTYPE_p_int swig_types[140]
#define SWIGTYPE_p_long swig_types[141]
#define SWIGTYPE_p_long_long swig_types[142]
#define SWIGTYPE_p_multicast_s swig_types[143]
#define SWIGTYPE_p_p_char swig_types[144]
#define SWIGTYPE_p_p_void swig_types[145]
#define SWIGTYPE_p_powerstat_t swig_types[146]
#define SWIGTYPE_p_pthread_mutex_t swig_types[147]
#define SWIGTYPE_p_pthread_t swig_types[148]
#define SWIGTYPE_p_ptt_t swig_types[149]
#define SWIGTYPE_p_ptt_type_e swig_types[150]
#define SWIGTYPE_p_reset_t swig_types[151]
#define SWIGTYPE_p_rig_callbacks swig_types[152]
#define SWIGTYPE_p_rig_caps swig_types[153]
#define SWIGTYPE_p_rig_debug_level_e swig_types[154]
#define SWIGTYPE_p_rig_port_e swig_types[155]
#define SWIGTYPE_p_rig_ptrx_t swig_types[156]
#define SWIGTYPE_p_rig_spectrum_avg_mode swig_types[157]
#define SWIGTYPE_p_rig_spectrum_line swig_types[158]
#define SWIGTYPE_p_rig_spectrum_mode_e swig_types[159]
#define SWIGTYPE_p_rig_spectrum_scope swig_types[160]
#define SWIGTYPE_p_rig_state swig_types[161]
#define SWIGTYPE_p_rig_type_t swig_types[162]
#define SWIGTYPE_p_rot_caps swig_types[163]
#define SWIGTYPE_p_rot_state swig_types[164]
#define SWIGTYPE_p_rot_status_t swig_types[165]
#define SWIGTYPE_p_rot_type_t swig_types[166]
#define SWIGTYPE_p_rptr_shift_t swig_types[167]
#define SWIGTYPE_p_s_rig swig_types[168]
#define SWIGTYPE_p_s_rot swig_types[169]
#define SWIGTYPE_p_scan_t swig_types[170]
#define SWIGTYPE_p_short swig_types[171]
#define SWIGTYPE_p_signed_char swig_types[172]
#define SWIGTYPE_p_sockaddr_in swig_types[173]
#define SWIGTYPE_p_split_t swig_types[174]
#define SWIGTYPE_p_time_t swig_types[175]
#define SWIGTYPE_p_timespec swig_types[176]
#define SWIGTYPE_p_toneArray swig_types[177]
#define SWIGTYPE_p_tuning_step_list swig_types[178]
#define SWIGTYPE_p_twiddle_state_t swig_types[179]
#define SWIGTYPE_p_unsigned_char swig_types[180]
#define SWIGTYPE_p_unsigned_int swig_types[181]
#define SWIGTYPE_p_unsigned_long swig_types[182]
#define SWIGTYPE_p_unsigned_long_long swig_types[183]
#define SWIGTYPE_p_unsigned_short swig_types[184]
#define SWIGTYPE_p_value_t swig_types[185]
#define SWIGTYPE_p_value_t_b swig_types[186]
#define SWIGTYPE_p_vfo_op_t swig_types[187]
#define SWIGTYPE_p_void swig_types[188]
static swig_type_info *swig_types[190];
static swig_module_info swig_module = {swig_types, 189, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#define SWIG_init    Hamlib_Init
#define SWIG_name    "hamlib"
#define SWIG_prefix  ""
#define SWIG_version "4.7"

#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 



#ifdef __cplusplus
extern "C" {
#endif
#ifdef MAC_TCL
#pragma export on
#endif
SWIGEXPORT int SWIG_init(Tcl_Interp *);
#ifdef MAC_TCL
#pragma export off
#endif
#ifdef __cplusplus
}
#endif

/* Compatibility version for TCL stubs */
#ifndef SWIG_TCL_STUBS_VERSION
#define SWIG_TCL_STUBS_VERSION "8.4-"
#endif



/*
 *  Hamlib bindings - swig interface file
 *  Copyright (c) 2001-2008 by Stephane Fillod
 *
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <hamlib/config.h>

//mdblack98 -- Commented this out -- is anybody using the functions in misc.h?
//If so, then those functions should be moved to rig.h
//#include "misc.h"

#include <limits.h>



#include <stdint.h>		// Use the C99 official header


typedef const char * const_char_string;


#include <hamlib/rig.h>


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


SWIGINTERNINLINE Tcl_Obj* 
SWIG_From_long  (long value)
{
  if (((long) INT_MIN <= value) && (value <= (long) INT_MAX)) {
    return Tcl_NewIntObj((int)(value));
  } else {
    return Tcl_NewLongObj(value);
  }
}


SWIGINTERNINLINE Tcl_Obj *
SWIG_From_int  (int value)
{    
  return SWIG_From_long  (value);
}


SWIGINTERN int
SWIG_AsVal_long SWIG_TCL_DECL_ARGS_2(Tcl_Obj *obj, long* val)
{
  long v;
  if (Tcl_GetLongFromObj(0,obj, &v) == TCL_OK) {
    if (val) *val = (long) v;
    return SWIG_OK;
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_int SWIG_TCL_DECL_ARGS_2(Tcl_Obj * obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (int)(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE Tcl_Obj *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  return (size < TCL_SIZE_MAX) ? Tcl_NewStringObj(carray, (Tcl_Size)(size)) : NULL;
}


SWIGINTERNINLINE Tcl_Obj * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}


#include <stdio.h>
#if (defined(_MSC_VER) && (_MSC_VER < 1900)) || defined(__BORLANDC__) || defined(_WATCOM)
# ifndef snprintf
#  define snprintf _snprintf
# endif
#endif


SWIGINTERNINLINE Tcl_Obj* 
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  if (value < (unsigned long) LONG_MAX) {
    return SWIG_From_long  ((long)(value));
  } else {
    char temp[256]; 
    SWIG_snprintf(temp, sizeof(temp), "%lu", value);
    return Tcl_NewStringObj(temp,-1);
  }
}


SWIGINTERNINLINE Tcl_Obj *
SWIG_From_unsigned_SS_int  (unsigned int value)
{    
  return SWIG_From_unsigned_SS_long  (value);
}

typedef union {
  struct {
    float min;                           
    float max;                           
    float step;                 
  } n;                                    
  struct {
    const char *combostr[16];                      
  } c;                                  
} confparams_u;

typedef struct {
  float min;                           
  float max;                           
  float step;                 
} confparams_u_n;

typedef struct {
  const char *combostr[16];                      
} confparams_u_c;


#include <float.h>


#include <math.h>


/* Getting isfinite working pre C99 across multiple platforms is non-trivial. Users can provide SWIG_isfinite on older platforms. */
#ifndef SWIG_isfinite
/* isfinite() is a macro for C99 */
# if defined(isfinite)
#  define SWIG_isfinite(X) (isfinite(X))
# elif defined(__cplusplus) && __cplusplus >= 201103L
/* Use a template so that this works whether isfinite() is std::isfinite() or
 * in the global namespace.  The reality seems to vary between compiler
 * versions.
 *
 * Make sure namespace std exists to avoid compiler warnings.
 *
 * extern "C++" is required as this fragment can end up inside an extern "C" { } block
 */
namespace std { }
extern "C++" template<typename T>
inline int SWIG_isfinite_func(T x) {
  using namespace std;
  return isfinite(x);
}
#  define SWIG_isfinite(X) (SWIG_isfinite_func(X))
# elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 2))
#  define SWIG_isfinite(X) (__builtin_isfinite(X))
# elif defined(_MSC_VER)
#  define SWIG_isfinite(X) (_finite(X))
# elif defined(__sun) && defined(__SVR4)
#  include <ieeefp.h>
#  define SWIG_isfinite(X) (finite(X))
# endif
#endif


/* Accept infinite as a valid float value unless we are unable to check if a value is finite */
#ifdef SWIG_isfinite
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX) && SWIG_isfinite(X))
#else
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX))
#endif


SWIGINTERN int
SWIG_AsVal_double SWIG_TCL_DECL_ARGS_2(Tcl_Obj *obj, double *val)
{
  double v;
  if (Tcl_GetDoubleFromObj(0, obj, &v) == TCL_OK) {
    if (val) *val = v;
    return SWIG_OK;
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_float SWIG_TCL_DECL_ARGS_2(Tcl_Obj * obj, float *val)
{
  double v;
  int res = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(obj, &v);
  if (SWIG_IsOK(res)) {
    if (SWIG_Float_Overflow_Check(v)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (float)(v);
    }
  }  
  return res;
}


  #define SWIG_From_double   Tcl_NewDoubleObj 


SWIGINTERNINLINE Tcl_Obj *
SWIG_From_float  (float value)
{    
  return SWIG_From_double  (value);
}


#include <errno.h>


SWIGINTERN int
SWIG_AsVal_unsigned_SS_long SWIG_TCL_DECL_ARGS_2(Tcl_Obj *obj, unsigned long *val) {
  long v;
  if (Tcl_GetLongFromObj(0,obj, &v) == TCL_OK) {
    if (v >= 0) {
      if (val) *val = (unsigned long) v;
      return SWIG_OK;
    }
    /* If v is negative, then this could be a negative number, or an
       unsigned value which doesn't fit in a signed long, so try to
       get it as a string so we can distinguish these cases. */
  }
  {
    Tcl_Size len = 0;
    const char *nptr = Tcl_GetStringFromObj(obj, &len);
    if (nptr && len > 0) {
      char *endptr;
      unsigned long v;
      if (*nptr == '-') return SWIG_OverflowError;
      errno = 0;
      v = strtoul(nptr, &endptr,0);
      if (nptr[0] == '\0' || *endptr != '\0')
	return SWIG_TypeError;
      if (v == ULONG_MAX && errno == ERANGE) {
	errno = 0;
	return SWIG_OverflowError;
      } else {
	if (*endptr == '\0') {
	  if (val) *val = v;
	  return SWIG_OK;
	}
      }
    }
  }
  
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_int SWIG_TCL_DECL_ARGS_2(Tcl_Obj * obj, unsigned int *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long SWIG_TCL_CALL_ARGS_2(obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UINT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (unsigned int)(v);
    }
  }  
  return res;
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(Tcl_Obj *obj, char** cptr, size_t* psize, int *alloc)
{ 
  Tcl_Size len = 0;
  char *cstr = Tcl_GetStringFromObj(obj, &len);
  if (cstr) {
    if (cptr)  *cptr = cstr;
    if (psize) *psize = len + 1;
    if (alloc) *alloc = SWIG_OLDOBJ;
    return SWIG_OK;
  }
  return SWIG_TypeError;
}




typedef struct {
  int l;                                
  unsigned char *d;                             
} value_t_b;


#if defined(LLONG_MAX) && !defined(SWIG_LONG_LONG_AVAILABLE)
#  define SWIG_LONG_LONG_AVAILABLE
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE Tcl_Obj* 
SWIG_From_long_SS_long  (long long value)
{
  if (((long long) LONG_MIN <= value) && (value <= (long long) LONG_MAX)) {
    return SWIG_From_long  ((long)(value));
  } else {    
    char temp[256]; 
    SWIG_snprintf(temp, sizeof(temp), "%lld", value);
    return Tcl_NewStringObj(temp,-1);
  }
}
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE Tcl_Obj* 
SWIG_From_unsigned_SS_long_SS_long  (unsigned long long value)
{
  if (value < (unsigned long long) LONG_MAX) {
    return SWIG_From_long_SS_long  ((long long)(value));
  } else {
    char temp[256]; 
    SWIG_snprintf(temp, sizeof(temp), "%llu", value);
    return Tcl_NewStringObj(temp,-1);
  }
}
#endif


#include <stdlib.h>
#ifdef _MSC_VER
# ifndef strtoull
#  define strtoull _strtoui64
# endif
# ifndef strtoll
#  define strtoll _strtoi64
# endif
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN int
SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_DECL_ARGS_2(Tcl_Obj *obj, unsigned long long *val)
{
  long v;
  if (Tcl_GetLongFromObj(0,obj, &v) == TCL_OK) {
    if (val) *val = (unsigned long) v;
    return SWIG_OK;
  } else {
    Tcl_Size len = 0;
    const char *nptr = Tcl_GetStringFromObj(obj, &len);
    if (nptr && len > 0) {
      char *endptr;
      unsigned long long v;
      if (*nptr == '-') return SWIG_OverflowError;
      errno = 0;
      v = strtoull(nptr, &endptr,0);
      if (nptr[0] == '\0' || *endptr != '\0')
	return SWIG_TypeError;
      if (v == ULLONG_MAX && errno == ERANGE) {
	errno = 0;
	return SWIG_OverflowError;
      } else {
	if (*endptr == '\0') {
	  if (val) *val = v;
	  return SWIG_OK;
	}
      }
    }
  }
  return SWIG_TypeError;
}
#endif


SWIGINTERN int
SWIG_AsCharArray(Tcl_Obj * obj, char *val, size_t size)
{ 
  char* cptr = 0; size_t csize = 0; int alloc = SWIG_OLDOBJ;
  int res = SWIG_AsCharPtrAndSize(obj, &cptr, &csize, &alloc);
  if (SWIG_IsOK(res)) {
    /* special case of single char conversion when we don't need space for NUL */
    if (size == 1 && csize == 2 && cptr && !cptr[1]) --csize;
    if (csize <= size) {
      if (val) {
	if (csize) memcpy(val, cptr, csize*sizeof(char));
	if (csize < size) memset(val + csize, 0, (size - csize)*sizeof(char));
      }
      if (alloc == SWIG_NEWOBJ) {
	free((char*)cptr);
	res = SWIG_DelNewMask(res);
      }      
      return res;
    }
    if (alloc == SWIG_NEWOBJ) free((char*)cptr);
  }
  return SWIG_TypeError;
}


SWIGINTERN size_t
SWIG_strnlen(const char* s, size_t maxlen)
{
  const char *p;
  for (p = s; maxlen-- && *p; p++)
    ;
  return p - s;
}

SWIGINTERN struct channel *new_channel(vfo_t vfo,int ch){
		channel_t *chan;

		chan = (channel_t*)malloc(sizeof(channel_t));
		if (!chan)
			return NULL;
		memset(chan, 0, sizeof(channel_t));
		chan->vfo = vfo;
		chan->channel_num = ch;
		return chan;
	}
SWIGINTERN void delete_channel(struct channel *self){
		/* TODO: free ext_levels */
		free(self);
	}
typedef struct {
  int raw;                                                                           
  int val;                                                                         
} cal_table_table;

typedef struct {
  int raw;                                         
  float val;                                        
} cal_table_float_table;


SWIGINTERNINLINE int
SWIG_AsVal_size_t SWIG_TCL_DECL_ARGS_2(Tcl_Obj * obj, size_t *val)
{
  int res = SWIG_TypeError;
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long SWIG_TCL_CALL_ARGS_2(obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = (size_t)(v);
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else if (sizeof(size_t) <= sizeof(unsigned long long)) {
    unsigned long long v;
    res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = (size_t)(v);
  }
#endif
  return res;
}


SWIGINTERNINLINE Tcl_Obj *
SWIG_From_size_t  (size_t value)
{    
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    return SWIG_From_unsigned_SS_long  ((unsigned long)(value));
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else {
    /* assume sizeof(size_t) <= sizeof(unsigned long long) */
    return SWIG_From_unsigned_SS_long_SS_long  ((unsigned long long)(value));
  }
#endif
}


SWIGINTERN int
SWIG_AsVal_short SWIG_TCL_DECL_ARGS_2(Tcl_Obj * obj, short *val)
{
  long v;
  int res = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < SHRT_MIN || v > SHRT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (short)(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE Tcl_Obj *
SWIG_From_short  (short value)
{    
  return SWIG_From_long  (value);
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_char SWIG_TCL_DECL_ARGS_2(Tcl_Obj * obj, unsigned char *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long SWIG_TCL_CALL_ARGS_2(obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UCHAR_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (unsigned char)(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE Tcl_Obj *
SWIG_From_unsigned_SS_char  (unsigned char value)
{    
  return SWIG_From_unsigned_SS_long  (value);
}



#pragma SWIG nowarn=451

#ifdef SWIGPYTHON
	// This mirrors "struct rig_callbacks" from rig.h using Python types
	typedef struct {
		PyObject *freq_event;     /*!< Frequency change event */
		PyObject *freq_arg;       /*!< Frequency change argument */
		PyObject *mode_event;     /*!< Mode change event */
		PyObject *mode_arg;       /*!< Mode change argument */
		PyObject *vfo_event;      /*!< VFO change event */
		PyObject *vfo_arg;        /*!< VFO change argument */
		PyObject *ptt_event;      /*!< PTT change event */
		PyObject *ptt_arg;        /*!< PTT change argument */
		PyObject *dcd_event;      /*!< DCD change event */
		PyObject *dcd_arg;        /*!< DCD change argument */
		PyObject *pltune_event;   /*!< Pipeline tuning module freq/mode/width callback NOTE: the C struct doesn't have the _event suffix*/
		PyObject *pltune_arg;     /*!< Pipeline tuning argument */
		PyObject *spectrum_event; /*!< Spectrum line reception event */
		PyObject *spectrum_arg;   /*!< Spectrum line reception argument */
		/* etc.. */
	} python_callbacks_t;
#endif

typedef struct Rig {
	RIG *rig;
	struct rig_caps *caps;		/* shortcut to RIG->caps */
	struct rig_state *state;	/* shortcut to RIG->state */
	int error_status;
	int do_exception;
#ifdef SWIGPYTHON
	python_callbacks_t *python_callbacks;
#endif
} Rig;

typedef char * char_string;
typedef channel_t * channel_t_p;
typedef channel_t * const_channel_t_p;

#ifdef SWIGPYTHON
int rig_freq_cb_python(RIG *rig, vfo_t vfo, freq_t freq, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(3,
                                    PyLong_FromLong(vfo),
                                    PyFloat_FromDouble(freq),
                                    self->python_callbacks->freq_arg
                                    );

    PyObject_CallObject(self->python_callbacks->freq_event, python_arguments);
    Py_XDECREF(python_arguments);

    return RIG_OK;
}

int rig_mode_cb_python(RIG *rig, vfo_t vfo, rmode_t rmode, pbwidth_t pbwidth, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(4,
                                    PyLong_FromLong(vfo),
                                    PyLong_FromLong(rmode),
                                    PyLong_FromLong(pbwidth),
                                    self->python_callbacks->mode_arg
                                    );

    PyObject_CallObject(self->python_callbacks->mode_event, python_arguments);

    Py_XDECREF(python_arguments);

    return RIG_OK;
}

int rig_vfo_cb_python(RIG *rig, vfo_t vfo, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(2,
                                    PyLong_FromLong(vfo),
                                    self->python_callbacks->vfo_arg
                                    );

    PyObject_CallObject(self->python_callbacks->vfo_event, python_arguments);

    Py_XDECREF(python_arguments);

    return RIG_OK;
}

int rig_ptt_cb_python(RIG *rig, vfo_t vfo, ptt_t ptt, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(3,
                                    PyLong_FromLong(vfo),
                                    PyLong_FromLong(ptt),
                                    self->python_callbacks->ptt_arg
                                    );

    PyObject_CallObject(self->python_callbacks->ptt_event, python_arguments);

    Py_XDECREF(python_arguments);

    return RIG_OK;
}

int rig_dcd_cb_python(RIG *rig, vfo_t vfo, dcd_t dcd, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(3,
                                    PyLong_FromLong(vfo),
                                    PyLong_FromLong(dcd),
                                    self->python_callbacks->dcd_arg
                                    );

    PyObject_CallObject(self->python_callbacks->dcd_event, python_arguments);

    Py_XDECREF(python_arguments);

    return RIG_OK;
}

int rig_pltune_cb_python(RIG *rig, vfo_t vfo, freq_t *freq, rmode_t *rmode, pbwidth_t *pbwidth, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(1,
                                    self->python_callbacks->pltune_arg
                                    );

    PyObject_CallObject(self->python_callbacks->pltune_event, python_arguments);

    Py_XDECREF(python_arguments);

    return RIG_OK;
}

int rig_spectrum_cb_python(RIG *rig, struct rig_spectrum_line *rig_spectrum_line, rig_ptr_t arg)
{
    Rig *self = arg;
    PyObject *python_arguments;

    python_arguments = PyTuple_Pack(1,
                                    self->python_callbacks->spectrum_arg
                                    );

    PyObject_CallObject(self->python_callbacks->spectrum_event, python_arguments);

    Py_XDECREF(python_arguments);

    return RIG_OK;
}
#endif

SWIGINTERN struct Rig *new_Rig(int rig_model){
		Rig *r;

		r = (Rig*)malloc(sizeof(Rig));
		if (!r)
			return NULL;
		r->rig = rig_init(rig_model);
		if (!r->rig) {
			free(r);
			return NULL;
		}







		/* install shortcuts */
		r->caps = r->rig->caps;
		r->state = rig_data_pointer(r->rig, RIG_PTRX_STATE);
		r->do_exception = 0;	/* default is disabled */
		r->error_status = RIG_OK;

		return r;
	}
SWIGINTERN void delete_Rig(struct Rig *self){
		rig_cleanup(self->rig);



		free(self);
	}
SWIGINTERN void Rig_open(struct Rig *self){
		self->error_status = rig_open(self->rig);
	}
SWIGINTERN void Rig_close(struct Rig *self){
		self->error_status = rig_close(self->rig);
	}
SWIGINTERN void Rig_set_freq(struct Rig *self,vfo_t vfo,freq_t _freq_t){ self->error_status = rig_set_freq(self->rig  ,vfo, _freq_t); }
SWIGINTERN void Rig_set_mode(struct Rig *self,rmode_t _rmode_t_1,pbwidth_t _pbwidth_t_2,vfo_t vfo){ self->error_status = rig_set_mode(self->rig  ,vfo, _rmode_t_1, _pbwidth_t_2); }
SWIGINTERN void Rig_set_ptt(struct Rig *self,vfo_t vfo,ptt_t _ptt_t){ self->error_status = rig_set_ptt(self->rig  ,vfo, _ptt_t); }
SWIGINTERN void Rig_set_rptr_shift(struct Rig *self,vfo_t vfo,rptr_shift_t _rptr_shift_t){ self->error_status = rig_set_rptr_shift(self->rig  ,vfo, _rptr_shift_t); }
SWIGINTERN void Rig_set_rptr_offs(struct Rig *self,vfo_t vfo,shortfreq_t _shortfreq_t){ self->error_status = rig_set_rptr_offs(self->rig  ,vfo, _shortfreq_t); }
SWIGINTERN void Rig_set_ctcss_tone(struct Rig *self,vfo_t vfo,tone_t _tone_t){ self->error_status = rig_set_ctcss_tone(self->rig  ,vfo, _tone_t); }
SWIGINTERN void Rig_set_dcs_code(struct Rig *self,vfo_t vfo,tone_t _tone_t){ self->error_status = rig_set_dcs_code(self->rig  ,vfo, _tone_t); }
SWIGINTERN void Rig_set_ctcss_sql(struct Rig *self,vfo_t vfo,tone_t _tone_t){ self->error_status = rig_set_ctcss_sql(self->rig  ,vfo, _tone_t); }
SWIGINTERN void Rig_set_dcs_sql(struct Rig *self,vfo_t vfo,tone_t _tone_t){ self->error_status = rig_set_dcs_sql(self->rig  ,vfo, _tone_t); }
SWIGINTERN void Rig_set_split_freq(struct Rig *self,vfo_t vfo,freq_t _freq_t){ self->error_status = rig_set_split_freq(self->rig  ,vfo, _freq_t); }
SWIGINTERN void Rig_set_split_mode(struct Rig *self,rmode_t _rmode_t_1,pbwidth_t _pbwidth_t_2,vfo_t vfo){ self->error_status = rig_set_split_mode(self->rig  ,vfo, _rmode_t_1, _pbwidth_t_2); }
SWIGINTERN void Rig_set_split_freq_mode(struct Rig *self,freq_t _freq_t_1,rmode_t _rmode_t_2,pbwidth_t _pbwidth_t_3,vfo_t vfo){ self->error_status = rig_set_split_freq_mode(self->rig  ,vfo, _freq_t_1, _rmode_t_2, _pbwidth_t_3); }
SWIGINTERN void Rig_set_split_vfo(struct Rig *self,split_t _split_t_1,vfo_t _vfo_t_2,vfo_t vfo){ self->error_status = rig_set_split_vfo(self->rig  ,vfo, _split_t_1, _vfo_t_2); }
SWIGINTERN void Rig_set_rit(struct Rig *self,vfo_t vfo,shortfreq_t _shortfreq_t){ self->error_status = rig_set_rit(self->rig  ,vfo, _shortfreq_t); }
SWIGINTERN void Rig_set_xit(struct Rig *self,vfo_t vfo,shortfreq_t _shortfreq_t){ self->error_status = rig_set_xit(self->rig  ,vfo, _shortfreq_t); }
SWIGINTERN void Rig_set_ts(struct Rig *self,vfo_t vfo,shortfreq_t _shortfreq_t){ self->error_status = rig_set_ts(self->rig  ,vfo, _shortfreq_t); }
SWIGINTERN void Rig_set_ant(struct Rig *self,ant_t _ant_t_1,value_t _value_t_2,vfo_t vfo){ self->error_status = rig_set_ant(self->rig  ,vfo, _ant_t_1, _value_t_2); }
SWIGINTERN void Rig_set_func(struct Rig *self,setting_t _setting_t_1,int _int_2,vfo_t vfo){ self->error_status = rig_set_func(self->rig  ,vfo, _setting_t_1, _int_2); }
SWIGINTERN void Rig_set_ext_func(struct Rig *self,hamlib_token_t _hamlib_token_t_1,int _int_2,vfo_t vfo){ self->error_status = rig_set_ext_func(self->rig  ,vfo, _hamlib_token_t_1, _int_2); }
SWIGINTERN void Rig_set_bank(struct Rig *self,vfo_t vfo,int _int){ self->error_status = rig_set_bank(self->rig  ,vfo, _int); }
SWIGINTERN void Rig_set_mem(struct Rig *self,vfo_t vfo,int _int){ self->error_status = rig_set_mem(self->rig  ,vfo, _int); }
SWIGINTERN void Rig_send_dtmf(struct Rig *self,vfo_t vfo,const_char_string _const_char_string){ self->error_status = rig_send_dtmf(self->rig  ,vfo, _const_char_string); }
SWIGINTERN void Rig_send_morse(struct Rig *self,vfo_t vfo,const_char_string _const_char_string){ self->error_status = rig_send_morse(self->rig  ,vfo, _const_char_string); }
SWIGINTERN void Rig_vfo_op(struct Rig *self,vfo_t vfo,vfo_op_t _vfo_op_t){ self->error_status = rig_vfo_op(self->rig  ,vfo, _vfo_op_t); }
SWIGINTERN void Rig_scan(struct Rig *self,scan_t _scan_t_1,int _int_2,vfo_t vfo){ self->error_status = rig_scan(self->rig  ,vfo, _scan_t_1, _int_2); }
SWIGINTERN void Rig_set_level__SWIG_0(struct Rig *self,setting_t stg,int i,vfo_t vfo){ value_t val = {0}; 	if (((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.i = i;	 		self->error_status = rig_set_level(self->rig  ,vfo , stg, val);	 	}
SWIGINTERN void Rig_set_level__SWIG_1(struct Rig *self,setting_t stg,float f,vfo_t vfo){ value_t val = {0}; 	if (!((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.f = f;	 		self->error_status = rig_set_level(self->rig  ,vfo , stg, val);	 	}
SWIGINTERN void Rig_set_level__SWIG_2(struct Rig *self,char const *name,int i,vfo_t vfo){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_level(name);	 		if (!rig_has_set_level(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return;	 			}  			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				val.i = i;	 				break;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				val.i = (int)i;	 				break;	 			case RIG_CONF_STRING:		 				self->error_status = -RIG_EINVAL;		 				return;	 			default:	 				self->error_status = -RIG_ECONF;	 				return;	 			}	 			self->error_status = rig_set_ext_level(self->rig  ,vfo, cfp->token, val);	 			return;	 		}	 		if (((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.i = i;	 		self->error_status = rig_set_level(self->rig  ,vfo , stg, val);	 	}
SWIGINTERN void Rig_set_level__SWIG_3(struct Rig *self,char const *name,float f,vfo_t vfo){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_level(name);	 		if (!rig_has_set_level(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return;	 			}  			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				val.f = f;	 				break;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				val.i = (int)f;	 				break;	 			case RIG_CONF_STRING:		 				self->error_status = -RIG_EINVAL;		 				return;	 			default:	 				self->error_status = -RIG_ECONF;	 				return;	 			}	 			self->error_status = rig_set_ext_level(self->rig  ,vfo, cfp->token, val);	 			return;	 		}	 		if (!((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.f = f;	 		self->error_status = rig_set_level(self->rig  ,vfo , stg, val);	 	}
SWIGINTERN void Rig_set_level__SWIG_4(struct Rig *self,char const *name,char const *s,vfo_t vfo){ value_t val = {0};    		const struct confparams *cfp;  		cfp = rig_ext_lookup(self->rig, name);  		if (!cfp) {	 			self->error_status = -RIG_EINVAL;	  			return;  		}  		if (cfp->type != RIG_CONF_STRING) {  			self->error_status = -RIG_EINVAL;	  			return;  		}  		val.cs = s;  		self->error_status = rig_set_ext_level(self->rig ,vfo, cfp->token, val);  	}
SWIGINTERN void Rig_set_ext_level(struct Rig *self,hamlib_token_t _hamlib_token_t_1,value_t _value_t_2,vfo_t vfo){ self->error_status = rig_set_ext_level(self->rig  ,vfo, _hamlib_token_t_1, _value_t_2); }
SWIGINTERN int Rig_get_level_i__SWIG_0(struct Rig *self,setting_t stg,vfo_t vfo){ value_t val = {0}; 	if (((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return val.i;			 		}	 		self->error_status = rig_get_level(self->rig  ,vfo , stg, &val);	 		return val.i;	 	}
SWIGINTERN float Rig_get_level_f__SWIG_0(struct Rig *self,setting_t stg,vfo_t vfo){ value_t val = {0}; 	if (!((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return val.f;			 		}	 		self->error_status = rig_get_level(self->rig  ,vfo , stg, &val);	 		return val.f;	 	}
SWIGINTERN int Rig_get_level_i__SWIG_1(struct Rig *self,char const *name,vfo_t vfo){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_level(name);	 		if (!rig_has_get_level(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return val.i;		  			}  			if (cfp->type == RIG_CONF_STRING) {  				self->error_status = -RIG_EINVAL;	 				return val.i;	  			}	 			self->error_status = rig_get_ext_level(self->rig  ,vfo, cfp->token, &val);	 			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				return (int)val.f;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				return (int)val.i;	 			default:	 				self->error_status = -RIG_ECONF;	 				return val.i;	  			}	 		}	 		if (((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return val.i;	  		}	 		self->error_status = rig_get_level(self->rig  ,vfo , stg, &val);	 		return val.i;	 	}
SWIGINTERN float Rig_get_level_f__SWIG_1(struct Rig *self,char const *name,vfo_t vfo){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_level(name);	 		if (!rig_has_get_level(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return val.f;		  			}  			if (cfp->type == RIG_CONF_STRING) {  				self->error_status = -RIG_EINVAL;	 				return val.f;	  			}	 			self->error_status = rig_get_ext_level(self->rig  ,vfo, cfp->token, &val);	 			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				return (float)val.f;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				return (float)val.i;	 			default:	 				self->error_status = -RIG_ECONF;	 				return val.f;	  			}	 		}	 		if (!((stg)&((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))) {	 			self->error_status = -RIG_EINVAL;		 			return val.f;	  		}	 		self->error_status = rig_get_level(self->rig  ,vfo , stg, &val);	 		return val.f;	 	}
SWIGINTERN void Rig_get_level__SWIG_0(struct Rig *self,char const *name,char *returnstr,vfo_t vfo){ value_t val = {0};   		const struct confparams *cfp;  		returnstr[0] = '\0';  		cfp = rig_ext_lookup(self->rig, name);  		if (!cfp || cfp->type != RIG_CONF_STRING) {	 			self->error_status = -RIG_EINVAL;	  			return;  		}  		val.s = returnstr;  		self->error_status = rig_get_ext_level(self->rig ,vfo, cfp->token, &val);  	}
SWIGINTERN void Rig_set_vfo(struct Rig *self,vfo_t _vfo_t){ self->error_status = rig_set_vfo(self->rig  , _vfo_t); }
SWIGINTERN void Rig_set_powerstat(struct Rig *self,powerstat_t _powerstat_t){ self->error_status = rig_set_powerstat(self->rig  , _powerstat_t); }
SWIGINTERN void Rig_has_get_level(struct Rig *self,setting_t _setting_t){ self->error_status = rig_has_get_level(self->rig  , _setting_t); }
SWIGINTERN void Rig_has_get_parm(struct Rig *self,setting_t _setting_t){ self->error_status = rig_has_get_parm(self->rig  , _setting_t); }
SWIGINTERN void Rig_has_set_parm(struct Rig *self,setting_t _setting_t){ self->error_status = rig_has_set_parm(self->rig  , _setting_t); }
SWIGINTERN void Rig_has_get_func(struct Rig *self,setting_t _setting_t){ self->error_status = rig_has_get_func(self->rig  , _setting_t); }
SWIGINTERN void Rig_has_set_func(struct Rig *self,setting_t _setting_t){ self->error_status = rig_has_set_func(self->rig  , _setting_t); }
SWIGINTERN void Rig_reset(struct Rig *self,reset_t _reset_t){ self->error_status = rig_reset(self->rig  , _reset_t); }
SWIGINTERN void Rig_has_scan(struct Rig *self,scan_t _scan_t){ self->error_status = rig_has_scan(self->rig  , _scan_t); }
SWIGINTERN void Rig_has_vfo_op(struct Rig *self,vfo_op_t _vfo_op_t){ self->error_status = rig_has_vfo_op(self->rig  , _vfo_op_t); }
SWIGINTERN void Rig_passband_normal(struct Rig *self,rmode_t _rmode_t){ self->error_status = rig_passband_normal(self->rig  , _rmode_t); }
SWIGINTERN void Rig_passband_narrow(struct Rig *self,rmode_t _rmode_t){ self->error_status = rig_passband_narrow(self->rig  , _rmode_t); }
SWIGINTERN void Rig_passband_wide(struct Rig *self,rmode_t _rmode_t){ self->error_status = rig_passband_wide(self->rig  , _rmode_t); }
SWIGINTERN void Rig_set_vfo_opt(struct Rig *self,int _int){ self->error_status = rig_set_vfo_opt(self->rig  , _int); }
SWIGINTERN void Rig_ext_token_lookup(struct Rig *self,const_char_string _const_char_string){ self->error_status = rig_ext_token_lookup(self->rig  , _const_char_string); }
SWIGINTERN void Rig_token_lookup(struct Rig *self,const_char_string _const_char_string){ self->error_status = rig_token_lookup(self->rig  , _const_char_string); }
SWIGINTERN void Rig_set_conf__SWIG_0(struct Rig *self,hamlib_token_t _hamlib_token_t_1,const_char_string _const_char_string_2){ self->error_status = rig_set_conf(self->rig  , _hamlib_token_t_1, _const_char_string_2); }
SWIGINTERN void Rig_set_ext_parm(struct Rig *self,hamlib_token_t _hamlib_token_t_1,value_t _value_t_2){ self->error_status = rig_set_ext_parm(self->rig  , _hamlib_token_t_1, _value_t_2); }
SWIGINTERN void Rig_set_parm__SWIG_0(struct Rig *self,setting_t stg,int i){ value_t val = {0}; 	if (((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.i = i;	 		self->error_status = rig_set_parm(self->rig   , stg, val);	 	}
SWIGINTERN void Rig_set_parm__SWIG_1(struct Rig *self,setting_t stg,float f){ value_t val = {0}; 	if (!((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.f = f;	 		self->error_status = rig_set_parm(self->rig   , stg, val);	 	}
SWIGINTERN void Rig_set_parm__SWIG_2(struct Rig *self,char const *name,int i){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_parm(name);	 		if (!rig_has_set_parm(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return;	 			}  			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				val.i = i;	 				break;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				val.i = (int)i;	 				break;	 			case RIG_CONF_STRING:		 				self->error_status = -RIG_EINVAL;		 				return;	 			default:	 				self->error_status = -RIG_ECONF;	 				return;	 			}	 			self->error_status = rig_set_ext_parm(self->rig  , cfp->token, val);	 			return;	 		}	 		if (((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.i = i;	 		self->error_status = rig_set_parm(self->rig   , stg, val);	 	}
SWIGINTERN void Rig_set_parm__SWIG_3(struct Rig *self,char const *name,float f){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_parm(name);	 		if (!rig_has_set_parm(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return;	 			}  			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				val.f = f;	 				break;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				val.i = (int)f;	 				break;	 			case RIG_CONF_STRING:		 				self->error_status = -RIG_EINVAL;		 				return;	 			default:	 				self->error_status = -RIG_ECONF;	 				return;	 			}	 			self->error_status = rig_set_ext_parm(self->rig  , cfp->token, val);	 			return;	 		}	 		if (!((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return;	 		}	 		val.f = f;	 		self->error_status = rig_set_parm(self->rig   , stg, val);	 	}
SWIGINTERN void Rig_set_parm__SWIG_4(struct Rig *self,char const *name,char const *s){ value_t val = {0};    		const struct confparams *cfp;  		cfp = rig_ext_lookup(self->rig, name);  		if (!cfp) {	 			self->error_status = -RIG_EINVAL;	  			return;  		}  		if (cfp->type != RIG_CONF_STRING) {  			self->error_status = -RIG_EINVAL;	  			return;  		}  		val.cs = s;  		self->error_status = rig_set_ext_parm(self->rig , cfp->token, val);  	}
SWIGINTERN int Rig_get_parm_i__SWIG_0(struct Rig *self,setting_t stg){ value_t val = {0}; 	if (((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return val.i;			 		}	 		self->error_status = rig_get_parm(self->rig   , stg, &val);	 		return val.i;	 	}
SWIGINTERN float Rig_get_parm_f__SWIG_0(struct Rig *self,setting_t stg){ value_t val = {0}; 	if (!((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return val.f;			 		}	 		self->error_status = rig_get_parm(self->rig   , stg, &val);	 		return val.f;	 	}
SWIGINTERN int Rig_get_parm_i__SWIG_1(struct Rig *self,char const *name){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_parm(name);	 		if (!rig_has_get_parm(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return val.i;		  			}  			if (cfp->type == RIG_CONF_STRING) {  				self->error_status = -RIG_EINVAL;	 				return val.i;	  			}	 			self->error_status = rig_get_ext_parm(self->rig  , cfp->token, &val);	 			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				return (int)val.f;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				return (int)val.i;	 			default:	 				self->error_status = -RIG_ECONF;	 				return val.i;	  			}	 		}	 		if (((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return val.i;	  		}	 		self->error_status = rig_get_parm(self->rig   , stg, &val);	 		return val.i;	 	}
SWIGINTERN float Rig_get_parm_f__SWIG_1(struct Rig *self,char const *name){ setting_t stg; value_t val = {0}; 	 		stg = rig_parse_parm(name);	 		if (!rig_has_get_parm(self->rig, stg)) {	 			const struct confparams *cfp;	 			cfp = rig_ext_lookup(self->rig, name);	 			if (!cfp) {	 				self->error_status = -RIG_EINVAL;	  				return val.f;		  			}  			if (cfp->type == RIG_CONF_STRING) {  				self->error_status = -RIG_EINVAL;	 				return val.f;	  			}	 			self->error_status = rig_get_ext_parm(self->rig  , cfp->token, &val);	 			switch (cfp->type) {	 			case RIG_CONF_NUMERIC:	 				return (float)val.f;	 			case RIG_CONF_CHECKBUTTON:	 			case RIG_CONF_COMBO:		 				return (float)val.i;	 			default:	 				self->error_status = -RIG_ECONF;	 				return val.f;	  			}	 		}	 		if (!((stg)&(RIG_PARM_BACKLIGHT|RIG_PARM_BAT|RIG_PARM_KEYLIGHT|RIG_PARM_BACKLIGHT))) {	 			self->error_status = -RIG_EINVAL;		 			return val.f;	  		}	 		self->error_status = rig_get_parm(self->rig   , stg, &val);	 		return val.f;	 	}
SWIGINTERN void Rig_get_parm__SWIG_0(struct Rig *self,char const *name,char *returnstr){ value_t val = {0};   		const struct confparams *cfp;  		returnstr[0] = '\0';  		cfp = rig_ext_lookup(self->rig, name);  		if (!cfp || cfp->type != RIG_CONF_STRING) {	 			self->error_status = -RIG_EINVAL;	  			return;  		}  		val.s = returnstr;  		self->error_status = rig_get_ext_parm(self->rig , cfp->token, &val);  	}
SWIGINTERN void Rig_set_conf__SWIG_1(struct Rig *self,char const *name,char const *val){
		hamlib_token_t tok = rig_token_lookup(self->rig, name);
		if (tok == 0)
			self->error_status = -RIG_EINVAL;
		else
			self->error_status = rig_set_conf(self->rig, tok, val);
	}
SWIGINTERN freq_t Rig_get_freq(struct Rig *self,vfo_t vfo){ freq_t _freq_t; self->error_status = rig_get_freq(self->rig, vfo, &_freq_t); return _freq_t; }
SWIGINTERN vfo_t Rig_get_vfo(struct Rig *self){ vfo_t _vfo_t; self->error_status = rig_get_vfo(self->rig, &_vfo_t); return _vfo_t; }
SWIGINTERN ptt_t Rig_get_ptt(struct Rig *self,vfo_t vfo){ ptt_t _ptt_t; self->error_status = rig_get_ptt(self->rig, vfo, &_ptt_t); return _ptt_t; }
SWIGINTERN rptr_shift_t Rig_get_rptr_shift(struct Rig *self,vfo_t vfo){ rptr_shift_t _rptr_shift_t; self->error_status = rig_get_rptr_shift(self->rig, vfo, &_rptr_shift_t); return _rptr_shift_t; }
SWIGINTERN shortfreq_t Rig_get_rptr_offs(struct Rig *self,vfo_t vfo){ shortfreq_t _shortfreq_t; self->error_status = rig_get_rptr_offs(self->rig, vfo, &_shortfreq_t); return _shortfreq_t; }
SWIGINTERN tone_t Rig_get_ctcss_tone(struct Rig *self,vfo_t vfo){ tone_t _tone_t; self->error_status = rig_get_ctcss_tone(self->rig, vfo, &_tone_t); return _tone_t; }
SWIGINTERN tone_t Rig_get_dcs_code(struct Rig *self,vfo_t vfo){ tone_t _tone_t; self->error_status = rig_get_dcs_code(self->rig, vfo, &_tone_t); return _tone_t; }
SWIGINTERN tone_t Rig_get_ctcss_sql(struct Rig *self,vfo_t vfo){ tone_t _tone_t; self->error_status = rig_get_ctcss_sql(self->rig, vfo, &_tone_t); return _tone_t; }
SWIGINTERN tone_t Rig_get_dcs_sql(struct Rig *self,vfo_t vfo){ tone_t _tone_t; self->error_status = rig_get_dcs_sql(self->rig, vfo, &_tone_t); return _tone_t; }
SWIGINTERN freq_t Rig_get_split_freq(struct Rig *self,vfo_t vfo){ freq_t _freq_t; self->error_status = rig_get_split_freq(self->rig, vfo, &_freq_t); return _freq_t; }
SWIGINTERN void Rig_get_split_vfo(struct Rig *self,split_t *split,vfo_t *tx_vfo,vfo_t vfo){ self->error_status = rig_get_split_vfo(self->rig, vfo, split, tx_vfo); }
SWIGINTERN shortfreq_t Rig_get_rit(struct Rig *self,vfo_t vfo){ shortfreq_t _shortfreq_t; self->error_status = rig_get_rit(self->rig, vfo, &_shortfreq_t); return _shortfreq_t; }
SWIGINTERN shortfreq_t Rig_get_xit(struct Rig *self,vfo_t vfo){ shortfreq_t _shortfreq_t; self->error_status = rig_get_xit(self->rig, vfo, &_shortfreq_t); return _shortfreq_t; }
SWIGINTERN shortfreq_t Rig_get_ts(struct Rig *self,vfo_t vfo){ shortfreq_t _shortfreq_t; self->error_status = rig_get_ts(self->rig, vfo, &_shortfreq_t); return _shortfreq_t; }
SWIGINTERN void Rig_get_vfo_info(struct Rig *self,int *satmode,split_t *split,pbwidth_t *width,rmode_t *mode,freq_t *freq,vfo_t vfo){ self->error_status = rig_get_vfo_info(self->rig, vfo, freq, mode, width, split, satmode); }
SWIGINTERN int Rig_get_mem(struct Rig *self,vfo_t vfo){ int _int; self->error_status = rig_get_mem(self->rig, vfo, &_int); return _int; }
SWIGINTERN powerstat_t Rig_get_powerstat(struct Rig *self){ powerstat_t _powerstat_t; self->error_status = rig_get_powerstat(self->rig, &_powerstat_t); return _powerstat_t; }
SWIGINTERN dcd_t Rig_get_dcd(struct Rig *self,vfo_t vfo){ dcd_t _dcd_t; self->error_status = rig_get_dcd(self->rig, vfo, &_dcd_t); return _dcd_t; }
SWIGINTERN int Rig_mem_count(struct Rig *self){
		return rig_mem_count(self->rig);
	}
SWIGINTERN chan_t const *Rig_lookup_mem_caps(struct Rig *self,int channel_num){
		return rig_lookup_mem_caps(self->rig, channel_num);
	}
SWIGINTERN void Rig_set_channel(struct Rig *self,struct channel const *chan){
		self->error_status = rig_set_channel(self->rig, 0, chan);
	}
SWIGINTERN void Rig_get_channel__SWIG_0(struct Rig *self,struct channel *chan,int read_only){
		self->error_status = rig_get_channel(self->rig, 0, chan, read_only);
		/* TODO: handle ext_level's */
	}
SWIGINTERN void Rig_get_chan_all__SWIG_0(struct Rig *self,struct channel *chans){
		self->error_status = rig_get_chan_all(self->rig, 0, chans);
		/* TODO: handle ext_level's */
	}
SWIGINTERN struct channel *Rig_get_channel__SWIG_1(struct Rig *self,int read_only,int channel_num,vfo_t vfo){
		struct channel *chan;
		chan = new_channel(channel_num != INT_MAX ? vfo : (1u<<(29)), channel_num);
		if (!chan)
		{
			self->error_status = -RIG_ENOMEM;
			return NULL;
		}
		self->error_status = rig_get_channel(self->rig, 0, chan, read_only);
		/* TODO: copy ext_level's */
		return chan;
	}
SWIGINTERN void Rig_get_conf__SWIG_0(struct Rig *self,hamlib_token_t tok,char *returnstr){
		returnstr[0] = '\0';
		self->error_status = rig_get_conf2(self->rig, tok, returnstr, 256);
	}
SWIGINTERN void Rig_get_conf__SWIG_1(struct Rig *self,char const *name,char *returnstr){
		returnstr[0] = '\0';
		hamlib_token_t tok = rig_token_lookup(self->rig, name);
		if (tok == 0)
			self->error_status = -RIG_EINVAL;
		else
			self->error_status = rig_get_conf2(self->rig, tok, returnstr, 256);
	}
SWIGINTERN void Rig_recv_dtmf(struct Rig *self,char *returnstr,vfo_t vfo){
		int len = 256;
		self->error_status = rig_recv_dtmf(self->rig, vfo, returnstr, &len);
		returnstr[len] = '\0';
	}
SWIGINTERN char const *Rig_get_info(struct Rig *self){
		const char *s;
		s = rig_get_info(self->rig);
		self->error_status = s ? RIG_OK : -RIG_EINVAL;
		return s;
	}
SWIGINTERN int Rig_get_func(struct Rig *self,setting_t func,vfo_t vfo){
		int status;
		self->error_status = rig_get_func(self->rig, vfo, func, &status);
		return status;
	}
SWIGINTERN int Rig_get_ext_func(struct Rig *self,hamlib_token_t func,vfo_t vfo){
		int status;
		self->error_status = rig_get_ext_func(self->rig, vfo, func, &status);
		return status;
	}
SWIGINTERN void Rig_get_level__SWIG_1(struct Rig *self,setting_t level,vfo_t vfo){ value_t val = {0}; self->error_status = rig_get_level(self->rig, vfo, level, &val);
			//if (RIG_LEVEL_IS_FLOAT(level))
			/* TODO: dynamic casting */
		}
SWIGINTERN void Rig_get_ext_level(struct Rig *self,hamlib_token_t token,vfo_t vfo){ value_t val = {0}; self->error_status = rig_get_level(self->rig, vfo, token, &val);
			//if (RIG_LEVEL_IS_FLOAT(level))
			/* TODO: dynamic casting */
		}
SWIGINTERN void Rig_get_parm__SWIG_1(struct Rig *self,setting_t parm){ value_t val = {0}; self->error_status = rig_get_parm(self->rig, parm, &val);
			//if (RIG_LEVEL_IS_FLOAT(parm))
			/* TODO: dynamic casting */
		}
SWIGINTERN void Rig_get_ext_parm(struct Rig *self,hamlib_token_t parm){ value_t val = {0}; self->error_status = rig_get_parm(self->rig, parm, &val);
			//if (RIG_LEVEL_IS_FLOAT(parm))
			/* TODO: dynamic casting */
		}

typedef struct channel channelArray;

SWIGINTERN channelArray *new_channelArray(size_t nelements){
    return (struct channel *)calloc(nelements, sizeof(struct channel));
  }
SWIGINTERN void delete_channelArray(channelArray *self){
    free((char*)self);
  }
SWIGINTERN struct channel channelArray_getitem(channelArray *self,size_t index){
    return self[index];
  }
SWIGINTERN void channelArray_setitem(channelArray *self,size_t index,struct channel value){
    self[index] = value;
  }
SWIGINTERN struct channel *channelArray_cast(channelArray *self){
    return self;
  }
SWIGINTERN channelArray *channelArray_frompointer(struct channel *t){
    return (channelArray *)(t);
  }

typedef tone_t toneArray;

SWIGINTERN toneArray *new_toneArray(size_t nelements){
    return (tone_t *)calloc(nelements, sizeof(tone_t));
  }
SWIGINTERN void delete_toneArray(toneArray *self){
    free((char*)self);
  }
SWIGINTERN tone_t toneArray_getitem(toneArray *self,size_t index){
    return self[index];
  }
SWIGINTERN void toneArray_setitem(toneArray *self,size_t index,tone_t value){
    self[index] = value;
  }
SWIGINTERN tone_t *toneArray_cast(toneArray *self){
    return self;
  }
SWIGINTERN toneArray *toneArray_frompointer(tone_t *t){
    return (toneArray *)(t);
  }


/*
 * these ones return 2 values, here is a perl example:
 * 	($mode, $width) = $rig->get_mode();
 */
void Rig_get_mode(Rig *self, rmode_t *mode, pbwidth_t *width, vfo_t vfo)
{
	self->error_status = rig_get_mode(self->rig, vfo, mode, width);
}

void Rig_get_split_mode(Rig *self, rmode_t *mode, pbwidth_t *width, vfo_t vfo)
{
	self->error_status = rig_get_split_mode(self->rig, vfo, mode, width);
}

/*
 * these ones return 3 values
 */
void Rig_get_split_freq_mode(Rig *self, vfo_t vfo, freq_t *tx_freq, rmode_t *tx_mode, pbwidth_t *tx_width)
{
	self->error_status = rig_get_split_freq_mode(self->rig, vfo, tx_freq, tx_mode, tx_width);
}


/*
 * these ones return 4 values
 */
void Rig_get_ant(Rig *self, ant_t *ant_rx, ant_t *ant_tx, ant_t *ant_curr, signed int *option, ant_t ant, vfo_t vfo)
{
	value_t value;
	self->error_status = rig_get_ant(self->rig, vfo, ant, &value, ant_curr, ant_tx, ant_rx);
	*option = value.i;
}

struct channel *Rig_get_chan_all(Rig *self)
{
	struct channel *chans;
	int nb_chans = rig_mem_count(self->rig);

	/* TODO: memleak: typemap to release that memory */
	chans = calloc(sizeof (struct channel), nb_chans);
	if (!chans)
	{
		self->error_status = -RIG_ENOMEM;
		return NULL;
	}
	self->error_status = rig_get_chan_all(self->rig, RIG_VFO_NONE, chans);
	/* TODO: copy ext_level's */
	return chans;
}



#include <hamlib/rotator.h>



typedef struct Rot {
	ROT *rot;
	struct rot_caps *caps;		/* shortcut to ROT->caps */
	struct rot_state *state;	/* shortcut to ROT->state */
	int error_status;
	int do_exception;
} Rot;


SWIGINTERN struct Rot *new_Rot(rot_model_t rot_model){
		Rot *r;

		r = (Rot*)malloc(sizeof(Rot));
		if (!r)
			return NULL;
		r->rot = rot_init(rot_model);
		if (!r->rot) {
			free(r);
			return NULL;
		}
		/* install shortcuts */
		r->caps = r->rot->caps;
		r->state = &r->rot->state; rot_data_pointer(r->rot, RIG_PTRX_ROTSTATE);
		r->do_exception = 0;    /* default is disabled */
		r->error_status = RIG_OK;
		return r;
	}
SWIGINTERN void delete_Rot(struct Rot *self){
		rot_cleanup(self->rot);
		free(self);
	}
SWIGINTERN void Rot_open(struct Rot *self){ self->error_status = rot_open(self->rot); }
SWIGINTERN void Rot_close(struct Rot *self){ self->error_status = rot_close(self->rot); }
SWIGINTERN void Rot_set_ext_func(struct Rot *self,hamlib_token_t _hamlib_token_t_1,int _int_2){ self->error_status = rot_set_ext_func(self->rot, _hamlib_token_t_1, _int_2); }
SWIGINTERN void Rot_set_ext_level(struct Rot *self,hamlib_token_t _hamlib_token_t_1,value_t _value_t_2){ self->error_status = rot_set_ext_level(self->rot, _hamlib_token_t_1, _value_t_2); }
SWIGINTERN void Rot_set_ext_parm(struct Rot *self,hamlib_token_t _hamlib_token_t_1,value_t _value_t_2){ self->error_status = rot_set_ext_parm(self->rot, _hamlib_token_t_1, _value_t_2); }
SWIGINTERN void Rot_set_func(struct Rot *self,setting_t _setting_t_1,int _int_2){ self->error_status = rot_set_func(self->rot, _setting_t_1, _int_2); }
SWIGINTERN void Rot_set_level(struct Rot *self,setting_t _setting_t_1,value_t _value_t_2){ self->error_status = rot_set_level(self->rot, _setting_t_1, _value_t_2); }
SWIGINTERN void Rot_set_parm(struct Rot *self,setting_t _setting_t_1,value_t _value_t_2){ self->error_status = rot_set_parm(self->rot, _setting_t_1, _value_t_2); }
SWIGINTERN void Rot_set_position(struct Rot *self,azimuth_t _azimuth_t_1,elevation_t _elevation_t_2){ self->error_status = rot_set_position(self->rot, _azimuth_t_1, _elevation_t_2); }
SWIGINTERN void Rot_stop(struct Rot *self){ self->error_status = rot_stop(self->rot); }
SWIGINTERN void Rot_park(struct Rot *self){ self->error_status = rot_park(self->rot); }
SWIGINTERN void Rot_reset(struct Rot *self,rot_reset_t _rot_reset_t){ self->error_status = rot_reset(self->rot, _rot_reset_t); }
SWIGINTERN void Rot_move(struct Rot *self,int _int_1,int _int_2){ self->error_status = rot_move(self->rot, _int_1, _int_2); }
SWIGINTERN void Rot_token_lookup(struct Rot *self,const_char_string _const_char_string){ self->error_status = rot_token_lookup(self->rot, _const_char_string); }
SWIGINTERN void Rot_set_conf__SWIG_0(struct Rot *self,char const *name,char const *val){
		hamlib_token_t tok = rot_token_lookup(self->rot, name);
		if (tok == 0)
			self->error_status = -RIG_EINVAL;
		else
			self->error_status = rot_set_conf(self->rot, tok, val);
	}
SWIGINTERN void Rot_set_conf__SWIG_1(struct Rot *self,hamlib_token_t _hamlib_token_t_1,const_char_string _const_char_string_2){ self->error_status = rot_set_conf(self->rot, _hamlib_token_t_1, _const_char_string_2); }
SWIGINTERN setting_t Rot_has_get_func(struct Rot *self,setting_t func){
		return rot_has_get_func(self->rot, func);
	}
SWIGINTERN setting_t Rot_has_set_func(struct Rot *self,setting_t func){
		return rot_has_set_func(self->rot, func);
	}
SWIGINTERN setting_t Rot_has_get_level(struct Rot *self,setting_t level){
		return rot_has_get_level(self->rot, level);
	}
SWIGINTERN setting_t Rot_has_set_level(struct Rot *self,setting_t level){
		return rot_has_set_level(self->rot, level);
	}
SWIGINTERN setting_t Rot_has_get_parm(struct Rot *self,setting_t parm){
		return rot_has_get_parm(self->rot, parm);
	}
SWIGINTERN setting_t Rot_has_set_parm(struct Rot *self,setting_t parm){
		return rot_has_set_parm(self->rot, parm);
	}
SWIGINTERN char const *Rot_get_conf__SWIG_0(struct Rot *self,hamlib_token_t tok){
                static char s[128] = "";
                self->error_status = rot_get_conf2(self->rot, tok, s, sizeof s);
                return s;
        }
SWIGINTERN char const *Rot_get_conf__SWIG_1(struct Rot *self,char const *name){
                hamlib_token_t tok = rot_token_lookup(self->rot, name);
                static char s[128] = "";
                if (tok == 0)
                        self->error_status = -RIG_EINVAL;
                else
                        self->error_status = rot_get_conf2(self->rot, tok, s, sizeof s);
                return s;
        }
SWIGINTERN char const *Rot_get_info(struct Rot *self){
                const char *s;
                s = rot_get_info(self->rot);
                self->error_status = s ? RIG_OK : -RIG_EINVAL;
                return s;
        }


/*
 * this one returns 2 values, here is a perl example:
 *      ($az, $elevation) = $rig->get_position();
 */
void Rot_get_position(Rot *self, azimuth_t *azimuth, elevation_t *elevation)
{
        self->error_status = rot_get_position(self->rot, azimuth, elevation);
}



#include <hamlib/amplifier.h>



typedef struct Amp {
	AMP *amp;
	struct amp_caps *caps;		/* shortcut to AMP->caps */
	struct amp_state *state;	/* shortcut to AMP->state */
	int error_status;
	int do_exception;
} Amp;


SWIGINTERN struct Amp *new_Amp(amp_model_t amp_model){
		Amp *r;

		r = (Amp*)malloc(sizeof(Amp));
		if (!r)
			return NULL;
		r->amp = amp_init(amp_model);
		if (!r->amp) {
			free(r);
			return NULL;
		}
		/* install shortcuts */
		r->caps = r->amp->caps;
		r->state = amp_data_pointer(r->amp, RIG_PTRX_AMPSTATE);
		r->do_exception = 0;    /* default is disabled */
		r->error_status = RIG_OK;
		return r;
	}
SWIGINTERN void delete_Amp(struct Amp *self){
		amp_cleanup(self->amp);
		free(self);
	}
SWIGINTERN void Amp_open(struct Amp *self){ self->error_status = amp_open(self->amp); }
SWIGINTERN void Amp_close(struct Amp *self){ self->error_status = amp_close(self->amp); }
SWIGINTERN void Amp_reset(struct Amp *self,amp_reset_t _amp_reset_t){ self->error_status = amp_reset(self->amp, _amp_reset_t); }
SWIGINTERN void Amp_token_lookup(struct Amp *self,const_char_string _const_char_string){ self->error_status = amp_token_lookup(self->amp, _const_char_string); }
SWIGINTERN setting_t Amp_has_get_level(struct Amp *self,setting_t level){
		return amp_has_get_level(self->amp, level);
	}
SWIGINTERN setting_t Amp_has_set_level(struct Amp *self,setting_t level){
		return amp_has_set_level(self->amp, level);
	}
SWIGINTERN void Amp_set_freq(struct Amp *self,freq_t _freq_t){ self->error_status = amp_set_freq(self->amp, _freq_t); }
SWIGINTERN void Amp_set_powerstat(struct Amp *self,powerstat_t _powerstat_t){ self->error_status = amp_set_powerstat(self->amp, _powerstat_t); }
SWIGINTERN void Amp_set_level(struct Amp *self,setting_t _setting_t_1,value_t _value_t_2){ self->error_status = amp_set_level(self->amp, _setting_t_1, _value_t_2); }
SWIGINTERN void Amp_set_ext_level(struct Amp *self,hamlib_token_t _hamlib_token_t_1,value_t _value_t_2){ self->error_status = amp_set_ext_level(self->amp, _hamlib_token_t_1, _value_t_2); }
SWIGINTERN void Amp_set_conf__SWIG_0(struct Amp *self,char const *name,char const *val){
		hamlib_token_t tok = amp_token_lookup(self->amp, name);
		if (tok == 0)
			self->error_status = -RIG_EINVAL;
		else
			self->error_status = amp_set_conf(self->amp, tok, val);
	}
SWIGINTERN void Amp_set_conf__SWIG_1(struct Amp *self,hamlib_token_t _hamlib_token_t_1,const_char_string _const_char_string_2){ self->error_status = amp_set_conf(self->amp, _hamlib_token_t_1, _const_char_string_2); }
SWIGINTERN char const *Amp_get_conf__SWIG_0(struct Amp *self,hamlib_token_t tok){
          static char s[128] = "";
          self->error_status = amp_get_conf2(self->amp, tok, s, sizeof s);
          return s;
  }
SWIGINTERN char const *Amp_get_conf__SWIG_1(struct Amp *self,char const *name){
          hamlib_token_t tok = amp_token_lookup(self->amp, name);
          static char s[128] = "";
          if (tok == 0)
                  self->error_status = -RIG_EINVAL;
          else
                  self->error_status = amp_get_conf2(self->amp, tok, s, sizeof s);
          return s;
  }
SWIGINTERN freq_t Amp_get_freq(struct Amp *self){ freq_t _freq_t; self->error_status = amp_get_freq(self->amp, &_freq_t); return _freq_t; }
SWIGINTERN char const *Amp_get_info(struct Amp *self){
          const char *s;
          s = amp_get_info(self->amp);
          self->error_status = s ? RIG_OK : -RIG_EINVAL;
          return s;
  }
SWIGINTERN powerstat_t Amp_get_powerstat(struct Amp *self){ powerstat_t _powerstat_t; self->error_status = amp_get_powerstat(self->amp, &_powerstat_t); return _powerstat_t; }
#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN const char *_wrap_cookie_use_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, char *name1, char *name2, int flags) {
  Tcl_Obj *value = 0;
  
  value = SWIG_From_int((int)(cookie_use));
  if (value) {
    Tcl_SetVar2(interp,name1,name2,Tcl_GetString(value), flags);
    Tcl_DecrRefCount(value);
  }
  return NULL;
}


SWIGINTERN const char *_wrap_cookie_use_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, char *name1, char *name2 SWIGUNUSED, int flags) {
  Tcl_Obj *value = 0;
  Tcl_Obj *name1o = 0;
  
  name1o = Tcl_NewStringObj(name1,-1);
  value = Tcl_ObjGetVar2(interp, name1o, 0, flags);
  Tcl_DecrRefCount(name1o);
  if (!value) SWIG_fail;
  {
    int val;
    int res = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(value, &val);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in variable '""cookie_use""' of type '""int""'");
    }
    cookie_use = (int)(val);
  }
  return NULL;
fail:
  return "cookie_use";
}


SWIGINTERN const char *_wrap_skip_init_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, char *name1, char *name2, int flags) {
  Tcl_Obj *value = 0;
  
  value = SWIG_From_int((int)(skip_init));
  if (value) {
    Tcl_SetVar2(interp,name1,name2,Tcl_GetString(value), flags);
    Tcl_DecrRefCount(value);
  }
  return NULL;
}


SWIGINTERN const char *_wrap_skip_init_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, char *name1, char *name2 SWIGUNUSED, int flags) {
  Tcl_Obj *value = 0;
  Tcl_Obj *name1o = 0;
  
  name1o = Tcl_NewStringObj(name1,-1);
  value = Tcl_ObjGetVar2(interp, name1o, 0, flags);
  Tcl_DecrRefCount(name1o);
  if (!value) SWIG_fail;
  {
    int val;
    int res = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(value, &val);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in variable '""skip_init""' of type '""int""'");
    }
    skip_init = (int)(val);
  }
  return NULL;
fail:
  return "skip_init";
}


SWIGINTERN const char *_wrap_hamlib_version_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, char *name1, char *name2, int flags) {
  Tcl_Obj *value = 0;
  
  value = SWIG_FromCharPtr(hamlib_version);
  if (value) {
    Tcl_SetVar2(interp,name1,name2,Tcl_GetString(value), flags);
    Tcl_DecrRefCount(value);
  }
  return NULL;
}


SWIGINTERN const char *swig_readonly(ClientData clientData SWIGUNUSED, Tcl_Interp *interp SWIGUNUSED, char *name1 SWIGUNUSED, char *name2 SWIGUNUSED, int flags SWIGUNUSED) {
  return "Variable is read-only";
}


SWIGINTERN const char *_wrap_hamlib_copyright_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, char *name1, char *name2, int flags) {
  Tcl_Obj *value = 0;
  
  value = SWIG_FromCharPtr(hamlib_copyright);
  if (value) {
    Tcl_SetVar2(interp,name1,name2,Tcl_GetString(value), flags);
    Tcl_DecrRefCount(value);
  }
  return NULL;
}


SWIGINTERN int
_wrap_confparams_token_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct confparams *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:confparams_token_set self token ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_token_set" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "confparams_token_set" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  if (arg1) (arg1)->token = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_token_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  hamlib_token_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_token_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_token_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (hamlib_token_t) ((arg1)->token);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_name_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_name_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_name_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (char *) ((arg1)->name);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_label_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_label_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_label_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (char *) ((arg1)->label);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_tooltip_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_tooltip_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_tooltip_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (char *) ((arg1)->tooltip);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_dflt_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_dflt_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_dflt_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (char *) ((arg1)->dflt);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct confparams *arg1 = 0 ;
  enum rig_conf_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:confparams_type_set self type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_type_set" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "confparams_type_set" "', argument " "2"" of type '" "enum rig_conf_e""'");
  } 
  arg2 = (enum rig_conf_e)(val2);
  if (arg1) (arg1)->type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum rig_conf_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_type_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (enum rig_conf_e) ((arg1)->type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_u_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  confparams_u *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_u_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_get" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  result = (confparams_u *)& ((arg1)->u);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams_u,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_confparams(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct confparams *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_confparams ") == TCL_ERROR) SWIG_fail;
  result = (struct confparams *)calloc(1, sizeof(struct confparams));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_confparams(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct confparams *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_confparams self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_confparams" "', argument " "1"" of type '" "struct confparams *""'"); 
  }
  arg1 = (struct confparams *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_confparams(void *obj) {
struct confparams *arg1 = (struct confparams *) obj;
free((char *) arg1);
}
static swig_method swig_confparams_methods[] = {
    {0,0}
};
static swig_attribute swig_confparams_attributes[] = {
    { "-token",_wrap_confparams_token_get, _wrap_confparams_token_set},
    { "-name",_wrap_confparams_name_get, 0 },
    { "-label",_wrap_confparams_label_get, 0 },
    { "-tooltip",_wrap_confparams_tooltip_get, 0 },
    { "-dflt",_wrap_confparams_dflt_get, 0 },
    { "-type",_wrap_confparams_type_get, _wrap_confparams_type_set},
    { "-u",_wrap_confparams_u_get, 0 },
    {0,0,0}
};
static swig_class *swig_confparams_bases[] = {0};
static const char * swig_confparams_base_names[] = {0};
static swig_class _wrap_class_confparams = { "confparams", &SWIGTYPE_p_confparams,_wrap_new_confparams, swig_delete_confparams, swig_confparams_methods, swig_confparams_attributes, swig_confparams_bases,swig_confparams_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_confparams_u_n_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  confparams_u_n *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_u_n_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_get" "', argument " "1"" of type '" "confparams_u *""'"); 
  }
  arg1 = (confparams_u *)(argp1);
  result = (confparams_u_n *)& ((arg1)->n);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams_u_n,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_u_c_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  confparams_u_c *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_u_c_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_c_get" "', argument " "1"" of type '" "confparams_u *""'"); 
  }
  arg1 = (confparams_u *)(argp1);
  result = (confparams_u_c *)& ((arg1)->c);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams_u_c,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_confparams_u(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_confparams_u ") == TCL_ERROR) SWIG_fail;
  result = (confparams_u *)calloc(1, sizeof(confparams_u));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams_u,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_confparams_u(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_confparams_u self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_confparams_u" "', argument " "1"" of type '" "confparams_u *""'"); 
  }
  arg1 = (confparams_u *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_confparams_u(void *obj) {
confparams_u *arg1 = (confparams_u *) obj;
free((char *) arg1);
}
static swig_method swig_confparams_u_methods[] = {
    {0,0}
};
static swig_attribute swig_confparams_u_attributes[] = {
    { "-n",_wrap_confparams_u_n_get, 0 },
    { "-c",_wrap_confparams_u_c_get, 0 },
    {0,0,0}
};
static swig_class *swig_confparams_u_bases[] = {0};
static const char * swig_confparams_u_base_names[] = {0};
static swig_class _wrap_class_confparams_u = { "confparams_u", &SWIGTYPE_p_confparams_u,_wrap_new_confparams_u, swig_delete_confparams_u, swig_confparams_u_methods, swig_confparams_u_attributes, swig_confparams_u_bases,swig_confparams_u_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_confparams_u_c_combostr_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u_c *arg1 = 0 ;
  char **arg2 = (char **)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:confparams_u_c_combostr_set self combostr ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u_c, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_c_combostr_set" "', argument " "1"" of type '" "confparams_u_c *""'"); 
  }
  arg1 = (confparams_u_c *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "confparams_u_c_combostr_set" "', argument " "2"" of type '" "char const *[16]""'"); 
  } 
  arg2 = (char **)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)16; ++ii) *(char const * *)&arg1->combostr[ii] = *((char const * *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""combostr""' of type '""char const *[16]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_u_c_combostr_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u_c *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char **result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_u_c_combostr_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u_c, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_c_combostr_get" "', argument " "1"" of type '" "confparams_u_c *""'"); 
  }
  arg1 = (confparams_u_c *)(argp1);
  result = (char **)(char **) ((arg1)->combostr);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_p_char,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_confparams_u_c(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u_c *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_confparams_u_c ") == TCL_ERROR) SWIG_fail;
  result = (confparams_u_c *)calloc(1, sizeof(confparams_u_c));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams_u_c,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_confparams_u_c(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u_c *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_confparams_u_c self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u_c, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_confparams_u_c" "', argument " "1"" of type '" "confparams_u_c *""'"); 
  }
  arg1 = (confparams_u_c *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_confparams_u_c(void *obj) {
confparams_u_c *arg1 = (confparams_u_c *) obj;
free((char *) arg1);
}
static swig_method swig_confparams_u_c_methods[] = {
    {0,0}
};
static swig_attribute swig_confparams_u_c_attributes[] = {
    { "-combostr",_wrap_confparams_u_c_combostr_get, _wrap_confparams_u_c_combostr_set},
    {0,0,0}
};
static swig_class *swig_confparams_u_c_bases[] = {0};
static const char * swig_confparams_u_c_base_names[] = {0};
static swig_class _wrap_class_confparams_u_c = { "confparams_u_c", &SWIGTYPE_p_confparams_u_c,_wrap_new_confparams_u_c, swig_delete_confparams_u_c, swig_confparams_u_c_methods, swig_confparams_u_c_attributes, swig_confparams_u_c_bases,swig_confparams_u_c_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_confparams_u_n_min_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u_n *arg1 = 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:confparams_u_n_min_set self min ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u_n, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_min_set" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "confparams_u_n_min_set" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = (float)(val2);
  if (arg1) (arg1)->min = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_u_n_min_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u_n *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_u_n_min_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u_n, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_min_get" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  result = (float) ((arg1)->min);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_u_n_max_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u_n *arg1 = 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:confparams_u_n_max_set self max ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u_n, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_max_set" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "confparams_u_n_max_set" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = (float)(val2);
  if (arg1) (arg1)->max = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_u_n_max_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u_n *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_u_n_max_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u_n, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_max_get" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  result = (float) ((arg1)->max);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_u_n_step_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u_n *arg1 = 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:confparams_u_n_step_set self step ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u_n, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_step_set" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "confparams_u_n_step_set" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = (float)(val2);
  if (arg1) (arg1)->step = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_confparams_u_n_step_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u_n *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:confparams_u_n_step_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u_n, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "confparams_u_n_step_get" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  result = (float) ((arg1)->step);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_confparams_u_n(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u_n *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_confparams_u_n ") == TCL_ERROR) SWIG_fail;
  result = (confparams_u_n *)calloc(1, sizeof(confparams_u_n));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams_u_n,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_confparams_u_n(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  confparams_u_n *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_confparams_u_n self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_confparams_u_n, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_confparams_u_n" "', argument " "1"" of type '" "confparams_u_n *""'"); 
  }
  arg1 = (confparams_u_n *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_confparams_u_n(void *obj) {
confparams_u_n *arg1 = (confparams_u_n *) obj;
free((char *) arg1);
}
static swig_method swig_confparams_u_n_methods[] = {
    {0,0}
};
static swig_attribute swig_confparams_u_n_attributes[] = {
    { "-min",_wrap_confparams_u_n_min_get, _wrap_confparams_u_n_min_set},
    { "-max",_wrap_confparams_u_n_max_get, _wrap_confparams_u_n_max_set},
    { "-step",_wrap_confparams_u_n_step_get, _wrap_confparams_u_n_step_set},
    {0,0,0}
};
static swig_class *swig_confparams_u_n_bases[] = {0};
static const char * swig_confparams_u_n_base_names[] = {0};
static swig_class _wrap_class_confparams_u_n = { "confparams_u_n", &SWIGTYPE_p_confparams_u_n,_wrap_new_confparams_u_n, swig_delete_confparams_u_n, swig_confparams_u_n_methods, swig_confparams_u_n_attributes, swig_confparams_u_n_bases,swig_confparams_u_n_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_value_t_i_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:value_t_i_set self i ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_i_set" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "value_t_i_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->i = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_value_t_i_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:value_t_i_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_i_get" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  result = (int) ((arg1)->i);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_value_t_u_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:value_t_u_set self u ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_u_set" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "value_t_u_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->u = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_value_t_u_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:value_t_u_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_u_get" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  result = (unsigned int) ((arg1)->u);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_value_t_f_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t *arg1 = 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:value_t_f_set self f ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_f_set" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "value_t_f_set" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = (float)(val2);
  if (arg1) (arg1)->f = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_value_t_f_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:value_t_f_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_f_get" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  result = (float) ((arg1)->f);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_value_t_s_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:value_t_s_set self s ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_s_set" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "value_t_s_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->s);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->s = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->s = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_value_t_s_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:value_t_s_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_s_get" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  result = (char *) ((arg1)->s);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_value_t_cs_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:value_t_cs_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_cs_get" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  result = (char *) ((arg1)->cs);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_value_t_b_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  value_t_b *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:value_t_b_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_b_get" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  result = (value_t_b *)& ((arg1)->b);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_value_t_b,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_value_t(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_value_t ") == TCL_ERROR) SWIG_fail;
  result = (value_t *)calloc(1, sizeof(value_t));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_value_t,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_value_t(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_value_t self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_value_t" "', argument " "1"" of type '" "value_t *""'"); 
  }
  arg1 = (value_t *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_value_t(void *obj) {
value_t *arg1 = (value_t *) obj;
free((char *) arg1);
}
static swig_method swig_value_t_methods[] = {
    {0,0}
};
static swig_attribute swig_value_t_attributes[] = {
    { "-i",_wrap_value_t_i_get, _wrap_value_t_i_set},
    { "-u",_wrap_value_t_u_get, _wrap_value_t_u_set},
    { "-f",_wrap_value_t_f_get, _wrap_value_t_f_set},
    { "-s",_wrap_value_t_s_get, _wrap_value_t_s_set},
    { "-cs",_wrap_value_t_cs_get, 0 },
    { "-b",_wrap_value_t_b_get, 0 },
    {0,0,0}
};
static swig_class *swig_value_t_bases[] = {0};
static const char * swig_value_t_base_names[] = {0};
static swig_class _wrap_class_value_t = { "value_t", &SWIGTYPE_p_value_t,_wrap_new_value_t, swig_delete_value_t, swig_value_t_methods, swig_value_t_attributes, swig_value_t_bases,swig_value_t_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_value_t_b_l_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t_b *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:value_t_b_l_set self l ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t_b, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_b_l_set" "', argument " "1"" of type '" "value_t_b *""'"); 
  }
  arg1 = (value_t_b *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "value_t_b_l_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->l = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_value_t_b_l_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t_b *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:value_t_b_l_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t_b, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_b_l_get" "', argument " "1"" of type '" "value_t_b *""'"); 
  }
  arg1 = (value_t_b *)(argp1);
  result = (int) ((arg1)->l);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_value_t_b_d_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t_b *arg1 = 0 ;
  unsigned char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:value_t_b_d_set self d ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t_b, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_b_d_set" "', argument " "1"" of type '" "value_t_b *""'"); 
  }
  arg1 = (value_t_b *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_unsigned_char, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "value_t_b_d_set" "', argument " "2"" of type '" "unsigned char *""'"); 
  }
  arg2 = (unsigned char *)(argp2);
  if (arg1) (arg1)->d = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_value_t_b_d_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t_b *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:value_t_b_d_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t_b, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "value_t_b_d_get" "', argument " "1"" of type '" "value_t_b *""'"); 
  }
  arg1 = (value_t_b *)(argp1);
  result = (unsigned char *) ((arg1)->d);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_value_t_b(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t_b *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_value_t_b ") == TCL_ERROR) SWIG_fail;
  result = (value_t_b *)calloc(1, sizeof(value_t_b));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_value_t_b,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_value_t_b(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  value_t_b *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_value_t_b self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_value_t_b, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_value_t_b" "', argument " "1"" of type '" "value_t_b *""'"); 
  }
  arg1 = (value_t_b *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_value_t_b(void *obj) {
value_t_b *arg1 = (value_t_b *) obj;
free((char *) arg1);
}
static swig_method swig_value_t_b_methods[] = {
    {0,0}
};
static swig_attribute swig_value_t_b_attributes[] = {
    { "-l",_wrap_value_t_b_l_get, _wrap_value_t_b_l_set},
    { "-d",_wrap_value_t_b_d_get, _wrap_value_t_b_d_set},
    {0,0,0}
};
static swig_class *swig_value_t_b_bases[] = {0};
static const char * swig_value_t_b_base_names[] = {0};
static swig_class _wrap_class_value_t_b = { "value_t_b", &SWIGTYPE_p_value_t_b,_wrap_new_value_t_b, swig_delete_value_t_b, swig_value_t_b_methods, swig_value_t_b_attributes, swig_value_t_b_bases,swig_value_t_b_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_freq_range_t_startf_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:freq_range_t_startf_set self startf ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_startf_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_startf_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->startf = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_startf_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:freq_range_t_startf_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_startf_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (freq_t) ((arg1)->startf);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_endf_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:freq_range_t_endf_set self endf ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_endf_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_endf_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->endf = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_endf_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:freq_range_t_endf_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_endf_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (freq_t) ((arg1)->endf);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_modes_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:freq_range_t_modes_set self modes ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_modes_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_modes_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->modes = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_modes_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rmode_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:freq_range_t_modes_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_modes_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (rmode_t) ((arg1)->modes);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_low_power_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:freq_range_t_low_power_set self low_power ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_low_power_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_low_power_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->low_power = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_low_power_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:freq_range_t_low_power_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_low_power_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (int) ((arg1)->low_power);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_high_power_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:freq_range_t_high_power_set self high_power ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_high_power_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_high_power_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->high_power = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_high_power_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:freq_range_t_high_power_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_high_power_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (int) ((arg1)->high_power);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:freq_range_t_vfo_set self vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_vfo_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_vfo_set" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  if (arg1) (arg1)->vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  vfo_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:freq_range_t_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_vfo_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (vfo_t) ((arg1)->vfo);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_ant_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  ant_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:freq_range_t_ant_set self ant ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_ant_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "freq_range_t_ant_set" "', argument " "2"" of type '" "ant_t""'");
  } 
  arg2 = (ant_t)(val2);
  if (arg1) (arg1)->ant = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_ant_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ant_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:freq_range_t_ant_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_ant_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (ant_t) ((arg1)->ant);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_label_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:freq_range_t_label_set self label ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_label_set" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "freq_range_t_label_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->label);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->label = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->label = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_freq_range_t_label_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:freq_range_t_label_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "freq_range_t_label_get" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  result = (char *) ((arg1)->label);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_freq_range_t(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_freq_range_t ") == TCL_ERROR) SWIG_fail;
  result = (struct freq_range_list *)calloc(1, sizeof(struct freq_range_list));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_freq_range_t(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct freq_range_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_freq_range_t self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_freq_range_list, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_freq_range_t" "', argument " "1"" of type '" "struct freq_range_list *""'"); 
  }
  arg1 = (struct freq_range_list *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_freq_range_t(void *obj) {
struct freq_range_list *arg1 = (struct freq_range_list *) obj;
free((char *) arg1);
}
static swig_method swig_freq_range_list_methods[] = {
    {0,0}
};
static swig_attribute swig_freq_range_list_attributes[] = {
    { "-startf",_wrap_freq_range_t_startf_get, _wrap_freq_range_t_startf_set},
    { "-endf",_wrap_freq_range_t_endf_get, _wrap_freq_range_t_endf_set},
    { "-modes",_wrap_freq_range_t_modes_get, _wrap_freq_range_t_modes_set},
    { "-low_power",_wrap_freq_range_t_low_power_get, _wrap_freq_range_t_low_power_set},
    { "-high_power",_wrap_freq_range_t_high_power_get, _wrap_freq_range_t_high_power_set},
    { "-vfo",_wrap_freq_range_t_vfo_get, _wrap_freq_range_t_vfo_set},
    { "-ant",_wrap_freq_range_t_ant_get, _wrap_freq_range_t_ant_set},
    { "-label",_wrap_freq_range_t_label_get, _wrap_freq_range_t_label_set},
    {0,0,0}
};
static swig_class *swig_freq_range_list_bases[] = {0};
static const char * swig_freq_range_list_base_names[] = {0};
static swig_class _wrap_class_freq_range_list = { "freq_range_t", &SWIGTYPE_p_freq_range_list,_wrap_new_freq_range_t, swig_delete_freq_range_t, swig_freq_range_list_methods, swig_freq_range_list_attributes, swig_freq_range_list_bases,swig_freq_range_list_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_tuning_step_list_modes_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct tuning_step_list *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:tuning_step_list_modes_set self modes ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_tuning_step_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "tuning_step_list_modes_set" "', argument " "1"" of type '" "struct tuning_step_list *""'"); 
  }
  arg1 = (struct tuning_step_list *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "tuning_step_list_modes_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->modes = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_tuning_step_list_modes_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct tuning_step_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rmode_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:tuning_step_list_modes_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_tuning_step_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "tuning_step_list_modes_get" "', argument " "1"" of type '" "struct tuning_step_list *""'"); 
  }
  arg1 = (struct tuning_step_list *)(argp1);
  result = (rmode_t) ((arg1)->modes);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_tuning_step_list_ts_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct tuning_step_list *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:tuning_step_list_ts_set self ts ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_tuning_step_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "tuning_step_list_ts_set" "', argument " "1"" of type '" "struct tuning_step_list *""'"); 
  }
  arg1 = (struct tuning_step_list *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "tuning_step_list_ts_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->ts = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_tuning_step_list_ts_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct tuning_step_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:tuning_step_list_ts_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_tuning_step_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "tuning_step_list_ts_get" "', argument " "1"" of type '" "struct tuning_step_list *""'"); 
  }
  arg1 = (struct tuning_step_list *)(argp1);
  result = (shortfreq_t) ((arg1)->ts);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_tuning_step_list(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct tuning_step_list *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_tuning_step_list ") == TCL_ERROR) SWIG_fail;
  result = (struct tuning_step_list *)calloc(1, sizeof(struct tuning_step_list));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_tuning_step_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_tuning_step_list(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct tuning_step_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_tuning_step_list self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_tuning_step_list, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_tuning_step_list" "', argument " "1"" of type '" "struct tuning_step_list *""'"); 
  }
  arg1 = (struct tuning_step_list *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_tuning_step_list(void *obj) {
struct tuning_step_list *arg1 = (struct tuning_step_list *) obj;
free((char *) arg1);
}
static swig_method swig_tuning_step_list_methods[] = {
    {0,0}
};
static swig_attribute swig_tuning_step_list_attributes[] = {
    { "-modes",_wrap_tuning_step_list_modes_get, _wrap_tuning_step_list_modes_set},
    { "-ts",_wrap_tuning_step_list_ts_get, _wrap_tuning_step_list_ts_set},
    {0,0,0}
};
static swig_class *swig_tuning_step_list_bases[] = {0};
static const char * swig_tuning_step_list_base_names[] = {0};
static swig_class _wrap_class_tuning_step_list = { "tuning_step_list", &SWIGTYPE_p_tuning_step_list,_wrap_new_tuning_step_list, swig_delete_tuning_step_list, swig_tuning_step_list_methods, swig_tuning_step_list_attributes, swig_tuning_step_list_bases,swig_tuning_step_list_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_filter_list_modes_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct filter_list *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:filter_list_modes_set self modes ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_filter_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "filter_list_modes_set" "', argument " "1"" of type '" "struct filter_list *""'"); 
  }
  arg1 = (struct filter_list *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "filter_list_modes_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->modes = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_filter_list_modes_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct filter_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rmode_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:filter_list_modes_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_filter_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "filter_list_modes_get" "', argument " "1"" of type '" "struct filter_list *""'"); 
  }
  arg1 = (struct filter_list *)(argp1);
  result = (rmode_t) ((arg1)->modes);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_filter_list_width_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct filter_list *arg1 = 0 ;
  pbwidth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:filter_list_width_set self width ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_filter_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "filter_list_width_set" "', argument " "1"" of type '" "struct filter_list *""'"); 
  }
  arg1 = (struct filter_list *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "filter_list_width_set" "', argument " "2"" of type '" "pbwidth_t""'");
  } 
  arg2 = (pbwidth_t)(val2);
  if (arg1) (arg1)->width = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_filter_list_width_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct filter_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  pbwidth_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:filter_list_width_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_filter_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "filter_list_width_get" "', argument " "1"" of type '" "struct filter_list *""'"); 
  }
  arg1 = (struct filter_list *)(argp1);
  result = (pbwidth_t) ((arg1)->width);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_filter_list(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct filter_list *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_filter_list ") == TCL_ERROR) SWIG_fail;
  result = (struct filter_list *)calloc(1, sizeof(struct filter_list));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_filter_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_filter_list(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct filter_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_filter_list self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_filter_list, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_filter_list" "', argument " "1"" of type '" "struct filter_list *""'"); 
  }
  arg1 = (struct filter_list *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_filter_list(void *obj) {
struct filter_list *arg1 = (struct filter_list *) obj;
free((char *) arg1);
}
static swig_method swig_filter_list_methods[] = {
    {0,0}
};
static swig_attribute swig_filter_list_attributes[] = {
    { "-modes",_wrap_filter_list_modes_get, _wrap_filter_list_modes_set},
    { "-width",_wrap_filter_list_width_get, _wrap_filter_list_width_set},
    {0,0,0}
};
static swig_class *swig_filter_list_bases[] = {0};
static const char * swig_filter_list_base_names[] = {0};
static swig_class _wrap_class_filter_list = { "filter_list", &SWIGTYPE_p_filter_list,_wrap_new_filter_list, swig_delete_filter_list, swig_filter_list_methods, swig_filter_list_attributes, swig_filter_list_bases,swig_filter_list_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_ext_list_token_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct ext_list *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:ext_list_token_set self token ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_ext_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ext_list_token_set" "', argument " "1"" of type '" "struct ext_list *""'"); 
  }
  arg1 = (struct ext_list *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ext_list_token_set" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  if (arg1) (arg1)->token = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_ext_list_token_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct ext_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  hamlib_token_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:ext_list_token_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_ext_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ext_list_token_get" "', argument " "1"" of type '" "struct ext_list *""'"); 
  }
  arg1 = (struct ext_list *)(argp1);
  result = (hamlib_token_t) ((arg1)->token);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_ext_list_val_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct ext_list *arg1 = 0 ;
  value_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:ext_list_val_set self val ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_ext_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ext_list_val_set" "', argument " "1"" of type '" "struct ext_list *""'"); 
  }
  arg1 = (struct ext_list *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ext_list_val_set" "', argument " "2"" of type '" "value_t *""'"); 
  }
  arg2 = (value_t *)(argp2);
  if (arg1) (arg1)->val = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_ext_list_val_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct ext_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  value_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:ext_list_val_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_ext_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ext_list_val_get" "', argument " "1"" of type '" "struct ext_list *""'"); 
  }
  arg1 = (struct ext_list *)(argp1);
  result = (value_t *)& ((arg1)->val);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_value_t,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_ext_list(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct ext_list *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_ext_list ") == TCL_ERROR) SWIG_fail;
  result = (struct ext_list *)calloc(1, sizeof(struct ext_list));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_ext_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_ext_list(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct ext_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_ext_list self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_ext_list, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ext_list" "', argument " "1"" of type '" "struct ext_list *""'"); 
  }
  arg1 = (struct ext_list *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_ext_list(void *obj) {
struct ext_list *arg1 = (struct ext_list *) obj;
free((char *) arg1);
}
static swig_method swig_ext_list_methods[] = {
    {0,0}
};
static swig_attribute swig_ext_list_attributes[] = {
    { "-token",_wrap_ext_list_token_get, _wrap_ext_list_token_set},
    { "-val",_wrap_ext_list_val_get, _wrap_ext_list_val_set},
    {0,0,0}
};
static swig_class *swig_ext_list_bases[] = {0};
static const char * swig_ext_list_base_names[] = {0};
static swig_class _wrap_class_ext_list = { "ext_list", &SWIGTYPE_p_ext_list,_wrap_new_ext_list, swig_delete_ext_list, swig_ext_list_methods, swig_ext_list_attributes, swig_ext_list_bases,swig_ext_list_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_channel_channel_num_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_channel_num_set self channel_num ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_channel_num_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_channel_num_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->channel_num = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_channel_num_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_channel_num_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_channel_num_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (int) ((arg1)->channel_num);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_bank_num_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_bank_num_set self bank_num ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_bank_num_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_bank_num_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->bank_num = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_bank_num_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_bank_num_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_bank_num_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (int) ((arg1)->bank_num);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_vfo_set self vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_vfo_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_vfo_set" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  if (arg1) (arg1)->vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  vfo_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_vfo_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (vfo_t) ((arg1)->vfo);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_ant_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  ant_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_ant_set self ant ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ant_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_ant_set" "', argument " "2"" of type '" "ant_t""'");
  } 
  arg2 = (ant_t)(val2);
  if (arg1) (arg1)->ant = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_ant_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ant_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_ant_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ant_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (ant_t) ((arg1)->ant);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_freq_set self freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_freq_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_freq_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (freq_t) ((arg1)->freq);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_mode_set self mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_mode_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_mode_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rmode_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_mode_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (rmode_t) ((arg1)->mode);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_width_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  pbwidth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_width_set self width ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_width_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_width_set" "', argument " "2"" of type '" "pbwidth_t""'");
  } 
  arg2 = (pbwidth_t)(val2);
  if (arg1) (arg1)->width = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_width_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  pbwidth_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_width_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_width_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (pbwidth_t) ((arg1)->width);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_tx_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_tx_freq_set self tx_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_freq_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_tx_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->tx_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_tx_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_tx_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_freq_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (freq_t) ((arg1)->tx_freq);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_tx_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_tx_mode_set self tx_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_mode_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_tx_mode_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->tx_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_tx_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rmode_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_tx_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_mode_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (rmode_t) ((arg1)->tx_mode);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_tx_width_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  pbwidth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_tx_width_set self tx_width ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_width_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_tx_width_set" "', argument " "2"" of type '" "pbwidth_t""'");
  } 
  arg2 = (pbwidth_t)(val2);
  if (arg1) (arg1)->tx_width = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_tx_width_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  pbwidth_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_tx_width_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_width_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (pbwidth_t) ((arg1)->tx_width);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_split_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  split_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_split_set self split ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_split_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_split_set" "', argument " "2"" of type '" "split_t""'");
  } 
  arg2 = (split_t)(val2);
  if (arg1) (arg1)->split = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_split_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  split_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_split_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_split_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (split_t) ((arg1)->split);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_tx_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_tx_vfo_set self tx_vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_vfo_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_tx_vfo_set" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  if (arg1) (arg1)->tx_vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_tx_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  vfo_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_tx_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tx_vfo_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (vfo_t) ((arg1)->tx_vfo);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_rptr_shift_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  rptr_shift_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_rptr_shift_set self rptr_shift ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_rptr_shift_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_rptr_shift_set" "', argument " "2"" of type '" "rptr_shift_t""'");
  } 
  arg2 = (rptr_shift_t)(val2);
  if (arg1) (arg1)->rptr_shift = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_rptr_shift_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rptr_shift_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_rptr_shift_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_rptr_shift_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (rptr_shift_t) ((arg1)->rptr_shift);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_rptr_offs_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_rptr_offs_set self rptr_offs ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_rptr_offs_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_rptr_offs_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->rptr_offs = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_rptr_offs_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_rptr_offs_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_rptr_offs_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (shortfreq_t) ((arg1)->rptr_offs);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_tuning_step_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_tuning_step_set self tuning_step ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tuning_step_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_tuning_step_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->tuning_step = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_tuning_step_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_tuning_step_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tuning_step_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (shortfreq_t) ((arg1)->tuning_step);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_rit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_rit_set self rit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_rit_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_rit_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->rit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_rit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_rit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_rit_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (shortfreq_t) ((arg1)->rit);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_xit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_xit_set self xit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_xit_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_xit_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->xit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_xit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_xit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_xit_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (shortfreq_t) ((arg1)->xit);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_funcs_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_funcs_set self funcs ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_funcs_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_funcs_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->funcs = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_funcs_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_funcs_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_funcs_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (setting_t) ((arg1)->funcs);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_levels_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  value_t *arg2 = (value_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_levels_set self levels ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_levels_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "channel_levels_set" "', argument " "2"" of type '" "value_t [64]""'"); 
  } 
  arg2 = (value_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(value_t *)&arg1->levels[ii] = *((value_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""levels""' of type '""value_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_levels_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  value_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_levels_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_levels_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (value_t *)(value_t *) ((arg1)->levels);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_value_t,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_ctcss_tone_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  tone_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_ctcss_tone_set self ctcss_tone ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ctcss_tone_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_ctcss_tone_set" "', argument " "2"" of type '" "tone_t""'");
  } 
  arg2 = (tone_t)(val2);
  if (arg1) (arg1)->ctcss_tone = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_ctcss_tone_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  tone_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_ctcss_tone_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ctcss_tone_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (tone_t) ((arg1)->ctcss_tone);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_ctcss_sql_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  tone_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_ctcss_sql_set self ctcss_sql ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ctcss_sql_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_ctcss_sql_set" "', argument " "2"" of type '" "tone_t""'");
  } 
  arg2 = (tone_t)(val2);
  if (arg1) (arg1)->ctcss_sql = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_ctcss_sql_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  tone_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_ctcss_sql_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ctcss_sql_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (tone_t) ((arg1)->ctcss_sql);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_dcs_code_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  tone_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_dcs_code_set self dcs_code ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_dcs_code_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_dcs_code_set" "', argument " "2"" of type '" "tone_t""'");
  } 
  arg2 = (tone_t)(val2);
  if (arg1) (arg1)->dcs_code = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_dcs_code_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  tone_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_dcs_code_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_dcs_code_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (tone_t) ((arg1)->dcs_code);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_dcs_sql_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  tone_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_dcs_sql_set self dcs_sql ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_dcs_sql_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_dcs_sql_set" "', argument " "2"" of type '" "tone_t""'");
  } 
  arg2 = (tone_t)(val2);
  if (arg1) (arg1)->dcs_sql = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_dcs_sql_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  tone_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_dcs_sql_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_dcs_sql_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (tone_t) ((arg1)->dcs_sql);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_scan_group_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_scan_group_set self scan_group ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_scan_group_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_scan_group_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->scan_group = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_scan_group_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_scan_group_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_scan_group_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (int) ((arg1)->scan_group);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_flags_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_flags_set self flags ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_flags_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_flags_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->flags = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_flags_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_flags_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_flags_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (unsigned int) ((arg1)->flags);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_channel_desc_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  char *arg2 = (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char temp2[30] ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_channel_desc_set self channel_desc ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_channel_desc_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  res2 = SWIG_AsCharArray(objv[2], temp2, 30);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "channel_channel_desc_set" "', argument " "2"" of type '" "char [30]""'");
  }
  arg2 = (char *)(temp2);
  if (arg2) memcpy(arg1->channel_desc,arg2,30*sizeof(char));
  else memset(arg1->channel_desc,0,30*sizeof(char));
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_channel_desc_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_channel_desc_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_channel_desc_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (char *)(char *) ((arg1)->channel_desc);
  {
    size_t size = SWIG_strnlen(result, 30);
    
    
    
    Tcl_SetObjResult(interp,SWIG_FromCharPtrAndSize(result, size));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_ext_levels_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  struct ext_list *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_ext_levels_set self ext_levels ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ext_levels_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_ext_list, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "channel_ext_levels_set" "', argument " "2"" of type '" "struct ext_list *""'"); 
  }
  arg2 = (struct ext_list *)(argp2);
  if (arg1) (arg1)->ext_levels = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_ext_levels_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct ext_list *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_ext_levels_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_ext_levels_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (struct ext_list *) ((arg1)->ext_levels);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_ext_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_tag_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  char *arg2 = (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char temp2[32] ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_tag_set self tag ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tag_set" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  res2 = SWIG_AsCharArray(objv[2], temp2, 32);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "channel_tag_set" "', argument " "2"" of type '" "char [32]""'");
  }
  arg2 = (char *)(temp2);
  if (arg2) memcpy(arg1->tag,arg2,32*sizeof(char));
  else memset(arg1->tag,0,32*sizeof(char));
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_tag_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_tag_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_tag_get" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (char *)(char *) ((arg1)->tag);
  {
    size_t size = SWIG_strnlen(result, 32);
    
    
    
    Tcl_SetObjResult(interp,SWIG_FromCharPtrAndSize(result, size));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_channel(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  vfo_t arg1 = (1u << (29)) ;
  int arg2 = 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  struct channel *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"|oo:new_channel ?vfo? ?ch? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  if (objc > 1) {
    ecode1 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_channel" "', argument " "1"" of type '" "vfo_t""'");
    } 
    arg1 = (vfo_t)(val1);
  }
  if (objc > 2) {
    ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_channel" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = (struct channel *)new_channel(arg1,arg2);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_channel,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_channel(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_channel self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_channel" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  delete_channel(arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_channel(void *obj) {
struct channel *arg1 = (struct channel *) obj;
delete_channel(arg1);
}
static swig_method swig_channel_methods[] = {
    {0,0}
};
static swig_attribute swig_channel_attributes[] = {
    { "-channel_num",_wrap_channel_channel_num_get, _wrap_channel_channel_num_set},
    { "-bank_num",_wrap_channel_bank_num_get, _wrap_channel_bank_num_set},
    { "-vfo",_wrap_channel_vfo_get, _wrap_channel_vfo_set},
    { "-ant",_wrap_channel_ant_get, _wrap_channel_ant_set},
    { "-freq",_wrap_channel_freq_get, _wrap_channel_freq_set},
    { "-mode",_wrap_channel_mode_get, _wrap_channel_mode_set},
    { "-width",_wrap_channel_width_get, _wrap_channel_width_set},
    { "-tx_freq",_wrap_channel_tx_freq_get, _wrap_channel_tx_freq_set},
    { "-tx_mode",_wrap_channel_tx_mode_get, _wrap_channel_tx_mode_set},
    { "-tx_width",_wrap_channel_tx_width_get, _wrap_channel_tx_width_set},
    { "-split",_wrap_channel_split_get, _wrap_channel_split_set},
    { "-tx_vfo",_wrap_channel_tx_vfo_get, _wrap_channel_tx_vfo_set},
    { "-rptr_shift",_wrap_channel_rptr_shift_get, _wrap_channel_rptr_shift_set},
    { "-rptr_offs",_wrap_channel_rptr_offs_get, _wrap_channel_rptr_offs_set},
    { "-tuning_step",_wrap_channel_tuning_step_get, _wrap_channel_tuning_step_set},
    { "-rit",_wrap_channel_rit_get, _wrap_channel_rit_set},
    { "-xit",_wrap_channel_xit_get, _wrap_channel_xit_set},
    { "-funcs",_wrap_channel_funcs_get, _wrap_channel_funcs_set},
    { "-levels",_wrap_channel_levels_get, _wrap_channel_levels_set},
    { "-ctcss_tone",_wrap_channel_ctcss_tone_get, _wrap_channel_ctcss_tone_set},
    { "-ctcss_sql",_wrap_channel_ctcss_sql_get, _wrap_channel_ctcss_sql_set},
    { "-dcs_code",_wrap_channel_dcs_code_get, _wrap_channel_dcs_code_set},
    { "-dcs_sql",_wrap_channel_dcs_sql_get, _wrap_channel_dcs_sql_set},
    { "-scan_group",_wrap_channel_scan_group_get, _wrap_channel_scan_group_set},
    { "-flags",_wrap_channel_flags_get, _wrap_channel_flags_set},
    { "-channel_desc",_wrap_channel_channel_desc_get, _wrap_channel_channel_desc_set},
    { "-ext_levels",_wrap_channel_ext_levels_get, _wrap_channel_ext_levels_set},
    { "-tag",_wrap_channel_tag_get, _wrap_channel_tag_set},
    {0,0,0}
};
static swig_class *swig_channel_bases[] = {0};
static const char * swig_channel_base_names[] = {0};
static swig_class _wrap_class_channel = { "channel", &SWIGTYPE_p_channel,_wrap_new_channel, swig_delete_channel, swig_channel_methods, swig_channel_attributes, swig_channel_bases,swig_channel_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_channel_cap_bank_num_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_bank_num_set self bank_num ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_bank_num_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_bank_num_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->bank_num = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_bank_num_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_bank_num_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_bank_num_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->bank_num);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_vfo_set self vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_vfo_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_vfo_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_vfo_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->vfo);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_ant_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_ant_set self ant ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ant_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_ant_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->ant = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_ant_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_ant_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ant_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->ant);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_freq_set self freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_freq_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_freq_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_freq_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->freq);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_mode_set self mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_mode_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_mode_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_mode_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->mode);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_width_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_width_set self width ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_width_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_width_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->width = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_width_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_width_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_width_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->width);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_tx_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_tx_freq_set self tx_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_freq_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_tx_freq_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->tx_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_tx_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_tx_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_freq_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->tx_freq);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_tx_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_tx_mode_set self tx_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_mode_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_tx_mode_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->tx_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_tx_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_tx_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_mode_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->tx_mode);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_tx_width_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_tx_width_set self tx_width ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_width_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_tx_width_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->tx_width = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_tx_width_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_tx_width_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_width_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->tx_width);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_split_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_split_set self split ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_split_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_split_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->split = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_split_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_split_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_split_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->split);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_tx_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_tx_vfo_set self tx_vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_vfo_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_tx_vfo_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->tx_vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_tx_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_tx_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tx_vfo_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->tx_vfo);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_rptr_shift_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_rptr_shift_set self rptr_shift ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_rptr_shift_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_rptr_shift_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->rptr_shift = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_rptr_shift_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_rptr_shift_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_rptr_shift_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->rptr_shift);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_rptr_offs_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_rptr_offs_set self rptr_offs ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_rptr_offs_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_rptr_offs_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->rptr_offs = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_rptr_offs_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_rptr_offs_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_rptr_offs_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->rptr_offs);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_tuning_step_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_tuning_step_set self tuning_step ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tuning_step_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_tuning_step_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->tuning_step = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_tuning_step_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_tuning_step_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tuning_step_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->tuning_step);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_rit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_rit_set self rit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_rit_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_rit_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->rit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_rit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_rit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_rit_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->rit);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_xit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_xit_set self xit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_xit_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_xit_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->xit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_xit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_xit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_xit_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->xit);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_funcs_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_funcs_set self funcs ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_funcs_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_funcs_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->funcs = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_funcs_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_funcs_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_funcs_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (setting_t) ((arg1)->funcs);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_levels_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_levels_set self levels ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_levels_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_levels_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->levels = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_levels_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_levels_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_levels_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (setting_t) ((arg1)->levels);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_ctcss_tone_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_ctcss_tone_set self ctcss_tone ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ctcss_tone_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_ctcss_tone_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->ctcss_tone = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_ctcss_tone_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_ctcss_tone_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ctcss_tone_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->ctcss_tone);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_ctcss_sql_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_ctcss_sql_set self ctcss_sql ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ctcss_sql_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_ctcss_sql_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->ctcss_sql = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_ctcss_sql_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_ctcss_sql_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ctcss_sql_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->ctcss_sql);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_dcs_code_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_dcs_code_set self dcs_code ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_dcs_code_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_dcs_code_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->dcs_code = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_dcs_code_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_dcs_code_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_dcs_code_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->dcs_code);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_dcs_sql_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_dcs_sql_set self dcs_sql ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_dcs_sql_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_dcs_sql_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->dcs_sql = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_dcs_sql_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_dcs_sql_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_dcs_sql_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->dcs_sql);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_scan_group_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_scan_group_set self scan_group ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_scan_group_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_scan_group_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->scan_group = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_scan_group_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_scan_group_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_scan_group_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->scan_group);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_flags_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_flags_set self flags ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_flags_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_flags_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->flags = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_flags_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_flags_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_flags_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->flags);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_channel_desc_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_channel_desc_set self channel_desc ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_channel_desc_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_channel_desc_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->channel_desc = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_channel_desc_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_channel_desc_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_channel_desc_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->channel_desc);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_ext_levels_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_ext_levels_set self ext_levels ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ext_levels_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_ext_levels_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->ext_levels = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_ext_levels_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_ext_levels_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_ext_levels_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->ext_levels);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_tag_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channel_cap_tag_set self tag ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tag_set" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channel_cap_tag_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->tag = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channel_cap_tag_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channel_cap_tag_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channel_cap_tag_get" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  result = (unsigned int) ((arg1)->tag);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_channel_cap(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_channel_cap ") == TCL_ERROR) SWIG_fail;
  result = (struct channel_cap *)calloc(1, sizeof(struct channel_cap));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_channel_cap,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_channel_cap(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel_cap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_channel_cap self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel_cap, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_channel_cap" "', argument " "1"" of type '" "struct channel_cap *""'"); 
  }
  arg1 = (struct channel_cap *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_channel_cap(void *obj) {
struct channel_cap *arg1 = (struct channel_cap *) obj;
free((char *) arg1);
}
static swig_method swig_channel_cap_methods[] = {
    {0,0}
};
static swig_attribute swig_channel_cap_attributes[] = {
    { "-bank_num",_wrap_channel_cap_bank_num_get, _wrap_channel_cap_bank_num_set},
    { "-vfo",_wrap_channel_cap_vfo_get, _wrap_channel_cap_vfo_set},
    { "-ant",_wrap_channel_cap_ant_get, _wrap_channel_cap_ant_set},
    { "-freq",_wrap_channel_cap_freq_get, _wrap_channel_cap_freq_set},
    { "-mode",_wrap_channel_cap_mode_get, _wrap_channel_cap_mode_set},
    { "-width",_wrap_channel_cap_width_get, _wrap_channel_cap_width_set},
    { "-tx_freq",_wrap_channel_cap_tx_freq_get, _wrap_channel_cap_tx_freq_set},
    { "-tx_mode",_wrap_channel_cap_tx_mode_get, _wrap_channel_cap_tx_mode_set},
    { "-tx_width",_wrap_channel_cap_tx_width_get, _wrap_channel_cap_tx_width_set},
    { "-split",_wrap_channel_cap_split_get, _wrap_channel_cap_split_set},
    { "-tx_vfo",_wrap_channel_cap_tx_vfo_get, _wrap_channel_cap_tx_vfo_set},
    { "-rptr_shift",_wrap_channel_cap_rptr_shift_get, _wrap_channel_cap_rptr_shift_set},
    { "-rptr_offs",_wrap_channel_cap_rptr_offs_get, _wrap_channel_cap_rptr_offs_set},
    { "-tuning_step",_wrap_channel_cap_tuning_step_get, _wrap_channel_cap_tuning_step_set},
    { "-rit",_wrap_channel_cap_rit_get, _wrap_channel_cap_rit_set},
    { "-xit",_wrap_channel_cap_xit_get, _wrap_channel_cap_xit_set},
    { "-funcs",_wrap_channel_cap_funcs_get, _wrap_channel_cap_funcs_set},
    { "-levels",_wrap_channel_cap_levels_get, _wrap_channel_cap_levels_set},
    { "-ctcss_tone",_wrap_channel_cap_ctcss_tone_get, _wrap_channel_cap_ctcss_tone_set},
    { "-ctcss_sql",_wrap_channel_cap_ctcss_sql_get, _wrap_channel_cap_ctcss_sql_set},
    { "-dcs_code",_wrap_channel_cap_dcs_code_get, _wrap_channel_cap_dcs_code_set},
    { "-dcs_sql",_wrap_channel_cap_dcs_sql_get, _wrap_channel_cap_dcs_sql_set},
    { "-scan_group",_wrap_channel_cap_scan_group_get, _wrap_channel_cap_scan_group_set},
    { "-flags",_wrap_channel_cap_flags_get, _wrap_channel_cap_flags_set},
    { "-channel_desc",_wrap_channel_cap_channel_desc_get, _wrap_channel_cap_channel_desc_set},
    { "-ext_levels",_wrap_channel_cap_ext_levels_get, _wrap_channel_cap_ext_levels_set},
    { "-tag",_wrap_channel_cap_tag_get, _wrap_channel_cap_tag_set},
    {0,0,0}
};
static swig_class *swig_channel_cap_bases[] = {0};
static const char * swig_channel_cap_base_names[] = {0};
static swig_class _wrap_class_channel_cap = { "channel_cap", &SWIGTYPE_p_channel_cap,_wrap_new_channel_cap, swig_delete_channel_cap, swig_channel_cap_methods, swig_channel_cap_attributes, swig_channel_cap_bases,swig_channel_cap_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_chan_list_startc_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct chan_list *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:chan_list_startc_set self startc ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_startc_set" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "chan_list_startc_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->startc = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_chan_list_startc_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct chan_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:chan_list_startc_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_startc_get" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  result = (int) ((arg1)->startc);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_chan_list_endc_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct chan_list *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:chan_list_endc_set self endc ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_endc_set" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "chan_list_endc_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->endc = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_chan_list_endc_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct chan_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:chan_list_endc_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_endc_get" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  result = (int) ((arg1)->endc);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_chan_list_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct chan_list *arg1 = 0 ;
  chan_type_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:chan_list_type_set self type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_type_set" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "chan_list_type_set" "', argument " "2"" of type '" "chan_type_t""'");
  } 
  arg2 = (chan_type_t)(val2);
  if (arg1) (arg1)->type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_chan_list_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct chan_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  chan_type_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:chan_list_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_type_get" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  result = (chan_type_t) ((arg1)->type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_chan_list_mem_caps_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct chan_list *arg1 = 0 ;
  channel_cap_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:chan_list_mem_caps_set self mem_caps ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_mem_caps_set" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_channel_cap, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "chan_list_mem_caps_set" "', argument " "2"" of type '" "channel_cap_t *""'"); 
  }
  arg2 = (channel_cap_t *)(argp2);
  if (arg1) (arg1)->mem_caps = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_chan_list_mem_caps_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct chan_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  channel_cap_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:chan_list_mem_caps_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "chan_list_mem_caps_get" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  result = (channel_cap_t *)& ((arg1)->mem_caps);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_channel_cap,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_chan_list(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct chan_list *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_chan_list ") == TCL_ERROR) SWIG_fail;
  result = (struct chan_list *)calloc(1, sizeof(struct chan_list));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_chan_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_chan_list(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct chan_list *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_chan_list self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_chan_list, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_chan_list" "', argument " "1"" of type '" "struct chan_list *""'"); 
  }
  arg1 = (struct chan_list *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_chan_list(void *obj) {
struct chan_list *arg1 = (struct chan_list *) obj;
free((char *) arg1);
}
static swig_method swig_chan_list_methods[] = {
    {0,0}
};
static swig_attribute swig_chan_list_attributes[] = {
    { "-startc",_wrap_chan_list_startc_get, _wrap_chan_list_startc_set},
    { "-endc",_wrap_chan_list_endc_get, _wrap_chan_list_endc_set},
    { "-type",_wrap_chan_list_type_get, _wrap_chan_list_type_set},
    { "-mem_caps",_wrap_chan_list_mem_caps_get, _wrap_chan_list_mem_caps_set},
    {0,0,0}
};
static swig_class *swig_chan_list_bases[] = {0};
static const char * swig_chan_list_base_names[] = {0};
static swig_class _wrap_class_chan_list = { "chan_list", &SWIGTYPE_p_chan_list,_wrap_new_chan_list, swig_delete_chan_list, swig_chan_list_methods, swig_chan_list_attributes, swig_chan_list_bases,swig_chan_list_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_gran_min_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct gran *arg1 = 0 ;
  value_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:gran_min_set self min ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "gran_min_set" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "gran_min_set" "', argument " "2"" of type '" "value_t *""'"); 
  }
  arg2 = (value_t *)(argp2);
  if (arg1) (arg1)->min = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_gran_min_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct gran *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  value_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:gran_min_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "gran_min_get" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  result = (value_t *)& ((arg1)->min);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_value_t,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_gran_max_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct gran *arg1 = 0 ;
  value_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:gran_max_set self max ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "gran_max_set" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "gran_max_set" "', argument " "2"" of type '" "value_t *""'"); 
  }
  arg2 = (value_t *)(argp2);
  if (arg1) (arg1)->max = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_gran_max_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct gran *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  value_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:gran_max_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "gran_max_get" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  result = (value_t *)& ((arg1)->max);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_value_t,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_gran_step_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct gran *arg1 = 0 ;
  value_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:gran_step_set self step ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "gran_step_set" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_value_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "gran_step_set" "', argument " "2"" of type '" "value_t *""'"); 
  }
  arg2 = (value_t *)(argp2);
  if (arg1) (arg1)->step = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_gran_step_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct gran *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  value_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:gran_step_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "gran_step_get" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  result = (value_t *)& ((arg1)->step);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_value_t,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_gran(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct gran *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_gran ") == TCL_ERROR) SWIG_fail;
  result = (struct gran *)calloc(1, sizeof(struct gran));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_gran(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct gran *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_gran self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_gran, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_gran" "', argument " "1"" of type '" "struct gran *""'"); 
  }
  arg1 = (struct gran *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_gran(void *obj) {
struct gran *arg1 = (struct gran *) obj;
free((char *) arg1);
}
static swig_method swig_gran_methods[] = {
    {0,0}
};
static swig_attribute swig_gran_attributes[] = {
    { "-min",_wrap_gran_min_get, _wrap_gran_min_set},
    { "-max",_wrap_gran_max_get, _wrap_gran_max_set},
    { "-step",_wrap_gran_step_get, _wrap_gran_step_set},
    {0,0,0}
};
static swig_class *swig_gran_bases[] = {0};
static const char * swig_gran_base_names[] = {0};
static swig_class _wrap_class_gran = { "gran", &SWIGTYPE_p_gran,_wrap_new_gran, swig_delete_gran, swig_gran_methods, swig_gran_attributes, swig_gran_bases,swig_gran_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_cal_table_size_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct cal_table *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:cal_table_size_set self size ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_size_set" "', argument " "1"" of type '" "struct cal_table *""'"); 
  }
  arg1 = (struct cal_table *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "cal_table_size_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->size = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_cal_table_size_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct cal_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:cal_table_size_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_size_get" "', argument " "1"" of type '" "struct cal_table *""'"); 
  }
  arg1 = (struct cal_table *)(argp1);
  result = (int) ((arg1)->size);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_cal_table_table_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct cal_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cal_table_table *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:cal_table_table_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_table_get" "', argument " "1"" of type '" "struct cal_table *""'"); 
  }
  arg1 = (struct cal_table *)(argp1);
  result = (cal_table_table *)(cal_table_table *) ((arg1)->table);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_table,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_cal_table(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct cal_table *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_cal_table ") == TCL_ERROR) SWIG_fail;
  result = (struct cal_table *)calloc(1, sizeof(struct cal_table));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_cal_table(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct cal_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_cal_table self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_cal_table" "', argument " "1"" of type '" "struct cal_table *""'"); 
  }
  arg1 = (struct cal_table *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_cal_table(void *obj) {
struct cal_table *arg1 = (struct cal_table *) obj;
free((char *) arg1);
}
static swig_method swig_cal_table_methods[] = {
    {0,0}
};
static swig_attribute swig_cal_table_attributes[] = {
    { "-size",_wrap_cal_table_size_get, _wrap_cal_table_size_set},
    { "-table",_wrap_cal_table_table_get, 0 },
    {0,0,0}
};
static swig_class *swig_cal_table_bases[] = {0};
static const char * swig_cal_table_base_names[] = {0};
static swig_class _wrap_class_cal_table = { "cal_table", &SWIGTYPE_p_cal_table,_wrap_new_cal_table, swig_delete_cal_table, swig_cal_table_methods, swig_cal_table_attributes, swig_cal_table_bases,swig_cal_table_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_cal_table_table_raw_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  cal_table_table *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:cal_table_table_raw_set self raw ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_table_raw_set" "', argument " "1"" of type '" "cal_table_table *""'"); 
  }
  arg1 = (cal_table_table *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "cal_table_table_raw_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->raw = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_cal_table_table_raw_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  cal_table_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:cal_table_table_raw_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_table_raw_get" "', argument " "1"" of type '" "cal_table_table *""'"); 
  }
  arg1 = (cal_table_table *)(argp1);
  result = (int) ((arg1)->raw);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_cal_table_table_val_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  cal_table_table *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:cal_table_table_val_set self val ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_table_val_set" "', argument " "1"" of type '" "cal_table_table *""'"); 
  }
  arg1 = (cal_table_table *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "cal_table_table_val_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->val = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_cal_table_table_val_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  cal_table_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:cal_table_table_val_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_table_val_get" "', argument " "1"" of type '" "cal_table_table *""'"); 
  }
  arg1 = (cal_table_table *)(argp1);
  result = (int) ((arg1)->val);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_cal_table_table(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  cal_table_table *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_cal_table_table ") == TCL_ERROR) SWIG_fail;
  result = (cal_table_table *)calloc(1, sizeof(cal_table_table));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_table,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_cal_table_table(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  cal_table_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_cal_table_table self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_table, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_cal_table_table" "', argument " "1"" of type '" "cal_table_table *""'"); 
  }
  arg1 = (cal_table_table *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_cal_table_table(void *obj) {
cal_table_table *arg1 = (cal_table_table *) obj;
free((char *) arg1);
}
static swig_method swig_cal_table_table_methods[] = {
    {0,0}
};
static swig_attribute swig_cal_table_table_attributes[] = {
    { "-raw",_wrap_cal_table_table_raw_get, _wrap_cal_table_table_raw_set},
    { "-val",_wrap_cal_table_table_val_get, _wrap_cal_table_table_val_set},
    {0,0,0}
};
static swig_class *swig_cal_table_table_bases[] = {0};
static const char * swig_cal_table_table_base_names[] = {0};
static swig_class _wrap_class_cal_table_table = { "cal_table_table", &SWIGTYPE_p_cal_table_table,_wrap_new_cal_table_table, swig_delete_cal_table_table, swig_cal_table_table_methods, swig_cal_table_table_attributes, swig_cal_table_table_bases,swig_cal_table_table_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_cal_table_float_size_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct cal_table_float *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:cal_table_float_size_set self size ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_size_set" "', argument " "1"" of type '" "struct cal_table_float *""'"); 
  }
  arg1 = (struct cal_table_float *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "cal_table_float_size_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->size = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_cal_table_float_size_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct cal_table_float *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:cal_table_float_size_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_size_get" "', argument " "1"" of type '" "struct cal_table_float *""'"); 
  }
  arg1 = (struct cal_table_float *)(argp1);
  result = (int) ((arg1)->size);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_cal_table_float_table_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct cal_table_float *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cal_table_float_table *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:cal_table_float_table_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_table_get" "', argument " "1"" of type '" "struct cal_table_float *""'"); 
  }
  arg1 = (struct cal_table_float *)(argp1);
  result = (cal_table_float_table *)(cal_table_float_table *) ((arg1)->table);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float_table,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_cal_table_float(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct cal_table_float *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_cal_table_float ") == TCL_ERROR) SWIG_fail;
  result = (struct cal_table_float *)calloc(1, sizeof(struct cal_table_float));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_cal_table_float(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct cal_table_float *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_cal_table_float self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_float, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_cal_table_float" "', argument " "1"" of type '" "struct cal_table_float *""'"); 
  }
  arg1 = (struct cal_table_float *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_cal_table_float(void *obj) {
struct cal_table_float *arg1 = (struct cal_table_float *) obj;
free((char *) arg1);
}
static swig_method swig_cal_table_float_methods[] = {
    {0,0}
};
static swig_attribute swig_cal_table_float_attributes[] = {
    { "-size",_wrap_cal_table_float_size_get, _wrap_cal_table_float_size_set},
    { "-table",_wrap_cal_table_float_table_get, 0 },
    {0,0,0}
};
static swig_class *swig_cal_table_float_bases[] = {0};
static const char * swig_cal_table_float_base_names[] = {0};
static swig_class _wrap_class_cal_table_float = { "cal_table_float", &SWIGTYPE_p_cal_table_float,_wrap_new_cal_table_float, swig_delete_cal_table_float, swig_cal_table_float_methods, swig_cal_table_float_attributes, swig_cal_table_float_bases,swig_cal_table_float_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_cal_table_float_table_raw_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  cal_table_float_table *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:cal_table_float_table_raw_set self raw ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_float_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_table_raw_set" "', argument " "1"" of type '" "cal_table_float_table *""'"); 
  }
  arg1 = (cal_table_float_table *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "cal_table_float_table_raw_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->raw = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_cal_table_float_table_raw_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  cal_table_float_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:cal_table_float_table_raw_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_float_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_table_raw_get" "', argument " "1"" of type '" "cal_table_float_table *""'"); 
  }
  arg1 = (cal_table_float_table *)(argp1);
  result = (int) ((arg1)->raw);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_cal_table_float_table_val_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  cal_table_float_table *arg1 = 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:cal_table_float_table_val_set self val ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_float_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_table_val_set" "', argument " "1"" of type '" "cal_table_float_table *""'"); 
  }
  arg1 = (cal_table_float_table *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "cal_table_float_table_val_set" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = (float)(val2);
  if (arg1) (arg1)->val = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_cal_table_float_table_val_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  cal_table_float_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:cal_table_float_table_val_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_float_table, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cal_table_float_table_val_get" "', argument " "1"" of type '" "cal_table_float_table *""'"); 
  }
  arg1 = (cal_table_float_table *)(argp1);
  result = (float) ((arg1)->val);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_cal_table_float_table(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  cal_table_float_table *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_cal_table_float_table ") == TCL_ERROR) SWIG_fail;
  result = (cal_table_float_table *)calloc(1, sizeof(cal_table_float_table));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float_table,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_cal_table_float_table(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  cal_table_float_table *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_cal_table_float_table self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_cal_table_float_table, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_cal_table_float_table" "', argument " "1"" of type '" "cal_table_float_table *""'"); 
  }
  arg1 = (cal_table_float_table *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_cal_table_float_table(void *obj) {
cal_table_float_table *arg1 = (cal_table_float_table *) obj;
free((char *) arg1);
}
static swig_method swig_cal_table_float_table_methods[] = {
    {0,0}
};
static swig_attribute swig_cal_table_float_table_attributes[] = {
    { "-raw",_wrap_cal_table_float_table_raw_get, _wrap_cal_table_float_table_raw_set},
    { "-val",_wrap_cal_table_float_table_val_get, _wrap_cal_table_float_table_val_set},
    {0,0,0}
};
static swig_class *swig_cal_table_float_table_bases[] = {0};
static const char * swig_cal_table_float_table_base_names[] = {0};
static swig_class _wrap_class_cal_table_float_table = { "cal_table_float_table", &SWIGTYPE_p_cal_table_float_table,_wrap_new_cal_table_float_table, swig_delete_cal_table_float_table, swig_cal_table_float_table_methods, swig_cal_table_float_table_attributes, swig_cal_table_float_table_bases,swig_cal_table_float_table_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_rig_spectrum_scope_id_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_scope *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_scope_id_set self id ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_scope_id_set" "', argument " "1"" of type '" "struct rig_spectrum_scope *""'"); 
  }
  arg1 = (struct rig_spectrum_scope *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_scope_id_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->id = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_scope_id_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_scope *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_scope_id_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_scope_id_get" "', argument " "1"" of type '" "struct rig_spectrum_scope *""'"); 
  }
  arg1 = (struct rig_spectrum_scope *)(argp1);
  result = (int) ((arg1)->id);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_scope_name_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_scope *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_scope_name_set self name ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_scope_name_set" "', argument " "1"" of type '" "struct rig_spectrum_scope *""'"); 
  }
  arg1 = (struct rig_spectrum_scope *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_spectrum_scope_name_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->name);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->name = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->name = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_scope_name_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_scope *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_scope_name_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_scope_name_get" "', argument " "1"" of type '" "struct rig_spectrum_scope *""'"); 
  }
  arg1 = (struct rig_spectrum_scope *)(argp1);
  result = (char *) ((arg1)->name);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_rig_spectrum_scope(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_scope *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_rig_spectrum_scope ") == TCL_ERROR) SWIG_fail;
  result = (struct rig_spectrum_scope *)calloc(1, sizeof(struct rig_spectrum_scope));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_scope,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_rig_spectrum_scope(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_scope *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_rig_spectrum_scope self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_scope, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rig_spectrum_scope" "', argument " "1"" of type '" "struct rig_spectrum_scope *""'"); 
  }
  arg1 = (struct rig_spectrum_scope *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_rig_spectrum_scope(void *obj) {
struct rig_spectrum_scope *arg1 = (struct rig_spectrum_scope *) obj;
free((char *) arg1);
}
static swig_method swig_rig_spectrum_scope_methods[] = {
    {0,0}
};
static swig_attribute swig_rig_spectrum_scope_attributes[] = {
    { "-id",_wrap_rig_spectrum_scope_id_get, _wrap_rig_spectrum_scope_id_set},
    { "-name",_wrap_rig_spectrum_scope_name_get, _wrap_rig_spectrum_scope_name_set},
    {0,0,0}
};
static swig_class *swig_rig_spectrum_scope_bases[] = {0};
static const char * swig_rig_spectrum_scope_base_names[] = {0};
static swig_class _wrap_class_rig_spectrum_scope = { "rig_spectrum_scope", &SWIGTYPE_p_rig_spectrum_scope,_wrap_new_rig_spectrum_scope, swig_delete_rig_spectrum_scope, swig_rig_spectrum_scope_methods, swig_rig_spectrum_scope_attributes, swig_rig_spectrum_scope_bases,swig_rig_spectrum_scope_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_rig_spectrum_avg_mode_id_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_avg_mode *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_avg_mode_id_set self id ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_avg_mode_id_set" "', argument " "1"" of type '" "struct rig_spectrum_avg_mode *""'"); 
  }
  arg1 = (struct rig_spectrum_avg_mode *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_avg_mode_id_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->id = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_avg_mode_id_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_avg_mode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_avg_mode_id_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_avg_mode_id_get" "', argument " "1"" of type '" "struct rig_spectrum_avg_mode *""'"); 
  }
  arg1 = (struct rig_spectrum_avg_mode *)(argp1);
  result = (int) ((arg1)->id);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_avg_mode_name_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_avg_mode *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_avg_mode_name_set self name ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_avg_mode_name_set" "', argument " "1"" of type '" "struct rig_spectrum_avg_mode *""'"); 
  }
  arg1 = (struct rig_spectrum_avg_mode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_spectrum_avg_mode_name_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->name);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->name = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->name = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_avg_mode_name_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_avg_mode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_avg_mode_name_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_avg_mode_name_get" "', argument " "1"" of type '" "struct rig_spectrum_avg_mode *""'"); 
  }
  arg1 = (struct rig_spectrum_avg_mode *)(argp1);
  result = (char *) ((arg1)->name);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_rig_spectrum_avg_mode(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_avg_mode *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_rig_spectrum_avg_mode ") == TCL_ERROR) SWIG_fail;
  result = (struct rig_spectrum_avg_mode *)calloc(1, sizeof(struct rig_spectrum_avg_mode));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_avg_mode,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_rig_spectrum_avg_mode(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_avg_mode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_rig_spectrum_avg_mode self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_avg_mode, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rig_spectrum_avg_mode" "', argument " "1"" of type '" "struct rig_spectrum_avg_mode *""'"); 
  }
  arg1 = (struct rig_spectrum_avg_mode *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_rig_spectrum_avg_mode(void *obj) {
struct rig_spectrum_avg_mode *arg1 = (struct rig_spectrum_avg_mode *) obj;
free((char *) arg1);
}
static swig_method swig_rig_spectrum_avg_mode_methods[] = {
    {0,0}
};
static swig_attribute swig_rig_spectrum_avg_mode_attributes[] = {
    { "-id",_wrap_rig_spectrum_avg_mode_id_get, _wrap_rig_spectrum_avg_mode_id_set},
    { "-name",_wrap_rig_spectrum_avg_mode_name_get, _wrap_rig_spectrum_avg_mode_name_set},
    {0,0,0}
};
static swig_class *swig_rig_spectrum_avg_mode_bases[] = {0};
static const char * swig_rig_spectrum_avg_mode_base_names[] = {0};
static swig_class _wrap_class_rig_spectrum_avg_mode = { "rig_spectrum_avg_mode", &SWIGTYPE_p_rig_spectrum_avg_mode,_wrap_new_rig_spectrum_avg_mode, swig_delete_rig_spectrum_avg_mode, swig_rig_spectrum_avg_mode_methods, swig_rig_spectrum_avg_mode_attributes, swig_rig_spectrum_avg_mode_bases,swig_rig_spectrum_avg_mode_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_rig_spectrum_line_id_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_line_id_set self id ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_id_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_id_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->id = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_id_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_line_id_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_id_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (int) ((arg1)->id);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_data_level_min_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_line_data_level_min_set self data_level_min ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_data_level_min_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_data_level_min_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->data_level_min = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_data_level_min_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_line_data_level_min_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_data_level_min_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (int) ((arg1)->data_level_min);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_data_level_max_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_line_data_level_max_set self data_level_max ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_data_level_max_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_data_level_max_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->data_level_max = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_data_level_max_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_line_data_level_max_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_data_level_max_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (int) ((arg1)->data_level_max);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_signal_strength_min_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_line_signal_strength_min_set self signal_strength_min ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_signal_strength_min_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_signal_strength_min_set" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  if (arg1) (arg1)->signal_strength_min = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_signal_strength_min_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_line_signal_strength_min_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_signal_strength_min_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (double) ((arg1)->signal_strength_min);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_signal_strength_max_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_line_signal_strength_max_set self signal_strength_max ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_signal_strength_max_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_signal_strength_max_set" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  if (arg1) (arg1)->signal_strength_max = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_signal_strength_max_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_line_signal_strength_max_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_signal_strength_max_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (double) ((arg1)->signal_strength_max);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_spectrum_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  enum rig_spectrum_mode_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_line_spectrum_mode_set self spectrum_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_spectrum_mode_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_spectrum_mode_set" "', argument " "2"" of type '" "enum rig_spectrum_mode_e""'");
  } 
  arg2 = (enum rig_spectrum_mode_e)(val2);
  if (arg1) (arg1)->spectrum_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_spectrum_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum rig_spectrum_mode_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_line_spectrum_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_spectrum_mode_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (enum rig_spectrum_mode_e) ((arg1)->spectrum_mode);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_center_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_line_center_freq_set self center_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_center_freq_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_center_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->center_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_center_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_line_center_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_center_freq_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (freq_t) ((arg1)->center_freq);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_span_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_line_span_freq_set self span_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_span_freq_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_span_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->span_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_span_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_line_span_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_span_freq_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (freq_t) ((arg1)->span_freq);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_low_edge_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_line_low_edge_freq_set self low_edge_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_low_edge_freq_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_low_edge_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->low_edge_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_low_edge_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_line_low_edge_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_low_edge_freq_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (freq_t) ((arg1)->low_edge_freq);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_high_edge_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_line_high_edge_freq_set self high_edge_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_high_edge_freq_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_high_edge_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->high_edge_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_high_edge_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_line_high_edge_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_high_edge_freq_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (freq_t) ((arg1)->high_edge_freq);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_spectrum_data_length_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_line_spectrum_data_length_set self spectrum_data_length ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_spectrum_data_length_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  ecode2 = SWIG_AsVal_size_t SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_spectrum_line_spectrum_data_length_set" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  if (arg1) (arg1)->spectrum_data_length = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_spectrum_data_length_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_line_spectrum_data_length_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_spectrum_data_length_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result =  ((arg1)->spectrum_data_length);
  Tcl_SetObjResult(interp,SWIG_From_size_t((size_t)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_spectrum_data_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  unsigned char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_spectrum_line_spectrum_data_set self spectrum_data ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_spectrum_data_set" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_unsigned_char, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_spectrum_line_spectrum_data_set" "', argument " "2"" of type '" "unsigned char *""'"); 
  }
  arg2 = (unsigned char *)(argp2);
  if (arg1) (arg1)->spectrum_data = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_spectrum_line_spectrum_data_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_spectrum_line_spectrum_data_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_spectrum_line_spectrum_data_get" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  result = (unsigned char *) ((arg1)->spectrum_data);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_rig_spectrum_line(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_rig_spectrum_line ") == TCL_ERROR) SWIG_fail;
  result = (struct rig_spectrum_line *)calloc(1, sizeof(struct rig_spectrum_line));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_line,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_rig_spectrum_line(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_spectrum_line *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_rig_spectrum_line self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_spectrum_line, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rig_spectrum_line" "', argument " "1"" of type '" "struct rig_spectrum_line *""'"); 
  }
  arg1 = (struct rig_spectrum_line *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_rig_spectrum_line(void *obj) {
struct rig_spectrum_line *arg1 = (struct rig_spectrum_line *) obj;
free((char *) arg1);
}
static swig_method swig_rig_spectrum_line_methods[] = {
    {0,0}
};
static swig_attribute swig_rig_spectrum_line_attributes[] = {
    { "-id",_wrap_rig_spectrum_line_id_get, _wrap_rig_spectrum_line_id_set},
    { "-data_level_min",_wrap_rig_spectrum_line_data_level_min_get, _wrap_rig_spectrum_line_data_level_min_set},
    { "-data_level_max",_wrap_rig_spectrum_line_data_level_max_get, _wrap_rig_spectrum_line_data_level_max_set},
    { "-signal_strength_min",_wrap_rig_spectrum_line_signal_strength_min_get, _wrap_rig_spectrum_line_signal_strength_min_set},
    { "-signal_strength_max",_wrap_rig_spectrum_line_signal_strength_max_get, _wrap_rig_spectrum_line_signal_strength_max_set},
    { "-spectrum_mode",_wrap_rig_spectrum_line_spectrum_mode_get, _wrap_rig_spectrum_line_spectrum_mode_set},
    { "-center_freq",_wrap_rig_spectrum_line_center_freq_get, _wrap_rig_spectrum_line_center_freq_set},
    { "-span_freq",_wrap_rig_spectrum_line_span_freq_get, _wrap_rig_spectrum_line_span_freq_set},
    { "-low_edge_freq",_wrap_rig_spectrum_line_low_edge_freq_get, _wrap_rig_spectrum_line_low_edge_freq_set},
    { "-high_edge_freq",_wrap_rig_spectrum_line_high_edge_freq_get, _wrap_rig_spectrum_line_high_edge_freq_set},
    { "-spectrum_data_length",_wrap_rig_spectrum_line_spectrum_data_length_get, _wrap_rig_spectrum_line_spectrum_data_length_set},
    { "-spectrum_data",_wrap_rig_spectrum_line_spectrum_data_get, _wrap_rig_spectrum_line_spectrum_data_set},
    {0,0,0}
};
static swig_class *swig_rig_spectrum_line_bases[] = {0};
static const char * swig_rig_spectrum_line_base_names[] = {0};
static swig_class _wrap_class_rig_spectrum_line = { "rig_spectrum_line", &SWIGTYPE_p_rig_spectrum_line,_wrap_new_rig_spectrum_line, swig_delete_rig_spectrum_line, swig_rig_spectrum_line_methods, swig_rig_spectrum_line_attributes, swig_rig_spectrum_line_bases,swig_rig_spectrum_line_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_deferred_config_item_nextt_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_item *arg1 = 0 ;
  struct deferred_config_item *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:deferred_config_item_nextt_set self nextt ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_deferred_config_item, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_item_nextt_set" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_deferred_config_item, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "deferred_config_item_nextt_set" "', argument " "2"" of type '" "struct deferred_config_item *""'"); 
  }
  arg2 = (struct deferred_config_item *)(argp2);
  if (arg1) (arg1)->nextt = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_deferred_config_item_nextt_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_item *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct deferred_config_item *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:deferred_config_item_nextt_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_deferred_config_item, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_item_nextt_get" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  result = (struct deferred_config_item *) ((arg1)->nextt);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_deferred_config_item,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_deferred_config_item_token_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_item *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:deferred_config_item_token_set self token ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_deferred_config_item, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_item_token_set" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "deferred_config_item_token_set" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  if (arg1) (arg1)->token = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_deferred_config_item_token_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_item *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  hamlib_token_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:deferred_config_item_token_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_deferred_config_item, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_item_token_get" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  result = (hamlib_token_t) ((arg1)->token);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_deferred_config_item_value_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_item *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:deferred_config_item_value_set self value ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_deferred_config_item, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_item_value_set" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "deferred_config_item_value_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->value);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->value = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->value = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_deferred_config_item_value_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_item *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:deferred_config_item_value_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_deferred_config_item, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_item_value_get" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  result = (char *) ((arg1)->value);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_deferred_config_item(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_item *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_deferred_config_item ") == TCL_ERROR) SWIG_fail;
  result = (struct deferred_config_item *)calloc(1, sizeof(struct deferred_config_item));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_deferred_config_item,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_deferred_config_item(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_item *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_deferred_config_item self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_deferred_config_item, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_deferred_config_item" "', argument " "1"" of type '" "struct deferred_config_item *""'"); 
  }
  arg1 = (struct deferred_config_item *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_deferred_config_item(void *obj) {
struct deferred_config_item *arg1 = (struct deferred_config_item *) obj;
free((char *) arg1);
}
static swig_method swig_deferred_config_item_methods[] = {
    {0,0}
};
static swig_attribute swig_deferred_config_item_attributes[] = {
    { "-nextt",_wrap_deferred_config_item_nextt_get, _wrap_deferred_config_item_nextt_set},
    { "-token",_wrap_deferred_config_item_token_get, _wrap_deferred_config_item_token_set},
    { "-value",_wrap_deferred_config_item_value_get, _wrap_deferred_config_item_value_set},
    {0,0,0}
};
static swig_class *swig_deferred_config_item_bases[] = {0};
static const char * swig_deferred_config_item_base_names[] = {0};
static swig_class _wrap_class_deferred_config_item = { "deferred_config_item", &SWIGTYPE_p_deferred_config_item,_wrap_new_deferred_config_item, swig_delete_deferred_config_item, swig_deferred_config_item_methods, swig_deferred_config_item_attributes, swig_deferred_config_item_bases,swig_deferred_config_item_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_deferred_config_header_firstt_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_header *arg1 = 0 ;
  struct deferred_config_item *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:deferred_config_header_firstt_set self firstt ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_deferred_config_header, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_header_firstt_set" "', argument " "1"" of type '" "struct deferred_config_header *""'"); 
  }
  arg1 = (struct deferred_config_header *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_deferred_config_item, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "deferred_config_header_firstt_set" "', argument " "2"" of type '" "struct deferred_config_item *""'"); 
  }
  arg2 = (struct deferred_config_item *)(argp2);
  if (arg1) (arg1)->firstt = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_deferred_config_header_firstt_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_header *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct deferred_config_item *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:deferred_config_header_firstt_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_deferred_config_header, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_header_firstt_get" "', argument " "1"" of type '" "struct deferred_config_header *""'"); 
  }
  arg1 = (struct deferred_config_header *)(argp1);
  result = (struct deferred_config_item *) ((arg1)->firstt);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_deferred_config_item,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_deferred_config_header_lastt_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_header *arg1 = 0 ;
  struct deferred_config_item *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:deferred_config_header_lastt_set self lastt ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_deferred_config_header, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_header_lastt_set" "', argument " "1"" of type '" "struct deferred_config_header *""'"); 
  }
  arg1 = (struct deferred_config_header *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_deferred_config_item, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "deferred_config_header_lastt_set" "', argument " "2"" of type '" "struct deferred_config_item *""'"); 
  }
  arg2 = (struct deferred_config_item *)(argp2);
  if (arg1) (arg1)->lastt = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_deferred_config_header_lastt_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_header *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct deferred_config_item *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:deferred_config_header_lastt_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_deferred_config_header, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "deferred_config_header_lastt_get" "', argument " "1"" of type '" "struct deferred_config_header *""'"); 
  }
  arg1 = (struct deferred_config_header *)(argp1);
  result = (struct deferred_config_item *) ((arg1)->lastt);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_deferred_config_item,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_deferred_config_header(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_header *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_deferred_config_header ") == TCL_ERROR) SWIG_fail;
  result = (struct deferred_config_header *)calloc(1, sizeof(struct deferred_config_header));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_deferred_config_header,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_deferred_config_header(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct deferred_config_header *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_deferred_config_header self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_deferred_config_header, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_deferred_config_header" "', argument " "1"" of type '" "struct deferred_config_header *""'"); 
  }
  arg1 = (struct deferred_config_header *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_deferred_config_header(void *obj) {
struct deferred_config_header *arg1 = (struct deferred_config_header *) obj;
free((char *) arg1);
}
static swig_method swig_deferred_config_header_methods[] = {
    {0,0}
};
static swig_attribute swig_deferred_config_header_attributes[] = {
    { "-firstt",_wrap_deferred_config_header_firstt_get, _wrap_deferred_config_header_firstt_set},
    { "-lastt",_wrap_deferred_config_header_lastt_get, _wrap_deferred_config_header_lastt_set},
    {0,0,0}
};
static swig_class *swig_deferred_config_header_bases[] = {0};
static const char * swig_deferred_config_header_base_names[] = {0};
static swig_class _wrap_class_deferred_config_header = { "deferred_config_header", &SWIGTYPE_p_deferred_config_header,_wrap_new_deferred_config_header, swig_delete_deferred_config_header, swig_deferred_config_header_methods, swig_deferred_config_header_attributes, swig_deferred_config_header_bases,swig_deferred_config_header_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_rig_caps_rig_model_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  rig_model_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rig_model_set self rig_model ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_model_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_rig_model_set" "', argument " "2"" of type '" "rig_model_t""'");
  } 
  arg2 = (rig_model_t)(val2);
  if (arg1) (arg1)->rig_model = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_model_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rig_model_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rig_model_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_model_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (rig_model_t) ((arg1)->rig_model);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_model_name_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_model_name_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_model_name_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->model_name);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_mfg_name_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_mfg_name_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_mfg_name_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->mfg_name);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_version_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_version_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_version_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->version);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_copyright_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_copyright_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_copyright_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->copyright);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_status_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  enum rig_status_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_status_set self status ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_status_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_status_set" "', argument " "2"" of type '" "enum rig_status_e""'");
  } 
  arg2 = (enum rig_status_e)(val2);
  if (arg1) (arg1)->status = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_status_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum rig_status_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_status_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_status_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (enum rig_status_e) ((arg1)->status);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rig_type_set self rig_type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_type_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_rig_type_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->rig_type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rig_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_type_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->rig_type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_ptt_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  ptt_type_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_ptt_type_set self ptt_type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_ptt_type_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_ptt_type_set" "', argument " "2"" of type '" "ptt_type_t""'");
  } 
  arg2 = (ptt_type_t)(val2);
  if (arg1) (arg1)->ptt_type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_ptt_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptt_type_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_ptt_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_ptt_type_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (ptt_type_t) ((arg1)->ptt_type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_dcd_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  dcd_type_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_dcd_type_set self dcd_type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_dcd_type_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_dcd_type_set" "', argument " "2"" of type '" "dcd_type_t""'");
  } 
  arg2 = (dcd_type_t)(val2);
  if (arg1) (arg1)->dcd_type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_dcd_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  dcd_type_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_dcd_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_dcd_type_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (dcd_type_t) ((arg1)->dcd_type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_port_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  rig_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_port_type_set self port_type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_port_type_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_port_type_set" "', argument " "2"" of type '" "rig_port_t""'");
  } 
  arg2 = (rig_port_t)(val2);
  if (arg1) (arg1)->port_type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_port_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rig_port_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_port_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_port_type_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (rig_port_t) ((arg1)->port_type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_serial_rate_min_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_serial_rate_min_set self serial_rate_min ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_rate_min_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_serial_rate_min_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_min = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_serial_rate_min_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_serial_rate_min_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_rate_min_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->serial_rate_min);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_serial_rate_max_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_serial_rate_max_set self serial_rate_max ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_rate_max_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_serial_rate_max_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_max = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_serial_rate_max_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_serial_rate_max_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_rate_max_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->serial_rate_max);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_serial_data_bits_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_serial_data_bits_set self serial_data_bits ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_data_bits_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_serial_data_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_data_bits = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_serial_data_bits_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_serial_data_bits_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_data_bits_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->serial_data_bits);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_serial_stop_bits_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_serial_stop_bits_set self serial_stop_bits ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_stop_bits_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_serial_stop_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_stop_bits = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_serial_stop_bits_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_serial_stop_bits_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_stop_bits_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->serial_stop_bits);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_serial_parity_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  enum serial_parity_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_serial_parity_set self serial_parity ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_parity_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_serial_parity_set" "', argument " "2"" of type '" "enum serial_parity_e""'");
  } 
  arg2 = (enum serial_parity_e)(val2);
  if (arg1) (arg1)->serial_parity = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_serial_parity_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum serial_parity_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_serial_parity_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_parity_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (enum serial_parity_e) ((arg1)->serial_parity);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_serial_handshake_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  enum serial_handshake_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_serial_handshake_set self serial_handshake ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_handshake_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_serial_handshake_set" "', argument " "2"" of type '" "enum serial_handshake_e""'");
  } 
  arg2 = (enum serial_handshake_e)(val2);
  if (arg1) (arg1)->serial_handshake = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_serial_handshake_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum serial_handshake_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_serial_handshake_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_serial_handshake_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (enum serial_handshake_e) ((arg1)->serial_handshake);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_write_delay_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_write_delay_set self write_delay ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_write_delay_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->write_delay = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_write_delay_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_write_delay_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_write_delay_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->write_delay);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_post_write_delay_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_post_write_delay_set self post_write_delay ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_post_write_delay_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_post_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->post_write_delay = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_post_write_delay_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_post_write_delay_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_post_write_delay_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->post_write_delay);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_timeout_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_timeout_set self timeout ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_timeout_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_timeout_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->timeout = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_timeout_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_timeout_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_timeout_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->timeout);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_retry_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_retry_set self retry ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_retry_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_retry_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->retry = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_retry_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_retry_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_retry_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->retry);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_has_get_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_has_get_func_set self has_get_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_get_func_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_has_get_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_has_get_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_has_get_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_get_func_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_func);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_has_set_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_has_set_func_set self has_set_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_set_func_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_has_set_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_has_set_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_has_set_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_set_func_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_func);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_has_get_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_has_get_level_set self has_get_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_get_level_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_has_get_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_has_get_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_has_get_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_get_level_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_level);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_has_set_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_has_set_level_set self has_set_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_set_level_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_has_set_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_has_set_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_has_set_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_set_level_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_level);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_has_get_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_has_get_parm_set self has_get_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_get_parm_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_has_get_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_has_get_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_has_get_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_get_parm_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_parm);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_has_set_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_has_set_parm_set self has_set_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_set_parm_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_has_set_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_has_set_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_has_set_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_has_set_parm_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_parm);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_level_gran_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_level_gran_set self level_gran ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_level_gran_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_level_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->level_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""level_gran""' of type '""gran_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_level_gran_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gran_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_level_gran_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_level_gran_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->level_gran);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_parm_gran_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_parm_gran_set self parm_gran ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_parm_gran_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_parm_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->parm_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""parm_gran""' of type '""gran_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_parm_gran_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gran_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_parm_gran_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_parm_gran_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->parm_gran);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_extparms_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_extparms_set self extparms ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_extparms_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_extparms_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extparms = (struct confparams const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_extparms_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct confparams *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_extparms_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_extparms_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extparms);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_extlevels_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_extlevels_set self extlevels ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_extlevels_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_extlevels_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extlevels = (struct confparams const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_extlevels_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct confparams *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_extlevels_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_extlevels_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extlevels);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_extfuncs_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_extfuncs_set self extfuncs ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_extfuncs_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_extfuncs_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extfuncs = (struct confparams const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_extfuncs_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct confparams *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_extfuncs_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_extfuncs_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extfuncs);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_ext_tokens_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_ext_tokens_set self ext_tokens ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_ext_tokens_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_int, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_ext_tokens_set" "', argument " "2"" of type '" "int *""'"); 
  }
  arg2 = (int *)(argp2);
  if (arg1) (arg1)->ext_tokens = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_ext_tokens_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_ext_tokens_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_ext_tokens_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int *) ((arg1)->ext_tokens);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_int,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_ctcss_list_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  tone_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_ctcss_list_set self ctcss_list ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_ctcss_list_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_unsigned_int, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_ctcss_list_set" "', argument " "2"" of type '" "tone_t *""'"); 
  }
  arg2 = (tone_t *)(argp2);
  if (arg1) (arg1)->ctcss_list = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_ctcss_list_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  tone_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_ctcss_list_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_ctcss_list_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (tone_t *) ((arg1)->ctcss_list);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_dcs_list_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  tone_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_dcs_list_set self dcs_list ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_dcs_list_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_unsigned_int, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_dcs_list_set" "', argument " "2"" of type '" "tone_t *""'"); 
  }
  arg2 = (tone_t *)(argp2);
  if (arg1) (arg1)->dcs_list = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_dcs_list_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  tone_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_dcs_list_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_dcs_list_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (tone_t *) ((arg1)->dcs_list);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_preamp_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int *arg2 = (int *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_preamp_set self preamp ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_preamp_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_preamp_set" "', argument " "2"" of type '" "int [8]""'"); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(int *)&arg1->preamp[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""preamp""' of type '""int [8]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_preamp_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_preamp_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_preamp_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int *)(int *) ((arg1)->preamp);
  {
    int i, len, l;
    len = 8;
    
    Tcl_Obj * list = Tcl_NewListObj(len, NULL);
    for(i=0; i < len; i++) {
      Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(result[i]));
    }
    Tcl_SetObjResult(interp, list);
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_attenuator_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int *arg2 = (int *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_attenuator_set self attenuator ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_attenuator_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_attenuator_set" "', argument " "2"" of type '" "int [8]""'"); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(int *)&arg1->attenuator[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""attenuator""' of type '""int [8]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_attenuator_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_attenuator_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_attenuator_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int *)(int *) ((arg1)->attenuator);
  {
    int i, len, l;
    len = 8;
    
    Tcl_Obj * list = Tcl_NewListObj(len, NULL);
    for(i=0; i < len; i++) {
      Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(result[i]));
    }
    Tcl_SetObjResult(interp, list);
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_max_rit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_max_rit_set self max_rit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_max_rit_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_max_rit_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->max_rit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_max_rit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_max_rit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_max_rit_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (shortfreq_t) ((arg1)->max_rit);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_max_xit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_max_xit_set self max_xit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_max_xit_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_max_xit_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->max_xit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_max_xit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_max_xit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_max_xit_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (shortfreq_t) ((arg1)->max_xit);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_max_ifshift_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_max_ifshift_set self max_ifshift ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_max_ifshift_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_max_ifshift_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->max_ifshift = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_max_ifshift_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_max_ifshift_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_max_ifshift_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (shortfreq_t) ((arg1)->max_ifshift);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_agc_level_count_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_agc_level_count_set self agc_level_count ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_agc_level_count_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_agc_level_count_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->agc_level_count = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_agc_level_count_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_agc_level_count_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_agc_level_count_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->agc_level_count);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_agc_levels_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  enum agc_level_e *arg2 = (enum agc_level_e *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_agc_levels_set self agc_levels ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_agc_levels_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_agc_level_e, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_agc_levels_set" "', argument " "2"" of type '" "enum agc_level_e [8]""'"); 
  } 
  arg2 = (enum agc_level_e *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(enum agc_level_e *)&arg1->agc_levels[ii] = *((enum agc_level_e *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""agc_levels""' of type '""enum agc_level_e [8]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_agc_levels_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum agc_level_e *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_agc_levels_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_agc_levels_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (enum agc_level_e *)(enum agc_level_e *) ((arg1)->agc_levels);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_agc_level_e,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_announces_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  ann_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_announces_set self announces ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_announces_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_announces_set" "', argument " "2"" of type '" "ann_t""'");
  } 
  arg2 = (ann_t)(val2);
  if (arg1) (arg1)->announces = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_announces_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ann_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_announces_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_announces_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (ann_t) ((arg1)->announces);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_vfo_ops_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  vfo_op_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_vfo_ops_set self vfo_ops ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_vfo_ops_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_vfo_ops_set" "', argument " "2"" of type '" "vfo_op_t""'");
  } 
  arg2 = (vfo_op_t)(val2);
  if (arg1) (arg1)->vfo_ops = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_vfo_ops_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  vfo_op_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_vfo_ops_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_vfo_ops_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (vfo_op_t) ((arg1)->vfo_ops);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_scan_ops_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  scan_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_scan_ops_set self scan_ops ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_scan_ops_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_scan_ops_set" "', argument " "2"" of type '" "scan_t""'");
  } 
  arg2 = (scan_t)(val2);
  if (arg1) (arg1)->scan_ops = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_scan_ops_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  scan_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_scan_ops_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_scan_ops_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (scan_t) ((arg1)->scan_ops);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_targetable_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_targetable_vfo_set self targetable_vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_targetable_vfo_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_targetable_vfo_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->targetable_vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_targetable_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_targetable_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_targetable_vfo_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->targetable_vfo);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_transceive_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_transceive_set self transceive ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_transceive_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_transceive_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->transceive = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_transceive_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_transceive_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_transceive_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->transceive);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_bank_qty_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_bank_qty_set self bank_qty ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_bank_qty_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_bank_qty_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->bank_qty = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_bank_qty_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_bank_qty_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_bank_qty_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->bank_qty);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_chan_desc_sz_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_chan_desc_sz_set self chan_desc_sz ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_chan_desc_sz_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_chan_desc_sz_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->chan_desc_sz = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_chan_desc_sz_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_chan_desc_sz_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_chan_desc_sz_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->chan_desc_sz);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_chan_list_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  chan_t *arg2 = (chan_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_chan_list_set self chan_list ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_chan_list_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_chan_list_set" "', argument " "2"" of type '" "chan_t [16]""'"); 
  } 
  arg2 = (chan_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)16; ++ii) *(chan_t *)&arg1->chan_list[ii] = *((chan_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""chan_list""' of type '""chan_t [16]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_chan_list_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  chan_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_chan_list_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_chan_list_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (chan_t *)(chan_t *) ((arg1)->chan_list);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_chan_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rx_range_list1_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rx_range_list1_set self rx_range_list1 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list1_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_rx_range_list1_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list1[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list1""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rx_range_list1_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rx_range_list1_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list1_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_tx_range_list1_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_tx_range_list1_set self tx_range_list1 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list1_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_tx_range_list1_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list1[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list1""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_tx_range_list1_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_tx_range_list1_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list1_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rx_range_list2_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rx_range_list2_set self rx_range_list2 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list2_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_rx_range_list2_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list2[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list2""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rx_range_list2_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rx_range_list2_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list2_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list2);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_tx_range_list2_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_tx_range_list2_set self tx_range_list2 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list2_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_tx_range_list2_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list2[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list2""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_tx_range_list2_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_tx_range_list2_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list2_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list2);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rx_range_list3_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rx_range_list3_set self rx_range_list3 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list3_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_rx_range_list3_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list3[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list3""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rx_range_list3_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rx_range_list3_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list3_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list3);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_tx_range_list3_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_tx_range_list3_set self tx_range_list3 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list3_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_tx_range_list3_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list3[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list3""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_tx_range_list3_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_tx_range_list3_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list3_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list3);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rx_range_list4_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rx_range_list4_set self rx_range_list4 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list4_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_rx_range_list4_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list4[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list4""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rx_range_list4_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rx_range_list4_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list4_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list4);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_tx_range_list4_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_tx_range_list4_set self tx_range_list4 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list4_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_tx_range_list4_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list4[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list4""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_tx_range_list4_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_tx_range_list4_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list4_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list4);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rx_range_list5_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rx_range_list5_set self rx_range_list5 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list5_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_rx_range_list5_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list5[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list5""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rx_range_list5_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rx_range_list5_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rx_range_list5_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list5);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_tx_range_list5_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_tx_range_list5_set self tx_range_list5 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list5_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_tx_range_list5_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list5[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list5""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_tx_range_list5_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_tx_range_list5_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tx_range_list5_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list5);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_tuning_steps_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  struct tuning_step_list *arg2 = (struct tuning_step_list *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_tuning_steps_set self tuning_steps ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tuning_steps_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_tuning_step_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_tuning_steps_set" "', argument " "2"" of type '" "struct tuning_step_list [20]""'"); 
  } 
  arg2 = (struct tuning_step_list *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)20; ++ii) *(struct tuning_step_list *)&arg1->tuning_steps[ii] = *((struct tuning_step_list *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tuning_steps""' of type '""struct tuning_step_list [20]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_tuning_steps_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct tuning_step_list *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_tuning_steps_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_tuning_steps_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct tuning_step_list *)(struct tuning_step_list *) ((arg1)->tuning_steps);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_tuning_step_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_filters_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  struct filter_list *arg2 = (struct filter_list *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_filters_set self filters ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_filters_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_filter_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_filters_set" "', argument " "2"" of type '" "struct filter_list [60]""'"); 
  } 
  arg2 = (struct filter_list *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)60; ++ii) *(struct filter_list *)&arg1->filters[ii] = *((struct filter_list *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""filters""' of type '""struct filter_list [60]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_filters_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct filter_list *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_filters_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_filters_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct filter_list *)(struct filter_list *) ((arg1)->filters);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_filter_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_str_cal_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  cal_table_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_str_cal_set self str_cal ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_str_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_cal_table, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_str_cal_set" "', argument " "2"" of type '" "cal_table_t *""'"); 
  }
  arg2 = (cal_table_t *)(argp2);
  if (arg1) (arg1)->str_cal = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_str_cal_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cal_table_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_str_cal_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_str_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_t *)& ((arg1)->str_cal);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_swr_cal_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  cal_table_float_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_swr_cal_set self swr_cal ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_swr_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_swr_cal_set" "', argument " "2"" of type '" "cal_table_float_t *""'"); 
  }
  arg2 = (cal_table_float_t *)(argp2);
  if (arg1) (arg1)->swr_cal = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_swr_cal_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cal_table_float_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_swr_cal_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_swr_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_float_t *)& ((arg1)->swr_cal);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_alc_cal_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  cal_table_float_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_alc_cal_set self alc_cal ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_alc_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_alc_cal_set" "', argument " "2"" of type '" "cal_table_float_t *""'"); 
  }
  arg2 = (cal_table_float_t *)(argp2);
  if (arg1) (arg1)->alc_cal = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_alc_cal_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cal_table_float_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_alc_cal_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_alc_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_float_t *)& ((arg1)->alc_cal);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rfpower_meter_cal_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  cal_table_float_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rfpower_meter_cal_set self rfpower_meter_cal ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rfpower_meter_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_rfpower_meter_cal_set" "', argument " "2"" of type '" "cal_table_float_t *""'"); 
  }
  arg2 = (cal_table_float_t *)(argp2);
  if (arg1) (arg1)->rfpower_meter_cal = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rfpower_meter_cal_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cal_table_float_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rfpower_meter_cal_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rfpower_meter_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_float_t *)& ((arg1)->rfpower_meter_cal);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_comp_meter_cal_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  cal_table_float_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_comp_meter_cal_set self comp_meter_cal ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_comp_meter_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_comp_meter_cal_set" "', argument " "2"" of type '" "cal_table_float_t *""'"); 
  }
  arg2 = (cal_table_float_t *)(argp2);
  if (arg1) (arg1)->comp_meter_cal = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_comp_meter_cal_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cal_table_float_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_comp_meter_cal_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_comp_meter_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_float_t *)& ((arg1)->comp_meter_cal);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_vd_meter_cal_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  cal_table_float_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_vd_meter_cal_set self vd_meter_cal ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_vd_meter_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_vd_meter_cal_set" "', argument " "2"" of type '" "cal_table_float_t *""'"); 
  }
  arg2 = (cal_table_float_t *)(argp2);
  if (arg1) (arg1)->vd_meter_cal = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_vd_meter_cal_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cal_table_float_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_vd_meter_cal_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_vd_meter_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_float_t *)& ((arg1)->vd_meter_cal);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_id_meter_cal_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  cal_table_float_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_id_meter_cal_set self id_meter_cal ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_id_meter_cal_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_cal_table_float, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_id_meter_cal_set" "', argument " "2"" of type '" "cal_table_float_t *""'"); 
  }
  arg2 = (cal_table_float_t *)(argp2);
  if (arg1) (arg1)->id_meter_cal = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_id_meter_cal_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cal_table_float_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_id_meter_cal_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_id_meter_cal_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (cal_table_float_t *)& ((arg1)->id_meter_cal);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table_float,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_spectrum_scopes_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  struct rig_spectrum_scope *arg2 = (struct rig_spectrum_scope *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_spectrum_scopes_set self spectrum_scopes ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_scopes_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_spectrum_scopes_set" "', argument " "2"" of type '" "struct rig_spectrum_scope [4]""'"); 
  } 
  arg2 = (struct rig_spectrum_scope *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)4; ++ii) *(struct rig_spectrum_scope *)&arg1->spectrum_scopes[ii] = *((struct rig_spectrum_scope *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_scopes""' of type '""struct rig_spectrum_scope [4]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_spectrum_scopes_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct rig_spectrum_scope *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_spectrum_scopes_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_scopes_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct rig_spectrum_scope *)(struct rig_spectrum_scope *) ((arg1)->spectrum_scopes);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_scope,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_spectrum_modes_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  enum rig_spectrum_mode_e *arg2 = (enum rig_spectrum_mode_e *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_spectrum_modes_set self spectrum_modes ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_modes_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_rig_spectrum_mode_e, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_spectrum_modes_set" "', argument " "2"" of type '" "enum rig_spectrum_mode_e [5]""'"); 
  } 
  arg2 = (enum rig_spectrum_mode_e *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)5; ++ii) *(enum rig_spectrum_mode_e *)&arg1->spectrum_modes[ii] = *((enum rig_spectrum_mode_e *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_modes""' of type '""enum rig_spectrum_mode_e [5]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_spectrum_modes_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum rig_spectrum_mode_e *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_spectrum_modes_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_modes_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (enum rig_spectrum_mode_e *)(enum rig_spectrum_mode_e *) ((arg1)->spectrum_modes);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_mode_e,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_spectrum_spans_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  freq_t *arg2 = (freq_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_spectrum_spans_set self spectrum_spans ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_spans_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_spectrum_spans_set" "', argument " "2"" of type '" "freq_t [20]""'"); 
  } 
  arg2 = (freq_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)20; ++ii) *(freq_t *)&arg1->spectrum_spans[ii] = *((freq_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_spans""' of type '""freq_t [20]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_spectrum_spans_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_spectrum_spans_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_spans_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (freq_t *)(freq_t *) ((arg1)->spectrum_spans);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_double,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_spectrum_avg_modes_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  struct rig_spectrum_avg_mode *arg2 = (struct rig_spectrum_avg_mode *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_spectrum_avg_modes_set self spectrum_avg_modes ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_avg_modes_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_spectrum_avg_modes_set" "', argument " "2"" of type '" "struct rig_spectrum_avg_mode [12]""'"); 
  } 
  arg2 = (struct rig_spectrum_avg_mode *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)12; ++ii) *(struct rig_spectrum_avg_mode *)&arg1->spectrum_avg_modes[ii] = *((struct rig_spectrum_avg_mode *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_avg_modes""' of type '""struct rig_spectrum_avg_mode [12]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_spectrum_avg_modes_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct rig_spectrum_avg_mode *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_spectrum_avg_modes_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_avg_modes_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct rig_spectrum_avg_mode *)(struct rig_spectrum_avg_mode *) ((arg1)->spectrum_avg_modes);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_avg_mode,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_spectrum_attenuator_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int *arg2 = (int *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_spectrum_attenuator_set self spectrum_attenuator ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_attenuator_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_spectrum_attenuator_set" "', argument " "2"" of type '" "int [8]""'"); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(int *)&arg1->spectrum_attenuator[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_attenuator""' of type '""int [8]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_spectrum_attenuator_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_spectrum_attenuator_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_spectrum_attenuator_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int *)(int *) ((arg1)->spectrum_attenuator);
  {
    int i, len, l;
    len = 8;
    
    Tcl_Obj * list = Tcl_NewListObj(len, NULL);
    for(i=0; i < len; i++) {
      Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(result[i]));
    }
    Tcl_SetObjResult(interp, list);
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_cfgparams_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_cfgparams_set self cfgparams ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_cfgparams_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_cfgparams_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->cfgparams = (struct confparams const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_cfgparams_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct confparams *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_cfgparams_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_cfgparams_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (struct confparams *) ((arg1)->cfgparams);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_priv_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_priv_set self priv ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_priv_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_priv_set" "', argument " "2"" of type '" "void const *""'"); 
  }
  if (arg1) (arg1)->priv = (void const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_priv_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_priv_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_priv_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (void *) ((arg1)->priv);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_init_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rig_init_set self rig_init ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_init_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_rig_init_set" "', argument " "2"" of type '" "int (*)(RIG *)""'"); 
    }
  }
  if (arg1) (arg1)->rig_init = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_init_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rig_init_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_init_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *)) ((arg1)->rig_init);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_cleanup_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rig_cleanup_set self rig_cleanup ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_cleanup_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_rig_cleanup_set" "', argument " "2"" of type '" "int (*)(RIG *)""'"); 
    }
  }
  if (arg1) (arg1)->rig_cleanup = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_cleanup_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rig_cleanup_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_cleanup_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *)) ((arg1)->rig_cleanup);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_open_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rig_open_set self rig_open ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_open_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_rig_open_set" "', argument " "2"" of type '" "int (*)(RIG *)""'"); 
    }
  }
  if (arg1) (arg1)->rig_open = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_open_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rig_open_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_open_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *)) ((arg1)->rig_open);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_close_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rig_close_set self rig_close ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_close_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_rig_close_set" "', argument " "2"" of type '" "int (*)(RIG *)""'"); 
    }
  }
  if (arg1) (arg1)->rig_close = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_close_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rig_close_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_close_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *)) ((arg1)->rig_close);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_freq_set self set_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_freq_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_freq_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,freq_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_freq_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t)) ((arg1)->set_freq);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_freq_set self get_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_freq_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_freq_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,freq_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_freq_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t *)) ((arg1)->get_freq);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,rmode_t,pbwidth_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_mode_set self set_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,rmode_t,pbwidth_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,rmode_t,pbwidth_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,rmode_t,pbwidth_t)) ((arg1)->set_mode);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,rmode_t *,pbwidth_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_mode_set self get_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,rmode_t *,pbwidth_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,rmode_t *,pbwidth_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,rmode_t *,pbwidth_t *)) ((arg1)->get_mode);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_vfo_set self set_vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_vfo_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_vfo_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_vfo_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t)) ((arg1)->set_vfo);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_vfo_set self get_vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_vfo_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_vfo_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_vfo_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t *)) ((arg1)->get_vfo);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ptt_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,ptt_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_ptt_set self set_ptt ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ptt_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ptt_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,ptt_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ptt = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ptt_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,ptt_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_ptt_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ptt_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,ptt_t)) ((arg1)->set_ptt);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ptt_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,ptt_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_ptt_set self get_ptt ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ptt_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ptt_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,ptt_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ptt = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ptt_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,ptt_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_ptt_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ptt_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,ptt_t *)) ((arg1)->get_ptt);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_dcd_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,dcd_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_dcd_set self get_dcd ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_dcd_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_dcd_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,dcd_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_dcd = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_dcd_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,dcd_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_dcd_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_dcd_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,dcd_t *)) ((arg1)->get_dcd);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_rptr_shift_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,rptr_shift_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_rptr_shift_set self set_rptr_shift ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_rptr_shift_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_rptr_shift_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,rptr_shift_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_rptr_shift = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_rptr_shift_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,rptr_shift_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_rptr_shift_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_rptr_shift_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,rptr_shift_t)) ((arg1)->set_rptr_shift);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_rptr_shift_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,rptr_shift_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_rptr_shift_set self get_rptr_shift ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_rptr_shift_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_rptr_shift_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,rptr_shift_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_rptr_shift = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_rptr_shift_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,rptr_shift_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_rptr_shift_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_rptr_shift_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,rptr_shift_t *)) ((arg1)->get_rptr_shift);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_rptr_offs_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_rptr_offs_set self set_rptr_offs ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_rptr_offs_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_rptr_offs_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_rptr_offs = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_rptr_offs_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,shortfreq_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_rptr_offs_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_rptr_offs_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t)) ((arg1)->set_rptr_offs);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_rptr_offs_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_rptr_offs_set self get_rptr_offs ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_rptr_offs_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_rptr_offs_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_rptr_offs = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_rptr_offs_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_rptr_offs_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_rptr_offs_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t *)) ((arg1)->get_rptr_offs);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_split_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_split_freq_set self set_split_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_freq_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_split_freq_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_split_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_split_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,freq_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_split_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_freq_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t)) ((arg1)->set_split_freq);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_split_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_split_freq_set self get_split_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_freq_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_split_freq_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_split_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_split_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,freq_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_split_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_freq_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t *)) ((arg1)->get_split_freq);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_split_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,rmode_t,pbwidth_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_split_mode_set self set_split_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_split_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,rmode_t,pbwidth_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_split_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_split_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,rmode_t,pbwidth_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_split_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,rmode_t,pbwidth_t)) ((arg1)->set_split_mode);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_split_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,rmode_t *,pbwidth_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_split_mode_set self get_split_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_split_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,rmode_t *,pbwidth_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_split_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_split_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,rmode_t *,pbwidth_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_split_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,rmode_t *,pbwidth_t *)) ((arg1)->get_split_mode);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_split_freq_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t,rmode_t,pbwidth_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_split_freq_mode_set self set_split_freq_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_freq_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_split_freq_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t,rmode_t,pbwidth_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_split_freq_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_split_freq_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,freq_t,rmode_t,pbwidth_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_split_freq_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_freq_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t,rmode_t,pbwidth_t)) ((arg1)->set_split_freq_mode);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_split_freq_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_split_freq_mode_set self get_split_freq_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_freq_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_split_freq_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_split_freq_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_split_freq_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_split_freq_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_freq_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *)) ((arg1)->get_split_freq_mode);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_split_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,split_t,vfo_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_split_vfo_set self set_split_vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_vfo_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_split_vfo_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,split_t,vfo_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_split_vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_split_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,split_t,vfo_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_split_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_split_vfo_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,split_t,vfo_t)) ((arg1)->set_split_vfo);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_split_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,split_t *,vfo_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_split_vfo_set self get_split_vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_vfo_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_split_vfo_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,split_t *,vfo_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_split_vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_split_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,split_t *,vfo_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_split_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_split_vfo_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,split_t *,vfo_t *)) ((arg1)->get_split_vfo);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_rit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_rit_set self set_rit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_rit_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_rit_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_rit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_rit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,shortfreq_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_rit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_rit_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t)) ((arg1)->set_rit);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_rit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_rit_set self get_rit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_rit_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_rit_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_rit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_rit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_rit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_rit_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t *)) ((arg1)->get_rit);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_xit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_xit_set self set_xit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_xit_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_xit_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_xit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_xit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,shortfreq_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_xit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_xit_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t)) ((arg1)->set_xit);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_xit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_xit_set self get_xit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_xit_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_xit_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_xit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_xit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_xit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_xit_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t *)) ((arg1)->get_xit);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ts_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_ts_set self set_ts ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ts_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ts_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ts = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ts_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,shortfreq_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_ts_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ts_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t)) ((arg1)->set_ts);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ts_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_ts_set self get_ts ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ts_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ts_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,shortfreq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ts = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ts_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,shortfreq_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_ts_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ts_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,shortfreq_t *)) ((arg1)->get_ts);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_dcs_code_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_dcs_code_set self set_dcs_code ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_dcs_code_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_dcs_code_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_dcs_code = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_dcs_code_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,tone_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_dcs_code_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_dcs_code_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t)) ((arg1)->set_dcs_code);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_dcs_code_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_dcs_code_set self get_dcs_code ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_dcs_code_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_dcs_code_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_dcs_code = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_dcs_code_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,tone_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_dcs_code_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_dcs_code_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t *)) ((arg1)->get_dcs_code);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_tone_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_tone_set self set_tone ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_tone_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_tone_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_tone = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_tone_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,tone_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_tone_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_tone_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t)) ((arg1)->set_tone);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_tone_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_tone_set self get_tone ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_tone_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_tone_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_tone = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_tone_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,tone_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_tone_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_tone_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t *)) ((arg1)->get_tone);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ctcss_tone_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_ctcss_tone_set self set_ctcss_tone ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ctcss_tone_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ctcss_tone_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ctcss_tone = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ctcss_tone_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,tone_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_ctcss_tone_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ctcss_tone_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t)) ((arg1)->set_ctcss_tone);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ctcss_tone_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_ctcss_tone_set self get_ctcss_tone ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ctcss_tone_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ctcss_tone_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ctcss_tone = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ctcss_tone_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,tone_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_ctcss_tone_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ctcss_tone_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t *)) ((arg1)->get_ctcss_tone);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_dcs_sql_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_dcs_sql_set self set_dcs_sql ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_dcs_sql_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_dcs_sql_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_dcs_sql = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_dcs_sql_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,tone_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_dcs_sql_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_dcs_sql_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t)) ((arg1)->set_dcs_sql);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_dcs_sql_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_dcs_sql_set self get_dcs_sql ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_dcs_sql_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_dcs_sql_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_dcs_sql = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_dcs_sql_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,tone_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_dcs_sql_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_dcs_sql_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t *)) ((arg1)->get_dcs_sql);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_tone_sql_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_tone_sql_set self set_tone_sql ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_tone_sql_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_tone_sql_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_tone_sql = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_tone_sql_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,tone_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_tone_sql_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_tone_sql_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t)) ((arg1)->set_tone_sql);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_tone_sql_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_tone_sql_set self get_tone_sql ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_tone_sql_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_tone_sql_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_tone_sql = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_tone_sql_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,tone_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_tone_sql_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_tone_sql_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t *)) ((arg1)->get_tone_sql);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ctcss_sql_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_ctcss_sql_set self set_ctcss_sql ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ctcss_sql_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ctcss_sql_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ctcss_sql = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ctcss_sql_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,tone_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_ctcss_sql_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ctcss_sql_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t)) ((arg1)->set_ctcss_sql);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ctcss_sql_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,tone_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_ctcss_sql_set self get_ctcss_sql ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ctcss_sql_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ctcss_sql_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,tone_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ctcss_sql = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ctcss_sql_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,tone_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_ctcss_sql_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ctcss_sql_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,tone_t *)) ((arg1)->get_ctcss_sql);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_power2mW_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,unsigned int *,float,freq_t,rmode_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_power2mW_set self power2mW ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_power2mW_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_power2mW_set" "', argument " "2"" of type '" "int (*)(RIG *,unsigned int *,float,freq_t,rmode_t)""'"); 
    }
  }
  if (arg1) (arg1)->power2mW = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_power2mW_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,unsigned int *,float,freq_t,rmode_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_power2mW_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_power2mW_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,unsigned int *,float,freq_t,rmode_t)) ((arg1)->power2mW);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_mW2power_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,float *,unsigned int,freq_t,rmode_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_mW2power_set self mW2power ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_mW2power_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_mW2power_set" "', argument " "2"" of type '" "int (*)(RIG *,float *,unsigned int,freq_t,rmode_t)""'"); 
    }
  }
  if (arg1) (arg1)->mW2power = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_mW2power_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,float *,unsigned int,freq_t,rmode_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_mW2power_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_mW2power_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,float *,unsigned int,freq_t,rmode_t)) ((arg1)->mW2power);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_powerstat_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,powerstat_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_powerstat_set self set_powerstat ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_powerstat_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_enum_powerstat_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_powerstat_set" "', argument " "2"" of type '" "int (*)(RIG *,powerstat_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_powerstat = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_powerstat_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,powerstat_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_powerstat_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_powerstat_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,powerstat_t)) ((arg1)->set_powerstat);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_enum_powerstat_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_powerstat_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,powerstat_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_powerstat_set self get_powerstat ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_powerstat_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_enum_powerstat_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_powerstat_set" "', argument " "2"" of type '" "int (*)(RIG *,powerstat_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_powerstat = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_powerstat_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,powerstat_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_powerstat_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_powerstat_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,powerstat_t *)) ((arg1)->get_powerstat);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_enum_powerstat_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_reset_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,reset_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_reset_set self reset ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_reset_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_enum_reset_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_reset_set" "', argument " "2"" of type '" "int (*)(RIG *,reset_t)""'"); 
    }
  }
  if (arg1) (arg1)->reset = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_reset_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,reset_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_reset_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_reset_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,reset_t)) ((arg1)->reset);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_enum_reset_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ant_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,ant_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_ant_set self set_ant ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ant_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ant_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,ant_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ant = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ant_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,ant_t,value_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_ant_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ant_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,ant_t,value_t)) ((arg1)->set_ant);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ant_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,ant_t,value_t *,ant_t *,ant_t *,ant_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_ant_set self get_ant ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ant_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ant_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,ant_t,value_t *,ant_t *,ant_t *,ant_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ant = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ant_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,ant_t,value_t *,ant_t *,ant_t *,ant_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_ant_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ant_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,ant_t,value_t *,ant_t *,ant_t *,ant_t *)) ((arg1)->get_ant);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,setting_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_level_set self set_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_level_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_level_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,setting_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,setting_t,value_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_level_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,setting_t,value_t)) ((arg1)->set_level);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,setting_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_level_set self get_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_level_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_level_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,setting_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,setting_t,value_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_level_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,setting_t,value_t *)) ((arg1)->get_level);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,setting_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_func_set self set_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_func_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_func_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,setting_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,setting_t,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_func_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,setting_t,int)) ((arg1)->set_func);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,setting_t,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_func_set self get_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_func_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_func_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,setting_t,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,setting_t,int *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_func_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,setting_t,int *)) ((arg1)->get_func);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,setting_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_parm_set self set_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_parm_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_long_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_parm_set" "', argument " "2"" of type '" "int (*)(RIG *,setting_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,setting_t,value_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_parm_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,setting_t,value_t)) ((arg1)->set_parm);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_long_long_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,setting_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_parm_set self get_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_parm_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_parm_set" "', argument " "2"" of type '" "int (*)(RIG *,setting_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,setting_t,value_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_parm_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,setting_t,value_t *)) ((arg1)->get_parm);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ext_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,hamlib_token_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_ext_level_set self set_ext_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ext_level_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ext_level_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,hamlib_token_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ext_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,hamlib_token_t,value_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_ext_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ext_level_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,hamlib_token_t,value_t)) ((arg1)->set_ext_level);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ext_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,hamlib_token_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_ext_level_set self get_ext_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ext_level_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ext_level_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,hamlib_token_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ext_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,hamlib_token_t,value_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_ext_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ext_level_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,hamlib_token_t,value_t *)) ((arg1)->get_ext_level);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ext_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,hamlib_token_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_ext_func_set self set_ext_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ext_func_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ext_func_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,hamlib_token_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ext_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,hamlib_token_t,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_ext_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ext_func_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,hamlib_token_t,int)) ((arg1)->set_ext_func);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ext_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,hamlib_token_t,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_ext_func_set self get_ext_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ext_func_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ext_func_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,hamlib_token_t,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ext_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,hamlib_token_t,int *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_ext_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ext_func_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,hamlib_token_t,int *)) ((arg1)->get_ext_func);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_long_p_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ext_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,hamlib_token_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_ext_parm_set self set_ext_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ext_parm_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_ext_parm_set" "', argument " "2"" of type '" "int (*)(RIG *,hamlib_token_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_ext_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,hamlib_token_t,value_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_ext_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_ext_parm_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,hamlib_token_t,value_t)) ((arg1)->set_ext_parm);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_long_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ext_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,hamlib_token_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_ext_parm_set self get_ext_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ext_parm_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_ext_parm_set" "', argument " "2"" of type '" "int (*)(RIG *,hamlib_token_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_ext_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,hamlib_token_t,value_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_ext_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_ext_parm_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,hamlib_token_t,value_t *)) ((arg1)->get_ext_parm);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_long_p_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_conf_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,hamlib_token_t,char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_conf_set self set_conf ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_conf_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_long_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_conf_set" "', argument " "2"" of type '" "int (*)(RIG *,hamlib_token_t,char const *)""'"); 
    }
  }
  if (arg1) (arg1)->set_conf = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_conf_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,hamlib_token_t,char const *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_conf_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_conf_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,hamlib_token_t,char const *)) ((arg1)->set_conf);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_long_p_q_const__char__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_conf_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,hamlib_token_t,char *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_conf_set self get_conf ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_conf_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_long_p_char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_conf_set" "', argument " "2"" of type '" "int (*)(RIG *,hamlib_token_t,char *)""'"); 
    }
  }
  if (arg1) (arg1)->get_conf = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_conf_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,hamlib_token_t,char *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_conf_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_conf_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,hamlib_token_t,char *)) ((arg1)->get_conf);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_long_p_char__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_send_dtmf_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_send_dtmf_set self send_dtmf ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_send_dtmf_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_send_dtmf_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,char const *)""'"); 
    }
  }
  if (arg1) (arg1)->send_dtmf = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_send_dtmf_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,char const *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_send_dtmf_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_send_dtmf_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,char const *)) ((arg1)->send_dtmf);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_recv_dtmf_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,char *,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_recv_dtmf_set self recv_dtmf ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_recv_dtmf_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_recv_dtmf_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,char *,int *)""'"); 
    }
  }
  if (arg1) (arg1)->recv_dtmf = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_recv_dtmf_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,char *,int *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_recv_dtmf_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_recv_dtmf_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,char *,int *)) ((arg1)->recv_dtmf);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_send_morse_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_send_morse_set self send_morse ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_send_morse_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_send_morse_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,char const *)""'"); 
    }
  }
  if (arg1) (arg1)->send_morse = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_send_morse_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,char const *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_send_morse_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_send_morse_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,char const *)) ((arg1)->send_morse);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_stop_morse_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_stop_morse_set self stop_morse ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_stop_morse_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_stop_morse_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t)""'"); 
    }
  }
  if (arg1) (arg1)->stop_morse = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_stop_morse_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_stop_morse_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_stop_morse_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t)) ((arg1)->stop_morse);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_wait_morse_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_wait_morse_set self wait_morse ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_wait_morse_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_wait_morse_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t)""'"); 
    }
  }
  if (arg1) (arg1)->wait_morse = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_wait_morse_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_wait_morse_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_wait_morse_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t)) ((arg1)->wait_morse);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_send_voice_mem_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_send_voice_mem_set self send_voice_mem ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_send_voice_mem_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_send_voice_mem_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->send_voice_mem = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_send_voice_mem_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_send_voice_mem_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_send_voice_mem_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,int)) ((arg1)->send_voice_mem);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_stop_voice_mem_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_stop_voice_mem_set self stop_voice_mem ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_stop_voice_mem_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_stop_voice_mem_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t)""'"); 
    }
  }
  if (arg1) (arg1)->stop_voice_mem = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_stop_voice_mem_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_stop_voice_mem_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_stop_voice_mem_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t)) ((arg1)->stop_voice_mem);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_bank_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_bank_set self set_bank ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_bank_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_bank_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_bank = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_bank_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_bank_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_bank_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,int)) ((arg1)->set_bank);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_mem_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_mem_set self set_mem ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_mem_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_mem_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_mem = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_mem_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_mem_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_mem_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,int)) ((arg1)->set_mem);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_mem_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_mem_set self get_mem ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_mem_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_mem_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_mem = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_mem_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,int *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_mem_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_mem_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,int *)) ((arg1)->get_mem);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_vfo_op_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,vfo_op_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_vfo_op_set self vfo_op ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_vfo_op_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_vfo_op_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,vfo_op_t)""'"); 
    }
  }
  if (arg1) (arg1)->vfo_op = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_vfo_op_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,vfo_op_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_vfo_op_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_vfo_op_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,vfo_op_t)) ((arg1)->vfo_op);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_scan_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,scan_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_scan_set self scan ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_scan_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_scan_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,scan_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->scan = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_scan_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,scan_t,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_scan_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_scan_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,scan_t,int)) ((arg1)->scan);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_trn_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_trn_set self set_trn ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_trn_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_trn_set" "', argument " "2"" of type '" "int (*)(RIG *,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_trn = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_trn_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_trn_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_trn_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int)) ((arg1)->set_trn);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_trn_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_trn_set self get_trn ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_trn_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_trn_set" "', argument " "2"" of type '" "int (*)(RIG *,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_trn = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_trn_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,int *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_trn_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_trn_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int *)) ((arg1)->get_trn);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_decode_event_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_decode_event_set self decode_event ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_decode_event_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_decode_event_set" "', argument " "2"" of type '" "int (*)(RIG *)""'"); 
    }
  }
  if (arg1) (arg1)->decode_event = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_decode_event_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_decode_event_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_decode_event_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *)) ((arg1)->decode_event);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_channel_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,channel_t const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_channel_set self set_channel ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_channel_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_channel_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,channel_t const *)""'"); 
    }
  }
  if (arg1) (arg1)->set_channel = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_channel_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,channel_t const *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_channel_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_channel_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,channel_t const *)) ((arg1)->set_channel);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_channel_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,channel_t *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_channel_set self get_channel ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_channel_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_channel_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,channel_t *,int)""'"); 
    }
  }
  if (arg1) (arg1)->get_channel = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_channel_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,channel_t *,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_channel_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_channel_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,channel_t *,int)) ((arg1)->get_channel);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_info_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  char *(*arg2)(RIG *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_info_set self get_info ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_info_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig__p_char);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_info_set" "', argument " "2"" of type '" "char const *(*)(RIG *)""'"); 
    }
  }
  if (arg1) (arg1)->get_info = (char const *(*)(RIG *))arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_info_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *(*result)(RIG *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_info_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_info_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *(*)(RIG *)) ((arg1)->get_info);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig__p_char));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_chan_all_cb_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,chan_cb_t,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_chan_all_cb_set self set_chan_all_cb ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_chan_all_cb_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_chan_all_cb_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,chan_cb_t,void *)""'"); 
    }
  }
  if (arg1) (arg1)->set_chan_all_cb = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_chan_all_cb_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,chan_cb_t,void *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_chan_all_cb_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_chan_all_cb_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,chan_cb_t,void *)) ((arg1)->set_chan_all_cb);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_chan_all_cb_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,chan_cb_t,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_chan_all_cb_set self get_chan_all_cb ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_chan_all_cb_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_chan_all_cb_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,chan_cb_t,void *)""'"); 
    }
  }
  if (arg1) (arg1)->get_chan_all_cb = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_chan_all_cb_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,chan_cb_t,void *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_chan_all_cb_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_chan_all_cb_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,chan_cb_t,void *)) ((arg1)->get_chan_all_cb);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_mem_all_cb_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_mem_all_cb_set self set_mem_all_cb ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_mem_all_cb_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_mem_all_cb_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *)""'"); 
    }
  }
  if (arg1) (arg1)->set_mem_all_cb = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_mem_all_cb_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_mem_all_cb_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_mem_all_cb_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *)) ((arg1)->set_mem_all_cb);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_mem_all_cb_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_mem_all_cb_set self get_mem_all_cb ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_mem_all_cb_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_mem_all_cb_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *)""'"); 
    }
  }
  if (arg1) (arg1)->get_mem_all_cb = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_mem_all_cb_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_mem_all_cb_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_mem_all_cb_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *)) ((arg1)->get_mem_all_cb);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_vfo_opt_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_vfo_opt_set self set_vfo_opt ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_vfo_opt_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_vfo_opt_set" "', argument " "2"" of type '" "int (*)(RIG *,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_vfo_opt = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_vfo_opt_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_vfo_opt_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_vfo_opt_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int)) ((arg1)->set_vfo_opt);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_get_vfo_info_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *,split_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_rig_get_vfo_info_set self rig_get_vfo_info ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_get_vfo_info_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_rig_get_vfo_info_set" "', argument " "2"" of type '" "int (*)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *,split_t *)""'"); 
    }
  }
  if (arg1) (arg1)->rig_get_vfo_info = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_rig_get_vfo_info_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *,split_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_rig_get_vfo_info_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_rig_get_vfo_info_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *,split_t *)) ((arg1)->rig_get_vfo_info);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_clock_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int,int,int,int,int,int,double,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_clock_set self set_clock ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_clock_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_clock_set" "', argument " "2"" of type '" "int (*)(RIG *,int,int,int,int,int,int,double,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_clock = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_clock_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,int,int,int,int,int,int,double,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_clock_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_clock_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int,int,int,int,int,int,double,int)) ((arg1)->set_clock);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_clock_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int *,int *,int *,int *,int *,int *,double *,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_clock_set self get_clock ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_clock_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_clock_set" "', argument " "2"" of type '" "int (*)(RIG *,int *,int *,int *,int *,int *,int *,double *,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_clock = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_clock_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,int *,int *,int *,int *,int *,int *,double *,int *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_clock_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_clock_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int *,int *,int *,int *,int *,int *,double *,int *)) ((arg1)->get_clock);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_clone_combo_set_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_clone_combo_set_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_clone_combo_set_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->clone_combo_set);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_clone_combo_get_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_clone_combo_get_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_clone_combo_get_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->clone_combo_get);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_async_data_supported_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_async_data_supported_set self async_data_supported ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_async_data_supported_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_async_data_supported_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->async_data_supported = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_async_data_supported_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_async_data_supported_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_async_data_supported_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int) ((arg1)->async_data_supported);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_read_frame_direct_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,size_t,unsigned char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_read_frame_direct_set self read_frame_direct ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_read_frame_direct_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_read_frame_direct_set" "', argument " "2"" of type '" "int (*)(RIG *,size_t,unsigned char const *)""'"); 
    }
  }
  if (arg1) (arg1)->read_frame_direct = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_read_frame_direct_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,size_t,unsigned char const *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_read_frame_direct_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_read_frame_direct_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,size_t,unsigned char const *)) ((arg1)->read_frame_direct);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_is_async_frame_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,size_t,unsigned char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_is_async_frame_set self is_async_frame ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_is_async_frame_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_is_async_frame_set" "', argument " "2"" of type '" "int (*)(RIG *,size_t,unsigned char const *)""'"); 
    }
  }
  if (arg1) (arg1)->is_async_frame = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_is_async_frame_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,size_t,unsigned char const *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_is_async_frame_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_is_async_frame_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,size_t,unsigned char const *)) ((arg1)->is_async_frame);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_process_async_frame_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,size_t,unsigned char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_process_async_frame_set self process_async_frame ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_process_async_frame_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_process_async_frame_set" "', argument " "2"" of type '" "int (*)(RIG *,size_t,unsigned char const *)""'"); 
    }
  }
  if (arg1) (arg1)->process_async_frame = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_process_async_frame_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,size_t,unsigned char const *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_process_async_frame_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_process_async_frame_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,size_t,unsigned char const *)) ((arg1)->process_async_frame);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_hamlib_check_rig_caps_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_hamlib_check_rig_caps_set self hamlib_check_rig_caps ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_hamlib_check_rig_caps_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_caps_hamlib_check_rig_caps_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->hamlib_check_rig_caps);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->hamlib_check_rig_caps = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->hamlib_check_rig_caps = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_hamlib_check_rig_caps_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_hamlib_check_rig_caps_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_hamlib_check_rig_caps_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (char *) ((arg1)->hamlib_check_rig_caps);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_conf2_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,hamlib_token_t,char *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_conf2_set self get_conf2 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_conf2_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_long_p_char_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_conf2_set" "', argument " "2"" of type '" "int (*)(RIG *,hamlib_token_t,char *,int)""'"); 
    }
  }
  if (arg1) (arg1)->get_conf2 = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_conf2_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,hamlib_token_t,char *,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_conf2_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_conf2_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,hamlib_token_t,char *,int)) ((arg1)->get_conf2);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_long_p_char_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_password_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_password_set self password ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_password_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_password_set" "', argument " "2"" of type '" "int (*)(RIG *,char const *)""'"); 
    }
  }
  if (arg1) (arg1)->password = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_password_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,char const *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_password_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_password_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,char const *)) ((arg1)->password);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_q_const__char__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_lock_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_set_lock_mode_set self set_lock_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_lock_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_set_lock_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_lock_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_set_lock_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_set_lock_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_set_lock_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int)) ((arg1)->set_lock_mode);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_lock_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  int (*arg2)(RIG *,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_get_lock_mode_set self get_lock_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_lock_mode_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_caps_get_lock_mode_set" "', argument " "2"" of type '" "int (*)(RIG *,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_lock_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_get_lock_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(RIG *,int *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_get_lock_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_get_lock_mode_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (int (*)(RIG *,int *)) ((arg1)->get_lock_mode);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_timeout_retry_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  short arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  short val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_timeout_retry_set self timeout_retry ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_timeout_retry_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_short SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_timeout_retry_set" "', argument " "2"" of type '" "short""'");
  } 
  arg2 = (short)(val2);
  if (arg1) (arg1)->timeout_retry = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_timeout_retry_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  short result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_timeout_retry_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_timeout_retry_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (short) ((arg1)->timeout_retry);
  Tcl_SetObjResult(interp,SWIG_From_short((short)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_morse_qsize_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  short arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  short val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_caps_morse_qsize_set self morse_qsize ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_morse_qsize_set" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  ecode2 = SWIG_AsVal_short SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_caps_morse_qsize_set" "', argument " "2"" of type '" "short""'");
  } 
  arg2 = (short)(val2);
  if (arg1) (arg1)->morse_qsize = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_caps_morse_qsize_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  short result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_caps_morse_qsize_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_caps_morse_qsize_get" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  result = (short) ((arg1)->morse_qsize);
  Tcl_SetObjResult(interp,SWIG_From_short((short)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_rig_caps(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_rig_caps ") == TCL_ERROR) SWIG_fail;
  result = (struct rig_caps *)calloc(1, sizeof(struct rig_caps));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_caps,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_rig_caps(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_rig_caps self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_caps, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rig_caps" "', argument " "1"" of type '" "struct rig_caps *""'"); 
  }
  arg1 = (struct rig_caps *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_rig_caps(void *obj) {
struct rig_caps *arg1 = (struct rig_caps *) obj;
free((char *) arg1);
}
static swig_method swig_rig_caps_methods[] = {
    {0,0}
};
static swig_attribute swig_rig_caps_attributes[] = {
    { "-rig_model",_wrap_rig_caps_rig_model_get, _wrap_rig_caps_rig_model_set},
    { "-model_name",_wrap_rig_caps_model_name_get, 0 },
    { "-mfg_name",_wrap_rig_caps_mfg_name_get, 0 },
    { "-version",_wrap_rig_caps_version_get, 0 },
    { "-copyright",_wrap_rig_caps_copyright_get, 0 },
    { "-status",_wrap_rig_caps_status_get, _wrap_rig_caps_status_set},
    { "-rig_type",_wrap_rig_caps_rig_type_get, _wrap_rig_caps_rig_type_set},
    { "-ptt_type",_wrap_rig_caps_ptt_type_get, _wrap_rig_caps_ptt_type_set},
    { "-dcd_type",_wrap_rig_caps_dcd_type_get, _wrap_rig_caps_dcd_type_set},
    { "-port_type",_wrap_rig_caps_port_type_get, _wrap_rig_caps_port_type_set},
    { "-serial_rate_min",_wrap_rig_caps_serial_rate_min_get, _wrap_rig_caps_serial_rate_min_set},
    { "-serial_rate_max",_wrap_rig_caps_serial_rate_max_get, _wrap_rig_caps_serial_rate_max_set},
    { "-serial_data_bits",_wrap_rig_caps_serial_data_bits_get, _wrap_rig_caps_serial_data_bits_set},
    { "-serial_stop_bits",_wrap_rig_caps_serial_stop_bits_get, _wrap_rig_caps_serial_stop_bits_set},
    { "-serial_parity",_wrap_rig_caps_serial_parity_get, _wrap_rig_caps_serial_parity_set},
    { "-serial_handshake",_wrap_rig_caps_serial_handshake_get, _wrap_rig_caps_serial_handshake_set},
    { "-write_delay",_wrap_rig_caps_write_delay_get, _wrap_rig_caps_write_delay_set},
    { "-post_write_delay",_wrap_rig_caps_post_write_delay_get, _wrap_rig_caps_post_write_delay_set},
    { "-timeout",_wrap_rig_caps_timeout_get, _wrap_rig_caps_timeout_set},
    { "-retry",_wrap_rig_caps_retry_get, _wrap_rig_caps_retry_set},
    { "-has_get_func",_wrap_rig_caps_has_get_func_get, _wrap_rig_caps_has_get_func_set},
    { "-has_set_func",_wrap_rig_caps_has_set_func_get, _wrap_rig_caps_has_set_func_set},
    { "-has_get_level",_wrap_rig_caps_has_get_level_get, _wrap_rig_caps_has_get_level_set},
    { "-has_set_level",_wrap_rig_caps_has_set_level_get, _wrap_rig_caps_has_set_level_set},
    { "-has_get_parm",_wrap_rig_caps_has_get_parm_get, _wrap_rig_caps_has_get_parm_set},
    { "-has_set_parm",_wrap_rig_caps_has_set_parm_get, _wrap_rig_caps_has_set_parm_set},
    { "-level_gran",_wrap_rig_caps_level_gran_get, _wrap_rig_caps_level_gran_set},
    { "-parm_gran",_wrap_rig_caps_parm_gran_get, _wrap_rig_caps_parm_gran_set},
    { "-extparms",_wrap_rig_caps_extparms_get, _wrap_rig_caps_extparms_set},
    { "-extlevels",_wrap_rig_caps_extlevels_get, _wrap_rig_caps_extlevels_set},
    { "-extfuncs",_wrap_rig_caps_extfuncs_get, _wrap_rig_caps_extfuncs_set},
    { "-ext_tokens",_wrap_rig_caps_ext_tokens_get, _wrap_rig_caps_ext_tokens_set},
    { "-ctcss_list",_wrap_rig_caps_ctcss_list_get, _wrap_rig_caps_ctcss_list_set},
    { "-dcs_list",_wrap_rig_caps_dcs_list_get, _wrap_rig_caps_dcs_list_set},
    { "-preamp",_wrap_rig_caps_preamp_get, _wrap_rig_caps_preamp_set},
    { "-attenuator",_wrap_rig_caps_attenuator_get, _wrap_rig_caps_attenuator_set},
    { "-max_rit",_wrap_rig_caps_max_rit_get, _wrap_rig_caps_max_rit_set},
    { "-max_xit",_wrap_rig_caps_max_xit_get, _wrap_rig_caps_max_xit_set},
    { "-max_ifshift",_wrap_rig_caps_max_ifshift_get, _wrap_rig_caps_max_ifshift_set},
    { "-agc_level_count",_wrap_rig_caps_agc_level_count_get, _wrap_rig_caps_agc_level_count_set},
    { "-agc_levels",_wrap_rig_caps_agc_levels_get, _wrap_rig_caps_agc_levels_set},
    { "-announces",_wrap_rig_caps_announces_get, _wrap_rig_caps_announces_set},
    { "-vfo_ops",_wrap_rig_caps_vfo_ops_get, _wrap_rig_caps_vfo_ops_set},
    { "-scan_ops",_wrap_rig_caps_scan_ops_get, _wrap_rig_caps_scan_ops_set},
    { "-targetable_vfo",_wrap_rig_caps_targetable_vfo_get, _wrap_rig_caps_targetable_vfo_set},
    { "-transceive",_wrap_rig_caps_transceive_get, _wrap_rig_caps_transceive_set},
    { "-bank_qty",_wrap_rig_caps_bank_qty_get, _wrap_rig_caps_bank_qty_set},
    { "-chan_desc_sz",_wrap_rig_caps_chan_desc_sz_get, _wrap_rig_caps_chan_desc_sz_set},
    { "-chan_list",_wrap_rig_caps_chan_list_get, _wrap_rig_caps_chan_list_set},
    { "-rx_range_list1",_wrap_rig_caps_rx_range_list1_get, _wrap_rig_caps_rx_range_list1_set},
    { "-tx_range_list1",_wrap_rig_caps_tx_range_list1_get, _wrap_rig_caps_tx_range_list1_set},
    { "-rx_range_list2",_wrap_rig_caps_rx_range_list2_get, _wrap_rig_caps_rx_range_list2_set},
    { "-tx_range_list2",_wrap_rig_caps_tx_range_list2_get, _wrap_rig_caps_tx_range_list2_set},
    { "-rx_range_list3",_wrap_rig_caps_rx_range_list3_get, _wrap_rig_caps_rx_range_list3_set},
    { "-tx_range_list3",_wrap_rig_caps_tx_range_list3_get, _wrap_rig_caps_tx_range_list3_set},
    { "-rx_range_list4",_wrap_rig_caps_rx_range_list4_get, _wrap_rig_caps_rx_range_list4_set},
    { "-tx_range_list4",_wrap_rig_caps_tx_range_list4_get, _wrap_rig_caps_tx_range_list4_set},
    { "-rx_range_list5",_wrap_rig_caps_rx_range_list5_get, _wrap_rig_caps_rx_range_list5_set},
    { "-tx_range_list5",_wrap_rig_caps_tx_range_list5_get, _wrap_rig_caps_tx_range_list5_set},
    { "-tuning_steps",_wrap_rig_caps_tuning_steps_get, _wrap_rig_caps_tuning_steps_set},
    { "-filters",_wrap_rig_caps_filters_get, _wrap_rig_caps_filters_set},
    { "-str_cal",_wrap_rig_caps_str_cal_get, _wrap_rig_caps_str_cal_set},
    { "-swr_cal",_wrap_rig_caps_swr_cal_get, _wrap_rig_caps_swr_cal_set},
    { "-alc_cal",_wrap_rig_caps_alc_cal_get, _wrap_rig_caps_alc_cal_set},
    { "-rfpower_meter_cal",_wrap_rig_caps_rfpower_meter_cal_get, _wrap_rig_caps_rfpower_meter_cal_set},
    { "-comp_meter_cal",_wrap_rig_caps_comp_meter_cal_get, _wrap_rig_caps_comp_meter_cal_set},
    { "-vd_meter_cal",_wrap_rig_caps_vd_meter_cal_get, _wrap_rig_caps_vd_meter_cal_set},
    { "-id_meter_cal",_wrap_rig_caps_id_meter_cal_get, _wrap_rig_caps_id_meter_cal_set},
    { "-spectrum_scopes",_wrap_rig_caps_spectrum_scopes_get, _wrap_rig_caps_spectrum_scopes_set},
    { "-spectrum_modes",_wrap_rig_caps_spectrum_modes_get, _wrap_rig_caps_spectrum_modes_set},
    { "-spectrum_spans",_wrap_rig_caps_spectrum_spans_get, _wrap_rig_caps_spectrum_spans_set},
    { "-spectrum_avg_modes",_wrap_rig_caps_spectrum_avg_modes_get, _wrap_rig_caps_spectrum_avg_modes_set},
    { "-spectrum_attenuator",_wrap_rig_caps_spectrum_attenuator_get, _wrap_rig_caps_spectrum_attenuator_set},
    { "-cfgparams",_wrap_rig_caps_cfgparams_get, _wrap_rig_caps_cfgparams_set},
    { "-priv",_wrap_rig_caps_priv_get, _wrap_rig_caps_priv_set},
    { "-rig_init",_wrap_rig_caps_rig_init_get, _wrap_rig_caps_rig_init_set},
    { "-rig_cleanup",_wrap_rig_caps_rig_cleanup_get, _wrap_rig_caps_rig_cleanup_set},
    { "-rig_open",_wrap_rig_caps_rig_open_get, _wrap_rig_caps_rig_open_set},
    { "-rig_close",_wrap_rig_caps_rig_close_get, _wrap_rig_caps_rig_close_set},
    { "-set_freq",_wrap_rig_caps_set_freq_get, _wrap_rig_caps_set_freq_set},
    { "-get_freq",_wrap_rig_caps_get_freq_get, _wrap_rig_caps_get_freq_set},
    { "-set_mode",_wrap_rig_caps_set_mode_get, _wrap_rig_caps_set_mode_set},
    { "-get_mode",_wrap_rig_caps_get_mode_get, _wrap_rig_caps_get_mode_set},
    { "-set_vfo",_wrap_rig_caps_set_vfo_get, _wrap_rig_caps_set_vfo_set},
    { "-get_vfo",_wrap_rig_caps_get_vfo_get, _wrap_rig_caps_get_vfo_set},
    { "-set_ptt",_wrap_rig_caps_set_ptt_get, _wrap_rig_caps_set_ptt_set},
    { "-get_ptt",_wrap_rig_caps_get_ptt_get, _wrap_rig_caps_get_ptt_set},
    { "-get_dcd",_wrap_rig_caps_get_dcd_get, _wrap_rig_caps_get_dcd_set},
    { "-set_rptr_shift",_wrap_rig_caps_set_rptr_shift_get, _wrap_rig_caps_set_rptr_shift_set},
    { "-get_rptr_shift",_wrap_rig_caps_get_rptr_shift_get, _wrap_rig_caps_get_rptr_shift_set},
    { "-set_rptr_offs",_wrap_rig_caps_set_rptr_offs_get, _wrap_rig_caps_set_rptr_offs_set},
    { "-get_rptr_offs",_wrap_rig_caps_get_rptr_offs_get, _wrap_rig_caps_get_rptr_offs_set},
    { "-set_split_freq",_wrap_rig_caps_set_split_freq_get, _wrap_rig_caps_set_split_freq_set},
    { "-get_split_freq",_wrap_rig_caps_get_split_freq_get, _wrap_rig_caps_get_split_freq_set},
    { "-set_split_mode",_wrap_rig_caps_set_split_mode_get, _wrap_rig_caps_set_split_mode_set},
    { "-get_split_mode",_wrap_rig_caps_get_split_mode_get, _wrap_rig_caps_get_split_mode_set},
    { "-set_split_freq_mode",_wrap_rig_caps_set_split_freq_mode_get, _wrap_rig_caps_set_split_freq_mode_set},
    { "-get_split_freq_mode",_wrap_rig_caps_get_split_freq_mode_get, _wrap_rig_caps_get_split_freq_mode_set},
    { "-set_split_vfo",_wrap_rig_caps_set_split_vfo_get, _wrap_rig_caps_set_split_vfo_set},
    { "-get_split_vfo",_wrap_rig_caps_get_split_vfo_get, _wrap_rig_caps_get_split_vfo_set},
    { "-set_rit",_wrap_rig_caps_set_rit_get, _wrap_rig_caps_set_rit_set},
    { "-get_rit",_wrap_rig_caps_get_rit_get, _wrap_rig_caps_get_rit_set},
    { "-set_xit",_wrap_rig_caps_set_xit_get, _wrap_rig_caps_set_xit_set},
    { "-get_xit",_wrap_rig_caps_get_xit_get, _wrap_rig_caps_get_xit_set},
    { "-set_ts",_wrap_rig_caps_set_ts_get, _wrap_rig_caps_set_ts_set},
    { "-get_ts",_wrap_rig_caps_get_ts_get, _wrap_rig_caps_get_ts_set},
    { "-set_dcs_code",_wrap_rig_caps_set_dcs_code_get, _wrap_rig_caps_set_dcs_code_set},
    { "-get_dcs_code",_wrap_rig_caps_get_dcs_code_get, _wrap_rig_caps_get_dcs_code_set},
    { "-set_tone",_wrap_rig_caps_set_tone_get, _wrap_rig_caps_set_tone_set},
    { "-get_tone",_wrap_rig_caps_get_tone_get, _wrap_rig_caps_get_tone_set},
    { "-set_ctcss_tone",_wrap_rig_caps_set_ctcss_tone_get, _wrap_rig_caps_set_ctcss_tone_set},
    { "-get_ctcss_tone",_wrap_rig_caps_get_ctcss_tone_get, _wrap_rig_caps_get_ctcss_tone_set},
    { "-set_dcs_sql",_wrap_rig_caps_set_dcs_sql_get, _wrap_rig_caps_set_dcs_sql_set},
    { "-get_dcs_sql",_wrap_rig_caps_get_dcs_sql_get, _wrap_rig_caps_get_dcs_sql_set},
    { "-set_tone_sql",_wrap_rig_caps_set_tone_sql_get, _wrap_rig_caps_set_tone_sql_set},
    { "-get_tone_sql",_wrap_rig_caps_get_tone_sql_get, _wrap_rig_caps_get_tone_sql_set},
    { "-set_ctcss_sql",_wrap_rig_caps_set_ctcss_sql_get, _wrap_rig_caps_set_ctcss_sql_set},
    { "-get_ctcss_sql",_wrap_rig_caps_get_ctcss_sql_get, _wrap_rig_caps_get_ctcss_sql_set},
    { "-power2mW",_wrap_rig_caps_power2mW_get, _wrap_rig_caps_power2mW_set},
    { "-mW2power",_wrap_rig_caps_mW2power_get, _wrap_rig_caps_mW2power_set},
    { "-set_powerstat",_wrap_rig_caps_set_powerstat_get, _wrap_rig_caps_set_powerstat_set},
    { "-get_powerstat",_wrap_rig_caps_get_powerstat_get, _wrap_rig_caps_get_powerstat_set},
    { "-reset",_wrap_rig_caps_reset_get, _wrap_rig_caps_reset_set},
    { "-set_ant",_wrap_rig_caps_set_ant_get, _wrap_rig_caps_set_ant_set},
    { "-get_ant",_wrap_rig_caps_get_ant_get, _wrap_rig_caps_get_ant_set},
    { "-set_level",_wrap_rig_caps_set_level_get, _wrap_rig_caps_set_level_set},
    { "-get_level",_wrap_rig_caps_get_level_get, _wrap_rig_caps_get_level_set},
    { "-set_func",_wrap_rig_caps_set_func_get, _wrap_rig_caps_set_func_set},
    { "-get_func",_wrap_rig_caps_get_func_get, _wrap_rig_caps_get_func_set},
    { "-set_parm",_wrap_rig_caps_set_parm_get, _wrap_rig_caps_set_parm_set},
    { "-get_parm",_wrap_rig_caps_get_parm_get, _wrap_rig_caps_get_parm_set},
    { "-set_ext_level",_wrap_rig_caps_set_ext_level_get, _wrap_rig_caps_set_ext_level_set},
    { "-get_ext_level",_wrap_rig_caps_get_ext_level_get, _wrap_rig_caps_get_ext_level_set},
    { "-set_ext_func",_wrap_rig_caps_set_ext_func_get, _wrap_rig_caps_set_ext_func_set},
    { "-get_ext_func",_wrap_rig_caps_get_ext_func_get, _wrap_rig_caps_get_ext_func_set},
    { "-set_ext_parm",_wrap_rig_caps_set_ext_parm_get, _wrap_rig_caps_set_ext_parm_set},
    { "-get_ext_parm",_wrap_rig_caps_get_ext_parm_get, _wrap_rig_caps_get_ext_parm_set},
    { "-set_conf",_wrap_rig_caps_set_conf_get, _wrap_rig_caps_set_conf_set},
    { "-get_conf",_wrap_rig_caps_get_conf_get, _wrap_rig_caps_get_conf_set},
    { "-send_dtmf",_wrap_rig_caps_send_dtmf_get, _wrap_rig_caps_send_dtmf_set},
    { "-recv_dtmf",_wrap_rig_caps_recv_dtmf_get, _wrap_rig_caps_recv_dtmf_set},
    { "-send_morse",_wrap_rig_caps_send_morse_get, _wrap_rig_caps_send_morse_set},
    { "-stop_morse",_wrap_rig_caps_stop_morse_get, _wrap_rig_caps_stop_morse_set},
    { "-wait_morse",_wrap_rig_caps_wait_morse_get, _wrap_rig_caps_wait_morse_set},
    { "-send_voice_mem",_wrap_rig_caps_send_voice_mem_get, _wrap_rig_caps_send_voice_mem_set},
    { "-stop_voice_mem",_wrap_rig_caps_stop_voice_mem_get, _wrap_rig_caps_stop_voice_mem_set},
    { "-set_bank",_wrap_rig_caps_set_bank_get, _wrap_rig_caps_set_bank_set},
    { "-set_mem",_wrap_rig_caps_set_mem_get, _wrap_rig_caps_set_mem_set},
    { "-get_mem",_wrap_rig_caps_get_mem_get, _wrap_rig_caps_get_mem_set},
    { "-vfo_op",_wrap_rig_caps_vfo_op_get, _wrap_rig_caps_vfo_op_set},
    { "-scan",_wrap_rig_caps_scan_get, _wrap_rig_caps_scan_set},
    { "-set_trn",_wrap_rig_caps_set_trn_get, _wrap_rig_caps_set_trn_set},
    { "-get_trn",_wrap_rig_caps_get_trn_get, _wrap_rig_caps_get_trn_set},
    { "-decode_event",_wrap_rig_caps_decode_event_get, _wrap_rig_caps_decode_event_set},
    { "-set_channel",_wrap_rig_caps_set_channel_get, _wrap_rig_caps_set_channel_set},
    { "-get_channel",_wrap_rig_caps_get_channel_get, _wrap_rig_caps_get_channel_set},
    { "-get_info",_wrap_rig_caps_get_info_get, _wrap_rig_caps_get_info_set},
    { "-set_chan_all_cb",_wrap_rig_caps_set_chan_all_cb_get, _wrap_rig_caps_set_chan_all_cb_set},
    { "-get_chan_all_cb",_wrap_rig_caps_get_chan_all_cb_get, _wrap_rig_caps_get_chan_all_cb_set},
    { "-set_mem_all_cb",_wrap_rig_caps_set_mem_all_cb_get, _wrap_rig_caps_set_mem_all_cb_set},
    { "-get_mem_all_cb",_wrap_rig_caps_get_mem_all_cb_get, _wrap_rig_caps_get_mem_all_cb_set},
    { "-set_vfo_opt",_wrap_rig_caps_set_vfo_opt_get, _wrap_rig_caps_set_vfo_opt_set},
    { "-rig_get_vfo_info",_wrap_rig_caps_rig_get_vfo_info_get, _wrap_rig_caps_rig_get_vfo_info_set},
    { "-set_clock",_wrap_rig_caps_set_clock_get, _wrap_rig_caps_set_clock_set},
    { "-get_clock",_wrap_rig_caps_get_clock_get, _wrap_rig_caps_get_clock_set},
    { "-clone_combo_set",_wrap_rig_caps_clone_combo_set_get, 0 },
    { "-clone_combo_get",_wrap_rig_caps_clone_combo_get_get, 0 },
    { "-async_data_supported",_wrap_rig_caps_async_data_supported_get, _wrap_rig_caps_async_data_supported_set},
    { "-read_frame_direct",_wrap_rig_caps_read_frame_direct_get, _wrap_rig_caps_read_frame_direct_set},
    { "-is_async_frame",_wrap_rig_caps_is_async_frame_get, _wrap_rig_caps_is_async_frame_set},
    { "-process_async_frame",_wrap_rig_caps_process_async_frame_get, _wrap_rig_caps_process_async_frame_set},
    { "-hamlib_check_rig_caps",_wrap_rig_caps_hamlib_check_rig_caps_get, _wrap_rig_caps_hamlib_check_rig_caps_set},
    { "-get_conf2",_wrap_rig_caps_get_conf2_get, _wrap_rig_caps_get_conf2_set},
    { "-password",_wrap_rig_caps_password_get, _wrap_rig_caps_password_set},
    { "-set_lock_mode",_wrap_rig_caps_set_lock_mode_get, _wrap_rig_caps_set_lock_mode_set},
    { "-get_lock_mode",_wrap_rig_caps_get_lock_mode_get, _wrap_rig_caps_get_lock_mode_set},
    { "-timeout_retry",_wrap_rig_caps_timeout_retry_get, _wrap_rig_caps_timeout_retry_set},
    { "-morse_qsize",_wrap_rig_caps_morse_qsize_get, _wrap_rig_caps_morse_qsize_set},
    {0,0,0}
};
static swig_class *swig_rig_caps_bases[] = {0};
static const char * swig_rig_caps_base_names[] = {0};
static swig_class _wrap_class_rig_caps = { "rig_caps", &SWIGTYPE_p_rig_caps,_wrap_new_rig_caps, swig_delete_rig_caps, swig_rig_caps_methods, swig_rig_caps_attributes, swig_rig_caps_bases,swig_rig_caps_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_multicast_s_multicast_running_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:multicast_s_multicast_running_set self multicast_running ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_multicast_running_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "multicast_s_multicast_running_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->multicast_running = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_multicast_running_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:multicast_s_multicast_running_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_multicast_running_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result = (int) ((arg1)->multicast_running);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_sock_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:multicast_s_sock_set self sock ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_sock_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "multicast_s_sock_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->sock = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_sock_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:multicast_s_sock_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_sock_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result = (int) ((arg1)->sock);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_seqnumber_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:multicast_s_seqnumber_set self seqnumber ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_seqnumber_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "multicast_s_seqnumber_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->seqnumber = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_seqnumber_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:multicast_s_seqnumber_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_seqnumber_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result = (int) ((arg1)->seqnumber);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_runflag_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:multicast_s_runflag_set self runflag ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_runflag_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "multicast_s_runflag_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->runflag = (int volatile )arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_runflag_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:multicast_s_runflag_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_runflag_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result = (int)(int) ((arg1)->runflag);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_threadid_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  pthread_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:multicast_s_threadid_set self threadid ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_threadid_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_pthread_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "multicast_s_threadid_set" "', argument " "2"" of type '" "pthread_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "multicast_s_threadid_set" "', argument " "2"" of type '" "pthread_t""'");
    } else {
      arg2 = *((pthread_t *)(argp2));
    }
  }
  if (arg1) (arg1)->threadid = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_threadid_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  pthread_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:multicast_s_threadid_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_threadid_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result =  ((arg1)->threadid);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((pthread_t *)memcpy((pthread_t *)calloc(1,sizeof(pthread_t)),&result,sizeof(pthread_t)), SWIGTYPE_p_pthread_t, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_mutex_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  pthread_mutex_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:multicast_s_mutex_set self mutex ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_mutex_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_pthread_mutex_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "multicast_s_mutex_set" "', argument " "2"" of type '" "pthread_mutex_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "multicast_s_mutex_set" "', argument " "2"" of type '" "pthread_mutex_t""'");
    } else {
      arg2 = *((pthread_mutex_t *)(argp2));
    }
  }
  if (arg1) (arg1)->mutex = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_mutex_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  pthread_mutex_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:multicast_s_mutex_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_mutex_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result =  ((arg1)->mutex);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((pthread_mutex_t *)memcpy((pthread_mutex_t *)calloc(1,sizeof(pthread_mutex_t)),&result,sizeof(pthread_mutex_t)), SWIGTYPE_p_pthread_mutex_t, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_mutex_initialized_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:multicast_s_mutex_initialized_set self mutex_initialized ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_mutex_initialized_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "multicast_s_mutex_initialized_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->mutex_initialized = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_mutex_initialized_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:multicast_s_mutex_initialized_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_mutex_initialized_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result = (int) ((arg1)->mutex_initialized);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_dest_addr_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  struct sockaddr_in arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:multicast_s_dest_addr_set self dest_addr ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_dest_addr_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_sockaddr_in,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "multicast_s_dest_addr_set" "', argument " "2"" of type '" "struct sockaddr_in""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "multicast_s_dest_addr_set" "', argument " "2"" of type '" "struct sockaddr_in""'");
    } else {
      arg2 = *((struct sockaddr_in *)(argp2));
    }
  }
  if (arg1) (arg1)->dest_addr = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_dest_addr_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct sockaddr_in result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:multicast_s_dest_addr_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_dest_addr_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result =  ((arg1)->dest_addr);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((struct sockaddr_in *)memcpy((struct sockaddr_in *)calloc(1,sizeof(struct sockaddr_in)),&result,sizeof(struct sockaddr_in)), SWIGTYPE_p_sockaddr_in, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_port_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:multicast_s_port_set self port ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_port_set" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "multicast_s_port_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->port = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_multicast_s_port_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:multicast_s_port_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "multicast_s_port_get" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  result = (int) ((arg1)->port);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_multicast_s(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_multicast_s ") == TCL_ERROR) SWIG_fail;
  result = (struct multicast_s *)calloc(1, sizeof(struct multicast_s));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_multicast_s,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_multicast_s(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct multicast_s *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_multicast_s self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_multicast_s, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_multicast_s" "', argument " "1"" of type '" "struct multicast_s *""'"); 
  }
  arg1 = (struct multicast_s *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_multicast_s(void *obj) {
struct multicast_s *arg1 = (struct multicast_s *) obj;
free((char *) arg1);
}
static swig_method swig_multicast_s_methods[] = {
    {0,0}
};
static swig_attribute swig_multicast_s_attributes[] = {
    { "-multicast_running",_wrap_multicast_s_multicast_running_get, _wrap_multicast_s_multicast_running_set},
    { "-sock",_wrap_multicast_s_sock_get, _wrap_multicast_s_sock_set},
    { "-seqnumber",_wrap_multicast_s_seqnumber_get, _wrap_multicast_s_seqnumber_set},
    { "-runflag",_wrap_multicast_s_runflag_get, _wrap_multicast_s_runflag_set},
    { "-threadid",_wrap_multicast_s_threadid_get, _wrap_multicast_s_threadid_set},
    { "-mutex",_wrap_multicast_s_mutex_get, _wrap_multicast_s_mutex_set},
    { "-mutex_initialized",_wrap_multicast_s_mutex_initialized_get, _wrap_multicast_s_mutex_initialized_set},
    { "-dest_addr",_wrap_multicast_s_dest_addr_get, _wrap_multicast_s_dest_addr_set},
    { "-port",_wrap_multicast_s_port_get, _wrap_multicast_s_port_set},
    {0,0,0}
};
static swig_class *swig_multicast_s_bases[] = {0};
static const char * swig_multicast_s_base_names[] = {0};
static swig_class _wrap_class_multicast_s = { "multicast_s", &SWIGTYPE_p_multicast_s,_wrap_new_multicast_s, swig_delete_multicast_s, swig_multicast_s_methods, swig_multicast_s_attributes, swig_multicast_s_bases,swig_multicast_s_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_rig_callbacks_freq_event_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  freq_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_freq_event_set self freq_event ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_freq_event_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_freq_event_set" "', argument " "2"" of type '" "freq_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->freq_event = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_freq_event_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_cb_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_freq_event_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_freq_event_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (freq_cb_t) ((arg1)->freq_event);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_double_p_void__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_freq_arg_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_freq_arg_set self freq_arg ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_freq_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_freq_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->freq_arg = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_freq_arg_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_freq_arg_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_freq_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->freq_arg);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_mode_event_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  mode_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_mode_event_set self mode_event ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_mode_event_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_mode_event_set" "', argument " "2"" of type '" "mode_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->mode_event = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_mode_event_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  mode_cb_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_mode_event_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_mode_event_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (mode_cb_t) ((arg1)->mode_event);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_mode_arg_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_mode_arg_set self mode_arg ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_mode_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_mode_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->mode_arg = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_mode_arg_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_mode_arg_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_mode_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->mode_arg);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_vfo_event_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  vfo_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_vfo_event_set self vfo_event ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_vfo_event_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_vfo_event_set" "', argument " "2"" of type '" "vfo_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->vfo_event = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_vfo_event_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  vfo_cb_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_vfo_event_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_vfo_event_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (vfo_cb_t) ((arg1)->vfo_event);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_void__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_vfo_arg_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_vfo_arg_set self vfo_arg ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_vfo_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_vfo_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->vfo_arg = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_vfo_arg_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_vfo_arg_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_vfo_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->vfo_arg);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_ptt_event_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  ptt_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_ptt_event_set self ptt_event ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_ptt_event_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_ptt_event_set" "', argument " "2"" of type '" "ptt_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->ptt_event = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_ptt_event_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptt_cb_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_ptt_event_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_ptt_event_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (ptt_cb_t) ((arg1)->ptt_event);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_ptt_arg_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_ptt_arg_set self ptt_arg ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_ptt_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_ptt_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->ptt_arg = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_ptt_arg_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_ptt_arg_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_ptt_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->ptt_arg);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_dcd_event_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  dcd_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_dcd_event_set self dcd_event ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_dcd_event_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_dcd_event_set" "', argument " "2"" of type '" "dcd_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->dcd_event = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_dcd_event_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  dcd_cb_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_dcd_event_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_dcd_event_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (dcd_cb_t) ((arg1)->dcd_event);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_dcd_arg_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_dcd_arg_set self dcd_arg ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_dcd_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_dcd_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->dcd_arg = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_dcd_arg_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_dcd_arg_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_dcd_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->dcd_arg);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_pltune_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  pltune_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_pltune_set self pltune ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_pltune_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_pltune_set" "', argument " "2"" of type '" "pltune_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->pltune = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_pltune_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  pltune_cb_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_pltune_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_pltune_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (pltune_cb_t) ((arg1)->pltune);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_pltune_arg_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_pltune_arg_set self pltune_arg ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_pltune_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_pltune_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->pltune_arg = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_pltune_arg_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_pltune_arg_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_pltune_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->pltune_arg);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_spectrum_event_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  spectrum_cb_t arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_spectrum_event_set self spectrum_event ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_spectrum_event_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rig_callbacks_spectrum_event_set" "', argument " "2"" of type '" "spectrum_cb_t""'"); 
    }
  }
  if (arg1) (arg1)->spectrum_event = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_spectrum_event_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  spectrum_cb_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_spectrum_event_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_spectrum_event_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (spectrum_cb_t) ((arg1)->spectrum_event);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_spectrum_arg_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_callbacks_spectrum_arg_set self spectrum_arg ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_spectrum_arg_set" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_callbacks_spectrum_arg_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->spectrum_arg = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_callbacks_spectrum_arg_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_callbacks_spectrum_arg_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_callbacks_spectrum_arg_get" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  result = (void *) ((arg1)->spectrum_arg);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_rig_callbacks(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_rig_callbacks ") == TCL_ERROR) SWIG_fail;
  result = (struct rig_callbacks *)calloc(1, sizeof(struct rig_callbacks));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_callbacks,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_rig_callbacks(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_callbacks *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_rig_callbacks self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_callbacks, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rig_callbacks" "', argument " "1"" of type '" "struct rig_callbacks *""'"); 
  }
  arg1 = (struct rig_callbacks *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_rig_callbacks(void *obj) {
struct rig_callbacks *arg1 = (struct rig_callbacks *) obj;
free((char *) arg1);
}
static swig_method swig_rig_callbacks_methods[] = {
    {0,0}
};
static swig_attribute swig_rig_callbacks_attributes[] = {
    { "-freq_event",_wrap_rig_callbacks_freq_event_get, _wrap_rig_callbacks_freq_event_set},
    { "-freq_arg",_wrap_rig_callbacks_freq_arg_get, _wrap_rig_callbacks_freq_arg_set},
    { "-mode_event",_wrap_rig_callbacks_mode_event_get, _wrap_rig_callbacks_mode_event_set},
    { "-mode_arg",_wrap_rig_callbacks_mode_arg_get, _wrap_rig_callbacks_mode_arg_set},
    { "-vfo_event",_wrap_rig_callbacks_vfo_event_get, _wrap_rig_callbacks_vfo_event_set},
    { "-vfo_arg",_wrap_rig_callbacks_vfo_arg_get, _wrap_rig_callbacks_vfo_arg_set},
    { "-ptt_event",_wrap_rig_callbacks_ptt_event_get, _wrap_rig_callbacks_ptt_event_set},
    { "-ptt_arg",_wrap_rig_callbacks_ptt_arg_get, _wrap_rig_callbacks_ptt_arg_set},
    { "-dcd_event",_wrap_rig_callbacks_dcd_event_get, _wrap_rig_callbacks_dcd_event_set},
    { "-dcd_arg",_wrap_rig_callbacks_dcd_arg_get, _wrap_rig_callbacks_dcd_arg_set},
    { "-pltune",_wrap_rig_callbacks_pltune_get, _wrap_rig_callbacks_pltune_set},
    { "-pltune_arg",_wrap_rig_callbacks_pltune_arg_get, _wrap_rig_callbacks_pltune_arg_set},
    { "-spectrum_event",_wrap_rig_callbacks_spectrum_event_get, _wrap_rig_callbacks_spectrum_event_set},
    { "-spectrum_arg",_wrap_rig_callbacks_spectrum_arg_get, _wrap_rig_callbacks_spectrum_arg_set},
    {0,0,0}
};
static swig_class *swig_rig_callbacks_bases[] = {0};
static const char * swig_rig_callbacks_base_names[] = {0};
static swig_class _wrap_class_rig_callbacks = { "rig_callbacks", &SWIGTYPE_p_rig_callbacks,_wrap_new_rig_callbacks, swig_delete_rig_callbacks, swig_rig_callbacks_methods, swig_rig_callbacks_attributes, swig_rig_callbacks_bases,swig_rig_callbacks_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_netrigctl_get_vfo_mode(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  RIG *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:netrigctl_get_vfo_mode rig ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_s_rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "netrigctl_get_vfo_mode" "', argument " "1"" of type '" "RIG *""'"); 
  }
  arg1 = (RIG *)(argp1);
  result = (int)netrigctl_get_vfo_mode(arg1);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rigerror(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rigerror errnum ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rigerror" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = (int)(val1);
  result = (char *)rigerror(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rigerror2(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rigerror2 errnum ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rigerror2" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = (int)(val1);
  result = (char *)rigerror2(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_set_debug(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  enum rig_debug_level_e arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_set_debug debug_level ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_set_debug" "', argument " "1"" of type '" "enum rig_debug_level_e""'");
  } 
  arg1 = (enum rig_debug_level_e)(val1);
  rig_set_debug(arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_get_debug(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  enum rig_debug_level_e *arg1 = 0 ;
  enum rig_debug_level_e temp1 ;
  int res1 = SWIG_TMPOBJ ;
  
  arg1 = &temp1;
  if (SWIG_GetArgs(interp, objc, objv,":rig_get_debug ") == TCL_ERROR) SWIG_fail;
  rig_get_debug(arg1);
  
  if (SWIG_IsTmpObj(res1)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_unsigned_SS_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_rig_debug_level_e, new_flags));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_set_debug_time_stamp(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_set_debug_time_stamp flag ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_set_debug_time_stamp" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = (int)(val1);
  rig_set_debug_time_stamp(arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_add2debugmsgsave(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  char *arg1 = 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:add2debugmsgsave s ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(objv[1], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "add2debugmsgsave" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  add2debugmsgsave((char const *)arg1);
  
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return TCL_OK;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strrmode(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  rmode_t arg1 ;
  unsigned long long val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_strrmode mode ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strrmode" "', argument " "1"" of type '" "rmode_t""'");
  } 
  arg1 = (rmode_t)(val1);
  result = (char *)rig_strrmode(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strrmodes(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  rmode_t arg1 ;
  char *arg2 = 0 ;
  int arg3 ;
  unsigned long long val1 ;
  int ecode1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:rig_strrmodes modes buf buflen ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strrmodes" "', argument " "1"" of type '" "rmode_t""'");
  } 
  arg1 = (rmode_t)(val1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_strrmodes" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "rig_strrmodes" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  result = (int)rig_strrmodes(arg1,arg2,arg3);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strvfo(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  vfo_t arg1 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_strvfo vfo ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strvfo" "', argument " "1"" of type '" "vfo_t""'");
  } 
  arg1 = (vfo_t)(val1);
  result = (char *)rig_strvfo(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strfunc(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  setting_t arg1 ;
  unsigned long long val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_strfunc setting_t ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strfunc" "', argument " "1"" of type '" "setting_t""'");
  } 
  arg1 = (setting_t)(val1);
  result = (char *)rig_strfunc(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strlevel(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  setting_t arg1 ;
  unsigned long long val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_strlevel setting_t ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strlevel" "', argument " "1"" of type '" "setting_t""'");
  } 
  arg1 = (setting_t)(val1);
  result = (char *)rig_strlevel(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strparm(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  setting_t arg1 ;
  unsigned long long val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_strparm setting_t ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strparm" "', argument " "1"" of type '" "setting_t""'");
  } 
  arg1 = (setting_t)(val1);
  result = (char *)rig_strparm(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_stragclevel(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  enum agc_level_e arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_stragclevel level ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_stragclevel" "', argument " "1"" of type '" "enum agc_level_e""'");
  } 
  arg1 = (enum agc_level_e)(val1);
  result = (char *)rig_stragclevel(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strptrshift(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  rptr_shift_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_strptrshift rptr_shift_t ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strptrshift" "', argument " "1"" of type '" "rptr_shift_t""'");
  } 
  arg1 = (rptr_shift_t)(val1);
  result = (char *)rig_strptrshift(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strvfop(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  vfo_op_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_strvfop op ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strvfop" "', argument " "1"" of type '" "vfo_op_t""'");
  } 
  arg1 = (vfo_op_t)(val1);
  result = (char *)rig_strvfop(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strscan(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  scan_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_strscan scan ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strscan" "', argument " "1"" of type '" "scan_t""'");
  } 
  arg1 = (scan_t)(val1);
  result = (char *)rig_strscan(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strstatus(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  enum rig_status_e arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_strstatus status ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strstatus" "', argument " "1"" of type '" "enum rig_status_e""'");
  } 
  arg1 = (enum rig_status_e)(val1);
  result = (char *)rig_strstatus(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strmtype(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  chan_type_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_strmtype mtype ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strmtype" "', argument " "1"" of type '" "chan_type_t""'");
  } 
  arg1 = (chan_type_t)(val1);
  result = (char *)rig_strmtype(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strspectrummode(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  enum rig_spectrum_mode_e arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_strspectrummode mode ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strspectrummode" "', argument " "1"" of type '" "enum rig_spectrum_mode_e""'");
  } 
  arg1 = (enum rig_spectrum_mode_e)(val1);
  result = (char *)rig_strspectrummode(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_strcommstatus(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  rig_comm_status_t arg1 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_strcommstatus vfo ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "rig_strcommstatus" "', argument " "1"" of type '" "rig_comm_status_t""'");
  } 
  arg1 = (rig_comm_status_t)(val1);
  result = (char *)rig_strcommstatus(arg1);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_hl_usleep(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  rig_useconds_t arg1 ;
  unsigned long val1 ;
  int ecode1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:hl_usleep msec ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_long SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "hl_usleep" "', argument " "1"" of type '" "rig_useconds_t""'");
  } 
  arg1 = (rig_useconds_t)(val1);
  result = (int)hl_usleep(arg1);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_longlat2locator(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  double arg1 ;
  double arg2 ;
  char *arg3 = 0 ;
  int arg4 ;
  double val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  char temp3[13+1] ;
  int val4 ;
  int ecode4 = 0 ;
  int result;
  
  arg3 = (char *) temp3;
  if (SWIG_GetArgs(interp, objc, objv,"ooo:longlat2locator longitude latitude pair_count ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "longlat2locator" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = (double)(val1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "longlat2locator" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  ecode4 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "longlat2locator" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = (int)(val4);
  result = (int)longlat2locator(arg1,arg2,arg3,arg4);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  arg3[13] = 0;  
  Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_FromCharPtr(arg3));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_locator2longlat(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  double *arg1 = 0 ;
  double *arg2 = 0 ;
  char *arg3 = 0 ;
  double temp1 ;
  int res1 = SWIG_TMPOBJ ;
  double temp2 ;
  int res2 = SWIG_TMPOBJ ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int result;
  
  arg1 = &temp1;
  arg2 = &temp2;
  if (SWIG_GetArgs(interp, objc, objv,"o:locator2longlat locator ",(void *)0) == TCL_ERROR) SWIG_fail;
  res3 = SWIG_AsCharPtrAndSize(objv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "locator2longlat" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  result = (int)locator2longlat(arg1,arg2,(char const *)arg3);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  if (SWIG_IsTmpObj(res1)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_double((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_double, new_flags));
  }
  if (SWIG_IsTmpObj(res2)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_double((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_double, new_flags));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_OK;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_vfo_comp_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_vfo_comp_set self vfo_comp ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_comp_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_vfo_comp_set" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  if (arg1) (arg1)->vfo_comp = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_vfo_comp_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_vfo_comp_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_comp_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (double) ((arg1)->vfo_comp);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_range_list_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_rx_range_list_set self rx_range_list ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rx_range_list_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_range_list_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_rx_range_list_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_range_list_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_tx_range_list_set self tx_range_list ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tx_range_list_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_range_list_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_tx_range_list_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tuning_steps_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  struct tuning_step_list *arg2 = (struct tuning_step_list *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_tuning_steps_set self tuning_steps ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tuning_steps_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_tuning_step_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tuning_steps_set" "', argument " "2"" of type '" "struct tuning_step_list [20]""'"); 
  } 
  arg2 = (struct tuning_step_list *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)20; ++ii) *(struct tuning_step_list *)&arg1->tuning_steps[ii] = *((struct tuning_step_list *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tuning_steps""' of type '""struct tuning_step_list [20]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tuning_steps_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct tuning_step_list *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_tuning_steps_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tuning_steps_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (struct tuning_step_list *)(struct tuning_step_list *) ((arg1)->tuning_steps);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_tuning_step_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_filters_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  struct filter_list *arg2 = (struct filter_list *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_filters_set self filters ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_filters_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_filter_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_filters_set" "', argument " "2"" of type '" "struct filter_list [60]""'"); 
  } 
  arg2 = (struct filter_list *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)60; ++ii) *(struct filter_list *)&arg1->filters[ii] = *((struct filter_list *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""filters""' of type '""struct filter_list [60]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_filters_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct filter_list *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_filters_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_filters_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (struct filter_list *)(struct filter_list *) ((arg1)->filters);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_filter_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_str_cal_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  cal_table_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_str_cal_set self str_cal ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_str_cal_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_cal_table, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_str_cal_set" "', argument " "2"" of type '" "cal_table_t *""'"); 
  }
  arg2 = (cal_table_t *)(argp2);
  if (arg1) (arg1)->str_cal = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_str_cal_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cal_table_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_str_cal_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_str_cal_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (cal_table_t *)& ((arg1)->str_cal);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_cal_table,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_chan_list_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  chan_t *arg2 = (chan_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_chan_list_set self chan_list ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_chan_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_chan_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_chan_list_set" "', argument " "2"" of type '" "chan_t [16]""'"); 
  } 
  arg2 = (chan_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)16; ++ii) *(chan_t *)&arg1->chan_list[ii] = *((chan_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""chan_list""' of type '""chan_t [16]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_chan_list_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  chan_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_chan_list_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_chan_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (chan_t *)(chan_t *) ((arg1)->chan_list);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_chan_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_max_rit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_max_rit_set self max_rit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_max_rit_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_max_rit_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->max_rit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_max_rit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_max_rit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_max_rit_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (shortfreq_t) ((arg1)->max_rit);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_max_xit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_max_xit_set self max_xit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_max_xit_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_max_xit_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->max_xit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_max_xit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_max_xit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_max_xit_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (shortfreq_t) ((arg1)->max_xit);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_max_ifshift_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  shortfreq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_max_ifshift_set self max_ifshift ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_max_ifshift_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_max_ifshift_set" "', argument " "2"" of type '" "shortfreq_t""'");
  } 
  arg2 = (shortfreq_t)(val2);
  if (arg1) (arg1)->max_ifshift = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_max_ifshift_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_max_ifshift_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_max_ifshift_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (shortfreq_t) ((arg1)->max_ifshift);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_announces_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  ann_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_announces_set self announces ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_announces_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_announces_set" "', argument " "2"" of type '" "ann_t""'");
  } 
  arg2 = (ann_t)(val2);
  if (arg1) (arg1)->announces = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_announces_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ann_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_announces_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_announces_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (ann_t) ((arg1)->announces);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_preamp_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int *arg2 = (int *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_preamp_set self preamp ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_preamp_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_preamp_set" "', argument " "2"" of type '" "int [8]""'"); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(int *)&arg1->preamp[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""preamp""' of type '""int [8]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_preamp_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_preamp_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_preamp_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int *)(int *) ((arg1)->preamp);
  {
    int i, len, l;
    len = 8;
    
    Tcl_Obj * list = Tcl_NewListObj(len, NULL);
    for(i=0; i < len; i++) {
      Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(result[i]));
    }
    Tcl_SetObjResult(interp, list);
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_attenuator_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int *arg2 = (int *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_attenuator_set self attenuator ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_attenuator_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_attenuator_set" "', argument " "2"" of type '" "int [8]""'"); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(int *)&arg1->attenuator[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""attenuator""' of type '""int [8]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_attenuator_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_attenuator_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_attenuator_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int *)(int *) ((arg1)->attenuator);
  {
    int i, len, l;
    len = 8;
    
    Tcl_Obj * list = Tcl_NewListObj(len, NULL);
    for(i=0; i < len; i++) {
      Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(result[i]));
    }
    Tcl_SetObjResult(interp, list);
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_has_get_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_has_get_func_set self has_get_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_get_func_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_has_get_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_has_get_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_has_get_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_get_func_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_func);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_has_set_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_has_set_func_set self has_set_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_set_func_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_has_set_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_has_set_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_has_set_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_set_func_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_func);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_has_get_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_has_get_level_set self has_get_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_get_level_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_has_get_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_has_get_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_has_get_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_get_level_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_level);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_has_set_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_has_set_level_set self has_set_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_set_level_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_has_set_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_has_set_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_has_set_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_set_level_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_level);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_has_get_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_has_get_parm_set self has_get_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_get_parm_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_has_get_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_has_get_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_has_get_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_get_parm_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_parm);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_has_set_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_has_set_parm_set self has_set_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_set_parm_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_has_set_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_has_set_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_has_set_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_has_set_parm_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_parm);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_level_gran_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_level_gran_set self level_gran ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_level_gran_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_level_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->level_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""level_gran""' of type '""gran_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_level_gran_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gran_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_level_gran_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_level_gran_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->level_gran);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_parm_gran_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_parm_gran_set self parm_gran ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_parm_gran_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_parm_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->parm_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""parm_gran""' of type '""gran_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_parm_gran_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gran_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_parm_gran_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_parm_gran_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->parm_gran);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_transaction_active_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_transaction_active_set self transaction_active ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_transaction_active_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_transaction_active_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->transaction_active = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_transaction_active_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_transaction_active_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_transaction_active_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->transaction_active);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_current_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_current_vfo_set self current_vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_vfo_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_current_vfo_set" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  if (arg1) (arg1)->current_vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_current_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  vfo_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_current_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_vfo_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (vfo_t) ((arg1)->current_vfo);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_vfo_list_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_vfo_list_set self vfo_list ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_vfo_list_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->vfo_list = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_vfo_list_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_vfo_list_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->vfo_list);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_comm_state_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_comm_state_set self comm_state ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_comm_state_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_comm_state_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->comm_state = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_comm_state_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_comm_state_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_comm_state_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->comm_state);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_priv_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_priv_set self priv ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_priv_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_priv_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->priv = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_priv_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_priv_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_priv_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->priv);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_obj_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_obj_set self obj ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_obj_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_obj_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->obj = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_obj_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_obj_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_obj_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->obj);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_async_data_enabled_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_async_data_enabled_set self async_data_enabled ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_enabled_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_async_data_enabled_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->async_data_enabled = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_async_data_enabled_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_async_data_enabled_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_enabled_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->async_data_enabled);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_poll_interval_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_poll_interval_set self poll_interval ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_poll_interval_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_poll_interval_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->poll_interval = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_poll_interval_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_poll_interval_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_poll_interval_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->poll_interval);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_current_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_current_freq_set self current_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_freq_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_current_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->current_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_current_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_current_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_freq_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_t) ((arg1)->current_freq);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_current_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_current_mode_set self current_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_mode_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_current_mode_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->current_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_current_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rmode_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_current_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_mode_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (rmode_t) ((arg1)->current_mode);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_current_width_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  pbwidth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_current_width_set self current_width ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_width_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_current_width_set" "', argument " "2"" of type '" "pbwidth_t""'");
  } 
  arg2 = (pbwidth_t)(val2);
  if (arg1) (arg1)->current_width = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_current_width_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  pbwidth_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_current_width_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_current_width_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (pbwidth_t) ((arg1)->current_width);
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_tx_vfo_set self tx_vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_vfo_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_tx_vfo_set" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  if (arg1) (arg1)->tx_vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  vfo_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_tx_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_vfo_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (vfo_t) ((arg1)->tx_vfo);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_mode_list_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_mode_list_set self mode_list ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_mode_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_mode_list_set" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (arg1) (arg1)->mode_list = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_mode_list_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rmode_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_mode_list_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_mode_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (rmode_t) ((arg1)->mode_list);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_transmit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_transmit_set self transmit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_transmit_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_transmit_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->transmit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_transmit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_transmit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_transmit_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->transmit);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_lo_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_lo_freq_set self lo_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_lo_freq_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_lo_freq_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->lo_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_lo_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_lo_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_lo_freq_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_t) ((arg1)->lo_freq);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_twiddle_time_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_twiddle_time_set self twiddle_time ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_time_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_time_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_twiddle_time_set" "', argument " "2"" of type '" "time_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_twiddle_time_set" "', argument " "2"" of type '" "time_t""'");
    } else {
      arg2 = *((time_t *)(argp2));
    }
  }
  if (arg1) (arg1)->twiddle_time = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_twiddle_time_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  time_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_twiddle_time_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_time_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->twiddle_time);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((time_t *)memcpy((time_t *)calloc(1,sizeof(time_t)),&result,sizeof(time_t)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_twiddle_timeout_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_twiddle_timeout_set self twiddle_timeout ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_timeout_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_twiddle_timeout_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->twiddle_timeout = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_twiddle_timeout_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_twiddle_timeout_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_timeout_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->twiddle_timeout);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_uplink_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_uplink_set self uplink ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_uplink_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_uplink_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->uplink = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_uplink_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_uplink_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_uplink_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->uplink);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_vfo_opt_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_vfo_opt_set self vfo_opt ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_opt_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_vfo_opt_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->vfo_opt = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_vfo_opt_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_vfo_opt_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_opt_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->vfo_opt);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_auto_power_on_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_auto_power_on_set self auto_power_on ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_auto_power_on_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_auto_power_on_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->auto_power_on = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_auto_power_on_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_auto_power_on_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_auto_power_on_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->auto_power_on);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_auto_power_off_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_auto_power_off_set self auto_power_off ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_auto_power_off_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_auto_power_off_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->auto_power_off = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_auto_power_off_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_auto_power_off_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_auto_power_off_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->auto_power_off);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_auto_disable_screensaver_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_auto_disable_screensaver_set self auto_disable_screensaver ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_auto_disable_screensaver_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_auto_disable_screensaver_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->auto_disable_screensaver = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_auto_disable_screensaver_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_auto_disable_screensaver_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_auto_disable_screensaver_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->auto_disable_screensaver);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_ptt_share_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_ptt_share_set self ptt_share ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_ptt_share_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_ptt_share_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->ptt_share = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_ptt_share_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_ptt_share_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_ptt_share_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->ptt_share);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_power_now_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_power_now_set self power_now ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_power_now_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_power_now_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->power_now = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_power_now_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_power_now_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_power_now_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->power_now);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_power_min_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_power_min_set self power_min ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_power_min_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_power_min_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->power_min = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_power_min_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_power_min_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_power_min_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->power_min);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_power_max_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_power_max_set self power_max ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_power_max_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_power_max_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->power_max = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_power_max_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_power_max_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_power_max_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->power_max);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_disable_yaesu_bandselect_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  unsigned char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_disable_yaesu_bandselect_set self disable_yaesu_bandselect ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_disable_yaesu_bandselect_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_char SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_disable_yaesu_bandselect_set" "', argument " "2"" of type '" "unsigned char""'");
  } 
  arg2 = (unsigned char)(val2);
  if (arg1) (arg1)->disable_yaesu_bandselect = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_disable_yaesu_bandselect_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_disable_yaesu_bandselect_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_disable_yaesu_bandselect_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (unsigned char) ((arg1)->disable_yaesu_bandselect);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_char((unsigned char)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_twiddle_rit_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_twiddle_rit_set self twiddle_rit ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_rit_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_twiddle_rit_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->twiddle_rit = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_twiddle_rit_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_twiddle_rit_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_rit_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->twiddle_rit);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_twiddle_state_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_twiddle_state_set self twiddle_state ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_state_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_twiddle_state_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->twiddle_state = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_twiddle_state_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_twiddle_state_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_twiddle_state_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->twiddle_state);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_rx_vfo_set self rx_vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_vfo_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_rx_vfo_set" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  if (arg1) (arg1)->rx_vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  vfo_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_rx_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_vfo_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (vfo_t) ((arg1)->rx_vfo);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_snapshot_packet_sequence_number_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_snapshot_packet_sequence_number_set self snapshot_packet_sequence_number ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_snapshot_packet_sequence_number_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_snapshot_packet_sequence_number_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->snapshot_packet_sequence_number = (unsigned int volatile )arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_snapshot_packet_sequence_number_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_snapshot_packet_sequence_number_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_snapshot_packet_sequence_number_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (unsigned int)(unsigned int) ((arg1)->snapshot_packet_sequence_number);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_publisher_run_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_multicast_publisher_run_set self multicast_publisher_run ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_publisher_run_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_multicast_publisher_run_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->multicast_publisher_run = (int volatile )arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_publisher_run_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_multicast_publisher_run_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_publisher_run_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int)(int) ((arg1)->multicast_publisher_run);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_publisher_priv_data_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_multicast_publisher_priv_data_set self multicast_publisher_priv_data ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_publisher_priv_data_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_multicast_publisher_priv_data_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->multicast_publisher_priv_data = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_publisher_priv_data_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_multicast_publisher_priv_data_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_publisher_priv_data_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->multicast_publisher_priv_data);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_async_data_handler_thread_run_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_async_data_handler_thread_run_set self async_data_handler_thread_run ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_handler_thread_run_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_async_data_handler_thread_run_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->async_data_handler_thread_run = (int volatile )arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_async_data_handler_thread_run_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_async_data_handler_thread_run_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_handler_thread_run_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int)(int) ((arg1)->async_data_handler_thread_run);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_async_data_handler_priv_data_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_async_data_handler_priv_data_set self async_data_handler_priv_data ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_handler_priv_data_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_async_data_handler_priv_data_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->async_data_handler_priv_data = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_async_data_handler_priv_data_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_async_data_handler_priv_data_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_handler_priv_data_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->async_data_handler_priv_data);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_poll_routine_thread_run_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_poll_routine_thread_run_set self poll_routine_thread_run ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_poll_routine_thread_run_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_poll_routine_thread_run_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->poll_routine_thread_run = (int volatile )arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_poll_routine_thread_run_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_poll_routine_thread_run_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_poll_routine_thread_run_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int)(int) ((arg1)->poll_routine_thread_run);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_poll_routine_priv_data_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_poll_routine_priv_data_set self poll_routine_priv_data ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_poll_routine_priv_data_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_poll_routine_priv_data_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->poll_routine_priv_data = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_poll_routine_priv_data_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_poll_routine_priv_data_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_poll_routine_priv_data_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->poll_routine_priv_data);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_mutex_set_transaction_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  pthread_mutex_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_mutex_set_transaction_set self mutex_set_transaction ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_mutex_set_transaction_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_pthread_mutex_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_mutex_set_transaction_set" "', argument " "2"" of type '" "pthread_mutex_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_mutex_set_transaction_set" "', argument " "2"" of type '" "pthread_mutex_t""'");
    } else {
      arg2 = *((pthread_mutex_t *)(argp2));
    }
  }
  if (arg1) (arg1)->mutex_set_transaction = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_mutex_set_transaction_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  pthread_mutex_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_mutex_set_transaction_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_mutex_set_transaction_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->mutex_set_transaction);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((pthread_mutex_t *)memcpy((pthread_mutex_t *)calloc(1,sizeof(pthread_mutex_t)),&result,sizeof(pthread_mutex_t)), SWIGTYPE_p_pthread_mutex_t, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rigport_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  hamlib_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_rigport_set self rigport ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rigport_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_hamlib_port,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rigport_set" "', argument " "2"" of type '" "hamlib_port_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_rigport_set" "', argument " "2"" of type '" "hamlib_port_t""'");
    } else {
      arg2 = *((hamlib_port_t *)(argp2));
    }
  }
  if (arg1) (arg1)->rigport = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rigport_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  hamlib_port_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_rigport_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rigport_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->rigport);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((hamlib_port_t *)memcpy((hamlib_port_t *)calloc(1,sizeof(hamlib_port_t)),&result,sizeof(hamlib_port_t)), SWIGTYPE_p_hamlib_port, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_pttport_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  hamlib_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_pttport_set self pttport ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_pttport_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_hamlib_port,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_pttport_set" "', argument " "2"" of type '" "hamlib_port_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_pttport_set" "', argument " "2"" of type '" "hamlib_port_t""'");
    } else {
      arg2 = *((hamlib_port_t *)(argp2));
    }
  }
  if (arg1) (arg1)->pttport = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_pttport_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  hamlib_port_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_pttport_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_pttport_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->pttport);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((hamlib_port_t *)memcpy((hamlib_port_t *)calloc(1,sizeof(hamlib_port_t)),&result,sizeof(hamlib_port_t)), SWIGTYPE_p_hamlib_port, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_dcdport_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  hamlib_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_dcdport_set self dcdport ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dcdport_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_hamlib_port,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_dcdport_set" "', argument " "2"" of type '" "hamlib_port_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_dcdport_set" "', argument " "2"" of type '" "hamlib_port_t""'");
    } else {
      arg2 = *((hamlib_port_t *)(argp2));
    }
  }
  if (arg1) (arg1)->dcdport = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_dcdport_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  hamlib_port_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_dcdport_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dcdport_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->dcdport);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((hamlib_port_t *)memcpy((hamlib_port_t *)calloc(1,sizeof(hamlib_port_t)),&result,sizeof(hamlib_port_t)), SWIGTYPE_p_hamlib_port, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_use_cached_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_use_cached_freq_set self use_cached_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_use_cached_freq_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_use_cached_freq_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->use_cached_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_use_cached_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_use_cached_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_use_cached_freq_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->use_cached_freq);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_use_cached_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_use_cached_mode_set self use_cached_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_use_cached_mode_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_use_cached_mode_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->use_cached_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_use_cached_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_use_cached_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_use_cached_mode_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->use_cached_mode);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_use_cached_ptt_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_use_cached_ptt_set self use_cached_ptt ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_use_cached_ptt_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_use_cached_ptt_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->use_cached_ptt = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_use_cached_ptt_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_use_cached_ptt_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_use_cached_ptt_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->use_cached_ptt);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_depth_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_depth_set self depth ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_depth_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_depth_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->depth = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_depth_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_depth_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_depth_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->depth);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_lock_mode_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_lock_mode_set self lock_mode ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_lock_mode_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_lock_mode_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->lock_mode = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_lock_mode_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_lock_mode_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_lock_mode_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->lock_mode);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_powerstat_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  powerstat_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_powerstat_set self powerstat ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_powerstat_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_powerstat_set" "', argument " "2"" of type '" "powerstat_t""'");
  } 
  arg2 = (powerstat_t)(val2);
  if (arg1) (arg1)->powerstat = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_powerstat_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  powerstat_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_powerstat_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_powerstat_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (powerstat_t) ((arg1)->powerstat);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tuner_control_pathname_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_tuner_control_pathname_set self tuner_control_pathname ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tuner_control_pathname_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tuner_control_pathname_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->tuner_control_pathname);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->tuner_control_pathname = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->tuner_control_pathname = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tuner_control_pathname_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_tuner_control_pathname_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tuner_control_pathname_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->tuner_control_pathname);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_client_version_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  char *arg2 = (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char temp2[32] ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_client_version_set self client_version ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_client_version_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharArray(objv[2], temp2, 32);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_client_version_set" "', argument " "2"" of type '" "char [32]""'");
  }
  arg2 = (char *)(temp2);
  if (arg2) memcpy(arg1->client_version,arg2,32*sizeof(char));
  else memset(arg1->client_version,0,32*sizeof(char));
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_client_version_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_client_version_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_client_version_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *)(char *) ((arg1)->client_version);
  {
    size_t size = SWIG_strnlen(result, 32);
    
    
    
    Tcl_SetObjResult(interp,SWIG_FromCharPtrAndSize(result, size));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_offset_vfoa_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_offset_vfoa_set self offset_vfoa ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_offset_vfoa_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_offset_vfoa_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->offset_vfoa = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_offset_vfoa_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_offset_vfoa_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_offset_vfoa_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_t) ((arg1)->offset_vfoa);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_offset_vfob_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_offset_vfob_set self offset_vfob ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_offset_vfob_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_offset_vfob_set" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  if (arg1) (arg1)->offset_vfob = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_offset_vfob_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_offset_vfob_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_offset_vfob_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_t) ((arg1)->offset_vfob);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  struct multicast_s *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_multicast_set self multicast ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_multicast_s, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_multicast_set" "', argument " "2"" of type '" "struct multicast_s *""'"); 
  }
  arg2 = (struct multicast_s *)(argp2);
  if (arg1) (arg1)->multicast = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct multicast_s *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_multicast_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (struct multicast_s *) ((arg1)->multicast);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_multicast_s,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rig_model_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  rig_model_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_rig_model_set self rig_model ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rig_model_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_rig_model_set" "', argument " "2"" of type '" "rig_model_t""'");
  } 
  arg2 = (rig_model_t)(val2);
  if (arg1) (arg1)->rig_model = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rig_model_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rig_model_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_rig_model_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rig_model_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (rig_model_t) ((arg1)->rig_model);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_model_name_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_model_name_set self model_name ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_model_name_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_model_name_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->model_name = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->model_name = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_model_name_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_model_name_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_model_name_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->model_name);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_mfg_name_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_mfg_name_set self mfg_name ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_mfg_name_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_mfg_name_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->mfg_name = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->mfg_name = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_mfg_name_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_mfg_name_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_mfg_name_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->mfg_name);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_version_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_version_set self version ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_version_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_version_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->version = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->version = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_version_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_version_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_version_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->version);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_copyright_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_copyright_set self copyright ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_copyright_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_copyright_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->copyright = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->copyright = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_copyright_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_copyright_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_copyright_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->copyright);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_status_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  enum rig_status_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_status_set self status ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_status_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_status_set" "', argument " "2"" of type '" "enum rig_status_e""'");
  } 
  arg2 = (enum rig_status_e)(val2);
  if (arg1) (arg1)->status = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_status_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum rig_status_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_status_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_status_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (enum rig_status_e) ((arg1)->status);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rig_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_rig_type_set self rig_type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rig_type_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_rig_type_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->rig_type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rig_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_rig_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rig_type_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->rig_type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_ptt_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  ptt_type_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_ptt_type_set self ptt_type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_ptt_type_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_ptt_type_set" "', argument " "2"" of type '" "ptt_type_t""'");
  } 
  arg2 = (ptt_type_t)(val2);
  if (arg1) (arg1)->ptt_type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_ptt_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptt_type_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_ptt_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_ptt_type_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (ptt_type_t) ((arg1)->ptt_type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_dcd_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  dcd_type_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_dcd_type_set self dcd_type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dcd_type_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_dcd_type_set" "', argument " "2"" of type '" "dcd_type_t""'");
  } 
  arg2 = (dcd_type_t)(val2);
  if (arg1) (arg1)->dcd_type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_dcd_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  dcd_type_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_dcd_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dcd_type_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (dcd_type_t) ((arg1)->dcd_type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_port_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  rig_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_port_type_set self port_type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_port_type_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_port_type_set" "', argument " "2"" of type '" "rig_port_t""'");
  } 
  arg2 = (rig_port_t)(val2);
  if (arg1) (arg1)->port_type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_port_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rig_port_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_port_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_port_type_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (rig_port_t) ((arg1)->port_type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_serial_rate_min_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_serial_rate_min_set self serial_rate_min ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_rate_min_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_serial_rate_min_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_min = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_serial_rate_min_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_serial_rate_min_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_rate_min_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->serial_rate_min);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_serial_rate_max_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_serial_rate_max_set self serial_rate_max ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_rate_max_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_serial_rate_max_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_max = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_serial_rate_max_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_serial_rate_max_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_rate_max_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->serial_rate_max);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_serial_data_bits_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_serial_data_bits_set self serial_data_bits ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_data_bits_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_serial_data_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_data_bits = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_serial_data_bits_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_serial_data_bits_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_data_bits_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->serial_data_bits);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_serial_stop_bits_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_serial_stop_bits_set self serial_stop_bits ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_stop_bits_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_serial_stop_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_stop_bits = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_serial_stop_bits_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_serial_stop_bits_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_stop_bits_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->serial_stop_bits);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_serial_parity_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  enum serial_parity_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_serial_parity_set self serial_parity ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_parity_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_serial_parity_set" "', argument " "2"" of type '" "enum serial_parity_e""'");
  } 
  arg2 = (enum serial_parity_e)(val2);
  if (arg1) (arg1)->serial_parity = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_serial_parity_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum serial_parity_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_serial_parity_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_parity_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (enum serial_parity_e) ((arg1)->serial_parity);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_serial_handshake_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  enum serial_handshake_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_serial_handshake_set self serial_handshake ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_handshake_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_serial_handshake_set" "', argument " "2"" of type '" "enum serial_handshake_e""'");
  } 
  arg2 = (enum serial_handshake_e)(val2);
  if (arg1) (arg1)->serial_handshake = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_serial_handshake_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum serial_handshake_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_serial_handshake_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_serial_handshake_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (enum serial_handshake_e) ((arg1)->serial_handshake);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_write_delay_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_write_delay_set self write_delay ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_write_delay_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->write_delay = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_write_delay_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_write_delay_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_write_delay_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->write_delay);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_post_write_delay_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_post_write_delay_set self post_write_delay ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_post_write_delay_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_post_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->post_write_delay = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_post_write_delay_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_post_write_delay_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_post_write_delay_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->post_write_delay);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_timeout_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_timeout_set self timeout ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_timeout_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_timeout_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->timeout = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_timeout_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_timeout_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_timeout_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->timeout);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_retry_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_retry_set self retry ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_retry_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_retry_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->retry = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_retry_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_retry_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_retry_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->retry);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_targetable_vfo_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_targetable_vfo_set self targetable_vfo ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_targetable_vfo_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_targetable_vfo_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->targetable_vfo = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_targetable_vfo_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_targetable_vfo_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_targetable_vfo_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->targetable_vfo);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_async_data_supported_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_async_data_supported_set self async_data_supported ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_supported_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_async_data_supported_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->async_data_supported = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_async_data_supported_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_async_data_supported_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_async_data_supported_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->async_data_supported);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_agc_level_count_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_agc_level_count_set self agc_level_count ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_agc_level_count_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_agc_level_count_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->agc_level_count = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_agc_level_count_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_agc_level_count_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_agc_level_count_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->agc_level_count);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_agc_levels_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  enum agc_level_e *arg2 = (enum agc_level_e *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_agc_levels_set self agc_levels ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_agc_levels_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_agc_level_e, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_agc_levels_set" "', argument " "2"" of type '" "enum agc_level_e [8]""'"); 
  } 
  arg2 = (enum agc_level_e *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(enum agc_level_e *)&arg1->agc_levels[ii] = *((enum agc_level_e *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""agc_levels""' of type '""enum agc_level_e [8]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_agc_levels_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum agc_level_e *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_agc_levels_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_agc_levels_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (enum agc_level_e *)(enum agc_level_e *) ((arg1)->agc_levels);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_agc_level_e,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_ctcss_list_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  tone_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_ctcss_list_set self ctcss_list ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_ctcss_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_unsigned_int, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_ctcss_list_set" "', argument " "2"" of type '" "tone_t *""'"); 
  }
  arg2 = (tone_t *)(argp2);
  if (arg1) (arg1)->ctcss_list = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_ctcss_list_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  tone_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_ctcss_list_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_ctcss_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (tone_t *) ((arg1)->ctcss_list);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_dcs_list_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  tone_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_dcs_list_set self dcs_list ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dcs_list_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_unsigned_int, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_dcs_list_set" "', argument " "2"" of type '" "tone_t *""'"); 
  }
  arg2 = (tone_t *)(argp2);
  if (arg1) (arg1)->dcs_list = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_dcs_list_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  tone_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_dcs_list_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dcs_list_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (tone_t *) ((arg1)->dcs_list);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_vfo_ops_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  vfo_op_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_vfo_ops_set self vfo_ops ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_ops_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_vfo_ops_set" "', argument " "2"" of type '" "vfo_op_t""'");
  } 
  arg2 = (vfo_op_t)(val2);
  if (arg1) (arg1)->vfo_ops = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_vfo_ops_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  vfo_op_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_vfo_ops_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_vfo_ops_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (vfo_op_t) ((arg1)->vfo_ops);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_scan_ops_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  scan_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_scan_ops_set self scan_ops ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_scan_ops_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_scan_ops_set" "', argument " "2"" of type '" "scan_t""'");
  } 
  arg2 = (scan_t)(val2);
  if (arg1) (arg1)->scan_ops = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_scan_ops_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  scan_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_scan_ops_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_scan_ops_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (scan_t) ((arg1)->scan_ops);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_transceive_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_transceive_set self transceive ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_transceive_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_transceive_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->transceive = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_transceive_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_transceive_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_transceive_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->transceive);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_bank_qty_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_bank_qty_set self bank_qty ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_bank_qty_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_bank_qty_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->bank_qty = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_bank_qty_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_bank_qty_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_bank_qty_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->bank_qty);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_chan_desc_sz_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_chan_desc_sz_set self chan_desc_sz ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_chan_desc_sz_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_chan_desc_sz_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->chan_desc_sz = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_chan_desc_sz_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_chan_desc_sz_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_chan_desc_sz_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->chan_desc_sz);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_range_list1_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_rx_range_list1_set self rx_range_list1 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list1_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rx_range_list1_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list1[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list1""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_range_list1_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_rx_range_list1_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list1_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_range_list1_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_tx_range_list1_set self tx_range_list1 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list1_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tx_range_list1_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list1[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list1""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_range_list1_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_tx_range_list1_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list1_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_range_list2_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_rx_range_list2_set self rx_range_list2 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list2_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rx_range_list2_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list2[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list2""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_range_list2_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_rx_range_list2_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list2_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list2);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_range_list2_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_tx_range_list2_set self tx_range_list2 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list2_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tx_range_list2_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list2[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list2""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_range_list2_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_tx_range_list2_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list2_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list2);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_range_list3_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_rx_range_list3_set self rx_range_list3 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list3_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rx_range_list3_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list3[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list3""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_range_list3_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_rx_range_list3_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list3_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list3);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_range_list3_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_tx_range_list3_set self tx_range_list3 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list3_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tx_range_list3_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list3[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list3""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_range_list3_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_tx_range_list3_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list3_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list3);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_range_list4_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_rx_range_list4_set self rx_range_list4 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list4_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rx_range_list4_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list4[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list4""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_range_list4_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_rx_range_list4_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list4_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list4);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_range_list4_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_tx_range_list4_set self tx_range_list4 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list4_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tx_range_list4_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list4[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list4""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_range_list4_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_tx_range_list4_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list4_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list4);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_range_list5_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_rx_range_list5_set self rx_range_list5 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list5_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_rx_range_list5_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->rx_range_list5[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""rx_range_list5""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_rx_range_list5_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_rx_range_list5_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_rx_range_list5_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->rx_range_list5);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_range_list5_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_range_t *arg2 = (freq_range_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_tx_range_list5_set self tx_range_list5 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list5_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_freq_range_list, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_tx_range_list5_set" "', argument " "2"" of type '" "freq_range_t [30]""'"); 
  } 
  arg2 = (freq_range_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)30; ++ii) *(freq_range_t *)&arg1->tx_range_list5[ii] = *((freq_range_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""tx_range_list5""' of type '""freq_range_t [30]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_tx_range_list5_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_range_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_tx_range_list5_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_tx_range_list5_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_range_t *)(freq_range_t *) ((arg1)->tx_range_list5);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_freq_range_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_spectrum_scopes_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  struct rig_spectrum_scope *arg2 = (struct rig_spectrum_scope *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_spectrum_scopes_set self spectrum_scopes ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_scopes_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_rig_spectrum_scope, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_spectrum_scopes_set" "', argument " "2"" of type '" "struct rig_spectrum_scope [4]""'"); 
  } 
  arg2 = (struct rig_spectrum_scope *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)4; ++ii) *(struct rig_spectrum_scope *)&arg1->spectrum_scopes[ii] = *((struct rig_spectrum_scope *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_scopes""' of type '""struct rig_spectrum_scope [4]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_spectrum_scopes_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct rig_spectrum_scope *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_spectrum_scopes_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_scopes_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (struct rig_spectrum_scope *)(struct rig_spectrum_scope *) ((arg1)->spectrum_scopes);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_scope,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_spectrum_modes_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  enum rig_spectrum_mode_e *arg2 = (enum rig_spectrum_mode_e *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_spectrum_modes_set self spectrum_modes ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_modes_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_rig_spectrum_mode_e, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_spectrum_modes_set" "', argument " "2"" of type '" "enum rig_spectrum_mode_e [5]""'"); 
  } 
  arg2 = (enum rig_spectrum_mode_e *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)5; ++ii) *(enum rig_spectrum_mode_e *)&arg1->spectrum_modes[ii] = *((enum rig_spectrum_mode_e *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_modes""' of type '""enum rig_spectrum_mode_e [5]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_spectrum_modes_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum rig_spectrum_mode_e *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_spectrum_modes_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_modes_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (enum rig_spectrum_mode_e *)(enum rig_spectrum_mode_e *) ((arg1)->spectrum_modes);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_mode_e,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_spectrum_spans_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  freq_t *arg2 = (freq_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_spectrum_spans_set self spectrum_spans ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_spans_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_spectrum_spans_set" "', argument " "2"" of type '" "freq_t [20]""'"); 
  } 
  arg2 = (freq_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)20; ++ii) *(freq_t *)&arg1->spectrum_spans[ii] = *((freq_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_spans""' of type '""freq_t [20]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_spectrum_spans_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_spectrum_spans_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_spans_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (freq_t *)(freq_t *) ((arg1)->spectrum_spans);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_double,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_spectrum_avg_modes_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  struct rig_spectrum_avg_mode *arg2 = (struct rig_spectrum_avg_mode *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_spectrum_avg_modes_set self spectrum_avg_modes ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_avg_modes_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_rig_spectrum_avg_mode, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_spectrum_avg_modes_set" "', argument " "2"" of type '" "struct rig_spectrum_avg_mode [12]""'"); 
  } 
  arg2 = (struct rig_spectrum_avg_mode *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)12; ++ii) *(struct rig_spectrum_avg_mode *)&arg1->spectrum_avg_modes[ii] = *((struct rig_spectrum_avg_mode *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_avg_modes""' of type '""struct rig_spectrum_avg_mode [12]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_spectrum_avg_modes_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct rig_spectrum_avg_mode *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_spectrum_avg_modes_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_avg_modes_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (struct rig_spectrum_avg_mode *)(struct rig_spectrum_avg_mode *) ((arg1)->spectrum_avg_modes);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_spectrum_avg_mode,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_spectrum_attenuator_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int *arg2 = (int *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_spectrum_attenuator_set self spectrum_attenuator ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_attenuator_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_spectrum_attenuator_set" "', argument " "2"" of type '" "int [8]""'"); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(int *)&arg1->spectrum_attenuator[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""spectrum_attenuator""' of type '""int [8]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_spectrum_attenuator_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_spectrum_attenuator_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_spectrum_attenuator_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int *)(int *) ((arg1)->spectrum_attenuator);
  {
    int i, len, l;
    len = 8;
    
    Tcl_Obj * list = Tcl_NewListObj(len, NULL);
    for(i=0; i < len; i++) {
      Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(result[i]));
    }
    Tcl_SetObjResult(interp, list);
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_morse_data_handler_thread_run_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_morse_data_handler_thread_run_set self morse_data_handler_thread_run ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_morse_data_handler_thread_run_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_morse_data_handler_thread_run_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->morse_data_handler_thread_run = (int volatile )arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_morse_data_handler_thread_run_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_morse_data_handler_thread_run_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_morse_data_handler_thread_run_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int)(int) ((arg1)->morse_data_handler_thread_run);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_morse_data_handler_priv_data_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_morse_data_handler_priv_data_set self morse_data_handler_priv_data ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_morse_data_handler_priv_data_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_morse_data_handler_priv_data_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->morse_data_handler_priv_data = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_morse_data_handler_priv_data_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_morse_data_handler_priv_data_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_morse_data_handler_priv_data_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->morse_data_handler_priv_data);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_fifo_morse_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  struct FIFO_RIG_s *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_fifo_morse_set self fifo_morse ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_fifo_morse_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_FIFO_RIG_s, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_fifo_morse_set" "', argument " "2"" of type '" "struct FIFO_RIG_s *""'"); 
  }
  arg2 = (struct FIFO_RIG_s *)(argp2);
  if (arg1) (arg1)->fifo_morse = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_fifo_morse_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct FIFO_RIG_s *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_fifo_morse_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_fifo_morse_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (struct FIFO_RIG_s *) ((arg1)->fifo_morse);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_FIFO_RIG_s,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_doppler_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_doppler_set self doppler ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_doppler_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_doppler_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->doppler = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_doppler_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_doppler_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_doppler_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->doppler);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_data_addr_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_multicast_data_addr_set self multicast_data_addr ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_data_addr_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_multicast_data_addr_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->multicast_data_addr);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->multicast_data_addr = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->multicast_data_addr = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_data_addr_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_multicast_data_addr_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_data_addr_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->multicast_data_addr);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_data_port_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_multicast_data_port_set self multicast_data_port ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_data_port_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_multicast_data_port_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->multicast_data_port = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_data_port_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_multicast_data_port_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_data_port_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->multicast_data_port);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_cmd_addr_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_multicast_cmd_addr_set self multicast_cmd_addr ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_cmd_addr_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_multicast_cmd_addr_set" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->multicast_cmd_addr);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->multicast_cmd_addr = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->multicast_cmd_addr = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_cmd_addr_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_multicast_cmd_addr_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_cmd_addr_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *) ((arg1)->multicast_cmd_addr);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_cmd_port_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_multicast_cmd_port_set self multicast_cmd_port ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_cmd_port_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_multicast_cmd_port_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->multicast_cmd_port = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_cmd_port_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_multicast_cmd_port_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_cmd_port_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->multicast_cmd_port);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_receiver_run_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_multicast_receiver_run_set self multicast_receiver_run ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_receiver_run_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_multicast_receiver_run_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->multicast_receiver_run = (int volatile )arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_receiver_run_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_multicast_receiver_run_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_receiver_run_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int)(int) ((arg1)->multicast_receiver_run);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_receiver_priv_data_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_multicast_receiver_priv_data_set self multicast_receiver_priv_data ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_receiver_priv_data_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_multicast_receiver_priv_data_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->multicast_receiver_priv_data = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_multicast_receiver_priv_data_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_multicast_receiver_priv_data_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_multicast_receiver_priv_data_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (void *) ((arg1)->multicast_receiver_priv_data);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_comm_status_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  rig_comm_status_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_comm_status_set self comm_status ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_comm_status_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_comm_status_set" "', argument " "2"" of type '" "rig_comm_status_t""'");
  } 
  arg2 = (rig_comm_status_t)(val2);
  if (arg1) (arg1)->comm_status = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_comm_status_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rig_comm_status_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_comm_status_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_comm_status_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (rig_comm_status_t) ((arg1)->comm_status);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_device_id_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  char *arg2 = (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char temp2[30] ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_device_id_set self device_id ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_device_id_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  res2 = SWIG_AsCharArray(objv[2], temp2, 30);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_device_id_set" "', argument " "2"" of type '" "char [30]""'");
  }
  arg2 = (char *)(temp2);
  if (arg2) memcpy(arg1->device_id,arg2,30*sizeof(char));
  else memset(arg1->device_id,0,30*sizeof(char));
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_device_id_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_device_id_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_device_id_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (char *)(char *) ((arg1)->device_id);
  {
    size_t size = SWIG_strnlen(result, 30);
    
    
    
    Tcl_SetObjResult(interp,SWIG_FromCharPtrAndSize(result, size));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_dual_watch_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_dual_watch_set self dual_watch ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dual_watch_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_dual_watch_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->dual_watch = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_dual_watch_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_dual_watch_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_dual_watch_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->dual_watch);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_post_ptt_delay_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_post_ptt_delay_set self post_ptt_delay ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_post_ptt_delay_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_post_ptt_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->post_ptt_delay = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_post_ptt_delay_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_post_ptt_delay_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_post_ptt_delay_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->post_ptt_delay);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_freq_event_elapsed_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  struct timespec arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_freq_event_elapsed_set self freq_event_elapsed ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_freq_event_elapsed_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_timespec,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_freq_event_elapsed_set" "', argument " "2"" of type '" "struct timespec""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_freq_event_elapsed_set" "', argument " "2"" of type '" "struct timespec""'");
    } else {
      arg2 = *((struct timespec *)(argp2));
    }
  }
  if (arg1) (arg1)->freq_event_elapsed = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_freq_event_elapsed_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct timespec result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_freq_event_elapsed_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_freq_event_elapsed_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->freq_event_elapsed);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((struct timespec *)memcpy((struct timespec *)calloc(1,sizeof(struct timespec)),&result,sizeof(struct timespec)), SWIGTYPE_p_timespec, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_freq_skip_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_freq_skip_set self freq_skip ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_freq_skip_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_freq_skip_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->freq_skip = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_freq_skip_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_freq_skip_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_freq_skip_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (int) ((arg1)->freq_skip);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_client_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  client_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_client_set self client ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_client_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rig_state_client_set" "', argument " "2"" of type '" "client_t""'");
  } 
  arg2 = (client_t)(val2);
  if (arg1) (arg1)->client = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_client_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  client_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_client_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_client_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result = (client_t) ((arg1)->client);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_api_mutex_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  pthread_mutex_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rig_state_api_mutex_set self api_mutex ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_api_mutex_set" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_pthread_mutex_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rig_state_api_mutex_set" "', argument " "2"" of type '" "pthread_mutex_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rig_state_api_mutex_set" "', argument " "2"" of type '" "pthread_mutex_t""'");
    } else {
      arg2 = *((pthread_mutex_t *)(argp2));
    }
  }
  if (arg1) (arg1)->api_mutex = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rig_state_api_mutex_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  pthread_mutex_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rig_state_api_mutex_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rig_state_api_mutex_get" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  result =  ((arg1)->api_mutex);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((pthread_mutex_t *)memcpy((pthread_mutex_t *)calloc(1,sizeof(pthread_mutex_t)),&result,sizeof(pthread_mutex_t)), SWIGTYPE_p_pthread_mutex_t, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_rig_state(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_rig_state ") == TCL_ERROR) SWIG_fail;
  result = (struct rig_state *)calloc(1, sizeof(struct rig_state));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_state,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_rig_state(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rig_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_rig_state self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rig_state, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rig_state" "', argument " "1"" of type '" "struct rig_state *""'"); 
  }
  arg1 = (struct rig_state *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_rig_state(void *obj) {
struct rig_state *arg1 = (struct rig_state *) obj;
free((char *) arg1);
}
static swig_method swig_rig_state_methods[] = {
    {0,0}
};
static swig_attribute swig_rig_state_attributes[] = {
    { "-vfo_comp",_wrap_rig_state_vfo_comp_get, _wrap_rig_state_vfo_comp_set},
    { "-rx_range_list",_wrap_rig_state_rx_range_list_get, _wrap_rig_state_rx_range_list_set},
    { "-tx_range_list",_wrap_rig_state_tx_range_list_get, _wrap_rig_state_tx_range_list_set},
    { "-tuning_steps",_wrap_rig_state_tuning_steps_get, _wrap_rig_state_tuning_steps_set},
    { "-filters",_wrap_rig_state_filters_get, _wrap_rig_state_filters_set},
    { "-str_cal",_wrap_rig_state_str_cal_get, _wrap_rig_state_str_cal_set},
    { "-chan_list",_wrap_rig_state_chan_list_get, _wrap_rig_state_chan_list_set},
    { "-max_rit",_wrap_rig_state_max_rit_get, _wrap_rig_state_max_rit_set},
    { "-max_xit",_wrap_rig_state_max_xit_get, _wrap_rig_state_max_xit_set},
    { "-max_ifshift",_wrap_rig_state_max_ifshift_get, _wrap_rig_state_max_ifshift_set},
    { "-announces",_wrap_rig_state_announces_get, _wrap_rig_state_announces_set},
    { "-preamp",_wrap_rig_state_preamp_get, _wrap_rig_state_preamp_set},
    { "-attenuator",_wrap_rig_state_attenuator_get, _wrap_rig_state_attenuator_set},
    { "-has_get_func",_wrap_rig_state_has_get_func_get, _wrap_rig_state_has_get_func_set},
    { "-has_set_func",_wrap_rig_state_has_set_func_get, _wrap_rig_state_has_set_func_set},
    { "-has_get_level",_wrap_rig_state_has_get_level_get, _wrap_rig_state_has_get_level_set},
    { "-has_set_level",_wrap_rig_state_has_set_level_get, _wrap_rig_state_has_set_level_set},
    { "-has_get_parm",_wrap_rig_state_has_get_parm_get, _wrap_rig_state_has_get_parm_set},
    { "-has_set_parm",_wrap_rig_state_has_set_parm_get, _wrap_rig_state_has_set_parm_set},
    { "-level_gran",_wrap_rig_state_level_gran_get, _wrap_rig_state_level_gran_set},
    { "-parm_gran",_wrap_rig_state_parm_gran_get, _wrap_rig_state_parm_gran_set},
    { "-transaction_active",_wrap_rig_state_transaction_active_get, _wrap_rig_state_transaction_active_set},
    { "-current_vfo",_wrap_rig_state_current_vfo_get, _wrap_rig_state_current_vfo_set},
    { "-vfo_list",_wrap_rig_state_vfo_list_get, _wrap_rig_state_vfo_list_set},
    { "-comm_state",_wrap_rig_state_comm_state_get, _wrap_rig_state_comm_state_set},
    { "-priv",_wrap_rig_state_priv_get, _wrap_rig_state_priv_set},
    { "-obj",_wrap_rig_state_obj_get, _wrap_rig_state_obj_set},
    { "-async_data_enabled",_wrap_rig_state_async_data_enabled_get, _wrap_rig_state_async_data_enabled_set},
    { "-poll_interval",_wrap_rig_state_poll_interval_get, _wrap_rig_state_poll_interval_set},
    { "-current_freq",_wrap_rig_state_current_freq_get, _wrap_rig_state_current_freq_set},
    { "-current_mode",_wrap_rig_state_current_mode_get, _wrap_rig_state_current_mode_set},
    { "-current_width",_wrap_rig_state_current_width_get, _wrap_rig_state_current_width_set},
    { "-tx_vfo",_wrap_rig_state_tx_vfo_get, _wrap_rig_state_tx_vfo_set},
    { "-mode_list",_wrap_rig_state_mode_list_get, _wrap_rig_state_mode_list_set},
    { "-transmit",_wrap_rig_state_transmit_get, _wrap_rig_state_transmit_set},
    { "-lo_freq",_wrap_rig_state_lo_freq_get, _wrap_rig_state_lo_freq_set},
    { "-twiddle_time",_wrap_rig_state_twiddle_time_get, _wrap_rig_state_twiddle_time_set},
    { "-twiddle_timeout",_wrap_rig_state_twiddle_timeout_get, _wrap_rig_state_twiddle_timeout_set},
    { "-uplink",_wrap_rig_state_uplink_get, _wrap_rig_state_uplink_set},
    { "-vfo_opt",_wrap_rig_state_vfo_opt_get, _wrap_rig_state_vfo_opt_set},
    { "-auto_power_on",_wrap_rig_state_auto_power_on_get, _wrap_rig_state_auto_power_on_set},
    { "-auto_power_off",_wrap_rig_state_auto_power_off_get, _wrap_rig_state_auto_power_off_set},
    { "-auto_disable_screensaver",_wrap_rig_state_auto_disable_screensaver_get, _wrap_rig_state_auto_disable_screensaver_set},
    { "-ptt_share",_wrap_rig_state_ptt_share_get, _wrap_rig_state_ptt_share_set},
    { "-power_now",_wrap_rig_state_power_now_get, _wrap_rig_state_power_now_set},
    { "-power_min",_wrap_rig_state_power_min_get, _wrap_rig_state_power_min_set},
    { "-power_max",_wrap_rig_state_power_max_get, _wrap_rig_state_power_max_set},
    { "-disable_yaesu_bandselect",_wrap_rig_state_disable_yaesu_bandselect_get, _wrap_rig_state_disable_yaesu_bandselect_set},
    { "-twiddle_rit",_wrap_rig_state_twiddle_rit_get, _wrap_rig_state_twiddle_rit_set},
    { "-twiddle_state",_wrap_rig_state_twiddle_state_get, _wrap_rig_state_twiddle_state_set},
    { "-rx_vfo",_wrap_rig_state_rx_vfo_get, _wrap_rig_state_rx_vfo_set},
    { "-snapshot_packet_sequence_number",_wrap_rig_state_snapshot_packet_sequence_number_get, _wrap_rig_state_snapshot_packet_sequence_number_set},
    { "-multicast_publisher_run",_wrap_rig_state_multicast_publisher_run_get, _wrap_rig_state_multicast_publisher_run_set},
    { "-multicast_publisher_priv_data",_wrap_rig_state_multicast_publisher_priv_data_get, _wrap_rig_state_multicast_publisher_priv_data_set},
    { "-async_data_handler_thread_run",_wrap_rig_state_async_data_handler_thread_run_get, _wrap_rig_state_async_data_handler_thread_run_set},
    { "-async_data_handler_priv_data",_wrap_rig_state_async_data_handler_priv_data_get, _wrap_rig_state_async_data_handler_priv_data_set},
    { "-poll_routine_thread_run",_wrap_rig_state_poll_routine_thread_run_get, _wrap_rig_state_poll_routine_thread_run_set},
    { "-poll_routine_priv_data",_wrap_rig_state_poll_routine_priv_data_get, _wrap_rig_state_poll_routine_priv_data_set},
    { "-mutex_set_transaction",_wrap_rig_state_mutex_set_transaction_get, _wrap_rig_state_mutex_set_transaction_set},
    { "-rigport",_wrap_rig_state_rigport_get, _wrap_rig_state_rigport_set},
    { "-pttport",_wrap_rig_state_pttport_get, _wrap_rig_state_pttport_set},
    { "-dcdport",_wrap_rig_state_dcdport_get, _wrap_rig_state_dcdport_set},
    { "-use_cached_freq",_wrap_rig_state_use_cached_freq_get, _wrap_rig_state_use_cached_freq_set},
    { "-use_cached_mode",_wrap_rig_state_use_cached_mode_get, _wrap_rig_state_use_cached_mode_set},
    { "-use_cached_ptt",_wrap_rig_state_use_cached_ptt_get, _wrap_rig_state_use_cached_ptt_set},
    { "-depth",_wrap_rig_state_depth_get, _wrap_rig_state_depth_set},
    { "-lock_mode",_wrap_rig_state_lock_mode_get, _wrap_rig_state_lock_mode_set},
    { "-powerstat",_wrap_rig_state_powerstat_get, _wrap_rig_state_powerstat_set},
    { "-tuner_control_pathname",_wrap_rig_state_tuner_control_pathname_get, _wrap_rig_state_tuner_control_pathname_set},
    { "-client_version",_wrap_rig_state_client_version_get, _wrap_rig_state_client_version_set},
    { "-offset_vfoa",_wrap_rig_state_offset_vfoa_get, _wrap_rig_state_offset_vfoa_set},
    { "-offset_vfob",_wrap_rig_state_offset_vfob_get, _wrap_rig_state_offset_vfob_set},
    { "-multicast",_wrap_rig_state_multicast_get, _wrap_rig_state_multicast_set},
    { "-rig_model",_wrap_rig_state_rig_model_get, _wrap_rig_state_rig_model_set},
    { "-model_name",_wrap_rig_state_model_name_get, _wrap_rig_state_model_name_set},
    { "-mfg_name",_wrap_rig_state_mfg_name_get, _wrap_rig_state_mfg_name_set},
    { "-version",_wrap_rig_state_version_get, _wrap_rig_state_version_set},
    { "-copyright",_wrap_rig_state_copyright_get, _wrap_rig_state_copyright_set},
    { "-status",_wrap_rig_state_status_get, _wrap_rig_state_status_set},
    { "-rig_type",_wrap_rig_state_rig_type_get, _wrap_rig_state_rig_type_set},
    { "-ptt_type",_wrap_rig_state_ptt_type_get, _wrap_rig_state_ptt_type_set},
    { "-dcd_type",_wrap_rig_state_dcd_type_get, _wrap_rig_state_dcd_type_set},
    { "-port_type",_wrap_rig_state_port_type_get, _wrap_rig_state_port_type_set},
    { "-serial_rate_min",_wrap_rig_state_serial_rate_min_get, _wrap_rig_state_serial_rate_min_set},
    { "-serial_rate_max",_wrap_rig_state_serial_rate_max_get, _wrap_rig_state_serial_rate_max_set},
    { "-serial_data_bits",_wrap_rig_state_serial_data_bits_get, _wrap_rig_state_serial_data_bits_set},
    { "-serial_stop_bits",_wrap_rig_state_serial_stop_bits_get, _wrap_rig_state_serial_stop_bits_set},
    { "-serial_parity",_wrap_rig_state_serial_parity_get, _wrap_rig_state_serial_parity_set},
    { "-serial_handshake",_wrap_rig_state_serial_handshake_get, _wrap_rig_state_serial_handshake_set},
    { "-write_delay",_wrap_rig_state_write_delay_get, _wrap_rig_state_write_delay_set},
    { "-post_write_delay",_wrap_rig_state_post_write_delay_get, _wrap_rig_state_post_write_delay_set},
    { "-timeout",_wrap_rig_state_timeout_get, _wrap_rig_state_timeout_set},
    { "-retry",_wrap_rig_state_retry_get, _wrap_rig_state_retry_set},
    { "-targetable_vfo",_wrap_rig_state_targetable_vfo_get, _wrap_rig_state_targetable_vfo_set},
    { "-async_data_supported",_wrap_rig_state_async_data_supported_get, _wrap_rig_state_async_data_supported_set},
    { "-agc_level_count",_wrap_rig_state_agc_level_count_get, _wrap_rig_state_agc_level_count_set},
    { "-agc_levels",_wrap_rig_state_agc_levels_get, _wrap_rig_state_agc_levels_set},
    { "-ctcss_list",_wrap_rig_state_ctcss_list_get, _wrap_rig_state_ctcss_list_set},
    { "-dcs_list",_wrap_rig_state_dcs_list_get, _wrap_rig_state_dcs_list_set},
    { "-vfo_ops",_wrap_rig_state_vfo_ops_get, _wrap_rig_state_vfo_ops_set},
    { "-scan_ops",_wrap_rig_state_scan_ops_get, _wrap_rig_state_scan_ops_set},
    { "-transceive",_wrap_rig_state_transceive_get, _wrap_rig_state_transceive_set},
    { "-bank_qty",_wrap_rig_state_bank_qty_get, _wrap_rig_state_bank_qty_set},
    { "-chan_desc_sz",_wrap_rig_state_chan_desc_sz_get, _wrap_rig_state_chan_desc_sz_set},
    { "-rx_range_list1",_wrap_rig_state_rx_range_list1_get, _wrap_rig_state_rx_range_list1_set},
    { "-tx_range_list1",_wrap_rig_state_tx_range_list1_get, _wrap_rig_state_tx_range_list1_set},
    { "-rx_range_list2",_wrap_rig_state_rx_range_list2_get, _wrap_rig_state_rx_range_list2_set},
    { "-tx_range_list2",_wrap_rig_state_tx_range_list2_get, _wrap_rig_state_tx_range_list2_set},
    { "-rx_range_list3",_wrap_rig_state_rx_range_list3_get, _wrap_rig_state_rx_range_list3_set},
    { "-tx_range_list3",_wrap_rig_state_tx_range_list3_get, _wrap_rig_state_tx_range_list3_set},
    { "-rx_range_list4",_wrap_rig_state_rx_range_list4_get, _wrap_rig_state_rx_range_list4_set},
    { "-tx_range_list4",_wrap_rig_state_tx_range_list4_get, _wrap_rig_state_tx_range_list4_set},
    { "-rx_range_list5",_wrap_rig_state_rx_range_list5_get, _wrap_rig_state_rx_range_list5_set},
    { "-tx_range_list5",_wrap_rig_state_tx_range_list5_get, _wrap_rig_state_tx_range_list5_set},
    { "-spectrum_scopes",_wrap_rig_state_spectrum_scopes_get, _wrap_rig_state_spectrum_scopes_set},
    { "-spectrum_modes",_wrap_rig_state_spectrum_modes_get, _wrap_rig_state_spectrum_modes_set},
    { "-spectrum_spans",_wrap_rig_state_spectrum_spans_get, _wrap_rig_state_spectrum_spans_set},
    { "-spectrum_avg_modes",_wrap_rig_state_spectrum_avg_modes_get, _wrap_rig_state_spectrum_avg_modes_set},
    { "-spectrum_attenuator",_wrap_rig_state_spectrum_attenuator_get, _wrap_rig_state_spectrum_attenuator_set},
    { "-morse_data_handler_thread_run",_wrap_rig_state_morse_data_handler_thread_run_get, _wrap_rig_state_morse_data_handler_thread_run_set},
    { "-morse_data_handler_priv_data",_wrap_rig_state_morse_data_handler_priv_data_get, _wrap_rig_state_morse_data_handler_priv_data_set},
    { "-fifo_morse",_wrap_rig_state_fifo_morse_get, _wrap_rig_state_fifo_morse_set},
    { "-doppler",_wrap_rig_state_doppler_get, _wrap_rig_state_doppler_set},
    { "-multicast_data_addr",_wrap_rig_state_multicast_data_addr_get, _wrap_rig_state_multicast_data_addr_set},
    { "-multicast_data_port",_wrap_rig_state_multicast_data_port_get, _wrap_rig_state_multicast_data_port_set},
    { "-multicast_cmd_addr",_wrap_rig_state_multicast_cmd_addr_get, _wrap_rig_state_multicast_cmd_addr_set},
    { "-multicast_cmd_port",_wrap_rig_state_multicast_cmd_port_get, _wrap_rig_state_multicast_cmd_port_set},
    { "-multicast_receiver_run",_wrap_rig_state_multicast_receiver_run_get, _wrap_rig_state_multicast_receiver_run_set},
    { "-multicast_receiver_priv_data",_wrap_rig_state_multicast_receiver_priv_data_get, _wrap_rig_state_multicast_receiver_priv_data_set},
    { "-comm_status",_wrap_rig_state_comm_status_get, _wrap_rig_state_comm_status_set},
    { "-device_id",_wrap_rig_state_device_id_get, _wrap_rig_state_device_id_set},
    { "-dual_watch",_wrap_rig_state_dual_watch_get, _wrap_rig_state_dual_watch_set},
    { "-post_ptt_delay",_wrap_rig_state_post_ptt_delay_get, _wrap_rig_state_post_ptt_delay_set},
    { "-freq_event_elapsed",_wrap_rig_state_freq_event_elapsed_get, _wrap_rig_state_freq_event_elapsed_set},
    { "-freq_skip",_wrap_rig_state_freq_skip_get, _wrap_rig_state_freq_skip_set},
    { "-client",_wrap_rig_state_client_get, _wrap_rig_state_client_set},
    { "-api_mutex",_wrap_rig_state_api_mutex_get, _wrap_rig_state_api_mutex_set},
    {0,0,0}
};
static swig_class *swig_rig_state_bases[] = {0};
static const char * swig_rig_state_base_names[] = {0};
static swig_class _wrap_class_rig_state = { "rig_state", &SWIGTYPE_p_rig_state,_wrap_new_rig_state, swig_delete_rig_state, swig_rig_state_methods, swig_rig_state_attributes, swig_rig_state_bases,swig_rig_state_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_Rig_rig_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  RIG *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_rig_set self rig ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_rig_set" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_s_rig, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_rig_set" "', argument " "2"" of type '" "RIG *""'"); 
  }
  arg2 = (RIG *)(argp2);
  if (arg1) (arg1)->rig = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_rig_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  RIG *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rig_rig_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_rig_get" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  result = (RIG *) ((arg1)->rig);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_s_rig,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_caps_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  struct rig_caps *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_caps_set self caps ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_caps_set" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_rig_caps, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_caps_set" "', argument " "2"" of type '" "struct rig_caps *""'"); 
  }
  arg2 = (struct rig_caps *)(argp2);
  if (arg1) (arg1)->caps = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_caps_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct rig_caps *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rig_caps_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_caps_get" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  result = (struct rig_caps *) ((arg1)->caps);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_caps,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_state_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  struct rig_state *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_state_set self state ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_state_set" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_rig_state, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_state_set" "', argument " "2"" of type '" "struct rig_state *""'"); 
  }
  arg2 = (struct rig_state *)(argp2);
  if (arg1) (arg1)->state = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_state_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct rig_state *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rig_state_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_state_get" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  result = (struct rig_state *) ((arg1)->state);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rig_state,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_error_status_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_error_status_set self error_status ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_error_status_set" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_error_status_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->error_status = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_error_status_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rig_error_status_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_error_status_get" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  result = (int) ((arg1)->error_status);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_do_exception_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_do_exception_set self do_exception ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_do_exception_set" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_do_exception_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->do_exception = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_do_exception_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rig_do_exception_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_do_exception_get" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  result = (int) ((arg1)->do_exception);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_Rig(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  struct Rig *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:new_Rig rig_model ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Rig" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = (int)(val1);
  result = (struct Rig *)new_Rig(arg1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_Rig,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_Rig(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_Rig self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Rig" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  delete_Rig(arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_open(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rig_open self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_open" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rig_open(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_close(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rig_close self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_close" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rig_close(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_freq(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  freq_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_freq self vfo _freq_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_freq" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_freq" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_freq" "', argument " "3"" of type '" "freq_t""'");
  } 
  arg3 = (freq_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_freq(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_mode(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  rmode_t arg2 ;
  pbwidth_t arg3 = ((shortfreq_t) (0)) ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo|oo:Rig_set_mode self _rmode_t_1 ?_pbwidth_t_2? ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_mode" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_mode" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (objc > 3) {
    ecode3 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_mode" "', argument " "3"" of type '" "pbwidth_t""'");
    } 
    arg3 = (pbwidth_t)(val3);
  }
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_mode" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_mode(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_ptt(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  ptt_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_ptt self vfo _ptt_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ptt" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ptt" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_ptt" "', argument " "3"" of type '" "ptt_t""'");
  } 
  arg3 = (ptt_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_ptt(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_rptr_shift(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  rptr_shift_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_rptr_shift self vfo _rptr_shift_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_rptr_shift" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_rptr_shift" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_rptr_shift" "', argument " "3"" of type '" "rptr_shift_t""'");
  } 
  arg3 = (rptr_shift_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_rptr_shift(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_rptr_offs(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  shortfreq_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_rptr_offs self vfo _shortfreq_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_rptr_offs" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_rptr_offs" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_rptr_offs" "', argument " "3"" of type '" "shortfreq_t""'");
  } 
  arg3 = (shortfreq_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_rptr_offs(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_ctcss_tone(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  tone_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_ctcss_tone self vfo _tone_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ctcss_tone" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ctcss_tone" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_ctcss_tone" "', argument " "3"" of type '" "tone_t""'");
  } 
  arg3 = (tone_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_ctcss_tone(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_dcs_code(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  tone_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_dcs_code self vfo _tone_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_dcs_code" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_dcs_code" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_dcs_code" "', argument " "3"" of type '" "tone_t""'");
  } 
  arg3 = (tone_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_dcs_code(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_ctcss_sql(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  tone_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_ctcss_sql self vfo _tone_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ctcss_sql" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ctcss_sql" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_ctcss_sql" "', argument " "3"" of type '" "tone_t""'");
  } 
  arg3 = (tone_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_ctcss_sql(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_dcs_sql(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  tone_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_dcs_sql self vfo _tone_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_dcs_sql" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_dcs_sql" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_dcs_sql" "', argument " "3"" of type '" "tone_t""'");
  } 
  arg3 = (tone_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_dcs_sql(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_split_freq(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  freq_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_split_freq self vfo _freq_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_split_freq" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_split_freq" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_split_freq" "', argument " "3"" of type '" "freq_t""'");
  } 
  arg3 = (freq_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_split_freq(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_split_mode(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  rmode_t arg2 ;
  pbwidth_t arg3 = ((shortfreq_t) (0)) ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo|oo:Rig_set_split_mode self _rmode_t_1 ?_pbwidth_t_2? ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_split_mode" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_split_mode" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  if (objc > 3) {
    ecode3 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_split_mode" "', argument " "3"" of type '" "pbwidth_t""'");
    } 
    arg3 = (pbwidth_t)(val3);
  }
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_split_mode" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_split_mode(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_split_freq_mode(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  freq_t arg2 ;
  rmode_t arg3 ;
  pbwidth_t arg4 = ((shortfreq_t) (0)) ;
  vfo_t arg5 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  unsigned long long val3 ;
  int ecode3 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo|oo:Rig_set_split_freq_mode self _freq_t_1 _rmode_t_2 ?_pbwidth_t_3? ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_split_freq_mode" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_split_freq_mode" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_split_freq_mode" "', argument " "3"" of type '" "rmode_t""'");
  } 
  arg3 = (rmode_t)(val3);
  if (objc > 4) {
    ecode4 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_split_freq_mode" "', argument " "4"" of type '" "pbwidth_t""'");
    } 
    arg4 = (pbwidth_t)(val4);
  }
  if (objc > 5) {
    ecode5 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[5], &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Rig_set_split_freq_mode" "', argument " "5"" of type '" "vfo_t""'");
    } 
    arg5 = (vfo_t)(val5);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_split_freq_mode(arg1,arg2,arg3,arg4,arg5);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_split_vfo(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  split_t arg2 ;
  vfo_t arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo|o:Rig_set_split_vfo self _split_t_1 _vfo_t_2 ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_split_vfo" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_split_vfo" "', argument " "2"" of type '" "split_t""'");
  } 
  arg2 = (split_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_split_vfo" "', argument " "3"" of type '" "vfo_t""'");
  } 
  arg3 = (vfo_t)(val3);
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_split_vfo" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_split_vfo(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_rit(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  shortfreq_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_rit self vfo _shortfreq_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_rit" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_rit" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_rit" "', argument " "3"" of type '" "shortfreq_t""'");
  } 
  arg3 = (shortfreq_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_rit(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_xit(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  shortfreq_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_xit self vfo _shortfreq_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_xit" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_xit" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_xit" "', argument " "3"" of type '" "shortfreq_t""'");
  } 
  arg3 = (shortfreq_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_xit(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_ts(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  shortfreq_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_ts self vfo _shortfreq_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ts" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ts" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_ts" "', argument " "3"" of type '" "shortfreq_t""'");
  } 
  arg3 = (shortfreq_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_ts(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_ant(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  ant_t arg2 ;
  value_t arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo|o:Rig_set_ant self _ant_t_1 _value_t_2 ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ant" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ant" "', argument " "2"" of type '" "ant_t""'");
  } 
  arg2 = (ant_t)(val2);
  {
    res3 = SWIG_ConvertPtr(objv[3], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_ant" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rig_set_ant" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_ant" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_ant(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_func(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  int arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo|o:Rig_set_func self _setting_t_1 _int_2 ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_func" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_func" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_func" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_func(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_ext_func(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  int arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo|o:Rig_set_ext_func self _hamlib_token_t_1 _int_2 ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ext_func" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ext_func" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_ext_func" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_ext_func" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_ext_func(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_bank(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_bank self vfo _int ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_bank" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_bank" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_bank" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_bank(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_mem(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_mem self vfo _int ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_mem" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_mem" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_mem" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_mem(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_send_dtmf(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  const_char_string arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_send_dtmf self vfo _const_char_string ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_send_dtmf" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_send_dtmf" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(objv[3], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_send_dtmf" "', argument " "3"" of type '" "const_char_string""'");
  }
  arg3 = (const_char_string)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rig_send_dtmf(arg1,arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_OK;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_send_morse(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  const_char_string arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_send_morse self vfo _const_char_string ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_send_morse" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_send_morse" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(objv[3], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_send_morse" "', argument " "3"" of type '" "const_char_string""'");
  }
  arg3 = (const_char_string)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rig_send_morse(arg1,arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_OK;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_vfo_op(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  vfo_op_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_vfo_op self vfo _vfo_op_t ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_vfo_op" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_vfo_op" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_vfo_op" "', argument " "3"" of type '" "vfo_op_t""'");
  } 
  arg3 = (vfo_op_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_vfo_op(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_scan(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  scan_t arg2 ;
  int arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo|o:Rig_scan self _scan_t_1 _int_2 ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_scan" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_scan" "', argument " "2"" of type '" "scan_t""'");
  } 
  arg2 = (scan_t)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_scan" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_scan" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_scan(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_level__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  int arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo|o:Rig_set_level self stg i ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_level" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_level__SWIG_0(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_level__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  float arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo|o:Rig_set_level self stg f ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  ecode3 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_level" "', argument " "3"" of type '" "float""'");
  } 
  arg3 = (float)(val3);
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_level__SWIG_1(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_level__SWIG_2(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo|o:Rig_set_level self name i ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_level" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_level" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_level__SWIG_2(arg1,(char const *)arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_level__SWIG_3(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  float arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo|o:Rig_set_level self name f ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_level" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_level" "', argument " "3"" of type '" "float""'");
  } 
  arg3 = (float)(val3);
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_level__SWIG_3(arg1,(char const *)arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_level__SWIG_4(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo|o:Rig_set_level self name s ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_level" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(objv[3], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_level" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_level__SWIG_4(arg1,(char const *)arg2,(char const *)arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if ((argc >= 3) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          if (argc <= 3) {
            return _wrap_Rig_set_level__SWIG_0(clientData, interp, objc, argv - 1);
          }
          {
            int res = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Rig_set_level__SWIG_0(clientData, interp, objc, argv - 1);
          }
        }
      }
    }
  }
  if ((argc >= 3) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          if (argc <= 3) {
            return _wrap_Rig_set_level__SWIG_1(clientData, interp, objc, argv - 1);
          }
          {
            int res = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Rig_set_level__SWIG_1(clientData, interp, objc, argv - 1);
          }
        }
      }
    }
  }
  if ((argc >= 3) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          if (argc <= 3) {
            return _wrap_Rig_set_level__SWIG_2(clientData, interp, objc, argv - 1);
          }
          {
            int res = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Rig_set_level__SWIG_2(clientData, interp, objc, argv - 1);
          }
        }
      }
    }
  }
  if ((argc >= 3) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          if (argc <= 3) {
            return _wrap_Rig_set_level__SWIG_3(clientData, interp, objc, argv - 1);
          }
          {
            int res = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Rig_set_level__SWIG_3(clientData, interp, objc, argv - 1);
          }
        }
      }
    }
  }
  if ((argc >= 3) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          if (argc <= 3) {
            return _wrap_Rig_set_level__SWIG_4(clientData, interp, objc, argv - 1);
          }
          {
            int res = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Rig_set_level__SWIG_4(clientData, interp, objc, argv - 1);
          }
        }
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rig_set_level'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::set_level(setting_t,int,vfo_t)\n"
    "    Rig::set_level(setting_t,float,vfo_t)\n"
    "    Rig::set_level(char const *,int,vfo_t)\n"
    "    Rig::set_level(char const *,float,vfo_t)\n"
    "    Rig::set_level(char const *,char const *,vfo_t)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_ext_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  value_t arg3 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo|o:Rig_set_ext_level self _hamlib_token_t_1 _value_t_2 ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ext_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ext_level" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    res3 = SWIG_ConvertPtr(objv[3], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_ext_level" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rig_set_ext_level" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_set_ext_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_ext_level(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_level_i__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo|o:Rig_get_level_i self stg ?vfo? ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_level_i" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_level_i" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (objc > 3) {
    ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_level_i" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_level_i__SWIG_0(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_level_f__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  float result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo|o:Rig_get_level_f self stg ?vfo? ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_level_f" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_level_f" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (objc > 3) {
    ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_level_f" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    result = (float)Rig_get_level_f__SWIG_0(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_level_i__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo|o:Rig_get_level_i self name ?vfo? ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_level_i" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_level_i" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (objc > 3) {
    ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_level_i" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_level_i__SWIG_1(arg1,(char const *)arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_level_i(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_level_i__SWIG_0(clientData, interp, objc, argv - 1);
        }
        {
          int res = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_level_i__SWIG_0(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_level_i__SWIG_1(clientData, interp, objc, argv - 1);
        }
        {
          int res = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_level_i__SWIG_1(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rig_get_level_i'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_level_i(setting_t,vfo_t)\n"
    "    Rig::get_level_i(char const *,vfo_t)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_level_f__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  float result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo|o:Rig_get_level_f self name ?vfo? ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_level_f" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_level_f" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (objc > 3) {
    ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_level_f" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    result = (float)Rig_get_level_f__SWIG_1(arg1,(char const *)arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_level_f(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_level_f__SWIG_0(clientData, interp, objc, argv - 1);
        }
        {
          int res = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_level_f__SWIG_0(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_level_f__SWIG_1(clientData, interp, objc, argv - 1);
        }
        {
          int res = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_level_f__SWIG_1(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rig_get_level_f'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_level_f(setting_t,vfo_t)\n"
    "    Rig::get_level_f(char const *,vfo_t)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_level__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char temp3[256+1] ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  arg3 = (char *) temp3;
  if (SWIG_GetArgs(interp, objc, objv,"oo|o:Rig_get_level self name ?vfo? ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_level" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (objc > 3) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_get_level" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_level__SWIG_0(arg1,(char const *)arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  arg3[256] = 0;  
  Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_FromCharPtr(arg3));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_vfo(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_set_vfo self _vfo_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_vfo" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_vfo" "', argument " "2"" of type '" "vfo_t""'");
  } 
  arg2 = (vfo_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_set_vfo(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_powerstat(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  powerstat_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_set_powerstat self _powerstat_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_powerstat" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_powerstat" "', argument " "2"" of type '" "powerstat_t""'");
  } 
  arg2 = (powerstat_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_set_powerstat(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_has_get_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_has_get_level self _setting_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_get_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_get_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_get_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_has_get_parm(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_has_get_parm self _setting_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_get_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_get_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_get_parm(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_has_set_parm(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_has_set_parm self _setting_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_set_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_set_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_set_parm(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_has_get_func(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_has_get_func self _setting_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_get_func" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_get_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_get_func(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_has_set_func(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_has_set_func self _setting_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_set_func" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_set_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_set_func(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_reset(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  reset_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_reset self _reset_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_reset" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_reset" "', argument " "2"" of type '" "reset_t""'");
  } 
  arg2 = (reset_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_reset(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_has_scan(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  scan_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_has_scan self _scan_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_scan" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_scan" "', argument " "2"" of type '" "scan_t""'");
  } 
  arg2 = (scan_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_scan(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_has_vfo_op(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_op_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_has_vfo_op self _vfo_op_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_has_vfo_op" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_has_vfo_op" "', argument " "2"" of type '" "vfo_op_t""'");
  } 
  arg2 = (vfo_op_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_has_vfo_op(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_passband_normal(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_passband_normal self _rmode_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_passband_normal" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_passband_normal" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_passband_normal(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_passband_narrow(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_passband_narrow self _rmode_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_passband_narrow" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_passband_narrow" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_passband_narrow(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_passband_wide(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  rmode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_passband_wide self _rmode_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_passband_wide" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_passband_wide" "', argument " "2"" of type '" "rmode_t""'");
  } 
  arg2 = (rmode_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_passband_wide(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_vfo_opt(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_set_vfo_opt self _int ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_vfo_opt" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_vfo_opt" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_set_vfo_opt(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_ext_token_lookup(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  const_char_string arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_ext_token_lookup self _const_char_string ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_ext_token_lookup" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_ext_token_lookup" "', argument " "2"" of type '" "const_char_string""'");
  }
  arg2 = (const_char_string)(buf2);
  {
    arg1->error_status = RIG_OK;
    Rig_ext_token_lookup(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_token_lookup(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  const_char_string arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_token_lookup self _const_char_string ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_token_lookup" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_token_lookup" "', argument " "2"" of type '" "const_char_string""'");
  }
  arg2 = (const_char_string)(buf2);
  {
    arg1->error_status = RIG_OK;
    Rig_token_lookup(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_conf__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  const_char_string arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_conf self _hamlib_token_t_1 _const_char_string_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_conf" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_conf" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(objv[3], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_conf" "', argument " "3"" of type '" "const_char_string""'");
  }
  arg3 = (const_char_string)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_conf__SWIG_0(arg1,arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_OK;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_ext_parm(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_ext_parm self _hamlib_token_t_1 _value_t_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_ext_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_ext_parm" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    res3 = SWIG_ConvertPtr(objv[3], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_ext_parm" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rig_set_ext_parm" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Rig_set_ext_parm(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_parm__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_parm self stg i ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_parm" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_parm__SWIG_0(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_parm__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  float arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_parm self stg f ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_set_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  ecode3 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_parm" "', argument " "3"" of type '" "float""'");
  } 
  arg3 = (float)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_parm__SWIG_1(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_parm__SWIG_2(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_parm self name i ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_parm" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_parm" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_parm__SWIG_2(arg1,(char const *)arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_parm__SWIG_3(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  float arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_parm self name f ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_parm" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_set_parm" "', argument " "3"" of type '" "float""'");
  } 
  arg3 = (float)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_parm__SWIG_3(arg1,(char const *)arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_parm__SWIG_4(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_parm self name s ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_parm" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(objv[3], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_parm" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_parm__SWIG_4(arg1,(char const *)arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_parm(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_set_parm__SWIG_0(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_set_parm__SWIG_1(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_set_parm__SWIG_2(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_set_parm__SWIG_3(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Rig_set_parm__SWIG_4(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rig_set_parm'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::set_parm(setting_t,int)\n"
    "    Rig::set_parm(setting_t,float)\n"
    "    Rig::set_parm(char const *,int)\n"
    "    Rig::set_parm(char const *,float)\n"
    "    Rig::set_parm(char const *,char const *)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_parm_i__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_get_parm_i self stg ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_parm_i" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_parm_i" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_parm_i__SWIG_0(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_parm_f__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  float result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_get_parm_f self stg ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_parm_f" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_parm_f" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (float)Rig_get_parm_f__SWIG_0(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_parm_i__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_get_parm_i self name ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_parm_i" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_parm_i" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_parm_i__SWIG_1(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_parm_i(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Rig_get_parm_i__SWIG_0(clientData, interp, objc, argv - 1);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Rig_get_parm_i__SWIG_1(clientData, interp, objc, argv - 1);
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rig_get_parm_i'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_parm_i(setting_t)\n"
    "    Rig::get_parm_i(char const *)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_parm_f__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  float result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_get_parm_f self name ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_parm_f" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_parm_f" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    arg1->error_status = RIG_OK;
    result = (float)Rig_get_parm_f__SWIG_1(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_parm_f(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Rig_get_parm_f__SWIG_0(clientData, interp, objc, argv - 1);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Rig_get_parm_f__SWIG_1(clientData, interp, objc, argv - 1);
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rig_get_parm_f'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_parm_f(setting_t)\n"
    "    Rig::get_parm_f(char const *)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_parm__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char temp3[256+1] ;
  
  arg3 = (char *) temp3;
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_get_parm self name ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_parm" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    arg1->error_status = RIG_OK;
    Rig_get_parm__SWIG_0(arg1,(char const *)arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  arg3[256] = 0;  
  Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_FromCharPtr(arg3));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_conf__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_set_conf self name val ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_conf" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_conf" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(objv[3], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rig_set_conf" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rig_set_conf__SWIG_1(arg1,(char const *)arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_conf(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Rig_set_conf__SWIG_0(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Rig_set_conf__SWIG_1(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rig_set_conf'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::set_conf(hamlib_token_t,const_char_string)\n"
    "    Rig::set_conf(char const *,char const *)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_freq(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_freq self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_freq" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_freq" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (freq_t)Rig_get_freq(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_mode(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  rmode_t *arg2 = 0 ;
  pbwidth_t *arg3 = 0 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rmode_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  pbwidth_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_mode self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_mode" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_get_mode" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_mode(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (SWIG_IsTmpObj(res2)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_unsigned_SS_long_SS_long((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_unsigned_long_long, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_long((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_split_mode(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  rmode_t *arg2 = 0 ;
  pbwidth_t *arg3 = 0 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rmode_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  pbwidth_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_split_mode self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_split_mode" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_get_split_mode" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_split_mode(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (SWIG_IsTmpObj(res2)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_unsigned_SS_long_SS_long((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_unsigned_long_long, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_long((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_vfo(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  vfo_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rig_get_vfo self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_vfo" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (vfo_t)Rig_get_vfo(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_ptt(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  ptt_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_ptt self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ptt" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ptt" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (ptt_t)Rig_get_ptt(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_rptr_shift(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  rptr_shift_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_rptr_shift self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_rptr_shift" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_rptr_shift" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (rptr_shift_t)Rig_get_rptr_shift(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_rptr_offs(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_rptr_offs self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_rptr_offs" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_rptr_offs" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (shortfreq_t)Rig_get_rptr_offs(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_ctcss_tone(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  tone_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_ctcss_tone self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ctcss_tone" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ctcss_tone" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (tone_t)Rig_get_ctcss_tone(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_dcs_code(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  tone_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_dcs_code self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_dcs_code" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_dcs_code" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (tone_t)Rig_get_dcs_code(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_ctcss_sql(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  tone_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_ctcss_sql self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ctcss_sql" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ctcss_sql" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (tone_t)Rig_get_ctcss_sql(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_dcs_sql(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  tone_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_dcs_sql self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_dcs_sql" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_dcs_sql" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (tone_t)Rig_get_dcs_sql(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_split_freq(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_split_freq self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_split_freq" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_split_freq" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (freq_t)Rig_get_split_freq(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_split_vfo(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  split_t *arg2 = 0 ;
  vfo_t *arg3 = 0 ;
  vfo_t arg4 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  split_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  vfo_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_split_vfo self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_split_vfo" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_get_split_vfo" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_split_vfo(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (SWIG_IsTmpObj(res2)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_split_t, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_unsigned_SS_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_unsigned_int, new_flags));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_rit(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_rit self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_rit" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_rit" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (shortfreq_t)Rig_get_rit(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_xit(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_xit self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_xit" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_xit" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (shortfreq_t)Rig_get_xit(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_ts(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  shortfreq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_ts self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ts" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ts" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (shortfreq_t)Rig_get_ts(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_long((long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_ant(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  ant_t *arg2 = 0 ;
  ant_t *arg3 = 0 ;
  ant_t *arg4 = 0 ;
  int *arg5 = 0 ;
  ant_t arg6 ;
  vfo_t arg7 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ant_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  ant_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  ant_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  int temp5 ;
  int res5 = SWIG_TMPOBJ ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  arg5 = &temp5;
  if (SWIG_GetArgs(interp, objc, objv,"oo|o:Rig_get_ant self ant ?vfo? ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ant" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode6 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Rig_get_ant" "', argument " "6"" of type '" "ant_t""'");
  } 
  arg6 = (ant_t)(val6);
  if (objc > 3) {
    ecode7 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Rig_get_ant" "', argument " "7"" of type '" "vfo_t""'");
    } 
    arg7 = (vfo_t)(val7);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_ant(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (SWIG_IsTmpObj(res2)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_unsigned_SS_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_unsigned_int, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_unsigned_SS_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_unsigned_int, new_flags));
  }
  if (SWIG_IsTmpObj(res4)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_unsigned_SS_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_unsigned_int, new_flags));
  }
  if (SWIG_IsTmpObj(res5)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_int((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_int, new_flags));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_vfo_info(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  int *arg2 = 0 ;
  split_t *arg3 = 0 ;
  pbwidth_t *arg4 = 0 ;
  rmode_t *arg5 = 0 ;
  freq_t *arg6 = 0 ;
  vfo_t arg7 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  split_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  pbwidth_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  rmode_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  freq_t temp6 ;
  int res6 = SWIG_TMPOBJ ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  arg5 = &temp5;
  arg6 = &temp6;
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_vfo_info self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_vfo_info" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode7 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Rig_get_vfo_info" "', argument " "7"" of type '" "vfo_t""'");
    } 
    arg7 = (vfo_t)(val7);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_vfo_info(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (SWIG_IsTmpObj(res2)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_split_t, new_flags));
  }
  if (SWIG_IsTmpObj(res4)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_long((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_long, new_flags));
  }
  if (SWIG_IsTmpObj(res5)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_unsigned_SS_long_SS_long((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_unsigned_long_long, new_flags));
  }
  if (SWIG_IsTmpObj(res6)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_double((*arg6)));
  } else {
    int new_flags = SWIG_IsNewObj(res6) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg6), SWIGTYPE_p_double, new_flags));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_mem(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_mem self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_mem" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_mem" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_mem(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_powerstat(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  powerstat_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rig_get_powerstat self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_powerstat" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (powerstat_t)Rig_get_powerstat(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_dcd(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  vfo_t arg2 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  dcd_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_get_dcd self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_dcd" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_dcd" "', argument " "2"" of type '" "vfo_t""'");
    } 
    arg2 = (vfo_t)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (dcd_t)Rig_get_dcd(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_mem_count(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rig_mem_count self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_mem_count" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_mem_count(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_lookup_mem_caps(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  int arg2 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  chan_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_lookup_mem_caps self ?channel_num? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_lookup_mem_caps" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_lookup_mem_caps" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  {
    arg1->error_status = RIG_OK;
    result = (chan_t *)Rig_lookup_mem_caps(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_chan_list,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_set_channel(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  struct channel *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_set_channel self chan ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_set_channel" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_set_channel" "', argument " "2"" of type '" "struct channel const *""'"); 
  }
  arg2 = (struct channel *)(argp2);
  {
    arg1->error_status = RIG_OK;
    Rig_set_channel(arg1,(struct channel const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_channel__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  struct channel *arg2 = 0 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rig_get_channel self chan read_only ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_channel" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_channel" "', argument " "2"" of type '" "struct channel *""'"); 
  }
  arg2 = (struct channel *)(argp2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_channel" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rig_get_channel__SWIG_0(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_chan_all__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  struct channel *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_get_chan_all self chans ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_chan_all" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_chan_all" "', argument " "2"" of type '" "struct channel *""'"); 
  }
  arg2 = (struct channel *)(argp2);
  {
    arg1->error_status = RIG_OK;
    Rig_get_chan_all__SWIG_0(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_chan_all__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct channel *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rig_get_chan_all self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_chan_all" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (struct channel *)Rig_get_chan_all(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_channel,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_chan_all(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Rig_get_chan_all__SWIG_1(clientData, interp, objc, argv - 1);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_channel, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Rig_get_chan_all__SWIG_0(clientData, interp, objc, argv - 1);
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rig_get_chan_all'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_chan_all(struct channel *)\n"
    "    Rig::get_chan_all()\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_channel__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  int arg2 ;
  int arg3 = INT_MAX ;
  vfo_t arg4 = (1u << (28)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  struct channel *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo|oo:Rig_get_channel self read_only ?channel_num? ?vfo? ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_channel" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_channel" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (objc > 3) {
    ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_channel" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  if (objc > 4) {
    ecode4 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Rig_get_channel" "', argument " "4"" of type '" "vfo_t""'");
    } 
    arg4 = (vfo_t)(val4);
  }
  {
    arg1->error_status = RIG_OK;
    result = (struct channel *)Rig_get_channel__SWIG_1(arg1,arg2,arg3,arg4);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_channel,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_channel(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if ((argc >= 2) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_channel__SWIG_1(clientData, interp, objc, argv - 1);
        }
        {
          int res = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          if (argc <= 3) {
            return _wrap_Rig_get_channel__SWIG_1(clientData, interp, objc, argv - 1);
          }
          {
            int res = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Rig_get_channel__SWIG_1(clientData, interp, objc, argv - 1);
          }
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_channel, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_channel__SWIG_0(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rig_get_channel'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_channel(struct channel *,int)\n"
    "    Rig::get_channel(int,int,vfo_t)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_conf__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  char temp3[256+1] ;
  
  arg3 = (char *) temp3;
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_get_conf self tok ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_conf" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_conf" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_get_conf__SWIG_0(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  arg3[256] = 0;  
  Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_FromCharPtr(arg3));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_conf__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char temp3[256+1] ;
  
  arg3 = (char *) temp3;
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_get_conf self name ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_conf" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rig_get_conf" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    arg1->error_status = RIG_OK;
    Rig_get_conf__SWIG_1(arg1,(char const *)arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  arg3[256] = 0;  
  Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_FromCharPtr(arg3));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_conf(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Rig_get_conf__SWIG_0(clientData, interp, objc, argv - 1);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Rig_get_conf__SWIG_1(clientData, interp, objc, argv - 1);
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rig_get_conf'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_conf(hamlib_token_t,char *)\n"
    "    Rig::get_conf(char const *,char *)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_recv_dtmf(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  char *arg2 = 0 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char temp2[256+1] ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  
  arg2 = (char *) temp2;
  if (SWIG_GetArgs(interp, objc, objv,"o|o:Rig_recv_dtmf self ?vfo? ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_recv_dtmf" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  if (objc > 2) {
    ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_recv_dtmf" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_recv_dtmf(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  arg2[256] = 0;  
  Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_FromCharPtr(arg2));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_info(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rig_get_info self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_info" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Rig_get_info(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_func(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo|o:Rig_get_func self func ?vfo? ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_func" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (objc > 3) {
    ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_func" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_func(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_ext_func(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo|o:Rig_get_ext_func self func ?vfo? ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ext_func" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ext_func" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  if (objc > 3) {
    ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_ext_func" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    result = (int)Rig_get_ext_func(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_level__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo|o:Rig_get_level self level ?vfo? ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (objc > 3) {
    ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_level" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_level__SWIG_1(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_level__SWIG_1(clientData, interp, objc, argv - 1);
        }
        {
          int res = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_level__SWIG_1(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        if (argc <= 2) {
          return _wrap_Rig_get_level__SWIG_0(clientData, interp, objc, argv - 1);
        }
        {
          int res = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Rig_get_level__SWIG_0(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rig_get_level'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_level(char const *,char *,vfo_t)\n"
    "    Rig::get_level(setting_t,vfo_t)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_ext_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  vfo_t arg3 = (1u << (29)) ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo|o:Rig_get_ext_level self token ?vfo? ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ext_level" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ext_level" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  if (objc > 3) {
    ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rig_get_ext_level" "', argument " "3"" of type '" "vfo_t""'");
    } 
    arg3 = (vfo_t)(val3);
  }
  {
    arg1->error_status = RIG_OK;
    Rig_get_ext_level(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_parm__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_get_parm self parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_get_parm__SWIG_1(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_parm(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Rig_get_parm__SWIG_1(clientData, interp, objc, argv - 1);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Rig_get_parm__SWIG_0(clientData, interp, objc, argv - 1);
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rig_get_parm'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rig::get_parm(char const *,char *)\n"
    "    Rig::get_parm(setting_t)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rig_get_ext_parm(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rig *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rig_get_ext_parm self parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rig_get_ext_parm" "', argument " "1"" of type '" "struct Rig *""'"); 
  }
  arg1 = (struct Rig *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rig_get_ext_parm" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rig_get_ext_parm(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_Rig(void *obj) {
struct Rig *arg1 = (struct Rig *) obj;
delete_Rig(arg1);
}
static swig_method swig_Rig_methods[] = {
    {"open", _wrap_Rig_open}, 
    {"close", _wrap_Rig_close}, 
    {"set_freq", _wrap_Rig_set_freq}, 
    {"set_mode", _wrap_Rig_set_mode}, 
    {"set_ptt", _wrap_Rig_set_ptt}, 
    {"set_rptr_shift", _wrap_Rig_set_rptr_shift}, 
    {"set_rptr_offs", _wrap_Rig_set_rptr_offs}, 
    {"set_ctcss_tone", _wrap_Rig_set_ctcss_tone}, 
    {"set_dcs_code", _wrap_Rig_set_dcs_code}, 
    {"set_ctcss_sql", _wrap_Rig_set_ctcss_sql}, 
    {"set_dcs_sql", _wrap_Rig_set_dcs_sql}, 
    {"set_split_freq", _wrap_Rig_set_split_freq}, 
    {"set_split_mode", _wrap_Rig_set_split_mode}, 
    {"set_split_freq_mode", _wrap_Rig_set_split_freq_mode}, 
    {"set_split_vfo", _wrap_Rig_set_split_vfo}, 
    {"set_rit", _wrap_Rig_set_rit}, 
    {"set_xit", _wrap_Rig_set_xit}, 
    {"set_ts", _wrap_Rig_set_ts}, 
    {"set_ant", _wrap_Rig_set_ant}, 
    {"set_func", _wrap_Rig_set_func}, 
    {"set_ext_func", _wrap_Rig_set_ext_func}, 
    {"set_bank", _wrap_Rig_set_bank}, 
    {"set_mem", _wrap_Rig_set_mem}, 
    {"send_dtmf", _wrap_Rig_send_dtmf}, 
    {"send_morse", _wrap_Rig_send_morse}, 
    {"vfo_op", _wrap_Rig_vfo_op}, 
    {"scan", _wrap_Rig_scan}, 
    {"set_level", _wrap_Rig_set_level}, 
    {"set_ext_level", _wrap_Rig_set_ext_level}, 
    {"get_level_i", _wrap_Rig_get_level_i}, 
    {"get_level_f", _wrap_Rig_get_level_f}, 
    {"set_vfo", _wrap_Rig_set_vfo}, 
    {"set_powerstat", _wrap_Rig_set_powerstat}, 
    {"has_get_level", _wrap_Rig_has_get_level}, 
    {"has_get_parm", _wrap_Rig_has_get_parm}, 
    {"has_set_parm", _wrap_Rig_has_set_parm}, 
    {"has_get_func", _wrap_Rig_has_get_func}, 
    {"has_set_func", _wrap_Rig_has_set_func}, 
    {"reset", _wrap_Rig_reset}, 
    {"has_scan", _wrap_Rig_has_scan}, 
    {"has_vfo_op", _wrap_Rig_has_vfo_op}, 
    {"passband_normal", _wrap_Rig_passband_normal}, 
    {"passband_narrow", _wrap_Rig_passband_narrow}, 
    {"passband_wide", _wrap_Rig_passband_wide}, 
    {"set_vfo_opt", _wrap_Rig_set_vfo_opt}, 
    {"ext_token_lookup", _wrap_Rig_ext_token_lookup}, 
    {"token_lookup", _wrap_Rig_token_lookup}, 
    {"set_ext_parm", _wrap_Rig_set_ext_parm}, 
    {"set_parm", _wrap_Rig_set_parm}, 
    {"get_parm_i", _wrap_Rig_get_parm_i}, 
    {"get_parm_f", _wrap_Rig_get_parm_f}, 
    {"set_conf", _wrap_Rig_set_conf}, 
    {"get_freq", _wrap_Rig_get_freq}, 
    {"get_mode", _wrap_Rig_get_mode}, 
    {"get_split_mode", _wrap_Rig_get_split_mode}, 
    {"get_vfo", _wrap_Rig_get_vfo}, 
    {"get_ptt", _wrap_Rig_get_ptt}, 
    {"get_rptr_shift", _wrap_Rig_get_rptr_shift}, 
    {"get_rptr_offs", _wrap_Rig_get_rptr_offs}, 
    {"get_ctcss_tone", _wrap_Rig_get_ctcss_tone}, 
    {"get_dcs_code", _wrap_Rig_get_dcs_code}, 
    {"get_ctcss_sql", _wrap_Rig_get_ctcss_sql}, 
    {"get_dcs_sql", _wrap_Rig_get_dcs_sql}, 
    {"get_split_freq", _wrap_Rig_get_split_freq}, 
    {"get_split_vfo", _wrap_Rig_get_split_vfo}, 
    {"get_rit", _wrap_Rig_get_rit}, 
    {"get_xit", _wrap_Rig_get_xit}, 
    {"get_ts", _wrap_Rig_get_ts}, 
    {"get_ant", _wrap_Rig_get_ant}, 
    {"get_vfo_info", _wrap_Rig_get_vfo_info}, 
    {"get_mem", _wrap_Rig_get_mem}, 
    {"get_powerstat", _wrap_Rig_get_powerstat}, 
    {"get_dcd", _wrap_Rig_get_dcd}, 
    {"mem_count", _wrap_Rig_mem_count}, 
    {"lookup_mem_caps", _wrap_Rig_lookup_mem_caps}, 
    {"set_channel", _wrap_Rig_set_channel}, 
    {"get_chan_all", _wrap_Rig_get_chan_all}, 
    {"get_channel", _wrap_Rig_get_channel}, 
    {"get_conf", _wrap_Rig_get_conf}, 
    {"recv_dtmf", _wrap_Rig_recv_dtmf}, 
    {"get_info", _wrap_Rig_get_info}, 
    {"get_func", _wrap_Rig_get_func}, 
    {"get_ext_func", _wrap_Rig_get_ext_func}, 
    {"get_level", _wrap_Rig_get_level}, 
    {"get_ext_level", _wrap_Rig_get_ext_level}, 
    {"get_parm", _wrap_Rig_get_parm}, 
    {"get_ext_parm", _wrap_Rig_get_ext_parm}, 
    {0,0}
};
static swig_attribute swig_Rig_attributes[] = {
    { "-rig",_wrap_Rig_rig_get, _wrap_Rig_rig_set},
    { "-caps",_wrap_Rig_caps_get, _wrap_Rig_caps_set},
    { "-state",_wrap_Rig_state_get, _wrap_Rig_state_set},
    { "-error_status",_wrap_Rig_error_status_get, _wrap_Rig_error_status_set},
    { "-do_exception",_wrap_Rig_do_exception_get, _wrap_Rig_do_exception_set},
    {0,0,0}
};
static swig_class *swig_Rig_bases[] = {0};
static const char * swig_Rig_base_names[] = {0};
static swig_class _wrap_class_Rig = { "Rig", &SWIGTYPE_p_Rig,_wrap_new_Rig, swig_delete_Rig, swig_Rig_methods, swig_Rig_attributes, swig_Rig_bases,swig_Rig_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_new_channelArray(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  channelArray *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:new_channelArray nelements ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_channelArray" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (channelArray *)new_channelArray(arg1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_channelArray,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_channelArray(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  channelArray *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_channelArray self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channelArray, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_channelArray" "', argument " "1"" of type '" "channelArray *""'"); 
  }
  arg1 = (channelArray *)(argp1);
  delete_channelArray(arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channelArray_getitem(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  channelArray *arg1 = 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  struct channel result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:channelArray_getitem self index ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channelArray, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channelArray_getitem" "', argument " "1"" of type '" "channelArray *""'"); 
  }
  arg1 = (channelArray *)(argp1);
  ecode2 = SWIG_AsVal_size_t SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channelArray_getitem" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = channelArray_getitem(arg1,arg2);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((struct channel *)memcpy((struct channel *)calloc(1,sizeof(struct channel)),&result,sizeof(struct channel)), SWIGTYPE_p_channel, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channelArray_setitem(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  channelArray *arg1 = 0 ;
  size_t arg2 ;
  struct channel arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:channelArray_setitem self index value ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channelArray, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channelArray_setitem" "', argument " "1"" of type '" "channelArray *""'"); 
  }
  arg1 = (channelArray *)(argp1);
  ecode2 = SWIG_AsVal_size_t SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "channelArray_setitem" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(objv[3], &argp3, SWIGTYPE_p_channel,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "channelArray_setitem" "', argument " "3"" of type '" "struct channel""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "channelArray_setitem" "', argument " "3"" of type '" "struct channel""'");
    } else {
      arg3 = *((struct channel *)(argp3));
    }
  }
  channelArray_setitem(arg1,arg2,arg3);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channelArray_cast(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  channelArray *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct channel *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channelArray_cast self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channelArray, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channelArray_cast" "', argument " "1"" of type '" "channelArray *""'"); 
  }
  arg1 = (channelArray *)(argp1);
  result = (struct channel *)channelArray_cast(arg1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_channel,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_channelArray_frompointer(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct channel *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  channelArray *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:channelArray_frompointer t ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_channel, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "channelArray_frompointer" "', argument " "1"" of type '" "struct channel *""'"); 
  }
  arg1 = (struct channel *)(argp1);
  result = (channelArray *)channelArray_frompointer(arg1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_channelArray,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_channelArray(void *obj) {
channelArray *arg1 = (channelArray *) obj;
delete_channelArray(arg1);
}
static swig_method swig_channelArray_methods[] = {
    {"getitem", _wrap_channelArray_getitem}, 
    {"setitem", _wrap_channelArray_setitem}, 
    {"cast", _wrap_channelArray_cast}, 
    {0,0}
};
static swig_attribute swig_channelArray_attributes[] = {
    {0,0,0}
};
static swig_class *swig_channelArray_bases[] = {0};
static const char * swig_channelArray_base_names[] = {0};
static swig_class _wrap_class_channelArray = { "channelArray", &SWIGTYPE_p_channelArray,_wrap_new_channelArray, swig_delete_channelArray, swig_channelArray_methods, swig_channelArray_attributes, swig_channelArray_bases,swig_channelArray_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_new_toneArray(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  toneArray *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:new_toneArray nelements ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_toneArray" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (toneArray *)new_toneArray(arg1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_toneArray,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_toneArray(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  toneArray *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_toneArray self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_toneArray, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_toneArray" "', argument " "1"" of type '" "toneArray *""'"); 
  }
  arg1 = (toneArray *)(argp1);
  delete_toneArray(arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_toneArray_getitem(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  toneArray *arg1 = 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  tone_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:toneArray_getitem self index ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_toneArray, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toneArray_getitem" "', argument " "1"" of type '" "toneArray *""'"); 
  }
  arg1 = (toneArray *)(argp1);
  ecode2 = SWIG_AsVal_size_t SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "toneArray_getitem" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = (tone_t)toneArray_getitem(arg1,arg2);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_toneArray_setitem(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  toneArray *arg1 = 0 ;
  size_t arg2 ;
  tone_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:toneArray_setitem self index value ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_toneArray, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toneArray_setitem" "', argument " "1"" of type '" "toneArray *""'"); 
  }
  arg1 = (toneArray *)(argp1);
  ecode2 = SWIG_AsVal_size_t SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "toneArray_setitem" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "toneArray_setitem" "', argument " "3"" of type '" "tone_t""'");
  } 
  arg3 = (tone_t)(val3);
  toneArray_setitem(arg1,arg2,arg3);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_toneArray_cast(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  toneArray *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  tone_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:toneArray_cast self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_toneArray, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toneArray_cast" "', argument " "1"" of type '" "toneArray *""'"); 
  }
  arg1 = (toneArray *)(argp1);
  result = (tone_t *)toneArray_cast(arg1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_toneArray_frompointer(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  tone_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  toneArray *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:toneArray_frompointer t ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toneArray_frompointer" "', argument " "1"" of type '" "tone_t *""'"); 
  }
  arg1 = (tone_t *)(argp1);
  result = (toneArray *)toneArray_frompointer(arg1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_toneArray,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_toneArray(void *obj) {
toneArray *arg1 = (toneArray *) obj;
delete_toneArray(arg1);
}
static swig_method swig_toneArray_methods[] = {
    {"getitem", _wrap_toneArray_getitem}, 
    {"setitem", _wrap_toneArray_setitem}, 
    {"cast", _wrap_toneArray_cast}, 
    {0,0}
};
static swig_attribute swig_toneArray_attributes[] = {
    {0,0,0}
};
static swig_class *swig_toneArray_bases[] = {0};
static const char * swig_toneArray_base_names[] = {0};
static swig_class _wrap_class_toneArray = { "toneArray", &SWIGTYPE_p_toneArray,_wrap_new_toneArray, swig_delete_toneArray, swig_toneArray_methods, swig_toneArray_attributes, swig_toneArray_bases,swig_toneArray_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_rot_caps_rot_model_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  rot_model_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_rot_model_set self rot_model ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_model_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_rot_model_set" "', argument " "2"" of type '" "rot_model_t""'");
  } 
  arg2 = (rot_model_t)(val2);
  if (arg1) (arg1)->rot_model = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_rot_model_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rot_model_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_rot_model_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_model_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (rot_model_t) ((arg1)->rot_model);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_model_name_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_model_name_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_model_name_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (char *) ((arg1)->model_name);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_mfg_name_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_mfg_name_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_mfg_name_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (char *) ((arg1)->mfg_name);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_version_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_version_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_version_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (char *) ((arg1)->version);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_copyright_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_copyright_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_copyright_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (char *) ((arg1)->copyright);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_status_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  enum rig_status_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_status_set self status ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_status_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_status_set" "', argument " "2"" of type '" "enum rig_status_e""'");
  } 
  arg2 = (enum rig_status_e)(val2);
  if (arg1) (arg1)->status = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_status_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum rig_status_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_status_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_status_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (enum rig_status_e) ((arg1)->status);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_rot_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_rot_type_set self rot_type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_type_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_rot_type_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->rot_type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_rot_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_rot_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_type_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->rot_type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_port_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  enum rig_port_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_port_type_set self port_type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_port_type_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_port_type_set" "', argument " "2"" of type '" "enum rig_port_e""'");
  } 
  arg2 = (enum rig_port_e)(val2);
  if (arg1) (arg1)->port_type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_port_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum rig_port_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_port_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_port_type_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (enum rig_port_e) ((arg1)->port_type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_serial_rate_min_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_serial_rate_min_set self serial_rate_min ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_rate_min_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_serial_rate_min_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_min = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_serial_rate_min_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_serial_rate_min_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_rate_min_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->serial_rate_min);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_serial_rate_max_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_serial_rate_max_set self serial_rate_max ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_rate_max_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_serial_rate_max_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_max = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_serial_rate_max_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_serial_rate_max_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_rate_max_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->serial_rate_max);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_serial_data_bits_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_serial_data_bits_set self serial_data_bits ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_data_bits_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_serial_data_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_data_bits = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_serial_data_bits_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_serial_data_bits_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_data_bits_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->serial_data_bits);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_serial_stop_bits_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_serial_stop_bits_set self serial_stop_bits ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_stop_bits_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_serial_stop_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_stop_bits = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_serial_stop_bits_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_serial_stop_bits_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_stop_bits_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->serial_stop_bits);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_serial_parity_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  enum serial_parity_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_serial_parity_set self serial_parity ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_parity_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_serial_parity_set" "', argument " "2"" of type '" "enum serial_parity_e""'");
  } 
  arg2 = (enum serial_parity_e)(val2);
  if (arg1) (arg1)->serial_parity = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_serial_parity_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum serial_parity_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_serial_parity_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_parity_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (enum serial_parity_e) ((arg1)->serial_parity);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_serial_handshake_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  enum serial_handshake_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_serial_handshake_set self serial_handshake ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_handshake_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_serial_handshake_set" "', argument " "2"" of type '" "enum serial_handshake_e""'");
  } 
  arg2 = (enum serial_handshake_e)(val2);
  if (arg1) (arg1)->serial_handshake = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_serial_handshake_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum serial_handshake_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_serial_handshake_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_serial_handshake_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (enum serial_handshake_e) ((arg1)->serial_handshake);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_write_delay_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_write_delay_set self write_delay ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_write_delay_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->write_delay = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_write_delay_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_write_delay_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_write_delay_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->write_delay);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_post_write_delay_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_post_write_delay_set self post_write_delay ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_post_write_delay_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_post_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->post_write_delay = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_post_write_delay_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_post_write_delay_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_post_write_delay_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->post_write_delay);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_timeout_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_timeout_set self timeout ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_timeout_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_timeout_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->timeout = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_timeout_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_timeout_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_timeout_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->timeout);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_retry_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_retry_set self retry ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_retry_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_retry_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->retry = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_retry_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_retry_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_retry_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int) ((arg1)->retry);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_get_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_has_get_func_set self has_get_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_get_func_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_get_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_get_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_has_get_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_get_func_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_func);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_set_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_has_set_func_set self has_set_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_set_func_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_set_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_set_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_has_set_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_set_func_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_func);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_get_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_has_get_level_set self has_get_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_get_level_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_get_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_get_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_has_get_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_get_level_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_level);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_set_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_has_set_level_set self has_set_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_set_level_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_set_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_set_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_has_set_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_set_level_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_level);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_get_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_has_get_parm_set self has_get_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_get_parm_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_get_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_get_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_has_get_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_get_parm_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_parm);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_set_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_has_set_parm_set self has_set_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_set_parm_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_set_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_set_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_has_set_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_set_parm_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_parm);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_status_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  rot_status_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_has_status_set self has_status ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_status_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_has_status_set" "', argument " "2"" of type '" "rot_status_t""'");
  } 
  arg2 = (rot_status_t)(val2);
  if (arg1) (arg1)->has_status = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_has_status_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rot_status_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_has_status_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_has_status_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (rot_status_t) ((arg1)->has_status);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_level_gran_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_level_gran_set self level_gran ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_level_gran_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_level_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->level_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""level_gran""' of type '""gran_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_level_gran_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gran_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_level_gran_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_level_gran_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->level_gran);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_parm_gran_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_parm_gran_set self parm_gran ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_parm_gran_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_parm_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->parm_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""parm_gran""' of type '""gran_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_parm_gran_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gran_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_parm_gran_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_parm_gran_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->parm_gran);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_extparms_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_extparms_set self extparms ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_extparms_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_extparms_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extparms = (struct confparams const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_extparms_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct confparams *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_extparms_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_extparms_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extparms);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_extlevels_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_extlevels_set self extlevels ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_extlevels_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_extlevels_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extlevels = (struct confparams const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_extlevels_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct confparams *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_extlevels_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_extlevels_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extlevels);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_extfuncs_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_extfuncs_set self extfuncs ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_extfuncs_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_extfuncs_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extfuncs = (struct confparams const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_extfuncs_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct confparams *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_extfuncs_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_extfuncs_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extfuncs);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_ext_tokens_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  
  arg2 = &temp2;
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_ext_tokens_set self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_ext_tokens_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  if (arg1) (arg1)->ext_tokens = arg2;
  
  if (SWIG_IsTmpObj(res2)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_ext_tokens_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_ext_tokens_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_ext_tokens_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int *) ((arg1)->ext_tokens);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_int,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_min_az_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  azimuth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_min_az_set self min_az ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_min_az_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_min_az_set" "', argument " "2"" of type '" "azimuth_t""'");
  } 
  arg2 = (azimuth_t)(val2);
  if (arg1) (arg1)->min_az = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_min_az_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  azimuth_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_min_az_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_min_az_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (azimuth_t) ((arg1)->min_az);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_max_az_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  azimuth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_max_az_set self max_az ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_max_az_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_max_az_set" "', argument " "2"" of type '" "azimuth_t""'");
  } 
  arg2 = (azimuth_t)(val2);
  if (arg1) (arg1)->max_az = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_max_az_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  azimuth_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_max_az_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_max_az_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (azimuth_t) ((arg1)->max_az);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_min_el_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  elevation_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_min_el_set self min_el ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_min_el_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_min_el_set" "', argument " "2"" of type '" "elevation_t""'");
  } 
  arg2 = (elevation_t)(val2);
  if (arg1) (arg1)->min_el = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_min_el_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  elevation_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_min_el_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_min_el_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (elevation_t) ((arg1)->min_el);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_max_el_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  elevation_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_max_el_set self max_el ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_max_el_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_caps_max_el_set" "', argument " "2"" of type '" "elevation_t""'");
  } 
  arg2 = (elevation_t)(val2);
  if (arg1) (arg1)->max_el = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_max_el_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  elevation_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_max_el_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_max_el_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (elevation_t) ((arg1)->max_el);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_cfgparams_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_cfgparams_set self cfgparams ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_cfgparams_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_cfgparams_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->cfgparams = (struct confparams const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_cfgparams_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct confparams *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_cfgparams_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_cfgparams_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (struct confparams *) ((arg1)->cfgparams);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_priv_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_priv_set self priv ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_priv_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_caps_priv_set" "', argument " "2"" of type '" "void const *""'"); 
  }
  if (arg1) (arg1)->priv = (void const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_priv_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_priv_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_priv_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (void *) ((arg1)->priv);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_rot_init_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_rot_init_set self rot_init ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_init_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_rot_init_set" "', argument " "2"" of type '" "int (*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->rot_init = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_rot_init_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_rot_init_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_init_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *)) ((arg1)->rot_init);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_rot_cleanup_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_rot_cleanup_set self rot_cleanup ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_cleanup_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_rot_cleanup_set" "', argument " "2"" of type '" "int (*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->rot_cleanup = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_rot_cleanup_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_rot_cleanup_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_cleanup_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *)) ((arg1)->rot_cleanup);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_rot_open_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_rot_open_set self rot_open ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_open_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_rot_open_set" "', argument " "2"" of type '" "int (*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->rot_open = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_rot_open_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_rot_open_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_open_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *)) ((arg1)->rot_open);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_rot_close_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_rot_close_set self rot_close ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_close_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_rot_close_set" "', argument " "2"" of type '" "int (*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->rot_close = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_rot_close_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_rot_close_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_rot_close_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *)) ((arg1)->rot_close);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_conf_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_set_conf_set self set_conf ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_conf_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_conf_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,char const *)""'"); 
    }
  }
  if (arg1) (arg1)->set_conf = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_conf_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,hamlib_token_t,char const *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_set_conf_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_conf_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,char const *)) ((arg1)->set_conf);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_p_q_const__char__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_conf_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,char *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_get_conf_set self get_conf ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_conf_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_p_char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_conf_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,char *)""'"); 
    }
  }
  if (arg1) (arg1)->get_conf = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_conf_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,hamlib_token_t,char *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_get_conf_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_conf_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,char *)) ((arg1)->get_conf);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_p_char__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_position_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,azimuth_t,elevation_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_set_position_set self set_position ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_position_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_float_float__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_position_set" "', argument " "2"" of type '" "int (*)(ROT *,azimuth_t,elevation_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_position = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_position_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,azimuth_t,elevation_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_set_position_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_position_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,azimuth_t,elevation_t)) ((arg1)->set_position);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_float_float__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_position_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,azimuth_t *,elevation_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_get_position_set self get_position ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_position_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_p_float_p_float__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_position_set" "', argument " "2"" of type '" "int (*)(ROT *,azimuth_t *,elevation_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_position = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_position_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,azimuth_t *,elevation_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_get_position_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_position_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,azimuth_t *,elevation_t *)) ((arg1)->get_position);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_p_float_p_float__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_stop_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_stop_set self stop ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_stop_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_stop_set" "', argument " "2"" of type '" "int (*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->stop = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_stop_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_stop_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_stop_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *)) ((arg1)->stop);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_park_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_park_set self park ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_park_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_park_set" "', argument " "2"" of type '" "int (*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->park = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_park_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_park_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_park_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *)) ((arg1)->park);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_reset_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,rot_reset_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_reset_set self reset ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_reset_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_reset_set" "', argument " "2"" of type '" "int (*)(ROT *,rot_reset_t)""'"); 
    }
  }
  if (arg1) (arg1)->reset = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_reset_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,rot_reset_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_reset_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_reset_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,rot_reset_t)) ((arg1)->reset);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_move_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,int,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_move_set self move ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_move_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_int_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_move_set" "', argument " "2"" of type '" "int (*)(ROT *,int,int)""'"); 
    }
  }
  if (arg1) (arg1)->move = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_move_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,int,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_move_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_move_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,int,int)) ((arg1)->move);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_int_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_info_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  char *(*arg2)(ROT *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_get_info_set self get_info ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_info_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot__p_char);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_info_set" "', argument " "2"" of type '" "char const *(*)(ROT *)""'"); 
    }
  }
  if (arg1) (arg1)->get_info = (char const *(*)(ROT *))arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_info_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *(*result)(ROT *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_get_info_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_info_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (char *(*)(ROT *)) ((arg1)->get_info);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot__p_char));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,setting_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_set_level_set self set_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_level_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_level_set" "', argument " "2"" of type '" "int (*)(ROT *,setting_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,setting_t,value_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_set_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_level_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,setting_t,value_t)) ((arg1)->set_level);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,setting_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_get_level_set self get_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_level_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_level_set" "', argument " "2"" of type '" "int (*)(ROT *,setting_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,setting_t,value_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_get_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_level_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,setting_t,value_t *)) ((arg1)->get_level);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,setting_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_set_func_set self set_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_func_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_func_set" "', argument " "2"" of type '" "int (*)(ROT *,setting_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,setting_t,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_set_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_func_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,setting_t,int)) ((arg1)->set_func);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,setting_t,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_get_func_set self get_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_func_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_func_set" "', argument " "2"" of type '" "int (*)(ROT *,setting_t,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,setting_t,int *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_get_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_func_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,setting_t,int *)) ((arg1)->get_func);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,setting_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_set_parm_set self set_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_parm_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_parm_set" "', argument " "2"" of type '" "int (*)(ROT *,setting_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,setting_t,value_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_set_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_parm_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,setting_t,value_t)) ((arg1)->set_parm);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,setting_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_get_parm_set self get_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_parm_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_parm_set" "', argument " "2"" of type '" "int (*)(ROT *,setting_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,setting_t,value_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_get_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_parm_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,setting_t,value_t *)) ((arg1)->get_parm);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_ext_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_set_ext_level_set self set_ext_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_ext_level_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_ext_level_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_ext_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,hamlib_token_t,value_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_set_ext_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_ext_level_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,value_t)) ((arg1)->set_ext_level);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_ext_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_get_ext_level_set self get_ext_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_ext_level_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_ext_level_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_ext_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,hamlib_token_t,value_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_get_ext_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_ext_level_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,value_t *)) ((arg1)->get_ext_level);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_p_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_ext_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_set_ext_func_set self set_ext_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_ext_func_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_ext_func_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,int)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_ext_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,hamlib_token_t,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_set_ext_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_ext_func_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,int)) ((arg1)->set_ext_func);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_ext_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,int *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_get_ext_func_set self get_ext_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_ext_func_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_p_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_ext_func_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,int *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_ext_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,hamlib_token_t,int *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_get_ext_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_ext_func_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,int *)) ((arg1)->get_ext_func);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_p_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_ext_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_set_ext_parm_set self set_ext_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_ext_parm_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_set_ext_parm_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_set_ext_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,hamlib_token_t,value_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_set_ext_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_set_ext_parm_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,value_t)) ((arg1)->set_ext_parm);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_ext_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_get_ext_parm_set self get_ext_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_ext_parm_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_ext_parm_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_ext_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,hamlib_token_t,value_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_get_ext_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_ext_parm_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,value_t *)) ((arg1)->get_ext_parm);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_p_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_status_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,rot_status_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_get_status_set self get_status ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_status_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_p_enum_rot_status_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_status_set" "', argument " "2"" of type '" "int (*)(ROT *,rot_status_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_status = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_status_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,rot_status_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_get_status_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_status_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,rot_status_t *)) ((arg1)->get_status);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_p_enum_rot_status_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_conf2_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  int (*arg2)(ROT *,hamlib_token_t,char *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_caps_get_conf2_set self get_conf2 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_conf2_set" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_s_rot_long_p_char_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "rot_caps_get_conf2_set" "', argument " "2"" of type '" "int (*)(ROT *,hamlib_token_t,char *,int)""'"); 
    }
  }
  if (arg1) (arg1)->get_conf2 = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_caps_get_conf2_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(ROT *,hamlib_token_t,char *,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_caps_get_conf2_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_caps_get_conf2_get" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  result = (int (*)(ROT *,hamlib_token_t,char *,int)) ((arg1)->get_conf2);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_s_rot_long_p_char_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_rot_caps(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_rot_caps ") == TCL_ERROR) SWIG_fail;
  result = (struct rot_caps *)calloc(1, sizeof(struct rot_caps));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rot_caps,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_rot_caps(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_rot_caps self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_caps, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rot_caps" "', argument " "1"" of type '" "struct rot_caps *""'"); 
  }
  arg1 = (struct rot_caps *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_rot_caps(void *obj) {
struct rot_caps *arg1 = (struct rot_caps *) obj;
free((char *) arg1);
}
static swig_method swig_rot_caps_methods[] = {
    {0,0}
};
static swig_attribute swig_rot_caps_attributes[] = {
    { "-rot_model",_wrap_rot_caps_rot_model_get, _wrap_rot_caps_rot_model_set},
    { "-model_name",_wrap_rot_caps_model_name_get, 0 },
    { "-mfg_name",_wrap_rot_caps_mfg_name_get, 0 },
    { "-version",_wrap_rot_caps_version_get, 0 },
    { "-copyright",_wrap_rot_caps_copyright_get, 0 },
    { "-status",_wrap_rot_caps_status_get, _wrap_rot_caps_status_set},
    { "-rot_type",_wrap_rot_caps_rot_type_get, _wrap_rot_caps_rot_type_set},
    { "-port_type",_wrap_rot_caps_port_type_get, _wrap_rot_caps_port_type_set},
    { "-serial_rate_min",_wrap_rot_caps_serial_rate_min_get, _wrap_rot_caps_serial_rate_min_set},
    { "-serial_rate_max",_wrap_rot_caps_serial_rate_max_get, _wrap_rot_caps_serial_rate_max_set},
    { "-serial_data_bits",_wrap_rot_caps_serial_data_bits_get, _wrap_rot_caps_serial_data_bits_set},
    { "-serial_stop_bits",_wrap_rot_caps_serial_stop_bits_get, _wrap_rot_caps_serial_stop_bits_set},
    { "-serial_parity",_wrap_rot_caps_serial_parity_get, _wrap_rot_caps_serial_parity_set},
    { "-serial_handshake",_wrap_rot_caps_serial_handshake_get, _wrap_rot_caps_serial_handshake_set},
    { "-write_delay",_wrap_rot_caps_write_delay_get, _wrap_rot_caps_write_delay_set},
    { "-post_write_delay",_wrap_rot_caps_post_write_delay_get, _wrap_rot_caps_post_write_delay_set},
    { "-timeout",_wrap_rot_caps_timeout_get, _wrap_rot_caps_timeout_set},
    { "-retry",_wrap_rot_caps_retry_get, _wrap_rot_caps_retry_set},
    { "-has_get_func",_wrap_rot_caps_has_get_func_get, _wrap_rot_caps_has_get_func_set},
    { "-has_set_func",_wrap_rot_caps_has_set_func_get, _wrap_rot_caps_has_set_func_set},
    { "-has_get_level",_wrap_rot_caps_has_get_level_get, _wrap_rot_caps_has_get_level_set},
    { "-has_set_level",_wrap_rot_caps_has_set_level_get, _wrap_rot_caps_has_set_level_set},
    { "-has_get_parm",_wrap_rot_caps_has_get_parm_get, _wrap_rot_caps_has_get_parm_set},
    { "-has_set_parm",_wrap_rot_caps_has_set_parm_get, _wrap_rot_caps_has_set_parm_set},
    { "-has_status",_wrap_rot_caps_has_status_get, _wrap_rot_caps_has_status_set},
    { "-level_gran",_wrap_rot_caps_level_gran_get, _wrap_rot_caps_level_gran_set},
    { "-parm_gran",_wrap_rot_caps_parm_gran_get, _wrap_rot_caps_parm_gran_set},
    { "-extparms",_wrap_rot_caps_extparms_get, _wrap_rot_caps_extparms_set},
    { "-extlevels",_wrap_rot_caps_extlevels_get, _wrap_rot_caps_extlevels_set},
    { "-extfuncs",_wrap_rot_caps_extfuncs_get, _wrap_rot_caps_extfuncs_set},
    { "-ext_tokens",_wrap_rot_caps_ext_tokens_get, _wrap_rot_caps_ext_tokens_set},
    { "-min_az",_wrap_rot_caps_min_az_get, _wrap_rot_caps_min_az_set},
    { "-max_az",_wrap_rot_caps_max_az_get, _wrap_rot_caps_max_az_set},
    { "-min_el",_wrap_rot_caps_min_el_get, _wrap_rot_caps_min_el_set},
    { "-max_el",_wrap_rot_caps_max_el_get, _wrap_rot_caps_max_el_set},
    { "-cfgparams",_wrap_rot_caps_cfgparams_get, _wrap_rot_caps_cfgparams_set},
    { "-priv",_wrap_rot_caps_priv_get, _wrap_rot_caps_priv_set},
    { "-rot_init",_wrap_rot_caps_rot_init_get, _wrap_rot_caps_rot_init_set},
    { "-rot_cleanup",_wrap_rot_caps_rot_cleanup_get, _wrap_rot_caps_rot_cleanup_set},
    { "-rot_open",_wrap_rot_caps_rot_open_get, _wrap_rot_caps_rot_open_set},
    { "-rot_close",_wrap_rot_caps_rot_close_get, _wrap_rot_caps_rot_close_set},
    { "-set_conf",_wrap_rot_caps_set_conf_get, _wrap_rot_caps_set_conf_set},
    { "-get_conf",_wrap_rot_caps_get_conf_get, _wrap_rot_caps_get_conf_set},
    { "-set_position",_wrap_rot_caps_set_position_get, _wrap_rot_caps_set_position_set},
    { "-get_position",_wrap_rot_caps_get_position_get, _wrap_rot_caps_get_position_set},
    { "-stop",_wrap_rot_caps_stop_get, _wrap_rot_caps_stop_set},
    { "-park",_wrap_rot_caps_park_get, _wrap_rot_caps_park_set},
    { "-reset",_wrap_rot_caps_reset_get, _wrap_rot_caps_reset_set},
    { "-move",_wrap_rot_caps_move_get, _wrap_rot_caps_move_set},
    { "-get_info",_wrap_rot_caps_get_info_get, _wrap_rot_caps_get_info_set},
    { "-set_level",_wrap_rot_caps_set_level_get, _wrap_rot_caps_set_level_set},
    { "-get_level",_wrap_rot_caps_get_level_get, _wrap_rot_caps_get_level_set},
    { "-set_func",_wrap_rot_caps_set_func_get, _wrap_rot_caps_set_func_set},
    { "-get_func",_wrap_rot_caps_get_func_get, _wrap_rot_caps_get_func_set},
    { "-set_parm",_wrap_rot_caps_set_parm_get, _wrap_rot_caps_set_parm_set},
    { "-get_parm",_wrap_rot_caps_get_parm_get, _wrap_rot_caps_get_parm_set},
    { "-set_ext_level",_wrap_rot_caps_set_ext_level_get, _wrap_rot_caps_set_ext_level_set},
    { "-get_ext_level",_wrap_rot_caps_get_ext_level_get, _wrap_rot_caps_get_ext_level_set},
    { "-set_ext_func",_wrap_rot_caps_set_ext_func_get, _wrap_rot_caps_set_ext_func_set},
    { "-get_ext_func",_wrap_rot_caps_get_ext_func_get, _wrap_rot_caps_get_ext_func_set},
    { "-set_ext_parm",_wrap_rot_caps_set_ext_parm_get, _wrap_rot_caps_set_ext_parm_set},
    { "-get_ext_parm",_wrap_rot_caps_get_ext_parm_get, _wrap_rot_caps_get_ext_parm_set},
    { "-get_status",_wrap_rot_caps_get_status_get, _wrap_rot_caps_get_status_set},
    { "-get_conf2",_wrap_rot_caps_get_conf2_get, _wrap_rot_caps_get_conf2_set},
    {0,0,0}
};
static swig_class *swig_rot_caps_bases[] = {0};
static const char * swig_rot_caps_base_names[] = {0};
static swig_class _wrap_class_rot_caps = { "rot_caps", &SWIGTYPE_p_rot_caps,_wrap_new_rot_caps, swig_delete_rot_caps, swig_rot_caps_methods, swig_rot_caps_attributes, swig_rot_caps_bases,swig_rot_caps_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_qrb(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  double arg1 ;
  double arg2 ;
  double arg3 ;
  double arg4 ;
  double *arg5 = 0 ;
  double *arg6 = 0 ;
  double val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  double temp5 ;
  int res5 = SWIG_TMPOBJ ;
  double temp6 ;
  int res6 = SWIG_TMPOBJ ;
  int result;
  
  arg5 = &temp5;
  arg6 = &temp6;
  if (SWIG_GetArgs(interp, objc, objv,"oooo:qrb lon1 lat1 lon2 lat2 ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "qrb" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = (double)(val1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "qrb" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  ecode3 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "qrb" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = (double)(val3);
  ecode4 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "qrb" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = (double)(val4);
  result = (int)qrb(arg1,arg2,arg3,arg4,arg5,arg6);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  if (SWIG_IsTmpObj(res5)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_double((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_double, new_flags));
  }
  if (SWIG_IsTmpObj(res6)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_double((*arg6)));
  } else {
    int new_flags = SWIG_IsNewObj(res6) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg6), SWIGTYPE_p_double, new_flags));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_distance_long_path(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  double arg1 ;
  double val1 ;
  int ecode1 = 0 ;
  double result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:distance_long_path distance ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "distance_long_path" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = (double)(val1);
  result = (double)distance_long_path(arg1);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_azimuth_long_path(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  double arg1 ;
  double val1 ;
  int ecode1 = 0 ;
  double result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:azimuth_long_path azimuth ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "azimuth_long_path" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = (double)(val1);
  result = (double)azimuth_long_path(arg1);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_dms2dec(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  int arg1 ;
  int arg2 ;
  double arg3 ;
  int arg4 ;
  int val1 ;
  int ecode1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  double result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oooo:dms2dec degrees minutes seconds sw ",(void *)0,(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "dms2dec" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = (int)(val1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "dms2dec" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "dms2dec" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = (double)(val3);
  ecode4 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[4], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "dms2dec" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = (int)(val4);
  result = (double)dms2dec(arg1,arg2,arg3,arg4);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_dec2dms(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  double arg1 ;
  int *arg2 = 0 ;
  int *arg3 = 0 ;
  double *arg4 = 0 ;
  int *arg5 = 0 ;
  double val1 ;
  int ecode1 = 0 ;
  int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  int temp3 ;
  int res3 = SWIG_TMPOBJ ;
  double temp4 ;
  int res4 = SWIG_TMPOBJ ;
  int temp5 ;
  int res5 = SWIG_TMPOBJ ;
  int result;
  
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  arg5 = &temp5;
  if (SWIG_GetArgs(interp, objc, objv,"o:dec2dms dec ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "dec2dms" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = (double)(val1);
  result = (int)dec2dms(arg1,arg2,arg3,arg4,arg5);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  if (SWIG_IsTmpObj(res2)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res4)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_double, new_flags));
  }
  if (SWIG_IsTmpObj(res5)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_int((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_int, new_flags));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_dec2dmmm(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  double arg1 ;
  int *arg2 = 0 ;
  double *arg3 = 0 ;
  int *arg4 = 0 ;
  double val1 ;
  int ecode1 = 0 ;
  int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  double temp3 ;
  int res3 = SWIG_TMPOBJ ;
  int temp4 ;
  int res4 = SWIG_TMPOBJ ;
  int result;
  
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if (SWIG_GetArgs(interp, objc, objv,"o:dec2dmmm dec ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "dec2dmmm" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = (double)(val1);
  result = (int)dec2dmmm(arg1,arg2,arg3,arg4);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  if (SWIG_IsTmpObj(res2)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_double, new_flags));
  }
  if (SWIG_IsTmpObj(res4)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_dmmm2dec(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  int arg1 ;
  double arg2 ;
  int arg3 ;
  int val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  double result;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:dmmm2dec degrees minutes sw ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "dmmm2dec" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = (int)(val1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "dmmm2dec" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "dmmm2dec" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  result = (double)dmmm2dec(arg1,arg2,arg3);
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_min_az_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  azimuth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_min_az_set self min_az ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_min_az_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_min_az_set" "', argument " "2"" of type '" "azimuth_t""'");
  } 
  arg2 = (azimuth_t)(val2);
  if (arg1) (arg1)->min_az = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_min_az_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  azimuth_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_min_az_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_min_az_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (azimuth_t) ((arg1)->min_az);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_max_az_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  azimuth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_max_az_set self max_az ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_max_az_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_max_az_set" "', argument " "2"" of type '" "azimuth_t""'");
  } 
  arg2 = (azimuth_t)(val2);
  if (arg1) (arg1)->max_az = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_max_az_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  azimuth_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_max_az_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_max_az_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (azimuth_t) ((arg1)->max_az);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_min_el_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  elevation_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_min_el_set self min_el ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_min_el_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_min_el_set" "', argument " "2"" of type '" "elevation_t""'");
  } 
  arg2 = (elevation_t)(val2);
  if (arg1) (arg1)->min_el = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_min_el_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  elevation_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_min_el_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_min_el_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (elevation_t) ((arg1)->min_el);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_max_el_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  elevation_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_max_el_set self max_el ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_max_el_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_max_el_set" "', argument " "2"" of type '" "elevation_t""'");
  } 
  arg2 = (elevation_t)(val2);
  if (arg1) (arg1)->max_el = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_max_el_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  elevation_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_max_el_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_max_el_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (elevation_t) ((arg1)->max_el);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_south_zero_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_south_zero_set self south_zero ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_south_zero_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_south_zero_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->south_zero = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_south_zero_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_south_zero_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_south_zero_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (int) ((arg1)->south_zero);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_az_offset_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  azimuth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_az_offset_set self az_offset ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_az_offset_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_az_offset_set" "', argument " "2"" of type '" "azimuth_t""'");
  } 
  arg2 = (azimuth_t)(val2);
  if (arg1) (arg1)->az_offset = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_az_offset_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  azimuth_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_az_offset_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_az_offset_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (azimuth_t) ((arg1)->az_offset);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_el_offset_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  elevation_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_el_offset_set self el_offset ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_el_offset_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_el_offset_set" "', argument " "2"" of type '" "elevation_t""'");
  } 
  arg2 = (elevation_t)(val2);
  if (arg1) (arg1)->el_offset = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_el_offset_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  elevation_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_el_offset_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_el_offset_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (elevation_t) ((arg1)->el_offset);
  Tcl_SetObjResult(interp,SWIG_From_float((float)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_get_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_has_get_func_set self has_get_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_get_func_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_get_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_get_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_has_get_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_get_func_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_func);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_set_func_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_has_set_func_set self has_set_func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_set_func_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_set_func_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_func = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_set_func_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_has_set_func_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_set_func_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_func);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_get_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_has_get_level_set self has_get_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_get_level_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_get_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_get_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_has_get_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_get_level_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_level);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_set_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_has_set_level_set self has_set_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_set_level_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_set_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_set_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_has_set_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_set_level_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_level);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_get_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_has_get_parm_set self has_get_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_get_parm_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_get_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_get_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_has_get_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_get_parm_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_parm);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_set_parm_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_has_set_parm_set self has_set_parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_set_parm_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_set_parm_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_parm = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_set_parm_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_has_set_parm_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_set_parm_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_parm);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_status_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  rot_status_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_has_status_set self has_status ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_status_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_has_status_set" "', argument " "2"" of type '" "rot_status_t""'");
  } 
  arg2 = (rot_status_t)(val2);
  if (arg1) (arg1)->has_status = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_has_status_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  rot_status_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_has_status_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_has_status_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (rot_status_t) ((arg1)->has_status);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_level_gran_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_level_gran_set self level_gran ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_level_gran_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_level_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->level_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""level_gran""' of type '""gran_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_level_gran_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gran_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_level_gran_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_level_gran_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->level_gran);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_parm_gran_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_parm_gran_set self parm_gran ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_parm_gran_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_parm_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->parm_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""parm_gran""' of type '""gran_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_parm_gran_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gran_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_parm_gran_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_parm_gran_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->parm_gran);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_comm_state_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_comm_state_set self comm_state ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_comm_state_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_comm_state_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->comm_state = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_comm_state_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_comm_state_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_comm_state_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (int) ((arg1)->comm_state);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_priv_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_priv_set self priv ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_priv_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_priv_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->priv = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_priv_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_priv_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_priv_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (void *) ((arg1)->priv);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_obj_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_obj_set self obj ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_obj_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_obj_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->obj = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_obj_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_obj_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_obj_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (void *) ((arg1)->obj);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_current_speed_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_current_speed_set self current_speed ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_current_speed_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rot_state_current_speed_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->current_speed = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_current_speed_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_current_speed_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_current_speed_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (int) ((arg1)->current_speed);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_rotport_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  hamlib_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_rotport_set self rotport ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_rotport_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_hamlib_port,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_rotport_set" "', argument " "2"" of type '" "hamlib_port_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rot_state_rotport_set" "', argument " "2"" of type '" "hamlib_port_t""'");
    } else {
      arg2 = *((hamlib_port_t *)(argp2));
    }
  }
  if (arg1) (arg1)->rotport = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_rotport_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  hamlib_port_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_rotport_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_rotport_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result =  ((arg1)->rotport);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((hamlib_port_t *)memcpy((hamlib_port_t *)calloc(1,sizeof(hamlib_port_t)),&result,sizeof(hamlib_port_t)), SWIGTYPE_p_hamlib_port, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_rotport2_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  hamlib_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_rotport2_set self rotport2 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_rotport2_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_hamlib_port,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_rotport2_set" "', argument " "2"" of type '" "hamlib_port_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "rot_state_rotport2_set" "', argument " "2"" of type '" "hamlib_port_t""'");
    } else {
      arg2 = *((hamlib_port_t *)(argp2));
    }
  }
  if (arg1) (arg1)->rotport2 = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_rotport2_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  hamlib_port_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_rotport2_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_rotport2_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result =  ((arg1)->rotport2);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((hamlib_port_t *)memcpy((hamlib_port_t *)calloc(1,sizeof(hamlib_port_t)),&result,sizeof(hamlib_port_t)), SWIGTYPE_p_hamlib_port, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_pstrotator_handler_priv_data_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void **arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_pstrotator_handler_priv_data_set self pstrotator_handler_priv_data ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_pstrotator_handler_priv_data_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_p_void, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_pstrotator_handler_priv_data_set" "', argument " "2"" of type '" "void **""'"); 
  }
  arg2 = (void **)(argp2);
  if (arg1) (arg1)->pstrotator_handler_priv_data = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_pstrotator_handler_priv_data_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void **result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_pstrotator_handler_priv_data_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_pstrotator_handler_priv_data_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (void **) ((arg1)->pstrotator_handler_priv_data);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_config_queue_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  deferred_config_header_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:rot_state_config_queue_set self config_queue ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_config_queue_set" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_deferred_config_header, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rot_state_config_queue_set" "', argument " "2"" of type '" "deferred_config_header_t *""'"); 
  }
  arg2 = (deferred_config_header_t *)(argp2);
  if (arg1) (arg1)->config_queue = *arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_rot_state_config_queue_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  deferred_config_header_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:rot_state_config_queue_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rot_state_config_queue_get" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  result = (deferred_config_header_t *)& ((arg1)->config_queue);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_deferred_config_header,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_rot_state(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_rot_state ") == TCL_ERROR) SWIG_fail;
  result = (struct rot_state *)calloc(1, sizeof(struct rot_state));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rot_state,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_rot_state(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct rot_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_rot_state self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_rot_state, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rot_state" "', argument " "1"" of type '" "struct rot_state *""'"); 
  }
  arg1 = (struct rot_state *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_rot_state(void *obj) {
struct rot_state *arg1 = (struct rot_state *) obj;
free((char *) arg1);
}
static swig_method swig_rot_state_methods[] = {
    {0,0}
};
static swig_attribute swig_rot_state_attributes[] = {
    { "-min_az",_wrap_rot_state_min_az_get, _wrap_rot_state_min_az_set},
    { "-max_az",_wrap_rot_state_max_az_get, _wrap_rot_state_max_az_set},
    { "-min_el",_wrap_rot_state_min_el_get, _wrap_rot_state_min_el_set},
    { "-max_el",_wrap_rot_state_max_el_get, _wrap_rot_state_max_el_set},
    { "-south_zero",_wrap_rot_state_south_zero_get, _wrap_rot_state_south_zero_set},
    { "-az_offset",_wrap_rot_state_az_offset_get, _wrap_rot_state_az_offset_set},
    { "-el_offset",_wrap_rot_state_el_offset_get, _wrap_rot_state_el_offset_set},
    { "-has_get_func",_wrap_rot_state_has_get_func_get, _wrap_rot_state_has_get_func_set},
    { "-has_set_func",_wrap_rot_state_has_set_func_get, _wrap_rot_state_has_set_func_set},
    { "-has_get_level",_wrap_rot_state_has_get_level_get, _wrap_rot_state_has_get_level_set},
    { "-has_set_level",_wrap_rot_state_has_set_level_get, _wrap_rot_state_has_set_level_set},
    { "-has_get_parm",_wrap_rot_state_has_get_parm_get, _wrap_rot_state_has_get_parm_set},
    { "-has_set_parm",_wrap_rot_state_has_set_parm_get, _wrap_rot_state_has_set_parm_set},
    { "-has_status",_wrap_rot_state_has_status_get, _wrap_rot_state_has_status_set},
    { "-level_gran",_wrap_rot_state_level_gran_get, _wrap_rot_state_level_gran_set},
    { "-parm_gran",_wrap_rot_state_parm_gran_get, _wrap_rot_state_parm_gran_set},
    { "-comm_state",_wrap_rot_state_comm_state_get, _wrap_rot_state_comm_state_set},
    { "-priv",_wrap_rot_state_priv_get, _wrap_rot_state_priv_set},
    { "-obj",_wrap_rot_state_obj_get, _wrap_rot_state_obj_set},
    { "-current_speed",_wrap_rot_state_current_speed_get, _wrap_rot_state_current_speed_set},
    { "-rotport",_wrap_rot_state_rotport_get, _wrap_rot_state_rotport_set},
    { "-rotport2",_wrap_rot_state_rotport2_get, _wrap_rot_state_rotport2_set},
    { "-pstrotator_handler_priv_data",_wrap_rot_state_pstrotator_handler_priv_data_get, _wrap_rot_state_pstrotator_handler_priv_data_set},
    { "-config_queue",_wrap_rot_state_config_queue_get, _wrap_rot_state_config_queue_set},
    {0,0,0}
};
static swig_class *swig_rot_state_bases[] = {0};
static const char * swig_rot_state_base_names[] = {0};
static swig_class _wrap_class_rot_state = { "rot_state", &SWIGTYPE_p_rot_state,_wrap_new_rot_state, swig_delete_rot_state, swig_rot_state_methods, swig_rot_state_attributes, swig_rot_state_bases,swig_rot_state_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_Rot_rot_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  ROT *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_rot_set self rot ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_rot_set" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_s_rot, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rot_rot_set" "', argument " "2"" of type '" "ROT *""'"); 
  }
  arg2 = (ROT *)(argp2);
  if (arg1) (arg1)->rot = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_rot_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ROT *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rot_rot_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_rot_get" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  result = (ROT *) ((arg1)->rot);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_s_rot,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_caps_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  struct rot_caps *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_caps_set self caps ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_caps_set" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_rot_caps, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rot_caps_set" "', argument " "2"" of type '" "struct rot_caps *""'"); 
  }
  arg2 = (struct rot_caps *)(argp2);
  if (arg1) (arg1)->caps = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_caps_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct rot_caps *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rot_caps_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_caps_get" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  result = (struct rot_caps *) ((arg1)->caps);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rot_caps,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_state_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  struct rot_state *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_state_set self state ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_state_set" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_rot_state, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rot_state_set" "', argument " "2"" of type '" "struct rot_state *""'"); 
  }
  arg2 = (struct rot_state *)(argp2);
  if (arg1) (arg1)->state = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_state_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct rot_state *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rot_state_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_state_get" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  result = (struct rot_state *) ((arg1)->state);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_rot_state,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_error_status_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_error_status_set self error_status ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_error_status_set" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_error_status_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->error_status = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_error_status_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rot_error_status_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_error_status_get" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  result = (int) ((arg1)->error_status);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_do_exception_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_do_exception_set self do_exception ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_do_exception_set" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_do_exception_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->do_exception = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_do_exception_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rot_do_exception_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_do_exception_get" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  result = (int) ((arg1)->do_exception);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_Rot(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  rot_model_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  struct Rot *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:new_Rot rot_model ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Rot" "', argument " "1"" of type '" "rot_model_t""'");
  } 
  arg1 = (rot_model_t)(val1);
  result = (struct Rot *)new_Rot(arg1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_Rot,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_Rot(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_Rot self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Rot" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  delete_Rot(arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_open(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rot_open self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_open" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rot_open(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_close(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rot_close self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_close" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rot_close(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_set_ext_func(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rot_set_ext_func self _hamlib_token_t_1 _int_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_ext_func" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_ext_func" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rot_set_ext_func" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rot_set_ext_func(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_set_ext_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rot_set_ext_level self _hamlib_token_t_1 _value_t_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_ext_level" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_ext_level" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    res3 = SWIG_ConvertPtr(objv[3], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rot_set_ext_level" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rot_set_ext_level" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Rot_set_ext_level(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_set_ext_parm(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rot_set_ext_parm self _hamlib_token_t_1 _value_t_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_ext_parm" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_ext_parm" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    res3 = SWIG_ConvertPtr(objv[3], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rot_set_ext_parm" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rot_set_ext_parm" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Rot_set_ext_parm(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_set_func(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rot_set_func self _setting_t_1 _int_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_func" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rot_set_func" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rot_set_func(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_set_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rot_set_level self _setting_t_1 _value_t_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_level" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    res3 = SWIG_ConvertPtr(objv[3], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rot_set_level" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rot_set_level" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Rot_set_level(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_set_parm(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rot_set_parm self _setting_t_1 _value_t_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_parm" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    res3 = SWIG_ConvertPtr(objv[3], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rot_set_parm" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Rot_set_parm" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Rot_set_parm(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_set_position(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  azimuth_t arg2 ;
  elevation_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rot_set_position self _azimuth_t_1 _elevation_t_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_position" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_position" "', argument " "2"" of type '" "azimuth_t""'");
  } 
  arg2 = (azimuth_t)(val2);
  ecode3 = SWIG_AsVal_float SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rot_set_position" "', argument " "3"" of type '" "elevation_t""'");
  } 
  arg3 = (elevation_t)(val3);
  {
    arg1->error_status = RIG_OK;
    Rot_set_position(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_get_position(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  azimuth_t *arg2 = 0 ;
  elevation_t *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  azimuth_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  elevation_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  if (SWIG_GetArgs(interp, objc, objv,"o:Rot_get_position self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_position" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rot_get_position(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (SWIG_IsTmpObj(res2)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_float((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_float, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_From_float((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_float, new_flags));
  }
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_stop(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rot_stop self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_stop" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rot_stop(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_park(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rot_park self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_park" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Rot_park(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_reset(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  rot_reset_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_reset self _rot_reset_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_reset" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_reset" "', argument " "2"" of type '" "rot_reset_t""'");
  } 
  arg2 = (rot_reset_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Rot_reset(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_move(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  int arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rot_move self _int_1 _int_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_move" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_move" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[3], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Rot_move" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  {
    arg1->error_status = RIG_OK;
    Rot_move(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_token_lookup(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  const_char_string arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_token_lookup self _const_char_string ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_token_lookup" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rot_token_lookup" "', argument " "2"" of type '" "const_char_string""'");
  }
  arg2 = (const_char_string)(buf2);
  {
    arg1->error_status = RIG_OK;
    Rot_token_lookup(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_set_conf__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rot_set_conf self name val ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_conf" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rot_set_conf" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(objv[3], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rot_set_conf" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rot_set_conf__SWIG_0(arg1,(char const *)arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_set_conf__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  const_char_string arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Rot_set_conf self _hamlib_token_t_1 _const_char_string_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_set_conf" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_set_conf" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(objv[3], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Rot_set_conf" "', argument " "3"" of type '" "const_char_string""'");
  }
  arg3 = (const_char_string)(buf3);
  {
    arg1->error_status = RIG_OK;
    Rot_set_conf__SWIG_1(arg1,arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_OK;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_set_conf(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rot, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Rot_set_conf__SWIG_1(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rot, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Rot_set_conf__SWIG_0(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rot_set_conf'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rot::set_conf(char const *,char const *)\n"
    "    Rot::set_conf(hamlib_token_t,const_char_string)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_has_get_func(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_has_get_func self func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_has_get_func" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_has_get_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Rot_has_get_func(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_has_set_func(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_has_set_func self func ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_has_set_func" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_has_set_func" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Rot_has_set_func(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_has_get_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_has_get_level self level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_has_get_level" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_has_get_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Rot_has_get_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_has_set_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_has_set_level self level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_has_set_level" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_has_set_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Rot_has_set_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_has_get_parm(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_has_get_parm self parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_has_get_parm" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_has_get_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Rot_has_get_parm(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_has_set_parm(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_has_set_parm self parm ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_has_set_parm" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_has_set_parm" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Rot_has_set_parm(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_get_conf__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_get_conf self tok ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_conf" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Rot_get_conf" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Rot_get_conf__SWIG_0(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_get_conf__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Rot_get_conf self name ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_conf" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rot_get_conf" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Rot_get_conf__SWIG_1(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_get_conf(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rot, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Rot_get_conf__SWIG_0(clientData, interp, objc, argv - 1);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Rot, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Rot_get_conf__SWIG_1(clientData, interp, objc, argv - 1);
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Rot_get_conf'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Rot::get_conf(hamlib_token_t)\n"
    "    Rot::get_conf(char const *)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Rot_get_info(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Rot *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Rot_get_info self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Rot, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rot_get_info" "', argument " "1"" of type '" "struct Rot *""'"); 
  }
  arg1 = (struct Rot *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Rot_get_info(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_Rot(void *obj) {
struct Rot *arg1 = (struct Rot *) obj;
delete_Rot(arg1);
}
static swig_method swig_Rot_methods[] = {
    {"open", _wrap_Rot_open}, 
    {"close", _wrap_Rot_close}, 
    {"set_ext_func", _wrap_Rot_set_ext_func}, 
    {"set_ext_level", _wrap_Rot_set_ext_level}, 
    {"set_ext_parm", _wrap_Rot_set_ext_parm}, 
    {"set_func", _wrap_Rot_set_func}, 
    {"set_level", _wrap_Rot_set_level}, 
    {"set_parm", _wrap_Rot_set_parm}, 
    {"set_position", _wrap_Rot_set_position}, 
    {"get_position", _wrap_Rot_get_position}, 
    {"stop", _wrap_Rot_stop}, 
    {"park", _wrap_Rot_park}, 
    {"reset", _wrap_Rot_reset}, 
    {"move", _wrap_Rot_move}, 
    {"token_lookup", _wrap_Rot_token_lookup}, 
    {"set_conf", _wrap_Rot_set_conf}, 
    {"has_get_func", _wrap_Rot_has_get_func}, 
    {"has_set_func", _wrap_Rot_has_set_func}, 
    {"has_get_level", _wrap_Rot_has_get_level}, 
    {"has_set_level", _wrap_Rot_has_set_level}, 
    {"has_get_parm", _wrap_Rot_has_get_parm}, 
    {"has_set_parm", _wrap_Rot_has_set_parm}, 
    {"get_conf", _wrap_Rot_get_conf}, 
    {"get_info", _wrap_Rot_get_info}, 
    {0,0}
};
static swig_attribute swig_Rot_attributes[] = {
    { "-rot",_wrap_Rot_rot_get, _wrap_Rot_rot_set},
    { "-caps",_wrap_Rot_caps_get, _wrap_Rot_caps_set},
    { "-state",_wrap_Rot_state_get, _wrap_Rot_state_set},
    { "-error_status",_wrap_Rot_error_status_get, _wrap_Rot_error_status_set},
    { "-do_exception",_wrap_Rot_do_exception_get, _wrap_Rot_do_exception_set},
    {0,0,0}
};
static swig_class *swig_Rot_bases[] = {0};
static const char * swig_Rot_base_names[] = {0};
static swig_class _wrap_class_Rot = { "Rot", &SWIGTYPE_p_Rot,_wrap_new_Rot, swig_delete_Rot, swig_Rot_methods, swig_Rot_attributes, swig_Rot_bases,swig_Rot_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_amp_caps_amp_model_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  amp_model_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_amp_model_set self amp_model ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_model_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_amp_model_set" "', argument " "2"" of type '" "amp_model_t""'");
  } 
  arg2 = (amp_model_t)(val2);
  if (arg1) (arg1)->amp_model = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_model_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  amp_model_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_amp_model_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_model_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (amp_model_t) ((arg1)->amp_model);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_model_name_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_model_name_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_model_name_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (char *) ((arg1)->model_name);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_mfg_name_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_mfg_name_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_mfg_name_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (char *) ((arg1)->mfg_name);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_version_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_version_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_version_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (char *) ((arg1)->version);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_copyright_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_copyright_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_copyright_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (char *) ((arg1)->copyright);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_status_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  enum rig_status_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_status_set self status ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_status_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_status_set" "', argument " "2"" of type '" "enum rig_status_e""'");
  } 
  arg2 = (enum rig_status_e)(val2);
  if (arg1) (arg1)->status = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_status_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum rig_status_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_status_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_status_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (enum rig_status_e) ((arg1)->status);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_amp_type_set self amp_type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_type_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_amp_type_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->amp_type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_amp_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_type_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->amp_type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_port_type_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  enum rig_port_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_port_type_set self port_type ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_port_type_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_port_type_set" "', argument " "2"" of type '" "enum rig_port_e""'");
  } 
  arg2 = (enum rig_port_e)(val2);
  if (arg1) (arg1)->port_type = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_port_type_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum rig_port_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_port_type_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_port_type_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (enum rig_port_e) ((arg1)->port_type);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_serial_rate_min_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_serial_rate_min_set self serial_rate_min ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_rate_min_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_serial_rate_min_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_min = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_serial_rate_min_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_serial_rate_min_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_rate_min_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->serial_rate_min);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_serial_rate_max_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_serial_rate_max_set self serial_rate_max ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_rate_max_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_serial_rate_max_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_rate_max = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_serial_rate_max_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_serial_rate_max_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_rate_max_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->serial_rate_max);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_serial_data_bits_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_serial_data_bits_set self serial_data_bits ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_data_bits_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_serial_data_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_data_bits = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_serial_data_bits_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_serial_data_bits_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_data_bits_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->serial_data_bits);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_serial_stop_bits_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_serial_stop_bits_set self serial_stop_bits ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_stop_bits_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_serial_stop_bits_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->serial_stop_bits = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_serial_stop_bits_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_serial_stop_bits_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_stop_bits_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->serial_stop_bits);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_serial_parity_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  enum serial_parity_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_serial_parity_set self serial_parity ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_parity_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_serial_parity_set" "', argument " "2"" of type '" "enum serial_parity_e""'");
  } 
  arg2 = (enum serial_parity_e)(val2);
  if (arg1) (arg1)->serial_parity = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_serial_parity_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum serial_parity_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_serial_parity_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_parity_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (enum serial_parity_e) ((arg1)->serial_parity);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_serial_handshake_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  enum serial_handshake_e arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_serial_handshake_set self serial_handshake ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_handshake_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_serial_handshake_set" "', argument " "2"" of type '" "enum serial_handshake_e""'");
  } 
  arg2 = (enum serial_handshake_e)(val2);
  if (arg1) (arg1)->serial_handshake = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_serial_handshake_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum serial_handshake_e result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_serial_handshake_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_serial_handshake_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (enum serial_handshake_e) ((arg1)->serial_handshake);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_write_delay_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_write_delay_set self write_delay ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_write_delay_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->write_delay = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_write_delay_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_write_delay_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_write_delay_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->write_delay);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_post_write_delay_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_post_write_delay_set self post_write_delay ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_post_write_delay_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_post_write_delay_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->post_write_delay = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_post_write_delay_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_post_write_delay_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_post_write_delay_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->post_write_delay);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_timeout_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_timeout_set self timeout ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_timeout_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_timeout_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->timeout = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_timeout_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_timeout_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_timeout_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->timeout);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_retry_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_retry_set self retry ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_retry_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_retry_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->retry = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_retry_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_retry_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_retry_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int) ((arg1)->retry);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_cfgparams_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_cfgparams_set self cfgparams ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_cfgparams_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_cfgparams_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->cfgparams = (struct confparams const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_cfgparams_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct confparams *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_cfgparams_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_cfgparams_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (struct confparams *) ((arg1)->cfgparams);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_priv_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_priv_set self priv ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_priv_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_priv_set" "', argument " "2"" of type '" "void const *""'"); 
  }
  if (arg1) (arg1)->priv = (void const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_priv_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_priv_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_priv_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (void *) ((arg1)->priv);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_model_macro_name_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_amp_model_macro_name_set self amp_model_macro_name ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_model_macro_name_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_amp_model_macro_name_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->amp_model_macro_name = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->amp_model_macro_name = 0;
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_model_macro_name_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_amp_model_macro_name_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_model_macro_name_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (char *) ((arg1)->amp_model_macro_name);
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_has_get_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_has_get_level_set self has_get_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_has_get_level_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_has_get_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_has_get_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_has_get_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_has_get_level_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (setting_t) ((arg1)->has_get_level);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_has_set_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_has_set_level_set self has_set_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_has_set_level_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_has_set_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_has_set_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_has_set_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_has_set_level_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (setting_t) ((arg1)->has_set_level);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_level_gran_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_level_gran_set self level_gran ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_level_gran_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_level_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->level_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""level_gran""' of type '""gran_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_level_gran_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gran_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_level_gran_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_level_gran_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->level_gran);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_parm_gran_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_parm_gran_set self parm_gran ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_parm_gran_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_parm_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->parm_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""parm_gran""' of type '""gran_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_parm_gran_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gran_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_parm_gran_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_parm_gran_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->parm_gran);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_init_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_amp_init_set self amp_init ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_init_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_amp_init_set" "', argument " "2"" of type '" "int (*)(AMP *)""'"); 
    }
  }
  if (arg1) (arg1)->amp_init = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_init_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_amp_init_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_init_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *)) ((arg1)->amp_init);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_cleanup_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_amp_cleanup_set self amp_cleanup ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_cleanup_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_amp_cleanup_set" "', argument " "2"" of type '" "int (*)(AMP *)""'"); 
    }
  }
  if (arg1) (arg1)->amp_cleanup = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_cleanup_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_amp_cleanup_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_cleanup_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *)) ((arg1)->amp_cleanup);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_open_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_amp_open_set self amp_open ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_open_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_amp_open_set" "', argument " "2"" of type '" "int (*)(AMP *)""'"); 
    }
  }
  if (arg1) (arg1)->amp_open = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_open_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_amp_open_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_open_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *)) ((arg1)->amp_open);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_close_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_amp_close_set self amp_close ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_close_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_amp_close_set" "', argument " "2"" of type '" "int (*)(AMP *)""'"); 
    }
  }
  if (arg1) (arg1)->amp_close = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_amp_close_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_amp_close_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_amp_close_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *)) ((arg1)->amp_close);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_set_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,freq_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_set_freq_set self set_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_freq_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_double__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_set_freq_set" "', argument " "2"" of type '" "int (*)(AMP *,freq_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_set_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *,freq_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_set_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_freq_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,freq_t)) ((arg1)->set_freq);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_double__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_freq_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,freq_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_get_freq_set self get_freq ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_freq_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_p_double__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_freq_set" "', argument " "2"" of type '" "int (*)(AMP *,freq_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_freq = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_freq_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *,freq_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_get_freq_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_freq_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,freq_t *)) ((arg1)->get_freq);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_p_double__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_set_conf_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,hamlib_token_t,char const *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_set_conf_set self set_conf ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_conf_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_long_p_q_const__char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_set_conf_set" "', argument " "2"" of type '" "int (*)(AMP *,hamlib_token_t,char const *)""'"); 
    }
  }
  if (arg1) (arg1)->set_conf = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_set_conf_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *,hamlib_token_t,char const *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_set_conf_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_conf_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,hamlib_token_t,char const *)) ((arg1)->set_conf);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_long_p_q_const__char__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_conf2_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,hamlib_token_t,char *,int) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_get_conf2_set self get_conf2 ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_conf2_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_long_p_char_int__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_conf2_set" "', argument " "2"" of type '" "int (*)(AMP *,hamlib_token_t,char *,int)""'"); 
    }
  }
  if (arg1) (arg1)->get_conf2 = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_conf2_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *,hamlib_token_t,char *,int) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_get_conf2_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_conf2_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,hamlib_token_t,char *,int)) ((arg1)->get_conf2);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_long_p_char_int__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_conf_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,hamlib_token_t,char *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_get_conf_set self get_conf ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_conf_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_long_p_char__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_conf_set" "', argument " "2"" of type '" "int (*)(AMP *,hamlib_token_t,char *)""'"); 
    }
  }
  if (arg1) (arg1)->get_conf = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_conf_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *,hamlib_token_t,char *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_get_conf_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_conf_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,hamlib_token_t,char *)) ((arg1)->get_conf);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_long_p_char__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_reset_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,amp_reset_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_reset_set self reset ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_reset_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_enum_amp_reset_e__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_reset_set" "', argument " "2"" of type '" "int (*)(AMP *,amp_reset_t)""'"); 
    }
  }
  if (arg1) (arg1)->reset = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_reset_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *,amp_reset_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_reset_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_reset_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,amp_reset_t)) ((arg1)->reset);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_enum_amp_reset_e__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,setting_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_get_level_set self get_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_level_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_unsigned_long_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_level_set" "', argument " "2"" of type '" "int (*)(AMP *,setting_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *,setting_t,value_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_get_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_level_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,setting_t,value_t *)) ((arg1)->get_level);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_unsigned_long_long_p_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_set_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,setting_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_set_level_set self set_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_level_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_unsigned_long_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_set_level_set" "', argument " "2"" of type '" "int (*)(AMP *,setting_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_set_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *,setting_t,value_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_set_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_level_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,setting_t,value_t)) ((arg1)->set_level);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_unsigned_long_long_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_ext_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,hamlib_token_t,value_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_get_ext_level_set self get_ext_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_ext_level_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_long_p_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_ext_level_set" "', argument " "2"" of type '" "int (*)(AMP *,hamlib_token_t,value_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_ext_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_ext_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *,hamlib_token_t,value_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_get_ext_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_ext_level_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,hamlib_token_t,value_t *)) ((arg1)->get_ext_level);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_long_p_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_set_ext_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,hamlib_token_t,value_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_set_ext_level_set self set_ext_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_ext_level_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_long_value_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_set_ext_level_set" "', argument " "2"" of type '" "int (*)(AMP *,hamlib_token_t,value_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_ext_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_set_ext_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *,hamlib_token_t,value_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_set_ext_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_ext_level_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,hamlib_token_t,value_t)) ((arg1)->set_ext_level);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_long_value_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_set_powerstat_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,powerstat_t) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_set_powerstat_set self set_powerstat ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_powerstat_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_enum_powerstat_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_set_powerstat_set" "', argument " "2"" of type '" "int (*)(AMP *,powerstat_t)""'"); 
    }
  }
  if (arg1) (arg1)->set_powerstat = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_set_powerstat_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *,powerstat_t) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_set_powerstat_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_set_powerstat_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,powerstat_t)) ((arg1)->set_powerstat);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_enum_powerstat_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_powerstat_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  int (*arg2)(AMP *,powerstat_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_get_powerstat_set self get_powerstat ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_powerstat_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp_p_enum_powerstat_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_powerstat_set" "', argument " "2"" of type '" "int (*)(AMP *,powerstat_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_powerstat = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_powerstat_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int (*result)(AMP *,powerstat_t *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_get_powerstat_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_powerstat_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (int (*)(AMP *,powerstat_t *)) ((arg1)->get_powerstat);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp_p_enum_powerstat_t__int));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_info_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  char *(*arg2)(AMP *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_get_info_set self get_info ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_info_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(objv[2], (void**)(&arg2), SWIGTYPE_p_f_p_struct_amp__p_char);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "amp_caps_get_info_set" "', argument " "2"" of type '" "char const *(*)(AMP *)""'"); 
    }
  }
  if (arg1) (arg1)->get_info = (char const *(*)(AMP *))arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_get_info_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *(*result)(AMP *) = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_get_info_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_get_info_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (char *(*)(AMP *)) ((arg1)->get_info);
  Tcl_SetObjResult(interp,SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_struct_amp__p_char));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_levels_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_levels_set self levels ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_levels_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_levels_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->levels = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_levels_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_levels_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_levels_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (setting_t) ((arg1)->levels);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_ext_levels_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_ext_levels_set self ext_levels ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_ext_levels_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_caps_ext_levels_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  if (arg1) (arg1)->ext_levels = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_ext_levels_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_ext_levels_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_ext_levels_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (unsigned int) ((arg1)->ext_levels);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_int((unsigned int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_extlevels_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_extlevels_set self extlevels ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_extlevels_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_extlevels_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extlevels = (struct confparams const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_extlevels_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct confparams *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_extlevels_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_extlevels_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extlevels);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_extparms_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  struct confparams *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_caps_extparms_set self extparms ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_extparms_set" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_confparams, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_caps_extparms_set" "', argument " "2"" of type '" "struct confparams const *""'"); 
  }
  arg2 = (struct confparams *)(argp2);
  if (arg1) (arg1)->extparms = (struct confparams const *)arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_caps_extparms_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct confparams *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_caps_extparms_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_caps_extparms_get" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  result = (struct confparams *) ((arg1)->extparms);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_confparams,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_amp_caps(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_amp_caps ") == TCL_ERROR) SWIG_fail;
  result = (struct amp_caps *)calloc(1, sizeof(struct amp_caps));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_amp_caps,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_amp_caps(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_caps *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_amp_caps self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_caps, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_amp_caps" "', argument " "1"" of type '" "struct amp_caps *""'"); 
  }
  arg1 = (struct amp_caps *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_amp_caps(void *obj) {
struct amp_caps *arg1 = (struct amp_caps *) obj;
free((char *) arg1);
}
static swig_method swig_amp_caps_methods[] = {
    {0,0}
};
static swig_attribute swig_amp_caps_attributes[] = {
    { "-amp_model",_wrap_amp_caps_amp_model_get, _wrap_amp_caps_amp_model_set},
    { "-model_name",_wrap_amp_caps_model_name_get, 0 },
    { "-mfg_name",_wrap_amp_caps_mfg_name_get, 0 },
    { "-version",_wrap_amp_caps_version_get, 0 },
    { "-copyright",_wrap_amp_caps_copyright_get, 0 },
    { "-status",_wrap_amp_caps_status_get, _wrap_amp_caps_status_set},
    { "-amp_type",_wrap_amp_caps_amp_type_get, _wrap_amp_caps_amp_type_set},
    { "-port_type",_wrap_amp_caps_port_type_get, _wrap_amp_caps_port_type_set},
    { "-serial_rate_min",_wrap_amp_caps_serial_rate_min_get, _wrap_amp_caps_serial_rate_min_set},
    { "-serial_rate_max",_wrap_amp_caps_serial_rate_max_get, _wrap_amp_caps_serial_rate_max_set},
    { "-serial_data_bits",_wrap_amp_caps_serial_data_bits_get, _wrap_amp_caps_serial_data_bits_set},
    { "-serial_stop_bits",_wrap_amp_caps_serial_stop_bits_get, _wrap_amp_caps_serial_stop_bits_set},
    { "-serial_parity",_wrap_amp_caps_serial_parity_get, _wrap_amp_caps_serial_parity_set},
    { "-serial_handshake",_wrap_amp_caps_serial_handshake_get, _wrap_amp_caps_serial_handshake_set},
    { "-write_delay",_wrap_amp_caps_write_delay_get, _wrap_amp_caps_write_delay_set},
    { "-post_write_delay",_wrap_amp_caps_post_write_delay_get, _wrap_amp_caps_post_write_delay_set},
    { "-timeout",_wrap_amp_caps_timeout_get, _wrap_amp_caps_timeout_set},
    { "-retry",_wrap_amp_caps_retry_get, _wrap_amp_caps_retry_set},
    { "-cfgparams",_wrap_amp_caps_cfgparams_get, _wrap_amp_caps_cfgparams_set},
    { "-priv",_wrap_amp_caps_priv_get, _wrap_amp_caps_priv_set},
    { "-amp_model_macro_name",_wrap_amp_caps_amp_model_macro_name_get, _wrap_amp_caps_amp_model_macro_name_set},
    { "-has_get_level",_wrap_amp_caps_has_get_level_get, _wrap_amp_caps_has_get_level_set},
    { "-has_set_level",_wrap_amp_caps_has_set_level_get, _wrap_amp_caps_has_set_level_set},
    { "-level_gran",_wrap_amp_caps_level_gran_get, _wrap_amp_caps_level_gran_set},
    { "-parm_gran",_wrap_amp_caps_parm_gran_get, _wrap_amp_caps_parm_gran_set},
    { "-amp_init",_wrap_amp_caps_amp_init_get, _wrap_amp_caps_amp_init_set},
    { "-amp_cleanup",_wrap_amp_caps_amp_cleanup_get, _wrap_amp_caps_amp_cleanup_set},
    { "-amp_open",_wrap_amp_caps_amp_open_get, _wrap_amp_caps_amp_open_set},
    { "-amp_close",_wrap_amp_caps_amp_close_get, _wrap_amp_caps_amp_close_set},
    { "-set_freq",_wrap_amp_caps_set_freq_get, _wrap_amp_caps_set_freq_set},
    { "-get_freq",_wrap_amp_caps_get_freq_get, _wrap_amp_caps_get_freq_set},
    { "-set_conf",_wrap_amp_caps_set_conf_get, _wrap_amp_caps_set_conf_set},
    { "-get_conf2",_wrap_amp_caps_get_conf2_get, _wrap_amp_caps_get_conf2_set},
    { "-get_conf",_wrap_amp_caps_get_conf_get, _wrap_amp_caps_get_conf_set},
    { "-reset",_wrap_amp_caps_reset_get, _wrap_amp_caps_reset_set},
    { "-get_level",_wrap_amp_caps_get_level_get, _wrap_amp_caps_get_level_set},
    { "-set_level",_wrap_amp_caps_set_level_get, _wrap_amp_caps_set_level_set},
    { "-get_ext_level",_wrap_amp_caps_get_ext_level_get, _wrap_amp_caps_get_ext_level_set},
    { "-set_ext_level",_wrap_amp_caps_set_ext_level_get, _wrap_amp_caps_set_ext_level_set},
    { "-set_powerstat",_wrap_amp_caps_set_powerstat_get, _wrap_amp_caps_set_powerstat_set},
    { "-get_powerstat",_wrap_amp_caps_get_powerstat_get, _wrap_amp_caps_get_powerstat_set},
    { "-get_info",_wrap_amp_caps_get_info_get, _wrap_amp_caps_get_info_set},
    { "-levels",_wrap_amp_caps_levels_get, _wrap_amp_caps_levels_set},
    { "-ext_levels",_wrap_amp_caps_ext_levels_get, _wrap_amp_caps_ext_levels_set},
    { "-extlevels",_wrap_amp_caps_extlevels_get, _wrap_amp_caps_extlevels_set},
    { "-extparms",_wrap_amp_caps_extparms_get, _wrap_amp_caps_extparms_set},
    {0,0,0}
};
static swig_class *swig_amp_caps_bases[] = {0};
static const char * swig_amp_caps_base_names[] = {0};
static swig_class _wrap_class_amp_caps = { "amp_caps", &SWIGTYPE_p_amp_caps,_wrap_new_amp_caps, swig_delete_amp_caps, swig_amp_caps_methods, swig_amp_caps_attributes, swig_amp_caps_bases,swig_amp_caps_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_amp_state_comm_state_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_state_comm_state_set self comm_state ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_comm_state_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_state_comm_state_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->comm_state = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_comm_state_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_state_comm_state_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_comm_state_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (int) ((arg1)->comm_state);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_priv_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_state_priv_set self priv ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_priv_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_state_priv_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->priv = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_priv_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_state_priv_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_priv_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (void *) ((arg1)->priv);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_obj_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_state_obj_set self obj ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_obj_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_state_obj_set" "', argument " "2"" of type '" "void *""'"); 
  }
  if (arg1) (arg1)->obj = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_obj_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_state_obj_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_obj_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (void *) ((arg1)->obj);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_void,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_has_get_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_state_has_get_level_set self has_get_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_has_get_level_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_state_has_get_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_get_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_has_get_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_state_has_get_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_has_get_level_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (setting_t) ((arg1)->has_get_level);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_has_set_level_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_state_has_set_level_set self has_set_level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_has_set_level_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "amp_state_has_set_level_set" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  if (arg1) (arg1)->has_set_level = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_has_set_level_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_state_has_set_level_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_has_set_level_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (setting_t) ((arg1)->has_set_level);
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_level_gran_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_state_level_gran_set self level_gran ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_level_gran_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_state_level_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->level_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""level_gran""' of type '""gran_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_level_gran_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gran_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_state_level_gran_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_level_gran_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->level_gran);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_parm_gran_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  gran_t *arg2 = (gran_t *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_state_parm_gran_set self parm_gran ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_parm_gran_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_gran, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_state_parm_gran_set" "', argument " "2"" of type '" "gran_t [64]""'"); 
  } 
  arg2 = (gran_t *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(gran_t *)&arg1->parm_gran[ii] = *((gran_t *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""parm_gran""' of type '""gran_t [64]""'");
    }
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_parm_gran_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gran_t *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_state_parm_gran_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_parm_gran_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result = (gran_t *)(gran_t *) ((arg1)->parm_gran);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_gran,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_ampport_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  hamlib_port_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:amp_state_ampport_set self ampport ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_ampport_set" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  {
    res2 = SWIG_ConvertPtr(objv[2], &argp2, SWIGTYPE_p_hamlib_port,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "amp_state_ampport_set" "', argument " "2"" of type '" "hamlib_port_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "amp_state_ampport_set" "', argument " "2"" of type '" "hamlib_port_t""'");
    } else {
      arg2 = *((hamlib_port_t *)(argp2));
    }
  }
  if (arg1) (arg1)->ampport = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_amp_state_ampport_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  hamlib_port_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:amp_state_ampport_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "amp_state_ampport_get" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  result =  ((arg1)->ampport);
  Tcl_SetObjResult(interp,SWIG_NewInstanceObj((hamlib_port_t *)memcpy((hamlib_port_t *)calloc(1,sizeof(hamlib_port_t)),&result,sizeof(hamlib_port_t)), SWIGTYPE_p_hamlib_port, SWIG_POINTER_OWN | 0 ));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_amp_state(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,":new_amp_state ") == TCL_ERROR) SWIG_fail;
  result = (struct amp_state *)calloc(1, sizeof(struct amp_state));
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_amp_state,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_amp_state(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct amp_state *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_amp_state self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_amp_state, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_amp_state" "', argument " "1"" of type '" "struct amp_state *""'"); 
  }
  arg1 = (struct amp_state *)(argp1);
  free((char *) arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_amp_state(void *obj) {
struct amp_state *arg1 = (struct amp_state *) obj;
free((char *) arg1);
}
static swig_method swig_amp_state_methods[] = {
    {0,0}
};
static swig_attribute swig_amp_state_attributes[] = {
    { "-comm_state",_wrap_amp_state_comm_state_get, _wrap_amp_state_comm_state_set},
    { "-priv",_wrap_amp_state_priv_get, _wrap_amp_state_priv_set},
    { "-obj",_wrap_amp_state_obj_get, _wrap_amp_state_obj_set},
    { "-has_get_level",_wrap_amp_state_has_get_level_get, _wrap_amp_state_has_get_level_set},
    { "-has_set_level",_wrap_amp_state_has_set_level_get, _wrap_amp_state_has_set_level_set},
    { "-level_gran",_wrap_amp_state_level_gran_get, _wrap_amp_state_level_gran_set},
    { "-parm_gran",_wrap_amp_state_parm_gran_get, _wrap_amp_state_parm_gran_set},
    { "-ampport",_wrap_amp_state_ampport_get, _wrap_amp_state_ampport_set},
    {0,0,0}
};
static swig_class *swig_amp_state_bases[] = {0};
static const char * swig_amp_state_base_names[] = {0};
static swig_class _wrap_class_amp_state = { "amp_state", &SWIGTYPE_p_amp_state,_wrap_new_amp_state, swig_delete_amp_state, swig_amp_state_methods, swig_amp_state_attributes, swig_amp_state_bases,swig_amp_state_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };
SWIGINTERN int
_wrap_Amp_caps_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  struct amp_caps *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Amp_caps_set self caps ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_caps_set" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_amp_caps, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Amp_caps_set" "', argument " "2"" of type '" "struct amp_caps *""'"); 
  }
  arg2 = (struct amp_caps *)(argp2);
  if (arg1) (arg1)->caps = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_caps_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct amp_caps *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Amp_caps_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_caps_get" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  result = (struct amp_caps *) ((arg1)->caps);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_amp_caps,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_state_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  struct amp_state *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Amp_state_set self state ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_state_set" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  res2 = SWIG_ConvertPtr(objv[2], &argp2,SWIGTYPE_p_amp_state, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Amp_state_set" "', argument " "2"" of type '" "struct amp_state *""'"); 
  }
  arg2 = (struct amp_state *)(argp2);
  if (arg1) (arg1)->state = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_state_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct amp_state *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Amp_state_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_state_get" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  result = (struct amp_state *) ((arg1)->state);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_amp_state,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_error_status_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Amp_error_status_set self error_status ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_error_status_set" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_error_status_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->error_status = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_error_status_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Amp_error_status_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_error_status_get" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  result = (int) ((arg1)->error_status);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_do_exception_set(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Amp_do_exception_set self do_exception ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_do_exception_set" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_do_exception_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->do_exception = arg2;
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_do_exception_get(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Amp_do_exception_get self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_do_exception_get" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  result = (int) ((arg1)->do_exception);
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_new_Amp(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  amp_model_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  struct Amp *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:new_Amp amp_model ",(void *)0) == TCL_ERROR) SWIG_fail;
  ecode1 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[1], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Amp" "', argument " "1"" of type '" "amp_model_t""'");
  } 
  arg1 = (amp_model_t)(val1);
  result = (struct Amp *)new_Amp(arg1);
  Tcl_SetObjResult(interp, SWIG_NewInstanceObj( SWIG_as_voidptr(result), SWIGTYPE_p_Amp,0));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_delete_Amp(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:delete_Amp self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Amp" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  delete_Amp(arg1);
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_open(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Amp_open self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_open" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Amp_open(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_close(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Amp_close self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_close" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  {
    arg1->error_status = RIG_OK;
    Amp_close(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_reset(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  amp_reset_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Amp_reset self _amp_reset_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_reset" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_reset" "', argument " "2"" of type '" "amp_reset_t""'");
  } 
  arg2 = (amp_reset_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Amp_reset(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_token_lookup(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  const_char_string arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Amp_token_lookup self _const_char_string ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_token_lookup" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Amp_token_lookup" "', argument " "2"" of type '" "const_char_string""'");
  }
  arg2 = (const_char_string)(buf2);
  {
    arg1->error_status = RIG_OK;
    Amp_token_lookup(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_has_get_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Amp_has_get_level self level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_has_get_level" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_has_get_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Amp_has_get_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_has_set_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  setting_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  setting_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Amp_has_set_level self level ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_has_set_level" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_has_set_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (setting_t)Amp_has_set_level(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_set_freq(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  freq_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Amp_set_freq self _freq_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_set_freq" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_double SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_set_freq" "', argument " "2"" of type '" "freq_t""'");
  } 
  arg2 = (freq_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Amp_set_freq(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_set_powerstat(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  powerstat_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Amp_set_powerstat self _powerstat_t ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_set_powerstat" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_int SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_set_powerstat" "', argument " "2"" of type '" "powerstat_t""'");
  } 
  arg2 = (powerstat_t)(val2);
  {
    arg1->error_status = RIG_OK;
    Amp_set_powerstat(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_set_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  setting_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Amp_set_level self _setting_t_1 _value_t_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_set_level" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_set_level" "', argument " "2"" of type '" "setting_t""'");
  } 
  arg2 = (setting_t)(val2);
  {
    res3 = SWIG_ConvertPtr(objv[3], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Amp_set_level" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Amp_set_level" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Amp_set_level(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_set_ext_level(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  hamlib_token_t arg2 ;
  value_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Amp_set_ext_level self _hamlib_token_t_1 _value_t_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_set_ext_level" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_set_ext_level" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    res3 = SWIG_ConvertPtr(objv[3], &argp3, SWIGTYPE_p_value_t,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Amp_set_ext_level" "', argument " "3"" of type '" "value_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Amp_set_ext_level" "', argument " "3"" of type '" "value_t""'");
    } else {
      arg3 = *((value_t *)(argp3));
    }
  }
  {
    arg1->error_status = RIG_OK;
    Amp_set_ext_level(arg1,arg2,arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_set_conf__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Amp_set_conf self name val ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_set_conf" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Amp_set_conf" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(objv[3], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Amp_set_conf" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  {
    arg1->error_status = RIG_OK;
    Amp_set_conf__SWIG_0(arg1,(char const *)arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_set_conf__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  hamlib_token_t arg2 ;
  const_char_string arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"ooo:Amp_set_conf self _hamlib_token_t_1 _const_char_string_2 ",(void *)0,(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_set_conf" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_set_conf" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(objv[3], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Amp_set_conf" "', argument " "3"" of type '" "const_char_string""'");
  }
  arg3 = (const_char_string)(buf3);
  {
    arg1->error_status = RIG_OK;
    Amp_set_conf__SWIG_1(arg1,arg2,(char const *)arg3);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_OK;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_set_conf(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Amp, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Amp_set_conf__SWIG_1(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Amp, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Amp_set_conf__SWIG_0(clientData, interp, objc, argv - 1);
        }
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Amp_set_conf'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Amp::set_conf(char const *,char const *)\n"
    "    Amp::set_conf(hamlib_token_t,const_char_string)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_get_conf__SWIG_0(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  hamlib_token_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Amp_get_conf self tok ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_get_conf" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  ecode2 = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(objv[2], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Amp_get_conf" "', argument " "2"" of type '" "hamlib_token_t""'");
  } 
  arg2 = (hamlib_token_t)(val2);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Amp_get_conf__SWIG_0(arg1,arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_get_conf__SWIG_1(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"oo:Amp_get_conf self name ",(void *)0,(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_get_conf" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(objv[2], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Amp_get_conf" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Amp_get_conf__SWIG_1(arg1,(char const *)arg2);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_OK;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_get_conf(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  Tcl_Obj *const *argv = objv+1;
  int argc = objc-1;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Amp, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long SWIG_TCL_CALL_ARGS_2(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Amp_get_conf__SWIG_0(clientData, interp, objc, argv - 1);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Amp, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Amp_get_conf__SWIG_1(clientData, interp, objc, argv - 1);
      }
    }
  }
  
  Tcl_SetResult(interp,(char *) "Wrong number or type of arguments for overloaded function 'Amp_get_conf'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Amp::get_conf(hamlib_token_t)\n"
    "    Amp::get_conf(char const *)\n", TCL_STATIC);
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_get_freq(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  freq_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Amp_get_freq self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_get_freq" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (freq_t)Amp_get_freq(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_double((double)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_get_info(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Amp_get_info self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_get_info" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (char *)Amp_get_info(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_FromCharPtr((const char *)result));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN int
_wrap_Amp_get_powerstat(ClientData clientData SWIGUNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  struct Amp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  powerstat_t result;
  
  if (SWIG_GetArgs(interp, objc, objv,"o:Amp_get_powerstat self ",(void *)0) == TCL_ERROR) SWIG_fail;
  res1 = SWIG_ConvertPtr(objv[1], &argp1,SWIGTYPE_p_Amp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Amp_get_powerstat" "', argument " "1"" of type '" "struct Amp *""'"); 
  }
  arg1 = (struct Amp *)(argp1);
  {
    arg1->error_status = RIG_OK;
    result = (powerstat_t)Amp_get_powerstat(arg1);
    if (arg1->error_status != RIG_OK && arg1->do_exception)
    SWIG_exception(SWIG_UnknownError, rigerror(arg1->error_status));
  }
  Tcl_SetObjResult(interp,SWIG_From_int((int)(result)));
  return TCL_OK;
fail:
  return TCL_ERROR;
}


SWIGINTERN void swig_delete_Amp(void *obj) {
struct Amp *arg1 = (struct Amp *) obj;
delete_Amp(arg1);
}
static swig_method swig_Amp_methods[] = {
    {"open", _wrap_Amp_open}, 
    {"close", _wrap_Amp_close}, 
    {"reset", _wrap_Amp_reset}, 
    {"token_lookup", _wrap_Amp_token_lookup}, 
    {"has_get_level", _wrap_Amp_has_get_level}, 
    {"has_set_level", _wrap_Amp_has_set_level}, 
    {"set_freq", _wrap_Amp_set_freq}, 
    {"set_powerstat", _wrap_Amp_set_powerstat}, 
    {"set_level", _wrap_Amp_set_level}, 
    {"set_ext_level", _wrap_Amp_set_ext_level}, 
    {"set_conf", _wrap_Amp_set_conf}, 
    {"get_conf", _wrap_Amp_get_conf}, 
    {"get_freq", _wrap_Amp_get_freq}, 
    {"get_info", _wrap_Amp_get_info}, 
    {"get_powerstat", _wrap_Amp_get_powerstat}, 
    {0,0}
};
static swig_attribute swig_Amp_attributes[] = {
    { "-caps",_wrap_Amp_caps_get, _wrap_Amp_caps_set},
    { "-state",_wrap_Amp_state_get, _wrap_Amp_state_set},
    { "-error_status",_wrap_Amp_error_status_get, _wrap_Amp_error_status_set},
    { "-do_exception",_wrap_Amp_do_exception_get, _wrap_Amp_do_exception_set},
    {0,0,0}
};
static swig_class *swig_Amp_bases[] = {0};
static const char * swig_Amp_base_names[] = {0};
static swig_class _wrap_class_Amp = { "Amp", &SWIGTYPE_p_Amp,_wrap_new_Amp, swig_delete_Amp, swig_Amp_methods, swig_Amp_attributes, swig_Amp_bases,swig_Amp_base_names, &swig_module, SWIG_TCL_HASHTABLE_INIT };

static swig_command_info swig_commands[] = {
    { SWIG_prefix "confparams_token_set", (swig_wrapper_func) _wrap_confparams_token_set, NULL},
    { SWIG_prefix "confparams_token_get", (swig_wrapper_func) _wrap_confparams_token_get, NULL},
    { SWIG_prefix "confparams_name_get", (swig_wrapper_func) _wrap_confparams_name_get, NULL},
    { SWIG_prefix "confparams_label_get", (swig_wrapper_func) _wrap_confparams_label_get, NULL},
    { SWIG_prefix "confparams_tooltip_get", (swig_wrapper_func) _wrap_confparams_tooltip_get, NULL},
    { SWIG_prefix "confparams_dflt_get", (swig_wrapper_func) _wrap_confparams_dflt_get, NULL},
    { SWIG_prefix "confparams_type_set", (swig_wrapper_func) _wrap_confparams_type_set, NULL},
    { SWIG_prefix "confparams_type_get", (swig_wrapper_func) _wrap_confparams_type_get, NULL},
    { SWIG_prefix "confparams_u_get", (swig_wrapper_func) _wrap_confparams_u_get, NULL},
    { SWIG_prefix "new_confparams", (swig_wrapper_func) _wrap_new_confparams, NULL},
    { SWIG_prefix "delete_confparams", (swig_wrapper_func) _wrap_delete_confparams, NULL},
    { SWIG_prefix "confparams", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_confparams},
    { SWIG_prefix "confparams_u_n_get", (swig_wrapper_func) _wrap_confparams_u_n_get, NULL},
    { SWIG_prefix "confparams_u_c_get", (swig_wrapper_func) _wrap_confparams_u_c_get, NULL},
    { SWIG_prefix "new_confparams_u", (swig_wrapper_func) _wrap_new_confparams_u, NULL},
    { SWIG_prefix "delete_confparams_u", (swig_wrapper_func) _wrap_delete_confparams_u, NULL},
    { SWIG_prefix "confparams_u", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_confparams_u},
    { SWIG_prefix "confparams_u_c_combostr_set", (swig_wrapper_func) _wrap_confparams_u_c_combostr_set, NULL},
    { SWIG_prefix "confparams_u_c_combostr_get", (swig_wrapper_func) _wrap_confparams_u_c_combostr_get, NULL},
    { SWIG_prefix "new_confparams_u_c", (swig_wrapper_func) _wrap_new_confparams_u_c, NULL},
    { SWIG_prefix "delete_confparams_u_c", (swig_wrapper_func) _wrap_delete_confparams_u_c, NULL},
    { SWIG_prefix "confparams_u_c", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_confparams_u_c},
    { SWIG_prefix "confparams_u_n_min_set", (swig_wrapper_func) _wrap_confparams_u_n_min_set, NULL},
    { SWIG_prefix "confparams_u_n_min_get", (swig_wrapper_func) _wrap_confparams_u_n_min_get, NULL},
    { SWIG_prefix "confparams_u_n_max_set", (swig_wrapper_func) _wrap_confparams_u_n_max_set, NULL},
    { SWIG_prefix "confparams_u_n_max_get", (swig_wrapper_func) _wrap_confparams_u_n_max_get, NULL},
    { SWIG_prefix "confparams_u_n_step_set", (swig_wrapper_func) _wrap_confparams_u_n_step_set, NULL},
    { SWIG_prefix "confparams_u_n_step_get", (swig_wrapper_func) _wrap_confparams_u_n_step_get, NULL},
    { SWIG_prefix "new_confparams_u_n", (swig_wrapper_func) _wrap_new_confparams_u_n, NULL},
    { SWIG_prefix "delete_confparams_u_n", (swig_wrapper_func) _wrap_delete_confparams_u_n, NULL},
    { SWIG_prefix "confparams_u_n", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_confparams_u_n},
    { SWIG_prefix "value_t_i_set", (swig_wrapper_func) _wrap_value_t_i_set, NULL},
    { SWIG_prefix "value_t_i_get", (swig_wrapper_func) _wrap_value_t_i_get, NULL},
    { SWIG_prefix "value_t_u_set", (swig_wrapper_func) _wrap_value_t_u_set, NULL},
    { SWIG_prefix "value_t_u_get", (swig_wrapper_func) _wrap_value_t_u_get, NULL},
    { SWIG_prefix "value_t_f_set", (swig_wrapper_func) _wrap_value_t_f_set, NULL},
    { SWIG_prefix "value_t_f_get", (swig_wrapper_func) _wrap_value_t_f_get, NULL},
    { SWIG_prefix "value_t_s_set", (swig_wrapper_func) _wrap_value_t_s_set, NULL},
    { SWIG_prefix "value_t_s_get", (swig_wrapper_func) _wrap_value_t_s_get, NULL},
    { SWIG_prefix "value_t_cs_get", (swig_wrapper_func) _wrap_value_t_cs_get, NULL},
    { SWIG_prefix "value_t_b_get", (swig_wrapper_func) _wrap_value_t_b_get, NULL},
    { SWIG_prefix "new_value_t", (swig_wrapper_func) _wrap_new_value_t, NULL},
    { SWIG_prefix "delete_value_t", (swig_wrapper_func) _wrap_delete_value_t, NULL},
    { SWIG_prefix "value_t", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_value_t},
    { SWIG_prefix "value_t_b_l_set", (swig_wrapper_func) _wrap_value_t_b_l_set, NULL},
    { SWIG_prefix "value_t_b_l_get", (swig_wrapper_func) _wrap_value_t_b_l_get, NULL},
    { SWIG_prefix "value_t_b_d_set", (swig_wrapper_func) _wrap_value_t_b_d_set, NULL},
    { SWIG_prefix "value_t_b_d_get", (swig_wrapper_func) _wrap_value_t_b_d_get, NULL},
    { SWIG_prefix "new_value_t_b", (swig_wrapper_func) _wrap_new_value_t_b, NULL},
    { SWIG_prefix "delete_value_t_b", (swig_wrapper_func) _wrap_delete_value_t_b, NULL},
    { SWIG_prefix "value_t_b", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_value_t_b},
    { SWIG_prefix "freq_range_t_startf_set", (swig_wrapper_func) _wrap_freq_range_t_startf_set, NULL},
    { SWIG_prefix "freq_range_t_startf_get", (swig_wrapper_func) _wrap_freq_range_t_startf_get, NULL},
    { SWIG_prefix "freq_range_t_endf_set", (swig_wrapper_func) _wrap_freq_range_t_endf_set, NULL},
    { SWIG_prefix "freq_range_t_endf_get", (swig_wrapper_func) _wrap_freq_range_t_endf_get, NULL},
    { SWIG_prefix "freq_range_t_modes_set", (swig_wrapper_func) _wrap_freq_range_t_modes_set, NULL},
    { SWIG_prefix "freq_range_t_modes_get", (swig_wrapper_func) _wrap_freq_range_t_modes_get, NULL},
    { SWIG_prefix "freq_range_t_low_power_set", (swig_wrapper_func) _wrap_freq_range_t_low_power_set, NULL},
    { SWIG_prefix "freq_range_t_low_power_get", (swig_wrapper_func) _wrap_freq_range_t_low_power_get, NULL},
    { SWIG_prefix "freq_range_t_high_power_set", (swig_wrapper_func) _wrap_freq_range_t_high_power_set, NULL},
    { SWIG_prefix "freq_range_t_high_power_get", (swig_wrapper_func) _wrap_freq_range_t_high_power_get, NULL},
    { SWIG_prefix "freq_range_t_vfo_set", (swig_wrapper_func) _wrap_freq_range_t_vfo_set, NULL},
    { SWIG_prefix "freq_range_t_vfo_get", (swig_wrapper_func) _wrap_freq_range_t_vfo_get, NULL},
    { SWIG_prefix "freq_range_t_ant_set", (swig_wrapper_func) _wrap_freq_range_t_ant_set, NULL},
    { SWIG_prefix "freq_range_t_ant_get", (swig_wrapper_func) _wrap_freq_range_t_ant_get, NULL},
    { SWIG_prefix "freq_range_t_label_set", (swig_wrapper_func) _wrap_freq_range_t_label_set, NULL},
    { SWIG_prefix "freq_range_t_label_get", (swig_wrapper_func) _wrap_freq_range_t_label_get, NULL},
    { SWIG_prefix "new_freq_range_t", (swig_wrapper_func) _wrap_new_freq_range_t, NULL},
    { SWIG_prefix "delete_freq_range_t", (swig_wrapper_func) _wrap_delete_freq_range_t, NULL},
    { SWIG_prefix "freq_range_t", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_freq_range_list},
    { SWIG_prefix "tuning_step_list_modes_set", (swig_wrapper_func) _wrap_tuning_step_list_modes_set, NULL},
    { SWIG_prefix "tuning_step_list_modes_get", (swig_wrapper_func) _wrap_tuning_step_list_modes_get, NULL},
    { SWIG_prefix "tuning_step_list_ts_set", (swig_wrapper_func) _wrap_tuning_step_list_ts_set, NULL},
    { SWIG_prefix "tuning_step_list_ts_get", (swig_wrapper_func) _wrap_tuning_step_list_ts_get, NULL},
    { SWIG_prefix "new_tuning_step_list", (swig_wrapper_func) _wrap_new_tuning_step_list, NULL},
    { SWIG_prefix "delete_tuning_step_list", (swig_wrapper_func) _wrap_delete_tuning_step_list, NULL},
    { SWIG_prefix "tuning_step_list", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_tuning_step_list},
    { SWIG_prefix "filter_list_modes_set", (swig_wrapper_func) _wrap_filter_list_modes_set, NULL},
    { SWIG_prefix "filter_list_modes_get", (swig_wrapper_func) _wrap_filter_list_modes_get, NULL},
    { SWIG_prefix "filter_list_width_set", (swig_wrapper_func) _wrap_filter_list_width_set, NULL},
    { SWIG_prefix "filter_list_width_get", (swig_wrapper_func) _wrap_filter_list_width_get, NULL},
    { SWIG_prefix "new_filter_list", (swig_wrapper_func) _wrap_new_filter_list, NULL},
    { SWIG_prefix "delete_filter_list", (swig_wrapper_func) _wrap_delete_filter_list, NULL},
    { SWIG_prefix "filter_list", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_filter_list},
    { SWIG_prefix "ext_list_token_set", (swig_wrapper_func) _wrap_ext_list_token_set, NULL},
    { SWIG_prefix "ext_list_token_get", (swig_wrapper_func) _wrap_ext_list_token_get, NULL},
    { SWIG_prefix "ext_list_val_set", (swig_wrapper_func) _wrap_ext_list_val_set, NULL},
    { SWIG_prefix "ext_list_val_get", (swig_wrapper_func) _wrap_ext_list_val_get, NULL},
    { SWIG_prefix "new_ext_list", (swig_wrapper_func) _wrap_new_ext_list, NULL},
    { SWIG_prefix "delete_ext_list", (swig_wrapper_func) _wrap_delete_ext_list, NULL},
    { SWIG_prefix "ext_list", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_ext_list},
    { SWIG_prefix "channel_channel_num_set", (swig_wrapper_func) _wrap_channel_channel_num_set, NULL},
    { SWIG_prefix "channel_channel_num_get", (swig_wrapper_func) _wrap_channel_channel_num_get, NULL},
    { SWIG_prefix "channel_bank_num_set", (swig_wrapper_func) _wrap_channel_bank_num_set, NULL},
    { SWIG_prefix "channel_bank_num_get", (swig_wrapper_func) _wrap_channel_bank_num_get, NULL},
    { SWIG_prefix "channel_vfo_set", (swig_wrapper_func) _wrap_channel_vfo_set, NULL},
    { SWIG_prefix "channel_vfo_get", (swig_wrapper_func) _wrap_channel_vfo_get, NULL},
    { SWIG_prefix "channel_ant_set", (swig_wrapper_func) _wrap_channel_ant_set, NULL},
    { SWIG_prefix "channel_ant_get", (swig_wrapper_func) _wrap_channel_ant_get, NULL},
    { SWIG_prefix "channel_freq_set", (swig_wrapper_func) _wrap_channel_freq_set, NULL},
    { SWIG_prefix "channel_freq_get", (swig_wrapper_func) _wrap_channel_freq_get, NULL},
    { SWIG_prefix "channel_mode_set", (swig_wrapper_func) _wrap_channel_mode_set, NULL},
    { SWIG_prefix "channel_mode_get", (swig_wrapper_func) _wrap_channel_mode_get, NULL},
    { SWIG_prefix "channel_width_set", (swig_wrapper_func) _wrap_channel_width_set, NULL},
    { SWIG_prefix "channel_width_get", (swig_wrapper_func) _wrap_channel_width_get, NULL},
    { SWIG_prefix "channel_tx_freq_set", (swig_wrapper_func) _wrap_channel_tx_freq_set, NULL},
    { SWIG_prefix "channel_tx_freq_get", (swig_wrapper_func) _wrap_channel_tx_freq_get, NULL},
    { SWIG_prefix "channel_tx_mode_set", (swig_wrapper_func) _wrap_channel_tx_mode_set, NULL},
    { SWIG_prefix "channel_tx_mode_get", (swig_wrapper_func) _wrap_channel_tx_mode_get, NULL},
    { SWIG_prefix "channel_tx_width_set", (swig_wrapper_func) _wrap_channel_tx_width_set, NULL},
    { SWIG_prefix "channel_tx_width_get", (swig_wrapper_func) _wrap_channel_tx_width_get, NULL},
    { SWIG_prefix "channel_split_set", (swig_wrapper_func) _wrap_channel_split_set, NULL},
    { SWIG_prefix "channel_split_get", (swig_wrapper_func) _wrap_channel_split_get, NULL},
    { SWIG_prefix "channel_tx_vfo_set", (swig_wrapper_func) _wrap_channel_tx_vfo_set, NULL},
    { SWIG_prefix "channel_tx_vfo_get", (swig_wrapper_func) _wrap_channel_tx_vfo_get, NULL},
    { SWIG_prefix "channel_rptr_shift_set", (swig_wrapper_func) _wrap_channel_rptr_shift_set, NULL},
    { SWIG_prefix "channel_rptr_shift_get", (swig_wrapper_func) _wrap_channel_rptr_shift_get, NULL},
    { SWIG_prefix "channel_rptr_offs_set", (swig_wrapper_func) _wrap_channel_rptr_offs_set, NULL},
    { SWIG_prefix "channel_rptr_offs_get", (swig_wrapper_func) _wrap_channel_rptr_offs_get, NULL},
    { SWIG_prefix "channel_tuning_step_set", (swig_wrapper_func) _wrap_channel_tuning_step_set, NULL},
    { SWIG_prefix "channel_tuning_step_get", (swig_wrapper_func) _wrap_channel_tuning_step_get, NULL},
    { SWIG_prefix "channel_rit_set", (swig_wrapper_func) _wrap_channel_rit_set, NULL},
    { SWIG_prefix "channel_rit_get", (swig_wrapper_func) _wrap_channel_rit_get, NULL},
    { SWIG_prefix "channel_xit_set", (swig_wrapper_func) _wrap_channel_xit_set, NULL},
    { SWIG_prefix "channel_xit_get", (swig_wrapper_func) _wrap_channel_xit_get, NULL},
    { SWIG_prefix "channel_funcs_set", (swig_wrapper_func) _wrap_channel_funcs_set, NULL},
    { SWIG_prefix "channel_funcs_get", (swig_wrapper_func) _wrap_channel_funcs_get, NULL},
    { SWIG_prefix "channel_levels_set", (swig_wrapper_func) _wrap_channel_levels_set, NULL},
    { SWIG_prefix "channel_levels_get", (swig_wrapper_func) _wrap_channel_levels_get, NULL},
    { SWIG_prefix "channel_ctcss_tone_set", (swig_wrapper_func) _wrap_channel_ctcss_tone_set, NULL},
    { SWIG_prefix "channel_ctcss_tone_get", (swig_wrapper_func) _wrap_channel_ctcss_tone_get, NULL},
    { SWIG_prefix "channel_ctcss_sql_set", (swig_wrapper_func) _wrap_channel_ctcss_sql_set, NULL},
    { SWIG_prefix "channel_ctcss_sql_get", (swig_wrapper_func) _wrap_channel_ctcss_sql_get, NULL},
    { SWIG_prefix "channel_dcs_code_set", (swig_wrapper_func) _wrap_channel_dcs_code_set, NULL},
    { SWIG_prefix "channel_dcs_code_get", (swig_wrapper_func) _wrap_channel_dcs_code_get, NULL},
    { SWIG_prefix "channel_dcs_sql_set", (swig_wrapper_func) _wrap_channel_dcs_sql_set, NULL},
    { SWIG_prefix "channel_dcs_sql_get", (swig_wrapper_func) _wrap_channel_dcs_sql_get, NULL},
    { SWIG_prefix "channel_scan_group_set", (swig_wrapper_func) _wrap_channel_scan_group_set, NULL},
    { SWIG_prefix "channel_scan_group_get", (swig_wrapper_func) _wrap_channel_scan_group_get, NULL},
    { SWIG_prefix "channel_flags_set", (swig_wrapper_func) _wrap_channel_flags_set, NULL},
    { SWIG_prefix "channel_flags_get", (swig_wrapper_func) _wrap_channel_flags_get, NULL},
    { SWIG_prefix "channel_channel_desc_set", (swig_wrapper_func) _wrap_channel_channel_desc_set, NULL},
    { SWIG_prefix "channel_channel_desc_get", (swig_wrapper_func) _wrap_channel_channel_desc_get, NULL},
    { SWIG_prefix "channel_ext_levels_set", (swig_wrapper_func) _wrap_channel_ext_levels_set, NULL},
    { SWIG_prefix "channel_ext_levels_get", (swig_wrapper_func) _wrap_channel_ext_levels_get, NULL},
    { SWIG_prefix "channel_tag_set", (swig_wrapper_func) _wrap_channel_tag_set, NULL},
    { SWIG_prefix "channel_tag_get", (swig_wrapper_func) _wrap_channel_tag_get, NULL},
    { SWIG_prefix "new_channel", (swig_wrapper_func) _wrap_new_channel, NULL},
    { SWIG_prefix "delete_channel", (swig_wrapper_func) _wrap_delete_channel, NULL},
    { SWIG_prefix "channel", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_channel},
    { SWIG_prefix "channel_cap_bank_num_set", (swig_wrapper_func) _wrap_channel_cap_bank_num_set, NULL},
    { SWIG_prefix "channel_cap_bank_num_get", (swig_wrapper_func) _wrap_channel_cap_bank_num_get, NULL},
    { SWIG_prefix "channel_cap_vfo_set", (swig_wrapper_func) _wrap_channel_cap_vfo_set, NULL},
    { SWIG_prefix "channel_cap_vfo_get", (swig_wrapper_func) _wrap_channel_cap_vfo_get, NULL},
    { SWIG_prefix "channel_cap_ant_set", (swig_wrapper_func) _wrap_channel_cap_ant_set, NULL},
    { SWIG_prefix "channel_cap_ant_get", (swig_wrapper_func) _wrap_channel_cap_ant_get, NULL},
    { SWIG_prefix "channel_cap_freq_set", (swig_wrapper_func) _wrap_channel_cap_freq_set, NULL},
    { SWIG_prefix "channel_cap_freq_get", (swig_wrapper_func) _wrap_channel_cap_freq_get, NULL},
    { SWIG_prefix "channel_cap_mode_set", (swig_wrapper_func) _wrap_channel_cap_mode_set, NULL},
    { SWIG_prefix "channel_cap_mode_get", (swig_wrapper_func) _wrap_channel_cap_mode_get, NULL},
    { SWIG_prefix "channel_cap_width_set", (swig_wrapper_func) _wrap_channel_cap_width_set, NULL},
    { SWIG_prefix "channel_cap_width_get", (swig_wrapper_func) _wrap_channel_cap_width_get, NULL},
    { SWIG_prefix "channel_cap_tx_freq_set", (swig_wrapper_func) _wrap_channel_cap_tx_freq_set, NULL},
    { SWIG_prefix "channel_cap_tx_freq_get", (swig_wrapper_func) _wrap_channel_cap_tx_freq_get, NULL},
    { SWIG_prefix "channel_cap_tx_mode_set", (swig_wrapper_func) _wrap_channel_cap_tx_mode_set, NULL},
    { SWIG_prefix "channel_cap_tx_mode_get", (swig_wrapper_func) _wrap_channel_cap_tx_mode_get, NULL},
    { SWIG_prefix "channel_cap_tx_width_set", (swig_wrapper_func) _wrap_channel_cap_tx_width_set, NULL},
    { SWIG_prefix "channel_cap_tx_width_get", (swig_wrapper_func) _wrap_channel_cap_tx_width_get, NULL},
    { SWIG_prefix "channel_cap_split_set", (swig_wrapper_func) _wrap_channel_cap_split_set, NULL},
    { SWIG_prefix "channel_cap_split_get", (swig_wrapper_func) _wrap_channel_cap_split_get, NULL},
    { SWIG_prefix "channel_cap_tx_vfo_set", (swig_wrapper_func) _wrap_channel_cap_tx_vfo_set, NULL},
    { SWIG_prefix "channel_cap_tx_vfo_get", (swig_wrapper_func) _wrap_channel_cap_tx_vfo_get, NULL},
    { SWIG_prefix "channel_cap_rptr_shift_set", (swig_wrapper_func) _wrap_channel_cap_rptr_shift_set, NULL},
    { SWIG_prefix "channel_cap_rptr_shift_get", (swig_wrapper_func) _wrap_channel_cap_rptr_shift_get, NULL},
    { SWIG_prefix "channel_cap_rptr_offs_set", (swig_wrapper_func) _wrap_channel_cap_rptr_offs_set, NULL},
    { SWIG_prefix "channel_cap_rptr_offs_get", (swig_wrapper_func) _wrap_channel_cap_rptr_offs_get, NULL},
    { SWIG_prefix "channel_cap_tuning_step_set", (swig_wrapper_func) _wrap_channel_cap_tuning_step_set, NULL},
    { SWIG_prefix "channel_cap_tuning_step_get", (swig_wrapper_func) _wrap_channel_cap_tuning_step_get, NULL},
    { SWIG_prefix "channel_cap_rit_set", (swig_wrapper_func) _wrap_channel_cap_rit_set, NULL},
    { SWIG_prefix "channel_cap_rit_get", (swig_wrapper_func) _wrap_channel_cap_rit_get, NULL},
    { SWIG_prefix "channel_cap_xit_set", (swig_wrapper_func) _wrap_channel_cap_xit_set, NULL},
    { SWIG_prefix "channel_cap_xit_get", (swig_wrapper_func) _wrap_channel_cap_xit_get, NULL},
    { SWIG_prefix "channel_cap_funcs_set", (swig_wrapper_func) _wrap_channel_cap_funcs_set, NULL},
    { SWIG_prefix "channel_cap_funcs_get", (swig_wrapper_func) _wrap_channel_cap_funcs_get, NULL},
    { SWIG_prefix "channel_cap_levels_set", (swig_wrapper_func) _wrap_channel_cap_levels_set, NULL},
    { SWIG_prefix "channel_cap_levels_get", (swig_wrapper_func) _wrap_channel_cap_levels_get, NULL},
    { SWIG_prefix "channel_cap_ctcss_tone_set", (swig_wrapper_func) _wrap_channel_cap_ctcss_tone_set, NULL},
    { SWIG_prefix "channel_cap_ctcss_tone_get", (swig_wrapper_func) _wrap_channel_cap_ctcss_tone_get, NULL},
    { SWIG_prefix "channel_cap_ctcss_sql_set", (swig_wrapper_func) _wrap_channel_cap_ctcss_sql_set, NULL},
    { SWIG_prefix "channel_cap_ctcss_sql_get", (swig_wrapper_func) _wrap_channel_cap_ctcss_sql_get, NULL},
    { SWIG_prefix "channel_cap_dcs_code_set", (swig_wrapper_func) _wrap_channel_cap_dcs_code_set, NULL},
    { SWIG_prefix "channel_cap_dcs_code_get", (swig_wrapper_func) _wrap_channel_cap_dcs_code_get, NULL},
    { SWIG_prefix "channel_cap_dcs_sql_set", (swig_wrapper_func) _wrap_channel_cap_dcs_sql_set, NULL},
    { SWIG_prefix "channel_cap_dcs_sql_get", (swig_wrapper_func) _wrap_channel_cap_dcs_sql_get, NULL},
    { SWIG_prefix "channel_cap_scan_group_set", (swig_wrapper_func) _wrap_channel_cap_scan_group_set, NULL},
    { SWIG_prefix "channel_cap_scan_group_get", (swig_wrapper_func) _wrap_channel_cap_scan_group_get, NULL},
    { SWIG_prefix "channel_cap_flags_set", (swig_wrapper_func) _wrap_channel_cap_flags_set, NULL},
    { SWIG_prefix "channel_cap_flags_get", (swig_wrapper_func) _wrap_channel_cap_flags_get, NULL},
    { SWIG_prefix "channel_cap_channel_desc_set", (swig_wrapper_func) _wrap_channel_cap_channel_desc_set, NULL},
    { SWIG_prefix "channel_cap_channel_desc_get", (swig_wrapper_func) _wrap_channel_cap_channel_desc_get, NULL},
    { SWIG_prefix "channel_cap_ext_levels_set", (swig_wrapper_func) _wrap_channel_cap_ext_levels_set, NULL},
    { SWIG_prefix "channel_cap_ext_levels_get", (swig_wrapper_func) _wrap_channel_cap_ext_levels_get, NULL},
    { SWIG_prefix "channel_cap_tag_set", (swig_wrapper_func) _wrap_channel_cap_tag_set, NULL},
    { SWIG_prefix "channel_cap_tag_get", (swig_wrapper_func) _wrap_channel_cap_tag_get, NULL},
    { SWIG_prefix "new_channel_cap", (swig_wrapper_func) _wrap_new_channel_cap, NULL},
    { SWIG_prefix "delete_channel_cap", (swig_wrapper_func) _wrap_delete_channel_cap, NULL},
    { SWIG_prefix "channel_cap", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_channel_cap},
    { SWIG_prefix "chan_list_startc_set", (swig_wrapper_func) _wrap_chan_list_startc_set, NULL},
    { SWIG_prefix "chan_list_startc_get", (swig_wrapper_func) _wrap_chan_list_startc_get, NULL},
    { SWIG_prefix "chan_list_endc_set", (swig_wrapper_func) _wrap_chan_list_endc_set, NULL},
    { SWIG_prefix "chan_list_endc_get", (swig_wrapper_func) _wrap_chan_list_endc_get, NULL},
    { SWIG_prefix "chan_list_type_set", (swig_wrapper_func) _wrap_chan_list_type_set, NULL},
    { SWIG_prefix "chan_list_type_get", (swig_wrapper_func) _wrap_chan_list_type_get, NULL},
    { SWIG_prefix "chan_list_mem_caps_set", (swig_wrapper_func) _wrap_chan_list_mem_caps_set, NULL},
    { SWIG_prefix "chan_list_mem_caps_get", (swig_wrapper_func) _wrap_chan_list_mem_caps_get, NULL},
    { SWIG_prefix "new_chan_list", (swig_wrapper_func) _wrap_new_chan_list, NULL},
    { SWIG_prefix "delete_chan_list", (swig_wrapper_func) _wrap_delete_chan_list, NULL},
    { SWIG_prefix "chan_list", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_chan_list},
    { SWIG_prefix "gran_min_set", (swig_wrapper_func) _wrap_gran_min_set, NULL},
    { SWIG_prefix "gran_min_get", (swig_wrapper_func) _wrap_gran_min_get, NULL},
    { SWIG_prefix "gran_max_set", (swig_wrapper_func) _wrap_gran_max_set, NULL},
    { SWIG_prefix "gran_max_get", (swig_wrapper_func) _wrap_gran_max_get, NULL},
    { SWIG_prefix "gran_step_set", (swig_wrapper_func) _wrap_gran_step_set, NULL},
    { SWIG_prefix "gran_step_get", (swig_wrapper_func) _wrap_gran_step_get, NULL},
    { SWIG_prefix "new_gran", (swig_wrapper_func) _wrap_new_gran, NULL},
    { SWIG_prefix "delete_gran", (swig_wrapper_func) _wrap_delete_gran, NULL},
    { SWIG_prefix "gran", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_gran},
    { SWIG_prefix "cal_table_size_set", (swig_wrapper_func) _wrap_cal_table_size_set, NULL},
    { SWIG_prefix "cal_table_size_get", (swig_wrapper_func) _wrap_cal_table_size_get, NULL},
    { SWIG_prefix "cal_table_table_get", (swig_wrapper_func) _wrap_cal_table_table_get, NULL},
    { SWIG_prefix "new_cal_table", (swig_wrapper_func) _wrap_new_cal_table, NULL},
    { SWIG_prefix "delete_cal_table", (swig_wrapper_func) _wrap_delete_cal_table, NULL},
    { SWIG_prefix "cal_table", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_cal_table},
    { SWIG_prefix "cal_table_table_raw_set", (swig_wrapper_func) _wrap_cal_table_table_raw_set, NULL},
    { SWIG_prefix "cal_table_table_raw_get", (swig_wrapper_func) _wrap_cal_table_table_raw_get, NULL},
    { SWIG_prefix "cal_table_table_val_set", (swig_wrapper_func) _wrap_cal_table_table_val_set, NULL},
    { SWIG_prefix "cal_table_table_val_get", (swig_wrapper_func) _wrap_cal_table_table_val_get, NULL},
    { SWIG_prefix "new_cal_table_table", (swig_wrapper_func) _wrap_new_cal_table_table, NULL},
    { SWIG_prefix "delete_cal_table_table", (swig_wrapper_func) _wrap_delete_cal_table_table, NULL},
    { SWIG_prefix "cal_table_table", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_cal_table_table},
    { SWIG_prefix "cal_table_float_size_set", (swig_wrapper_func) _wrap_cal_table_float_size_set, NULL},
    { SWIG_prefix "cal_table_float_size_get", (swig_wrapper_func) _wrap_cal_table_float_size_get, NULL},
    { SWIG_prefix "cal_table_float_table_get", (swig_wrapper_func) _wrap_cal_table_float_table_get, NULL},
    { SWIG_prefix "new_cal_table_float", (swig_wrapper_func) _wrap_new_cal_table_float, NULL},
    { SWIG_prefix "delete_cal_table_float", (swig_wrapper_func) _wrap_delete_cal_table_float, NULL},
    { SWIG_prefix "cal_table_float", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_cal_table_float},
    { SWIG_prefix "cal_table_float_table_raw_set", (swig_wrapper_func) _wrap_cal_table_float_table_raw_set, NULL},
    { SWIG_prefix "cal_table_float_table_raw_get", (swig_wrapper_func) _wrap_cal_table_float_table_raw_get, NULL},
    { SWIG_prefix "cal_table_float_table_val_set", (swig_wrapper_func) _wrap_cal_table_float_table_val_set, NULL},
    { SWIG_prefix "cal_table_float_table_val_get", (swig_wrapper_func) _wrap_cal_table_float_table_val_get, NULL},
    { SWIG_prefix "new_cal_table_float_table", (swig_wrapper_func) _wrap_new_cal_table_float_table, NULL},
    { SWIG_prefix "delete_cal_table_float_table", (swig_wrapper_func) _wrap_delete_cal_table_float_table, NULL},
    { SWIG_prefix "cal_table_float_table", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_cal_table_float_table},
    { SWIG_prefix "rig_spectrum_scope_id_set", (swig_wrapper_func) _wrap_rig_spectrum_scope_id_set, NULL},
    { SWIG_prefix "rig_spectrum_scope_id_get", (swig_wrapper_func) _wrap_rig_spectrum_scope_id_get, NULL},
    { SWIG_prefix "rig_spectrum_scope_name_set", (swig_wrapper_func) _wrap_rig_spectrum_scope_name_set, NULL},
    { SWIG_prefix "rig_spectrum_scope_name_get", (swig_wrapper_func) _wrap_rig_spectrum_scope_name_get, NULL},
    { SWIG_prefix "new_rig_spectrum_scope", (swig_wrapper_func) _wrap_new_rig_spectrum_scope, NULL},
    { SWIG_prefix "delete_rig_spectrum_scope", (swig_wrapper_func) _wrap_delete_rig_spectrum_scope, NULL},
    { SWIG_prefix "rig_spectrum_scope", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_rig_spectrum_scope},
    { SWIG_prefix "rig_spectrum_avg_mode_id_set", (swig_wrapper_func) _wrap_rig_spectrum_avg_mode_id_set, NULL},
    { SWIG_prefix "rig_spectrum_avg_mode_id_get", (swig_wrapper_func) _wrap_rig_spectrum_avg_mode_id_get, NULL},
    { SWIG_prefix "rig_spectrum_avg_mode_name_set", (swig_wrapper_func) _wrap_rig_spectrum_avg_mode_name_set, NULL},
    { SWIG_prefix "rig_spectrum_avg_mode_name_get", (swig_wrapper_func) _wrap_rig_spectrum_avg_mode_name_get, NULL},
    { SWIG_prefix "new_rig_spectrum_avg_mode", (swig_wrapper_func) _wrap_new_rig_spectrum_avg_mode, NULL},
    { SWIG_prefix "delete_rig_spectrum_avg_mode", (swig_wrapper_func) _wrap_delete_rig_spectrum_avg_mode, NULL},
    { SWIG_prefix "rig_spectrum_avg_mode", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_rig_spectrum_avg_mode},
    { SWIG_prefix "rig_spectrum_line_id_set", (swig_wrapper_func) _wrap_rig_spectrum_line_id_set, NULL},
    { SWIG_prefix "rig_spectrum_line_id_get", (swig_wrapper_func) _wrap_rig_spectrum_line_id_get, NULL},
    { SWIG_prefix "rig_spectrum_line_data_level_min_set", (swig_wrapper_func) _wrap_rig_spectrum_line_data_level_min_set, NULL},
    { SWIG_prefix "rig_spectrum_line_data_level_min_get", (swig_wrapper_func) _wrap_rig_spectrum_line_data_level_min_get, NULL},
    { SWIG_prefix "rig_spectrum_line_data_level_max_set", (swig_wrapper_func) _wrap_rig_spectrum_line_data_level_max_set, NULL},
    { SWIG_prefix "rig_spectrum_line_data_level_max_get", (swig_wrapper_func) _wrap_rig_spectrum_line_data_level_max_get, NULL},
    { SWIG_prefix "rig_spectrum_line_signal_strength_min_set", (swig_wrapper_func) _wrap_rig_spectrum_line_signal_strength_min_set, NULL},
    { SWIG_prefix "rig_spectrum_line_signal_strength_min_get", (swig_wrapper_func) _wrap_rig_spectrum_line_signal_strength_min_get, NULL},
    { SWIG_prefix "rig_spectrum_line_signal_strength_max_set", (swig_wrapper_func) _wrap_rig_spectrum_line_signal_strength_max_set, NULL},
    { SWIG_prefix "rig_spectrum_line_signal_strength_max_get", (swig_wrapper_func) _wrap_rig_spectrum_line_signal_strength_max_get, NULL},
    { SWIG_prefix "rig_spectrum_line_spectrum_mode_set", (swig_wrapper_func) _wrap_rig_spectrum_line_spectrum_mode_set, NULL},
    { SWIG_prefix "rig_spectrum_line_spectrum_mode_get", (swig_wrapper_func) _wrap_rig_spectrum_line_spectrum_mode_get, NULL},
    { SWIG_prefix "rig_spectrum_line_center_freq_set", (swig_wrapper_func) _wrap_rig_spectrum_line_center_freq_set, NULL},
    { SWIG_prefix "rig_spectrum_line_center_freq_get", (swig_wrapper_func) _wrap_rig_spectrum_line_center_freq_get, NULL},
    { SWIG_prefix "rig_spectrum_line_span_freq_set", (swig_wrapper_func) _wrap_rig_spectrum_line_span_freq_set, NULL},
    { SWIG_prefix "rig_spectrum_line_span_freq_get", (swig_wrapper_func) _wrap_rig_spectrum_line_span_freq_get, NULL},
    { SWIG_prefix "rig_spectrum_line_low_edge_freq_set", (swig_wrapper_func) _wrap_rig_spectrum_line_low_edge_freq_set, NULL},
    { SWIG_prefix "rig_spectrum_line_low_edge_freq_get", (swig_wrapper_func) _wrap_rig_spectrum_line_low_edge_freq_get, NULL},
    { SWIG_prefix "rig_spectrum_line_high_edge_freq_set", (swig_wrapper_func) _wrap_rig_spectrum_line_high_edge_freq_set, NULL},
    { SWIG_prefix "rig_spectrum_line_high_edge_freq_get", (swig_wrapper_func) _wrap_rig_spectrum_line_high_edge_freq_get, NULL},
    { SWIG_prefix "rig_spectrum_line_spectrum_data_length_set", (swig_wrapper_func) _wrap_rig_spectrum_line_spectrum_data_length_set, NULL},
    { SWIG_prefix "rig_spectrum_line_spectrum_data_length_get", (swig_wrapper_func) _wrap_rig_spectrum_line_spectrum_data_length_get, NULL},
    { SWIG_prefix "rig_spectrum_line_spectrum_data_set", (swig_wrapper_func) _wrap_rig_spectrum_line_spectrum_data_set, NULL},
    { SWIG_prefix "rig_spectrum_line_spectrum_data_get", (swig_wrapper_func) _wrap_rig_spectrum_line_spectrum_data_get, NULL},
    { SWIG_prefix "new_rig_spectrum_line", (swig_wrapper_func) _wrap_new_rig_spectrum_line, NULL},
    { SWIG_prefix "delete_rig_spectrum_line", (swig_wrapper_func) _wrap_delete_rig_spectrum_line, NULL},
    { SWIG_prefix "rig_spectrum_line", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_rig_spectrum_line},
    { SWIG_prefix "deferred_config_item_nextt_set", (swig_wrapper_func) _wrap_deferred_config_item_nextt_set, NULL},
    { SWIG_prefix "deferred_config_item_nextt_get", (swig_wrapper_func) _wrap_deferred_config_item_nextt_get, NULL},
    { SWIG_prefix "deferred_config_item_token_set", (swig_wrapper_func) _wrap_deferred_config_item_token_set, NULL},
    { SWIG_prefix "deferred_config_item_token_get", (swig_wrapper_func) _wrap_deferred_config_item_token_get, NULL},
    { SWIG_prefix "deferred_config_item_value_set", (swig_wrapper_func) _wrap_deferred_config_item_value_set, NULL},
    { SWIG_prefix "deferred_config_item_value_get", (swig_wrapper_func) _wrap_deferred_config_item_value_get, NULL},
    { SWIG_prefix "new_deferred_config_item", (swig_wrapper_func) _wrap_new_deferred_config_item, NULL},
    { SWIG_prefix "delete_deferred_config_item", (swig_wrapper_func) _wrap_delete_deferred_config_item, NULL},
    { SWIG_prefix "deferred_config_item", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_deferred_config_item},
    { SWIG_prefix "deferred_config_header_firstt_set", (swig_wrapper_func) _wrap_deferred_config_header_firstt_set, NULL},
    { SWIG_prefix "deferred_config_header_firstt_get", (swig_wrapper_func) _wrap_deferred_config_header_firstt_get, NULL},
    { SWIG_prefix "deferred_config_header_lastt_set", (swig_wrapper_func) _wrap_deferred_config_header_lastt_set, NULL},
    { SWIG_prefix "deferred_config_header_lastt_get", (swig_wrapper_func) _wrap_deferred_config_header_lastt_get, NULL},
    { SWIG_prefix "new_deferred_config_header", (swig_wrapper_func) _wrap_new_deferred_config_header, NULL},
    { SWIG_prefix "delete_deferred_config_header", (swig_wrapper_func) _wrap_delete_deferred_config_header, NULL},
    { SWIG_prefix "deferred_config_header", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_deferred_config_header},
    { SWIG_prefix "rig_caps_rig_model_set", (swig_wrapper_func) _wrap_rig_caps_rig_model_set, NULL},
    { SWIG_prefix "rig_caps_rig_model_get", (swig_wrapper_func) _wrap_rig_caps_rig_model_get, NULL},
    { SWIG_prefix "rig_caps_model_name_get", (swig_wrapper_func) _wrap_rig_caps_model_name_get, NULL},
    { SWIG_prefix "rig_caps_mfg_name_get", (swig_wrapper_func) _wrap_rig_caps_mfg_name_get, NULL},
    { SWIG_prefix "rig_caps_version_get", (swig_wrapper_func) _wrap_rig_caps_version_get, NULL},
    { SWIG_prefix "rig_caps_copyright_get", (swig_wrapper_func) _wrap_rig_caps_copyright_get, NULL},
    { SWIG_prefix "rig_caps_status_set", (swig_wrapper_func) _wrap_rig_caps_status_set, NULL},
    { SWIG_prefix "rig_caps_status_get", (swig_wrapper_func) _wrap_rig_caps_status_get, NULL},
    { SWIG_prefix "rig_caps_rig_type_set", (swig_wrapper_func) _wrap_rig_caps_rig_type_set, NULL},
    { SWIG_prefix "rig_caps_rig_type_get", (swig_wrapper_func) _wrap_rig_caps_rig_type_get, NULL},
    { SWIG_prefix "rig_caps_ptt_type_set", (swig_wrapper_func) _wrap_rig_caps_ptt_type_set, NULL},
    { SWIG_prefix "rig_caps_ptt_type_get", (swig_wrapper_func) _wrap_rig_caps_ptt_type_get, NULL},
    { SWIG_prefix "rig_caps_dcd_type_set", (swig_wrapper_func) _wrap_rig_caps_dcd_type_set, NULL},
    { SWIG_prefix "rig_caps_dcd_type_get", (swig_wrapper_func) _wrap_rig_caps_dcd_type_get, NULL},
    { SWIG_prefix "rig_caps_port_type_set", (swig_wrapper_func) _wrap_rig_caps_port_type_set, NULL},
    { SWIG_prefix "rig_caps_port_type_get", (swig_wrapper_func) _wrap_rig_caps_port_type_get, NULL},
    { SWIG_prefix "rig_caps_serial_rate_min_set", (swig_wrapper_func) _wrap_rig_caps_serial_rate_min_set, NULL},
    { SWIG_prefix "rig_caps_serial_rate_min_get", (swig_wrapper_func) _wrap_rig_caps_serial_rate_min_get, NULL},
    { SWIG_prefix "rig_caps_serial_rate_max_set", (swig_wrapper_func) _wrap_rig_caps_serial_rate_max_set, NULL},
    { SWIG_prefix "rig_caps_serial_rate_max_get", (swig_wrapper_func) _wrap_rig_caps_serial_rate_max_get, NULL},
    { SWIG_prefix "rig_caps_serial_data_bits_set", (swig_wrapper_func) _wrap_rig_caps_serial_data_bits_set, NULL},
    { SWIG_prefix "rig_caps_serial_data_bits_get", (swig_wrapper_func) _wrap_rig_caps_serial_data_bits_get, NULL},
    { SWIG_prefix "rig_caps_serial_stop_bits_set", (swig_wrapper_func) _wrap_rig_caps_serial_stop_bits_set, NULL},
    { SWIG_prefix "rig_caps_serial_stop_bits_get", (swig_wrapper_func) _wrap_rig_caps_serial_stop_bits_get, NULL},
    { SWIG_prefix "rig_caps_serial_parity_set", (swig_wrapper_func) _wrap_rig_caps_serial_parity_set, NULL},
    { SWIG_prefix "rig_caps_serial_parity_get", (swig_wrapper_func) _wrap_rig_caps_serial_parity_get, NULL},
    { SWIG_prefix "rig_caps_serial_handshake_set", (swig_wrapper_func) _wrap_rig_caps_serial_handshake_set, NULL},
    { SWIG_prefix "rig_caps_serial_handshake_get", (swig_wrapper_func) _wrap_rig_caps_serial_handshake_get, NULL},
    { SWIG_prefix "rig_caps_write_delay_set", (swig_wrapper_func) _wrap_rig_caps_write_delay_set, NULL},
    { SWIG_prefix "rig_caps_write_delay_get", (swig_wrapper_func) _wrap_rig_caps_write_delay_get, NULL},
    { SWIG_prefix "rig_caps_post_write_delay_set", (swig_wrapper_func) _wrap_rig_caps_post_write_delay_set, NULL},
    { SWIG_prefix "rig_caps_post_write_delay_get", (swig_wrapper_func) _wrap_rig_caps_post_write_delay_get, NULL},
    { SWIG_prefix "rig_caps_timeout_set", (swig_wrapper_func) _wrap_rig_caps_timeout_set, NULL},
    { SWIG_prefix "rig_caps_timeout_get", (swig_wrapper_func) _wrap_rig_caps_timeout_get, NULL},
    { SWIG_prefix "rig_caps_retry_set", (swig_wrapper_func) _wrap_rig_caps_retry_set, NULL},
    { SWIG_prefix "rig_caps_retry_get", (swig_wrapper_func) _wrap_rig_caps_retry_get, NULL},
    { SWIG_prefix "rig_caps_has_get_func_set", (swig_wrapper_func) _wrap_rig_caps_has_get_func_set, NULL},
    { SWIG_prefix "rig_caps_has_get_func_get", (swig_wrapper_func) _wrap_rig_caps_has_get_func_get, NULL},
    { SWIG_prefix "rig_caps_has_set_func_set", (swig_wrapper_func) _wrap_rig_caps_has_set_func_set, NULL},
    { SWIG_prefix "rig_caps_has_set_func_get", (swig_wrapper_func) _wrap_rig_caps_has_set_func_get, NULL},
    { SWIG_prefix "rig_caps_has_get_level_set", (swig_wrapper_func) _wrap_rig_caps_has_get_level_set, NULL},
    { SWIG_prefix "rig_caps_has_get_level_get", (swig_wrapper_func) _wrap_rig_caps_has_get_level_get, NULL},
    { SWIG_prefix "rig_caps_has_set_level_set", (swig_wrapper_func) _wrap_rig_caps_has_set_level_set, NULL},
    { SWIG_prefix "rig_caps_has_set_level_get", (swig_wrapper_func) _wrap_rig_caps_has_set_level_get, NULL},
    { SWIG_prefix "rig_caps_has_get_parm_set", (swig_wrapper_func) _wrap_rig_caps_has_get_parm_set, NULL},
    { SWIG_prefix "rig_caps_has_get_parm_get", (swig_wrapper_func) _wrap_rig_caps_has_get_parm_get, NULL},
    { SWIG_prefix "rig_caps_has_set_parm_set", (swig_wrapper_func) _wrap_rig_caps_has_set_parm_set, NULL},
    { SWIG_prefix "rig_caps_has_set_parm_get", (swig_wrapper_func) _wrap_rig_caps_has_set_parm_get, NULL},
    { SWIG_prefix "rig_caps_level_gran_set", (swig_wrapper_func) _wrap_rig_caps_level_gran_set, NULL},
    { SWIG_prefix "rig_caps_level_gran_get", (swig_wrapper_func) _wrap_rig_caps_level_gran_get, NULL},
    { SWIG_prefix "rig_caps_parm_gran_set", (swig_wrapper_func) _wrap_rig_caps_parm_gran_set, NULL},
    { SWIG_prefix "rig_caps_parm_gran_get", (swig_wrapper_func) _wrap_rig_caps_parm_gran_get, NULL},
    { SWIG_prefix "rig_caps_extparms_set", (swig_wrapper_func) _wrap_rig_caps_extparms_set, NULL},
    { SWIG_prefix "rig_caps_extparms_get", (swig_wrapper_func) _wrap_rig_caps_extparms_get, NULL},
    { SWIG_prefix "rig_caps_extlevels_set", (swig_wrapper_func) _wrap_rig_caps_extlevels_set, NULL},
    { SWIG_prefix "rig_caps_extlevels_get", (swig_wrapper_func) _wrap_rig_caps_extlevels_get, NULL},
    { SWIG_prefix "rig_caps_extfuncs_set", (swig_wrapper_func) _wrap_rig_caps_extfuncs_set, NULL},
    { SWIG_prefix "rig_caps_extfuncs_get", (swig_wrapper_func) _wrap_rig_caps_extfuncs_get, NULL},
    { SWIG_prefix "rig_caps_ext_tokens_set", (swig_wrapper_func) _wrap_rig_caps_ext_tokens_set, NULL},
    { SWIG_prefix "rig_caps_ext_tokens_get", (swig_wrapper_func) _wrap_rig_caps_ext_tokens_get, NULL},
    { SWIG_prefix "rig_caps_ctcss_list_set", (swig_wrapper_func) _wrap_rig_caps_ctcss_list_set, NULL},
    { SWIG_prefix "rig_caps_ctcss_list_get", (swig_wrapper_func) _wrap_rig_caps_ctcss_list_get, NULL},
    { SWIG_prefix "rig_caps_dcs_list_set", (swig_wrapper_func) _wrap_rig_caps_dcs_list_set, NULL},
    { SWIG_prefix "rig_caps_dcs_list_get", (swig_wrapper_func) _wrap_rig_caps_dcs_list_get, NULL},
    { SWIG_prefix "rig_caps_preamp_set", (swig_wrapper_func) _wrap_rig_caps_preamp_set, NULL},
    { SWIG_prefix "rig_caps_preamp_get", (swig_wrapper_func) _wrap_rig_caps_preamp_get, NULL},
    { SWIG_prefix "rig_caps_attenuator_set", (swig_wrapper_func) _wrap_rig_caps_attenuator_set, NULL},
    { SWIG_prefix "rig_caps_attenuator_get", (swig_wrapper_func) _wrap_rig_caps_attenuator_get, NULL},
    { SWIG_prefix "rig_caps_max_rit_set", (swig_wrapper_func) _wrap_rig_caps_max_rit_set, NULL},
    { SWIG_prefix "rig_caps_max_rit_get", (swig_wrapper_func) _wrap_rig_caps_max_rit_get, NULL},
    { SWIG_prefix "rig_caps_max_xit_set", (swig_wrapper_func) _wrap_rig_caps_max_xit_set, NULL},
    { SWIG_prefix "rig_caps_max_xit_get", (swig_wrapper_func) _wrap_rig_caps_max_xit_get, NULL},
    { SWIG_prefix "rig_caps_max_ifshift_set", (swig_wrapper_func) _wrap_rig_caps_max_ifshift_set, NULL},
    { SWIG_prefix "rig_caps_max_ifshift_get", (swig_wrapper_func) _wrap_rig_caps_max_ifshift_get, NULL},
    { SWIG_prefix "rig_caps_agc_level_count_set", (swig_wrapper_func) _wrap_rig_caps_agc_level_count_set, NULL},
    { SWIG_prefix "rig_caps_agc_level_count_get", (swig_wrapper_func) _wrap_rig_caps_agc_level_count_get, NULL},
    { SWIG_prefix "rig_caps_agc_levels_set", (swig_wrapper_func) _wrap_rig_caps_agc_levels_set, NULL},
    { SWIG_prefix "rig_caps_agc_levels_get", (swig_wrapper_func) _wrap_rig_caps_agc_levels_get, NULL},
    { SWIG_prefix "rig_caps_announces_set", (swig_wrapper_func) _wrap_rig_caps_announces_set, NULL},
    { SWIG_prefix "rig_caps_announces_get", (swig_wrapper_func) _wrap_rig_caps_announces_get, NULL},
    { SWIG_prefix "rig_caps_vfo_ops_set", (swig_wrapper_func) _wrap_rig_caps_vfo_ops_set, NULL},
    { SWIG_prefix "rig_caps_vfo_ops_get", (swig_wrapper_func) _wrap_rig_caps_vfo_ops_get, NULL},
    { SWIG_prefix "rig_caps_scan_ops_set", (swig_wrapper_func) _wrap_rig_caps_scan_ops_set, NULL},
    { SWIG_prefix "rig_caps_scan_ops_get", (swig_wrapper_func) _wrap_rig_caps_scan_ops_get, NULL},
    { SWIG_prefix "rig_caps_targetable_vfo_set", (swig_wrapper_func) _wrap_rig_caps_targetable_vfo_set, NULL},
    { SWIG_prefix "rig_caps_targetable_vfo_get", (swig_wrapper_func) _wrap_rig_caps_targetable_vfo_get, NULL},
    { SWIG_prefix "rig_caps_transceive_set", (swig_wrapper_func) _wrap_rig_caps_transceive_set, NULL},
    { SWIG_prefix "rig_caps_transceive_get", (swig_wrapper_func) _wrap_rig_caps_transceive_get, NULL},
    { SWIG_prefix "rig_caps_bank_qty_set", (swig_wrapper_func) _wrap_rig_caps_bank_qty_set, NULL},
    { SWIG_prefix "rig_caps_bank_qty_get", (swig_wrapper_func) _wrap_rig_caps_bank_qty_get, NULL},
    { SWIG_prefix "rig_caps_chan_desc_sz_set", (swig_wrapper_func) _wrap_rig_caps_chan_desc_sz_set, NULL},
    { SWIG_prefix "rig_caps_chan_desc_sz_get", (swig_wrapper_func) _wrap_rig_caps_chan_desc_sz_get, NULL},
    { SWIG_prefix "rig_caps_chan_list_set", (swig_wrapper_func) _wrap_rig_caps_chan_list_set, NULL},
    { SWIG_prefix "rig_caps_chan_list_get", (swig_wrapper_func) _wrap_rig_caps_chan_list_get, NULL},
    { SWIG_prefix "rig_caps_rx_range_list1_set", (swig_wrapper_func) _wrap_rig_caps_rx_range_list1_set, NULL},
    { SWIG_prefix "rig_caps_rx_range_list1_get", (swig_wrapper_func) _wrap_rig_caps_rx_range_list1_get, NULL},
    { SWIG_prefix "rig_caps_tx_range_list1_set", (swig_wrapper_func) _wrap_rig_caps_tx_range_list1_set, NULL},
    { SWIG_prefix "rig_caps_tx_range_list1_get", (swig_wrapper_func) _wrap_rig_caps_tx_range_list1_get, NULL},
    { SWIG_prefix "rig_caps_rx_range_list2_set", (swig_wrapper_func) _wrap_rig_caps_rx_range_list2_set, NULL},
    { SWIG_prefix "rig_caps_rx_range_list2_get", (swig_wrapper_func) _wrap_rig_caps_rx_range_list2_get, NULL},
    { SWIG_prefix "rig_caps_tx_range_list2_set", (swig_wrapper_func) _wrap_rig_caps_tx_range_list2_set, NULL},
    { SWIG_prefix "rig_caps_tx_range_list2_get", (swig_wrapper_func) _wrap_rig_caps_tx_range_list2_get, NULL},
    { SWIG_prefix "rig_caps_rx_range_list3_set", (swig_wrapper_func) _wrap_rig_caps_rx_range_list3_set, NULL},
    { SWIG_prefix "rig_caps_rx_range_list3_get", (swig_wrapper_func) _wrap_rig_caps_rx_range_list3_get, NULL},
    { SWIG_prefix "rig_caps_tx_range_list3_set", (swig_wrapper_func) _wrap_rig_caps_tx_range_list3_set, NULL},
    { SWIG_prefix "rig_caps_tx_range_list3_get", (swig_wrapper_func) _wrap_rig_caps_tx_range_list3_get, NULL},
    { SWIG_prefix "rig_caps_rx_range_list4_set", (swig_wrapper_func) _wrap_rig_caps_rx_range_list4_set, NULL},
    { SWIG_prefix "rig_caps_rx_range_list4_get", (swig_wrapper_func) _wrap_rig_caps_rx_range_list4_get, NULL},
    { SWIG_prefix "rig_caps_tx_range_list4_set", (swig_wrapper_func) _wrap_rig_caps_tx_range_list4_set, NULL},
    { SWIG_prefix "rig_caps_tx_range_list4_get", (swig_wrapper_func) _wrap_rig_caps_tx_range_list4_get, NULL},
    { SWIG_prefix "rig_caps_rx_range_list5_set", (swig_wrapper_func) _wrap_rig_caps_rx_range_list5_set, NULL},
    { SWIG_prefix "rig_caps_rx_range_list5_get", (swig_wrapper_func) _wrap_rig_caps_rx_range_list5_get, NULL},
    { SWIG_prefix "rig_caps_tx_range_list5_set", (swig_wrapper_func) _wrap_rig_caps_tx_range_list5_set, NULL},
    { SWIG_prefix "rig_caps_tx_range_list5_get", (swig_wrapper_func) _wrap_rig_caps_tx_range_list5_get, NULL},
    { SWIG_prefix "rig_caps_tuning_steps_set", (swig_wrapper_func) _wrap_rig_caps_tuning_steps_set, NULL},
    { SWIG_prefix "rig_caps_tuning_steps_get", (swig_wrapper_func) _wrap_rig_caps_tuning_steps_get, NULL},
    { SWIG_prefix "rig_caps_filters_set", (swig_wrapper_func) _wrap_rig_caps_filters_set, NULL},
    { SWIG_prefix "rig_caps_filters_get", (swig_wrapper_func) _wrap_rig_caps_filters_get, NULL},
    { SWIG_prefix "rig_caps_str_cal_set", (swig_wrapper_func) _wrap_rig_caps_str_cal_set, NULL},
    { SWIG_prefix "rig_caps_str_cal_get", (swig_wrapper_func) _wrap_rig_caps_str_cal_get, NULL},
    { SWIG_prefix "rig_caps_swr_cal_set", (swig_wrapper_func) _wrap_rig_caps_swr_cal_set, NULL},
    { SWIG_prefix "rig_caps_swr_cal_get", (swig_wrapper_func) _wrap_rig_caps_swr_cal_get, NULL},
    { SWIG_prefix "rig_caps_alc_cal_set", (swig_wrapper_func) _wrap_rig_caps_alc_cal_set, NULL},
    { SWIG_prefix "rig_caps_alc_cal_get", (swig_wrapper_func) _wrap_rig_caps_alc_cal_get, NULL},
    { SWIG_prefix "rig_caps_rfpower_meter_cal_set", (swig_wrapper_func) _wrap_rig_caps_rfpower_meter_cal_set, NULL},
    { SWIG_prefix "rig_caps_rfpower_meter_cal_get", (swig_wrapper_func) _wrap_rig_caps_rfpower_meter_cal_get, NULL},
    { SWIG_prefix "rig_caps_comp_meter_cal_set", (swig_wrapper_func) _wrap_rig_caps_comp_meter_cal_set, NULL},
    { SWIG_prefix "rig_caps_comp_meter_cal_get", (swig_wrapper_func) _wrap_rig_caps_comp_meter_cal_get, NULL},
    { SWIG_prefix "rig_caps_vd_meter_cal_set", (swig_wrapper_func) _wrap_rig_caps_vd_meter_cal_set, NULL},
    { SWIG_prefix "rig_caps_vd_meter_cal_get", (swig_wrapper_func) _wrap_rig_caps_vd_meter_cal_get, NULL},
    { SWIG_prefix "rig_caps_id_meter_cal_set", (swig_wrapper_func) _wrap_rig_caps_id_meter_cal_set, NULL},
    { SWIG_prefix "rig_caps_id_meter_cal_get", (swig_wrapper_func) _wrap_rig_caps_id_meter_cal_get, NULL},
    { SWIG_prefix "rig_caps_spectrum_scopes_set", (swig_wrapper_func) _wrap_rig_caps_spectrum_scopes_set, NULL},
    { SWIG_prefix "rig_caps_spectrum_scopes_get", (swig_wrapper_func) _wrap_rig_caps_spectrum_scopes_get, NULL},
    { SWIG_prefix "rig_caps_spectrum_modes_set", (swig_wrapper_func) _wrap_rig_caps_spectrum_modes_set, NULL},
    { SWIG_prefix "rig_caps_spectrum_modes_get", (swig_wrapper_func) _wrap_rig_caps_spectrum_modes_get, NULL},
    { SWIG_prefix "rig_caps_spectrum_spans_set", (swig_wrapper_func) _wrap_rig_caps_spectrum_spans_set, NULL},
    { SWIG_prefix "rig_caps_spectrum_spans_get", (swig_wrapper_func) _wrap_rig_caps_spectrum_spans_get, NULL},
    { SWIG_prefix "rig_caps_spectrum_avg_modes_set", (swig_wrapper_func) _wrap_rig_caps_spectrum_avg_modes_set, NULL},
    { SWIG_prefix "rig_caps_spectrum_avg_modes_get", (swig_wrapper_func) _wrap_rig_caps_spectrum_avg_modes_get, NULL},
    { SWIG_prefix "rig_caps_spectrum_attenuator_set", (swig_wrapper_func) _wrap_rig_caps_spectrum_attenuator_set, NULL},
    { SWIG_prefix "rig_caps_spectrum_attenuator_get", (swig_wrapper_func) _wrap_rig_caps_spectrum_attenuator_get, NULL},
    { SWIG_prefix "rig_caps_cfgparams_set", (swig_wrapper_func) _wrap_rig_caps_cfgparams_set, NULL},
    { SWIG_prefix "rig_caps_cfgparams_get", (swig_wrapper_func) _wrap_rig_caps_cfgparams_get, NULL},
    { SWIG_prefix "rig_caps_priv_set", (swig_wrapper_func) _wrap_rig_caps_priv_set, NULL},
    { SWIG_prefix "rig_caps_priv_get", (swig_wrapper_func) _wrap_rig_caps_priv_get, NULL},
    { SWIG_prefix "rig_caps_rig_init_set", (swig_wrapper_func) _wrap_rig_caps_rig_init_set, NULL},
    { SWIG_prefix "rig_caps_rig_init_get", (swig_wrapper_func) _wrap_rig_caps_rig_init_get, NULL},
    { SWIG_prefix "rig_caps_rig_cleanup_set", (swig_wrapper_func) _wrap_rig_caps_rig_cleanup_set, NULL},
    { SWIG_prefix "rig_caps_rig_cleanup_get", (swig_wrapper_func) _wrap_rig_caps_rig_cleanup_get, NULL},
    { SWIG_prefix "rig_caps_rig_open_set", (swig_wrapper_func) _wrap_rig_caps_rig_open_set, NULL},
    { SWIG_prefix "rig_caps_rig_open_get", (swig_wrapper_func) _wrap_rig_caps_rig_open_get, NULL},
    { SWIG_prefix "rig_caps_rig_close_set", (swig_wrapper_func) _wrap_rig_caps_rig_close_set, NULL},
    { SWIG_prefix "rig_caps_rig_close_get", (swig_wrapper_func) _wrap_rig_caps_rig_close_get, NULL},
    { SWIG_prefix "rig_caps_set_freq_set", (swig_wrapper_func) _wrap_rig_caps_set_freq_set, NULL},
    { SWIG_prefix "rig_caps_set_freq_get", (swig_wrapper_func) _wrap_rig_caps_set_freq_get, NULL},
    { SWIG_prefix "rig_caps_get_freq_set", (swig_wrapper_func) _wrap_rig_caps_get_freq_set, NULL},
    { SWIG_prefix "rig_caps_get_freq_get", (swig_wrapper_func) _wrap_rig_caps_get_freq_get, NULL},
    { SWIG_prefix "rig_caps_set_mode_set", (swig_wrapper_func) _wrap_rig_caps_set_mode_set, NULL},
    { SWIG_prefix "rig_caps_set_mode_get", (swig_wrapper_func) _wrap_rig_caps_set_mode_get, NULL},
    { SWIG_prefix "rig_caps_get_mode_set", (swig_wrapper_func) _wrap_rig_caps_get_mode_set, NULL},
    { SWIG_prefix "rig_caps_get_mode_get", (swig_wrapper_func) _wrap_rig_caps_get_mode_get, NULL},
    { SWIG_prefix "rig_caps_set_vfo_set", (swig_wrapper_func) _wrap_rig_caps_set_vfo_set, NULL},
    { SWIG_prefix "rig_caps_set_vfo_get", (swig_wrapper_func) _wrap_rig_caps_set_vfo_get, NULL},
    { SWIG_prefix "rig_caps_get_vfo_set", (swig_wrapper_func) _wrap_rig_caps_get_vfo_set, NULL},
    { SWIG_prefix "rig_caps_get_vfo_get", (swig_wrapper_func) _wrap_rig_caps_get_vfo_get, NULL},
    { SWIG_prefix "rig_caps_set_ptt_set", (swig_wrapper_func) _wrap_rig_caps_set_ptt_set, NULL},
    { SWIG_prefix "rig_caps_set_ptt_get", (swig_wrapper_func) _wrap_rig_caps_set_ptt_get, NULL},
    { SWIG_prefix "rig_caps_get_ptt_set", (swig_wrapper_func) _wrap_rig_caps_get_ptt_set, NULL},
    { SWIG_prefix "rig_caps_get_ptt_get", (swig_wrapper_func) _wrap_rig_caps_get_ptt_get, NULL},
    { SWIG_prefix "rig_caps_get_dcd_set", (swig_wrapper_func) _wrap_rig_caps_get_dcd_set, NULL},
    { SWIG_prefix "rig_caps_get_dcd_get", (swig_wrapper_func) _wrap_rig_caps_get_dcd_get, NULL},
    { SWIG_prefix "rig_caps_set_rptr_shift_set", (swig_wrapper_func) _wrap_rig_caps_set_rptr_shift_set, NULL},
    { SWIG_prefix "rig_caps_set_rptr_shift_get", (swig_wrapper_func) _wrap_rig_caps_set_rptr_shift_get, NULL},
    { SWIG_prefix "rig_caps_get_rptr_shift_set", (swig_wrapper_func) _wrap_rig_caps_get_rptr_shift_set, NULL},
    { SWIG_prefix "rig_caps_get_rptr_shift_get", (swig_wrapper_func) _wrap_rig_caps_get_rptr_shift_get, NULL},
    { SWIG_prefix "rig_caps_set_rptr_offs_set", (swig_wrapper_func) _wrap_rig_caps_set_rptr_offs_set, NULL},
    { SWIG_prefix "rig_caps_set_rptr_offs_get", (swig_wrapper_func) _wrap_rig_caps_set_rptr_offs_get, NULL},
    { SWIG_prefix "rig_caps_get_rptr_offs_set", (swig_wrapper_func) _wrap_rig_caps_get_rptr_offs_set, NULL},
    { SWIG_prefix "rig_caps_get_rptr_offs_get", (swig_wrapper_func) _wrap_rig_caps_get_rptr_offs_get, NULL},
    { SWIG_prefix "rig_caps_set_split_freq_set", (swig_wrapper_func) _wrap_rig_caps_set_split_freq_set, NULL},
    { SWIG_prefix "rig_caps_set_split_freq_get", (swig_wrapper_func) _wrap_rig_caps_set_split_freq_get, NULL},
    { SWIG_prefix "rig_caps_get_split_freq_set", (swig_wrapper_func) _wrap_rig_caps_get_split_freq_set, NULL},
    { SWIG_prefix "rig_caps_get_split_freq_get", (swig_wrapper_func) _wrap_rig_caps_get_split_freq_get, NULL},
    { SWIG_prefix "rig_caps_set_split_mode_set", (swig_wrapper_func) _wrap_rig_caps_set_split_mode_set, NULL},
    { SWIG_prefix "rig_caps_set_split_mode_get", (swig_wrapper_func) _wrap_rig_caps_set_split_mode_get, NULL},
    { SWIG_prefix "rig_caps_get_split_mode_set", (swig_wrapper_func) _wrap_rig_caps_get_split_mode_set, NULL},
    { SWIG_prefix "rig_caps_get_split_mode_get", (swig_wrapper_func) _wrap_rig_caps_get_split_mode_get, NULL},
    { SWIG_prefix "rig_caps_set_split_freq_mode_set", (swig_wrapper_func) _wrap_rig_caps_set_split_freq_mode_set, NULL},
    { SWIG_prefix "rig_caps_set_split_freq_mode_get", (swig_wrapper_func) _wrap_rig_caps_set_split_freq_mode_get, NULL},
    { SWIG_prefix "rig_caps_get_split_freq_mode_set", (swig_wrapper_func) _wrap_rig_caps_get_split_freq_mode_set, NULL},
    { SWIG_prefix "rig_caps_get_split_freq_mode_get", (swig_wrapper_func) _wrap_rig_caps_get_split_freq_mode_get, NULL},
    { SWIG_prefix "rig_caps_set_split_vfo_set", (swig_wrapper_func) _wrap_rig_caps_set_split_vfo_set, NULL},
    { SWIG_prefix "rig_caps_set_split_vfo_get", (swig_wrapper_func) _wrap_rig_caps_set_split_vfo_get, NULL},
    { SWIG_prefix "rig_caps_get_split_vfo_set", (swig_wrapper_func) _wrap_rig_caps_get_split_vfo_set, NULL},
    { SWIG_prefix "rig_caps_get_split_vfo_get", (swig_wrapper_func) _wrap_rig_caps_get_split_vfo_get, NULL},
    { SWIG_prefix "rig_caps_set_rit_set", (swig_wrapper_func) _wrap_rig_caps_set_rit_set, NULL},
    { SWIG_prefix "rig_caps_set_rit_get", (swig_wrapper_func) _wrap_rig_caps_set_rit_get, NULL},
    { SWIG_prefix "rig_caps_get_rit_set", (swig_wrapper_func) _wrap_rig_caps_get_rit_set, NULL},
    { SWIG_prefix "rig_caps_get_rit_get", (swig_wrapper_func) _wrap_rig_caps_get_rit_get, NULL},
    { SWIG_prefix "rig_caps_set_xit_set", (swig_wrapper_func) _wrap_rig_caps_set_xit_set, NULL},
    { SWIG_prefix "rig_caps_set_xit_get", (swig_wrapper_func) _wrap_rig_caps_set_xit_get, NULL},
    { SWIG_prefix "rig_caps_get_xit_set", (swig_wrapper_func) _wrap_rig_caps_get_xit_set, NULL},
    { SWIG_prefix "rig_caps_get_xit_get", (swig_wrapper_func) _wrap_rig_caps_get_xit_get, NULL},
    { SWIG_prefix "rig_caps_set_ts_set", (swig_wrapper_func) _wrap_rig_caps_set_ts_set, NULL},
    { SWIG_prefix "rig_caps_set_ts_get", (swig_wrapper_func) _wrap_rig_caps_set_ts_get, NULL},
    { SWIG_prefix "rig_caps_get_ts_set", (swig_wrapper_func) _wrap_rig_caps_get_ts_set, NULL},
    { SWIG_prefix "rig_caps_get_ts_get", (swig_wrapper_func) _wrap_rig_caps_get_ts_get, NULL},
    { SWIG_prefix "rig_caps_set_dcs_code_set", (swig_wrapper_func) _wrap_rig_caps_set_dcs_code_set, NULL},
    { SWIG_prefix "rig_caps_set_dcs_code_get", (swig_wrapper_func) _wrap_rig_caps_set_dcs_code_get, NULL},
    { SWIG_prefix "rig_caps_get_dcs_code_set", (swig_wrapper_func) _wrap_rig_caps_get_dcs_code_set, NULL},
    { SWIG_prefix "rig_caps_get_dcs_code_get", (swig_wrapper_func) _wrap_rig_caps_get_dcs_code_get, NULL},
    { SWIG_prefix "rig_caps_set_tone_set", (swig_wrapper_func) _wrap_rig_caps_set_tone_set, NULL},
    { SWIG_prefix "rig_caps_set_tone_get", (swig_wrapper_func) _wrap_rig_caps_set_tone_get, NULL},
    { SWIG_prefix "rig_caps_get_tone_set", (swig_wrapper_func) _wrap_rig_caps_get_tone_set, NULL},
    { SWIG_prefix "rig_caps_get_tone_get", (swig_wrapper_func) _wrap_rig_caps_get_tone_get, NULL},
    { SWIG_prefix "rig_caps_set_ctcss_tone_set", (swig_wrapper_func) _wrap_rig_caps_set_ctcss_tone_set, NULL},
    { SWIG_prefix "rig_caps_set_ctcss_tone_get", (swig_wrapper_func) _wrap_rig_caps_set_ctcss_tone_get, NULL},
    { SWIG_prefix "rig_caps_get_ctcss_tone_set", (swig_wrapper_func) _wrap_rig_caps_get_ctcss_tone_set, NULL},
    { SWIG_prefix "rig_caps_get_ctcss_tone_get", (swig_wrapper_func) _wrap_rig_caps_get_ctcss_tone_get, NULL},
    { SWIG_prefix "rig_caps_set_dcs_sql_set", (swig_wrapper_func) _wrap_rig_caps_set_dcs_sql_set, NULL},
    { SWIG_prefix "rig_caps_set_dcs_sql_get", (swig_wrapper_func) _wrap_rig_caps_set_dcs_sql_get, NULL},
    { SWIG_prefix "rig_caps_get_dcs_sql_set", (swig_wrapper_func) _wrap_rig_caps_get_dcs_sql_set, NULL},
    { SWIG_prefix "rig_caps_get_dcs_sql_get", (swig_wrapper_func) _wrap_rig_caps_get_dcs_sql_get, NULL},
    { SWIG_prefix "rig_caps_set_tone_sql_set", (swig_wrapper_func) _wrap_rig_caps_set_tone_sql_set, NULL},
    { SWIG_prefix "rig_caps_set_tone_sql_get", (swig_wrapper_func) _wrap_rig_caps_set_tone_sql_get, NULL},
    { SWIG_prefix "rig_caps_get_tone_sql_set", (swig_wrapper_func) _wrap_rig_caps_get_tone_sql_set, NULL},
    { SWIG_prefix "rig_caps_get_tone_sql_get", (swig_wrapper_func) _wrap_rig_caps_get_tone_sql_get, NULL},
    { SWIG_prefix "rig_caps_set_ctcss_sql_set", (swig_wrapper_func) _wrap_rig_caps_set_ctcss_sql_set, NULL},
    { SWIG_prefix "rig_caps_set_ctcss_sql_get", (swig_wrapper_func) _wrap_rig_caps_set_ctcss_sql_get, NULL},
    { SWIG_prefix "rig_caps_get_ctcss_sql_set", (swig_wrapper_func) _wrap_rig_caps_get_ctcss_sql_set, NULL},
    { SWIG_prefix "rig_caps_get_ctcss_sql_get", (swig_wrapper_func) _wrap_rig_caps_get_ctcss_sql_get, NULL},
    { SWIG_prefix "rig_caps_power2mW_set", (swig_wrapper_func) _wrap_rig_caps_power2mW_set, NULL},
    { SWIG_prefix "rig_caps_power2mW_get", (swig_wrapper_func) _wrap_rig_caps_power2mW_get, NULL},
    { SWIG_prefix "rig_caps_mW2power_set", (swig_wrapper_func) _wrap_rig_caps_mW2power_set, NULL},
    { SWIG_prefix "rig_caps_mW2power_get", (swig_wrapper_func) _wrap_rig_caps_mW2power_get, NULL},
    { SWIG_prefix "rig_caps_set_powerstat_set", (swig_wrapper_func) _wrap_rig_caps_set_powerstat_set, NULL},
    { SWIG_prefix "rig_caps_set_powerstat_get", (swig_wrapper_func) _wrap_rig_caps_set_powerstat_get, NULL},
    { SWIG_prefix "rig_caps_get_powerstat_set", (swig_wrapper_func) _wrap_rig_caps_get_powerstat_set, NULL},
    { SWIG_prefix "rig_caps_get_powerstat_get", (swig_wrapper_func) _wrap_rig_caps_get_powerstat_get, NULL},
    { SWIG_prefix "rig_caps_reset_set", (swig_wrapper_func) _wrap_rig_caps_reset_set, NULL},
    { SWIG_prefix "rig_caps_reset_get", (swig_wrapper_func) _wrap_rig_caps_reset_get, NULL},
    { SWIG_prefix "rig_caps_set_ant_set", (swig_wrapper_func) _wrap_rig_caps_set_ant_set, NULL},
    { SWIG_prefix "rig_caps_set_ant_get", (swig_wrapper_func) _wrap_rig_caps_set_ant_get, NULL},
    { SWIG_prefix "rig_caps_get_ant_set", (swig_wrapper_func) _wrap_rig_caps_get_ant_set, NULL},
    { SWIG_prefix "rig_caps_get_ant_get", (swig_wrapper_func) _wrap_rig_caps_get_ant_get, NULL},
    { SWIG_prefix "rig_caps_set_level_set", (swig_wrapper_func) _wrap_rig_caps_set_level_set, NULL},
    { SWIG_prefix "rig_caps_set_level_get", (swig_wrapper_func) _wrap_rig_caps_set_level_get, NULL},
    { SWIG_prefix "rig_caps_get_level_set", (swig_wrapper_func) _wrap_rig_caps_get_level_set, NULL},
    { SWIG_prefix "rig_caps_get_level_get", (swig_wrapper_func) _wrap_rig_caps_get_level_get, NULL},
    { SWIG_prefix "rig_caps_set_func_set", (swig_wrapper_func) _wrap_rig_caps_set_func_set, NULL},
    { SWIG_prefix "rig_caps_set_func_get", (swig_wrapper_func) _wrap_rig_caps_set_func_get, NULL},
    { SWIG_prefix "rig_caps_get_func_set", (swig_wrapper_func) _wrap_rig_caps_get_func_set, NULL},
    { SWIG_prefix "rig_caps_get_func_get", (swig_wrapper_func) _wrap_rig_caps_get_func_get, NULL},
    { SWIG_prefix "rig_caps_set_parm_set", (swig_wrapper_func) _wrap_rig_caps_set_parm_set, NULL},
    { SWIG_prefix "rig_caps_set_parm_get", (swig_wrapper_func) _wrap_rig_caps_set_parm_get, NULL},
    { SWIG_prefix "rig_caps_get_parm_set", (swig_wrapper_func) _wrap_rig_caps_get_parm_set, NULL},
    { SWIG_prefix "rig_caps_get_parm_get", (swig_wrapper_func) _wrap_rig_caps_get_parm_get, NULL},
    { SWIG_prefix "rig_caps_set_ext_level_set", (swig_wrapper_func) _wrap_rig_caps_set_ext_level_set, NULL},
    { SWIG_prefix "rig_caps_set_ext_level_get", (swig_wrapper_func) _wrap_rig_caps_set_ext_level_get, NULL},
    { SWIG_prefix "rig_caps_get_ext_level_set", (swig_wrapper_func) _wrap_rig_caps_get_ext_level_set, NULL},
    { SWIG_prefix "rig_caps_get_ext_level_get", (swig_wrapper_func) _wrap_rig_caps_get_ext_level_get, NULL},
    { SWIG_prefix "rig_caps_set_ext_func_set", (swig_wrapper_func) _wrap_rig_caps_set_ext_func_set, NULL},
    { SWIG_prefix "rig_caps_set_ext_func_get", (swig_wrapper_func) _wrap_rig_caps_set_ext_func_get, NULL},
    { SWIG_prefix "rig_caps_get_ext_func_set", (swig_wrapper_func) _wrap_rig_caps_get_ext_func_set, NULL},
    { SWIG_prefix "rig_caps_get_ext_func_get", (swig_wrapper_func) _wrap_rig_caps_get_ext_func_get, NULL},
    { SWIG_prefix "rig_caps_set_ext_parm_set", (swig_wrapper_func) _wrap_rig_caps_set_ext_parm_set, NULL},
    { SWIG_prefix "rig_caps_set_ext_parm_get", (swig_wrapper_func) _wrap_rig_caps_set_ext_parm_get, NULL},
    { SWIG_prefix "rig_caps_get_ext_parm_set", (swig_wrapper_func) _wrap_rig_caps_get_ext_parm_set, NULL},
    { SWIG_prefix "rig_caps_get_ext_parm_get", (swig_wrapper_func) _wrap_rig_caps_get_ext_parm_get, NULL},
    { SWIG_prefix "rig_caps_set_conf_set", (swig_wrapper_func) _wrap_rig_caps_set_conf_set, NULL},
    { SWIG_prefix "rig_caps_set_conf_get", (swig_wrapper_func) _wrap_rig_caps_set_conf_get, NULL},
    { SWIG_prefix "rig_caps_get_conf_set", (swig_wrapper_func) _wrap_rig_caps_get_conf_set, NULL},
    { SWIG_prefix "rig_caps_get_conf_get", (swig_wrapper_func) _wrap_rig_caps_get_conf_get, NULL},
    { SWIG_prefix "rig_caps_send_dtmf_set", (swig_wrapper_func) _wrap_rig_caps_send_dtmf_set, NULL},
    { SWIG_prefix "rig_caps_send_dtmf_get", (swig_wrapper_func) _wrap_rig_caps_send_dtmf_get, NULL},
    { SWIG_prefix "rig_caps_recv_dtmf_set", (swig_wrapper_func) _wrap_rig_caps_recv_dtmf_set, NULL},
    { SWIG_prefix "rig_caps_recv_dtmf_get", (swig_wrapper_func) _wrap_rig_caps_recv_dtmf_get, NULL},
    { SWIG_prefix "rig_caps_send_morse_set", (swig_wrapper_func) _wrap_rig_caps_send_morse_set, NULL},
    { SWIG_prefix "rig_caps_send_morse_get", (swig_wrapper_func) _wrap_rig_caps_send_morse_get, NULL},
    { SWIG_prefix "rig_caps_stop_morse_set", (swig_wrapper_func) _wrap_rig_caps_stop_morse_set, NULL},
    { SWIG_prefix "rig_caps_stop_morse_get", (swig_wrapper_func) _wrap_rig_caps_stop_morse_get, NULL},
    { SWIG_prefix "rig_caps_wait_morse_set", (swig_wrapper_func) _wrap_rig_caps_wait_morse_set, NULL},
    { SWIG_prefix "rig_caps_wait_morse_get", (swig_wrapper_func) _wrap_rig_caps_wait_morse_get, NULL},
    { SWIG_prefix "rig_caps_send_voice_mem_set", (swig_wrapper_func) _wrap_rig_caps_send_voice_mem_set, NULL},
    { SWIG_prefix "rig_caps_send_voice_mem_get", (swig_wrapper_func) _wrap_rig_caps_send_voice_mem_get, NULL},
    { SWIG_prefix "rig_caps_stop_voice_mem_set", (swig_wrapper_func) _wrap_rig_caps_stop_voice_mem_set, NULL},
    { SWIG_prefix "rig_caps_stop_voice_mem_get", (swig_wrapper_func) _wrap_rig_caps_stop_voice_mem_get, NULL},
    { SWIG_prefix "rig_caps_set_bank_set", (swig_wrapper_func) _wrap_rig_caps_set_bank_set, NULL},
    { SWIG_prefix "rig_caps_set_bank_get", (swig_wrapper_func) _wrap_rig_caps_set_bank_get, NULL},
    { SWIG_prefix "rig_caps_set_mem_set", (swig_wrapper_func) _wrap_rig_caps_set_mem_set, NULL},
    { SWIG_prefix "rig_caps_set_mem_get", (swig_wrapper_func) _wrap_rig_caps_set_mem_get, NULL},
    { SWIG_prefix "rig_caps_get_mem_set", (swig_wrapper_func) _wrap_rig_caps_get_mem_set, NULL},
    { SWIG_prefix "rig_caps_get_mem_get", (swig_wrapper_func) _wrap_rig_caps_get_mem_get, NULL},
    { SWIG_prefix "rig_caps_vfo_op_set", (swig_wrapper_func) _wrap_rig_caps_vfo_op_set, NULL},
    { SWIG_prefix "rig_caps_vfo_op_get", (swig_wrapper_func) _wrap_rig_caps_vfo_op_get, NULL},
    { SWIG_prefix "rig_caps_scan_set", (swig_wrapper_func) _wrap_rig_caps_scan_set, NULL},
    { SWIG_prefix "rig_caps_scan_get", (swig_wrapper_func) _wrap_rig_caps_scan_get, NULL},
    { SWIG_prefix "rig_caps_set_trn_set", (swig_wrapper_func) _wrap_rig_caps_set_trn_set, NULL},
    { SWIG_prefix "rig_caps_set_trn_get", (swig_wrapper_func) _wrap_rig_caps_set_trn_get, NULL},
    { SWIG_prefix "rig_caps_get_trn_set", (swig_wrapper_func) _wrap_rig_caps_get_trn_set, NULL},
    { SWIG_prefix "rig_caps_get_trn_get", (swig_wrapper_func) _wrap_rig_caps_get_trn_get, NULL},
    { SWIG_prefix "rig_caps_decode_event_set", (swig_wrapper_func) _wrap_rig_caps_decode_event_set, NULL},
    { SWIG_prefix "rig_caps_decode_event_get", (swig_wrapper_func) _wrap_rig_caps_decode_event_get, NULL},
    { SWIG_prefix "rig_caps_set_channel_set", (swig_wrapper_func) _wrap_rig_caps_set_channel_set, NULL},
    { SWIG_prefix "rig_caps_set_channel_get", (swig_wrapper_func) _wrap_rig_caps_set_channel_get, NULL},
    { SWIG_prefix "rig_caps_get_channel_set", (swig_wrapper_func) _wrap_rig_caps_get_channel_set, NULL},
    { SWIG_prefix "rig_caps_get_channel_get", (swig_wrapper_func) _wrap_rig_caps_get_channel_get, NULL},
    { SWIG_prefix "rig_caps_get_info_set", (swig_wrapper_func) _wrap_rig_caps_get_info_set, NULL},
    { SWIG_prefix "rig_caps_get_info_get", (swig_wrapper_func) _wrap_rig_caps_get_info_get, NULL},
    { SWIG_prefix "rig_caps_set_chan_all_cb_set", (swig_wrapper_func) _wrap_rig_caps_set_chan_all_cb_set, NULL},
    { SWIG_prefix "rig_caps_set_chan_all_cb_get", (swig_wrapper_func) _wrap_rig_caps_set_chan_all_cb_get, NULL},
    { SWIG_prefix "rig_caps_get_chan_all_cb_set", (swig_wrapper_func) _wrap_rig_caps_get_chan_all_cb_set, NULL},
    { SWIG_prefix "rig_caps_get_chan_all_cb_get", (swig_wrapper_func) _wrap_rig_caps_get_chan_all_cb_get, NULL},
    { SWIG_prefix "rig_caps_set_mem_all_cb_set", (swig_wrapper_func) _wrap_rig_caps_set_mem_all_cb_set, NULL},
    { SWIG_prefix "rig_caps_set_mem_all_cb_get", (swig_wrapper_func) _wrap_rig_caps_set_mem_all_cb_get, NULL},
    { SWIG_prefix "rig_caps_get_mem_all_cb_set", (swig_wrapper_func) _wrap_rig_caps_get_mem_all_cb_set, NULL},
    { SWIG_prefix "rig_caps_get_mem_all_cb_get", (swig_wrapper_func) _wrap_rig_caps_get_mem_all_cb_get, NULL},
    { SWIG_prefix "rig_caps_set_vfo_opt_set", (swig_wrapper_func) _wrap_rig_caps_set_vfo_opt_set, NULL},
    { SWIG_prefix "rig_caps_set_vfo_opt_get", (swig_wrapper_func) _wrap_rig_caps_set_vfo_opt_get, NULL},
    { SWIG_prefix "rig_caps_rig_get_vfo_info_set", (swig_wrapper_func) _wrap_rig_caps_rig_get_vfo_info_set, NULL},
    { SWIG_prefix "rig_caps_rig_get_vfo_info_get", (swig_wrapper_func) _wrap_rig_caps_rig_get_vfo_info_get, NULL},
    { SWIG_prefix "rig_caps_set_clock_set", (swig_wrapper_func) _wrap_rig_caps_set_clock_set, NULL},
    { SWIG_prefix "rig_caps_set_clock_get", (swig_wrapper_func) _wrap_rig_caps_set_clock_get, NULL},
    { SWIG_prefix "rig_caps_get_clock_set", (swig_wrapper_func) _wrap_rig_caps_get_clock_set, NULL},
    { SWIG_prefix "rig_caps_get_clock_get", (swig_wrapper_func) _wrap_rig_caps_get_clock_get, NULL},
    { SWIG_prefix "rig_caps_clone_combo_set_get", (swig_wrapper_func) _wrap_rig_caps_clone_combo_set_get, NULL},
    { SWIG_prefix "rig_caps_clone_combo_get_get", (swig_wrapper_func) _wrap_rig_caps_clone_combo_get_get, NULL},
    { SWIG_prefix "rig_caps_async_data_supported_set", (swig_wrapper_func) _wrap_rig_caps_async_data_supported_set, NULL},
    { SWIG_prefix "rig_caps_async_data_supported_get", (swig_wrapper_func) _wrap_rig_caps_async_data_supported_get, NULL},
    { SWIG_prefix "rig_caps_read_frame_direct_set", (swig_wrapper_func) _wrap_rig_caps_read_frame_direct_set, NULL},
    { SWIG_prefix "rig_caps_read_frame_direct_get", (swig_wrapper_func) _wrap_rig_caps_read_frame_direct_get, NULL},
    { SWIG_prefix "rig_caps_is_async_frame_set", (swig_wrapper_func) _wrap_rig_caps_is_async_frame_set, NULL},
    { SWIG_prefix "rig_caps_is_async_frame_get", (swig_wrapper_func) _wrap_rig_caps_is_async_frame_get, NULL},
    { SWIG_prefix "rig_caps_process_async_frame_set", (swig_wrapper_func) _wrap_rig_caps_process_async_frame_set, NULL},
    { SWIG_prefix "rig_caps_process_async_frame_get", (swig_wrapper_func) _wrap_rig_caps_process_async_frame_get, NULL},
    { SWIG_prefix "rig_caps_hamlib_check_rig_caps_set", (swig_wrapper_func) _wrap_rig_caps_hamlib_check_rig_caps_set, NULL},
    { SWIG_prefix "rig_caps_hamlib_check_rig_caps_get", (swig_wrapper_func) _wrap_rig_caps_hamlib_check_rig_caps_get, NULL},
    { SWIG_prefix "rig_caps_get_conf2_set", (swig_wrapper_func) _wrap_rig_caps_get_conf2_set, NULL},
    { SWIG_prefix "rig_caps_get_conf2_get", (swig_wrapper_func) _wrap_rig_caps_get_conf2_get, NULL},
    { SWIG_prefix "rig_caps_password_set", (swig_wrapper_func) _wrap_rig_caps_password_set, NULL},
    { SWIG_prefix "rig_caps_password_get", (swig_wrapper_func) _wrap_rig_caps_password_get, NULL},
    { SWIG_prefix "rig_caps_set_lock_mode_set", (swig_wrapper_func) _wrap_rig_caps_set_lock_mode_set, NULL},
    { SWIG_prefix "rig_caps_set_lock_mode_get", (swig_wrapper_func) _wrap_rig_caps_set_lock_mode_get, NULL},
    { SWIG_prefix "rig_caps_get_lock_mode_set", (swig_wrapper_func) _wrap_rig_caps_get_lock_mode_set, NULL},
    { SWIG_prefix "rig_caps_get_lock_mode_get", (swig_wrapper_func) _wrap_rig_caps_get_lock_mode_get, NULL},
    { SWIG_prefix "rig_caps_timeout_retry_set", (swig_wrapper_func) _wrap_rig_caps_timeout_retry_set, NULL},
    { SWIG_prefix "rig_caps_timeout_retry_get", (swig_wrapper_func) _wrap_rig_caps_timeout_retry_get, NULL},
    { SWIG_prefix "rig_caps_morse_qsize_set", (swig_wrapper_func) _wrap_rig_caps_morse_qsize_set, NULL},
    { SWIG_prefix "rig_caps_morse_qsize_get", (swig_wrapper_func) _wrap_rig_caps_morse_qsize_get, NULL},
    { SWIG_prefix "new_rig_caps", (swig_wrapper_func) _wrap_new_rig_caps, NULL},
    { SWIG_prefix "delete_rig_caps", (swig_wrapper_func) _wrap_delete_rig_caps, NULL},
    { SWIG_prefix "rig_caps", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_rig_caps},
    { SWIG_prefix "multicast_s_multicast_running_set", (swig_wrapper_func) _wrap_multicast_s_multicast_running_set, NULL},
    { SWIG_prefix "multicast_s_multicast_running_get", (swig_wrapper_func) _wrap_multicast_s_multicast_running_get, NULL},
    { SWIG_prefix "multicast_s_sock_set", (swig_wrapper_func) _wrap_multicast_s_sock_set, NULL},
    { SWIG_prefix "multicast_s_sock_get", (swig_wrapper_func) _wrap_multicast_s_sock_get, NULL},
    { SWIG_prefix "multicast_s_seqnumber_set", (swig_wrapper_func) _wrap_multicast_s_seqnumber_set, NULL},
    { SWIG_prefix "multicast_s_seqnumber_get", (swig_wrapper_func) _wrap_multicast_s_seqnumber_get, NULL},
    { SWIG_prefix "multicast_s_runflag_set", (swig_wrapper_func) _wrap_multicast_s_runflag_set, NULL},
    { SWIG_prefix "multicast_s_runflag_get", (swig_wrapper_func) _wrap_multicast_s_runflag_get, NULL},
    { SWIG_prefix "multicast_s_threadid_set", (swig_wrapper_func) _wrap_multicast_s_threadid_set, NULL},
    { SWIG_prefix "multicast_s_threadid_get", (swig_wrapper_func) _wrap_multicast_s_threadid_get, NULL},
    { SWIG_prefix "multicast_s_mutex_set", (swig_wrapper_func) _wrap_multicast_s_mutex_set, NULL},
    { SWIG_prefix "multicast_s_mutex_get", (swig_wrapper_func) _wrap_multicast_s_mutex_get, NULL},
    { SWIG_prefix "multicast_s_mutex_initialized_set", (swig_wrapper_func) _wrap_multicast_s_mutex_initialized_set, NULL},
    { SWIG_prefix "multicast_s_mutex_initialized_get", (swig_wrapper_func) _wrap_multicast_s_mutex_initialized_get, NULL},
    { SWIG_prefix "multicast_s_dest_addr_set", (swig_wrapper_func) _wrap_multicast_s_dest_addr_set, NULL},
    { SWIG_prefix "multicast_s_dest_addr_get", (swig_wrapper_func) _wrap_multicast_s_dest_addr_get, NULL},
    { SWIG_prefix "multicast_s_port_set", (swig_wrapper_func) _wrap_multicast_s_port_set, NULL},
    { SWIG_prefix "multicast_s_port_get", (swig_wrapper_func) _wrap_multicast_s_port_get, NULL},
    { SWIG_prefix "new_multicast_s", (swig_wrapper_func) _wrap_new_multicast_s, NULL},
    { SWIG_prefix "delete_multicast_s", (swig_wrapper_func) _wrap_delete_multicast_s, NULL},
    { SWIG_prefix "multicast_s", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_multicast_s},
    { SWIG_prefix "rig_callbacks_freq_event_set", (swig_wrapper_func) _wrap_rig_callbacks_freq_event_set, NULL},
    { SWIG_prefix "rig_callbacks_freq_event_get", (swig_wrapper_func) _wrap_rig_callbacks_freq_event_get, NULL},
    { SWIG_prefix "rig_callbacks_freq_arg_set", (swig_wrapper_func) _wrap_rig_callbacks_freq_arg_set, NULL},
    { SWIG_prefix "rig_callbacks_freq_arg_get", (swig_wrapper_func) _wrap_rig_callbacks_freq_arg_get, NULL},
    { SWIG_prefix "rig_callbacks_mode_event_set", (swig_wrapper_func) _wrap_rig_callbacks_mode_event_set, NULL},
    { SWIG_prefix "rig_callbacks_mode_event_get", (swig_wrapper_func) _wrap_rig_callbacks_mode_event_get, NULL},
    { SWIG_prefix "rig_callbacks_mode_arg_set", (swig_wrapper_func) _wrap_rig_callbacks_mode_arg_set, NULL},
    { SWIG_prefix "rig_callbacks_mode_arg_get", (swig_wrapper_func) _wrap_rig_callbacks_mode_arg_get, NULL},
    { SWIG_prefix "rig_callbacks_vfo_event_set", (swig_wrapper_func) _wrap_rig_callbacks_vfo_event_set, NULL},
    { SWIG_prefix "rig_callbacks_vfo_event_get", (swig_wrapper_func) _wrap_rig_callbacks_vfo_event_get, NULL},
    { SWIG_prefix "rig_callbacks_vfo_arg_set", (swig_wrapper_func) _wrap_rig_callbacks_vfo_arg_set, NULL},
    { SWIG_prefix "rig_callbacks_vfo_arg_get", (swig_wrapper_func) _wrap_rig_callbacks_vfo_arg_get, NULL},
    { SWIG_prefix "rig_callbacks_ptt_event_set", (swig_wrapper_func) _wrap_rig_callbacks_ptt_event_set, NULL},
    { SWIG_prefix "rig_callbacks_ptt_event_get", (swig_wrapper_func) _wrap_rig_callbacks_ptt_event_get, NULL},
    { SWIG_prefix "rig_callbacks_ptt_arg_set", (swig_wrapper_func) _wrap_rig_callbacks_ptt_arg_set, NULL},
    { SWIG_prefix "rig_callbacks_ptt_arg_get", (swig_wrapper_func) _wrap_rig_callbacks_ptt_arg_get, NULL},
    { SWIG_prefix "rig_callbacks_dcd_event_set", (swig_wrapper_func) _wrap_rig_callbacks_dcd_event_set, NULL},
    { SWIG_prefix "rig_callbacks_dcd_event_get", (swig_wrapper_func) _wrap_rig_callbacks_dcd_event_get, NULL},
    { SWIG_prefix "rig_callbacks_dcd_arg_set", (swig_wrapper_func) _wrap_rig_callbacks_dcd_arg_set, NULL},
    { SWIG_prefix "rig_callbacks_dcd_arg_get", (swig_wrapper_func) _wrap_rig_callbacks_dcd_arg_get, NULL},
    { SWIG_prefix "rig_callbacks_pltune_set", (swig_wrapper_func) _wrap_rig_callbacks_pltune_set, NULL},
    { SWIG_prefix "rig_callbacks_pltune_get", (swig_wrapper_func) _wrap_rig_callbacks_pltune_get, NULL},
    { SWIG_prefix "rig_callbacks_pltune_arg_set", (swig_wrapper_func) _wrap_rig_callbacks_pltune_arg_set, NULL},
    { SWIG_prefix "rig_callbacks_pltune_arg_get", (swig_wrapper_func) _wrap_rig_callbacks_pltune_arg_get, NULL},
    { SWIG_prefix "rig_callbacks_spectrum_event_set", (swig_wrapper_func) _wrap_rig_callbacks_spectrum_event_set, NULL},
    { SWIG_prefix "rig_callbacks_spectrum_event_get", (swig_wrapper_func) _wrap_rig_callbacks_spectrum_event_get, NULL},
    { SWIG_prefix "rig_callbacks_spectrum_arg_set", (swig_wrapper_func) _wrap_rig_callbacks_spectrum_arg_set, NULL},
    { SWIG_prefix "rig_callbacks_spectrum_arg_get", (swig_wrapper_func) _wrap_rig_callbacks_spectrum_arg_get, NULL},
    { SWIG_prefix "new_rig_callbacks", (swig_wrapper_func) _wrap_new_rig_callbacks, NULL},
    { SWIG_prefix "delete_rig_callbacks", (swig_wrapper_func) _wrap_delete_rig_callbacks, NULL},
    { SWIG_prefix "rig_callbacks", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_rig_callbacks},
    { SWIG_prefix "netrigctl_get_vfo_mode", (swig_wrapper_func) _wrap_netrigctl_get_vfo_mode, NULL},
    { SWIG_prefix "rigerror", (swig_wrapper_func) _wrap_rigerror, NULL},
    { SWIG_prefix "rigerror2", (swig_wrapper_func) _wrap_rigerror2, NULL},
    { SWIG_prefix "rig_set_debug", (swig_wrapper_func) _wrap_rig_set_debug, NULL},
    { SWIG_prefix "rig_get_debug", (swig_wrapper_func) _wrap_rig_get_debug, NULL},
    { SWIG_prefix "rig_set_debug_time_stamp", (swig_wrapper_func) _wrap_rig_set_debug_time_stamp, NULL},
    { SWIG_prefix "add2debugmsgsave", (swig_wrapper_func) _wrap_add2debugmsgsave, NULL},
    { SWIG_prefix "rig_strrmode", (swig_wrapper_func) _wrap_rig_strrmode, NULL},
    { SWIG_prefix "rig_strrmodes", (swig_wrapper_func) _wrap_rig_strrmodes, NULL},
    { SWIG_prefix "rig_strvfo", (swig_wrapper_func) _wrap_rig_strvfo, NULL},
    { SWIG_prefix "rig_strfunc", (swig_wrapper_func) _wrap_rig_strfunc, NULL},
    { SWIG_prefix "rig_strlevel", (swig_wrapper_func) _wrap_rig_strlevel, NULL},
    { SWIG_prefix "rig_strparm", (swig_wrapper_func) _wrap_rig_strparm, NULL},
    { SWIG_prefix "rig_stragclevel", (swig_wrapper_func) _wrap_rig_stragclevel, NULL},
    { SWIG_prefix "rig_strptrshift", (swig_wrapper_func) _wrap_rig_strptrshift, NULL},
    { SWIG_prefix "rig_strvfop", (swig_wrapper_func) _wrap_rig_strvfop, NULL},
    { SWIG_prefix "rig_strscan", (swig_wrapper_func) _wrap_rig_strscan, NULL},
    { SWIG_prefix "rig_strstatus", (swig_wrapper_func) _wrap_rig_strstatus, NULL},
    { SWIG_prefix "rig_strmtype", (swig_wrapper_func) _wrap_rig_strmtype, NULL},
    { SWIG_prefix "rig_strspectrummode", (swig_wrapper_func) _wrap_rig_strspectrummode, NULL},
    { SWIG_prefix "rig_strcommstatus", (swig_wrapper_func) _wrap_rig_strcommstatus, NULL},
    { SWIG_prefix "hl_usleep", (swig_wrapper_func) _wrap_hl_usleep, NULL},
    { SWIG_prefix "longlat2locator", (swig_wrapper_func) _wrap_longlat2locator, NULL},
    { SWIG_prefix "locator2longlat", (swig_wrapper_func) _wrap_locator2longlat, NULL},
    { SWIG_prefix "rig_state_vfo_comp_set", (swig_wrapper_func) _wrap_rig_state_vfo_comp_set, NULL},
    { SWIG_prefix "rig_state_vfo_comp_get", (swig_wrapper_func) _wrap_rig_state_vfo_comp_get, NULL},
    { SWIG_prefix "rig_state_rx_range_list_set", (swig_wrapper_func) _wrap_rig_state_rx_range_list_set, NULL},
    { SWIG_prefix "rig_state_rx_range_list_get", (swig_wrapper_func) _wrap_rig_state_rx_range_list_get, NULL},
    { SWIG_prefix "rig_state_tx_range_list_set", (swig_wrapper_func) _wrap_rig_state_tx_range_list_set, NULL},
    { SWIG_prefix "rig_state_tx_range_list_get", (swig_wrapper_func) _wrap_rig_state_tx_range_list_get, NULL},
    { SWIG_prefix "rig_state_tuning_steps_set", (swig_wrapper_func) _wrap_rig_state_tuning_steps_set, NULL},
    { SWIG_prefix "rig_state_tuning_steps_get", (swig_wrapper_func) _wrap_rig_state_tuning_steps_get, NULL},
    { SWIG_prefix "rig_state_filters_set", (swig_wrapper_func) _wrap_rig_state_filters_set, NULL},
    { SWIG_prefix "rig_state_filters_get", (swig_wrapper_func) _wrap_rig_state_filters_get, NULL},
    { SWIG_prefix "rig_state_str_cal_set", (swig_wrapper_func) _wrap_rig_state_str_cal_set, NULL},
    { SWIG_prefix "rig_state_str_cal_get", (swig_wrapper_func) _wrap_rig_state_str_cal_get, NULL},
    { SWIG_prefix "rig_state_chan_list_set", (swig_wrapper_func) _wrap_rig_state_chan_list_set, NULL},
    { SWIG_prefix "rig_state_chan_list_get", (swig_wrapper_func) _wrap_rig_state_chan_list_get, NULL},
    { SWIG_prefix "rig_state_max_rit_set", (swig_wrapper_func) _wrap_rig_state_max_rit_set, NULL},
    { SWIG_prefix "rig_state_max_rit_get", (swig_wrapper_func) _wrap_rig_state_max_rit_get, NULL},
    { SWIG_prefix "rig_state_max_xit_set", (swig_wrapper_func) _wrap_rig_state_max_xit_set, NULL},
    { SWIG_prefix "rig_state_max_xit_get", (swig_wrapper_func) _wrap_rig_state_max_xit_get, NULL},
    { SWIG_prefix "rig_state_max_ifshift_set", (swig_wrapper_func) _wrap_rig_state_max_ifshift_set, NULL},
    { SWIG_prefix "rig_state_max_ifshift_get", (swig_wrapper_func) _wrap_rig_state_max_ifshift_get, NULL},
    { SWIG_prefix "rig_state_announces_set", (swig_wrapper_func) _wrap_rig_state_announces_set, NULL},
    { SWIG_prefix "rig_state_announces_get", (swig_wrapper_func) _wrap_rig_state_announces_get, NULL},
    { SWIG_prefix "rig_state_preamp_set", (swig_wrapper_func) _wrap_rig_state_preamp_set, NULL},
    { SWIG_prefix "rig_state_preamp_get", (swig_wrapper_func) _wrap_rig_state_preamp_get, NULL},
    { SWIG_prefix "rig_state_attenuator_set", (swig_wrapper_func) _wrap_rig_state_attenuator_set, NULL},
    { SWIG_prefix "rig_state_attenuator_get", (swig_wrapper_func) _wrap_rig_state_attenuator_get, NULL},
    { SWIG_prefix "rig_state_has_get_func_set", (swig_wrapper_func) _wrap_rig_state_has_get_func_set, NULL},
    { SWIG_prefix "rig_state_has_get_func_get", (swig_wrapper_func) _wrap_rig_state_has_get_func_get, NULL},
    { SWIG_prefix "rig_state_has_set_func_set", (swig_wrapper_func) _wrap_rig_state_has_set_func_set, NULL},
    { SWIG_prefix "rig_state_has_set_func_get", (swig_wrapper_func) _wrap_rig_state_has_set_func_get, NULL},
    { SWIG_prefix "rig_state_has_get_level_set", (swig_wrapper_func) _wrap_rig_state_has_get_level_set, NULL},
    { SWIG_prefix "rig_state_has_get_level_get", (swig_wrapper_func) _wrap_rig_state_has_get_level_get, NULL},
    { SWIG_prefix "rig_state_has_set_level_set", (swig_wrapper_func) _wrap_rig_state_has_set_level_set, NULL},
    { SWIG_prefix "rig_state_has_set_level_get", (swig_wrapper_func) _wrap_rig_state_has_set_level_get, NULL},
    { SWIG_prefix "rig_state_has_get_parm_set", (swig_wrapper_func) _wrap_rig_state_has_get_parm_set, NULL},
    { SWIG_prefix "rig_state_has_get_parm_get", (swig_wrapper_func) _wrap_rig_state_has_get_parm_get, NULL},
    { SWIG_prefix "rig_state_has_set_parm_set", (swig_wrapper_func) _wrap_rig_state_has_set_parm_set, NULL},
    { SWIG_prefix "rig_state_has_set_parm_get", (swig_wrapper_func) _wrap_rig_state_has_set_parm_get, NULL},
    { SWIG_prefix "rig_state_level_gran_set", (swig_wrapper_func) _wrap_rig_state_level_gran_set, NULL},
    { SWIG_prefix "rig_state_level_gran_get", (swig_wrapper_func) _wrap_rig_state_level_gran_get, NULL},
    { SWIG_prefix "rig_state_parm_gran_set", (swig_wrapper_func) _wrap_rig_state_parm_gran_set, NULL},
    { SWIG_prefix "rig_state_parm_gran_get", (swig_wrapper_func) _wrap_rig_state_parm_gran_get, NULL},
    { SWIG_prefix "rig_state_transaction_active_set", (swig_wrapper_func) _wrap_rig_state_transaction_active_set, NULL},
    { SWIG_prefix "rig_state_transaction_active_get", (swig_wrapper_func) _wrap_rig_state_transaction_active_get, NULL},
    { SWIG_prefix "rig_state_current_vfo_set", (swig_wrapper_func) _wrap_rig_state_current_vfo_set, NULL},
    { SWIG_prefix "rig_state_current_vfo_get", (swig_wrapper_func) _wrap_rig_state_current_vfo_get, NULL},
    { SWIG_prefix "rig_state_vfo_list_set", (swig_wrapper_func) _wrap_rig_state_vfo_list_set, NULL},
    { SWIG_prefix "rig_state_vfo_list_get", (swig_wrapper_func) _wrap_rig_state_vfo_list_get, NULL},
    { SWIG_prefix "rig_state_comm_state_set", (swig_wrapper_func) _wrap_rig_state_comm_state_set, NULL},
    { SWIG_prefix "rig_state_comm_state_get", (swig_wrapper_func) _wrap_rig_state_comm_state_get, NULL},
    { SWIG_prefix "rig_state_priv_set", (swig_wrapper_func) _wrap_rig_state_priv_set, NULL},
    { SWIG_prefix "rig_state_priv_get", (swig_wrapper_func) _wrap_rig_state_priv_get, NULL},
    { SWIG_prefix "rig_state_obj_set", (swig_wrapper_func) _wrap_rig_state_obj_set, NULL},
    { SWIG_prefix "rig_state_obj_get", (swig_wrapper_func) _wrap_rig_state_obj_get, NULL},
    { SWIG_prefix "rig_state_async_data_enabled_set", (swig_wrapper_func) _wrap_rig_state_async_data_enabled_set, NULL},
    { SWIG_prefix "rig_state_async_data_enabled_get", (swig_wrapper_func) _wrap_rig_state_async_data_enabled_get, NULL},
    { SWIG_prefix "rig_state_poll_interval_set", (swig_wrapper_func) _wrap_rig_state_poll_interval_set, NULL},
    { SWIG_prefix "rig_state_poll_interval_get", (swig_wrapper_func) _wrap_rig_state_poll_interval_get, NULL},
    { SWIG_prefix "rig_state_current_freq_set", (swig_wrapper_func) _wrap_rig_state_current_freq_set, NULL},
    { SWIG_prefix "rig_state_current_freq_get", (swig_wrapper_func) _wrap_rig_state_current_freq_get, NULL},
    { SWIG_prefix "rig_state_current_mode_set", (swig_wrapper_func) _wrap_rig_state_current_mode_set, NULL},
    { SWIG_prefix "rig_state_current_mode_get", (swig_wrapper_func) _wrap_rig_state_current_mode_get, NULL},
    { SWIG_prefix "rig_state_current_width_set", (swig_wrapper_func) _wrap_rig_state_current_width_set, NULL},
    { SWIG_prefix "rig_state_current_width_get", (swig_wrapper_func) _wrap_rig_state_current_width_get, NULL},
    { SWIG_prefix "rig_state_tx_vfo_set", (swig_wrapper_func) _wrap_rig_state_tx_vfo_set, NULL},
    { SWIG_prefix "rig_state_tx_vfo_get", (swig_wrapper_func) _wrap_rig_state_tx_vfo_get, NULL},
    { SWIG_prefix "rig_state_mode_list_set", (swig_wrapper_func) _wrap_rig_state_mode_list_set, NULL},
    { SWIG_prefix "rig_state_mode_list_get", (swig_wrapper_func) _wrap_rig_state_mode_list_get, NULL},
    { SWIG_prefix "rig_state_transmit_set", (swig_wrapper_func) _wrap_rig_state_transmit_set, NULL},
    { SWIG_prefix "rig_state_transmit_get", (swig_wrapper_func) _wrap_rig_state_transmit_get, NULL},
    { SWIG_prefix "rig_state_lo_freq_set", (swig_wrapper_func) _wrap_rig_state_lo_freq_set, NULL},
    { SWIG_prefix "rig_state_lo_freq_get", (swig_wrapper_func) _wrap_rig_state_lo_freq_get, NULL},
    { SWIG_prefix "rig_state_twiddle_time_set", (swig_wrapper_func) _wrap_rig_state_twiddle_time_set, NULL},
    { SWIG_prefix "rig_state_twiddle_time_get", (swig_wrapper_func) _wrap_rig_state_twiddle_time_get, NULL},
    { SWIG_prefix "rig_state_twiddle_timeout_set", (swig_wrapper_func) _wrap_rig_state_twiddle_timeout_set, NULL},
    { SWIG_prefix "rig_state_twiddle_timeout_get", (swig_wrapper_func) _wrap_rig_state_twiddle_timeout_get, NULL},
    { SWIG_prefix "rig_state_uplink_set", (swig_wrapper_func) _wrap_rig_state_uplink_set, NULL},
    { SWIG_prefix "rig_state_uplink_get", (swig_wrapper_func) _wrap_rig_state_uplink_get, NULL},
    { SWIG_prefix "rig_state_vfo_opt_set", (swig_wrapper_func) _wrap_rig_state_vfo_opt_set, NULL},
    { SWIG_prefix "rig_state_vfo_opt_get", (swig_wrapper_func) _wrap_rig_state_vfo_opt_get, NULL},
    { SWIG_prefix "rig_state_auto_power_on_set", (swig_wrapper_func) _wrap_rig_state_auto_power_on_set, NULL},
    { SWIG_prefix "rig_state_auto_power_on_get", (swig_wrapper_func) _wrap_rig_state_auto_power_on_get, NULL},
    { SWIG_prefix "rig_state_auto_power_off_set", (swig_wrapper_func) _wrap_rig_state_auto_power_off_set, NULL},
    { SWIG_prefix "rig_state_auto_power_off_get", (swig_wrapper_func) _wrap_rig_state_auto_power_off_get, NULL},
    { SWIG_prefix "rig_state_auto_disable_screensaver_set", (swig_wrapper_func) _wrap_rig_state_auto_disable_screensaver_set, NULL},
    { SWIG_prefix "rig_state_auto_disable_screensaver_get", (swig_wrapper_func) _wrap_rig_state_auto_disable_screensaver_get, NULL},
    { SWIG_prefix "rig_state_ptt_share_set", (swig_wrapper_func) _wrap_rig_state_ptt_share_set, NULL},
    { SWIG_prefix "rig_state_ptt_share_get", (swig_wrapper_func) _wrap_rig_state_ptt_share_get, NULL},
    { SWIG_prefix "rig_state_power_now_set", (swig_wrapper_func) _wrap_rig_state_power_now_set, NULL},
    { SWIG_prefix "rig_state_power_now_get", (swig_wrapper_func) _wrap_rig_state_power_now_get, NULL},
    { SWIG_prefix "rig_state_power_min_set", (swig_wrapper_func) _wrap_rig_state_power_min_set, NULL},
    { SWIG_prefix "rig_state_power_min_get", (swig_wrapper_func) _wrap_rig_state_power_min_get, NULL},
    { SWIG_prefix "rig_state_power_max_set", (swig_wrapper_func) _wrap_rig_state_power_max_set, NULL},
    { SWIG_prefix "rig_state_power_max_get", (swig_wrapper_func) _wrap_rig_state_power_max_get, NULL},
    { SWIG_prefix "rig_state_disable_yaesu_bandselect_set", (swig_wrapper_func) _wrap_rig_state_disable_yaesu_bandselect_set, NULL},
    { SWIG_prefix "rig_state_disable_yaesu_bandselect_get", (swig_wrapper_func) _wrap_rig_state_disable_yaesu_bandselect_get, NULL},
    { SWIG_prefix "rig_state_twiddle_rit_set", (swig_wrapper_func) _wrap_rig_state_twiddle_rit_set, NULL},
    { SWIG_prefix "rig_state_twiddle_rit_get", (swig_wrapper_func) _wrap_rig_state_twiddle_rit_get, NULL},
    { SWIG_prefix "rig_state_twiddle_state_set", (swig_wrapper_func) _wrap_rig_state_twiddle_state_set, NULL},
    { SWIG_prefix "rig_state_twiddle_state_get", (swig_wrapper_func) _wrap_rig_state_twiddle_state_get, NULL},
    { SWIG_prefix "rig_state_rx_vfo_set", (swig_wrapper_func) _wrap_rig_state_rx_vfo_set, NULL},
    { SWIG_prefix "rig_state_rx_vfo_get", (swig_wrapper_func) _wrap_rig_state_rx_vfo_get, NULL},
    { SWIG_prefix "rig_state_snapshot_packet_sequence_number_set", (swig_wrapper_func) _wrap_rig_state_snapshot_packet_sequence_number_set, NULL},
    { SWIG_prefix "rig_state_snapshot_packet_sequence_number_get", (swig_wrapper_func) _wrap_rig_state_snapshot_packet_sequence_number_get, NULL},
    { SWIG_prefix "rig_state_multicast_publisher_run_set", (swig_wrapper_func) _wrap_rig_state_multicast_publisher_run_set, NULL},
    { SWIG_prefix "rig_state_multicast_publisher_run_get", (swig_wrapper_func) _wrap_rig_state_multicast_publisher_run_get, NULL},
    { SWIG_prefix "rig_state_multicast_publisher_priv_data_set", (swig_wrapper_func) _wrap_rig_state_multicast_publisher_priv_data_set, NULL},
    { SWIG_prefix "rig_state_multicast_publisher_priv_data_get", (swig_wrapper_func) _wrap_rig_state_multicast_publisher_priv_data_get, NULL},
    { SWIG_prefix "rig_state_async_data_handler_thread_run_set", (swig_wrapper_func) _wrap_rig_state_async_data_handler_thread_run_set, NULL},
    { SWIG_prefix "rig_state_async_data_handler_thread_run_get", (swig_wrapper_func) _wrap_rig_state_async_data_handler_thread_run_get, NULL},
    { SWIG_prefix "rig_state_async_data_handler_priv_data_set", (swig_wrapper_func) _wrap_rig_state_async_data_handler_priv_data_set, NULL},
    { SWIG_prefix "rig_state_async_data_handler_priv_data_get", (swig_wrapper_func) _wrap_rig_state_async_data_handler_priv_data_get, NULL},
    { SWIG_prefix "rig_state_poll_routine_thread_run_set", (swig_wrapper_func) _wrap_rig_state_poll_routine_thread_run_set, NULL},
    { SWIG_prefix "rig_state_poll_routine_thread_run_get", (swig_wrapper_func) _wrap_rig_state_poll_routine_thread_run_get, NULL},
    { SWIG_prefix "rig_state_poll_routine_priv_data_set", (swig_wrapper_func) _wrap_rig_state_poll_routine_priv_data_set, NULL},
    { SWIG_prefix "rig_state_poll_routine_priv_data_get", (swig_wrapper_func) _wrap_rig_state_poll_routine_priv_data_get, NULL},
    { SWIG_prefix "rig_state_mutex_set_transaction_set", (swig_wrapper_func) _wrap_rig_state_mutex_set_transaction_set, NULL},
    { SWIG_prefix "rig_state_mutex_set_transaction_get", (swig_wrapper_func) _wrap_rig_state_mutex_set_transaction_get, NULL},
    { SWIG_prefix "rig_state_rigport_set", (swig_wrapper_func) _wrap_rig_state_rigport_set, NULL},
    { SWIG_prefix "rig_state_rigport_get", (swig_wrapper_func) _wrap_rig_state_rigport_get, NULL},
    { SWIG_prefix "rig_state_pttport_set", (swig_wrapper_func) _wrap_rig_state_pttport_set, NULL},
    { SWIG_prefix "rig_state_pttport_get", (swig_wrapper_func) _wrap_rig_state_pttport_get, NULL},
    { SWIG_prefix "rig_state_dcdport_set", (swig_wrapper_func) _wrap_rig_state_dcdport_set, NULL},
    { SWIG_prefix "rig_state_dcdport_get", (swig_wrapper_func) _wrap_rig_state_dcdport_get, NULL},
    { SWIG_prefix "rig_state_use_cached_freq_set", (swig_wrapper_func) _wrap_rig_state_use_cached_freq_set, NULL},
    { SWIG_prefix "rig_state_use_cached_freq_get", (swig_wrapper_func) _wrap_rig_state_use_cached_freq_get, NULL},
    { SWIG_prefix "rig_state_use_cached_mode_set", (swig_wrapper_func) _wrap_rig_state_use_cached_mode_set, NULL},
    { SWIG_prefix "rig_state_use_cached_mode_get", (swig_wrapper_func) _wrap_rig_state_use_cached_mode_get, NULL},
    { SWIG_prefix "rig_state_use_cached_ptt_set", (swig_wrapper_func) _wrap_rig_state_use_cached_ptt_set, NULL},
    { SWIG_prefix "rig_state_use_cached_ptt_get", (swig_wrapper_func) _wrap_rig_state_use_cached_ptt_get, NULL},
    { SWIG_prefix "rig_state_depth_set", (swig_wrapper_func) _wrap_rig_state_depth_set, NULL},
    { SWIG_prefix "rig_state_depth_get", (swig_wrapper_func) _wrap_rig_state_depth_get, NULL},
    { SWIG_prefix "rig_state_lock_mode_set", (swig_wrapper_func) _wrap_rig_state_lock_mode_set, NULL},
    { SWIG_prefix "rig_state_lock_mode_get", (swig_wrapper_func) _wrap_rig_state_lock_mode_get, NULL},
    { SWIG_prefix "rig_state_powerstat_set", (swig_wrapper_func) _wrap_rig_state_powerstat_set, NULL},
    { SWIG_prefix "rig_state_powerstat_get", (swig_wrapper_func) _wrap_rig_state_powerstat_get, NULL},
    { SWIG_prefix "rig_state_tuner_control_pathname_set", (swig_wrapper_func) _wrap_rig_state_tuner_control_pathname_set, NULL},
    { SWIG_prefix "rig_state_tuner_control_pathname_get", (swig_wrapper_func) _wrap_rig_state_tuner_control_pathname_get, NULL},
    { SWIG_prefix "rig_state_client_version_set", (swig_wrapper_func) _wrap_rig_state_client_version_set, NULL},
    { SWIG_prefix "rig_state_client_version_get", (swig_wrapper_func) _wrap_rig_state_client_version_get, NULL},
    { SWIG_prefix "rig_state_offset_vfoa_set", (swig_wrapper_func) _wrap_rig_state_offset_vfoa_set, NULL},
    { SWIG_prefix "rig_state_offset_vfoa_get", (swig_wrapper_func) _wrap_rig_state_offset_vfoa_get, NULL},
    { SWIG_prefix "rig_state_offset_vfob_set", (swig_wrapper_func) _wrap_rig_state_offset_vfob_set, NULL},
    { SWIG_prefix "rig_state_offset_vfob_get", (swig_wrapper_func) _wrap_rig_state_offset_vfob_get, NULL},
    { SWIG_prefix "rig_state_multicast_set", (swig_wrapper_func) _wrap_rig_state_multicast_set, NULL},
    { SWIG_prefix "rig_state_multicast_get", (swig_wrapper_func) _wrap_rig_state_multicast_get, NULL},
    { SWIG_prefix "rig_state_rig_model_set", (swig_wrapper_func) _wrap_rig_state_rig_model_set, NULL},
    { SWIG_prefix "rig_state_rig_model_get", (swig_wrapper_func) _wrap_rig_state_rig_model_get, NULL},
    { SWIG_prefix "rig_state_model_name_set", (swig_wrapper_func) _wrap_rig_state_model_name_set, NULL},
    { SWIG_prefix "rig_state_model_name_get", (swig_wrapper_func) _wrap_rig_state_model_name_get, NULL},
    { SWIG_prefix "rig_state_mfg_name_set", (swig_wrapper_func) _wrap_rig_state_mfg_name_set, NULL},
    { SWIG_prefix "rig_state_mfg_name_get", (swig_wrapper_func) _wrap_rig_state_mfg_name_get, NULL},
    { SWIG_prefix "rig_state_version_set", (swig_wrapper_func) _wrap_rig_state_version_set, NULL},
    { SWIG_prefix "rig_state_version_get", (swig_wrapper_func) _wrap_rig_state_version_get, NULL},
    { SWIG_prefix "rig_state_copyright_set", (swig_wrapper_func) _wrap_rig_state_copyright_set, NULL},
    { SWIG_prefix "rig_state_copyright_get", (swig_wrapper_func) _wrap_rig_state_copyright_get, NULL},
    { SWIG_prefix "rig_state_status_set", (swig_wrapper_func) _wrap_rig_state_status_set, NULL},
    { SWIG_prefix "rig_state_status_get", (swig_wrapper_func) _wrap_rig_state_status_get, NULL},
    { SWIG_prefix "rig_state_rig_type_set", (swig_wrapper_func) _wrap_rig_state_rig_type_set, NULL},
    { SWIG_prefix "rig_state_rig_type_get", (swig_wrapper_func) _wrap_rig_state_rig_type_get, NULL},
    { SWIG_prefix "rig_state_ptt_type_set", (swig_wrapper_func) _wrap_rig_state_ptt_type_set, NULL},
    { SWIG_prefix "rig_state_ptt_type_get", (swig_wrapper_func) _wrap_rig_state_ptt_type_get, NULL},
    { SWIG_prefix "rig_state_dcd_type_set", (swig_wrapper_func) _wrap_rig_state_dcd_type_set, NULL},
    { SWIG_prefix "rig_state_dcd_type_get", (swig_wrapper_func) _wrap_rig_state_dcd_type_get, NULL},
    { SWIG_prefix "rig_state_port_type_set", (swig_wrapper_func) _wrap_rig_state_port_type_set, NULL},
    { SWIG_prefix "rig_state_port_type_get", (swig_wrapper_func) _wrap_rig_state_port_type_get, NULL},
    { SWIG_prefix "rig_state_serial_rate_min_set", (swig_wrapper_func) _wrap_rig_state_serial_rate_min_set, NULL},
    { SWIG_prefix "rig_state_serial_rate_min_get", (swig_wrapper_func) _wrap_rig_state_serial_rate_min_get, NULL},
    { SWIG_prefix "rig_state_serial_rate_max_set", (swig_wrapper_func) _wrap_rig_state_serial_rate_max_set, NULL},
    { SWIG_prefix "rig_state_serial_rate_max_get", (swig_wrapper_func) _wrap_rig_state_serial_rate_max_get, NULL},
    { SWIG_prefix "rig_state_serial_data_bits_set", (swig_wrapper_func) _wrap_rig_state_serial_data_bits_set, NULL},
    { SWIG_prefix "rig_state_serial_data_bits_get", (swig_wrapper_func) _wrap_rig_state_serial_data_bits_get, NULL},
    { SWIG_prefix "rig_state_serial_stop_bits_set", (swig_wrapper_func) _wrap_rig_state_serial_stop_bits_set, NULL},
    { SWIG_prefix "rig_state_serial_stop_bits_get", (swig_wrapper_func) _wrap_rig_state_serial_stop_bits_get, NULL},
    { SWIG_prefix "rig_state_serial_parity_set", (swig_wrapper_func) _wrap_rig_state_serial_parity_set, NULL},
    { SWIG_prefix "rig_state_serial_parity_get", (swig_wrapper_func) _wrap_rig_state_serial_parity_get, NULL},
    { SWIG_prefix "rig_state_serial_handshake_set", (swig_wrapper_func) _wrap_rig_state_serial_handshake_set, NULL},
    { SWIG_prefix "rig_state_serial_handshake_get", (swig_wrapper_func) _wrap_rig_state_serial_handshake_get, NULL},
    { SWIG_prefix "rig_state_write_delay_set", (swig_wrapper_func) _wrap_rig_state_write_delay_set, NULL},
    { SWIG_prefix "rig_state_write_delay_get", (swig_wrapper_func) _wrap_rig_state_write_delay_get, NULL},
    { SWIG_prefix "rig_state_post_write_delay_set", (swig_wrapper_func) _wrap_rig_state_post_write_delay_set, NULL},
    { SWIG_prefix "rig_state_post_write_delay_get", (swig_wrapper_func) _wrap_rig_state_post_write_delay_get, NULL},
    { SWIG_prefix "rig_state_timeout_set", (swig_wrapper_func) _wrap_rig_state_timeout_set, NULL},
    { SWIG_prefix "rig_state_timeout_get", (swig_wrapper_func) _wrap_rig_state_timeout_get, NULL},
    { SWIG_prefix "rig_state_retry_set", (swig_wrapper_func) _wrap_rig_state_retry_set, NULL},
    { SWIG_prefix "rig_state_retry_get", (swig_wrapper_func) _wrap_rig_state_retry_get, NULL},
    { SWIG_prefix "rig_state_targetable_vfo_set", (swig_wrapper_func) _wrap_rig_state_targetable_vfo_set, NULL},
    { SWIG_prefix "rig_state_targetable_vfo_get", (swig_wrapper_func) _wrap_rig_state_targetable_vfo_get, NULL},
    { SWIG_prefix "rig_state_async_data_supported_set", (swig_wrapper_func) _wrap_rig_state_async_data_supported_set, NULL},
    { SWIG_prefix "rig_state_async_data_supported_get", (swig_wrapper_func) _wrap_rig_state_async_data_supported_get, NULL},
    { SWIG_prefix "rig_state_agc_level_count_set", (swig_wrapper_func) _wrap_rig_state_agc_level_count_set, NULL},
    { SWIG_prefix "rig_state_agc_level_count_get", (swig_wrapper_func) _wrap_rig_state_agc_level_count_get, NULL},
    { SWIG_prefix "rig_state_agc_levels_set", (swig_wrapper_func) _wrap_rig_state_agc_levels_set, NULL},
    { SWIG_prefix "rig_state_agc_levels_get", (swig_wrapper_func) _wrap_rig_state_agc_levels_get, NULL},
    { SWIG_prefix "rig_state_ctcss_list_set", (swig_wrapper_func) _wrap_rig_state_ctcss_list_set, NULL},
    { SWIG_prefix "rig_state_ctcss_list_get", (swig_wrapper_func) _wrap_rig_state_ctcss_list_get, NULL},
    { SWIG_prefix "rig_state_dcs_list_set", (swig_wrapper_func) _wrap_rig_state_dcs_list_set, NULL},
    { SWIG_prefix "rig_state_dcs_list_get", (swig_wrapper_func) _wrap_rig_state_dcs_list_get, NULL},
    { SWIG_prefix "rig_state_vfo_ops_set", (swig_wrapper_func) _wrap_rig_state_vfo_ops_set, NULL},
    { SWIG_prefix "rig_state_vfo_ops_get", (swig_wrapper_func) _wrap_rig_state_vfo_ops_get, NULL},
    { SWIG_prefix "rig_state_scan_ops_set", (swig_wrapper_func) _wrap_rig_state_scan_ops_set, NULL},
    { SWIG_prefix "rig_state_scan_ops_get", (swig_wrapper_func) _wrap_rig_state_scan_ops_get, NULL},
    { SWIG_prefix "rig_state_transceive_set", (swig_wrapper_func) _wrap_rig_state_transceive_set, NULL},
    { SWIG_prefix "rig_state_transceive_get", (swig_wrapper_func) _wrap_rig_state_transceive_get, NULL},
    { SWIG_prefix "rig_state_bank_qty_set", (swig_wrapper_func) _wrap_rig_state_bank_qty_set, NULL},
    { SWIG_prefix "rig_state_bank_qty_get", (swig_wrapper_func) _wrap_rig_state_bank_qty_get, NULL},
    { SWIG_prefix "rig_state_chan_desc_sz_set", (swig_wrapper_func) _wrap_rig_state_chan_desc_sz_set, NULL},
    { SWIG_prefix "rig_state_chan_desc_sz_get", (swig_wrapper_func) _wrap_rig_state_chan_desc_sz_get, NULL},
    { SWIG_prefix "rig_state_rx_range_list1_set", (swig_wrapper_func) _wrap_rig_state_rx_range_list1_set, NULL},
    { SWIG_prefix "rig_state_rx_range_list1_get", (swig_wrapper_func) _wrap_rig_state_rx_range_list1_get, NULL},
    { SWIG_prefix "rig_state_tx_range_list1_set", (swig_wrapper_func) _wrap_rig_state_tx_range_list1_set, NULL},
    { SWIG_prefix "rig_state_tx_range_list1_get", (swig_wrapper_func) _wrap_rig_state_tx_range_list1_get, NULL},
    { SWIG_prefix "rig_state_rx_range_list2_set", (swig_wrapper_func) _wrap_rig_state_rx_range_list2_set, NULL},
    { SWIG_prefix "rig_state_rx_range_list2_get", (swig_wrapper_func) _wrap_rig_state_rx_range_list2_get, NULL},
    { SWIG_prefix "rig_state_tx_range_list2_set", (swig_wrapper_func) _wrap_rig_state_tx_range_list2_set, NULL},
    { SWIG_prefix "rig_state_tx_range_list2_get", (swig_wrapper_func) _wrap_rig_state_tx_range_list2_get, NULL},
    { SWIG_prefix "rig_state_rx_range_list3_set", (swig_wrapper_func) _wrap_rig_state_rx_range_list3_set, NULL},
    { SWIG_prefix "rig_state_rx_range_list3_get", (swig_wrapper_func) _wrap_rig_state_rx_range_list3_get, NULL},
    { SWIG_prefix "rig_state_tx_range_list3_set", (swig_wrapper_func) _wrap_rig_state_tx_range_list3_set, NULL},
    { SWIG_prefix "rig_state_tx_range_list3_get", (swig_wrapper_func) _wrap_rig_state_tx_range_list3_get, NULL},
    { SWIG_prefix "rig_state_rx_range_list4_set", (swig_wrapper_func) _wrap_rig_state_rx_range_list4_set, NULL},
    { SWIG_prefix "rig_state_rx_range_list4_get", (swig_wrapper_func) _wrap_rig_state_rx_range_list4_get, NULL},
    { SWIG_prefix "rig_state_tx_range_list4_set", (swig_wrapper_func) _wrap_rig_state_tx_range_list4_set, NULL},
    { SWIG_prefix "rig_state_tx_range_list4_get", (swig_wrapper_func) _wrap_rig_state_tx_range_list4_get, NULL},
    { SWIG_prefix "rig_state_rx_range_list5_set", (swig_wrapper_func) _wrap_rig_state_rx_range_list5_set, NULL},
    { SWIG_prefix "rig_state_rx_range_list5_get", (swig_wrapper_func) _wrap_rig_state_rx_range_list5_get, NULL},
    { SWIG_prefix "rig_state_tx_range_list5_set", (swig_wrapper_func) _wrap_rig_state_tx_range_list5_set, NULL},
    { SWIG_prefix "rig_state_tx_range_list5_get", (swig_wrapper_func) _wrap_rig_state_tx_range_list5_get, NULL},
    { SWIG_prefix "rig_state_spectrum_scopes_set", (swig_wrapper_func) _wrap_rig_state_spectrum_scopes_set, NULL},
    { SWIG_prefix "rig_state_spectrum_scopes_get", (swig_wrapper_func) _wrap_rig_state_spectrum_scopes_get, NULL},
    { SWIG_prefix "rig_state_spectrum_modes_set", (swig_wrapper_func) _wrap_rig_state_spectrum_modes_set, NULL},
    { SWIG_prefix "rig_state_spectrum_modes_get", (swig_wrapper_func) _wrap_rig_state_spectrum_modes_get, NULL},
    { SWIG_prefix "rig_state_spectrum_spans_set", (swig_wrapper_func) _wrap_rig_state_spectrum_spans_set, NULL},
    { SWIG_prefix "rig_state_spectrum_spans_get", (swig_wrapper_func) _wrap_rig_state_spectrum_spans_get, NULL},
    { SWIG_prefix "rig_state_spectrum_avg_modes_set", (swig_wrapper_func) _wrap_rig_state_spectrum_avg_modes_set, NULL},
    { SWIG_prefix "rig_state_spectrum_avg_modes_get", (swig_wrapper_func) _wrap_rig_state_spectrum_avg_modes_get, NULL},
    { SWIG_prefix "rig_state_spectrum_attenuator_set", (swig_wrapper_func) _wrap_rig_state_spectrum_attenuator_set, NULL},
    { SWIG_prefix "rig_state_spectrum_attenuator_get", (swig_wrapper_func) _wrap_rig_state_spectrum_attenuator_get, NULL},
    { SWIG_prefix "rig_state_morse_data_handler_thread_run_set", (swig_wrapper_func) _wrap_rig_state_morse_data_handler_thread_run_set, NULL},
    { SWIG_prefix "rig_state_morse_data_handler_thread_run_get", (swig_wrapper_func) _wrap_rig_state_morse_data_handler_thread_run_get, NULL},
    { SWIG_prefix "rig_state_morse_data_handler_priv_data_set", (swig_wrapper_func) _wrap_rig_state_morse_data_handler_priv_data_set, NULL},
    { SWIG_prefix "rig_state_morse_data_handler_priv_data_get", (swig_wrapper_func) _wrap_rig_state_morse_data_handler_priv_data_get, NULL},
    { SWIG_prefix "rig_state_fifo_morse_set", (swig_wrapper_func) _wrap_rig_state_fifo_morse_set, NULL},
    { SWIG_prefix "rig_state_fifo_morse_get", (swig_wrapper_func) _wrap_rig_state_fifo_morse_get, NULL},
    { SWIG_prefix "rig_state_doppler_set", (swig_wrapper_func) _wrap_rig_state_doppler_set, NULL},
    { SWIG_prefix "rig_state_doppler_get", (swig_wrapper_func) _wrap_rig_state_doppler_get, NULL},
    { SWIG_prefix "rig_state_multicast_data_addr_set", (swig_wrapper_func) _wrap_rig_state_multicast_data_addr_set, NULL},
    { SWIG_prefix "rig_state_multicast_data_addr_get", (swig_wrapper_func) _wrap_rig_state_multicast_data_addr_get, NULL},
    { SWIG_prefix "rig_state_multicast_data_port_set", (swig_wrapper_func) _wrap_rig_state_multicast_data_port_set, NULL},
    { SWIG_prefix "rig_state_multicast_data_port_get", (swig_wrapper_func) _wrap_rig_state_multicast_data_port_get, NULL},
    { SWIG_prefix "rig_state_multicast_cmd_addr_set", (swig_wrapper_func) _wrap_rig_state_multicast_cmd_addr_set, NULL},
    { SWIG_prefix "rig_state_multicast_cmd_addr_get", (swig_wrapper_func) _wrap_rig_state_multicast_cmd_addr_get, NULL},
    { SWIG_prefix "rig_state_multicast_cmd_port_set", (swig_wrapper_func) _wrap_rig_state_multicast_cmd_port_set, NULL},
    { SWIG_prefix "rig_state_multicast_cmd_port_get", (swig_wrapper_func) _wrap_rig_state_multicast_cmd_port_get, NULL},
    { SWIG_prefix "rig_state_multicast_receiver_run_set", (swig_wrapper_func) _wrap_rig_state_multicast_receiver_run_set, NULL},
    { SWIG_prefix "rig_state_multicast_receiver_run_get", (swig_wrapper_func) _wrap_rig_state_multicast_receiver_run_get, NULL},
    { SWIG_prefix "rig_state_multicast_receiver_priv_data_set", (swig_wrapper_func) _wrap_rig_state_multicast_receiver_priv_data_set, NULL},
    { SWIG_prefix "rig_state_multicast_receiver_priv_data_get", (swig_wrapper_func) _wrap_rig_state_multicast_receiver_priv_data_get, NULL},
    { SWIG_prefix "rig_state_comm_status_set", (swig_wrapper_func) _wrap_rig_state_comm_status_set, NULL},
    { SWIG_prefix "rig_state_comm_status_get", (swig_wrapper_func) _wrap_rig_state_comm_status_get, NULL},
    { SWIG_prefix "rig_state_device_id_set", (swig_wrapper_func) _wrap_rig_state_device_id_set, NULL},
    { SWIG_prefix "rig_state_device_id_get", (swig_wrapper_func) _wrap_rig_state_device_id_get, NULL},
    { SWIG_prefix "rig_state_dual_watch_set", (swig_wrapper_func) _wrap_rig_state_dual_watch_set, NULL},
    { SWIG_prefix "rig_state_dual_watch_get", (swig_wrapper_func) _wrap_rig_state_dual_watch_get, NULL},
    { SWIG_prefix "rig_state_post_ptt_delay_set", (swig_wrapper_func) _wrap_rig_state_post_ptt_delay_set, NULL},
    { SWIG_prefix "rig_state_post_ptt_delay_get", (swig_wrapper_func) _wrap_rig_state_post_ptt_delay_get, NULL},
    { SWIG_prefix "rig_state_freq_event_elapsed_set", (swig_wrapper_func) _wrap_rig_state_freq_event_elapsed_set, NULL},
    { SWIG_prefix "rig_state_freq_event_elapsed_get", (swig_wrapper_func) _wrap_rig_state_freq_event_elapsed_get, NULL},
    { SWIG_prefix "rig_state_freq_skip_set", (swig_wrapper_func) _wrap_rig_state_freq_skip_set, NULL},
    { SWIG_prefix "rig_state_freq_skip_get", (swig_wrapper_func) _wrap_rig_state_freq_skip_get, NULL},
    { SWIG_prefix "rig_state_client_set", (swig_wrapper_func) _wrap_rig_state_client_set, NULL},
    { SWIG_prefix "rig_state_client_get", (swig_wrapper_func) _wrap_rig_state_client_get, NULL},
    { SWIG_prefix "rig_state_api_mutex_set", (swig_wrapper_func) _wrap_rig_state_api_mutex_set, NULL},
    { SWIG_prefix "rig_state_api_mutex_get", (swig_wrapper_func) _wrap_rig_state_api_mutex_get, NULL},
    { SWIG_prefix "new_rig_state", (swig_wrapper_func) _wrap_new_rig_state, NULL},
    { SWIG_prefix "delete_rig_state", (swig_wrapper_func) _wrap_delete_rig_state, NULL},
    { SWIG_prefix "rig_state", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_rig_state},
    { SWIG_prefix "Rig_rig_set", (swig_wrapper_func) _wrap_Rig_rig_set, NULL},
    { SWIG_prefix "Rig_rig_get", (swig_wrapper_func) _wrap_Rig_rig_get, NULL},
    { SWIG_prefix "Rig_caps_set", (swig_wrapper_func) _wrap_Rig_caps_set, NULL},
    { SWIG_prefix "Rig_caps_get", (swig_wrapper_func) _wrap_Rig_caps_get, NULL},
    { SWIG_prefix "Rig_state_set", (swig_wrapper_func) _wrap_Rig_state_set, NULL},
    { SWIG_prefix "Rig_state_get", (swig_wrapper_func) _wrap_Rig_state_get, NULL},
    { SWIG_prefix "Rig_error_status_set", (swig_wrapper_func) _wrap_Rig_error_status_set, NULL},
    { SWIG_prefix "Rig_error_status_get", (swig_wrapper_func) _wrap_Rig_error_status_get, NULL},
    { SWIG_prefix "Rig_do_exception_set", (swig_wrapper_func) _wrap_Rig_do_exception_set, NULL},
    { SWIG_prefix "Rig_do_exception_get", (swig_wrapper_func) _wrap_Rig_do_exception_get, NULL},
    { SWIG_prefix "new_Rig", (swig_wrapper_func) _wrap_new_Rig, NULL},
    { SWIG_prefix "delete_Rig", (swig_wrapper_func) _wrap_delete_Rig, NULL},
    { SWIG_prefix "Rig_open", (swig_wrapper_func) _wrap_Rig_open, NULL},
    { SWIG_prefix "Rig_close", (swig_wrapper_func) _wrap_Rig_close, NULL},
    { SWIG_prefix "Rig_set_freq", (swig_wrapper_func) _wrap_Rig_set_freq, NULL},
    { SWIG_prefix "Rig_set_mode", (swig_wrapper_func) _wrap_Rig_set_mode, NULL},
    { SWIG_prefix "Rig_set_ptt", (swig_wrapper_func) _wrap_Rig_set_ptt, NULL},
    { SWIG_prefix "Rig_set_rptr_shift", (swig_wrapper_func) _wrap_Rig_set_rptr_shift, NULL},
    { SWIG_prefix "Rig_set_rptr_offs", (swig_wrapper_func) _wrap_Rig_set_rptr_offs, NULL},
    { SWIG_prefix "Rig_set_ctcss_tone", (swig_wrapper_func) _wrap_Rig_set_ctcss_tone, NULL},
    { SWIG_prefix "Rig_set_dcs_code", (swig_wrapper_func) _wrap_Rig_set_dcs_code, NULL},
    { SWIG_prefix "Rig_set_ctcss_sql", (swig_wrapper_func) _wrap_Rig_set_ctcss_sql, NULL},
    { SWIG_prefix "Rig_set_dcs_sql", (swig_wrapper_func) _wrap_Rig_set_dcs_sql, NULL},
    { SWIG_prefix "Rig_set_split_freq", (swig_wrapper_func) _wrap_Rig_set_split_freq, NULL},
    { SWIG_prefix "Rig_set_split_mode", (swig_wrapper_func) _wrap_Rig_set_split_mode, NULL},
    { SWIG_prefix "Rig_set_split_freq_mode", (swig_wrapper_func) _wrap_Rig_set_split_freq_mode, NULL},
    { SWIG_prefix "Rig_set_split_vfo", (swig_wrapper_func) _wrap_Rig_set_split_vfo, NULL},
    { SWIG_prefix "Rig_set_rit", (swig_wrapper_func) _wrap_Rig_set_rit, NULL},
    { SWIG_prefix "Rig_set_xit", (swig_wrapper_func) _wrap_Rig_set_xit, NULL},
    { SWIG_prefix "Rig_set_ts", (swig_wrapper_func) _wrap_Rig_set_ts, NULL},
    { SWIG_prefix "Rig_set_ant", (swig_wrapper_func) _wrap_Rig_set_ant, NULL},
    { SWIG_prefix "Rig_set_func", (swig_wrapper_func) _wrap_Rig_set_func, NULL},
    { SWIG_prefix "Rig_set_ext_func", (swig_wrapper_func) _wrap_Rig_set_ext_func, NULL},
    { SWIG_prefix "Rig_set_bank", (swig_wrapper_func) _wrap_Rig_set_bank, NULL},
    { SWIG_prefix "Rig_set_mem", (swig_wrapper_func) _wrap_Rig_set_mem, NULL},
    { SWIG_prefix "Rig_send_dtmf", (swig_wrapper_func) _wrap_Rig_send_dtmf, NULL},
    { SWIG_prefix "Rig_send_morse", (swig_wrapper_func) _wrap_Rig_send_morse, NULL},
    { SWIG_prefix "Rig_vfo_op", (swig_wrapper_func) _wrap_Rig_vfo_op, NULL},
    { SWIG_prefix "Rig_scan", (swig_wrapper_func) _wrap_Rig_scan, NULL},
    { SWIG_prefix "Rig_set_level", (swig_wrapper_func) _wrap_Rig_set_level, NULL},
    { SWIG_prefix "Rig_set_ext_level", (swig_wrapper_func) _wrap_Rig_set_ext_level, NULL},
    { SWIG_prefix "Rig_get_level_i", (swig_wrapper_func) _wrap_Rig_get_level_i, NULL},
    { SWIG_prefix "Rig_get_level_f", (swig_wrapper_func) _wrap_Rig_get_level_f, NULL},
    { SWIG_prefix "Rig_set_vfo", (swig_wrapper_func) _wrap_Rig_set_vfo, NULL},
    { SWIG_prefix "Rig_set_powerstat", (swig_wrapper_func) _wrap_Rig_set_powerstat, NULL},
    { SWIG_prefix "Rig_has_get_level", (swig_wrapper_func) _wrap_Rig_has_get_level, NULL},
    { SWIG_prefix "Rig_has_get_parm", (swig_wrapper_func) _wrap_Rig_has_get_parm, NULL},
    { SWIG_prefix "Rig_has_set_parm", (swig_wrapper_func) _wrap_Rig_has_set_parm, NULL},
    { SWIG_prefix "Rig_has_get_func", (swig_wrapper_func) _wrap_Rig_has_get_func, NULL},
    { SWIG_prefix "Rig_has_set_func", (swig_wrapper_func) _wrap_Rig_has_set_func, NULL},
    { SWIG_prefix "Rig_reset", (swig_wrapper_func) _wrap_Rig_reset, NULL},
    { SWIG_prefix "Rig_has_scan", (swig_wrapper_func) _wrap_Rig_has_scan, NULL},
    { SWIG_prefix "Rig_has_vfo_op", (swig_wrapper_func) _wrap_Rig_has_vfo_op, NULL},
    { SWIG_prefix "Rig_passband_normal", (swig_wrapper_func) _wrap_Rig_passband_normal, NULL},
    { SWIG_prefix "Rig_passband_narrow", (swig_wrapper_func) _wrap_Rig_passband_narrow, NULL},
    { SWIG_prefix "Rig_passband_wide", (swig_wrapper_func) _wrap_Rig_passband_wide, NULL},
    { SWIG_prefix "Rig_set_vfo_opt", (swig_wrapper_func) _wrap_Rig_set_vfo_opt, NULL},
    { SWIG_prefix "Rig_ext_token_lookup", (swig_wrapper_func) _wrap_Rig_ext_token_lookup, NULL},
    { SWIG_prefix "Rig_token_lookup", (swig_wrapper_func) _wrap_Rig_token_lookup, NULL},
    { SWIG_prefix "Rig_set_ext_parm", (swig_wrapper_func) _wrap_Rig_set_ext_parm, NULL},
    { SWIG_prefix "Rig_set_parm", (swig_wrapper_func) _wrap_Rig_set_parm, NULL},
    { SWIG_prefix "Rig_get_parm_i", (swig_wrapper_func) _wrap_Rig_get_parm_i, NULL},
    { SWIG_prefix "Rig_get_parm_f", (swig_wrapper_func) _wrap_Rig_get_parm_f, NULL},
    { SWIG_prefix "Rig_set_conf", (swig_wrapper_func) _wrap_Rig_set_conf, NULL},
    { SWIG_prefix "Rig_get_freq", (swig_wrapper_func) _wrap_Rig_get_freq, NULL},
    { SWIG_prefix "Rig_get_mode", (swig_wrapper_func) _wrap_Rig_get_mode, NULL},
    { SWIG_prefix "Rig_get_split_mode", (swig_wrapper_func) _wrap_Rig_get_split_mode, NULL},
    { SWIG_prefix "Rig_get_vfo", (swig_wrapper_func) _wrap_Rig_get_vfo, NULL},
    { SWIG_prefix "Rig_get_ptt", (swig_wrapper_func) _wrap_Rig_get_ptt, NULL},
    { SWIG_prefix "Rig_get_rptr_shift", (swig_wrapper_func) _wrap_Rig_get_rptr_shift, NULL},
    { SWIG_prefix "Rig_get_rptr_offs", (swig_wrapper_func) _wrap_Rig_get_rptr_offs, NULL},
    { SWIG_prefix "Rig_get_ctcss_tone", (swig_wrapper_func) _wrap_Rig_get_ctcss_tone, NULL},
    { SWIG_prefix "Rig_get_dcs_code", (swig_wrapper_func) _wrap_Rig_get_dcs_code, NULL},
    { SWIG_prefix "Rig_get_ctcss_sql", (swig_wrapper_func) _wrap_Rig_get_ctcss_sql, NULL},
    { SWIG_prefix "Rig_get_dcs_sql", (swig_wrapper_func) _wrap_Rig_get_dcs_sql, NULL},
    { SWIG_prefix "Rig_get_split_freq", (swig_wrapper_func) _wrap_Rig_get_split_freq, NULL},
    { SWIG_prefix "Rig_get_split_vfo", (swig_wrapper_func) _wrap_Rig_get_split_vfo, NULL},
    { SWIG_prefix "Rig_get_rit", (swig_wrapper_func) _wrap_Rig_get_rit, NULL},
    { SWIG_prefix "Rig_get_xit", (swig_wrapper_func) _wrap_Rig_get_xit, NULL},
    { SWIG_prefix "Rig_get_ts", (swig_wrapper_func) _wrap_Rig_get_ts, NULL},
    { SWIG_prefix "Rig_get_ant", (swig_wrapper_func) _wrap_Rig_get_ant, NULL},
    { SWIG_prefix "Rig_get_vfo_info", (swig_wrapper_func) _wrap_Rig_get_vfo_info, NULL},
    { SWIG_prefix "Rig_get_mem", (swig_wrapper_func) _wrap_Rig_get_mem, NULL},
    { SWIG_prefix "Rig_get_powerstat", (swig_wrapper_func) _wrap_Rig_get_powerstat, NULL},
    { SWIG_prefix "Rig_get_dcd", (swig_wrapper_func) _wrap_Rig_get_dcd, NULL},
    { SWIG_prefix "Rig_mem_count", (swig_wrapper_func) _wrap_Rig_mem_count, NULL},
    { SWIG_prefix "Rig_lookup_mem_caps", (swig_wrapper_func) _wrap_Rig_lookup_mem_caps, NULL},
    { SWIG_prefix "Rig_set_channel", (swig_wrapper_func) _wrap_Rig_set_channel, NULL},
    { SWIG_prefix "Rig_get_chan_all", (swig_wrapper_func) _wrap_Rig_get_chan_all, NULL},
    { SWIG_prefix "Rig_get_channel", (swig_wrapper_func) _wrap_Rig_get_channel, NULL},
    { SWIG_prefix "Rig_get_conf", (swig_wrapper_func) _wrap_Rig_get_conf, NULL},
    { SWIG_prefix "Rig_recv_dtmf", (swig_wrapper_func) _wrap_Rig_recv_dtmf, NULL},
    { SWIG_prefix "Rig_get_info", (swig_wrapper_func) _wrap_Rig_get_info, NULL},
    { SWIG_prefix "Rig_get_func", (swig_wrapper_func) _wrap_Rig_get_func, NULL},
    { SWIG_prefix "Rig_get_ext_func", (swig_wrapper_func) _wrap_Rig_get_ext_func, NULL},
    { SWIG_prefix "Rig_get_level", (swig_wrapper_func) _wrap_Rig_get_level, NULL},
    { SWIG_prefix "Rig_get_ext_level", (swig_wrapper_func) _wrap_Rig_get_ext_level, NULL},
    { SWIG_prefix "Rig_get_parm", (swig_wrapper_func) _wrap_Rig_get_parm, NULL},
    { SWIG_prefix "Rig_get_ext_parm", (swig_wrapper_func) _wrap_Rig_get_ext_parm, NULL},
    { SWIG_prefix "Rig", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_Rig},
    { SWIG_prefix "new_channelArray", (swig_wrapper_func) _wrap_new_channelArray, NULL},
    { SWIG_prefix "delete_channelArray", (swig_wrapper_func) _wrap_delete_channelArray, NULL},
    { SWIG_prefix "channelArray_getitem", (swig_wrapper_func) _wrap_channelArray_getitem, NULL},
    { SWIG_prefix "channelArray_setitem", (swig_wrapper_func) _wrap_channelArray_setitem, NULL},
    { SWIG_prefix "channelArray_cast", (swig_wrapper_func) _wrap_channelArray_cast, NULL},
    { SWIG_prefix "channelArray_frompointer", (swig_wrapper_func) _wrap_channelArray_frompointer, NULL},
    { SWIG_prefix "channelArray", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_channelArray},
    { SWIG_prefix "new_toneArray", (swig_wrapper_func) _wrap_new_toneArray, NULL},
    { SWIG_prefix "delete_toneArray", (swig_wrapper_func) _wrap_delete_toneArray, NULL},
    { SWIG_prefix "toneArray_getitem", (swig_wrapper_func) _wrap_toneArray_getitem, NULL},
    { SWIG_prefix "toneArray_setitem", (swig_wrapper_func) _wrap_toneArray_setitem, NULL},
    { SWIG_prefix "toneArray_cast", (swig_wrapper_func) _wrap_toneArray_cast, NULL},
    { SWIG_prefix "toneArray_frompointer", (swig_wrapper_func) _wrap_toneArray_frompointer, NULL},
    { SWIG_prefix "toneArray", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_toneArray},
    { SWIG_prefix "rot_caps_rot_model_set", (swig_wrapper_func) _wrap_rot_caps_rot_model_set, NULL},
    { SWIG_prefix "rot_caps_rot_model_get", (swig_wrapper_func) _wrap_rot_caps_rot_model_get, NULL},
    { SWIG_prefix "rot_caps_model_name_get", (swig_wrapper_func) _wrap_rot_caps_model_name_get, NULL},
    { SWIG_prefix "rot_caps_mfg_name_get", (swig_wrapper_func) _wrap_rot_caps_mfg_name_get, NULL},
    { SWIG_prefix "rot_caps_version_get", (swig_wrapper_func) _wrap_rot_caps_version_get, NULL},
    { SWIG_prefix "rot_caps_copyright_get", (swig_wrapper_func) _wrap_rot_caps_copyright_get, NULL},
    { SWIG_prefix "rot_caps_status_set", (swig_wrapper_func) _wrap_rot_caps_status_set, NULL},
    { SWIG_prefix "rot_caps_status_get", (swig_wrapper_func) _wrap_rot_caps_status_get, NULL},
    { SWIG_prefix "rot_caps_rot_type_set", (swig_wrapper_func) _wrap_rot_caps_rot_type_set, NULL},
    { SWIG_prefix "rot_caps_rot_type_get", (swig_wrapper_func) _wrap_rot_caps_rot_type_get, NULL},
    { SWIG_prefix "rot_caps_port_type_set", (swig_wrapper_func) _wrap_rot_caps_port_type_set, NULL},
    { SWIG_prefix "rot_caps_port_type_get", (swig_wrapper_func) _wrap_rot_caps_port_type_get, NULL},
    { SWIG_prefix "rot_caps_serial_rate_min_set", (swig_wrapper_func) _wrap_rot_caps_serial_rate_min_set, NULL},
    { SWIG_prefix "rot_caps_serial_rate_min_get", (swig_wrapper_func) _wrap_rot_caps_serial_rate_min_get, NULL},
    { SWIG_prefix "rot_caps_serial_rate_max_set", (swig_wrapper_func) _wrap_rot_caps_serial_rate_max_set, NULL},
    { SWIG_prefix "rot_caps_serial_rate_max_get", (swig_wrapper_func) _wrap_rot_caps_serial_rate_max_get, NULL},
    { SWIG_prefix "rot_caps_serial_data_bits_set", (swig_wrapper_func) _wrap_rot_caps_serial_data_bits_set, NULL},
    { SWIG_prefix "rot_caps_serial_data_bits_get", (swig_wrapper_func) _wrap_rot_caps_serial_data_bits_get, NULL},
    { SWIG_prefix "rot_caps_serial_stop_bits_set", (swig_wrapper_func) _wrap_rot_caps_serial_stop_bits_set, NULL},
    { SWIG_prefix "rot_caps_serial_stop_bits_get", (swig_wrapper_func) _wrap_rot_caps_serial_stop_bits_get, NULL},
    { SWIG_prefix "rot_caps_serial_parity_set", (swig_wrapper_func) _wrap_rot_caps_serial_parity_set, NULL},
    { SWIG_prefix "rot_caps_serial_parity_get", (swig_wrapper_func) _wrap_rot_caps_serial_parity_get, NULL},
    { SWIG_prefix "rot_caps_serial_handshake_set", (swig_wrapper_func) _wrap_rot_caps_serial_handshake_set, NULL},
    { SWIG_prefix "rot_caps_serial_handshake_get", (swig_wrapper_func) _wrap_rot_caps_serial_handshake_get, NULL},
    { SWIG_prefix "rot_caps_write_delay_set", (swig_wrapper_func) _wrap_rot_caps_write_delay_set, NULL},
    { SWIG_prefix "rot_caps_write_delay_get", (swig_wrapper_func) _wrap_rot_caps_write_delay_get, NULL},
    { SWIG_prefix "rot_caps_post_write_delay_set", (swig_wrapper_func) _wrap_rot_caps_post_write_delay_set, NULL},
    { SWIG_prefix "rot_caps_post_write_delay_get", (swig_wrapper_func) _wrap_rot_caps_post_write_delay_get, NULL},
    { SWIG_prefix "rot_caps_timeout_set", (swig_wrapper_func) _wrap_rot_caps_timeout_set, NULL},
    { SWIG_prefix "rot_caps_timeout_get", (swig_wrapper_func) _wrap_rot_caps_timeout_get, NULL},
    { SWIG_prefix "rot_caps_retry_set", (swig_wrapper_func) _wrap_rot_caps_retry_set, NULL},
    { SWIG_prefix "rot_caps_retry_get", (swig_wrapper_func) _wrap_rot_caps_retry_get, NULL},
    { SWIG_prefix "rot_caps_has_get_func_set", (swig_wrapper_func) _wrap_rot_caps_has_get_func_set, NULL},
    { SWIG_prefix "rot_caps_has_get_func_get", (swig_wrapper_func) _wrap_rot_caps_has_get_func_get, NULL},
    { SWIG_prefix "rot_caps_has_set_func_set", (swig_wrapper_func) _wrap_rot_caps_has_set_func_set, NULL},
    { SWIG_prefix "rot_caps_has_set_func_get", (swig_wrapper_func) _wrap_rot_caps_has_set_func_get, NULL},
    { SWIG_prefix "rot_caps_has_get_level_set", (swig_wrapper_func) _wrap_rot_caps_has_get_level_set, NULL},
    { SWIG_prefix "rot_caps_has_get_level_get", (swig_wrapper_func) _wrap_rot_caps_has_get_level_get, NULL},
    { SWIG_prefix "rot_caps_has_set_level_set", (swig_wrapper_func) _wrap_rot_caps_has_set_level_set, NULL},
    { SWIG_prefix "rot_caps_has_set_level_get", (swig_wrapper_func) _wrap_rot_caps_has_set_level_get, NULL},
    { SWIG_prefix "rot_caps_has_get_parm_set", (swig_wrapper_func) _wrap_rot_caps_has_get_parm_set, NULL},
    { SWIG_prefix "rot_caps_has_get_parm_get", (swig_wrapper_func) _wrap_rot_caps_has_get_parm_get, NULL},
    { SWIG_prefix "rot_caps_has_set_parm_set", (swig_wrapper_func) _wrap_rot_caps_has_set_parm_set, NULL},
    { SWIG_prefix "rot_caps_has_set_parm_get", (swig_wrapper_func) _wrap_rot_caps_has_set_parm_get, NULL},
    { SWIG_prefix "rot_caps_has_status_set", (swig_wrapper_func) _wrap_rot_caps_has_status_set, NULL},
    { SWIG_prefix "rot_caps_has_status_get", (swig_wrapper_func) _wrap_rot_caps_has_status_get, NULL},
    { SWIG_prefix "rot_caps_level_gran_set", (swig_wrapper_func) _wrap_rot_caps_level_gran_set, NULL},
    { SWIG_prefix "rot_caps_level_gran_get", (swig_wrapper_func) _wrap_rot_caps_level_gran_get, NULL},
    { SWIG_prefix "rot_caps_parm_gran_set", (swig_wrapper_func) _wrap_rot_caps_parm_gran_set, NULL},
    { SWIG_prefix "rot_caps_parm_gran_get", (swig_wrapper_func) _wrap_rot_caps_parm_gran_get, NULL},
    { SWIG_prefix "rot_caps_extparms_set", (swig_wrapper_func) _wrap_rot_caps_extparms_set, NULL},
    { SWIG_prefix "rot_caps_extparms_get", (swig_wrapper_func) _wrap_rot_caps_extparms_get, NULL},
    { SWIG_prefix "rot_caps_extlevels_set", (swig_wrapper_func) _wrap_rot_caps_extlevels_set, NULL},
    { SWIG_prefix "rot_caps_extlevels_get", (swig_wrapper_func) _wrap_rot_caps_extlevels_get, NULL},
    { SWIG_prefix "rot_caps_extfuncs_set", (swig_wrapper_func) _wrap_rot_caps_extfuncs_set, NULL},
    { SWIG_prefix "rot_caps_extfuncs_get", (swig_wrapper_func) _wrap_rot_caps_extfuncs_get, NULL},
    { SWIG_prefix "rot_caps_ext_tokens_set", (swig_wrapper_func) _wrap_rot_caps_ext_tokens_set, NULL},
    { SWIG_prefix "rot_caps_ext_tokens_get", (swig_wrapper_func) _wrap_rot_caps_ext_tokens_get, NULL},
    { SWIG_prefix "rot_caps_min_az_set", (swig_wrapper_func) _wrap_rot_caps_min_az_set, NULL},
    { SWIG_prefix "rot_caps_min_az_get", (swig_wrapper_func) _wrap_rot_caps_min_az_get, NULL},
    { SWIG_prefix "rot_caps_max_az_set", (swig_wrapper_func) _wrap_rot_caps_max_az_set, NULL},
    { SWIG_prefix "rot_caps_max_az_get", (swig_wrapper_func) _wrap_rot_caps_max_az_get, NULL},
    { SWIG_prefix "rot_caps_min_el_set", (swig_wrapper_func) _wrap_rot_caps_min_el_set, NULL},
    { SWIG_prefix "rot_caps_min_el_get", (swig_wrapper_func) _wrap_rot_caps_min_el_get, NULL},
    { SWIG_prefix "rot_caps_max_el_set", (swig_wrapper_func) _wrap_rot_caps_max_el_set, NULL},
    { SWIG_prefix "rot_caps_max_el_get", (swig_wrapper_func) _wrap_rot_caps_max_el_get, NULL},
    { SWIG_prefix "rot_caps_cfgparams_set", (swig_wrapper_func) _wrap_rot_caps_cfgparams_set, NULL},
    { SWIG_prefix "rot_caps_cfgparams_get", (swig_wrapper_func) _wrap_rot_caps_cfgparams_get, NULL},
    { SWIG_prefix "rot_caps_priv_set", (swig_wrapper_func) _wrap_rot_caps_priv_set, NULL},
    { SWIG_prefix "rot_caps_priv_get", (swig_wrapper_func) _wrap_rot_caps_priv_get, NULL},
    { SWIG_prefix "rot_caps_rot_init_set", (swig_wrapper_func) _wrap_rot_caps_rot_init_set, NULL},
    { SWIG_prefix "rot_caps_rot_init_get", (swig_wrapper_func) _wrap_rot_caps_rot_init_get, NULL},
    { SWIG_prefix "rot_caps_rot_cleanup_set", (swig_wrapper_func) _wrap_rot_caps_rot_cleanup_set, NULL},
    { SWIG_prefix "rot_caps_rot_cleanup_get", (swig_wrapper_func) _wrap_rot_caps_rot_cleanup_get, NULL},
    { SWIG_prefix "rot_caps_rot_open_set", (swig_wrapper_func) _wrap_rot_caps_rot_open_set, NULL},
    { SWIG_prefix "rot_caps_rot_open_get", (swig_wrapper_func) _wrap_rot_caps_rot_open_get, NULL},
    { SWIG_prefix "rot_caps_rot_close_set", (swig_wrapper_func) _wrap_rot_caps_rot_close_set, NULL},
    { SWIG_prefix "rot_caps_rot_close_get", (swig_wrapper_func) _wrap_rot_caps_rot_close_get, NULL},
    { SWIG_prefix "rot_caps_set_conf_set", (swig_wrapper_func) _wrap_rot_caps_set_conf_set, NULL},
    { SWIG_prefix "rot_caps_set_conf_get", (swig_wrapper_func) _wrap_rot_caps_set_conf_get, NULL},
    { SWIG_prefix "rot_caps_get_conf_set", (swig_wrapper_func) _wrap_rot_caps_get_conf_set, NULL},
    { SWIG_prefix "rot_caps_get_conf_get", (swig_wrapper_func) _wrap_rot_caps_get_conf_get, NULL},
    { SWIG_prefix "rot_caps_set_position_set", (swig_wrapper_func) _wrap_rot_caps_set_position_set, NULL},
    { SWIG_prefix "rot_caps_set_position_get", (swig_wrapper_func) _wrap_rot_caps_set_position_get, NULL},
    { SWIG_prefix "rot_caps_get_position_set", (swig_wrapper_func) _wrap_rot_caps_get_position_set, NULL},
    { SWIG_prefix "rot_caps_get_position_get", (swig_wrapper_func) _wrap_rot_caps_get_position_get, NULL},
    { SWIG_prefix "rot_caps_stop_set", (swig_wrapper_func) _wrap_rot_caps_stop_set, NULL},
    { SWIG_prefix "rot_caps_stop_get", (swig_wrapper_func) _wrap_rot_caps_stop_get, NULL},
    { SWIG_prefix "rot_caps_park_set", (swig_wrapper_func) _wrap_rot_caps_park_set, NULL},
    { SWIG_prefix "rot_caps_park_get", (swig_wrapper_func) _wrap_rot_caps_park_get, NULL},
    { SWIG_prefix "rot_caps_reset_set", (swig_wrapper_func) _wrap_rot_caps_reset_set, NULL},
    { SWIG_prefix "rot_caps_reset_get", (swig_wrapper_func) _wrap_rot_caps_reset_get, NULL},
    { SWIG_prefix "rot_caps_move_set", (swig_wrapper_func) _wrap_rot_caps_move_set, NULL},
    { SWIG_prefix "rot_caps_move_get", (swig_wrapper_func) _wrap_rot_caps_move_get, NULL},
    { SWIG_prefix "rot_caps_get_info_set", (swig_wrapper_func) _wrap_rot_caps_get_info_set, NULL},
    { SWIG_prefix "rot_caps_get_info_get", (swig_wrapper_func) _wrap_rot_caps_get_info_get, NULL},
    { SWIG_prefix "rot_caps_set_level_set", (swig_wrapper_func) _wrap_rot_caps_set_level_set, NULL},
    { SWIG_prefix "rot_caps_set_level_get", (swig_wrapper_func) _wrap_rot_caps_set_level_get, NULL},
    { SWIG_prefix "rot_caps_get_level_set", (swig_wrapper_func) _wrap_rot_caps_get_level_set, NULL},
    { SWIG_prefix "rot_caps_get_level_get", (swig_wrapper_func) _wrap_rot_caps_get_level_get, NULL},
    { SWIG_prefix "rot_caps_set_func_set", (swig_wrapper_func) _wrap_rot_caps_set_func_set, NULL},
    { SWIG_prefix "rot_caps_set_func_get", (swig_wrapper_func) _wrap_rot_caps_set_func_get, NULL},
    { SWIG_prefix "rot_caps_get_func_set", (swig_wrapper_func) _wrap_rot_caps_get_func_set, NULL},
    { SWIG_prefix "rot_caps_get_func_get", (swig_wrapper_func) _wrap_rot_caps_get_func_get, NULL},
    { SWIG_prefix "rot_caps_set_parm_set", (swig_wrapper_func) _wrap_rot_caps_set_parm_set, NULL},
    { SWIG_prefix "rot_caps_set_parm_get", (swig_wrapper_func) _wrap_rot_caps_set_parm_get, NULL},
    { SWIG_prefix "rot_caps_get_parm_set", (swig_wrapper_func) _wrap_rot_caps_get_parm_set, NULL},
    { SWIG_prefix "rot_caps_get_parm_get", (swig_wrapper_func) _wrap_rot_caps_get_parm_get, NULL},
    { SWIG_prefix "rot_caps_set_ext_level_set", (swig_wrapper_func) _wrap_rot_caps_set_ext_level_set, NULL},
    { SWIG_prefix "rot_caps_set_ext_level_get", (swig_wrapper_func) _wrap_rot_caps_set_ext_level_get, NULL},
    { SWIG_prefix "rot_caps_get_ext_level_set", (swig_wrapper_func) _wrap_rot_caps_get_ext_level_set, NULL},
    { SWIG_prefix "rot_caps_get_ext_level_get", (swig_wrapper_func) _wrap_rot_caps_get_ext_level_get, NULL},
    { SWIG_prefix "rot_caps_set_ext_func_set", (swig_wrapper_func) _wrap_rot_caps_set_ext_func_set, NULL},
    { SWIG_prefix "rot_caps_set_ext_func_get", (swig_wrapper_func) _wrap_rot_caps_set_ext_func_get, NULL},
    { SWIG_prefix "rot_caps_get_ext_func_set", (swig_wrapper_func) _wrap_rot_caps_get_ext_func_set, NULL},
    { SWIG_prefix "rot_caps_get_ext_func_get", (swig_wrapper_func) _wrap_rot_caps_get_ext_func_get, NULL},
    { SWIG_prefix "rot_caps_set_ext_parm_set", (swig_wrapper_func) _wrap_rot_caps_set_ext_parm_set, NULL},
    { SWIG_prefix "rot_caps_set_ext_parm_get", (swig_wrapper_func) _wrap_rot_caps_set_ext_parm_get, NULL},
    { SWIG_prefix "rot_caps_get_ext_parm_set", (swig_wrapper_func) _wrap_rot_caps_get_ext_parm_set, NULL},
    { SWIG_prefix "rot_caps_get_ext_parm_get", (swig_wrapper_func) _wrap_rot_caps_get_ext_parm_get, NULL},
    { SWIG_prefix "rot_caps_get_status_set", (swig_wrapper_func) _wrap_rot_caps_get_status_set, NULL},
    { SWIG_prefix "rot_caps_get_status_get", (swig_wrapper_func) _wrap_rot_caps_get_status_get, NULL},
    { SWIG_prefix "rot_caps_get_conf2_set", (swig_wrapper_func) _wrap_rot_caps_get_conf2_set, NULL},
    { SWIG_prefix "rot_caps_get_conf2_get", (swig_wrapper_func) _wrap_rot_caps_get_conf2_get, NULL},
    { SWIG_prefix "new_rot_caps", (swig_wrapper_func) _wrap_new_rot_caps, NULL},
    { SWIG_prefix "delete_rot_caps", (swig_wrapper_func) _wrap_delete_rot_caps, NULL},
    { SWIG_prefix "rot_caps", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_rot_caps},
    { SWIG_prefix "qrb", (swig_wrapper_func) _wrap_qrb, NULL},
    { SWIG_prefix "distance_long_path", (swig_wrapper_func) _wrap_distance_long_path, NULL},
    { SWIG_prefix "azimuth_long_path", (swig_wrapper_func) _wrap_azimuth_long_path, NULL},
    { SWIG_prefix "dms2dec", (swig_wrapper_func) _wrap_dms2dec, NULL},
    { SWIG_prefix "dec2dms", (swig_wrapper_func) _wrap_dec2dms, NULL},
    { SWIG_prefix "dec2dmmm", (swig_wrapper_func) _wrap_dec2dmmm, NULL},
    { SWIG_prefix "dmmm2dec", (swig_wrapper_func) _wrap_dmmm2dec, NULL},
    { SWIG_prefix "rot_state_min_az_set", (swig_wrapper_func) _wrap_rot_state_min_az_set, NULL},
    { SWIG_prefix "rot_state_min_az_get", (swig_wrapper_func) _wrap_rot_state_min_az_get, NULL},
    { SWIG_prefix "rot_state_max_az_set", (swig_wrapper_func) _wrap_rot_state_max_az_set, NULL},
    { SWIG_prefix "rot_state_max_az_get", (swig_wrapper_func) _wrap_rot_state_max_az_get, NULL},
    { SWIG_prefix "rot_state_min_el_set", (swig_wrapper_func) _wrap_rot_state_min_el_set, NULL},
    { SWIG_prefix "rot_state_min_el_get", (swig_wrapper_func) _wrap_rot_state_min_el_get, NULL},
    { SWIG_prefix "rot_state_max_el_set", (swig_wrapper_func) _wrap_rot_state_max_el_set, NULL},
    { SWIG_prefix "rot_state_max_el_get", (swig_wrapper_func) _wrap_rot_state_max_el_get, NULL},
    { SWIG_prefix "rot_state_south_zero_set", (swig_wrapper_func) _wrap_rot_state_south_zero_set, NULL},
    { SWIG_prefix "rot_state_south_zero_get", (swig_wrapper_func) _wrap_rot_state_south_zero_get, NULL},
    { SWIG_prefix "rot_state_az_offset_set", (swig_wrapper_func) _wrap_rot_state_az_offset_set, NULL},
    { SWIG_prefix "rot_state_az_offset_get", (swig_wrapper_func) _wrap_rot_state_az_offset_get, NULL},
    { SWIG_prefix "rot_state_el_offset_set", (swig_wrapper_func) _wrap_rot_state_el_offset_set, NULL},
    { SWIG_prefix "rot_state_el_offset_get", (swig_wrapper_func) _wrap_rot_state_el_offset_get, NULL},
    { SWIG_prefix "rot_state_has_get_func_set", (swig_wrapper_func) _wrap_rot_state_has_get_func_set, NULL},
    { SWIG_prefix "rot_state_has_get_func_get", (swig_wrapper_func) _wrap_rot_state_has_get_func_get, NULL},
    { SWIG_prefix "rot_state_has_set_func_set", (swig_wrapper_func) _wrap_rot_state_has_set_func_set, NULL},
    { SWIG_prefix "rot_state_has_set_func_get", (swig_wrapper_func) _wrap_rot_state_has_set_func_get, NULL},
    { SWIG_prefix "rot_state_has_get_level_set", (swig_wrapper_func) _wrap_rot_state_has_get_level_set, NULL},
    { SWIG_prefix "rot_state_has_get_level_get", (swig_wrapper_func) _wrap_rot_state_has_get_level_get, NULL},
    { SWIG_prefix "rot_state_has_set_level_set", (swig_wrapper_func) _wrap_rot_state_has_set_level_set, NULL},
    { SWIG_prefix "rot_state_has_set_level_get", (swig_wrapper_func) _wrap_rot_state_has_set_level_get, NULL},
    { SWIG_prefix "rot_state_has_get_parm_set", (swig_wrapper_func) _wrap_rot_state_has_get_parm_set, NULL},
    { SWIG_prefix "rot_state_has_get_parm_get", (swig_wrapper_func) _wrap_rot_state_has_get_parm_get, NULL},
    { SWIG_prefix "rot_state_has_set_parm_set", (swig_wrapper_func) _wrap_rot_state_has_set_parm_set, NULL},
    { SWIG_prefix "rot_state_has_set_parm_get", (swig_wrapper_func) _wrap_rot_state_has_set_parm_get, NULL},
    { SWIG_prefix "rot_state_has_status_set", (swig_wrapper_func) _wrap_rot_state_has_status_set, NULL},
    { SWIG_prefix "rot_state_has_status_get", (swig_wrapper_func) _wrap_rot_state_has_status_get, NULL},
    { SWIG_prefix "rot_state_level_gran_set", (swig_wrapper_func) _wrap_rot_state_level_gran_set, NULL},
    { SWIG_prefix "rot_state_level_gran_get", (swig_wrapper_func) _wrap_rot_state_level_gran_get, NULL},
    { SWIG_prefix "rot_state_parm_gran_set", (swig_wrapper_func) _wrap_rot_state_parm_gran_set, NULL},
    { SWIG_prefix "rot_state_parm_gran_get", (swig_wrapper_func) _wrap_rot_state_parm_gran_get, NULL},
    { SWIG_prefix "rot_state_comm_state_set", (swig_wrapper_func) _wrap_rot_state_comm_state_set, NULL},
    { SWIG_prefix "rot_state_comm_state_get", (swig_wrapper_func) _wrap_rot_state_comm_state_get, NULL},
    { SWIG_prefix "rot_state_priv_set", (swig_wrapper_func) _wrap_rot_state_priv_set, NULL},
    { SWIG_prefix "rot_state_priv_get", (swig_wrapper_func) _wrap_rot_state_priv_get, NULL},
    { SWIG_prefix "rot_state_obj_set", (swig_wrapper_func) _wrap_rot_state_obj_set, NULL},
    { SWIG_prefix "rot_state_obj_get", (swig_wrapper_func) _wrap_rot_state_obj_get, NULL},
    { SWIG_prefix "rot_state_current_speed_set", (swig_wrapper_func) _wrap_rot_state_current_speed_set, NULL},
    { SWIG_prefix "rot_state_current_speed_get", (swig_wrapper_func) _wrap_rot_state_current_speed_get, NULL},
    { SWIG_prefix "rot_state_rotport_set", (swig_wrapper_func) _wrap_rot_state_rotport_set, NULL},
    { SWIG_prefix "rot_state_rotport_get", (swig_wrapper_func) _wrap_rot_state_rotport_get, NULL},
    { SWIG_prefix "rot_state_rotport2_set", (swig_wrapper_func) _wrap_rot_state_rotport2_set, NULL},
    { SWIG_prefix "rot_state_rotport2_get", (swig_wrapper_func) _wrap_rot_state_rotport2_get, NULL},
    { SWIG_prefix "rot_state_pstrotator_handler_priv_data_set", (swig_wrapper_func) _wrap_rot_state_pstrotator_handler_priv_data_set, NULL},
    { SWIG_prefix "rot_state_pstrotator_handler_priv_data_get", (swig_wrapper_func) _wrap_rot_state_pstrotator_handler_priv_data_get, NULL},
    { SWIG_prefix "rot_state_config_queue_set", (swig_wrapper_func) _wrap_rot_state_config_queue_set, NULL},
    { SWIG_prefix "rot_state_config_queue_get", (swig_wrapper_func) _wrap_rot_state_config_queue_get, NULL},
    { SWIG_prefix "new_rot_state", (swig_wrapper_func) _wrap_new_rot_state, NULL},
    { SWIG_prefix "delete_rot_state", (swig_wrapper_func) _wrap_delete_rot_state, NULL},
    { SWIG_prefix "rot_state", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_rot_state},
    { SWIG_prefix "Rot_rot_set", (swig_wrapper_func) _wrap_Rot_rot_set, NULL},
    { SWIG_prefix "Rot_rot_get", (swig_wrapper_func) _wrap_Rot_rot_get, NULL},
    { SWIG_prefix "Rot_caps_set", (swig_wrapper_func) _wrap_Rot_caps_set, NULL},
    { SWIG_prefix "Rot_caps_get", (swig_wrapper_func) _wrap_Rot_caps_get, NULL},
    { SWIG_prefix "Rot_state_set", (swig_wrapper_func) _wrap_Rot_state_set, NULL},
    { SWIG_prefix "Rot_state_get", (swig_wrapper_func) _wrap_Rot_state_get, NULL},
    { SWIG_prefix "Rot_error_status_set", (swig_wrapper_func) _wrap_Rot_error_status_set, NULL},
    { SWIG_prefix "Rot_error_status_get", (swig_wrapper_func) _wrap_Rot_error_status_get, NULL},
    { SWIG_prefix "Rot_do_exception_set", (swig_wrapper_func) _wrap_Rot_do_exception_set, NULL},
    { SWIG_prefix "Rot_do_exception_get", (swig_wrapper_func) _wrap_Rot_do_exception_get, NULL},
    { SWIG_prefix "new_Rot", (swig_wrapper_func) _wrap_new_Rot, NULL},
    { SWIG_prefix "delete_Rot", (swig_wrapper_func) _wrap_delete_Rot, NULL},
    { SWIG_prefix "Rot_open", (swig_wrapper_func) _wrap_Rot_open, NULL},
    { SWIG_prefix "Rot_close", (swig_wrapper_func) _wrap_Rot_close, NULL},
    { SWIG_prefix "Rot_set_ext_func", (swig_wrapper_func) _wrap_Rot_set_ext_func, NULL},
    { SWIG_prefix "Rot_set_ext_level", (swig_wrapper_func) _wrap_Rot_set_ext_level, NULL},
    { SWIG_prefix "Rot_set_ext_parm", (swig_wrapper_func) _wrap_Rot_set_ext_parm, NULL},
    { SWIG_prefix "Rot_set_func", (swig_wrapper_func) _wrap_Rot_set_func, NULL},
    { SWIG_prefix "Rot_set_level", (swig_wrapper_func) _wrap_Rot_set_level, NULL},
    { SWIG_prefix "Rot_set_parm", (swig_wrapper_func) _wrap_Rot_set_parm, NULL},
    { SWIG_prefix "Rot_set_position", (swig_wrapper_func) _wrap_Rot_set_position, NULL},
    { SWIG_prefix "Rot_get_position", (swig_wrapper_func) _wrap_Rot_get_position, NULL},
    { SWIG_prefix "Rot_stop", (swig_wrapper_func) _wrap_Rot_stop, NULL},
    { SWIG_prefix "Rot_park", (swig_wrapper_func) _wrap_Rot_park, NULL},
    { SWIG_prefix "Rot_reset", (swig_wrapper_func) _wrap_Rot_reset, NULL},
    { SWIG_prefix "Rot_move", (swig_wrapper_func) _wrap_Rot_move, NULL},
    { SWIG_prefix "Rot_token_lookup", (swig_wrapper_func) _wrap_Rot_token_lookup, NULL},
    { SWIG_prefix "Rot_set_conf", (swig_wrapper_func) _wrap_Rot_set_conf, NULL},
    { SWIG_prefix "Rot_has_get_func", (swig_wrapper_func) _wrap_Rot_has_get_func, NULL},
    { SWIG_prefix "Rot_has_set_func", (swig_wrapper_func) _wrap_Rot_has_set_func, NULL},
    { SWIG_prefix "Rot_has_get_level", (swig_wrapper_func) _wrap_Rot_has_get_level, NULL},
    { SWIG_prefix "Rot_has_set_level", (swig_wrapper_func) _wrap_Rot_has_set_level, NULL},
    { SWIG_prefix "Rot_has_get_parm", (swig_wrapper_func) _wrap_Rot_has_get_parm, NULL},
    { SWIG_prefix "Rot_has_set_parm", (swig_wrapper_func) _wrap_Rot_has_set_parm, NULL},
    { SWIG_prefix "Rot_get_conf", (swig_wrapper_func) _wrap_Rot_get_conf, NULL},
    { SWIG_prefix "Rot_get_info", (swig_wrapper_func) _wrap_Rot_get_info, NULL},
    { SWIG_prefix "Rot", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_Rot},
    { SWIG_prefix "amp_caps_amp_model_set", (swig_wrapper_func) _wrap_amp_caps_amp_model_set, NULL},
    { SWIG_prefix "amp_caps_amp_model_get", (swig_wrapper_func) _wrap_amp_caps_amp_model_get, NULL},
    { SWIG_prefix "amp_caps_model_name_get", (swig_wrapper_func) _wrap_amp_caps_model_name_get, NULL},
    { SWIG_prefix "amp_caps_mfg_name_get", (swig_wrapper_func) _wrap_amp_caps_mfg_name_get, NULL},
    { SWIG_prefix "amp_caps_version_get", (swig_wrapper_func) _wrap_amp_caps_version_get, NULL},
    { SWIG_prefix "amp_caps_copyright_get", (swig_wrapper_func) _wrap_amp_caps_copyright_get, NULL},
    { SWIG_prefix "amp_caps_status_set", (swig_wrapper_func) _wrap_amp_caps_status_set, NULL},
    { SWIG_prefix "amp_caps_status_get", (swig_wrapper_func) _wrap_amp_caps_status_get, NULL},
    { SWIG_prefix "amp_caps_amp_type_set", (swig_wrapper_func) _wrap_amp_caps_amp_type_set, NULL},
    { SWIG_prefix "amp_caps_amp_type_get", (swig_wrapper_func) _wrap_amp_caps_amp_type_get, NULL},
    { SWIG_prefix "amp_caps_port_type_set", (swig_wrapper_func) _wrap_amp_caps_port_type_set, NULL},
    { SWIG_prefix "amp_caps_port_type_get", (swig_wrapper_func) _wrap_amp_caps_port_type_get, NULL},
    { SWIG_prefix "amp_caps_serial_rate_min_set", (swig_wrapper_func) _wrap_amp_caps_serial_rate_min_set, NULL},
    { SWIG_prefix "amp_caps_serial_rate_min_get", (swig_wrapper_func) _wrap_amp_caps_serial_rate_min_get, NULL},
    { SWIG_prefix "amp_caps_serial_rate_max_set", (swig_wrapper_func) _wrap_amp_caps_serial_rate_max_set, NULL},
    { SWIG_prefix "amp_caps_serial_rate_max_get", (swig_wrapper_func) _wrap_amp_caps_serial_rate_max_get, NULL},
    { SWIG_prefix "amp_caps_serial_data_bits_set", (swig_wrapper_func) _wrap_amp_caps_serial_data_bits_set, NULL},
    { SWIG_prefix "amp_caps_serial_data_bits_get", (swig_wrapper_func) _wrap_amp_caps_serial_data_bits_get, NULL},
    { SWIG_prefix "amp_caps_serial_stop_bits_set", (swig_wrapper_func) _wrap_amp_caps_serial_stop_bits_set, NULL},
    { SWIG_prefix "amp_caps_serial_stop_bits_get", (swig_wrapper_func) _wrap_amp_caps_serial_stop_bits_get, NULL},
    { SWIG_prefix "amp_caps_serial_parity_set", (swig_wrapper_func) _wrap_amp_caps_serial_parity_set, NULL},
    { SWIG_prefix "amp_caps_serial_parity_get", (swig_wrapper_func) _wrap_amp_caps_serial_parity_get, NULL},
    { SWIG_prefix "amp_caps_serial_handshake_set", (swig_wrapper_func) _wrap_amp_caps_serial_handshake_set, NULL},
    { SWIG_prefix "amp_caps_serial_handshake_get", (swig_wrapper_func) _wrap_amp_caps_serial_handshake_get, NULL},
    { SWIG_prefix "amp_caps_write_delay_set", (swig_wrapper_func) _wrap_amp_caps_write_delay_set, NULL},
    { SWIG_prefix "amp_caps_write_delay_get", (swig_wrapper_func) _wrap_amp_caps_write_delay_get, NULL},
    { SWIG_prefix "amp_caps_post_write_delay_set", (swig_wrapper_func) _wrap_amp_caps_post_write_delay_set, NULL},
    { SWIG_prefix "amp_caps_post_write_delay_get", (swig_wrapper_func) _wrap_amp_caps_post_write_delay_get, NULL},
    { SWIG_prefix "amp_caps_timeout_set", (swig_wrapper_func) _wrap_amp_caps_timeout_set, NULL},
    { SWIG_prefix "amp_caps_timeout_get", (swig_wrapper_func) _wrap_amp_caps_timeout_get, NULL},
    { SWIG_prefix "amp_caps_retry_set", (swig_wrapper_func) _wrap_amp_caps_retry_set, NULL},
    { SWIG_prefix "amp_caps_retry_get", (swig_wrapper_func) _wrap_amp_caps_retry_get, NULL},
    { SWIG_prefix "amp_caps_cfgparams_set", (swig_wrapper_func) _wrap_amp_caps_cfgparams_set, NULL},
    { SWIG_prefix "amp_caps_cfgparams_get", (swig_wrapper_func) _wrap_amp_caps_cfgparams_get, NULL},
    { SWIG_prefix "amp_caps_priv_set", (swig_wrapper_func) _wrap_amp_caps_priv_set, NULL},
    { SWIG_prefix "amp_caps_priv_get", (swig_wrapper_func) _wrap_amp_caps_priv_get, NULL},
    { SWIG_prefix "amp_caps_amp_model_macro_name_set", (swig_wrapper_func) _wrap_amp_caps_amp_model_macro_name_set, NULL},
    { SWIG_prefix "amp_caps_amp_model_macro_name_get", (swig_wrapper_func) _wrap_amp_caps_amp_model_macro_name_get, NULL},
    { SWIG_prefix "amp_caps_has_get_level_set", (swig_wrapper_func) _wrap_amp_caps_has_get_level_set, NULL},
    { SWIG_prefix "amp_caps_has_get_level_get", (swig_wrapper_func) _wrap_amp_caps_has_get_level_get, NULL},
    { SWIG_prefix "amp_caps_has_set_level_set", (swig_wrapper_func) _wrap_amp_caps_has_set_level_set, NULL},
    { SWIG_prefix "amp_caps_has_set_level_get", (swig_wrapper_func) _wrap_amp_caps_has_set_level_get, NULL},
    { SWIG_prefix "amp_caps_level_gran_set", (swig_wrapper_func) _wrap_amp_caps_level_gran_set, NULL},
    { SWIG_prefix "amp_caps_level_gran_get", (swig_wrapper_func) _wrap_amp_caps_level_gran_get, NULL},
    { SWIG_prefix "amp_caps_parm_gran_set", (swig_wrapper_func) _wrap_amp_caps_parm_gran_set, NULL},
    { SWIG_prefix "amp_caps_parm_gran_get", (swig_wrapper_func) _wrap_amp_caps_parm_gran_get, NULL},
    { SWIG_prefix "amp_caps_amp_init_set", (swig_wrapper_func) _wrap_amp_caps_amp_init_set, NULL},
    { SWIG_prefix "amp_caps_amp_init_get", (swig_wrapper_func) _wrap_amp_caps_amp_init_get, NULL},
    { SWIG_prefix "amp_caps_amp_cleanup_set", (swig_wrapper_func) _wrap_amp_caps_amp_cleanup_set, NULL},
    { SWIG_prefix "amp_caps_amp_cleanup_get", (swig_wrapper_func) _wrap_amp_caps_amp_cleanup_get, NULL},
    { SWIG_prefix "amp_caps_amp_open_set", (swig_wrapper_func) _wrap_amp_caps_amp_open_set, NULL},
    { SWIG_prefix "amp_caps_amp_open_get", (swig_wrapper_func) _wrap_amp_caps_amp_open_get, NULL},
    { SWIG_prefix "amp_caps_amp_close_set", (swig_wrapper_func) _wrap_amp_caps_amp_close_set, NULL},
    { SWIG_prefix "amp_caps_amp_close_get", (swig_wrapper_func) _wrap_amp_caps_amp_close_get, NULL},
    { SWIG_prefix "amp_caps_set_freq_set", (swig_wrapper_func) _wrap_amp_caps_set_freq_set, NULL},
    { SWIG_prefix "amp_caps_set_freq_get", (swig_wrapper_func) _wrap_amp_caps_set_freq_get, NULL},
    { SWIG_prefix "amp_caps_get_freq_set", (swig_wrapper_func) _wrap_amp_caps_get_freq_set, NULL},
    { SWIG_prefix "amp_caps_get_freq_get", (swig_wrapper_func) _wrap_amp_caps_get_freq_get, NULL},
    { SWIG_prefix "amp_caps_set_conf_set", (swig_wrapper_func) _wrap_amp_caps_set_conf_set, NULL},
    { SWIG_prefix "amp_caps_set_conf_get", (swig_wrapper_func) _wrap_amp_caps_set_conf_get, NULL},
    { SWIG_prefix "amp_caps_get_conf2_set", (swig_wrapper_func) _wrap_amp_caps_get_conf2_set, NULL},
    { SWIG_prefix "amp_caps_get_conf2_get", (swig_wrapper_func) _wrap_amp_caps_get_conf2_get, NULL},
    { SWIG_prefix "amp_caps_get_conf_set", (swig_wrapper_func) _wrap_amp_caps_get_conf_set, NULL},
    { SWIG_prefix "amp_caps_get_conf_get", (swig_wrapper_func) _wrap_amp_caps_get_conf_get, NULL},
    { SWIG_prefix "amp_caps_reset_set", (swig_wrapper_func) _wrap_amp_caps_reset_set, NULL},
    { SWIG_prefix "amp_caps_reset_get", (swig_wrapper_func) _wrap_amp_caps_reset_get, NULL},
    { SWIG_prefix "amp_caps_get_level_set", (swig_wrapper_func) _wrap_amp_caps_get_level_set, NULL},
    { SWIG_prefix "amp_caps_get_level_get", (swig_wrapper_func) _wrap_amp_caps_get_level_get, NULL},
    { SWIG_prefix "amp_caps_set_level_set", (swig_wrapper_func) _wrap_amp_caps_set_level_set, NULL},
    { SWIG_prefix "amp_caps_set_level_get", (swig_wrapper_func) _wrap_amp_caps_set_level_get, NULL},
    { SWIG_prefix "amp_caps_get_ext_level_set", (swig_wrapper_func) _wrap_amp_caps_get_ext_level_set, NULL},
    { SWIG_prefix "amp_caps_get_ext_level_get", (swig_wrapper_func) _wrap_amp_caps_get_ext_level_get, NULL},
    { SWIG_prefix "amp_caps_set_ext_level_set", (swig_wrapper_func) _wrap_amp_caps_set_ext_level_set, NULL},
    { SWIG_prefix "amp_caps_set_ext_level_get", (swig_wrapper_func) _wrap_amp_caps_set_ext_level_get, NULL},
    { SWIG_prefix "amp_caps_set_powerstat_set", (swig_wrapper_func) _wrap_amp_caps_set_powerstat_set, NULL},
    { SWIG_prefix "amp_caps_set_powerstat_get", (swig_wrapper_func) _wrap_amp_caps_set_powerstat_get, NULL},
    { SWIG_prefix "amp_caps_get_powerstat_set", (swig_wrapper_func) _wrap_amp_caps_get_powerstat_set, NULL},
    { SWIG_prefix "amp_caps_get_powerstat_get", (swig_wrapper_func) _wrap_amp_caps_get_powerstat_get, NULL},
    { SWIG_prefix "amp_caps_get_info_set", (swig_wrapper_func) _wrap_amp_caps_get_info_set, NULL},
    { SWIG_prefix "amp_caps_get_info_get", (swig_wrapper_func) _wrap_amp_caps_get_info_get, NULL},
    { SWIG_prefix "amp_caps_levels_set", (swig_wrapper_func) _wrap_amp_caps_levels_set, NULL},
    { SWIG_prefix "amp_caps_levels_get", (swig_wrapper_func) _wrap_amp_caps_levels_get, NULL},
    { SWIG_prefix "amp_caps_ext_levels_set", (swig_wrapper_func) _wrap_amp_caps_ext_levels_set, NULL},
    { SWIG_prefix "amp_caps_ext_levels_get", (swig_wrapper_func) _wrap_amp_caps_ext_levels_get, NULL},
    { SWIG_prefix "amp_caps_extlevels_set", (swig_wrapper_func) _wrap_amp_caps_extlevels_set, NULL},
    { SWIG_prefix "amp_caps_extlevels_get", (swig_wrapper_func) _wrap_amp_caps_extlevels_get, NULL},
    { SWIG_prefix "amp_caps_extparms_set", (swig_wrapper_func) _wrap_amp_caps_extparms_set, NULL},
    { SWIG_prefix "amp_caps_extparms_get", (swig_wrapper_func) _wrap_amp_caps_extparms_get, NULL},
    { SWIG_prefix "new_amp_caps", (swig_wrapper_func) _wrap_new_amp_caps, NULL},
    { SWIG_prefix "delete_amp_caps", (swig_wrapper_func) _wrap_delete_amp_caps, NULL},
    { SWIG_prefix "amp_caps", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_amp_caps},
    { SWIG_prefix "amp_state_comm_state_set", (swig_wrapper_func) _wrap_amp_state_comm_state_set, NULL},
    { SWIG_prefix "amp_state_comm_state_get", (swig_wrapper_func) _wrap_amp_state_comm_state_get, NULL},
    { SWIG_prefix "amp_state_priv_set", (swig_wrapper_func) _wrap_amp_state_priv_set, NULL},
    { SWIG_prefix "amp_state_priv_get", (swig_wrapper_func) _wrap_amp_state_priv_get, NULL},
    { SWIG_prefix "amp_state_obj_set", (swig_wrapper_func) _wrap_amp_state_obj_set, NULL},
    { SWIG_prefix "amp_state_obj_get", (swig_wrapper_func) _wrap_amp_state_obj_get, NULL},
    { SWIG_prefix "amp_state_has_get_level_set", (swig_wrapper_func) _wrap_amp_state_has_get_level_set, NULL},
    { SWIG_prefix "amp_state_has_get_level_get", (swig_wrapper_func) _wrap_amp_state_has_get_level_get, NULL},
    { SWIG_prefix "amp_state_has_set_level_set", (swig_wrapper_func) _wrap_amp_state_has_set_level_set, NULL},
    { SWIG_prefix "amp_state_has_set_level_get", (swig_wrapper_func) _wrap_amp_state_has_set_level_get, NULL},
    { SWIG_prefix "amp_state_level_gran_set", (swig_wrapper_func) _wrap_amp_state_level_gran_set, NULL},
    { SWIG_prefix "amp_state_level_gran_get", (swig_wrapper_func) _wrap_amp_state_level_gran_get, NULL},
    { SWIG_prefix "amp_state_parm_gran_set", (swig_wrapper_func) _wrap_amp_state_parm_gran_set, NULL},
    { SWIG_prefix "amp_state_parm_gran_get", (swig_wrapper_func) _wrap_amp_state_parm_gran_get, NULL},
    { SWIG_prefix "amp_state_ampport_set", (swig_wrapper_func) _wrap_amp_state_ampport_set, NULL},
    { SWIG_prefix "amp_state_ampport_get", (swig_wrapper_func) _wrap_amp_state_ampport_get, NULL},
    { SWIG_prefix "new_amp_state", (swig_wrapper_func) _wrap_new_amp_state, NULL},
    { SWIG_prefix "delete_amp_state", (swig_wrapper_func) _wrap_delete_amp_state, NULL},
    { SWIG_prefix "amp_state", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_amp_state},
    { SWIG_prefix "Amp_caps_set", (swig_wrapper_func) _wrap_Amp_caps_set, NULL},
    { SWIG_prefix "Amp_caps_get", (swig_wrapper_func) _wrap_Amp_caps_get, NULL},
    { SWIG_prefix "Amp_state_set", (swig_wrapper_func) _wrap_Amp_state_set, NULL},
    { SWIG_prefix "Amp_state_get", (swig_wrapper_func) _wrap_Amp_state_get, NULL},
    { SWIG_prefix "Amp_error_status_set", (swig_wrapper_func) _wrap_Amp_error_status_set, NULL},
    { SWIG_prefix "Amp_error_status_get", (swig_wrapper_func) _wrap_Amp_error_status_get, NULL},
    { SWIG_prefix "Amp_do_exception_set", (swig_wrapper_func) _wrap_Amp_do_exception_set, NULL},
    { SWIG_prefix "Amp_do_exception_get", (swig_wrapper_func) _wrap_Amp_do_exception_get, NULL},
    { SWIG_prefix "new_Amp", (swig_wrapper_func) _wrap_new_Amp, NULL},
    { SWIG_prefix "delete_Amp", (swig_wrapper_func) _wrap_delete_Amp, NULL},
    { SWIG_prefix "Amp_open", (swig_wrapper_func) _wrap_Amp_open, NULL},
    { SWIG_prefix "Amp_close", (swig_wrapper_func) _wrap_Amp_close, NULL},
    { SWIG_prefix "Amp_reset", (swig_wrapper_func) _wrap_Amp_reset, NULL},
    { SWIG_prefix "Amp_token_lookup", (swig_wrapper_func) _wrap_Amp_token_lookup, NULL},
    { SWIG_prefix "Amp_has_get_level", (swig_wrapper_func) _wrap_Amp_has_get_level, NULL},
    { SWIG_prefix "Amp_has_set_level", (swig_wrapper_func) _wrap_Amp_has_set_level, NULL},
    { SWIG_prefix "Amp_set_freq", (swig_wrapper_func) _wrap_Amp_set_freq, NULL},
    { SWIG_prefix "Amp_set_powerstat", (swig_wrapper_func) _wrap_Amp_set_powerstat, NULL},
    { SWIG_prefix "Amp_set_level", (swig_wrapper_func) _wrap_Amp_set_level, NULL},
    { SWIG_prefix "Amp_set_ext_level", (swig_wrapper_func) _wrap_Amp_set_ext_level, NULL},
    { SWIG_prefix "Amp_set_conf", (swig_wrapper_func) _wrap_Amp_set_conf, NULL},
    { SWIG_prefix "Amp_get_conf", (swig_wrapper_func) _wrap_Amp_get_conf, NULL},
    { SWIG_prefix "Amp_get_freq", (swig_wrapper_func) _wrap_Amp_get_freq, NULL},
    { SWIG_prefix "Amp_get_info", (swig_wrapper_func) _wrap_Amp_get_info, NULL},
    { SWIG_prefix "Amp_get_powerstat", (swig_wrapper_func) _wrap_Amp_get_powerstat, NULL},
    { SWIG_prefix "Amp", (swig_wrapper_func) SWIG_ObjectConstructor, (ClientData)&_wrap_class_Amp},
    {0, 0, 0}
};

static swig_var_info swig_variables[] = {
    { SWIG_prefix "cookie_use", 0, (swig_variable_func) _wrap_cookie_use_get,(swig_variable_func) _wrap_cookie_use_set},
    { SWIG_prefix "skip_init", 0, (swig_variable_func) _wrap_skip_init_get,(swig_variable_func) _wrap_skip_init_set},
    { SWIG_prefix "hamlib_version", 0, (swig_variable_func) _wrap_hamlib_version_get,(swig_variable_func) swig_readonly},
    { SWIG_prefix "hamlib_copyright", 0, (swig_variable_func) _wrap_hamlib_copyright_get,(swig_variable_func) swig_readonly},
    {0,0,0,0}
};

static swig_const_info swig_constants[] = {
    {0,0,0,0,0,0}
};

/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static void *_p_channelArrayTo_p_channel(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((struct channel *)  ((channelArray *) x));
}
static swig_type_info _swigt__p_Amp = {"_p_Amp", "Amp *|struct Amp *", 0, 0, (void*)&_wrap_class_Amp, 0};
static swig_type_info _swigt__p_FIFO_RIG_s = {"_p_FIFO_RIG_s", "struct FIFO_RIG_s *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Rig = {"_p_Rig", "Rig *|struct Rig *", 0, 0, (void*)&_wrap_class_Rig, 0};
static swig_type_info _swigt__p_Rot = {"_p_Rot", "Rot *|struct Rot *", 0, 0, (void*)&_wrap_class_Rot, 0};
static swig_type_info _swigt__p_agc_level_e = {"_p_agc_level_e", "enum agc_level_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_amp = {"_p_amp", "AMP *|struct amp *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_amp_caps = {"_p_amp_caps", "struct amp_caps *|amp_caps *", 0, 0, (void*)&_wrap_class_amp_caps, 0};
static swig_type_info _swigt__p_amp_reset_e = {"_p_amp_reset_e", "amp_reset_t *|enum amp_reset_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_amp_state = {"_p_amp_state", "struct amp_state *|amp_state *", 0, 0, (void*)&_wrap_class_amp_state, 0};
static swig_type_info _swigt__p_amp_type_e = {"_p_amp_type_e", "amp_type_t *|enum amp_type_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ann_t = {"_p_ann_t", "ann_t *|enum ann_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_cal_table = {"_p_cal_table", "cal_table_t *|struct cal_table *|cal_table *", 0, 0, (void*)&_wrap_class_cal_table, 0};
static swig_type_info _swigt__p_cal_table_float = {"_p_cal_table_float", "cal_table_float_t *|struct cal_table_float *|cal_table_float *", 0, 0, (void*)&_wrap_class_cal_table_float, 0};
static swig_type_info _swigt__p_cal_table_float_table = {"_p_cal_table_float_table", "cal_table_float_table *", 0, 0, (void*)&_wrap_class_cal_table_float_table, 0};
static swig_type_info _swigt__p_cal_table_table = {"_p_cal_table_table", "cal_table_table *", 0, 0, (void*)&_wrap_class_cal_table_table, 0};
static swig_type_info _swigt__p_chan_list = {"_p_chan_list", "chan_t *|struct chan_list *|chan_list *", 0, 0, (void*)&_wrap_class_chan_list, 0};
static swig_type_info _swigt__p_chan_type_t = {"_p_chan_type_t", "chan_type_t *|enum chan_type_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_channel = {"_p_channel", "channel_t *|struct channel *|channel *", 0, 0, (void*)&_wrap_class_channel, 0};
static swig_type_info _swigt__p_channelArray = {"_p_channelArray", "channelArray *", 0, 0, (void*)&_wrap_class_channelArray, 0};
static swig_type_info _swigt__p_channel_cap = {"_p_channel_cap", "channel_cap_t *|struct channel_cap *|channel_cap *", 0, 0, (void*)&_wrap_class_channel_cap, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_client_e = {"_p_client_e", "client_t *|enum client_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_confparams = {"_p_confparams", "struct confparams *|confparams *", 0, 0, (void*)&_wrap_class_confparams, 0};
static swig_type_info _swigt__p_confparams_u = {"_p_confparams_u", "confparams_u *", 0, 0, (void*)&_wrap_class_confparams_u, 0};
static swig_type_info _swigt__p_confparams_u_c = {"_p_confparams_u_c", "confparams_u_c *", 0, 0, (void*)&_wrap_class_confparams_u_c, 0};
static swig_type_info _swigt__p_confparams_u_n = {"_p_confparams_u_n", "confparams_u_n *", 0, 0, (void*)&_wrap_class_confparams_u_n, 0};
static swig_type_info _swigt__p_dcd_e = {"_p_dcd_e", "dcd_t *|enum dcd_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_dcd_type_e = {"_p_dcd_type_e", "dcd_type_t *|enum dcd_type_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_deferred_config_header = {"_p_deferred_config_header", "deferred_config_header_t *|struct deferred_config_header *|deferred_config_header *", 0, 0, (void*)&_wrap_class_deferred_config_header, 0};
static swig_type_info _swigt__p_deferred_config_item = {"_p_deferred_config_item", "deferred_config_item_t *|struct deferred_config_item *|deferred_config_item *", 0, 0, (void*)&_wrap_class_deferred_config_item, 0};
static swig_type_info _swigt__p_double = {"_p_double", "freq_t *|double *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ext_list = {"_p_ext_list", "struct ext_list *|ext_list *", 0, 0, (void*)&_wrap_class_ext_list, 0};
static swig_type_info _swigt__p_f_p_struct_amp__int = {"_p_f_p_struct_amp__int", "int (*)(AMP *)|int (*)(struct amp *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp__p_char = {"_p_f_p_struct_amp__p_char", "char *(*)(AMP *)|char *(*)(struct amp *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_double__int = {"_p_f_p_struct_amp_double__int", "int (*)(AMP *,freq_t)|int (*)(struct amp *,double)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_enum_amp_reset_e__int = {"_p_f_p_struct_amp_enum_amp_reset_e__int", "int (*)(AMP *,amp_reset_t)|int (*)(struct amp *,enum amp_reset_e)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_enum_powerstat_t__int = {"_p_f_p_struct_amp_enum_powerstat_t__int", "int (*)(AMP *,powerstat_t)|int (*)(struct amp *,enum powerstat_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_long_p_char__int = {"_p_f_p_struct_amp_long_p_char__int", "int (*)(AMP *,hamlib_token_t,char *)|int (*)(struct amp *,long,char *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_long_p_char_int__int = {"_p_f_p_struct_amp_long_p_char_int__int", "int (*)(AMP *,hamlib_token_t,char *,int)|int (*)(struct amp *,long,char *,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_long_p_q_const__char__int = {"_p_f_p_struct_amp_long_p_q_const__char__int", "int (*)(AMP *,hamlib_token_t,char const *)|int (*)(struct amp *,long,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_long_p_value_t__int = {"_p_f_p_struct_amp_long_p_value_t__int", "int (*)(AMP *,hamlib_token_t,value_t *)|int (*)(struct amp *,long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_long_value_t__int = {"_p_f_p_struct_amp_long_value_t__int", "int (*)(AMP *,hamlib_token_t,value_t)|int (*)(struct amp *,long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_p_double__int = {"_p_f_p_struct_amp_p_double__int", "int (*)(AMP *,freq_t *)|int (*)(struct amp *,double *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_p_enum_powerstat_t__int = {"_p_f_p_struct_amp_p_enum_powerstat_t__int", "int (*)(AMP *,powerstat_t *)|int (*)(struct amp *,enum powerstat_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_unsigned_long_long_p_value_t__int = {"_p_f_p_struct_amp_unsigned_long_long_p_value_t__int", "int (*)(AMP *,setting_t,value_t *)|int (*)(struct amp *,unsigned long long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_amp_unsigned_long_long_value_t__int = {"_p_f_p_struct_amp_unsigned_long_long_value_t__int", "int (*)(AMP *,setting_t,value_t)|int (*)(struct amp *,unsigned long long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig__int = {"_p_f_p_struct_s_rig__int", "int (*)(RIG *)|int (*)(struct s_rig *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig__p_char = {"_p_f_p_struct_s_rig__p_char", "char *(*)(RIG *)|char *(*)(struct s_rig *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_enum_powerstat_t__int = {"_p_f_p_struct_s_rig_enum_powerstat_t__int", "int (*)(RIG *,powerstat_t)|int (*)(struct s_rig *,enum powerstat_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_enum_reset_t__int = {"_p_f_p_struct_s_rig_enum_reset_t__int", "int (*)(RIG *,reset_t)|int (*)(struct s_rig *,enum reset_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_int__int = {"_p_f_p_struct_s_rig_int__int", "int (*)(RIG *,int)|int (*)(struct s_rig *,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int = {"_p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int", "int (*)(RIG *,int,int,int,int,int,int,double,int)|int (*)(struct s_rig *,int,int,int,int,int,int,double,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_long_p_char__int = {"_p_f_p_struct_s_rig_long_p_char__int", "int (*)(RIG *,hamlib_token_t,char *)|int (*)(struct s_rig *,long,char *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_long_p_char_int__int = {"_p_f_p_struct_s_rig_long_p_char_int__int", "int (*)(RIG *,hamlib_token_t,char *,int)|int (*)(struct s_rig *,long,char *,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_long_p_q_const__char__int = {"_p_f_p_struct_s_rig_long_p_q_const__char__int", "int (*)(RIG *,hamlib_token_t,char const *)|int (*)(struct s_rig *,long,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_long_p_value_t__int = {"_p_f_p_struct_s_rig_long_p_value_t__int", "int (*)(RIG *,hamlib_token_t,value_t *)|int (*)(struct s_rig *,long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_long_value_t__int = {"_p_f_p_struct_s_rig_long_value_t__int", "int (*)(RIG *,hamlib_token_t,value_t)|int (*)(struct s_rig *,long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_enum_powerstat_t__int = {"_p_f_p_struct_s_rig_p_enum_powerstat_t__int", "int (*)(RIG *,powerstat_t *)|int (*)(struct s_rig *,enum powerstat_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int = {"_p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int", "int (*)(RIG *,float *,unsigned int,freq_t,rmode_t)|int (*)(struct s_rig *,float *,unsigned int,double,unsigned long long)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_int__int = {"_p_f_p_struct_s_rig_p_int__int", "int (*)(RIG *,int *)|int (*)(struct s_rig *,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int = {"_p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int", "int (*)(RIG *,int *,int *,int *,int *,int *,int *,double *,int *)|int (*)(struct s_rig *,int *,int *,int *,int *,int *,int *,double *,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_q_const__char__int = {"_p_f_p_struct_s_rig_p_q_const__char__int", "int (*)(RIG *,char const *)|int (*)(struct s_rig *,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int = {"_p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int", "spectrum_cb_t|int (*)(struct s_rig *,struct rig_spectrum_line *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_unsigned_int__int = {"_p_f_p_struct_s_rig_p_unsigned_int__int", "int (*)(RIG *,vfo_t *)|int (*)(struct s_rig *,unsigned int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int = {"_p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int", "int (*)(RIG *,unsigned int *,float,freq_t,rmode_t)|int (*)(struct s_rig *,unsigned int *,float,double,unsigned long long)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int = {"_p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int", "int (*)(RIG *,size_t,unsigned char const *)|int (*)(struct s_rig *,size_t,unsigned char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int__int = {"_p_f_p_struct_s_rig_unsigned_int__int", "int (*)(RIG *,vfo_t)|int (*)(struct s_rig *,unsigned int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_double__int = {"_p_f_p_struct_s_rig_unsigned_int_double__int", "int (*)(RIG *,vfo_t,freq_t)|int (*)(struct s_rig *,unsigned int,double)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_double_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_double_p_void__int", "freq_cb_t|int (*)(struct s_rig *,unsigned int,double,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int = {"_p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int", "int (*)(RIG *,vfo_t,freq_t,rmode_t,pbwidth_t)|int (*)(struct s_rig *,unsigned int,double,unsigned long long,long)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int", "dcd_cb_t|int (*)(struct s_rig *,unsigned int,enum dcd_e,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int", "int (*)(RIG *,vfo_t,ptt_t)|int (*)(struct s_rig *,unsigned int,enum ptt_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int", "ptt_cb_t|int (*)(struct s_rig *,unsigned int,enum ptt_t,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int", "int (*)(RIG *,vfo_t,rptr_shift_t)|int (*)(struct s_rig *,unsigned int,enum rptr_shift_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int", "int (*)(RIG *,vfo_t,scan_t,int)|int (*)(struct s_rig *,unsigned int,enum scan_t,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int", "int (*)(RIG *,vfo_t,split_t,vfo_t)|int (*)(struct s_rig *,unsigned int,enum split_t,unsigned int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int = {"_p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int", "int (*)(RIG *,vfo_t,vfo_op_t)|int (*)(struct s_rig *,unsigned int,enum vfo_op_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_int__int = {"_p_f_p_struct_s_rig_unsigned_int_int__int", "int (*)(RIG *,vfo_t,int)|int (*)(struct s_rig *,unsigned int,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_long__int = {"_p_f_p_struct_s_rig_unsigned_int_long__int", "int (*)(RIG *,vfo_t,shortfreq_t)|int (*)(struct s_rig *,unsigned int,long)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_long_int__int = {"_p_f_p_struct_s_rig_unsigned_int_long_int__int", "int (*)(RIG *,vfo_t,hamlib_token_t,int)|int (*)(struct s_rig *,unsigned int,long,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_long_p_int__int = {"_p_f_p_struct_s_rig_unsigned_int_long_p_int__int", "int (*)(RIG *,vfo_t,hamlib_token_t,int *)|int (*)(struct s_rig *,unsigned int,long,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int = {"_p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int", "int (*)(RIG *,vfo_t,hamlib_token_t,value_t *)|int (*)(struct s_rig *,unsigned int,long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_long_value_t__int = {"_p_f_p_struct_s_rig_unsigned_int_long_value_t__int", "int (*)(RIG *,vfo_t,hamlib_token_t,value_t)|int (*)(struct s_rig *,unsigned int,long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int = {"_p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int", "int (*)(RIG *,vfo_t,char *,int *)|int (*)(struct s_rig *,unsigned int,char *,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_double__int = {"_p_f_p_struct_s_rig_unsigned_int_p_double__int", "int (*)(RIG *,vfo_t,freq_t *)|int (*)(struct s_rig *,unsigned int,double *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int = {"_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int", "int (*)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *)|int (*)(struct s_rig *,unsigned int,double *,unsigned long long *,long *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int = {"_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int", "int (*)(RIG *,vfo_t,freq_t *,rmode_t *,pbwidth_t *,split_t *)|int (*)(struct s_rig *,unsigned int,double *,unsigned long long *,long *,enum split_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int", "pltune_cb_t|int (*)(struct s_rig *,unsigned int,double *,unsigned long long *,long *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int = {"_p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int", "int (*)(RIG *,vfo_t,dcd_t *)|int (*)(struct s_rig *,unsigned int,enum dcd_e *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int = {"_p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int", "int (*)(RIG *,vfo_t,ptt_t *)|int (*)(struct s_rig *,unsigned int,enum ptt_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int = {"_p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int", "int (*)(RIG *,vfo_t,rptr_shift_t *)|int (*)(struct s_rig *,unsigned int,enum rptr_shift_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int = {"_p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int", "int (*)(RIG *,vfo_t,split_t *,vfo_t *)|int (*)(struct s_rig *,unsigned int,enum split_t *,unsigned int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int", "int (*)(RIG *,vfo_t,chan_cb_t,confval_cb_t,void *)|int (*)(struct s_rig *,unsigned int,int (*)(struct s_rig *,unsigned int,struct channel **,int,struct chan_list const *,void *),int (*)(struct s_rig *,struct confparams const *,value_t *,void *),void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int", "int (*)(RIG *,vfo_t,chan_cb_t,void *)|int (*)(struct s_rig *,unsigned int,int (*)(struct s_rig *,unsigned int,struct channel **,int,struct chan_list const *,void *),void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_int__int = {"_p_f_p_struct_s_rig_unsigned_int_p_int__int", "int (*)(RIG *,vfo_t,int *)|int (*)(struct s_rig *,unsigned int,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_long__int = {"_p_f_p_struct_s_rig_unsigned_int_p_long__int", "int (*)(RIG *,vfo_t,shortfreq_t *)|int (*)(struct s_rig *,unsigned int,long *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int = {"_p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int", "int (*)(RIG *,vfo_t,char const *)|int (*)(struct s_rig *,unsigned int,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int = {"_p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int", "int (*)(RIG *,vfo_t,channel_t const *)|int (*)(struct s_rig *,unsigned int,struct channel const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int = {"_p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int", "int (*)(RIG *,vfo_t,channel_t *,int)|int (*)(struct s_rig *,unsigned int,struct channel *,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int = {"_p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int", "int (*)(RIG *,vfo_t,tone_t *)|int (*)(struct s_rig *,unsigned int,unsigned int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int = {"_p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int", "int (*)(RIG *,vfo_t,rmode_t *,pbwidth_t *)|int (*)(struct s_rig *,unsigned int,unsigned long long *,long *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_p_void__int", "vfo_cb_t|int (*)(struct s_rig *,unsigned int,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_int__int", "int (*)(RIG *,vfo_t,tone_t)|int (*)(struct s_rig *,unsigned int,unsigned int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int", "int (*)(RIG *,vfo_t,ant_t,value_t *,ant_t *,ant_t *,ant_t *)|int (*)(struct s_rig *,unsigned int,unsigned int,value_t *,unsigned int *,unsigned int *,unsigned int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int", "int (*)(RIG *,vfo_t,ant_t,value_t)|int (*)(struct s_rig *,unsigned int,unsigned int,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int", "int (*)(RIG *,vfo_t,setting_t,int)|int (*)(struct s_rig *,unsigned int,unsigned long long,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int", "int (*)(RIG *,vfo_t,rmode_t,pbwidth_t)|int (*)(struct s_rig *,unsigned int,unsigned long long,long)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int", "mode_cb_t|int (*)(struct s_rig *,unsigned int,unsigned long long,long,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int", "int (*)(RIG *,vfo_t,setting_t,int *)|int (*)(struct s_rig *,unsigned int,unsigned long long,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int", "int (*)(RIG *,vfo_t,setting_t,value_t *)|int (*)(struct s_rig *,unsigned int,unsigned long long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int = {"_p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int", "int (*)(RIG *,vfo_t,setting_t,value_t)|int (*)(struct s_rig *,unsigned int,unsigned long long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int = {"_p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int", "int (*)(RIG *,setting_t,value_t *)|int (*)(struct s_rig *,unsigned long long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rig_unsigned_long_long_value_t__int = {"_p_f_p_struct_s_rig_unsigned_long_long_value_t__int", "int (*)(RIG *,setting_t,value_t)|int (*)(struct s_rig *,unsigned long long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot__int = {"_p_f_p_struct_s_rot__int", "int (*)(ROT *)|int (*)(struct s_rot *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot__p_char = {"_p_f_p_struct_s_rot__p_char", "char *(*)(ROT *)|char *(*)(struct s_rot *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_float_float__int = {"_p_f_p_struct_s_rot_float_float__int", "int (*)(ROT *,azimuth_t,elevation_t)|int (*)(struct s_rot *,float,float)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_int__int = {"_p_f_p_struct_s_rot_int__int", "int (*)(ROT *,rot_reset_t)|int (*)(struct s_rot *,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_int_int__int = {"_p_f_p_struct_s_rot_int_int__int", "int (*)(ROT *,int,int)|int (*)(struct s_rot *,int,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_int__int = {"_p_f_p_struct_s_rot_long_int__int", "int (*)(ROT *,hamlib_token_t,int)|int (*)(struct s_rot *,long,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_p_char__int = {"_p_f_p_struct_s_rot_long_p_char__int", "int (*)(ROT *,hamlib_token_t,char *)|int (*)(struct s_rot *,long,char *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_p_char_int__int = {"_p_f_p_struct_s_rot_long_p_char_int__int", "int (*)(ROT *,hamlib_token_t,char *,int)|int (*)(struct s_rot *,long,char *,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_p_int__int = {"_p_f_p_struct_s_rot_long_p_int__int", "int (*)(ROT *,hamlib_token_t,int *)|int (*)(struct s_rot *,long,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_p_q_const__char__int = {"_p_f_p_struct_s_rot_long_p_q_const__char__int", "int (*)(ROT *,hamlib_token_t,char const *)|int (*)(struct s_rot *,long,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_p_value_t__int = {"_p_f_p_struct_s_rot_long_p_value_t__int", "int (*)(ROT *,hamlib_token_t,value_t *)|int (*)(struct s_rot *,long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_long_value_t__int = {"_p_f_p_struct_s_rot_long_value_t__int", "int (*)(ROT *,hamlib_token_t,value_t)|int (*)(struct s_rot *,long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_p_enum_rot_status_t__int = {"_p_f_p_struct_s_rot_p_enum_rot_status_t__int", "int (*)(ROT *,rot_status_t *)|int (*)(struct s_rot *,enum rot_status_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_p_float_p_float__int = {"_p_f_p_struct_s_rot_p_float_p_float__int", "int (*)(ROT *,azimuth_t *,elevation_t *)|int (*)(struct s_rot *,float *,float *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_unsigned_long_long_int__int = {"_p_f_p_struct_s_rot_unsigned_long_long_int__int", "int (*)(ROT *,setting_t,int)|int (*)(struct s_rot *,unsigned long long,int)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_unsigned_long_long_p_int__int = {"_p_f_p_struct_s_rot_unsigned_long_long_p_int__int", "int (*)(ROT *,setting_t,int *)|int (*)(struct s_rot *,unsigned long long,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int = {"_p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int", "int (*)(ROT *,setting_t,value_t *)|int (*)(struct s_rot *,unsigned long long,value_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_struct_s_rot_unsigned_long_long_value_t__int = {"_p_f_p_struct_s_rot_unsigned_long_long_value_t__int", "int (*)(ROT *,setting_t,value_t)|int (*)(struct s_rot *,unsigned long long,value_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_filter_list = {"_p_filter_list", "struct filter_list *|filter_list *", 0, 0, (void*)&_wrap_class_filter_list, 0};
static swig_type_info _swigt__p_float = {"_p_float", "agc_time_t *|azimuth_t *|elevation_t *|swr_t *|float *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_freq_range_list = {"_p_freq_range_list", "freq_range_t *|struct freq_range_list *|freq_range_list *", 0, 0, (void*)&_wrap_class_freq_range_list, 0};
static swig_type_info _swigt__p_gran = {"_p_gran", "gran_t *|struct gran *|gran *", 0, 0, (void*)&_wrap_class_gran, 0};
static swig_type_info _swigt__p_hamlib_async_pipe = {"_p_hamlib_async_pipe", "hamlib_async_pipe_t *|struct hamlib_async_pipe *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_hamlib_band_t = {"_p_hamlib_band_t", "hamlib_band_t *|enum hamlib_band_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_hamlib_bandselect_t = {"_p_hamlib_bandselect_t", "hamlib_bandselect_t *|enum hamlib_bandselect_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_hamlib_cache_t = {"_p_hamlib_cache_t", "hamlib_cache_t *|enum hamlib_cache_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_hamlib_port = {"_p_hamlib_port", "hamlib_port_t *|struct hamlib_port *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "amp_model_t *|int32_t *|int_fast16_t *|int_fast32_t *|int_least32_t *|intptr_t *|rot_model_t *|rot_reset_t *|tune_value_t *|int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_long = {"_p_long", "hamlib_token_t *|pbwidth_t *|shortfreq_t *|long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_long_long = {"_p_long_long", "int64_t *|int_fast64_t *|int_least64_t *|intmax_t *|long long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_multicast_s = {"_p_multicast_s", "struct multicast_s *|multicast_s *", 0, 0, (void*)&_wrap_class_multicast_s, 0};
static swig_type_info _swigt__p_p_char = {"_p_p_char", "char **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_void = {"_p_p_void", "void **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_powerstat_t = {"_p_powerstat_t", "powerstat_t *|enum powerstat_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_pthread_mutex_t = {"_p_pthread_mutex_t", "pthread_mutex_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_pthread_t = {"_p_pthread_t", "pthread_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ptt_t = {"_p_ptt_t", "ptt_t *|enum ptt_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ptt_type_e = {"_p_ptt_type_e", "ptt_type_t *|enum ptt_type_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_reset_t = {"_p_reset_t", "reset_t *|enum reset_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_callbacks = {"_p_rig_callbacks", "struct rig_callbacks *|rig_callbacks *", 0, 0, (void*)&_wrap_class_rig_callbacks, 0};
static swig_type_info _swigt__p_rig_caps = {"_p_rig_caps", "struct rig_caps *|rig_caps *", 0, 0, (void*)&_wrap_class_rig_caps, 0};
static swig_type_info _swigt__p_rig_debug_level_e = {"_p_rig_debug_level_e", "enum rig_debug_level_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_port_e = {"_p_rig_port_e", "rig_port_t *|enum rig_port_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_ptrx_t = {"_p_rig_ptrx_t", "rig_ptrx_t *|enum rig_ptrx_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_spectrum_avg_mode = {"_p_rig_spectrum_avg_mode", "struct rig_spectrum_avg_mode *|rig_spectrum_avg_mode *", 0, 0, (void*)&_wrap_class_rig_spectrum_avg_mode, 0};
static swig_type_info _swigt__p_rig_spectrum_line = {"_p_rig_spectrum_line", "struct rig_spectrum_line *|rig_spectrum_line *", 0, 0, (void*)&_wrap_class_rig_spectrum_line, 0};
static swig_type_info _swigt__p_rig_spectrum_mode_e = {"_p_rig_spectrum_mode_e", "enum rig_spectrum_mode_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rig_spectrum_scope = {"_p_rig_spectrum_scope", "struct rig_spectrum_scope *|rig_spectrum_scope *", 0, 0, (void*)&_wrap_class_rig_spectrum_scope, 0};
static swig_type_info _swigt__p_rig_state = {"_p_rig_state", "struct rig_state *|rig_state *", 0, 0, (void*)&_wrap_class_rig_state, 0};
static swig_type_info _swigt__p_rig_type_t = {"_p_rig_type_t", "rig_type_t *|enum rig_type_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rot_caps = {"_p_rot_caps", "struct rot_caps *|rot_caps *", 0, 0, (void*)&_wrap_class_rot_caps, 0};
static swig_type_info _swigt__p_rot_state = {"_p_rot_state", "struct rot_state *|rot_state *", 0, 0, (void*)&_wrap_class_rot_state, 0};
static swig_type_info _swigt__p_rot_status_t = {"_p_rot_status_t", "rot_status_t *|enum rot_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rot_type_t = {"_p_rot_type_t", "rot_type_t *|enum rot_type_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_rptr_shift_t = {"_p_rptr_shift_t", "rptr_shift_t *|enum rptr_shift_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_s_rig = {"_p_s_rig", "RIG *|struct s_rig *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_s_rot = {"_p_s_rot", "ROT *|struct s_rot *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_scan_t = {"_p_scan_t", "scan_t *|enum scan_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_short = {"_p_short", "int16_t *|int_least16_t *|short *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_signed_char = {"_p_signed_char", "int8_t *|int_fast8_t *|int_least8_t *|signed char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_sockaddr_in = {"_p_sockaddr_in", "struct sockaddr_in *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_split_t = {"_p_split_t", "split_t *|enum split_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_time_t = {"_p_time_t", "time_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_timespec = {"_p_timespec", "struct timespec *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_toneArray = {"_p_toneArray", "toneArray *", 0, 0, (void*)&_wrap_class_toneArray, 0};
static swig_type_info _swigt__p_tuning_step_list = {"_p_tuning_step_list", "struct tuning_step_list *|tuning_step_list *", 0, 0, (void*)&_wrap_class_tuning_step_list, 0};
static swig_type_info _swigt__p_twiddle_state_t = {"_p_twiddle_state_t", "twiddle_state_t *|enum twiddle_state_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_char = {"_p_unsigned_char", "uint8_t *|uint_fast8_t *|uint_least8_t *|unsigned char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_int = {"_p_unsigned_int", "ant_t *|rig_comm_status_t *|rig_model_t *|tone_t *|uint32_t *|uint_fast16_t *|uint_fast32_t *|uint_least32_t *|uintptr_t *|vfo_t *|unsigned int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_long = {"_p_unsigned_long", "rig_useconds_t *|unsigned long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_long_long = {"_p_unsigned_long_long", "rig_level_e *|rmode_t *|setting_t *|uint64_t *|uint_fast64_t *|uint_least64_t *|uintmax_t *|unsigned long long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_short = {"_p_unsigned_short", "uint16_t *|uint_least16_t *|unsigned short *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_value_t = {"_p_value_t", "value_t *", 0, 0, (void*)&_wrap_class_value_t, 0};
static swig_type_info _swigt__p_value_t_b = {"_p_value_t_b", "value_t_b *", 0, 0, (void*)&_wrap_class_value_t_b, 0};
static swig_type_info _swigt__p_vfo_op_t = {"_p_vfo_op_t", "vfo_op_t *|enum vfo_op_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_void = {"_p_void", "void *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_Amp,
  &_swigt__p_FIFO_RIG_s,
  &_swigt__p_Rig,
  &_swigt__p_Rot,
  &_swigt__p_agc_level_e,
  &_swigt__p_amp,
  &_swigt__p_amp_caps,
  &_swigt__p_amp_reset_e,
  &_swigt__p_amp_state,
  &_swigt__p_amp_type_e,
  &_swigt__p_ann_t,
  &_swigt__p_cal_table,
  &_swigt__p_cal_table_float,
  &_swigt__p_cal_table_float_table,
  &_swigt__p_cal_table_table,
  &_swigt__p_chan_list,
  &_swigt__p_chan_type_t,
  &_swigt__p_channel,
  &_swigt__p_channelArray,
  &_swigt__p_channel_cap,
  &_swigt__p_char,
  &_swigt__p_client_e,
  &_swigt__p_confparams,
  &_swigt__p_confparams_u,
  &_swigt__p_confparams_u_c,
  &_swigt__p_confparams_u_n,
  &_swigt__p_dcd_e,
  &_swigt__p_dcd_type_e,
  &_swigt__p_deferred_config_header,
  &_swigt__p_deferred_config_item,
  &_swigt__p_double,
  &_swigt__p_ext_list,
  &_swigt__p_f_p_struct_amp__int,
  &_swigt__p_f_p_struct_amp__p_char,
  &_swigt__p_f_p_struct_amp_double__int,
  &_swigt__p_f_p_struct_amp_enum_amp_reset_e__int,
  &_swigt__p_f_p_struct_amp_enum_powerstat_t__int,
  &_swigt__p_f_p_struct_amp_long_p_char__int,
  &_swigt__p_f_p_struct_amp_long_p_char_int__int,
  &_swigt__p_f_p_struct_amp_long_p_q_const__char__int,
  &_swigt__p_f_p_struct_amp_long_p_value_t__int,
  &_swigt__p_f_p_struct_amp_long_value_t__int,
  &_swigt__p_f_p_struct_amp_p_double__int,
  &_swigt__p_f_p_struct_amp_p_enum_powerstat_t__int,
  &_swigt__p_f_p_struct_amp_unsigned_long_long_p_value_t__int,
  &_swigt__p_f_p_struct_amp_unsigned_long_long_value_t__int,
  &_swigt__p_f_p_struct_s_rig__int,
  &_swigt__p_f_p_struct_s_rig__p_char,
  &_swigt__p_f_p_struct_s_rig_enum_powerstat_t__int,
  &_swigt__p_f_p_struct_s_rig_enum_reset_t__int,
  &_swigt__p_f_p_struct_s_rig_int__int,
  &_swigt__p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int,
  &_swigt__p_f_p_struct_s_rig_long_p_char__int,
  &_swigt__p_f_p_struct_s_rig_long_p_char_int__int,
  &_swigt__p_f_p_struct_s_rig_long_p_q_const__char__int,
  &_swigt__p_f_p_struct_s_rig_long_p_value_t__int,
  &_swigt__p_f_p_struct_s_rig_long_value_t__int,
  &_swigt__p_f_p_struct_s_rig_p_enum_powerstat_t__int,
  &_swigt__p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int,
  &_swigt__p_f_p_struct_s_rig_p_int__int,
  &_swigt__p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int,
  &_swigt__p_f_p_struct_s_rig_p_q_const__char__int,
  &_swigt__p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int,
  &_swigt__p_f_p_struct_s_rig_p_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int,
  &_swigt__p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_double__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_double_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_long__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_long_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_long_p_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_long_value_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_double__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_long__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int,
  &_swigt__p_f_p_struct_s_rig_unsigned_long_long_value_t__int,
  &_swigt__p_f_p_struct_s_rot__int,
  &_swigt__p_f_p_struct_s_rot__p_char,
  &_swigt__p_f_p_struct_s_rot_float_float__int,
  &_swigt__p_f_p_struct_s_rot_int__int,
  &_swigt__p_f_p_struct_s_rot_int_int__int,
  &_swigt__p_f_p_struct_s_rot_long_int__int,
  &_swigt__p_f_p_struct_s_rot_long_p_char__int,
  &_swigt__p_f_p_struct_s_rot_long_p_char_int__int,
  &_swigt__p_f_p_struct_s_rot_long_p_int__int,
  &_swigt__p_f_p_struct_s_rot_long_p_q_const__char__int,
  &_swigt__p_f_p_struct_s_rot_long_p_value_t__int,
  &_swigt__p_f_p_struct_s_rot_long_value_t__int,
  &_swigt__p_f_p_struct_s_rot_p_enum_rot_status_t__int,
  &_swigt__p_f_p_struct_s_rot_p_float_p_float__int,
  &_swigt__p_f_p_struct_s_rot_unsigned_long_long_int__int,
  &_swigt__p_f_p_struct_s_rot_unsigned_long_long_p_int__int,
  &_swigt__p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int,
  &_swigt__p_f_p_struct_s_rot_unsigned_long_long_value_t__int,
  &_swigt__p_filter_list,
  &_swigt__p_float,
  &_swigt__p_freq_range_list,
  &_swigt__p_gran,
  &_swigt__p_hamlib_async_pipe,
  &_swigt__p_hamlib_band_t,
  &_swigt__p_hamlib_bandselect_t,
  &_swigt__p_hamlib_cache_t,
  &_swigt__p_hamlib_port,
  &_swigt__p_int,
  &_swigt__p_long,
  &_swigt__p_long_long,
  &_swigt__p_multicast_s,
  &_swigt__p_p_char,
  &_swigt__p_p_void,
  &_swigt__p_powerstat_t,
  &_swigt__p_pthread_mutex_t,
  &_swigt__p_pthread_t,
  &_swigt__p_ptt_t,
  &_swigt__p_ptt_type_e,
  &_swigt__p_reset_t,
  &_swigt__p_rig_callbacks,
  &_swigt__p_rig_caps,
  &_swigt__p_rig_debug_level_e,
  &_swigt__p_rig_port_e,
  &_swigt__p_rig_ptrx_t,
  &_swigt__p_rig_spectrum_avg_mode,
  &_swigt__p_rig_spectrum_line,
  &_swigt__p_rig_spectrum_mode_e,
  &_swigt__p_rig_spectrum_scope,
  &_swigt__p_rig_state,
  &_swigt__p_rig_type_t,
  &_swigt__p_rot_caps,
  &_swigt__p_rot_state,
  &_swigt__p_rot_status_t,
  &_swigt__p_rot_type_t,
  &_swigt__p_rptr_shift_t,
  &_swigt__p_s_rig,
  &_swigt__p_s_rot,
  &_swigt__p_scan_t,
  &_swigt__p_short,
  &_swigt__p_signed_char,
  &_swigt__p_sockaddr_in,
  &_swigt__p_split_t,
  &_swigt__p_time_t,
  &_swigt__p_timespec,
  &_swigt__p_toneArray,
  &_swigt__p_tuning_step_list,
  &_swigt__p_twiddle_state_t,
  &_swigt__p_unsigned_char,
  &_swigt__p_unsigned_int,
  &_swigt__p_unsigned_long,
  &_swigt__p_unsigned_long_long,
  &_swigt__p_unsigned_short,
  &_swigt__p_value_t,
  &_swigt__p_value_t_b,
  &_swigt__p_vfo_op_t,
  &_swigt__p_void,
};

static swig_cast_info _swigc__p_Amp[] = {  {&_swigt__p_Amp, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_FIFO_RIG_s[] = {  {&_swigt__p_FIFO_RIG_s, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Rig[] = {  {&_swigt__p_Rig, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Rot[] = {  {&_swigt__p_Rot, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_agc_level_e[] = {  {&_swigt__p_agc_level_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_amp[] = {  {&_swigt__p_amp, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_amp_caps[] = {  {&_swigt__p_amp_caps, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_amp_reset_e[] = {  {&_swigt__p_amp_reset_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_amp_state[] = {  {&_swigt__p_amp_state, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_amp_type_e[] = {  {&_swigt__p_amp_type_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ann_t[] = {  {&_swigt__p_ann_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_cal_table[] = {  {&_swigt__p_cal_table, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_cal_table_float[] = {  {&_swigt__p_cal_table_float, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_cal_table_float_table[] = {  {&_swigt__p_cal_table_float_table, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_cal_table_table[] = {  {&_swigt__p_cal_table_table, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_chan_list[] = {  {&_swigt__p_chan_list, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_chan_type_t[] = {  {&_swigt__p_chan_type_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_channel[] = {  {&_swigt__p_channel, 0, 0, 0},  {&_swigt__p_channelArray, _p_channelArrayTo_p_channel, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_channelArray[] = {  {&_swigt__p_channelArray, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_channel_cap[] = {  {&_swigt__p_channel_cap, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_client_e[] = {  {&_swigt__p_client_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_confparams[] = {  {&_swigt__p_confparams, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_confparams_u[] = {  {&_swigt__p_confparams_u, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_confparams_u_c[] = {  {&_swigt__p_confparams_u_c, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_confparams_u_n[] = {  {&_swigt__p_confparams_u_n, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_dcd_e[] = {  {&_swigt__p_dcd_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_dcd_type_e[] = {  {&_swigt__p_dcd_type_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_deferred_config_header[] = {  {&_swigt__p_deferred_config_header, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_deferred_config_item[] = {  {&_swigt__p_deferred_config_item, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_double[] = {  {&_swigt__p_double, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ext_list[] = {  {&_swigt__p_ext_list, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp__int[] = {  {&_swigt__p_f_p_struct_amp__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp__p_char[] = {  {&_swigt__p_f_p_struct_amp__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_double__int[] = {  {&_swigt__p_f_p_struct_amp_double__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_enum_amp_reset_e__int[] = {  {&_swigt__p_f_p_struct_amp_enum_amp_reset_e__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_enum_powerstat_t__int[] = {  {&_swigt__p_f_p_struct_amp_enum_powerstat_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_long_p_char__int[] = {  {&_swigt__p_f_p_struct_amp_long_p_char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_long_p_char_int__int[] = {  {&_swigt__p_f_p_struct_amp_long_p_char_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_long_p_q_const__char__int[] = {  {&_swigt__p_f_p_struct_amp_long_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_amp_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_long_value_t__int[] = {  {&_swigt__p_f_p_struct_amp_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_p_double__int[] = {  {&_swigt__p_f_p_struct_amp_p_double__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_p_enum_powerstat_t__int[] = {  {&_swigt__p_f_p_struct_amp_p_enum_powerstat_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_unsigned_long_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_amp_unsigned_long_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_amp_unsigned_long_long_value_t__int[] = {  {&_swigt__p_f_p_struct_amp_unsigned_long_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig__int[] = {  {&_swigt__p_f_p_struct_s_rig__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig__p_char[] = {  {&_swigt__p_f_p_struct_s_rig__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_enum_powerstat_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_enum_powerstat_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_enum_reset_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_enum_reset_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_long_p_char__int[] = {  {&_swigt__p_f_p_struct_s_rig_long_p_char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_long_p_char_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_long_p_char_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_long_p_q_const__char__int[] = {  {&_swigt__p_f_p_struct_s_rig_long_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_long_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_enum_powerstat_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_enum_powerstat_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_q_const__char__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int[] = {  {&_swigt__p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_double__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_double__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_double_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_double_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_long_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_long_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_long_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_long_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_long_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_double__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_double__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rig_unsigned_long_long_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rig_unsigned_long_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot__int[] = {  {&_swigt__p_f_p_struct_s_rot__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot__p_char[] = {  {&_swigt__p_f_p_struct_s_rot__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_float_float__int[] = {  {&_swigt__p_f_p_struct_s_rot_float_float__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_int_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_int_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_p_char__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_p_char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_p_char_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_p_char_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_p_q_const__char__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_long_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rot_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_p_enum_rot_status_t__int[] = {  {&_swigt__p_f_p_struct_s_rot_p_enum_rot_status_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_p_float_p_float__int[] = {  {&_swigt__p_f_p_struct_s_rot_p_float_p_float__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_unsigned_long_long_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_unsigned_long_long_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_unsigned_long_long_p_int__int[] = {  {&_swigt__p_f_p_struct_s_rot_unsigned_long_long_p_int__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_struct_s_rot_unsigned_long_long_value_t__int[] = {  {&_swigt__p_f_p_struct_s_rot_unsigned_long_long_value_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_filter_list[] = {  {&_swigt__p_filter_list, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_float[] = {  {&_swigt__p_float, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_freq_range_list[] = {  {&_swigt__p_freq_range_list, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_gran[] = {  {&_swigt__p_gran, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_hamlib_async_pipe[] = {  {&_swigt__p_hamlib_async_pipe, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_hamlib_band_t[] = {  {&_swigt__p_hamlib_band_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_hamlib_bandselect_t[] = {  {&_swigt__p_hamlib_bandselect_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_hamlib_cache_t[] = {  {&_swigt__p_hamlib_cache_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_hamlib_port[] = {  {&_swigt__p_hamlib_port, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_long[] = {  {&_swigt__p_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_long_long[] = {  {&_swigt__p_long_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_multicast_s[] = {  {&_swigt__p_multicast_s, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_char[] = {  {&_swigt__p_p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_void[] = {  {&_swigt__p_p_void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_powerstat_t[] = {  {&_swigt__p_powerstat_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_pthread_mutex_t[] = {  {&_swigt__p_pthread_mutex_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_pthread_t[] = {  {&_swigt__p_pthread_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ptt_t[] = {  {&_swigt__p_ptt_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ptt_type_e[] = {  {&_swigt__p_ptt_type_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_reset_t[] = {  {&_swigt__p_reset_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_callbacks[] = {  {&_swigt__p_rig_callbacks, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_caps[] = {  {&_swigt__p_rig_caps, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_debug_level_e[] = {  {&_swigt__p_rig_debug_level_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_port_e[] = {  {&_swigt__p_rig_port_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_ptrx_t[] = {  {&_swigt__p_rig_ptrx_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_spectrum_avg_mode[] = {  {&_swigt__p_rig_spectrum_avg_mode, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_spectrum_line[] = {  {&_swigt__p_rig_spectrum_line, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_spectrum_mode_e[] = {  {&_swigt__p_rig_spectrum_mode_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_spectrum_scope[] = {  {&_swigt__p_rig_spectrum_scope, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_state[] = {  {&_swigt__p_rig_state, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rig_type_t[] = {  {&_swigt__p_rig_type_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rot_caps[] = {  {&_swigt__p_rot_caps, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rot_state[] = {  {&_swigt__p_rot_state, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rot_status_t[] = {  {&_swigt__p_rot_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rot_type_t[] = {  {&_swigt__p_rot_type_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_rptr_shift_t[] = {  {&_swigt__p_rptr_shift_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_s_rig[] = {  {&_swigt__p_s_rig, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_s_rot[] = {  {&_swigt__p_s_rot, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_scan_t[] = {  {&_swigt__p_scan_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_short[] = {  {&_swigt__p_short, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_signed_char[] = {  {&_swigt__p_signed_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_sockaddr_in[] = {  {&_swigt__p_sockaddr_in, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_split_t[] = {  {&_swigt__p_split_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_time_t[] = {  {&_swigt__p_time_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_timespec[] = {  {&_swigt__p_timespec, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_toneArray[] = {  {&_swigt__p_toneArray, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_tuning_step_list[] = {  {&_swigt__p_tuning_step_list, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_twiddle_state_t[] = {  {&_swigt__p_twiddle_state_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_char[] = {  {&_swigt__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_int[] = {  {&_swigt__p_unsigned_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_long[] = {  {&_swigt__p_unsigned_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_long_long[] = {  {&_swigt__p_unsigned_long_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_short[] = {  {&_swigt__p_unsigned_short, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_value_t[] = {  {&_swigt__p_value_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_value_t_b[] = {  {&_swigt__p_value_t_b, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_vfo_op_t[] = {  {&_swigt__p_vfo_op_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_void[] = {  {&_swigt__p_void, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_Amp,
  _swigc__p_FIFO_RIG_s,
  _swigc__p_Rig,
  _swigc__p_Rot,
  _swigc__p_agc_level_e,
  _swigc__p_amp,
  _swigc__p_amp_caps,
  _swigc__p_amp_reset_e,
  _swigc__p_amp_state,
  _swigc__p_amp_type_e,
  _swigc__p_ann_t,
  _swigc__p_cal_table,
  _swigc__p_cal_table_float,
  _swigc__p_cal_table_float_table,
  _swigc__p_cal_table_table,
  _swigc__p_chan_list,
  _swigc__p_chan_type_t,
  _swigc__p_channel,
  _swigc__p_channelArray,
  _swigc__p_channel_cap,
  _swigc__p_char,
  _swigc__p_client_e,
  _swigc__p_confparams,
  _swigc__p_confparams_u,
  _swigc__p_confparams_u_c,
  _swigc__p_confparams_u_n,
  _swigc__p_dcd_e,
  _swigc__p_dcd_type_e,
  _swigc__p_deferred_config_header,
  _swigc__p_deferred_config_item,
  _swigc__p_double,
  _swigc__p_ext_list,
  _swigc__p_f_p_struct_amp__int,
  _swigc__p_f_p_struct_amp__p_char,
  _swigc__p_f_p_struct_amp_double__int,
  _swigc__p_f_p_struct_amp_enum_amp_reset_e__int,
  _swigc__p_f_p_struct_amp_enum_powerstat_t__int,
  _swigc__p_f_p_struct_amp_long_p_char__int,
  _swigc__p_f_p_struct_amp_long_p_char_int__int,
  _swigc__p_f_p_struct_amp_long_p_q_const__char__int,
  _swigc__p_f_p_struct_amp_long_p_value_t__int,
  _swigc__p_f_p_struct_amp_long_value_t__int,
  _swigc__p_f_p_struct_amp_p_double__int,
  _swigc__p_f_p_struct_amp_p_enum_powerstat_t__int,
  _swigc__p_f_p_struct_amp_unsigned_long_long_p_value_t__int,
  _swigc__p_f_p_struct_amp_unsigned_long_long_value_t__int,
  _swigc__p_f_p_struct_s_rig__int,
  _swigc__p_f_p_struct_s_rig__p_char,
  _swigc__p_f_p_struct_s_rig_enum_powerstat_t__int,
  _swigc__p_f_p_struct_s_rig_enum_reset_t__int,
  _swigc__p_f_p_struct_s_rig_int__int,
  _swigc__p_f_p_struct_s_rig_int_int_int_int_int_int_double_int__int,
  _swigc__p_f_p_struct_s_rig_long_p_char__int,
  _swigc__p_f_p_struct_s_rig_long_p_char_int__int,
  _swigc__p_f_p_struct_s_rig_long_p_q_const__char__int,
  _swigc__p_f_p_struct_s_rig_long_p_value_t__int,
  _swigc__p_f_p_struct_s_rig_long_value_t__int,
  _swigc__p_f_p_struct_s_rig_p_enum_powerstat_t__int,
  _swigc__p_f_p_struct_s_rig_p_float_unsigned_int_double_unsigned_long_long__int,
  _swigc__p_f_p_struct_s_rig_p_int__int,
  _swigc__p_f_p_struct_s_rig_p_int_p_int_p_int_p_int_p_int_p_int_p_double_p_int__int,
  _swigc__p_f_p_struct_s_rig_p_q_const__char__int,
  _swigc__p_f_p_struct_s_rig_p_struct_rig_spectrum_line_p_void__int,
  _swigc__p_f_p_struct_s_rig_p_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_p_unsigned_int_float_double_unsigned_long_long__int,
  _swigc__p_f_p_struct_s_rig_size_t_p_q_const__unsigned_char__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_double__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_double_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_double_unsigned_long_long_long__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_dcd_e_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_ptt_t_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_rptr_shift_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_scan_t_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_split_t_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_enum_vfo_op_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_long__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_long_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_long_p_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_long_p_value_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_long_value_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_char_p_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_double__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_enum_split_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_double_p_unsigned_long_long_p_long_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_dcd_e__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_ptt_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_rptr_shift_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_enum_split_t_p_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_f_p_struct_s_rig_p_q_const__struct_confparams_p_value_t_p_void__int_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_f_p_struct_s_rig_unsigned_int_p_p_struct_channel_int_p_q_const__struct_chan_list_p_void__int_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_long__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_q_const__char__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_q_const__struct_channel__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_struct_channel_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_unsigned_long_long_p_long__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_int_p_value_t_p_unsigned_int_p_unsigned_int_p_unsigned_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_int_value_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_long_p_void__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_int__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_p_value_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_int_unsigned_long_long_value_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_long_long_p_value_t__int,
  _swigc__p_f_p_struct_s_rig_unsigned_long_long_value_t__int,
  _swigc__p_f_p_struct_s_rot__int,
  _swigc__p_f_p_struct_s_rot__p_char,
  _swigc__p_f_p_struct_s_rot_float_float__int,
  _swigc__p_f_p_struct_s_rot_int__int,
  _swigc__p_f_p_struct_s_rot_int_int__int,
  _swigc__p_f_p_struct_s_rot_long_int__int,
  _swigc__p_f_p_struct_s_rot_long_p_char__int,
  _swigc__p_f_p_struct_s_rot_long_p_char_int__int,
  _swigc__p_f_p_struct_s_rot_long_p_int__int,
  _swigc__p_f_p_struct_s_rot_long_p_q_const__char__int,
  _swigc__p_f_p_struct_s_rot_long_p_value_t__int,
  _swigc__p_f_p_struct_s_rot_long_value_t__int,
  _swigc__p_f_p_struct_s_rot_p_enum_rot_status_t__int,
  _swigc__p_f_p_struct_s_rot_p_float_p_float__int,
  _swigc__p_f_p_struct_s_rot_unsigned_long_long_int__int,
  _swigc__p_f_p_struct_s_rot_unsigned_long_long_p_int__int,
  _swigc__p_f_p_struct_s_rot_unsigned_long_long_p_value_t__int,
  _swigc__p_f_p_struct_s_rot_unsigned_long_long_value_t__int,
  _swigc__p_filter_list,
  _swigc__p_float,
  _swigc__p_freq_range_list,
  _swigc__p_gran,
  _swigc__p_hamlib_async_pipe,
  _swigc__p_hamlib_band_t,
  _swigc__p_hamlib_bandselect_t,
  _swigc__p_hamlib_cache_t,
  _swigc__p_hamlib_port,
  _swigc__p_int,
  _swigc__p_long,
  _swigc__p_long_long,
  _swigc__p_multicast_s,
  _swigc__p_p_char,
  _swigc__p_p_void,
  _swigc__p_powerstat_t,
  _swigc__p_pthread_mutex_t,
  _swigc__p_pthread_t,
  _swigc__p_ptt_t,
  _swigc__p_ptt_type_e,
  _swigc__p_reset_t,
  _swigc__p_rig_callbacks,
  _swigc__p_rig_caps,
  _swigc__p_rig_debug_level_e,
  _swigc__p_rig_port_e,
  _swigc__p_rig_ptrx_t,
  _swigc__p_rig_spectrum_avg_mode,
  _swigc__p_rig_spectrum_line,
  _swigc__p_rig_spectrum_mode_e,
  _swigc__p_rig_spectrum_scope,
  _swigc__p_rig_state,
  _swigc__p_rig_type_t,
  _swigc__p_rot_caps,
  _swigc__p_rot_state,
  _swigc__p_rot_status_t,
  _swigc__p_rot_type_t,
  _swigc__p_rptr_shift_t,
  _swigc__p_s_rig,
  _swigc__p_s_rot,
  _swigc__p_scan_t,
  _swigc__p_short,
  _swigc__p_signed_char,
  _swigc__p_sockaddr_in,
  _swigc__p_split_t,
  _swigc__p_time_t,
  _swigc__p_timespec,
  _swigc__p_toneArray,
  _swigc__p_tuning_step_list,
  _swigc__p_twiddle_state_t,
  _swigc__p_unsigned_char,
  _swigc__p_unsigned_int,
  _swigc__p_unsigned_long,
  _swigc__p_unsigned_long_long,
  _swigc__p_unsigned_short,
  _swigc__p_value_t,
  _swigc__p_value_t_b,
  _swigc__p_vfo_op_t,
  _swigc__p_void,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

#ifdef __cplusplus
}
#endif
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is made tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast dependencies. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns.
 *
 * We loop through the cast[] array associated with the type and mark casts
 * which have not been defined in previously loaded modules by assigning
 * cast pointer value to cast->next. We also hash cast->type->name string
 * and store the value in the cast->value field. If we encounter swig_cast_info
 * structure that represents a cast to self we move it to the beginning
 * of the cast array. One trick we need to do is making sure the type pointer
 * in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the list (because the cast->type) pointer will
 *     be correct.
 *
 * Once the cast array has been set up AND it does have new casts that need
 * to be added we sort non-self cast entries to move filtered out entries
 * to the end of the array and to arrange the rest in the increasing order
 * of their type pointer values. We store the index of the last added entry
 * in the cast->value field of the entry[0] (overwriting the name hash). Then
 * we sort <next|value> fields of the remaining entries to arrange hash values
 * in the increasing order. This way cast->next->type->name field matches
 * the cast->value hash.
 *
 * Example:
 * Array of casts for type stored at 0x5000, cast to type stored at 0x3000
 * has already been loaded
 *
 *      After sweep-and-hash:   After sort-by-type:    After sort-by-hash:
 *        ________________       ________________       ________________
 *       |                |     |                |     |                |
 * Entry |  type = 0x5000 |     |  type = 0x5000 |     |  type = 0x5000 |
 *   0   |                |     |                |     |                |
 *       |  next = Entry0 |     |  next = Entry0 |     |  next = Entry0 |
 *       | value = 1212   |     | value = 3      |     | value = 3      |
 *       |                |     |                |     |                |
 *       |================|     |================|     |================|
 *       |                |     |                |     |                |
 * Entry |  type = 0x2000 |     |  type = 0x1000 |     |  type = 0x1000 |
 *   1   |                |     |                |     |                |
 *       |  next = Entry1 |     |  next = Entry1 |     |  next = Entry3 |
 *       | value = 3434   |     | value = 4545   |     | value = 2323   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x3000 |     |  type = 0x2000 |     |  type = 0x2000 |
 *   2   |                |     |                |     |                |
 *       |  next = 0      |     |  next = Entry2 |     |  next = Entry2 |
 *       | value = 0      |     | value = 3434   |     | value = 3434   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x1000 |     |  type = 0x4000 |     |  type = 0x4000 |
 *   3   |                |     |                |     |                |
 *       |  next = Entry3 |     |  next = Entry3 |     |  next = Entry1 |
 *       | value = 4545   |     | value = 2323   |     | value = 4545   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x4000 |     |  type = 0x3000 |     |  type = 0x3000 |
 *   4   |                |     |                |     |                |
 *       |  next = Entry4 |     |  next = 0      |     |  next = 0      |
 *       | value = 2323   |     | value = 0      |     | value = 0      |
 *       |________________|     |________________|     |________________|
 *
 * Once the cast array has been initialized, we use cast[0]->next field to link
 * it into the list of cast arrays for the type.
 *              ____    ____    ____
 * type->cast->|next|->|next|->|next|->0
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *
 * Subsequent cast resolution works as follows:
 *
 * 1. Check whether the type matches the first entry in the current cast array.
 * 2. If not, then do a binary search over the (0:cast->value] interval using
 *    either type address or the hash value of the type name.
 * 3. If not found, then move over to the next cast array (cast[0]->next).
 *
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif

#ifndef SWIG_INIT_CLIENT_DATA_TYPE
#define SWIG_INIT_CLIENT_DATA_TYPE void *
#endif

/*
 * Sort function that puts cast entries with nonzero 'next' at the front
 * of the array while ordering them by addresses of their 'type' structs.
 */
SWIGINTERN int SWIG_CastCmpStruct(const void *pa, const void *pb) {
  swig_cast_info *pca = (swig_cast_info *)pa;
  swig_cast_info *pcb = (swig_cast_info *)pb;
  if (pca->type < pcb->type)
  return (pca->next || pcb->next == 0) ? -1 : 1;
  if (pca->type > pcb->type)
  return (pcb->next || pca->next == 0) ? 1 : -1;
  return 0;
}

/*
 * Shell-sort 'next' and 'value' field pairs to order them by 'value'.
 */
SWIGINTERN void SWIG_CastHashSort(swig_cast_info *cast, int size) {
  const int hmax = size/9;
  int h, i;
  for(h = 1; h <= hmax; h = 3*h+1);
  for(; h > 0; h /= 3)
  {
    for(i = h; i < size; ++i)
    {
      swig_cast_info *p = cast[i].next;
      unsigned int hash = cast[i].value;
      int j = i;
      while(j >= h && hash < cast[j-h].value)
      {
        cast[j].next = cast[j-h].next;
        cast[j].value = cast[j-h].value;
        j -= h;
      }
      cast[j].next = p;
      cast[j].value = hash;
    }
  }
}

SWIGRUNTIME void
SWIG_InitializeModule(SWIG_INIT_CLIENT_DATA_TYPE clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *target_type;
    swig_cast_info *cast, *first;
    int num_mapped = 0;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = first = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      target_type = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        target_type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
        if (target_type) {
          /* Target type already defined in another module */
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: found cast %s\n", target_type->name);
#endif
          if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
            printf("SWIG_InitializeModule: skip old type %s\n", target_type->name);
#endif
            cast->type = target_type;
            target_type = 0;
          } else {
            /* Check if this cast is already in the list */
            swig_cast_info *ocast = SWIG_TypeCheck(target_type->name, type);
#ifdef SWIGRUNTIME_DEBUG
            if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", target_type->name);
#endif
            if (!ocast) target_type = 0;
          }
        }
      }
      
      if (!target_type) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        /* Set inclusion mark for sorting */
        cast->next = cast;
        num_mapped++;
        
        if (type == cast->type) {
#ifdef SWIGRUNTIME_DEBUG
          printf("%s : self cast at pos [%li]\n", type->name, cast - first);
#endif
          if (cast - first) {
            /* Move cast to itself to the first entry in the array */
            
            swig_cast_info tmp = *cast;
            *cast = *first;
            *first = tmp;
          }
          first++;
          
        } else {
          cast->value = SWIG_Hash(cast->type->name, (unsigned int)strlen(cast->type->name));
        }
      }
      cast++;
    }
    
    if (num_mapped) {
      if (cast - first) {
        swig_cast_info *tmp;
        
        /* Sort casts by type address for binary search in SWIG_TypeCheckStruct */
        qsort(first, cast - first, sizeof(swig_cast_info), SWIG_CastCmpStruct);
        
        /* Remap back links for added entries */
        cast = swig_module.cast_initial[i] + num_mapped;
        for (tmp = first; tmp < cast; tmp++) {
          tmp->next = tmp;
        }
      }
      
      /* Set the value field of the first entry to the index of the last added entry */
      cast = swig_module.cast_initial[i];
      cast->value = num_mapped - 1;
      
      num_mapped -= (int)(first - cast);
      if (num_mapped > 1) {
        /* Sort <'next','value'> pairs by 'value' for binary search in SWIG_TypeCheck */
        
        SWIG_CastHashSort(first, num_mapped);
      }
      
      first = type->cast;
      if (first) {
        /* Link the current set into the list of cast arrays */
        cast->next = first->next;
        first->next = cast;
      } else {
        cast->next = 0;
        type->cast = cast;
      }
    }
    
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      swig_cast_info *head, *cast;
      head = swig_module.types[i]->cast;
      while (head) {
        for (cast = head; (cast - head) <= head->value; cast++) {
          if (!cast->converter) {
            if (cast->type && !cast->type->clientdata)
            SWIG_TypeClientData(cast->type, swig_module.types[i]->clientdata);
          }
        }
        head = head->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif


#ifdef __cplusplus
extern "C" {
#endif
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  
  SWIGINTERN void
  SWIG_Tcl_InstallConstants(Tcl_Interp *interp, swig_const_info constants[]) {
    size_t i;
    Tcl_Obj *obj;
    
    if (!swigconstTableinit) {
      Tcl_InitHashTable(&swigconstTable, TCL_STRING_KEYS);
      swigconstTableinit = 1;
    }
    for (i = 0; constants[i].type; i++) {
      switch(constants[i].type) {
      case SWIG_TCL_POINTER:
        obj = SWIG_NewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_TCL_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        SWIG_Tcl_SetConstantObj(interp, constants[i].name, obj);
      }
    }
  }
  
  /* Create fast method lookup tables */
  
  SWIGINTERN void
  SWIG_Tcl_InstallMethodLookupTables(void) {
    size_t i;
    
    for (i = 0; i < swig_module.size; ++i) {
      swig_type_info *type = swig_module.type_initial[i];
      if (type->clientdata) {
        swig_class* klass = (swig_class*) type->clientdata;
        swig_method* meth;
        Tcl_InitHashTable(&(klass->hashtable), TCL_STRING_KEYS);
        for (meth = klass->methods; meth && meth->name; ++meth) {
          int newEntry;
          Tcl_HashEntry* hashentry = Tcl_CreateHashEntry(&(klass->hashtable), meth->name, &newEntry);
          Tcl_SetHashValue(hashentry, (ClientData)meth->method);
        }
      }
    }
  }
  
#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

SWIGEXPORT int SWIG_init(Tcl_Interp *interp) {
  size_t i;
  if (interp == 0) return TCL_ERROR;
#ifdef USE_TCL_STUBS
  if (Tcl_InitStubs(interp, SWIG_TCL_STUBS_VERSION, 0) == NULL) {
    return TCL_ERROR;
  }
#endif  
#ifdef USE_TK_STUBS
  /* (char*) cast is required to avoid compiler warning/error. */
  if (Tk_InitStubs(interp, (char*)SWIG_TCL_STUBS_VERSION, 0) == NULL) {
    return TCL_ERROR;
  }
#endif
  
  Tcl_PkgProvide(interp, (char*)SWIG_name, (char*)SWIG_version);
  
#ifdef SWIG_namespace
  Tcl_Eval(interp, "namespace eval " SWIG_namespace " { }");
#endif
  
  SWIG_InitializeModule((void *) interp);
  SWIG_PropagateClientData();
  
  for (i = 0; swig_commands[i].name; i++) {
    Tcl_CreateObjCommand(interp, (char *) swig_commands[i].name, (swig_wrapper_func) swig_commands[i].wrapper,
      swig_commands[i].clientdata, NULL);
  }
  for (i = 0; swig_variables[i].name; i++) {
    Tcl_SetVar(interp, (char *) swig_variables[i].name, (char *) "", TCL_GLOBAL_ONLY);
    Tcl_TraceVar(interp, (char *) swig_variables[i].name, TCL_TRACE_READS | TCL_GLOBAL_ONLY, 
      (Tcl_VarTraceProc *) swig_variables[i].get, (ClientData) swig_variables[i].addr);
    Tcl_TraceVar(interp, (char *) swig_variables[i].name, TCL_TRACE_WRITES | TCL_GLOBAL_ONLY, 
      (Tcl_VarTraceProc *) swig_variables[i].set, (ClientData) swig_variables[i].addr);
  }
  
  SWIG_Tcl_InstallConstants(interp, swig_constants);
  SWIG_Tcl_InstallMethodLookupTables();
  
  
  SWIG_Tcl_SetConstantObj(interp, "MAX_MODELS_PER_BACKEND", SWIG_From_int((int)(1000)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_NONE", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DUMMY", SWIG_From_int((int)((1000*(0)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_NETRIGCTL", SWIG_From_int((int)((1000*(0)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ARMSTRONG", SWIG_From_int((int)((1000*(0)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FLRIG", SWIG_From_int((int)((1000*(0)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TRXMANAGER_RIG", SWIG_From_int((int)((1000*(0)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DUMMY_NOVFO", SWIG_From_int((int)((1000*(0)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TCI1X", SWIG_From_int((int)((1000*(0)+(7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ACLOG", SWIG_From_int((int)((1000*(0)+(8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SDRSHARP", SWIG_From_int((int)((1000*(0)+(9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_QUISK", SWIG_From_int((int)((1000*(0)+(10)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_GQRX", SWIG_From_int((int)((1000*(0)+(11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT847", SWIG_From_int((int)((1000*(1)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT1000", SWIG_From_int((int)((1000*(1)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT1000D", SWIG_From_int((int)((1000*(1)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT1000MPMKV", SWIG_From_int((int)((1000*(1)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT747", SWIG_From_int((int)((1000*(1)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT757", SWIG_From_int((int)((1000*(1)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT757GXII", SWIG_From_int((int)((1000*(1)+(7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT575", SWIG_From_int((int)((1000*(1)+(8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT767", SWIG_From_int((int)((1000*(1)+(9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT736R", SWIG_From_int((int)((1000*(1)+(10)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT840", SWIG_From_int((int)((1000*(1)+(11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT820", SWIG_From_int((int)((1000*(1)+(12)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT900", SWIG_From_int((int)((1000*(1)+(13)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT920", SWIG_From_int((int)((1000*(1)+(14)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT890", SWIG_From_int((int)((1000*(1)+(15)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT990", SWIG_From_int((int)((1000*(1)+(16)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FRG100", SWIG_From_int((int)((1000*(1)+(17)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FRG9600", SWIG_From_int((int)((1000*(1)+(18)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FRG8800", SWIG_From_int((int)((1000*(1)+(19)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT817", SWIG_From_int((int)((1000*(1)+(20)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT100", SWIG_From_int((int)((1000*(1)+(21)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT857", SWIG_From_int((int)((1000*(1)+(22)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT897", SWIG_From_int((int)((1000*(1)+(23)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT1000MP", SWIG_From_int((int)((1000*(1)+(24)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT1000MPMKVFLD", SWIG_From_int((int)((1000*(1)+(25)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_VR5000", SWIG_From_int((int)((1000*(1)+(26)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT450", SWIG_From_int((int)((1000*(1)+(27)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT950", SWIG_From_int((int)((1000*(1)+(28)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT2000", SWIG_From_int((int)((1000*(1)+(29)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT9000", SWIG_From_int((int)((1000*(1)+(30)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT980", SWIG_From_int((int)((1000*(1)+(31)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FTDX5000", SWIG_From_int((int)((1000*(1)+(32)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_VX1700", SWIG_From_int((int)((1000*(1)+(33)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FTDX1200", SWIG_From_int((int)((1000*(1)+(34)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT991", SWIG_From_int((int)((1000*(1)+(35)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT891", SWIG_From_int((int)((1000*(1)+(36)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FTDX3000", SWIG_From_int((int)((1000*(1)+(37)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT847UNI", SWIG_From_int((int)((1000*(1)+(38)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT600", SWIG_From_int((int)((1000*(1)+(39)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FTDX101D", SWIG_From_int((int)((1000*(1)+(40)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT818", SWIG_From_int((int)((1000*(1)+(41)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FTDX10", SWIG_From_int((int)((1000*(1)+(42)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT897D", SWIG_From_int((int)((1000*(1)+(43)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FTDX101MP", SWIG_From_int((int)((1000*(1)+(44)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_MCHFQRP", SWIG_From_int((int)((1000*(1)+(45)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT450D", SWIG_From_int((int)((1000*(1)+(46)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT650", SWIG_From_int((int)((1000*(1)+(47)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT990UNI", SWIG_From_int((int)((1000*(1)+(48)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT710", SWIG_From_int((int)((1000*(1)+(49)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FT9000OLD", SWIG_From_int((int)((1000*(1)+(50)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FTX1", SWIG_From_int((int)((1000*(1)+(51)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS50", SWIG_From_int((int)((1000*(2)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS440", SWIG_From_int((int)((1000*(2)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS450S", SWIG_From_int((int)((1000*(2)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS570D", SWIG_From_int((int)((1000*(2)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS690S", SWIG_From_int((int)((1000*(2)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS711", SWIG_From_int((int)((1000*(2)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS790", SWIG_From_int((int)((1000*(2)+(7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS811", SWIG_From_int((int)((1000*(2)+(8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS850", SWIG_From_int((int)((1000*(2)+(9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS870S", SWIG_From_int((int)((1000*(2)+(10)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS940", SWIG_From_int((int)((1000*(2)+(11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS950S", SWIG_From_int((int)((1000*(2)+(12)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS950SDX", SWIG_From_int((int)((1000*(2)+(13)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS2000", SWIG_From_int((int)((1000*(2)+(14)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_R5000", SWIG_From_int((int)((1000*(2)+(15)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS570S", SWIG_From_int((int)((1000*(2)+(16)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_THD7A", SWIG_From_int((int)((1000*(2)+(17)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_THD7AG", SWIG_From_int((int)((1000*(2)+(18)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_THF6A", SWIG_From_int((int)((1000*(2)+(19)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_THF7E", SWIG_From_int((int)((1000*(2)+(20)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_K2", SWIG_From_int((int)((1000*(2)+(21)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS930", SWIG_From_int((int)((1000*(2)+(22)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_THG71", SWIG_From_int((int)((1000*(2)+(23)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS680S", SWIG_From_int((int)((1000*(2)+(24)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS140S", SWIG_From_int((int)((1000*(2)+(25)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TMD700", SWIG_From_int((int)((1000*(2)+(26)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TMV7", SWIG_From_int((int)((1000*(2)+(27)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS480", SWIG_From_int((int)((1000*(2)+(28)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_K3", SWIG_From_int((int)((1000*(2)+(29)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TRC80", SWIG_From_int((int)((1000*(2)+(30)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS590S", SWIG_From_int((int)((1000*(2)+(31)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TRANSFOX", SWIG_From_int((int)((1000*(2)+(32)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_THD72A", SWIG_From_int((int)((1000*(2)+(33)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TMD710", SWIG_From_int((int)((1000*(2)+(34)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TMV71", SWIG_From_int((int)((1000*(2)+(35)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_F6K", SWIG_From_int((int)((1000*(2)+(36)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS590SG", SWIG_From_int((int)((1000*(2)+(37)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_XG3", SWIG_From_int((int)((1000*(2)+(38)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS990S", SWIG_From_int((int)((1000*(2)+(39)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_HPSDR", SWIG_From_int((int)((1000*(2)+(40)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TS890S", SWIG_From_int((int)((1000*(2)+(41)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_THD74", SWIG_From_int((int)((1000*(2)+(42)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_K3S", SWIG_From_int((int)((1000*(2)+(43)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_KX2", SWIG_From_int((int)((1000*(2)+(44)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_KX3", SWIG_From_int((int)((1000*(2)+(45)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PT8000A", SWIG_From_int((int)((1000*(2)+(46)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_K4", SWIG_From_int((int)((1000*(2)+(47)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_POWERSDR", SWIG_From_int((int)((1000*(2)+(48)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_MALACHITE", SWIG_From_int((int)((1000*(2)+(49)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_LAB599_TX500", SWIG_From_int((int)((1000*(2)+(50)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SDRUNO", SWIG_From_int((int)((1000*(2)+(51)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_QRPLABS", SWIG_From_int((int)((1000*(2)+(52)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FX4", SWIG_From_int((int)((1000*(2)+(53)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_THETIS", SWIG_From_int((int)((1000*(2)+(54)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TRUSDX", SWIG_From_int((int)((1000*(2)+(55)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SDRCONSOLE", SWIG_From_int((int)((1000*(2)+(56)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_QRPLABS_QMX", SWIG_From_int((int)((1000*(2)+(57)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC1271", SWIG_From_int((int)((1000*(3)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC1275", SWIG_From_int((int)((1000*(3)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC271", SWIG_From_int((int)((1000*(3)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC275", SWIG_From_int((int)((1000*(3)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC375", SWIG_From_int((int)((1000*(3)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC471", SWIG_From_int((int)((1000*(3)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC475", SWIG_From_int((int)((1000*(3)+(7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC575", SWIG_From_int((int)((1000*(3)+(8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC706", SWIG_From_int((int)((1000*(3)+(9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC706MKII", SWIG_From_int((int)((1000*(3)+(10)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC706MKIIG", SWIG_From_int((int)((1000*(3)+(11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC707", SWIG_From_int((int)((1000*(3)+(12)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC718", SWIG_From_int((int)((1000*(3)+(13)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC725", SWIG_From_int((int)((1000*(3)+(14)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC726", SWIG_From_int((int)((1000*(3)+(15)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC728", SWIG_From_int((int)((1000*(3)+(16)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC729", SWIG_From_int((int)((1000*(3)+(17)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC731", SWIG_From_int((int)((1000*(3)+(18)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC735", SWIG_From_int((int)((1000*(3)+(19)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC736", SWIG_From_int((int)((1000*(3)+(20)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC737", SWIG_From_int((int)((1000*(3)+(21)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC738", SWIG_From_int((int)((1000*(3)+(22)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC746", SWIG_From_int((int)((1000*(3)+(23)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC751", SWIG_From_int((int)((1000*(3)+(24)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC751A", SWIG_From_int((int)((1000*(3)+(25)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC756", SWIG_From_int((int)((1000*(3)+(26)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC756PRO", SWIG_From_int((int)((1000*(3)+(27)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC761", SWIG_From_int((int)((1000*(3)+(28)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC765", SWIG_From_int((int)((1000*(3)+(29)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC775", SWIG_From_int((int)((1000*(3)+(30)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC781", SWIG_From_int((int)((1000*(3)+(31)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC820", SWIG_From_int((int)((1000*(3)+(32)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC821H", SWIG_From_int((int)((1000*(3)+(34)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC970", SWIG_From_int((int)((1000*(3)+(35)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR10", SWIG_From_int((int)((1000*(3)+(36)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR71", SWIG_From_int((int)((1000*(3)+(37)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR72", SWIG_From_int((int)((1000*(3)+(38)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR75", SWIG_From_int((int)((1000*(3)+(39)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR7000", SWIG_From_int((int)((1000*(3)+(40)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR7100", SWIG_From_int((int)((1000*(3)+(41)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR8500", SWIG_From_int((int)((1000*(3)+(42)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR9000", SWIG_From_int((int)((1000*(3)+(43)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC910", SWIG_From_int((int)((1000*(3)+(44)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC78", SWIG_From_int((int)((1000*(3)+(45)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC746PRO", SWIG_From_int((int)((1000*(3)+(46)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC756PROII", SWIG_From_int((int)((1000*(3)+(47)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICID1", SWIG_From_int((int)((1000*(3)+(54)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC703", SWIG_From_int((int)((1000*(3)+(55)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC7800", SWIG_From_int((int)((1000*(3)+(56)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC756PROIII", SWIG_From_int((int)((1000*(3)+(57)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR20", SWIG_From_int((int)((1000*(3)+(58)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC7000", SWIG_From_int((int)((1000*(3)+(60)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC7200", SWIG_From_int((int)((1000*(3)+(61)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC7700", SWIG_From_int((int)((1000*(3)+(62)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC7600", SWIG_From_int((int)((1000*(3)+(63)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC92D", SWIG_From_int((int)((1000*(3)+(65)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR9500", SWIG_From_int((int)((1000*(3)+(66)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC7410", SWIG_From_int((int)((1000*(3)+(67)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC9100", SWIG_From_int((int)((1000*(3)+(68)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICRX7", SWIG_From_int((int)((1000*(3)+(69)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC7100", SWIG_From_int((int)((1000*(3)+(70)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ID5100", SWIG_From_int((int)((1000*(3)+(71)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC2730", SWIG_From_int((int)((1000*(3)+(72)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC7300", SWIG_From_int((int)((1000*(3)+(73)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PERSEUS", SWIG_From_int((int)((1000*(3)+(74)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC785x", SWIG_From_int((int)((1000*(3)+(75)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_X108G", SWIG_From_int((int)((1000*(3)+(76)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR6", SWIG_From_int((int)((1000*(3)+(77)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC7610", SWIG_From_int((int)((1000*(3)+(78)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR8600", SWIG_From_int((int)((1000*(3)+(79)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICR30", SWIG_From_int((int)((1000*(3)+(80)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC9700", SWIG_From_int((int)((1000*(3)+(81)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ID4100", SWIG_From_int((int)((1000*(3)+(82)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ID31", SWIG_From_int((int)((1000*(3)+(83)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ID51", SWIG_From_int((int)((1000*(3)+(84)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC705", SWIG_From_int((int)((1000*(3)+(85)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ICF8101", SWIG_From_int((int)((1000*(3)+(86)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_X6100", SWIG_From_int((int)((1000*(3)+(87)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_G90", SWIG_From_int((int)((1000*(3)+(88)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_X5105", SWIG_From_int((int)((1000*(3)+(89)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC905", SWIG_From_int((int)((1000*(3)+(90)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_X6200", SWIG_From_int((int)((1000*(3)+(91)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC7760", SWIG_From_int((int)((1000*(3)+(92)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ID52PLUS", SWIG_From_int((int)((1000*(3)+(93)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC7300MK2", SWIG_From_int((int)((1000*(3)+(94)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_MINISCOUT", SWIG_From_int((int)((1000*(3)+(48)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_XPLORER", SWIG_From_int((int)((1000*(3)+(49)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_OS535", SWIG_From_int((int)((1000*(3)+(52)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_OS456", SWIG_From_int((int)((1000*(3)+(53)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_OMNIVI", SWIG_From_int((int)((1000*(3)+(50)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_OMNIVIP", SWIG_From_int((int)((1000*(3)+(51)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PARAGON2", SWIG_From_int((int)((1000*(3)+(59)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DELTAII", SWIG_From_int((int)((1000*(3)+(64)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PCR1000", SWIG_From_int((int)((1000*(4)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PCR100", SWIG_From_int((int)((1000*(4)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PCR1500", SWIG_From_int((int)((1000*(4)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PCR2500", SWIG_From_int((int)((1000*(4)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR8200", SWIG_From_int((int)((1000*(5)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR8000", SWIG_From_int((int)((1000*(5)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR7030", SWIG_From_int((int)((1000*(5)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR5000", SWIG_From_int((int)((1000*(5)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR3030", SWIG_From_int((int)((1000*(5)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR3000A", SWIG_From_int((int)((1000*(5)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR3000", SWIG_From_int((int)((1000*(5)+(7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR2700", SWIG_From_int((int)((1000*(5)+(8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR2500", SWIG_From_int((int)((1000*(5)+(9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR16", SWIG_From_int((int)((1000*(5)+(10)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SDU5500", SWIG_From_int((int)((1000*(5)+(11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SDU5000", SWIG_From_int((int)((1000*(5)+(12)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR8600", SWIG_From_int((int)((1000*(5)+(13)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR5000A", SWIG_From_int((int)((1000*(5)+(14)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_AR7030P", SWIG_From_int((int)((1000*(5)+(15)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SR2200", SWIG_From_int((int)((1000*(5)+(16)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_JST145", SWIG_From_int((int)((1000*(6)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_JST245", SWIG_From_int((int)((1000*(6)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_CMH530", SWIG_From_int((int)((1000*(6)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_NRD345", SWIG_From_int((int)((1000*(6)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_NRD525", SWIG_From_int((int)((1000*(6)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_NRD535", SWIG_From_int((int)((1000*(6)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_NRD545", SWIG_From_int((int)((1000*(6)+(7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RS64", SWIG_From_int((int)((1000*(7)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RS2005", SWIG_From_int((int)((1000*(7)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RS2006", SWIG_From_int((int)((1000*(7)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RS2035", SWIG_From_int((int)((1000*(7)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RS2042", SWIG_From_int((int)((1000*(7)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RS2041", SWIG_From_int((int)((1000*(7)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BC780", SWIG_From_int((int)((1000*(8)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BC245", SWIG_From_int((int)((1000*(8)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BC895", SWIG_From_int((int)((1000*(8)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PRO2052", SWIG_From_int((int)((1000*(8)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BC235", SWIG_From_int((int)((1000*(8)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BC250", SWIG_From_int((int)((1000*(8)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BC785", SWIG_From_int((int)((1000*(8)+(7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BC786", SWIG_From_int((int)((1000*(8)+(8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BCT8", SWIG_From_int((int)((1000*(8)+(9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BCD396T", SWIG_From_int((int)((1000*(8)+(10)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BCD996T", SWIG_From_int((int)((1000*(8)+(11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BC898", SWIG_From_int((int)((1000*(8)+(12)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DKR8", SWIG_From_int((int)((1000*(9)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DKR8A", SWIG_From_int((int)((1000*(9)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DKR8B", SWIG_From_int((int)((1000*(9)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_HF150", SWIG_From_int((int)((1000*(10)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_HF225", SWIG_From_int((int)((1000*(10)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_HF250", SWIG_From_int((int)((1000*(10)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_HF235", SWIG_From_int((int)((1000*(10)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RA3790", SWIG_From_int((int)((1000*(11)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RA3720", SWIG_From_int((int)((1000*(11)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RA6790", SWIG_From_int((int)((1000*(11)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RA3710", SWIG_From_int((int)((1000*(11)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RA3702", SWIG_From_int((int)((1000*(11)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_HF1000", SWIG_From_int((int)((1000*(12)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_HF1000A", SWIG_From_int((int)((1000*(12)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_WJ8711", SWIG_From_int((int)((1000*(12)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_WJ8888", SWIG_From_int((int)((1000*(12)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ESM500", SWIG_From_int((int)((1000*(13)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_EK890", SWIG_From_int((int)((1000*(13)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_EK891", SWIG_From_int((int)((1000*(13)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_EK895", SWIG_From_int((int)((1000*(13)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_EK070", SWIG_From_int((int)((1000*(13)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TRP7000", SWIG_From_int((int)((1000*(14)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TRP8000", SWIG_From_int((int)((1000*(14)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TRP9000", SWIG_From_int((int)((1000*(14)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TRP8255", SWIG_From_int((int)((1000*(14)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_WR1000", SWIG_From_int((int)((1000*(15)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_WR1500", SWIG_From_int((int)((1000*(15)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_WR1550", SWIG_From_int((int)((1000*(15)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_WR3100", SWIG_From_int((int)((1000*(15)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_WR3150", SWIG_From_int((int)((1000*(15)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_WR3500", SWIG_From_int((int)((1000*(15)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_WR3700", SWIG_From_int((int)((1000*(15)+(7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_G303", SWIG_From_int((int)((1000*(15)+(8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_G313", SWIG_From_int((int)((1000*(15)+(9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_G305", SWIG_From_int((int)((1000*(15)+(10)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_G315", SWIG_From_int((int)((1000*(15)+(11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TT550", SWIG_From_int((int)((1000*(16)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TT538", SWIG_From_int((int)((1000*(16)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RX320", SWIG_From_int((int)((1000*(16)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RX340", SWIG_From_int((int)((1000*(16)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RX350", SWIG_From_int((int)((1000*(16)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TT526", SWIG_From_int((int)((1000*(16)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TT516", SWIG_From_int((int)((1000*(16)+(7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TT565", SWIG_From_int((int)((1000*(16)+(8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TT585", SWIG_From_int((int)((1000*(16)+(9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TT588", SWIG_From_int((int)((1000*(16)+(11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RX331", SWIG_From_int((int)((1000*(16)+(12)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_TT599", SWIG_From_int((int)((1000*(16)+(13)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DX77", SWIG_From_int((int)((1000*(17)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DXSR8", SWIG_From_int((int)((1000*(17)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_505DSP", SWIG_From_int((int)((1000*(18)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_GNURADIO", SWIG_From_int((int)((1000*(20)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_MC4020", SWIG_From_int((int)((1000*(20)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_GRAUDIO", SWIG_From_int((int)((1000*(20)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_GRAUDIOIQ", SWIG_From_int((int)((1000*(20)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_USRP_G", SWIG_From_int((int)((1000*(20)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_MICROTUNE_4937", SWIG_From_int((int)((1000*(21)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_MICROTUNE_4702", SWIG_From_int((int)((1000*(21)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_MICROTUNE_4707", SWIG_From_int((int)((1000*(21)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DSP10", SWIG_From_int((int)((1000*(22)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SDR1000", SWIG_From_int((int)((1000*(23)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SDR1000RFE", SWIG_From_int((int)((1000*(23)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DTTSP", SWIG_From_int((int)((1000*(23)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DTTSP_UDP", SWIG_From_int((int)((1000*(23)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SMARTSDR_A", SWIG_From_int((int)((1000*(23)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SMARTSDR_B", SWIG_From_int((int)((1000*(23)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SMARTSDR_C", SWIG_From_int((int)((1000*(23)+(7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SMARTSDR_D", SWIG_From_int((int)((1000*(23)+(8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SMARTSDR_E", SWIG_From_int((int)((1000*(23)+(9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SMARTSDR_F", SWIG_From_int((int)((1000*(23)+(10)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SMARTSDR_G", SWIG_From_int((int)((1000*(23)+(11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SMARTSDR_H", SWIG_From_int((int)((1000*(23)+(12)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_EKD500", SWIG_From_int((int)((1000*(24)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ELEKTOR304", SWIG_From_int((int)((1000*(25)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DRT1", SWIG_From_int((int)((1000*(25)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DWT", SWIG_From_int((int)((1000*(25)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_USRP0", SWIG_From_int((int)((1000*(25)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_USRP", SWIG_From_int((int)((1000*(25)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DDS60", SWIG_From_int((int)((1000*(25)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ELEKTOR507", SWIG_From_int((int)((1000*(25)+(7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_MINIVNA", SWIG_From_int((int)((1000*(25)+(8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SI570AVRUSB", SWIG_From_int((int)((1000*(25)+(9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PMSDR", SWIG_From_int((int)((1000*(25)+(10)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SI570PICUSB", SWIG_From_int((int)((1000*(25)+(11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FIFISDR", SWIG_From_int((int)((1000*(25)+(12)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FUNCUBEDONGLE", SWIG_From_int((int)((1000*(25)+(13)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_HIQSDR", SWIG_From_int((int)((1000*(25)+(14)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FASDR", SWIG_From_int((int)((1000*(25)+(15)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SI570PEABERRY1", SWIG_From_int((int)((1000*(25)+(16)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SI570PEABERRY2", SWIG_From_int((int)((1000*(25)+(17)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_FUNCUBEDONGLEPLUS", SWIG_From_int((int)((1000*(25)+(18)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_RSHFIQ", SWIG_From_int((int)((1000*(25)+(19)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_V4L", SWIG_From_int((int)((1000*(26)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_V4L2", SWIG_From_int((int)((1000*(26)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ESMC", SWIG_From_int((int)((1000*(27)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_EB200", SWIG_From_int((int)((1000*(27)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_XK2100", SWIG_From_int((int)((1000*(27)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_EK89X", SWIG_From_int((int)((1000*(27)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_XK852", SWIG_From_int((int)((1000*(27)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PRM8060", SWIG_From_int((int)((1000*(28)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PRM8070", SWIG_From_int((int)((1000*(28)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ADT_200A", SWIG_From_int((int)((1000*(29)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC_M700PRO", SWIG_From_int((int)((1000*(30)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC_M802", SWIG_From_int((int)((1000*(30)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC_M710", SWIG_From_int((int)((1000*(30)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_IC_M803", SWIG_From_int((int)((1000*(30)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DORJI_DRA818V", SWIG_From_int((int)((1000*(31)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_DORJI_DRA818U", SWIG_From_int((int)((1000*(31)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BARRETT_2050", SWIG_From_int((int)((1000*(32)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BARRETT_950", SWIG_From_int((int)((1000*(32)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BARRETT_4050", SWIG_From_int((int)((1000*(32)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_BARRETT_4100", SWIG_From_int((int)((1000*(32)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ELAD_FDM_DUO", SWIG_From_int((int)((1000*(33)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_CODAN_ENVOY", SWIG_From_int((int)((1000*(34)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_CODAN_NGT", SWIG_From_int((int)((1000*(34)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_CODAN_2110", SWIG_From_int((int)((1000*(34)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_GS100", SWIG_From_int((int)((1000*(35)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_MDS4710", SWIG_From_int((int)((1000*(36)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_MDS9710", SWIG_From_int((int)((1000*(36)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_ATD578UVIII", SWIG_From_int((int)((1000*(37)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_MICOM2", SWIG_From_int((int)((1000*(38)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_CTX10", SWIG_From_int((int)((1000*(39)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PMR171", SWIG_From_int((int)((1000*(40)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_Q900", SWIG_From_int((int)((1000*(40)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_SIMPLECAT", SWIG_From_int((int)((1000*(41)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODEL_PRC138", SWIG_From_int((int)((1000*(42)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "BUILTINFUNC", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_SECRET_LENGTH", SWIG_From_int((int)(32)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_COOKIE_SIZE", SWIG_From_int((int)(37)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OK", SWIG_From_int((int)(RIG_OK)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_EINVAL", SWIG_From_int((int)(RIG_EINVAL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ECONF", SWIG_From_int((int)(RIG_ECONF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ENOMEM", SWIG_From_int((int)(RIG_ENOMEM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ENIMPL", SWIG_From_int((int)(RIG_ENIMPL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ETIMEOUT", SWIG_From_int((int)(RIG_ETIMEOUT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_EIO", SWIG_From_int((int)(RIG_EIO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_EINTERNAL", SWIG_From_int((int)(RIG_EINTERNAL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_EPROTO", SWIG_From_int((int)(RIG_EPROTO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ERJCTED", SWIG_From_int((int)(RIG_ERJCTED)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ETRUNC", SWIG_From_int((int)(RIG_ETRUNC)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ENAVAIL", SWIG_From_int((int)(RIG_ENAVAIL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ENTARGET", SWIG_From_int((int)(RIG_ENTARGET)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BUSERROR", SWIG_From_int((int)(RIG_BUSERROR)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BUSBUSY", SWIG_From_int((int)(RIG_BUSBUSY)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_EARG", SWIG_From_int((int)(RIG_EARG)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_EVFO", SWIG_From_int((int)(RIG_EVFO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_EDOM", SWIG_From_int((int)(RIG_EDOM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_EDEPRECATED", SWIG_From_int((int)(RIG_EDEPRECATED)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ESECURITY", SWIG_From_int((int)(RIG_ESECURITY)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_EPOWER", SWIG_From_int((int)(RIG_EPOWER)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ELIMIT", SWIG_From_int((int)(RIG_ELIMIT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_EACCESS", SWIG_From_int((int)(RIG_EACCESS)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_EEND", SWIG_From_int((int)(RIG_EEND)));
  SWIG_Tcl_SetConstantObj(interp, "NETRIGCTL_RET", SWIG_FromCharPtr("RPRT "));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DEBUG_NONE", SWIG_From_int((int)(RIG_DEBUG_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DEBUG_BUG", SWIG_From_int((int)(RIG_DEBUG_BUG)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DEBUG_ERR", SWIG_From_int((int)(RIG_DEBUG_ERR)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DEBUG_WARN", SWIG_From_int((int)(RIG_DEBUG_WARN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DEBUG_VERBOSE", SWIG_From_int((int)(RIG_DEBUG_VERBOSE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DEBUG_TRACE", SWIG_From_int((int)(RIG_DEBUG_TRACE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DEBUG_CACHE", SWIG_From_int((int)(RIG_DEBUG_CACHE)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_RIGNAMSIZ", SWIG_From_int((int)(30)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_RIGVERSIZ", SWIG_From_int((int)(8)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_FILPATHLEN", SWIG_From_int((int)(512)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_FRQRANGESIZ", SWIG_From_int((int)(30)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAXCHANDESC", SWIG_From_int((int)(30)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_TSLSTSIZ", SWIG_From_int((int)(20)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_FLTLSTSIZ", SWIG_From_int((int)(60)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAXDBLSTSIZ", SWIG_From_int((int)(8)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_CHANLSTSIZ", SWIG_From_int((int)(16)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_AGC_LEVELS", SWIG_From_int((int)(8)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_SPECTRUM_SCOPES", SWIG_From_int((int)(4)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_SPECTRUM_MODES", SWIG_From_int((int)(5)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_SPECTRUM_AVG_MODES", SWIG_From_int((int)(12)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_SPECTRUM_SPANS", SWIG_From_int((int)(20)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_SPECTRUM_DATA", SWIG_From_int((int)(2048)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_CAL_LENGTH", SWIG_From_int((int)(32)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_MODES", SWIG_From_int((int)(63)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_VFOS", SWIG_From_int((int)(31)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_ROTORS", SWIG_From_int((int)(31)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_VFO_OPS", SWIG_From_int((int)(31)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_RSCANS", SWIG_From_int((int)(31)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_MAX_SNAPSHOT_PACKET_SIZE", SWIG_From_int((int)(16384)));
  SWIG_Tcl_SetConstantObj(interp, "CTCSS_LIST_SIZE", SWIG_From_int((int)(60)));
  SWIG_Tcl_SetConstantObj(interp, "DCS_LIST_SIZE", SWIG_From_int((int)(128)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_NONE", SWIG_From_int((int)(RIG_PORT_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_SERIAL", SWIG_From_int((int)(RIG_PORT_SERIAL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_NETWORK", SWIG_From_int((int)(RIG_PORT_NETWORK)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_DEVICE", SWIG_From_int((int)(RIG_PORT_DEVICE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_PACKET", SWIG_From_int((int)(RIG_PORT_PACKET)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_DTMF", SWIG_From_int((int)(RIG_PORT_DTMF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_ULTRA", SWIG_From_int((int)(RIG_PORT_ULTRA)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_RPC", SWIG_From_int((int)(RIG_PORT_RPC)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_PARALLEL", SWIG_From_int((int)(RIG_PORT_PARALLEL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_USB", SWIG_From_int((int)(RIG_PORT_USB)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_UDP_NETWORK", SWIG_From_int((int)(RIG_PORT_UDP_NETWORK)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_CM108", SWIG_From_int((int)(RIG_PORT_CM108)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_GPIO", SWIG_From_int((int)(RIG_PORT_GPIO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PORT_GPION", SWIG_From_int((int)(RIG_PORT_GPION)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARITY_NONE", SWIG_From_int((int)(RIG_PARITY_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARITY_ODD", SWIG_From_int((int)(RIG_PARITY_ODD)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARITY_EVEN", SWIG_From_int((int)(RIG_PARITY_EVEN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARITY_MARK", SWIG_From_int((int)(RIG_PARITY_MARK)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARITY_SPACE", SWIG_From_int((int)(RIG_PARITY_SPACE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_HANDSHAKE_NONE", SWIG_From_int((int)(RIG_HANDSHAKE_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_HANDSHAKE_XONXOFF", SWIG_From_int((int)(RIG_HANDSHAKE_XONXOFF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_HANDSHAKE_HARDWARE", SWIG_From_int((int)(RIG_HANDSHAKE_HARDWARE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SIGNAL_UNSET", SWIG_From_int((int)(RIG_SIGNAL_UNSET)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SIGNAL_ON", SWIG_From_int((int)(RIG_SIGNAL_ON)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SIGNAL_OFF", SWIG_From_int((int)(RIG_SIGNAL_OFF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FLAG_RECEIVER", SWIG_From_int((int)(RIG_FLAG_RECEIVER)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FLAG_TRANSMITTER", SWIG_From_int((int)(RIG_FLAG_TRANSMITTER)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FLAG_SCANNER", SWIG_From_int((int)(RIG_FLAG_SCANNER)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FLAG_MOBILE", SWIG_From_int((int)(RIG_FLAG_MOBILE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FLAG_HANDHELD", SWIG_From_int((int)(RIG_FLAG_HANDHELD)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FLAG_COMPUTER", SWIG_From_int((int)(RIG_FLAG_COMPUTER)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FLAG_TRUNKING", SWIG_From_int((int)(RIG_FLAG_TRUNKING)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FLAG_APRS", SWIG_From_int((int)(RIG_FLAG_APRS)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FLAG_TNC", SWIG_From_int((int)(RIG_FLAG_TNC)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FLAG_DXCLUSTER", SWIG_From_int((int)(RIG_FLAG_DXCLUSTER)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FLAG_TUNER", SWIG_From_int((int)(RIG_FLAG_TUNER)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_AGC_OFF", SWIG_From_int((int)(RIG_AGC_OFF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_AGC_SUPERFAST", SWIG_From_int((int)(RIG_AGC_SUPERFAST)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_AGC_FAST", SWIG_From_int((int)(RIG_AGC_FAST)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_AGC_SLOW", SWIG_From_int((int)(RIG_AGC_SLOW)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_AGC_USER", SWIG_From_int((int)(RIG_AGC_USER)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_AGC_MEDIUM", SWIG_From_int((int)(RIG_AGC_MEDIUM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_AGC_AUTO", SWIG_From_int((int)(RIG_AGC_AUTO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_AGC_LONG", SWIG_From_int((int)(RIG_AGC_LONG)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_AGC_ON", SWIG_From_int((int)(RIG_AGC_ON)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_AGC_NONE", SWIG_From_int((int)(RIG_AGC_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TYPE_OTHER", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_STATUS_ALPHA", SWIG_From_int((int)(RIG_STATUS_ALPHA)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_STATUS_UNTESTED", SWIG_From_int((int)(RIG_STATUS_UNTESTED)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_STATUS_BETA", SWIG_From_int((int)(RIG_STATUS_BETA)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_STATUS_STABLE", SWIG_From_int((int)(RIG_STATUS_STABLE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_STATUS_BUGGY", SWIG_From_int((int)(RIG_STATUS_BUGGY)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_RPT_SHIFT_NONE", SWIG_From_int((int)(RIG_RPT_SHIFT_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_RPT_SHIFT_MINUS", SWIG_From_int((int)(RIG_RPT_SHIFT_MINUS)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_RPT_SHIFT_PLUS", SWIG_From_int((int)(RIG_RPT_SHIFT_PLUS)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SPLIT_OFF", SWIG_From_int((int)(RIG_SPLIT_OFF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SPLIT_ON", SWIG_From_int((int)(RIG_SPLIT_ON)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_NONE", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_A", SWIG_From_unsigned_SS_int((unsigned int)((1u << (0)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_B", SWIG_From_unsigned_SS_int((unsigned int)((1u << (1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_C", SWIG_From_unsigned_SS_int((unsigned int)((1u << (2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_SUB_A", SWIG_From_unsigned_SS_int((unsigned int)((1u << (21)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_SUB_B", SWIG_From_unsigned_SS_int((unsigned int)((1u << (22)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_SUB_C", SWIG_From_unsigned_SS_int((unsigned int)((1u << (3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_MAIN_A", SWIG_From_unsigned_SS_int((unsigned int)((1u << (23)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_MAIN_B", SWIG_From_unsigned_SS_int((unsigned int)((1u << (24)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_MAIN_C", SWIG_From_unsigned_SS_int((unsigned int)((1u << (4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_OTHER", SWIG_From_unsigned_SS_int((unsigned int)((1u << (5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_SUB", SWIG_From_unsigned_SS_int((unsigned int)((1u << (25)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_MAIN", SWIG_From_unsigned_SS_int((unsigned int)((1u << (26)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_VFO", SWIG_From_unsigned_SS_int((unsigned int)((1u << (27)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_MEM", SWIG_From_unsigned_SS_int((unsigned int)((1u << (28)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_CURR", SWIG_From_unsigned_SS_int((unsigned int)((1u << (29)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_TX_FLAG", SWIG_From_unsigned_SS_int((unsigned int)((1u << (30)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_ALL", SWIG_From_unsigned_SS_int((unsigned int)((1u << (31)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_TX", SWIG_From_unsigned_SS_int((unsigned int)((((1u << (29)))|(1u << (30))))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_VFO_RX", SWIG_From_unsigned_SS_int((unsigned int)((1u << (29)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_NONE", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_FREQ", SWIG_From_int((int)((1 << 0))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_MODE", SWIG_From_int((int)((1 << 1))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_PURE", SWIG_From_int((int)((1 << 2))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_TONE", SWIG_From_int((int)((1 << 3))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_FUNC", SWIG_From_int((int)((1 << 4))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_LEVEL", SWIG_From_int((int)((1 << 5))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_RITXIT", SWIG_From_int((int)((1 << 6))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_PTT", SWIG_From_int((int)((1 << 7))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_MEM", SWIG_From_int((int)((1 << 8))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_BANK", SWIG_From_int((int)((1 << 9))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_ANT", SWIG_From_int((int)((1 << 10))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_ROOFING", SWIG_From_int((int)((1 << 11))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_SPECTRUM", SWIG_From_int((int)((1 << 12))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_BAND", SWIG_From_int((int)((1 << 13))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_COMMON", SWIG_From_int((int)(((1 << 6)|(1 << 7)|(1 << 8)|(1 << 9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TARGETABLE_ALL", SWIG_From_int((int)(0x7fffffff)));
  SWIG_Tcl_SetConstantObj(interp, "VFO_DUAL", SWIG_From_unsigned_SS_int((unsigned int)(((1u << (23))|(1u << (24))|(1u << (21))|(1u << (22))))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DCD_OFF", SWIG_From_int((int)(RIG_DCD_OFF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DCD_ON", SWIG_From_int((int)(RIG_DCD_ON)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DCD_NONE", SWIG_From_int((int)(RIG_DCD_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DCD_RIG", SWIG_From_int((int)(RIG_DCD_RIG)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DCD_SERIAL_DSR", SWIG_From_int((int)(RIG_DCD_SERIAL_DSR)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DCD_SERIAL_CTS", SWIG_From_int((int)(RIG_DCD_SERIAL_CTS)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DCD_SERIAL_CAR", SWIG_From_int((int)(RIG_DCD_SERIAL_CAR)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DCD_PARALLEL", SWIG_From_int((int)(RIG_DCD_PARALLEL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DCD_CM108", SWIG_From_int((int)(RIG_DCD_CM108)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DCD_GPIO", SWIG_From_int((int)(RIG_DCD_GPIO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_DCD_GPION", SWIG_From_int((int)(RIG_DCD_GPION)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_OFF", SWIG_From_int((int)(RIG_PTT_OFF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_ON", SWIG_From_int((int)(RIG_PTT_ON)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_ON_MIC", SWIG_From_int((int)(RIG_PTT_ON_MIC)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_ON_DATA", SWIG_From_int((int)(RIG_PTT_ON_DATA)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_NONE", SWIG_From_int((int)(RIG_PTT_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_RIG", SWIG_From_int((int)(RIG_PTT_RIG)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_SERIAL_DTR", SWIG_From_int((int)(RIG_PTT_SERIAL_DTR)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_SERIAL_RTS", SWIG_From_int((int)(RIG_PTT_SERIAL_RTS)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_PARALLEL", SWIG_From_int((int)(RIG_PTT_PARALLEL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_RIG_MICDATA", SWIG_From_int((int)(RIG_PTT_RIG_MICDATA)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_CM108", SWIG_From_int((int)(RIG_PTT_CM108)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_GPIO", SWIG_From_int((int)(RIG_PTT_GPIO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTT_GPION", SWIG_From_int((int)(RIG_PTT_GPION)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_POWER_OFF", SWIG_From_int((int)(RIG_POWER_OFF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_POWER_ON", SWIG_From_int((int)(RIG_POWER_ON)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_POWER_STANDBY", SWIG_From_int((int)(RIG_POWER_STANDBY)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_POWER_OPERATE", SWIG_From_int((int)(RIG_POWER_OPERATE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_POWER_UNKNOWN", SWIG_From_int((int)(RIG_POWER_UNKNOWN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_RESET_NONE", SWIG_From_int((int)(RIG_RESET_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_RESET_SOFT", SWIG_From_int((int)(RIG_RESET_SOFT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_RESET_VFO", SWIG_From_int((int)(RIG_RESET_VFO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_RESET_MCALL", SWIG_From_int((int)(RIG_RESET_MCALL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_RESET_MASTER", SWIG_From_int((int)(RIG_RESET_MASTER)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CLIENT_UNKNOWN", SWIG_From_int((int)(RIG_CLIENT_UNKNOWN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CLIENT_WSJTX", SWIG_From_int((int)(RIG_CLIENT_WSJTX)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CLIENT_GPREDICT", SWIG_From_int((int)(RIG_CLIENT_GPREDICT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_NONE", SWIG_From_int((int)(RIG_OP_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_CPY", SWIG_From_int((int)(RIG_OP_CPY)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_XCHG", SWIG_From_int((int)(RIG_OP_XCHG)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_FROM_VFO", SWIG_From_int((int)(RIG_OP_FROM_VFO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_TO_VFO", SWIG_From_int((int)(RIG_OP_TO_VFO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_MCL", SWIG_From_int((int)(RIG_OP_MCL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_UP", SWIG_From_int((int)(RIG_OP_UP)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_DOWN", SWIG_From_int((int)(RIG_OP_DOWN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_BAND_UP", SWIG_From_int((int)(RIG_OP_BAND_UP)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_BAND_DOWN", SWIG_From_int((int)(RIG_OP_BAND_DOWN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_LEFT", SWIG_From_int((int)(RIG_OP_LEFT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_RIGHT", SWIG_From_int((int)(RIG_OP_RIGHT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_TUNE", SWIG_From_int((int)(RIG_OP_TUNE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_OP_TOGGLE", SWIG_From_int((int)(RIG_OP_TOGGLE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_160M", SWIG_From_int((int)(RIG_BAND_160M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_80M", SWIG_From_int((int)(RIG_BAND_80M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_60M", SWIG_From_int((int)(RIG_BAND_60M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_40M", SWIG_From_int((int)(RIG_BAND_40M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_30M", SWIG_From_int((int)(RIG_BAND_30M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_20M", SWIG_From_int((int)(RIG_BAND_20M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_17M", SWIG_From_int((int)(RIG_BAND_17M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_15M", SWIG_From_int((int)(RIG_BAND_15M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_12M", SWIG_From_int((int)(RIG_BAND_12M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_10M", SWIG_From_int((int)(RIG_BAND_10M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_6M", SWIG_From_int((int)(RIG_BAND_6M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_GEN", SWIG_From_int((int)(RIG_BAND_GEN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_MW", SWIG_From_int((int)(RIG_BAND_MW)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_UNUSED", SWIG_From_int((int)(RIG_BAND_UNUSED)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_AIR", SWIG_From_int((int)(RIG_BAND_AIR)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_144MHZ", SWIG_From_int((int)(RIG_BAND_144MHZ)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BAND_430MHZ", SWIG_From_int((int)(RIG_BAND_430MHZ)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_UNUSED", SWIG_From_int((int)(RIG_BANDSELECT_UNUSED)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_2200M", SWIG_From_int((int)(RIG_BANDSELECT_2200M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_600M", SWIG_From_int((int)(RIG_BANDSELECT_600M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_160M", SWIG_From_int((int)(RIG_BANDSELECT_160M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_80M", SWIG_From_int((int)(RIG_BANDSELECT_80M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_60M", SWIG_From_int((int)(RIG_BANDSELECT_60M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_40M", SWIG_From_int((int)(RIG_BANDSELECT_40M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_30M", SWIG_From_int((int)(RIG_BANDSELECT_30M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_20M", SWIG_From_int((int)(RIG_BANDSELECT_20M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_17M", SWIG_From_int((int)(RIG_BANDSELECT_17M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_15M", SWIG_From_int((int)(RIG_BANDSELECT_15M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_12M", SWIG_From_int((int)(RIG_BANDSELECT_12M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_10M", SWIG_From_int((int)(RIG_BANDSELECT_10M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_6M", SWIG_From_int((int)(RIG_BANDSELECT_6M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_WFM", SWIG_From_int((int)(RIG_BANDSELECT_WFM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_GEN", SWIG_From_int((int)(RIG_BANDSELECT_GEN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_MW", SWIG_From_int((int)(RIG_BANDSELECT_MW)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_AIR", SWIG_From_int((int)(RIG_BANDSELECT_AIR)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_4M", SWIG_From_int((int)(RIG_BANDSELECT_4M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_2M", SWIG_From_int((int)(RIG_BANDSELECT_2M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_1_25M", SWIG_From_int((int)(RIG_BANDSELECT_1_25M)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_70CM", SWIG_From_int((int)(RIG_BANDSELECT_70CM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_33CM", SWIG_From_int((int)(RIG_BANDSELECT_33CM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_23CM", SWIG_From_int((int)(RIG_BANDSELECT_23CM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_13CM", SWIG_From_int((int)(RIG_BANDSELECT_13CM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_9CM", SWIG_From_int((int)(RIG_BANDSELECT_9CM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_5CM", SWIG_From_int((int)(RIG_BANDSELECT_5CM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_BANDSELECT_3CM", SWIG_From_int((int)(RIG_BANDSELECT_3CM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SCAN_NONE", SWIG_From_int((int)(RIG_SCAN_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SCAN_MEM", SWIG_From_int((int)(RIG_SCAN_MEM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SCAN_SLCT", SWIG_From_int((int)(RIG_SCAN_SLCT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SCAN_PRIO", SWIG_From_int((int)(RIG_SCAN_PRIO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SCAN_PROG", SWIG_From_int((int)(RIG_SCAN_PROG)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SCAN_DELTA", SWIG_From_int((int)(RIG_SCAN_DELTA)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SCAN_VFO", SWIG_From_int((int)(RIG_SCAN_VFO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SCAN_PLT", SWIG_From_int((int)(RIG_SCAN_PLT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SCAN_STOP", SWIG_From_int((int)(RIG_SCAN_STOP)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CONF_END", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CONF_STRING", SWIG_From_int((int)(RIG_CONF_STRING)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CONF_COMBO", SWIG_From_int((int)(RIG_CONF_COMBO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CONF_NUMERIC", SWIG_From_int((int)(RIG_CONF_NUMERIC)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CONF_CHECKBUTTON", SWIG_From_int((int)(RIG_CONF_CHECKBUTTON)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CONF_BUTTON", SWIG_From_int((int)(RIG_CONF_BUTTON)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CONF_BINARY", SWIG_From_int((int)(RIG_CONF_BINARY)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CONF_INT", SWIG_From_int((int)(RIG_CONF_INT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ANN_NONE", SWIG_From_int((int)(RIG_ANN_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ANN_OFF", SWIG_From_int((int)(RIG_ANN_OFF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ANN_FREQ", SWIG_From_int((int)(RIG_ANN_FREQ)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ANN_RXMODE", SWIG_From_int((int)(RIG_ANN_RXMODE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ANN_CW", SWIG_From_int((int)(RIG_ANN_CW)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ANN_ENG", SWIG_From_int((int)(RIG_ANN_ENG)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ANN_JAP", SWIG_From_int((int)(RIG_ANN_JAP)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ANT_NONE", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_ANT_MAX", SWIG_From_int((int)(32)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_AGC_LAST", SWIG_From_int((int)(99999)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_METER_NONE", SWIG_From_int((int)(RIG_METER_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_METER_SWR", SWIG_From_int((int)(RIG_METER_SWR)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_METER_COMP", SWIG_From_int((int)(RIG_METER_COMP)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_METER_ALC", SWIG_From_int((int)(RIG_METER_ALC)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_METER_IC", SWIG_From_int((int)(RIG_METER_IC)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_METER_DB", SWIG_From_int((int)(RIG_METER_DB)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_METER_PO", SWIG_From_int((int)(RIG_METER_PO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_METER_VDD", SWIG_From_int((int)(RIG_METER_VDD)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_METER_TEMP", SWIG_From_int((int)(RIG_METER_TEMP)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_NONE", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_PREAMP", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (0)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_ATT", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_VOXDELAY", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_AF", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_RF", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_SQL", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_IF", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_APF", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_NR", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_PBT_IN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_PBT_OUT", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (10)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_CWPITCH", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_RFPOWER", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (12)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_MICGAIN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (13)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_KEYSPD", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (14)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_NOTCHF", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (15)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_COMP", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (16)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_AGC", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (17)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_BKINDL", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (18)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_BALANCE", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (19)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_METER", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (20)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_VOXGAIN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (21)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_ANTIVOX", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (22)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_SLOPE_LOW", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (23)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_SLOPE_HIGH", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (24)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_BKIN_DLYMS", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (25)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_RAWSTR", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (26)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_SWR", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (28)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_ALC", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (29)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_STRENGTH", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (30)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_RFPOWER_METER", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (32)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_COMP_METER", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (33)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_VD_METER", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (34)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_ID_METER", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (35)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_NOTCHF_RAW", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (36)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_MONITOR_GAIN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (37)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_NB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (38)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_RFPOWER_METER_WATTS", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (39)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_SPECTRUM_MODE", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (40)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_SPECTRUM_SPAN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (41)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_SPECTRUM_EDGE_LOW", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (42)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_SPECTRUM_EDGE_HIGH", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (43)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_SPECTRUM_SPEED", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (44)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_SPECTRUM_REF", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (45)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_SPECTRUM_AVG", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (46)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_SPECTRUM_ATT", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (47)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_TEMP_METER", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (48)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_BAND_SELECT", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (49)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_USB_AF", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (50)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_USB_AF_INPUT", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (51)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_AGC_TIME", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (52)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_53", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (53)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_54", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (54)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_55", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (55)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_56", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (56)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_57", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (57)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_58", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (58)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_59", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (59)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_60", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (60)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_61", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (61)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_62", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (62)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_63", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (63)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_FLOAT_LIST", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(((1ull << (3))|(1ull << (4))|(1ull << (5))|(1ull << (7))|(1ull << (8))|(1ull << (9))|(1ull << (10))|(1ull << (12))|(1ull << (13))|(1ull << (16))|(1ull << (19))|(1ull << (28))|(1ull << (29))|(1ull << (21))|(1ull << (22))|(1ull << (32))|(1ull << (39))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (36))|(1ull << (37))|(1ull << (38))|(1ull << (45))|(1ull << (48))|(1ull << (50))|(1ull << (51))|(1ull << (52))))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_LEVEL_READONLY_LIST", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(((1ull << (28))|(1ull << (29))|(1ull << (30))|(1ull << (26))|(1ull << (33))|(1ull << (34))|(1ull << (35))|(1ull << (48))|(1ull << (32))|(1ull << (39))))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_NONE", SWIG_From_int((int)(RIG_PARM_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_ANN", SWIG_From_int((int)(RIG_PARM_ANN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_APO", SWIG_From_int((int)(RIG_PARM_APO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_BACKLIGHT", SWIG_From_int((int)(RIG_PARM_BACKLIGHT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_BEEP", SWIG_From_int((int)(RIG_PARM_BEEP)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_TIME", SWIG_From_int((int)(RIG_PARM_TIME)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_BAT", SWIG_From_int((int)(RIG_PARM_BAT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_KEYLIGHT", SWIG_From_int((int)(RIG_PARM_KEYLIGHT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_SCREENSAVER", SWIG_From_int((int)(RIG_PARM_SCREENSAVER)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_AFIF", SWIG_From_int((int)(RIG_PARM_AFIF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_BANDSELECT", SWIG_From_int((int)(RIG_PARM_BANDSELECT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_KEYERTYPE", SWIG_From_int((int)(RIG_PARM_KEYERTYPE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_AFIF_LAN", SWIG_From_int((int)(RIG_PARM_AFIF_LAN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_AFIF_WLAN", SWIG_From_int((int)(RIG_PARM_AFIF_WLAN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PARM_AFIF_ACC", SWIG_From_int((int)(RIG_PARM_AFIF_ACC)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_KEYERTYPE_STRAIGHT", SWIG_From_int((int)(RIG_KEYERTYPE_STRAIGHT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_KEYERTYPE_BUG", SWIG_From_int((int)(RIG_KEYERTYPE_BUG)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_KEYERTYPE_PADDLE", SWIG_From_int((int)(RIG_KEYERTYPE_PADDLE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_KEYERTYPE_UNKNOWN", SWIG_From_int((int)(RIG_KEYERTYPE_UNKNOWN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_COOKIE_GET", SWIG_From_int((int)(RIG_COOKIE_GET)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_COOKIE_RELEASE", SWIG_From_int((int)(RIG_COOKIE_RELEASE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_COOKIE_RENEW", SWIG_From_int((int)(RIG_COOKIE_RENEW)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MULTICAST_POLL", SWIG_From_int((int)(RIG_MULTICAST_POLL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MULTICAST_TRANSCEIVE", SWIG_From_int((int)(RIG_MULTICAST_TRANSCEIVE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MULTICAST_SPECTRUM", SWIG_From_int((int)(RIG_MULTICAST_SPECTRUM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SETTING_MAX", SWIG_From_int((int)(64)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TRN_OFF", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TRN_RIG", SWIG_From_int((int)(1)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TRN_POLL", SWIG_From_int((int)(2)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_NONE", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_FAGC", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (0)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_NB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_COMP", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_VOX", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_TONE", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_TSQL", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_SBKIN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_FBKIN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_ANF", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_NR", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_AIP", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (10)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_APF", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_MON", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (12)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_MN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (13)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_RF", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (14)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_ARO", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (15)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_LOCK", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (16)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_MUTE", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (17)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_VSC", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (18)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_REV", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (19)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_SQL", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (20)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_ABM", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (21)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BC", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (22)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_MBC", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (23)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_RIT", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (24)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_AFC", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (25)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_SATMODE", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (26)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_SCOPE", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (27)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_RESUME", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (28)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_TBURST", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (29)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_TUNER", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (30)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_XIT", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (31)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_NB2", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (32)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_CSQL", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (33)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_AFLT", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (34)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_ANL", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (35)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BC2", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (36)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_DUAL_WATCH", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (37)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_DIVERSITY", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (38)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_DSQL", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (39)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_SCEN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (40)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_SLICE", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (41)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_TRANSCEIVE", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (42)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_SPECTRUM", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (43)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_SPECTRUM_HOLD", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (44)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_SEND_MORSE", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (45)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_SEND_VOICE_MEM", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (46)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_OVF_STATUS", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (47)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_SYNC", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (48)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT49", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (49)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT50", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (50)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT51", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (51)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT52", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (52)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT53", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (53)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT54", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (54)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT55", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (55)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT56", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (56)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT57", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (57)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT58", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (58)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT59", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (59)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT60", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (60)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT61", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (61)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT62", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (62)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNC_BIT63", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (63)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_NONE", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_AM", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (0)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_CW", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (1)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_USB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (2)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_LSB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (3)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_RTTY", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (4)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_FM", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (5)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_WFM", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (6)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_CWR", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (7)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_RTTYR", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (8)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_AMS", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (9)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_PKTLSB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (10)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_PKTUSB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (11)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_PKTFM", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (12)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_ECSSUSB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (13)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_ECSSLSB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (14)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_FAX", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (15)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_SAM", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (16)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_SAL", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (17)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_SAH", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (18)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_DSB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (19)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_FMN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (21)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_PKTAM", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (22)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_P25", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (23)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_DSTAR", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (24)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_DPMR", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (25)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_NXDNVN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (26)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_NXDN_N", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (27)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_DCR", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (28)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_AMN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (29)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_PSK", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (30)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_PSKR", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (31)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_DD", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (32)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_C4FM", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (33)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_PKTFMN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (34)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_SPEC", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (35)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_CWN", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (36)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_IQ", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (37)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_ISBUSB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (38)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_ISBLSB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (39)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_USBD1", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (40)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_USBD2", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (41)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_USBD3", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (42)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_LSBD1", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (43)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_LSBD2", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (44)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_LSBD3", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (45)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_WFMS", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (46)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT47", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (47)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT48", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (48)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT49", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (49)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT50", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (50)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT51", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (51)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT52", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (52)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT53", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (53)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT54", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (54)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT55", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (55)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT56", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (56)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT57", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (57)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT58", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (58)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT59", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (59)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT60", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (60)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT61", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (61)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_BIT62", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (62)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_TESTS_MAX", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (63)))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_ALL", SWIG_From_int((int)((0xffffffff))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_SSB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(((1ull << (2))|(1ull << (3))))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_PKTSSB", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(((1ull << (11))|(1ull << (10))))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MODE_ECSS", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(((1ull << (13))|(1ull << (14))))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_TS_ANY", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FLT_ANY", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "DEBUGMSGSAVE_SIZE", SWIG_From_int((int)(24000)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CHFLAG_NONE", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CHFLAG_SKIP", SWIG_From_int((int)((1 << 0))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CHFLAG_DATA", SWIG_From_int((int)((1 << 1))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CHFLAG_PSKIP", SWIG_From_int((int)((1 << 2))));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MTYPE_NONE", SWIG_From_int((int)(RIG_MTYPE_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MTYPE_MEM", SWIG_From_int((int)(RIG_MTYPE_MEM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MTYPE_EDGE", SWIG_From_int((int)(RIG_MTYPE_EDGE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MTYPE_CALL", SWIG_From_int((int)(RIG_MTYPE_CALL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MTYPE_MEMOPAD", SWIG_From_int((int)(RIG_MTYPE_MEMOPAD)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MTYPE_SAT", SWIG_From_int((int)(RIG_MTYPE_SAT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MTYPE_BAND", SWIG_From_int((int)(RIG_MTYPE_BAND)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MTYPE_PRIO", SWIG_From_int((int)(RIG_MTYPE_PRIO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MTYPE_VOICE", SWIG_From_int((int)(RIG_MTYPE_VOICE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MTYPE_MORSE", SWIG_From_int((int)(RIG_MTYPE_MORSE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MTYPE_SPLIT", SWIG_From_int((int)(RIG_MTYPE_SPLIT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_MEM_CAPS_ALL", SWIG_From_int((int)(-1)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SPECTRUM_MODE_NONE", SWIG_From_int((int)(RIG_SPECTRUM_MODE_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SPECTRUM_MODE_CENTER", SWIG_From_int((int)(RIG_SPECTRUM_MODE_CENTER)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SPECTRUM_MODE_FIXED", SWIG_From_int((int)(RIG_SPECTRUM_MODE_FIXED)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SPECTRUM_MODE_CENTER_SCROLL", SWIG_From_int((int)(RIG_SPECTRUM_MODE_CENTER_SCROLL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_SPECTRUM_MODE_FIXED_SCROLL", SWIG_From_int((int)(RIG_SPECTRUM_MODE_FIXED_SCROLL)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_CHECK_RIG_CAPS", SWIG_FromCharPtr("HAMLIB_CHECK_RIG_CAPS"));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_INIT", SWIG_From_int((int)(RIG_FUNCTION_INIT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_CLEANUP", SWIG_From_int((int)(RIG_FUNCTION_CLEANUP)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_OPEN", SWIG_From_int((int)(RIG_FUNCTION_OPEN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_CLOSE", SWIG_From_int((int)(RIG_FUNCTION_CLOSE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_FREQ", SWIG_From_int((int)(RIG_FUNCTION_SET_FREQ)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_FREQ", SWIG_From_int((int)(RIG_FUNCTION_GET_FREQ)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_MODE", SWIG_From_int((int)(RIG_FUNCTION_SET_MODE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_MODE", SWIG_From_int((int)(RIG_FUNCTION_GET_MODE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_VFO", SWIG_From_int((int)(RIG_FUNCTION_SET_VFO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_VFO", SWIG_From_int((int)(RIG_FUNCTION_GET_VFO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_PTT", SWIG_From_int((int)(RIG_FUNCTION_SET_PTT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_PTT", SWIG_From_int((int)(RIG_FUNCTION_GET_PTT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_DCD", SWIG_From_int((int)(RIG_FUNCTION_GET_DCD)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_RPTR_SHIFT", SWIG_From_int((int)(RIG_FUNCTION_SET_RPTR_SHIFT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_RPTR_SHIFT", SWIG_From_int((int)(RIG_FUNCTION_GET_RPTR_SHIFT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_RPTR_OFFS", SWIG_From_int((int)(RIG_FUNCTION_SET_RPTR_OFFS)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_RPTR_OFFS", SWIG_From_int((int)(RIG_FUNCTION_GET_RPTR_OFFS)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_SPLIT_FREQ", SWIG_From_int((int)(RIG_FUNCTION_SET_SPLIT_FREQ)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_SPLIT_FREQ", SWIG_From_int((int)(RIG_FUNCTION_GET_SPLIT_FREQ)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_SPLIT_MODE", SWIG_From_int((int)(RIG_FUNCTION_SET_SPLIT_MODE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_SPLIT_FREQ_MODE", SWIG_From_int((int)(RIG_FUNCTION_SET_SPLIT_FREQ_MODE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_SPLIT_FREQ_MODE", SWIG_From_int((int)(RIG_FUNCTION_GET_SPLIT_FREQ_MODE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_SPLIT_VFO", SWIG_From_int((int)(RIG_FUNCTION_SET_SPLIT_VFO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_SPLIT_VFO", SWIG_From_int((int)(RIG_FUNCTION_GET_SPLIT_VFO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_RIT", SWIG_From_int((int)(RIG_FUNCTION_SET_RIT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_RIT", SWIG_From_int((int)(RIG_FUNCTION_GET_RIT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_XIT", SWIG_From_int((int)(RIG_FUNCTION_SET_XIT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_XIT", SWIG_From_int((int)(RIG_FUNCTION_GET_XIT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_TS", SWIG_From_int((int)(RIG_FUNCTION_SET_TS)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_TS", SWIG_From_int((int)(RIG_FUNCTION_GET_TS)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_DCS_CODE", SWIG_From_int((int)(RIG_FUNCTION_SET_DCS_CODE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_DCS_CODE", SWIG_From_int((int)(RIG_FUNCTION_GET_DCS_CODE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_TONE", SWIG_From_int((int)(RIG_FUNCTION_SET_TONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_TONE", SWIG_From_int((int)(RIG_FUNCTION_GET_TONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_CTCSS_TONE", SWIG_From_int((int)(RIG_FUNCTION_SET_CTCSS_TONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_CTCSS_TONE", SWIG_From_int((int)(RIG_FUNCTION_GET_CTCSS_TONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_DCS_SQL", SWIG_From_int((int)(RIG_FUNCTION_SET_DCS_SQL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_DCS_SQL", SWIG_From_int((int)(RIG_FUNCTION_GET_DCS_SQL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_TONE_SQL", SWIG_From_int((int)(RIG_FUNCTION_SET_TONE_SQL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_TONE_SQL", SWIG_From_int((int)(RIG_FUNCTION_GET_TONE_SQL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_CTCSS_SQL", SWIG_From_int((int)(RIG_FUNCTION_SET_CTCSS_SQL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_CTCSS_SQL", SWIG_From_int((int)(RIG_FUNCTION_GET_CTCSS_SQL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_POWER2MW", SWIG_From_int((int)(RIG_FUNCTION_POWER2MW)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_MW2POWER", SWIG_From_int((int)(RIG_FUNCTION_MW2POWER)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_POWERSTAT", SWIG_From_int((int)(RIG_FUNCTION_SET_POWERSTAT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_POWERSTAT", SWIG_From_int((int)(RIG_FUNCTION_GET_POWERSTAT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_RESET", SWIG_From_int((int)(RIG_FUNCTION_RESET)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_ANT", SWIG_From_int((int)(RIG_FUNCTION_SET_ANT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_ANT", SWIG_From_int((int)(RIG_FUNCTION_GET_ANT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_LEVEL", SWIG_From_int((int)(RIG_FUNCTION_SET_LEVEL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_LEVEL", SWIG_From_int((int)(RIG_FUNCTION_GET_LEVEL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_FUNC", SWIG_From_int((int)(RIG_FUNCTION_SET_FUNC)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_FUNC", SWIG_From_int((int)(RIG_FUNCTION_GET_FUNC)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_PARM", SWIG_From_int((int)(RIG_FUNCTION_SET_PARM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_PARM", SWIG_From_int((int)(RIG_FUNCTION_GET_PARM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_EXT_LEVEL", SWIG_From_int((int)(RIG_FUNCTION_SET_EXT_LEVEL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_EXT_LEVEL", SWIG_From_int((int)(RIG_FUNCTION_GET_EXT_LEVEL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_EXT_FUNC", SWIG_From_int((int)(RIG_FUNCTION_SET_EXT_FUNC)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_EXT_FUNC", SWIG_From_int((int)(RIG_FUNCTION_GET_EXT_FUNC)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_EXT_PARM", SWIG_From_int((int)(RIG_FUNCTION_SET_EXT_PARM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_EXT_PARM", SWIG_From_int((int)(RIG_FUNCTION_GET_EXT_PARM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_CONF", SWIG_From_int((int)(RIG_FUNCTION_SET_CONF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_CONF", SWIG_From_int((int)(RIG_FUNCTION_GET_CONF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SEND_DTMF", SWIG_From_int((int)(RIG_FUNCTION_SEND_DTMF)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SEND_MORSE", SWIG_From_int((int)(RIG_FUNCTION_SEND_MORSE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_STOP_MORSE", SWIG_From_int((int)(RIG_FUNCTION_STOP_MORSE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_WAIT_MORSE", SWIG_From_int((int)(RIG_FUNCTION_WAIT_MORSE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SEND_VOICE_MEM", SWIG_From_int((int)(RIG_FUNCTION_SEND_VOICE_MEM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_BANK", SWIG_From_int((int)(RIG_FUNCTION_SET_BANK)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_MEM", SWIG_From_int((int)(RIG_FUNCTION_SET_MEM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_MEM", SWIG_From_int((int)(RIG_FUNCTION_GET_MEM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_VFO_OP", SWIG_From_int((int)(RIG_FUNCTION_VFO_OP)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SCAN", SWIG_From_int((int)(RIG_FUNCTION_SCAN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_TRN", SWIG_From_int((int)(RIG_FUNCTION_SET_TRN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_TRN", SWIG_From_int((int)(RIG_FUNCTION_GET_TRN)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_DECODE_EVENT", SWIG_From_int((int)(RIG_FUNCTION_DECODE_EVENT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_CHANNEL", SWIG_From_int((int)(RIG_FUNCTION_SET_CHANNEL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_CHANNEL", SWIG_From_int((int)(RIG_FUNCTION_GET_CHANNEL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_INFO", SWIG_From_int((int)(RIG_FUNCTION_GET_INFO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_CHAN_ALL_CB", SWIG_From_int((int)(RIG_FUNCTION_SET_CHAN_ALL_CB)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_CHAN_ALL_CB", SWIG_From_int((int)(RIG_FUNCTION_GET_CHAN_ALL_CB)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_MEM_ALL_CB", SWIG_From_int((int)(RIG_FUNCTION_SET_MEM_ALL_CB)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_MEM_ALL_CB", SWIG_From_int((int)(RIG_FUNCTION_GET_MEM_ALL_CB)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_SET_VFO_OPT", SWIG_From_int((int)(RIG_FUNCTION_SET_VFO_OPT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_READ_FRAME_DIRECT", SWIG_From_int((int)(RIG_FUNCTION_READ_FRAME_DIRECT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_IS_ASYNC_FRAME", SWIG_From_int((int)(RIG_FUNCTION_IS_ASYNC_FRAME)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_PROCESS_ASYNC_FRAME", SWIG_From_int((int)(RIG_FUNCTION_PROCESS_ASYNC_FRAME)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_GET_CONF2", SWIG_From_int((int)(RIG_FUNCTION_GET_CONF2)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_FUNCTION_STOP_VOICE_MEM", SWIG_From_int((int)(RIG_FUNCTION_STOP_VOICE_MEM)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CAPS_TARGETABLE_VFO", SWIG_From_int((int)(RIG_CAPS_TARGETABLE_VFO)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CAPS_RIG_MODEL", SWIG_From_int((int)(RIG_CAPS_RIG_MODEL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CAPS_PORT_TYPE", SWIG_From_int((int)(RIG_CAPS_PORT_TYPE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CAPS_PTT_TYPE", SWIG_From_int((int)(RIG_CAPS_PTT_TYPE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CAPS_HAS_GET_LEVEL", SWIG_From_int((int)(RIG_CAPS_HAS_GET_LEVEL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CAPS_HAS_SET_LEVEL", SWIG_From_int((int)(RIG_CAPS_HAS_SET_LEVEL)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CAPS_VERSION_CPTR", SWIG_From_int((int)(RIG_CAPS_VERSION_CPTR)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CAPS_MFG_NAME_CPTR", SWIG_From_int((int)(RIG_CAPS_MFG_NAME_CPTR)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CAPS_MODEL_NAME_CPTR", SWIG_From_int((int)(RIG_CAPS_MODEL_NAME_CPTR)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_CAPS_STATUS_CPTR", SWIG_From_int((int)(RIG_CAPS_STATUS_CPTR)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTRX_NONE", SWIG_From_int((int)(RIG_PTRX_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTRX_RIGPORT", SWIG_From_int((int)(RIG_PTRX_RIGPORT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTRX_PTTPORT", SWIG_From_int((int)(RIG_PTRX_PTTPORT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTRX_DCDPORT", SWIG_From_int((int)(RIG_PTRX_DCDPORT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTRX_CACHE", SWIG_From_int((int)(RIG_PTRX_CACHE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTRX_AMPPORT", SWIG_From_int((int)(RIG_PTRX_AMPPORT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTRX_ROTPORT", SWIG_From_int((int)(RIG_PTRX_ROTPORT)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTRX_ROTPORT2", SWIG_From_int((int)(RIG_PTRX_ROTPORT2)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTRX_STATE", SWIG_From_int((int)(RIG_PTRX_STATE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTRX_AMPSTATE", SWIG_From_int((int)(RIG_PTRX_AMPSTATE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTRX_ROTSTATE", SWIG_From_int((int)(RIG_PTRX_ROTSTATE)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_PTRX_MAXIMUM", SWIG_From_int((int)(RIG_PTRX_MAXIMUM)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_ELAPSED_GET", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_ELAPSED_SET", SWIG_From_int((int)(1)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_ELAPSED_INVALIDATE", SWIG_From_int((int)(2)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_CACHE_ALWAYS", SWIG_From_int((int)((-1))));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_CACHE_ALL", SWIG_From_int((int)(HAMLIB_CACHE_ALL)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_CACHE_VFO", SWIG_From_int((int)(HAMLIB_CACHE_VFO)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_CACHE_FREQ", SWIG_From_int((int)(HAMLIB_CACHE_FREQ)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_CACHE_MODE", SWIG_From_int((int)(HAMLIB_CACHE_MODE)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_CACHE_PTT", SWIG_From_int((int)(HAMLIB_CACHE_PTT)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_CACHE_SPLIT", SWIG_From_int((int)(HAMLIB_CACHE_SPLIT)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_CACHE_WIDTH", SWIG_From_int((int)(HAMLIB_CACHE_WIDTH)));
  SWIG_Tcl_SetConstantObj(interp, "TWIDDLE_OFF", SWIG_From_int((int)(TWIDDLE_OFF)));
  SWIG_Tcl_SetConstantObj(interp, "TWIDDLE_ON", SWIG_From_int((int)(TWIDDLE_ON)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_COMM_STATUS_OK", SWIG_From_int((int)(0x00)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_COMM_STATUS_CONNECTING", SWIG_From_int((int)(0x01)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_COMM_STATUS_DISCONNECTED", SWIG_From_int((int)(0x02)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_COMM_STATUS_TERMINATED", SWIG_From_int((int)(0x03)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_COMM_STATUS_WARNING", SWIG_From_int((int)(0x04)));
  SWIG_Tcl_SetConstantObj(interp, "RIG_COMM_STATUS_ERROR", SWIG_From_int((int)(0x05)));
  SWIG_Tcl_SetConstantObj(interp, "HAMLIB_SETTINGS_FILE", SWIG_FromCharPtr("hamlib_settings"));
  SWIG_Tcl_SetConstantObj(interp, "GPIO1", SWIG_From_int((int)(GPIO1)));
  SWIG_Tcl_SetConstantObj(interp, "GPIO2", SWIG_From_int((int)(GPIO2)));
  SWIG_Tcl_SetConstantObj(interp, "GPIO3", SWIG_From_int((int)(GPIO3)));
  SWIG_Tcl_SetConstantObj(interp, "GPIO4", SWIG_From_int((int)(GPIO4)));
  SWIG_Tcl_SetConstantObj(interp, "NETROTCTL_RET", SWIG_FromCharPtr("RPRT "));
  SWIG_Tcl_SetConstantObj(interp, "ROT_RESET_ALL", SWIG_From_int((int)(1)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_FLAG_AZIMUTH", SWIG_From_int((int)(ROT_FLAG_AZIMUTH)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_FLAG_ELEVATION", SWIG_From_int((int)(ROT_FLAG_ELEVATION)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_TYPE_OTHER", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_UP", SWIG_From_int((int)((1 << 1))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_DOWN", SWIG_From_int((int)((1 << 2))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_LEFT", SWIG_From_int((int)((1 << 3))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_CCW", SWIG_From_int((int)((1 << 3))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_RIGHT", SWIG_From_int((int)((1 << 4))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_CW", SWIG_From_int((int)((1 << 4))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_UP_LEFT", SWIG_From_int((int)((1 << 5))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_UP_CCW", SWIG_From_int((int)((1 << 5))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_UP_RIGHT", SWIG_From_int((int)((1 << 6))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_UP_CW", SWIG_From_int((int)((1 << 6))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_DOWN_LEFT", SWIG_From_int((int)((1 << 7))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_DOWN_CCW", SWIG_From_int((int)((1 << 7))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_DOWN_RIGHT", SWIG_From_int((int)((1 << 8))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MOVE_DOWN_CW", SWIG_From_int((int)((1 << 8))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_NONE", SWIG_From_int((int)(ROT_STATUS_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_BUSY", SWIG_From_int((int)(ROT_STATUS_BUSY)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_MOVING", SWIG_From_int((int)(ROT_STATUS_MOVING)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_MOVING_AZ", SWIG_From_int((int)(ROT_STATUS_MOVING_AZ)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_MOVING_LEFT", SWIG_From_int((int)(ROT_STATUS_MOVING_LEFT)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_MOVING_RIGHT", SWIG_From_int((int)(ROT_STATUS_MOVING_RIGHT)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_MOVING_EL", SWIG_From_int((int)(ROT_STATUS_MOVING_EL)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_MOVING_UP", SWIG_From_int((int)(ROT_STATUS_MOVING_UP)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_MOVING_DOWN", SWIG_From_int((int)(ROT_STATUS_MOVING_DOWN)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_LIMIT_UP", SWIG_From_int((int)(ROT_STATUS_LIMIT_UP)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_LIMIT_DOWN", SWIG_From_int((int)(ROT_STATUS_LIMIT_DOWN)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_LIMIT_LEFT", SWIG_From_int((int)(ROT_STATUS_LIMIT_LEFT)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_LIMIT_RIGHT", SWIG_From_int((int)(ROT_STATUS_LIMIT_RIGHT)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_OVERLAP_UP", SWIG_From_int((int)(ROT_STATUS_OVERLAP_UP)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_OVERLAP_DOWN", SWIG_From_int((int)(ROT_STATUS_OVERLAP_DOWN)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_OVERLAP_LEFT", SWIG_From_int((int)(ROT_STATUS_OVERLAP_LEFT)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_STATUS_OVERLAP_RIGHT", SWIG_From_int((int)(ROT_STATUS_OVERLAP_RIGHT)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_SPEED_NOCHANGE", SWIG_From_int((int)((-1))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_LEVEL_NONE", SWIG_From_int((int)(ROT_LEVEL_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_LEVEL_SPEED", SWIG_From_int((int)(ROT_LEVEL_SPEED)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_LEVEL_63", SWIG_From_int((int)(ROT_LEVEL_63)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_LEVEL_FLOAT_LIST", SWIG_From_int((int)((0))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_LEVEL_READONLY_LIST", SWIG_From_int((int)((0))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_PARM_NONE", SWIG_From_int((int)(ROT_PARM_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_PARM_FLOAT_LIST", SWIG_From_int((int)((0))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_PARM_READONLY_LIST", SWIG_From_int((int)((0))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_FUNC_NONE", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_FUNC_BIT63", SWIG_From_unsigned_SS_long_SS_long((unsigned long long)((1ull << (63)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_NONE", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_DUMMY", SWIG_From_int((int)((100*(0)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_NETROTCTL", SWIG_From_int((int)((100*(0)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_PSTROTATOR", SWIG_From_int((int)((100*(0)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_SATROTCTL", SWIG_From_int((int)((100*(0)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_EASYCOMM1", SWIG_From_int((int)((100*(2)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_EASYCOMM2", SWIG_From_int((int)((100*(2)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_EASYCOMM3", SWIG_From_int((int)((100*(2)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_FODTRACK", SWIG_From_int((int)((100*(3)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_ROTOREZ", SWIG_From_int((int)((100*(4)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_ROTORCARD", SWIG_From_int((int)((100*(4)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_DCU", SWIG_From_int((int)((100*(4)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_ERC", SWIG_From_int((int)((100*(4)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_RT21", SWIG_From_int((int)((100*(4)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_YRC1", SWIG_From_int((int)((100*(4)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_SARTEK1", SWIG_From_int((int)((100*(5)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_GS232A", SWIG_From_int((int)((100*(6)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_GS232_GENERIC", SWIG_From_int((int)((100*(6)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_GS232B", SWIG_From_int((int)((100*(6)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_F1TETRACKER", SWIG_From_int((int)((100*(6)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_GS23", SWIG_From_int((int)((100*(6)+(5)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_GS232", SWIG_From_int((int)((100*(6)+(6)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_LVB", SWIG_From_int((int)((100*(6)+(7)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_ST2", SWIG_From_int((int)((100*(6)+(8)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_GS232A_AZ", SWIG_From_int((int)((100*(6)+(9)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_GS232A_EL", SWIG_From_int((int)((100*(6)+(10)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_GS232B_AZ", SWIG_From_int((int)((100*(6)+(11)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_GS232B_EL", SWIG_From_int((int)((100*(6)+(12)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_GS23_AZ", SWIG_From_int((int)((100*(6)+(13)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_AF6SA_WRC", SWIG_From_int((int)((100*(6)+(14)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_ARRAYSOLUTIONS_SAL_12_20_30", SWIG_From_int((int)((100*(327)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_PCROTOR", SWIG_From_int((int)((100*(7)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_HD1780", SWIG_From_int((int)((100*(8)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_SPID_ROT2PROG", SWIG_From_int((int)((100*(9)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_SPID_ROT1PROG", SWIG_From_int((int)((100*(9)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_SPID_MD01_ROT2PROG", SWIG_From_int((int)((100*(9)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_RC2800", SWIG_From_int((int)((100*(10)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_RC2800_EARLY_AZ", SWIG_From_int((int)((100*(10)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_RC2800_EARLY_AZEL", SWIG_From_int((int)((100*(10)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_RCI_AZEL", SWIG_From_int((int)((100*(11)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_RCI_AZ", SWIG_From_int((int)((100*(11)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_IF100", SWIG_From_int((int)((100*(12)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_TS7400", SWIG_From_int((int)((100*(13)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_NEXSTAR", SWIG_From_int((int)((100*(14)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_ETHER6", SWIG_From_int((int)((100*(15)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_CNCTRK", SWIG_From_int((int)((100*(16)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_PROSISTEL_D_AZ", SWIG_From_int((int)((100*(17)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_PROSISTEL_D_EL", SWIG_From_int((int)((100*(17)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_PROSISTEL_COMBI_TRACK_AZEL", SWIG_From_int((int)((100*(17)+(3)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_PROSISTEL_D_EL_CBOXAZ", SWIG_From_int((int)((100*(17)+(4)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_MEADE", SWIG_From_int((int)((100*(18)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_IOPTRON", SWIG_From_int((int)((100*(19)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_INDI", SWIG_From_int((int)((100*(20)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_SATEL", SWIG_From_int((int)((100*(21)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_RADANT", SWIG_From_int((int)((100*(22)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_ANDROIDSENSOR", SWIG_From_int((int)((100*(23)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_GRBLTRK_SER", SWIG_From_int((int)((100*(24)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_GRBLTRK_NET", SWIG_From_int((int)((100*(24)+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_FLIR", SWIG_From_int((int)((100*(25)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_APEX_SHARED_LOOP", SWIG_From_int((int)((100*(26)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_SAEBRTRACK", SWIG_From_int((int)((100*(27)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "ROT_MODEL_SKYWATCHER", SWIG_From_int((int)((100*(28)+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "AMP_MODEL_NONE", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_MODEL_DUMMY", SWIG_From_int((int)(((0)*100+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "AMP_MODEL_NETAMPCTL", SWIG_From_int((int)(((0)*100+(2)))));
  SWIG_Tcl_SetConstantObj(interp, "AMP_MODEL_ELECRAFT_KPA1500", SWIG_From_int((int)(((2)*100+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "AMP_MODEL_GEMINI_DX1200", SWIG_From_int((int)(((3)*100+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "AMP_MODEL_EXPERT_FA", SWIG_From_int((int)(((4)*100+(1)))));
  SWIG_Tcl_SetConstantObj(interp, "NETAMPCTL_RET", SWIG_FromCharPtr("RPRT "));
  SWIG_Tcl_SetConstantObj(interp, "AMP_RESET_MEM", SWIG_From_int((int)(AMP_RESET_MEM)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_RESET_FAULT", SWIG_From_int((int)(AMP_RESET_FAULT)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_RESET_AMP", SWIG_From_int((int)(AMP_RESET_AMP)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_FLAG_1", SWIG_From_int((int)(AMP_FLAG_1)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_FLAG_2", SWIG_From_int((int)(AMP_FLAG_2)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_TYPE_OTHER", SWIG_From_int((int)(0)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_LEVEL_NONE", SWIG_From_int((int)(AMP_LEVEL_NONE)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_LEVEL_SWR", SWIG_From_int((int)(AMP_LEVEL_SWR)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_LEVEL_NH", SWIG_From_int((int)(AMP_LEVEL_NH)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_LEVEL_PF", SWIG_From_int((int)(AMP_LEVEL_PF)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_LEVEL_PWR_INPUT", SWIG_From_int((int)(AMP_LEVEL_PWR_INPUT)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_LEVEL_PWR_FWD", SWIG_From_int((int)(AMP_LEVEL_PWR_FWD)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_LEVEL_PWR_REFLECTED", SWIG_From_int((int)(AMP_LEVEL_PWR_REFLECTED)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_LEVEL_PWR_PEAK", SWIG_From_int((int)(AMP_LEVEL_PWR_PEAK)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_LEVEL_FAULT", SWIG_From_int((int)(AMP_LEVEL_FAULT)));
  SWIG_Tcl_SetConstantObj(interp, "AMP_LEVEL_PWR", SWIG_From_int((int)(AMP_LEVEL_PWR)));
  return TCL_OK;
}
SWIGEXPORT int Hamlib_SafeInit(Tcl_Interp *interp) {
  return SWIG_init(interp);
}

